/**
 * This file is intended to serve as the
 * translator from OS independent calls to kernel
 * specific calls.
 *
 */
/*
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/device.h>
#include <linux/random.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <stdarg.h>
#include "apfo.h"

#define DEBUG_BUFFER_SIZE 4096
#define DEBUG_LINE_SIZE 200

int pfo_debug_flag;
static char debug_buffer[DEBUG_BUFFER_SIZE] = "\001";
static int debug_tail = 1;

static struct mutex path_mutex[MAX_PATH_NUMBER];
static struct mutex fdev_mutex[FDEV_MUTEX_COUNT];
static int path_mutex_need_init = 1;
static int fdev_mutex_need_init = 1;

int lock_apfo_path(int path_num)
{
	int i;

	if (path_mutex_need_init) {
		path_mutex_need_init = 0;
		for (i = 0; i < MAX_PATH_NUMBER; i++) {
			mutex_init(&path_mutex[i]);
		}
	}
	return mutex_lock_interruptible(&path_mutex[path_num]);
}

void unlock_apfo_path(int path_num)
{
	mutex_unlock(&path_mutex[path_num]);
}

int lock_apfo_device(int fdev_num)
{
	int i;

	if (fdev_mutex_need_init) {
		fdev_mutex_need_init = 0;
		for (i = 0; i < FDEV_MUTEX_COUNT; i++) {
			mutex_init(&fdev_mutex[i]);
		}
	}
	return mutex_lock_interruptible(&fdev_mutex[fdev_num]);
}

void unlock_apfo_device(int fdev_num)
{
	mutex_unlock(&fdev_mutex[fdev_num]);
}

void sleep_half_second(void)
{
	msleep(500);
}

void sleep_six_seconds(void)
{
	msleep(6000);
}

/** Print to messages file, unformatted.
 * Intended for output that an end customer
 * would need to see.
 */
void print_log(char *string)
{
	printk(KERN_WARNING "pfo: %s", string);
}

unsigned char random_byte(void)
{
	unsigned char thebyte[1];
	get_random_bytes(&thebyte[0], 1);
	return thebyte[0];
}

/** Scan the string for a valid path descriptor.
 * If valid then modify the string to make it able
 * to be compared to another valid path string.
 * @return PFO_GOOD if this is now a valid path string.
 */
int check_string_for_valid_path(char *string)
{
	int res, a, b, c, d;

	res = sscanf(string, "%d:%d:%d:%d", &a, &b, &c, &d);
	if (res == 4 &&
	    a >= 0 && b >= 0 && c >= 0 && d >= 0) {
		sprintf(string, "%d:%d:%d:%d", a, b, c, d);
		return PFO_GOOD;
	}
	return PFO_ERROR;
}

void debug(const char *_file, int _line, const char *_func,
				int level, const char *fmt, ...)
{
	char tmp[DEBUG_LINE_SIZE];
	int count = 0;
	int mask = level;
	va_list argptr;

	if (strstr(__FILE__, "lower.c")) {
		mask = level << 4;
	} else if (strstr(__FILE__, "decide.c")) {
		mask = level << 8;
	} else if (strstr(__FILE__, "upper.c")) {
		mask = level << 12;
	}
	/* Default to any other file, intercept.c */

	/* Create the text to be put in the trace log
	 * if the level is set in the debug flag,
	 * or if this is a serious item.
	 */
	if ((pfo_debug_flag & (mask)) || (level & 3)) {
		count += sprintf(tmp+count, "%d:%s:%d:%s\t",
					level, _file, _line, _func);
		va_start(argptr, fmt);
		count += vsprintf(tmp+count, fmt, argptr);
		va_end(argptr);
		if (tmp[count-1] != '\n') {
			count += sprintf(tmp+count, "\n");
		}
		if (pfo_debug_flag & (mask)) {
			printk(tmp);
		}
		if (count <= (DEBUG_BUFFER_SIZE-1-debug_tail)) {
			strcpy(debug_buffer+debug_tail, tmp);
			debug_tail += count;
		} else {
			strncpy(debug_buffer+debug_tail, tmp,
					DEBUG_BUFFER_SIZE-1-debug_tail);
			debug_buffer[DEBUG_BUFFER_SIZE-1] = '\0';
			strcpy(debug_buffer,
					tmp+DEBUG_BUFFER_SIZE-1-debug_tail);
			debug_tail = strlen(debug_buffer);
		}
	}
}

/** Enable the change of the debug control flag.
 *  /sys/bus/scsi/drivers/pfo/debug_flag */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static ssize_t pfoattr_debug_flag_set(struct device_driver *dev,
					const char *buf, size_t count)
{
	sscanf(buf, "0x%x", &pfo_debug_flag);
	DBG_INFO("debug flag set to 0x%04x\n", pfo_debug_flag);
	return count;
}

/** Enable the cat of the debug control flag.
 *  /sys/bus/scsi/drivers/pfo/debug_flag */
static ssize_t pfoattr_debug_flag_get(struct device_driver *ddp, char *buf)
{
	DBG_DEBUG("");
	return scnprintf(buf, PAGE_SIZE,
		"debug flag=%04x\necho 0x<hexvalue> > debug_flag to change\n",
		pfo_debug_flag);
}
static DRIVER_ATTR(debug_flag, S_IRUGO|S_IWUSR|S_IWGRP,
			pfoattr_debug_flag_get, pfoattr_debug_flag_set);

/** Enable the cat of the debug buffer.
 *  /sys/bus/scsi/drivers/pfo/debug_buffer */
static ssize_t pfoattr_debug_buffer_get(struct device_driver *ddp, char *buf)
{
	DBG_DEBUG("");
	if (debug_buffer[0] == '\001') {
		return scnprintf(buf, PAGE_SIZE, "%s", debug_buffer+1);
	} else {
		strcpy(buf, debug_buffer+debug_tail+1);
		strcat(buf, debug_buffer);
		return strlen(buf);
	}
}
static DRIVER_ATTR(debug_buffer, S_IRUGO, pfoattr_debug_buffer_get, NULL);

/** Enable the cat of the trace buffer.
 *  /sys/bus/scsi/drivers/pfo/trace */
static ssize_t pfoattr_debug_trace_get(struct device_driver *ddp, char *buf)
{
	char *the_buffer;
	int strlenbuf = 0;
	unsigned char *ptr = NULL;

	the_buffer = kmalloc(2*PAGE_SIZE, GFP_KERNEL);
	if (the_buffer != NULL) {
		trace_get(the_buffer);
		strlenbuf = strlen(the_buffer);
		if (the_buffer[strlenbuf+1] != '\0') {
			ptr = strchr(&the_buffer[strlenbuf+1], '\n');
			if (ptr != NULL && *ptr == '\n' && *(++ptr)) {
				strcpy(buf, ptr);
				strcat(buf, the_buffer);
				strlenbuf = strlen(buf);
			} else {
				strlenbuf = 0;
			}
		} else {
			strlenbuf = scnprintf(buf, PAGE_SIZE, "%s", the_buffer);
		}
		kfree(the_buffer);
	}
	return strlenbuf;
}
static DRIVER_ATTR(trace, S_IRUGO, pfoattr_debug_trace_get, NULL);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static ssize_t debug_flag_store (struct device_driver *dev,
		const char *buf, size_t count)
{
sscanf(buf, "0x%x", &pfo_debug_flag);
DBG_INFO("debug flag set to 0x%04x\n", pfo_debug_flag);
return count;
}
static ssize_t debug_flag_show(struct device_driver *ddp, char *buf)
{
	DBG_DEBUG("");
	return scnprintf(buf, PAGE_SIZE,
		"debug flag=%04x\necho 0x<hexvalue> > debug_flag to change\n",
		pfo_debug_flag);
}
static DRIVER_ATTR_RW(debug_flag);

static ssize_t debug_buffer_show (struct device_driver *ddp, char *buf)
{
	DBG_DEBUG("");
	if (debug_buffer[0] == '\001') {
		return scnprintf(buf, PAGE_SIZE, "%s", debug_buffer+1);
	} else {
		strcpy(buf, debug_buffer+debug_tail+1);
		strcat(buf, debug_buffer);
		return strlen(buf);
	}
}
static DRIVER_ATTR_RO(debug_buffer);

static ssize_t trace_show(struct device_driver *ddp, char *buf)
{
	char *the_buffer;
	int strlenbuf = 0;
	unsigned char *ptr = NULL;

	the_buffer = kmalloc(2*PAGE_SIZE, GFP_KERNEL);
	if (the_buffer != NULL) {
		trace_get(the_buffer);
		strlenbuf = strlen(the_buffer);
		if (the_buffer[strlenbuf+1] != '\0') {
			ptr = strchr(&the_buffer[strlenbuf+1], '\n');
			if (ptr != NULL && *ptr == '\n' && *(++ptr)) {
				strcpy(buf, ptr);
				strcat(buf, the_buffer);
				strlenbuf = strlen(buf);
			} else {
				strlenbuf = 0;
			}
		} else {
			strlenbuf = scnprintf(buf, PAGE_SIZE, "%s", the_buffer);
		}
		kfree(the_buffer);
	}
	return strlenbuf;
}
static DRIVER_ATTR_RO(trace);
#endif

int init_debug(struct device_driver *pfo_gendrv)
{
	int rc;

	DBG_DEBUG("");

	rc = driver_create_file(pfo_gendrv, &driver_attr_debug_flag);
	if (rc) {
		DBG_ERROR("failed to create debug_flag file");
		goto create_file_error_debug_flag;
	}

	rc = driver_create_file(pfo_gendrv, &driver_attr_debug_buffer);
	if (rc) {
		DBG_ERROR("failed to create debug_buffer file");
		goto create_file_error_debug_buffer;
	}
	rc = driver_create_file(pfo_gendrv, &driver_attr_trace);
	if (rc) {
		DBG_ERROR("failed to create trace file");
		goto create_file_error_trace;
	}
	return 0;
create_file_error_trace:
	driver_remove_file(pfo_gendrv, &driver_attr_debug_buffer);
create_file_error_debug_buffer:
	driver_remove_file(pfo_gendrv, &driver_attr_debug_flag);
create_file_error_debug_flag:
	return 1;
}

void stop_debug(struct device_driver *pfo_gendrv)
{
	DBG_DEBUG("");
	driver_remove_file(pfo_gendrv, &driver_attr_debug_buffer);
	driver_remove_file(pfo_gendrv, &driver_attr_debug_flag);
	driver_remove_file(pfo_gendrv, &driver_attr_trace);
}
