/**
 * This file contains the code that
 * processes path adds and removes from the SCSI mid-layer.
 * It sends notifications to all of the other code, as needed,
 * about the additions of new APFO paths and the failure of known paths.
 * It simply passes non-APFO devices up to st and sg unaltered.
 * It handles module init and exit callbacks.
 * It handles the sysfs interface for this module.
 *
 */
/*
 * (c) Copyright International Business Machines  Corp., 2020
 *
 * (C) Copyright 2016 Hewlett Packard Enterprise Development LP
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/idr.h>
#include <scsi/scsi_driver.h>

#include "scsi.h"
#include "pfo.h"
#include "apfo.h"

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
#include "sysfs.h"
#endif

char *intercept_version_date = "20170113A";

/* we don't remove sg/st device during regular operation, but they must be
 * removed upon module unload, thus we use this flag to tell the difference */
static int REMOVE_ALL_TRUE = 1;
static int REMOVE_ALL_FALSE = 2;
static int DELETE_PFODEV = 1;
static int RETAIN_PFODEV = 0;
int PFO_MAJOR;
/** !=0 for NetBackup operation.
 * Must be read-only because things are
 * setup at insert and then cleaned up at remove. */
static int netbackup;

/* pfo common defines */
#define MAXSUBPATH 64
#define MAXPFO 1024 //lintape supports up to 1024 tape devices

/* main pfo structures */
struct pfo_subpath {
	struct device *scsidev;
	struct class_interface *clsinterface;
	int path_num;
	int sfmp_registered;
};

#define VENDOR_SIZE 8
#define MODEL_SIZE 16
#define REV_SIZE 4

struct pfo_device {
	int pfonum;
	int scsitype;
	unsigned char wwnn[WWN_SIZE];
	unsigned char serial[SN_SIZE];
	unsigned char vendor[VENDOR_SIZE+1];
	unsigned char model[MODEL_SIZE+1];
	unsigned char rev[REV_SIZE+1];
	unsigned int host;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;
	int activepath;
	int subpathcount;
	int sg_num;
	int st_num;
	int sfmp_determined;	/* sfmp driver has been consulted */
	int sfmp_supported;	/* sfmp driver claims support */
	int sfmp_fo_determined;	/* sfmp driver has responded */
	int sfmp_fo_enabled;	/* sfmp driver claims fo enabled */
	int fdev_num;
	struct device *upper_dev;
	struct device *gen_dev;
	struct kobject *kobj_scsi_generic;	/* pfo/pfo#/scsi_generic     */
	struct kobject *kobj_scsi_sg;		/* pfo/pfo#/scsi_generic/sg# */
	struct kobject *kobj_scsi_tape;		/* pfo/pfo#/scsi_tape        */
	struct kobject *kobj_scsi_st;		/* pfo/pfo#/scsi_tape/st#    */
	struct kobject *kobj_scsi_nst;		/* pfo/pfo#/scsi_tape/nst#   */
	struct pfo_subpath subpath[MAXSUBPATH];
};

static int intercept_add(struct device *dev, struct class_interface *pclass);
static void intercept_remove(struct device *dev,
					struct class_interface *pclass);
static void intercept_ctrl(struct pfo_device *pdev, const char *buf,
							size_t count);
static int intercept_create_unique_device(struct pfo_device *pitem,
							struct device *dev);

static struct class_interface pfo_interface = {
    .add_dev        = intercept_add,
    .remove_dev     = intercept_remove,
};

static struct scsi_driver pfo_template = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
	.owner	= THIS_MODULE,
#endif
	.gendrv	= {
		.name	= "pfo",
		.probe	= NULL,
		.remove	= NULL,
/*#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
		.groups = pfo_drv_groups,
#endif*/
	},
};
static struct class *pfo_sysfs_class;

struct _upper_device_callback {
	int (*sg_add_dev)(struct device *, struct class_interface *);
	void (*sg_remove_dev)(struct device *, struct class_interface *);
	void (*sg_switch)(struct device *, struct device *);
	int (*st_probe) (struct device *);
	int (*st_remove) (struct device *);
	void (*st_switch)(struct device *, struct device *);
	int (*join_add)(struct device *, struct class_interface *);
	void (*join_remove)(struct device *, struct class_interface *);
	int (*join_open)(struct join_device_identifier *gizmo);
	int (*join_close)(struct join_device_identifier *gizmo);
	int (*join_send)(struct join_device_identifier *gizmo,
			unchar *cdb,
			unchar cmd_length,
			unchar *buffer,
			uint buffer_length,
			uint data_direction,
			uint timeout,
			unchar *sense,
			int *sense_length,
			int *resid,
			int *result);

} upper_device_callback;

/* the lock and IDR list to keep pfo devices in order */
DEFINE_IDR(pfo_devices);
DEFINE_RWLOCK(pfo_devices_lock);
DEFINE_SPINLOCK(pfo_subpath_lock);

/** This value controls whether we check
 * APFO capabilities with decide layer or not */
static int ctrl_apfo_on = 1;
static volatile int max_pfonum = -1;

int intercept_print_device(char *string, int path)
{
	return 0;
}

/* the unused attributes must be set to -1 or NULL */
struct device_predicate {
	int fdev_num;		/* in */
	int path_num;		/* in */
	struct device *scsidev;	/* in */
	struct device *gendev;	/* in */
	unsigned char wwnn[WWN_SIZE];	/* in/out */
	unsigned char serial[SN_SIZE]; /* in/out */
	struct pfo_device *pdev;/* out */
	int pfonum;		/* out */
	int pathid;		/* out */
};
static void dump_pdev(struct pfo_device *pdev)
{
	int index;

	DBG_INFO("dev pfo%d\n", pdev->pfonum);
	DBG_INFO("fdev_num=%d\n", pdev->fdev_num);
	DBG_INFO("upper_dev=%p", pdev->upper_dev);
	DBG_INFO("gen_dev=%p", pdev->gen_dev);
	DBG_INFO("activesubpath=%d\n", pdev->activepath);
	DBG_INFO("sg_num=%d\n", pdev->sg_num);
	DBG_INFO("st_num=%d\n", pdev->st_num);
	if (pdev->subpathcount > 20 || pdev->subpathcount < 0) {
		DBG_INFO("subpathcount=%d\n", pdev->subpathcount);
		return;
	}
	for (index = 0; index < (pdev->subpathcount)+1; index++) {
		DBG_INFO("%d %p path=%d\n", index,
			pdev->subpath[index].scsidev,
			pdev->subpath[index].path_num);
	}
}
/** Find a pfo entry that meets given criteria.
 * @return 0 if not find, 1 if find, device detail is in the predicate structure */
static int find_device(int id, void *p, void *predicate)
{
	int i;
	struct pfo_device *pdev = (struct pfo_device *)p;
	struct device_predicate *cand = (struct device_predicate *)predicate;

	if (cand->fdev_num != -1) {
		if (cand->fdev_num == pdev->fdev_num) {
			cand->pfonum = pdev->pfonum;
			cand->pdev = pdev;
			memcpy(cand->wwnn, pdev->wwnn, WWN_SIZE);
			memcpy(cand->serial, pdev->serial, SN_SIZE);
			return 1;
		}
	} else if (cand->path_num != -1) {
		for (i = 0; i < pdev->subpathcount; i++) {
			if (cand->path_num == pdev->subpath[i].path_num) {
				cand->pdev = pdev;
				cand->pfonum = pdev->pfonum;
				cand->pathid = i;
				return 1;
			}
		}
	} else if (cand->gendev != NULL) {
		if (cand->gendev == pdev->gen_dev) {
			cand->pdev = pdev;
			return 1;
		}
	} else if (cand->scsidev != NULL) {
		for (i = 0; i < pdev->subpathcount; i++) {
			if (cand->scsidev == pdev->subpath[i].scsidev) {
				cand->pdev = pdev;
				cand->pfonum = pdev->pfonum;
				cand->pathid = i;
				cand->path_num = pdev->subpath[i].path_num;
				cand->fdev_num = pdev->fdev_num;
				return 1;
			}
		}
	} else {
		if (memcmp(cand->wwnn, pdev->wwnn, sizeof(pdev->wwnn)) == 0) {
			cand->pdev = pdev;
			cand->pfonum = pdev->pfonum;
			cand->pathid = -1;
			return 1;
		}
		if (cand->serial[0] &&
		    strncmp(cand->serial, pdev->serial, SN_SIZE) == 0) {
			cand->pdev = pdev;
			cand->pfonum = pdev->pfonum;
			cand->pathid = -1;
			return 1;
		}
	}

	return 0;
}

/** Locate the pfo_device structure to use
 * for the given device structure of the device.
 */
struct pfo_device *pfo_dev_get_drvdata(struct device *dev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
	struct device_predicate cand;

	cand.fdev_num = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.pdev = NULL;
	cand.serial[0] = '\0';
	cand.gendev = dev;
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);
	return cand.pdev;
#else
	return dev_get_drvdata(dev);
#endif
}

/** Perform the operations the customer has requested when they wrote to
 * /sys/bus/scsi/drivers/pfo/ctrl
 */
static ssize_t ctrl_store(struct device_driver *dev, const char *buf,
				size_t count)
{
	intercept_ctrl(NULL, buf, count);
	return count;
}

/** Create the text the customer will see when they read
 * /sys/bus/scsi/drivers/pfo/ctrl
 * @return string length
 */
static ssize_t ctrl_show(struct device_driver *ddp, char *buf)
{
	return resolve_ctrl_get(buf, PAGE_SIZE);
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
DRIVER_ATTR_RW(ctrl);
#else
static DRIVER_ATTR(ctrl, S_IRUGO|S_IWUSR|S_IWGRP, ctrl_show,
				ctrl_store);
#endif

/** Create the text the customer will see when they read
 * /sys/bus/scsi/drivers/pfo/version
 * @return string length
 */
static ssize_t pfo_version_show(struct device_driver *ddd, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "[%s]\n", PFO_DATE_STR);
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
DRIVER_ATTR_RO(pfo_version);
#else
static DRIVER_ATTR(pfo_version, S_IRUGO, pfo_version_show, NULL);
#endif
static ssize_t devattr_vendor(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%s\n", pdev->vendor);
	return count;
}
DEVICE_ATTR(vendor, S_IRUGO, devattr_vendor, NULL);

static ssize_t devattr_model(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%s\n", pdev->model);
	return count;
}
DEVICE_ATTR(model, S_IRUGO, devattr_model, NULL);

static ssize_t devattr_rev(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%s\n", pdev->rev);
	return count;
}
DEVICE_ATTR(rev, S_IRUGO, devattr_rev, NULL);

static ssize_t devattr_serial(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%s\n", pdev->serial);
	return count;
}
DEVICE_ATTR(serial, S_IRUGO, devattr_serial, NULL);

static ssize_t devattr_type_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%d\n", pdev->scsitype);
	return count;
}
DEVICE_ATTR(type, S_IRUGO, devattr_type_get, NULL);

static ssize_t devattr_wwnn_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "%02x%02x%02x%02x%02x%02x%02x%02x\n",
		pdev->wwnn[0], pdev->wwnn[1], pdev->wwnn[2], pdev->wwnn[3],
		pdev->wwnn[4], pdev->wwnn[5], pdev->wwnn[6], pdev->wwnn[7]);
	return count;
}
DEVICE_ATTR(wwnn, S_IRUGO, devattr_wwnn_get, NULL);

static ssize_t devattr_internal_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int i, count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		count += sprintf(buf+count, "NULL pointer.\n");
		return count;
	}
	count += sprintf(buf+count, "dev pfo%d: \n", pdev->pfonum);
	count += sprintf(buf+count, " type = %d\n", pdev->scsitype);
	count += sprintf(buf+count, " wwnn = %02x%02x%02x%02x%02x%02x%02x%02x\n",
		pdev->wwnn[0], pdev->wwnn[1], pdev->wwnn[2], pdev->wwnn[3],
		pdev->wwnn[4], pdev->wwnn[5], pdev->wwnn[6], pdev->wwnn[7]);
	count += sprintf(buf+count, " pfo_device *= %p\n", pdev);
	count += sprintf(buf+count, " dev   *= %p\n", dev);
	count += sprintf(buf+count, " gendev*= %p\n", pdev->gen_dev);
	count += sprintf(buf+count, " upper *= %p\n", pdev->upper_dev);
	count += sprintf(buf+count, " fdev_num = %d\n", pdev->fdev_num);
	count += sprintf(buf+count, " activepath = %d\n", pdev->activepath);
	count += sprintf(buf+count, " pdev.get_drvdata = %p\n",
		 dev_get_drvdata(pdev->upper_dev->parent));
	count += sprintf(buf+count, " upper.get_drvdata = %p\n",
		 dev_get_drvdata(pdev->upper_dev));
	for (i = 0; i < pdev->subpathcount; i++) {
		count += sprintf(buf+count,
			" subpath %2d : dev*=%p scsidev*=%p "
			" sys*=%p, path=%d sf_reg=%d\n",
			i, pdev->subpath[i].scsidev,
			to_scsi_device(pdev->subpath[i].scsidev->parent),
			pdev->subpath[i].clsinterface,
			pdev->subpath[i].path_num,
			pdev->subpath[i].sfmp_registered);
	}

	return count;
}
DEVICE_ATTR(internal, S_IRUGO, devattr_internal_get, NULL);

static ssize_t devattr_ctrl_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	return scnprintf(buf, PAGE_SIZE,
		"This file is write only.\nIt is for test and debug.\n");
}

/** Perform the operations the customer has requested when they wrote to
 * /sys/class/pfo/pfo?/ctrl
 */
static ssize_t devattr_ctrl_set(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		return sprintf((char *)buf, "NULL pointer.\n");
	}
	intercept_ctrl(pdev, buf, size);
	return size;
}
DEVICE_ATTR(ctrl, S_IRUGO|S_IWUSR|S_IWGRP, devattr_ctrl_get, devattr_ctrl_set);

#ifdef APFO
static ssize_t devattr_sessions_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		return sprintf(buf, "NULL pointer.\n");
	}
	strcpy(buf, "List of sessions on the device -\n");
	if (pdev->fdev_num != -1) {
		decide_print_sessions(pdev->fdev_num, buf, 4095);
	} else {
		strcat(buf, "Not an APFO device.\n");
	}
	return strlen(buf);
}
DEVICE_ATTR(sessions, S_IRUGO, devattr_sessions_get, NULL);
#endif

static ssize_t devattr_paths_get(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	int count = 0;
	struct pfo_device *pdev = (struct pfo_device *)pfo_dev_get_drvdata(dev);

	if (pdev == NULL) {
		return sprintf(buf+count, "NULL pointer.\n");
	}
	count += sprintf(buf+count, "pfo%-2d ", pdev->pfonum);
	if (pdev->sg_num == -1) {
		count += sprintf(buf+count, "sg=none ");
	} else {
		count += sprintf(buf+count, "sg=/dev/sg%-2d ", pdev->sg_num);
	}
	if (pdev->st_num == -1) {
		count += sprintf(buf+count, "st=none ");
	} else {
		count += sprintf(buf+count, "st=/dev/st%-2d ", pdev->st_num);
	}
	if (pdev->sfmp_determined) {
		if (pdev->sfmp_supported) {
			count += sprintf(buf+count, "sf=yes ");
			if (pdev->sfmp_fo_determined) {
				if (pdev->sfmp_fo_enabled) {
					count += sprintf(buf+count, "fo=yes ");
				} else {
					count += sprintf(buf+count, "fo=no  ");
				}
			} else {
				count += sprintf(buf+count, "fo=??  ");
			}
		}
	} else {
		count += sprintf(buf+count, "sf=??  ");
	}
	if (pdev->fdev_num != -1) {
		resolve_print_device(pdev->fdev_num, buf, 4095);
	} else {
		strcat(buf, "\n");
	}
	return strlen(buf);
}
DEVICE_ATTR(paths, S_IRUGO, devattr_paths_get, NULL);
/** Find the device structure pointer that is being
 * used by the sg or st module for this device.
 */
struct device *get_realupper(struct pfo_device *pdev)
{
	if (pdev->fdev_num == -1) {
		return pdev->subpath[0].scsidev;
	} else {
		return pdev->upper_dev;
	}
}

/** intercept layer triggered path change,
 this is also used by intercept_notify_path_change, don't be surprised */
static void my_activepathchange(struct pfo_device *pdev, struct device *newpath)
{
	DBG_DEBUG("upper_dev=%p, newpath=%p", pdev->upper_dev, newpath);
	if (upper_device_callback.st_switch != NULL && pdev->scsitype == 1) {
		(upper_device_callback.st_switch)(get_realupper(pdev)->parent,
			((newpath == NULL) ? NULL : newpath->parent));
	}

	if (upper_device_callback.sg_switch != NULL) {
		(upper_device_callback.sg_switch)(get_realupper(pdev), newpath);
	}
}

/** Modify the link in /sys/class so NetBackup will work.
 * @param[in] scsi_type generic or type
 * @param[in] link_name st# or sg#
 * @param[in] target where to link to
 */
static void netbackup_hack(char *scsi_type, char *link_name,
			struct kobject *kobj_target)
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	int result;
	struct kobject *kobj = kobj_target;
	struct sysfs_dirent *root_sysfs = NULL;
	struct sysfs_dirent *this_sysfs = NULL;
	struct sysfs_dirent *scsi_sysfs = NULL;

	/* Find sysfs root */
	while (kobj->parent) {
		kobj = kobj->parent;
	}
	/* Find the kobject that holds the /sys/class/"scsi_type" */
	root_sysfs = kobj->sd->s_parent;
	if (root_sysfs) {
		this_sysfs = sysfs_get_dirent(root_sysfs, "class");
		if (this_sysfs) {
			scsi_sysfs = sysfs_get_dirent(this_sysfs, scsi_type);
			if (scsi_sysfs) {
				kobj = scsi_sysfs->s_dir.kobj;
				sysfs_remove_link(kobj, link_name);
				result = sysfs_create_link(kobj, kobj_target, link_name);
				if (!result) {
					return;
				}
			}
		}
	}
	print_log("Failed to create netbackup links\n");
#endif
}

/** Break this device apart if needed.
 * The /dev/sg and /dev/st files have not yet been created,
 * so this is preparation for that. */
static void split_device(struct pfo_device *pdev)
{
	int fdev_num;
	fdev_num = pdev->fdev_num;
	DBG_INFO("I see fdev=%d pfo%d\n", fdev_num, pdev->pfonum);
	/* Before lin_tape is inserted we assume that every drive
	 * and library will be supported by lin_tape.  If that is
	 * not the case then we fix it here.
	 */
	if (!pdev->sfmp_supported) {
		DBG_INFO("Not a lin_tape device\n");
		/* Tell upper to have the original Linux
		 * code handle the SCSI commands. */
		upper_not_fo_device(fdev_num);
		/* Let resolve know.  */
		resolve_set_is_not_sfmp_device(fdev_num);
		/* Remember that this is non-fo */
		pdev->fdev_num = -1;
	}
	/* Before lin_tape is inserted we assume that every drive
	 * and library will have failover enabled.  If that is
	 * not the case then we fix it here.
	 */
	if (!pdev->sfmp_supported ||
	    (pdev->sfmp_fo_determined && !pdev->sfmp_fo_enabled)) {
		DBG_INFO("Need split\n");
		/* Break all of the sub-paths to this device into
		 * separate devices and create the /dev files.
		 * Resolve calls back into intercept to create
		 * new devices and the /dev files as needed.
		 */
		resolve_not_fo_device(pdev, fdev_num);
	}
}

/** Split any non-failover multi-path devices
 * into multiple, single path devices
 * in preaparation for creating sg and st device files.
 */
static void split_up_non_sfmp_multi_path_devices(void)
{
	int next_id;
	struct pfo_device *pdev;

	/* Make sure to get all of the way through the list
	 * because the list length WILL change when we
	 * add new devices from old paths.
	 */
	read_lock(&pfo_devices_lock);
	for (next_id = 0; next_id <= max_pfonum; next_id++) {
		pdev = idr_find(&pfo_devices, next_id);
		if (pdev) {
			if (pdev->fdev_num >= 0 && pdev->sfmp_determined) {
				read_unlock(&pfo_devices_lock);
				split_device(pdev);
				read_lock(&pfo_devices_lock);
			}
		}
	}
	read_unlock(&pfo_devices_lock);
}

/** Add a device to sfmp.
 * Makes sure that every path to the device is registered.
 * Note: An assumption is made that every path to a device
 * will get the same pass/fail response from sfmp add.
 * @return can be ignored as it's always 0
 */
static int add_device_to_sf(int pfonum, void *p, void *DONTUSE)
{
	int path_num;
	int sfmp_result;
	struct pfo_device *pdev = (struct pfo_device *)p;

	if (upper_device_callback.join_add == NULL) {
		/* lin_tape is not yet loaded. */
		return 0;
	}
	if (pdev->fdev_num < 0) {
		/* Do not bother to add devices
		 * that are known to not be failover. */
		DBG_INFO("Not fo pfonum=%d pdev=%p\n", pfonum, pdev);
		return 0;
	}
	for (path_num = 0; path_num < pdev->subpathcount; path_num++) {
		if (!pdev->subpath[path_num].sfmp_registered) {
			DBG_DEBUG("Add dev pfo%d (%p) to sfmp\n",
					pfonum,
					pdev->subpath[path_num].scsidev);
			sfmp_result = upper_device_callback.join_add(
				pdev->subpath[path_num].scsidev,
				pdev->subpath[path_num].clsinterface);
			if (sfmp_result == 0) {
				pdev->sfmp_supported = 1;
				pdev->sfmp_determined = 1;
				pdev->subpath[path_num].sfmp_registered = 1;
				/* fo enabled/disabled status decided elsewhere */
			} else {
				/* lin_tape rejected this device. */
				pdev->sfmp_supported = 0;
				pdev->sfmp_fo_enabled = 0;
				pdev->sfmp_determined = 1;
				pdev->sfmp_fo_determined = 1;
				DBG_INFO("lin_tape rejected fdev=%d\n",
							pdev->fdev_num);
				return 0;
			}
		}
	}
	return 0;
}

/** Add a device to sg.
 * The function only registers the active path on a device.
 * Sub-paths are skipped.
 * @return can be ignored as it's always 0
 */
static int add_device_to_sg(int pfonum, void *p, void *DONTUSE)
{
	int i;
	struct pfo_device *pdev = (struct pfo_device *)p;
	if (pdev->sg_num >= 0) {
		/* Do not create duplicate /dev files.  */
		DBG_DEBUG("Skip duplicate add dev pfo%d (%p) to sg%d\n",
				pfonum, get_realupper(pdev), pdev->sg_num);
		return 0;
	}
	if (!pdev->sfmp_determined) {
		/* Do not create /dev files until we know
		 * whether this is an sfmp supported device.
		 */
		return 0;
	}
	if (pdev->sfmp_supported && !pdev->sfmp_fo_determined) {
		/* Do not create /dev files until sfmp has said
		 * whether failover has been enabled or not.
		 */
		return 0;
	}
	if (upper_device_callback.sg_add_dev != NULL) {
		i = upper_device_callback.sg_add_dev(get_realupper(pdev),
			pdev->subpath[pdev->activepath].clsinterface);
		DBG_INFO("Add dev pfo%d (%p) to /dev/sg%d\n",
					pfonum, get_realupper(pdev), i);
		if (i >= 0) {
			pdev->sg_num = i;
			/* pdev->upper_dev->kobj points to -
			 * /sys/devices/pci0000...scsi_device/6:0:0:1
			 */
		} else {
			DBG_ERROR("sg_add failed with %d\n", i);
			return 0;
		}
		my_activepathchange(pdev, pdev->subpath[pdev->activepath].scsidev);
		if (netbackup && pdev->fdev_num != -1) {
			/* It is an APFO device so
			 * create special subdir and link for NetBackup */
			int result;
			char buf[10];
			struct device *sdev = get_realupper(pdev);
			sprintf(buf, "sg%d", pdev->sg_num);
			pdev->kobj_scsi_generic = kobject_create_and_add(
						"scsi_generic",
						&sdev->parent->kobj);
			if (pdev->kobj_scsi_generic == NULL) {
				print_log("Failed to add scsi_generic directory.\n");
				return 0;
			}
			pdev->kobj_scsi_sg = kobject_create_and_add(buf,
						pdev->kobj_scsi_generic);
			if (pdev->kobj_scsi_sg == NULL) {
				print_log("Failed to add scsi_generic/sg# directory.\n");
				return 0;
			}
			result = sysfs_create_link(&sdev->parent->kobj,
						pdev->kobj_scsi_sg,
						"generic");
			if (result) {
				print_log("Failed to add generic to sg# link\n");
			}
			netbackup_hack("scsi_generic", buf, pdev->kobj_scsi_sg);
		}
	}
	return 0;
}

/** Add a device to st.
 * The function check and skips if this is only a subpath of an APFO device
 * @return can be ignored as it's always 0
 */
static int add_device_to_st(int pfonum, void *p, void *DONTUSE)
{
	int i;
	struct pfo_device *pdev = (struct pfo_device *)p;
	if (pdev->st_num >= 0) {
		/* Do not create duplicate /dev files.  */
		DBG_DEBUG("Skip duplicate add dev pfo%d (%p) to st%d\n",
				pfonum, get_realupper(pdev), pdev->st_num);
		return 0;
	}
	if (!pdev->sfmp_determined) {
		/* Do not create /dev files until we know
		 * whether this is an sfmp supported device.
		 */
		return 0;
	}
	if (pdev->sfmp_supported && !pdev->sfmp_fo_determined) {
		/* Do not create /dev files until sfmp has said
		 * whether failover has been enabled or not.
		 */
		return 0;
	}
	if (upper_device_callback.st_probe != NULL && pdev->scsitype == 1) {
		DBG_DEBUG("Add dev pfo%d (%p) to st\n",
					pfonum, get_realupper(pdev));
		i = upper_device_callback.st_probe(get_realupper(pdev)->parent);
		if (i >= 0) {
			pdev->st_num = i;
		} else {
			DBG_INFO("st_add failed with %d\n", i);
			return 0;
		}
		my_activepathchange(pdev, pdev->subpath[pdev->activepath].scsidev);
		if (netbackup && pdev->fdev_num != -1) {
			/* It is an APFO device so
			 * create special subdir and link for NetBackup */
			int result;
			char buf[10];
			struct device *sdev = get_realupper(pdev);
			/*
			 * parent->kobj = /sys/devices/virtual/pfo/pfo#
			 * Create /sys/devices/virtual/pfo/pfo#/scsi_tape
			 */
			pdev->kobj_scsi_tape = kobject_create_and_add(
						"scsi_tape",
						&sdev->parent->kobj);
			if (pdev->kobj_scsi_tape == NULL) {
				print_log("Failed to add scsi_tape directory.\n");
				return 0;
			}
			/*
			 * Create st# related items.
			 */
			sprintf(buf, "st%d", pdev->st_num);
			/*
			 * Create /sys/devices/virtual/pfo/pfo#/scsi_tape/st#
			 */
			pdev->kobj_scsi_st = kobject_create_and_add(buf,
						pdev->kobj_scsi_tape);
			if (pdev->kobj_scsi_st == NULL) {
				print_log("Failed to add scsi_tape/st# directory.\n");
				return 0;
			}
			/*
			 * Create /sys/devices/virtual/pfo/pfo#/tape
			 * pointing to
			 * /sys/devices/virtual/pfo/pfo#/scsi_tape/st#
			 */
			result = sysfs_create_link(&sdev->parent->kobj,
						pdev->kobj_scsi_st,
						"tape");
			if (result) {
				print_log("Failed to add tape link to st# link\n");
			}
			/*
			 * Create link /sys/class/scsi_tape/st#
			 * pointing to
			 * /sys/devices/virtual/pfo/pfo#/scsi_tape/st#
			 */
			netbackup_hack("scsi_tape", buf, pdev->kobj_scsi_st);
			/*
			 * Create nst# related items.
			 */
			sprintf(buf, "nst%d", pdev->st_num);
			/*
			 * Create /sys/devices/virtual/pfo/pfo#/scsi_tape/nst#
			 */
			pdev->kobj_scsi_nst = kobject_create_and_add(buf,
						pdev->kobj_scsi_tape);
			if (pdev->kobj_scsi_nst == NULL) {
				print_log("Failed to add scsi_tape/nst# directory.\n");
				return 0;
			}
			/*
			 * Create link /sys/class/scsi_tape/nst#
			 * pointing to
			 * /sys/devices/virtual/pfo/pfo#/scsi_tape/nst#
			 */
			netbackup_hack("scsi_tape", buf, pdev->kobj_scsi_nst);
		}
	}
	return 0;
}

/** Remove a device from sfmp.  It removes all paths.
 * @return can be ignored as it's always 0
 */
static int remove_device_from_sf(int id, void *p, void *remove_all)
{
	int path_num;
	struct pfo_device *pdev = (struct pfo_device *)p;

	DBG_INFO("\n");
	if (upper_device_callback.join_remove == NULL) {
		return 0;
	}
	for (path_num = 0; path_num < pdev->subpathcount; path_num++) {
		if (pdev->subpath[path_num].sfmp_registered) {
			DBG_INFO("remove device %d from sf\n", id);
			upper_device_callback.join_remove(
				pdev->subpath[path_num].scsidev,
				pdev->subpath[path_num].clsinterface);
			pdev->subpath[path_num].sfmp_registered = 0;
		}
	}
	return 0;
}

/** Remove a device from sg, the function checks and
 * skips if this is only a subpath of an APFO device.
 * @return can be ignored as it's always 0
 */
static int remove_device_from_sg(int id, void *p, void *remove_all)
{
	struct pfo_device *pdev = (struct pfo_device *)p;

	if (pdev->fdev_num != -1 && (*(int *)remove_all) == REMOVE_ALL_FALSE) {
		return 0;
	}

	if (pdev->sg_num < 0) {
		/* Do not remove /dev file that was never created. */
		return 0;
	}
	if (netbackup) {
		/* Remove the stuff for NetBackup. */
		struct device *sdev;
		sdev = get_realupper(pdev);
		if (sdev && sdev->parent && &sdev->parent->kobj) {
			sysfs_remove_link(&sdev->parent->kobj, "generic");
		}
		if (pdev->kobj_scsi_sg) {
			kobject_put(pdev->kobj_scsi_sg);
		}
		if (pdev->kobj_scsi_generic) {
			kobject_put(pdev->kobj_scsi_generic);
		}
	}

	if (upper_device_callback.sg_remove_dev != NULL) {
		DBG_INFO("remove device %d from sg\n", id);
		upper_device_callback.sg_remove_dev(get_realupper(pdev),
			pdev->subpath[pdev->activepath].clsinterface);
		pdev->sg_num = -1;
	}
	return 0;
}

/** Remove a device from st,
 * the function check and skips if this is only a subpath of an APFO device.
 * @return can be ignored as it's always 0
 */
static int remove_device_from_st(int id, void *p, void *remove_all)
{
	struct pfo_device *pdev = (struct pfo_device *)p;

	if (pdev->fdev_num != -1 && (*(int *)remove_all) == REMOVE_ALL_FALSE) {
		return 0;
	}

	if (pdev->st_num < 0) {
		/* Do not remove /dev file that was never created. */
		return 0;
	}
	if (netbackup) {
		/* Remove the stuff for NetBackup. */
		struct device *sdev;
		sdev = get_realupper(pdev);
		if (sdev && sdev->parent && &sdev->parent->kobj) {
			sysfs_remove_link(&sdev->parent->kobj, "tape");
		}
		if (pdev->kobj_scsi_st) {
			kobject_put(pdev->kobj_scsi_st);
		}
		if (pdev->kobj_scsi_nst) {
			kobject_put(pdev->kobj_scsi_nst);
		}
		if (pdev->kobj_scsi_tape) {
			kobject_put(pdev->kobj_scsi_tape);
		}
	}

	if (upper_device_callback.st_remove != NULL && pdev->scsitype == 1) {
		DBG_DEBUG("remove device %d (%p) from st\n", id, get_realupper(pdev));
		upper_device_callback.st_remove(get_realupper(pdev)->parent);
		pdev->st_num = -1;
	}
	return 0;
}

int pfo_register_st(struct device_driver *drv,
	void (*path_switch_callback)(struct device *, struct device *))
{
	upper_device_callback.st_probe = drv->probe;
	upper_device_callback.st_remove = drv->remove;
	upper_device_callback.st_switch = path_switch_callback;

	drv->probe = NULL;
	drv->remove = NULL;
	scsi_register_driver(drv);

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, add_device_to_st, NULL);
	read_unlock(&pfo_devices_lock);

	return 0;
}
EXPORT_SYMBOL(pfo_register_st);

void pfo_unregister_st(struct device_driver *drv)
{
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, remove_device_from_st, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);

	scsi_unregister_driver(drv);

	upper_device_callback.st_probe = NULL;
	upper_device_callback.st_remove = NULL;
	upper_device_callback.st_switch = NULL;
}
EXPORT_SYMBOL(pfo_unregister_st);

int pfo_register_sf(struct join_class_interface *intf)
{
	DBG_INFO("%p\n", intf);
	upper_device_callback.join_add = intf->join_add;
	upper_device_callback.join_remove = intf->join_remove;
	upper_device_callback.join_open = intf->join_open;
	upper_device_callback.join_close = intf->join_close;
	upper_device_callback.join_send = intf->join_send;

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, add_device_to_sf, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);
	/* Split up any rejected devices, must be done outside of the lock. */
	split_up_non_sfmp_multi_path_devices();
	/* If sg and st are already loaded then attach any pending devices. */
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, add_device_to_sg, &REMOVE_ALL_TRUE);
	idr_for_each(&pfo_devices, add_device_to_st, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);

	return 0;
}
EXPORT_SYMBOL(pfo_register_sf);

void pfo_unregister_sf(void)
{
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, remove_device_from_sf, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);

	upper_device_callback.join_add = NULL;
	upper_device_callback.join_remove = NULL;
	upper_device_callback.join_open = NULL;
	upper_device_callback.join_close = NULL;
	upper_device_callback.join_send = NULL;
}
EXPORT_SYMBOL(pfo_unregister_sf);

int pfo_register_sg(struct class_interface *intf,
		void (*path_switch_callback)(struct device *, struct device *))
{
	upper_device_callback.sg_add_dev = intf->add_dev;
	upper_device_callback.sg_remove_dev = intf->remove_dev;
	upper_device_callback.sg_switch = path_switch_callback;

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, add_device_to_sg, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);

	return 0;
}
EXPORT_SYMBOL(pfo_register_sg);

void pfo_unregister_sg(struct class_interface *intf)
{
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, remove_device_from_sg, &REMOVE_ALL_TRUE);
	read_unlock(&pfo_devices_lock);

	upper_device_callback.sg_add_dev = NULL;
	upper_device_callback.sg_remove_dev = NULL;
	upper_device_callback.sg_switch = NULL;
}
EXPORT_SYMBOL(pfo_unregister_sg);

/** The gateway function for all driver/device ctrl calls.
 * Intercept functions are filtered and handled here. The rest will be passed
 * to resolve_ctrl for further processing
 */
static void intercept_ctrl(struct pfo_device *pdev, const char *buf, size_t count)
{
	int handled = 0;
	if (pdev == NULL) {
		/* driver-scope ctrl command */
		if (strcmp(buf, "apfo off")  == 0) {
			DBG_INFO("apfo passthru activated");
			handled = 1;
			ctrl_apfo_on = 0;
		} else if (strcmp(buf, "apfo on")  == 0) {
			DBG_INFO("apfo passthru de-activated");
			handled = 1;
			ctrl_apfo_on = 1;
		}
	} else {
		/* device-scope ctrl command */
		if (strncmp(buf, "upper", 5) == 0) {
			handled = 1;
			print_hex_dump(KERN_ERR, "upper: ", DUMP_PREFIX_ADDRESS,
				16, 1, pdev->upper_dev,
				sizeof(*pdev->upper_dev), true);
		} else if (strncmp(buf, "kobj", 4) == 0) {
			handled = 1;
			DBG_DEBUG("upper->kobj.name = %s",
				kobject_name(&pdev->upper_dev->kobj));
			print_hex_dump(KERN_ERR, "upper->kobj: ",
				DUMP_PREFIX_ADDRESS,
				16, 1, &pdev->upper_dev->kobj,
				sizeof(pdev->upper_dev->kobj), true);
			DBG_DEBUG("upper->parent->kobj.name = %s",
				kobject_name(&pdev->upper_dev->parent->kobj));
			print_hex_dump(KERN_ERR, "upper->parent->kobj: ",
				DUMP_PREFIX_ADDRESS,
				16, 1, &pdev->upper_dev->parent->kobj,
				sizeof(pdev->upper_dev->parent->kobj), true);
		} else if (strncmp(buf, "pdev", 4) == 0) {
			handled = 1;
			dump_pdev(pdev);
		}
	}

	if (handled != 0)
		return;
	/* pass up to decide if the command can't be handled by intercept layer */
	if (pdev == NULL) {
		resolve_ctrl(-1, buf, count);
	} else if (pdev->fdev_num != -1) {
		resolve_ctrl(pdev->fdev_num, buf, count);
	} else {
		DBG_WARNING("failed to handle ctrl command\n");
	}
}
/** Receive notification from decide layer
 * when the given device chooses to use the given path,
 * pass in path_num = -1 means all paths are dead
 */
void intercept_notify_path_change(int fdev_num, int path_num)
{
	struct device_predicate cand;

	DBG_INFO("fdev_num:%d changed to use path=%d\n",
			fdev_num, path_num);

	cand.path_num = path_num;
	cand.fdev_num = -1;
	cand.scsidev = NULL;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.gendev = NULL;
	cand.serial[0] = '\0';

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);

	if (cand.pfonum != -1) {
		struct pfo_device *pdev = cand.pdev;
		DBG_INFO("dev pfo%d changed to use path=%d\n",
			cand.pfonum, path_num);
		pdev->activepath = cand.pathid;
		my_activepathchange(pdev, pdev->subpath[pdev->activepath].scsidev);
		return;
	}

	if (path_num != -1) {
		DBG_ERROR("bad switch call, can't find paths needed for the operations\n");
		return;
	}

	/* we won't find entry with path_num = -1,
	 * reinitiate the search for fdev_num */
	cand.fdev_num = fdev_num;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pdev = NULL;
	cand.gendev = NULL;
	cand.serial[0] = '\0';
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);

	if (cand.pfonum != -1) {
		struct pfo_device *pdev = cand.pdev;
		DBG_WARNING("dev pfo%d changed to use new NULL path\n",
			cand.pfonum);
		pdev->activepath = -1;
		my_activepathchange(pdev, NULL);
	}
}

static int create_device_files(struct pfo_device *pdev, struct device *dev)
{
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_serial)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_type)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_wwnn)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_vendor)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_model)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_rev)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_internal)) {
		goto out;
	}
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_paths)) {
		goto out;
	}
#ifdef APFO
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_sessions)) {
		goto out;
	}
#endif
	if (device_create_file(pdev->upper_dev->parent, &dev_attr_ctrl)) {
		goto out;
	}

	return 0;

out:
return 0;
	return -1;
}

static int cleanup_device(int id, void *p, void *delpfo)
{
	struct pfo_device *pdev = (struct pfo_device *)p;
	struct device *dummy_dev = pdev->upper_dev;
	struct scsi_device *dummy_sdev = to_scsi_device(pdev->upper_dev->parent);

	DBG_INFO("destroy pfo%d\n", id);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_serial);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_type);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_wwnn);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_vendor);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_model);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_rev);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_internal);
	device_remove_file(pdev->upper_dev->parent, &dev_attr_paths);
#ifdef APFO
	device_remove_file(pdev->upper_dev->parent, &dev_attr_sessions);
#endif
	device_remove_file(pdev->upper_dev->parent, &dev_attr_ctrl);
	/* free pfo_device */
	DBG_DEBUG("unregister pfo%d\n", id);
	device_unregister(pdev->upper_dev->parent);
	if (*((int *)delpfo) == DELETE_PFODEV) {
		/* free dummy sdev only if certain it is safe */
		DBG_DEBUG("free pfo%d memory\n", id);
		kfree(dummy_sdev);
		kfree(dummy_dev);
	}
	kfree(pdev);

	return 0;
}

static void pfo_release_device(struct device *dev)
{
	DBG_DEBUG("releasing %p", dev);
}

/** Clone a scsi_device struct from mid-layer and
 * then instantiate virtual device for pfo.
 *
 * Input:
 * The pfo device fdev num = -1 says that we know that
 * this device is never intended to have failover enabled.
 * Any other value implies that this is (or may become)
 * a path to a failover enabled device.
 *
 */
static struct device *create_dummy_dev(struct pfo_device *pdev, struct device *dev)
{
	struct scsi_device *org_sdev = to_scsi_device(dev->parent);
	struct scsi_device *sdev = NULL;
	struct device *dummy_dev = NULL;
	struct device *sysdev;
	int result = 0;

	/* clone scsi dev that's passed in from scsi mid layer */
	sdev = kzalloc(sizeof(*sdev), GFP_ATOMIC);
	if (!sdev) {
		DBG_ERROR("failed to allocate sdev\n");
		goto out;
	}
	memcpy(sdev, org_sdev, sizeof(*org_sdev));

	/* clone leaf dev */
	dummy_dev = kzalloc(sizeof(*dummy_dev), GFP_ATOMIC);
	if (!dummy_dev) {
		DBG_ERROR("failed to allocate dummy_dev\n");
		goto out;
	}
	memcpy(dummy_dev, dev, sizeof(*dev));

	/* set the device personality to be of a PFO device */
	memset(&sdev->sdev_gendev, 0, sizeof(*sysdev));
	sysdev = &sdev->sdev_gendev;
	sysdev->devt = MKDEV(PFO_MAJOR, pdev->pfonum);
	sysdev->class = pfo_sysfs_class;
	sysdev->release = pfo_release_device;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
	pdev->gen_dev = sysdev;
#else
	dev_set_drvdata(sysdev, pdev);
#endif
	sysdev->parent = NULL;
	/* If this is an APFO device then
	 * we need to create a private queue that's
	 * not tied to the queue created by scsi mid layer */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,11,0)
	if (pdev->fdev_num != -1) {
		/* Only for fo devices */
		/* This 'if' must match the one above in create unique */
		sdev->request_queue = __scsi_alloc_queue(sdev->host, NULL);
	}
#endif
	/* set device hierachy so the new dummy dev
	 * is hinged on our PFO device */
	dummy_dev->parent = sysdev;
	dummy_dev->p = NULL;
	dev_set_drvdata(dummy_dev, NULL);

	/* init my kobj, this looks clumsy here because we are trying
	 * to reinit an already inited kobj, this is not how it's done
	 * normally */
	dummy_dev->kobj.state_initialized = 0;
	kobject_init(&dummy_dev->kobj, dev->kobj.ktype);

	kobject_set_name(&sysdev->kobj, "pfo%d", pdev->pfonum);
	result = device_register(sysdev);
	if (result) {
		DBG_ERROR("Failed to register sdev for dev pfo%d\n",
							pdev->pfonum);
		goto out;
	}
	/* sysdev->kobj points to /sys/devices/virtual/pfo/pfo_ */
	return dummy_dev;
out:
	return NULL;
}

static struct pfo_device *intercept_create_pfo_struct(struct device *dev,
					struct class_interface *pclass)
{
	int k;
	struct pfo_device *pitem;
	struct scsi_device *sdev;

	/* Build out pfo device struct.  Must zero it 1st for kobjects. */
	pitem = kzalloc(sizeof(struct pfo_device), GFP_KERNEL);
	if (pitem == NULL) {
		return pitem;
	}
	sdev = to_scsi_device(dev->parent);
	pitem->scsitype = sdev->type;
	if (sdev->inquiry_len > 35) {
		memcpy(pitem->vendor, &(sdev->inquiry[8]), VENDOR_SIZE);
		pitem->vendor[VENDOR_SIZE] = '\0';
		memcpy(pitem->model, &(sdev->inquiry[16]), MODEL_SIZE);
		pitem->model[MODEL_SIZE] = '\0';
		memcpy(pitem->rev, &(sdev->inquiry[32]), REV_SIZE);
		pitem->rev[REV_SIZE] = '\0';
	} else {
		pitem->vendor[0] = '\0';
		pitem->model[0] = '\0';
		pitem->rev[0] = '\0';
	}
	pitem->activepath = 0;
	pitem->subpath[pitem->activepath].scsidev = dev;
	pitem->subpath[pitem->activepath].sfmp_registered = 0;
	pitem->sfmp_supported = 0;
	pitem->sfmp_determined = 0;
	pitem->sfmp_fo_determined = 0;
	pitem->sfmp_fo_enabled = 0;
	if (pitem->scsitype != 1 && pitem->scsitype != 8) {
		/* Already know that only tape and library
		 * might have failover enabled. */
		pitem->sfmp_determined = 1;
		pitem->sfmp_fo_determined = 1;
	}
	pitem->subpath[pitem->activepath].clsinterface = pclass;
	pitem->subpathcount = 1;
	pitem->sg_num = -1;
	pitem->st_num = -1;
	pitem->host = sdev->host->host_no;
	pitem->channel = sdev->channel;
	pitem->target = sdev->id;
	pitem->lun = sdev->lun;

	/* Initialize the whole structure for debugging. */
	for (k = 1; k < MAXSUBPATH; k++) {
		pitem->subpath[k].path_num = 9900 + k;
#if __SIZEOF_POINTER__ == 4
/* 32-bit */
		pitem->subpath[k].scsidev = (void *)(0xbadad000 + k);
#elif __SIZEOF_POINTER__ == 8
/* 64-bit */
		pitem->subpath[k].scsidev = (void *)(0xbadf0add000 + k);
#else
/* error */
#endif
		pitem->subpath[k].clsinterface = NULL;
		pitem->subpath[k].sfmp_registered = 0;
	}
	return pitem;
}

static int intercept_add(struct device *dev, struct class_interface *pclass)
{
	int error = 0;
	int path_num;
	int fdev_num;
	struct pfo_device *pitem;
	struct device_predicate cand;

	DBG_DEBUG("handle %p", dev);

	/* Create a pfo device structure. */
	pitem = intercept_create_pfo_struct(dev, pclass);
	if (pitem == NULL) {
		DBG_WARNING("pfo device allocation failed\n");
		error = -ENOMEM;
		goto out;
	}

	/* Create a path in lower for resolve to use. */
	path_num = lower_add_new_path(to_scsi_device(dev->parent));
	if (path_num == -1) {
		error = -ENOMEM;
		goto out;
	} else {
		pitem->subpath[pitem->activepath].path_num = path_num;
	}
	/* Add this path to the resolve data base.
	 * Learn wwnn and serial number.
	 * Resolve assumes every tape and library is failover,
	 * until told otherwise.
	 */
	error = resolve_add_new_path(path_num, &fdev_num,
					pitem->wwnn, pitem->serial);
	/* See what lin_tape thinks of it.
	 * Must do this after resolve add to make sure
	 * that the resolve database is ready to accept
	 * the failover decision call from lin_tape.
	 * Only bother to try it if resolve approves.
	 */
	if (error == PFO_NEW_APFO_DEVICE || error == PFO_OLD_APFO_DEVICE) {
		if (!pitem->sfmp_determined) {
			pitem->fdev_num = fdev_num;
			add_device_to_sf(-1, pitem, NULL);
		}
		if (pitem->sfmp_determined && !pitem->sfmp_supported) {
			DBG_INFO("Detected is not sfmp supported\n");
			resolve_set_is_not_sfmp_device(pitem->fdev_num);
			error = PFO_NOT_APFO_DEVICE;
		}
	}

	if (error == PFO_NEW_APFO_DEVICE) {
		/* continue to add to upper and sg/st later */
		pitem->fdev_num = fdev_num;
		pitem->upper_dev = NULL;
	} else if (error == PFO_OLD_APFO_DEVICE) {
		/* convert pfo_device into a subpath */
		cand.fdev_num = fdev_num;
		cand.path_num = -1;
		cand.scsidev = NULL;
		cand.pdev = NULL;
		cand.pfonum = -1;
		cand.gendev = NULL;
		cand.serial[0] = '\0';

		read_lock(&pfo_devices_lock);
		idr_for_each(&pfo_devices, find_device, &cand);
		read_unlock(&pfo_devices_lock);

		if (cand.pfonum != -1) {
			/* good old apfo device is found,
			 * now link this as a subpath. */
			struct pfo_device *p = cand.pdev;
			DBG_INFO("add subpath %d to dev pfo%d\n",
					p->subpathcount, cand.pfonum);
			spin_lock(&pfo_subpath_lock);
			p->subpath[p->subpathcount].scsidev = dev;
			p->subpath[p->subpathcount].path_num = path_num;
			p->subpath[p->subpathcount].clsinterface = pclass;
			p->subpath[p->subpathcount].sfmp_registered = 0;
			if (pitem->sfmp_determined && pitem->sfmp_supported) {
				p->subpath[p->subpathcount].sfmp_registered = 1;
			}
			p->subpathcount++;
			spin_unlock(&pfo_subpath_lock);
			if (p->sfmp_fo_determined && !p->sfmp_fo_enabled) {
				/* We already know that this is not a failover
				 * device because it has been previously
				 * reported by lin_tape as such.
				 */
				DBG_INFO("Detected is not fo enabled\n");
				resolve_not_fo_device(p, p->fdev_num);
				error = 0;
				goto out;
			}

			if (p->subpathcount == 1) {
				/* Had no paths. Now have one.
				 * Make new one the active one. */
				my_activepathchange(p, dev);
			}

			/* we are done here, out will free the pitem */
			error = 0;
			DBG_INFO("path=%d added to dev pfo%d\n",
					path_num, cand.pfonum);
			goto out;
		} else {
			DBG_ERROR("decide says known apfo device"
			" but unable to locate in our table fdev_num=%d\n",
			fdev_num);
			resolve_remove_path(path_num);
			lower_remove_path(path_num);
			error = -ENOMEM;
			goto out;
		}
	} else if (error == PFO_NOT_APFO_DEVICE) {
		print_log("Not an APFO device.\n");
		/* save wwnn */
		error = resolve_get_wwnn_from_path(path_num, pitem->wwnn);
		/* Must always remove the path
		 * after attempting the Inquiry command */
		lower_remove_path(path_num);

		/* apfo will not handle this.
		 * This device will be handled
		 * by intercept with a pfo_struct. */
		pitem->fdev_num = -1;
		pitem->subpath[pitem->activepath].path_num = -1;

		if (error != 0) {
			/* Must support devices that
			 * do not support Fibre Channel */
			DBG_INFO("failed to obtain wwnn, treat as non-APFO\n");
		} else {
			/* There's a possibility that we have seen
			 * this device as an APFO dev before,
			 * thus we need to search our list to clear
			 * that old APFO dev out if that's the case.
			 */
			cand.fdev_num = -1;
			cand.path_num = -1;
			cand.scsidev = NULL;
			cand.pfonum = -1;
			cand.gendev = NULL;
			cand.serial[0] = '\0';
			memcpy(cand.wwnn, pitem->wwnn, sizeof(pitem->wwnn));

			read_lock(&pfo_devices_lock);
			idr_for_each(&pfo_devices, find_device, &cand);
			read_unlock(&pfo_devices_lock);

			if (cand.pfonum != -1 && cand.pdev->fdev_num != -1) {
				/* Old apfo device is found,
				 * remove it first, the new non-apfo device
				 * will be added later. */
				DBG_INFO("Remove old APFO dev pfo%d\n",
							cand.pfonum);
				remove_device_from_sg(cand.pfonum, cand.pdev,
							&REMOVE_ALL_TRUE);
				remove_device_from_st(cand.pfonum, cand.pdev,
							&REMOVE_ALL_TRUE);
				remove_device_from_sf(cand.pfonum, cand.pdev,
							&REMOVE_ALL_TRUE);
				write_lock(&pfo_devices_lock);
				idr_remove(&pfo_devices, cand.pfonum);
				write_unlock(&pfo_devices_lock);
				/* Retain structures in case still open by user.
				 * Yes, this results in a memory leak,
				 * but only on apfo->non-apfo transition. */
				cleanup_device(cand.pfonum, cand.pdev,
								&RETAIN_PFODEV);
			}
		}
	} else if (error == PFO_PATH_DOWN) {
		DBG_ERROR("decide can't probe this device\n");
		lower_remove_path(path_num);
		error = -ENOMEM;
		goto out;
	} else {
		DBG_ERROR("Fatal error return fdev_num=%d\n" , fdev_num);
		lower_remove_path(path_num);
		error = -ENOMEM;
		goto out;
	}
	return intercept_create_unique_device(pitem, dev);
out:
	if (error != 0)
		print_log("Unable to use that path.\n");
	if (pitem != NULL)
		kfree(pitem);
	return error;
}

/** Build structures for a single device with one path.
 *
 * Input:
 * The pfo device fdev num = -1 says that we know that
 * this device is never intended to have failover enabled.
 * Any other value implies that this is (or may become)
 * a path to a failover enabled device.
 */
static int intercept_create_unique_device(struct pfo_device *pitem,
			struct device *dev)
{
	int error;
	int pfonum;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	idr_preload(GFP_KERNEL);
#else
	/* We have a unique device now, common to APFO and non-APFO */
	if (!idr_pre_get(&pfo_devices, GFP_KERNEL)) {
		DBG_WARNING("idr expansion for pfo failed\n");
		error = -ENOMEM;
		goto out;
	}
#endif
	write_lock(&pfo_devices_lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	error = idr_alloc(&pfo_devices, pitem, 0, MAXPFO, GFP_NOWAIT);
	write_unlock(&pfo_devices_lock);
	if (error < 0) {
		if (error == -ENOSPC) {
			error = -ENODEV;
		} else {
			printk(KERN_WARNING
			       "idr allocation pfo_device failure: %d\n", error);
		}
		idr_preload_end();
		goto out;
	}
	pfonum = error;
#else
	error = idr_get_new(&pfo_devices, pitem, &pfonum);
	write_unlock(&pfo_devices_lock);
	if (error) {
		DBG_WARNING("idr allocation pfo failed: %d\n", error);
		error = -ENOMEM;
		goto out;
	}
#endif
	if (pfonum > max_pfonum) {
		max_pfonum = pfonum;
	}
	pitem->pfonum = pfonum;

	/* populate dummy device */
	pitem->upper_dev = create_dummy_dev(pitem, dev);
	if (pitem->upper_dev == NULL) {
		error = -ENOMEM;
		goto out_remove_idr;
	}
	if (pitem->fdev_num != -1) {
		/* Only for fo devices */
		/* This 'if' must match the one in create dummy */
		upper_add_new_device(to_scsi_device(pitem->upper_dev->parent),
					pitem->fdev_num);
	}
	error = create_device_files(pitem, dev);
	if (error) {
		error = -ENOMEM;
		goto out_remove_device_files;
	}
	DBG_INFO("Created dev pfo%d\n", pfonum);
	if (pitem->fdev_num != -1) {
		DBG_INFO("fdev=%d added to dev pfo%d\n",
				pitem->fdev_num, pfonum);
	}

	add_device_to_sg(pfonum, pitem, NULL);
	add_device_to_st(pfonum, pitem, NULL);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
		idr_preload_end();
#endif

	return 0;

out_remove_device_files:
	cleanup_device(pitem->pfonum, pitem, &DELETE_PFODEV);

out_remove_idr:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
    idr_preload_end();
#else
	write_lock(&pfo_devices_lock);
	idr_remove(&pfo_devices, pitem->pfonum);
	write_unlock(&pfo_devices_lock);
#endif

out:
	if (error != 0)
		print_log("Unable to use that path.\n");
	if (pitem != NULL)
		kfree(pitem);
	return error;
}

/** Remove a subpath of an APFO device.  The caller
 * guarantees that it only calls this on an APFO device. */
static void remove_subpath(struct pfo_device *pdev, int path)
{
	int index;
	int path_changed = 0;
	struct device *newpath = NULL;

	/* Adjust the active path number to new spot, if needed. */
	spin_lock(&pfo_subpath_lock);
	if (pdev->activepath > path) {
		/* Adjust because of shift done below. */
		pdev->activepath--;
	} else if (pdev->subpathcount == 1) {
		/* Lost the last path. */
		path_changed = 1;
		pdev->activepath = -1;
	} else if (pdev->activepath == path) {
		/* Lost active path, so reset active path to the first path */
		path_changed = 1;
		pdev->activepath = 0;
	}

	/* shift the subpath array to fill the hole */
	pdev->subpathcount--;
	for (index = path; index < MAXSUBPATH-1; index++) {
		pdev->subpath[index].scsidev = pdev->subpath[index+1].scsidev;
		pdev->subpath[index].path_num = pdev->subpath[index+1].path_num;
		pdev->subpath[index].clsinterface =
					pdev->subpath[index+1].clsinterface;
		pdev->subpath[index].sfmp_registered =
					pdev->subpath[index+1].sfmp_registered;
	}
	spin_unlock(&pfo_subpath_lock);

	if (path_changed && pdev->subpathcount != 0) {
		DBG_DEBUG("new activesubpath=%d\n", pdev->activepath);
		newpath = pdev->subpath[pdev->activepath].scsidev;
		my_activepathchange(pdev, newpath);
	} else if (path_changed) {
		DBG_DEBUG("no new path is provided, just lost the last path\n");
		my_activepathchange(pdev, NULL);
	}
}

static void intercept_remove(struct device *dev, struct class_interface *pclass)
{
	int is_failover_path;
	struct device_predicate cand;
	DBG_DEBUG("handle %p", dev);

	/* If we have a pfo path gone, just need to remove that path.
	 * If it's not a pfo device,
	 * then we need to remove the device and notify st/sg too. */
	cand.fdev_num = -1;
	cand.path_num = -1;
	cand.scsidev = dev;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.gendev = NULL;
	cand.serial[0] = '\0';

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);

	if (cand.pfonum != -1) {
		struct pfo_device *p = cand.pdev;
		int path_num = p->subpath[cand.pathid].path_num;

		DBG_INFO("subpath[pathid].path_num=%d",path_num);
		DBG_INFO("scsidev=%p, pdev=%p\n", cand.scsidev, cand.pdev);
		DBG_INFO("cand.fdev=%d, path=%d, pfo%d, pathid=%d\n",
			cand.fdev_num, cand.path_num, cand.pfonum, cand.pathid);
		DBG_INFO("p->subpath[cand.pathid].scsidev=%p\n",
					p->subpath[cand.pathid].scsidev);
		/* Determine whether to treat this path as
		 * failover or non-failover.
		 * NOTE: The following logic assumes no APFO devices.
		 * Yes, these if-else statements are tacky,
		 * but I wrote it this way so I could doucment
		 * each decision.
		 */
		if (path_num < 0) {
			/* Path not found in search,
			 * this is definitely not failover,
			 * so remove all.
			 */
			is_failover_path = 0;
		} else
		if (!p->sfmp_determined) {
			/* lin_tape not loaded,
			 * No /dev files have been created.
			 * Code is treating this as a
			 * failover device at this time
			 * so remove just the one path.
			 */
			is_failover_path = 1;
		} else
		if (!p->sfmp_supported) {
			/* Not a lin_tape device, so remove all.  */
			is_failover_path = 0;
		} else
		if (!p->sfmp_fo_determined) {
			/* lin_tape has not responded,
			 * lin_taped may be stopped,
			 * Code is treating this as a
			 * failover device at this time
			 * so remove just the one path.
			 */
			is_failover_path = 1;
		} else
		if (!p->sfmp_fo_enabled) {
			/* lin_tape has responded,
			 * this is definitely not failover,
			 * so remove all.
			 */
			is_failover_path = 0;
		} else {
			/* lin_tape has responded,
			 * this is definitely a failover path,
			 * so remove only this one path.
			 */
			is_failover_path = 1;
		}
		if (is_failover_path) {
			/* This is a valid failover path. */
			DBG_INFO("Remove path=%d from dev pfo%d\n",
				path_num, cand.pfonum);
			resolve_remove_path(path_num);
			lower_remove_path(path_num);
			/* Tell sfmp that this path is gone. */
			if (p->subpath[cand.pathid].sfmp_registered &&
			    upper_device_callback.join_remove) {
				upper_device_callback.join_remove(
					p->subpath[cand.pathid].scsidev,
					p->subpath[cand.pathid].clsinterface);
			}
			remove_subpath(p, cand.pathid);
		} else {
			/* This is a non-failover device. */
			DBG_INFO("Remove non-fo dev pfo%d\n", cand.pfonum);
			/* detach from high level driver first */
			remove_device_from_sg(p->pfonum, p, &REMOVE_ALL_TRUE);
			remove_device_from_st(p->pfonum, p, &REMOVE_ALL_TRUE);
			remove_device_from_sf(p->pfonum, p, &REMOVE_ALL_TRUE);
			/* Tell resolve to remove it from the data base. */
			resolve_remove_non_fo_device(p->fdev_num);
			/* remove from pfo lists */
			write_lock(&pfo_devices_lock);
			idr_remove(&pfo_devices, p->pfonum);
			write_unlock(&pfo_devices_lock);
			cleanup_device(p->pfonum, p, &DELETE_PFODEV);
		}
	} else {
		DBG_ERROR("no device/path to remove\n");
	}
}

static void init_callbacks(void)
{
	upper_device_callback.sg_add_dev = NULL;
	upper_device_callback.sg_remove_dev = NULL;
	upper_device_callback.sg_switch = NULL;
	upper_device_callback.st_probe = NULL;
	upper_device_callback.st_remove = NULL;
	upper_device_callback.st_switch = NULL;
	upper_device_callback.join_add = NULL;
	upper_device_callback.join_remove = NULL;
	upper_device_callback.join_open = NULL;
	upper_device_callback.join_close = NULL;
	upper_device_callback.join_send = NULL;
}

static int __init init_pfo(void)
{
	int rc = 0;
	dev_t dev;

	if (netbackup) {
		print_log("Driver installed with netbackup mode enabled.\n");
	} else {
		print_log("Driver installed.\n");
	}
	DBG_INFO("Version dates u=%s r=%s i=%s l=%s\n",
		upper_version_date,
		resolve_version_date,
		intercept_version_date,
		lower_version_date);

#if !(defined(SFMPENABLED))
	return 0;
#endif

	init_callbacks();
	pfo_sysfs_class = class_create(THIS_MODULE, "pfo");
	if (IS_ERR(pfo_sysfs_class)) {
		DBG_ERROR("pfo class registration failed\n");
		goto init_pfo_class_create_errout;
	}
	rc = scsi_register_driver(&pfo_template.gendrv);
	if (rc != 0) {
		DBG_ERROR("pfo scsi_register_driver failed\n");
		goto init_pfo_scsi_register_driver_errout;
	}

	rc = init_debug(&pfo_template.gendrv);
	if (rc != 0) {
		DBG_ERROR("pfo sysfs registration failed\n");
		goto init_pfo_create_driver_errout;
	}
	rc = driver_create_file(&pfo_template.gendrv, &driver_attr_ctrl);
	if (rc != 0) {
		DBG_ERROR("failed to create ctrl file\n");
		goto init_pfo_ioctol_errout;
	}

	rc = driver_create_file(&pfo_template.gendrv, &driver_attr_pfo_version);
	if (rc != 0) {
		DBG_ERROR("failed to create version file\n");
		goto init_pfo_ioctol_errout;
	}
	rc = alloc_chrdev_region(&dev, 0, 32768, "pfo");
	if (rc != 0) {
		DBG_ERROR("pfo device major allocation failed\n");
		goto init_pfo_alloc_chrdev_errout;
	} else {
		PFO_MAJOR = MAJOR(dev);
		DBG_INFO("pfo device will be using major %d\n", PFO_MAJOR);
	}

	rc = scsi_register_interface(&pfo_interface);
	if (rc != 0) {
		DBG_ERROR("pfo interface registration failed\n");
		goto init_pfo_register_interface_errout;
	}

	return 0;
	/*
	 * There is no need to deal with adding
	 * devices here. They will get added as
	 * needed during normal operation.
	 */

init_pfo_register_interface_errout:
	unregister_chrdev_region(MKDEV(PFO_MAJOR, 0), 32768);


init_pfo_alloc_chrdev_errout:
	driver_remove_file(&pfo_template.gendrv, &driver_attr_ctrl);
	driver_remove_file(&pfo_template.gendrv, &driver_attr_pfo_version);
init_pfo_ioctol_errout:
	stop_debug(&pfo_template.gendrv);
init_pfo_create_driver_errout:
	scsi_unregister_driver(&pfo_template.gendrv);

init_pfo_scsi_register_driver_errout:
	class_destroy(pfo_sysfs_class);

init_pfo_class_create_errout:
	return rc;

}


/*###############################################################*/
/* The sfmp translation functions.                               */
/*###############################################################*/
int intercept_sfmp_open(int fdev_num, struct file *filp)
{
	int sfmp_result;
	struct pfo_device *pdev;
	struct device_predicate cand;
	struct join_device_identifier gizmo;

	if (upper_device_callback.join_open == NULL ) {
		DBG_ERROR("lin_tape not attached to pfo\n");
		return -ENODEV;
	}
	if (fdev_num < 0) {
		DBG_ERROR("Invalid fdev=%d\n", fdev_num);
		return -ENODEV;
	}
	/* Convert the decide fdev_num to our pdev */
	cand.fdev_num = fdev_num;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pdev = NULL;
	cand.serial[0] = '\0';
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);
	pdev = cand.pdev;
	if (pdev == NULL) {
		DBG_ERROR("pdev==NULL\n");
		return -ENODEV;
	}
	/* Fill the gizmo device structure. */
	gizmo.fdev_num = fdev_num;
	gizmo.dev_type = pdev->scsitype;
	gizmo.host = pdev->host;
	gizmo.channel = pdev->channel;
	gizmo.target = pdev->target;
	gizmo.lun = pdev->lun;
	gizmo.filp = filp;
	gizmo.serial = pdev->serial;
	gizmo.wwnn = pdev->wwnn;

	DBG_INFO("fdev=%d type=%d %d:%d:%d:%d %p %s\n",
		gizmo.fdev_num, gizmo.dev_type,
		gizmo.host, gizmo.channel, gizmo.target, gizmo.lun,
		gizmo.filp, gizmo.serial);

	sfmp_result = upper_device_callback.join_open(&gizmo);
	return sfmp_result;
}

int intercept_sfmp_close(int fdev_num, struct file *filp)
{
	int sfmp_result;
	struct pfo_device *pdev;
	struct device_predicate cand;
	struct join_device_identifier gizmo;

	if (upper_device_callback.join_close == NULL ) {
		return -ENODEV;
	}
	if (fdev_num < 0) {
		DBG_ERROR("Invalid fdev=%d\n", fdev_num);
		return -ENODEV;
	}
	/* Convert the decide fdev_num to our pdev */
	cand.fdev_num = fdev_num;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pdev = NULL;
	cand.serial[0] = '\0';
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);
	pdev = cand.pdev;
	if (pdev == NULL) {
		DBG_ERROR("pdev==NULL\n");
		return -ENODEV;
	}
	/* Fill the gizmo device structure. */
	gizmo.fdev_num = fdev_num;
	gizmo.dev_type = pdev->scsitype;
	gizmo.host = pdev->host;
	gizmo.channel = pdev->channel;
	gizmo.target = pdev->target;
	gizmo.lun = pdev->lun;
	gizmo.filp = filp;
	gizmo.serial = pdev->serial;
	gizmo.wwnn = pdev->wwnn;

	DBG_INFO("fdev=%d type=%d %d:%d:%d:%d %p %s\n",
		gizmo.fdev_num, gizmo.dev_type,
		gizmo.host, gizmo.channel, gizmo.target, gizmo.lun,
		gizmo.filp, gizmo.serial);

	sfmp_result = upper_device_callback.join_close(&gizmo);
	return sfmp_result;
}

/** Send one SCSI command via the sfmp driver.
 * Makes the call to the other driver.
 * buffer is always in kernel memory.
 * @return 0=pass, else standard Linux error codes.
 */
int intercept_sfmp_send(int fdev_num,
	unchar *cdb,
	unchar cmd_length,
	unchar *buffer,
	uint buffer_length,
	uint data_direction,
	uint timeout,
	unchar *sense,
	int *sense_length,
	int *resid,
	int *result)
{
	int sfmp_result;
	struct pfo_device *pdev;
	struct device_predicate cand;
	struct join_device_identifier gizmo;

	if (upper_device_callback.join_send == NULL ) {
		return -ENODEV;
	}
	if (fdev_num < 0) {
		DBG_ERROR("Invalid fdev=%d\n", fdev_num);
		return -ENODEV;
	}
	/* Convert the decide fdev_num to our pdev */
	cand.fdev_num = fdev_num;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pdev = NULL;
	cand.serial[0] = '\0';
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);
	pdev = cand.pdev;
	if (pdev == NULL) {
		DBG_ERROR("pdev==NULL\n");
		return -ENODEV;
	}
	/* Fill the gizmo device structure. */
	gizmo.fdev_num = fdev_num;
	gizmo.dev_type = pdev->scsitype;
	gizmo.host = pdev->host;
	gizmo.channel = pdev->channel;
	gizmo.target = pdev->target;
	gizmo.lun = pdev->lun;
	gizmo.filp = NULL;
	gizmo.serial = pdev->serial;
	gizmo.wwnn = pdev->wwnn;

	DBG_INFO("fdev=%d type=%d %d:%d:%d:%d %p %s\n",
		gizmo.fdev_num, gizmo.dev_type,
		gizmo.host, gizmo.channel, gizmo.target, gizmo.lun,
		gizmo.filp, gizmo.serial);

	sfmp_result = upper_device_callback.join_send(&gizmo,
		cdb,
		cmd_length,
		buffer,
		buffer_length,
		data_direction,
		timeout,
		sense,
		sense_length,
		resid,
		result);
	return sfmp_result;
}

/** lin_tape has told us that this is not a failover device,
 * either by rejecting support or by the failover enabled call-back,
 * so break this one path out into an individual device.
 * Create the related /dev files, as needed.
 * @return 0=pass
 */
int intercept_not_fo_device(void *pdev_in, int fdev_num, int path_num)
{
	int path_index;
	int registered;
	struct device *dev;
	struct pfo_device *pdev_new;
	struct class_interface *pclass;
	struct pfo_device *pdev_old = pdev_in;

	/* Locks are not needed because this struct
	 * is not visible to other code, yet. */
	DBG_INFO("Enter pdev=%p fdev=%d path=%d\n",
				pdev_old, fdev_num, path_num);

	/* Which sub-path of the old device. */
	for (path_index = 0; path_index < pdev_old->subpathcount; path_index++) {
		if (pdev_old->subpath[path_index].path_num == path_num) {
			goto path_found;
		}
	}
	/* No cleanup is needed here because nothing was changed. */
	print_log("Device path not found.\n");
	goto error_out;

path_found:
	/* Create a pfo device structure. */
	dev = pdev_old->subpath[path_index].scsidev;
	pclass = pdev_old->subpath[path_index].clsinterface;
	registered = pdev_old->subpath[path_index].sfmp_registered;
	pdev_new = intercept_create_pfo_struct(dev, pclass);
	if (pdev_new == NULL) {
		/* No cleanup is needed here because nothing was changed. */
		print_log("Device allocation failed.\n");
		goto error_out;
	}
	/* Delete the sub path since it has been moved to its own device. */
	remove_subpath(pdev_old, path_index);
	/* Fill in the newly created device structure. */
	memcpy(pdev_new->wwnn, pdev_old->wwnn, sizeof(pdev_new->wwnn));
	memcpy(pdev_new->serial, pdev_old->serial, SN_SIZE);
	pdev_new->subpath[0].path_num = path_num;
	pdev_new->subpath[0].scsidev = dev;
	pdev_new->subpath[0].clsinterface = pclass;
	pdev_new->subpath[0].sfmp_registered = registered;
	pdev_new->sfmp_supported = pdev_old->sfmp_supported;
	pdev_new->sfmp_fo_enabled = 0;
	pdev_new->sfmp_fo_determined = 1;
	pdev_new->sfmp_determined = pdev_old->sfmp_determined;
	if (pdev_new->sfmp_determined && pdev_new->sfmp_supported) {
		pdev_new->fdev_num = fdev_num; /* lin_tape supported device */
	} else {
		pdev_new->fdev_num = -1; /* Not lin_tape supported device */
	}

	/* Create the pfo, sg, and st /dev files. */
	if (intercept_create_unique_device(pdev_new, dev)) {
		/* No cleanup is needed because no /dev file was created. */
		print_log("Device creation failed.\n");
		goto error_out;
	}
	return 0;

error_out:
	/* The /dev files are left in a bad state. */
	print_log("Creation of a non-failover device failed.\n");
	DBG_ERROR("FAILED fdev=%d path=%d\n", fdev_num, path_num);
	return 1;
}

/** See if this device is indicated by this identifier.
 * return =0 if a match.
 */
static int ident_not_match(unsigned char *identifier, struct pfo_device *pdev)
{
	if (identifier[1] == sizeof(pdev->wwnn) &&
	    memcmp(&identifier[2], pdev->wwnn, sizeof(pdev->wwnn)) == 0) {
		return 0;
	}
	if (identifier[1] < SN_SIZE &&
	    strncmp(&identifier[2], pdev->serial, identifier[1]) == 0) {
		return 0;
	}
	return 1;
}


/** Inform the pfo module whether an sfmp device has failover enabled.
 * @param[in] identifier might be wwnn or serial number of SN_SIZE.
 * @param[in] enabled =0 if not a failover device.
 */
void pfo_failover_device(unsigned char *identifier, unsigned char enabled)
{
	struct pfo_device *pdev;
	struct device_predicate cand;

	/* Convert the wwnn to our pdev */
	cand.fdev_num = -1;
	cand.path_num = -1;
	cand.gendev = NULL;
	cand.scsidev = NULL;
	cand.pfonum = -1;
	cand.pdev = NULL;
	memcpy(cand.wwnn, &identifier[2], identifier[1]);
	memcpy(cand.serial, &identifier[2], identifier[1]);
	cand.serial[identifier[1]] = '\0';
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);
	pdev = cand.pdev;
	if (pdev == NULL) {
		DBG_ERROR("ident=%02x%02x%02x%02x%02x%02x%02x%02x\n",
			cand.wwnn[0], cand.wwnn[1], cand.wwnn[2], cand.wwnn[3],
			cand.wwnn[4], cand.wwnn[5], cand.wwnn[6], cand.wwnn[7]);
		return;
	}
	if (pdev->fdev_num == -1 || pdev->pfonum == -1) {
		DBG_ERROR("pdev=%p fdev_num=%d pfo%d\n",
				pdev, pdev->fdev_num, pdev->pfonum);
		return;
	}
	if (enabled) {
		/* This device has sfmp enabled.
		 * This has been the assumption
		 * up until now, so not much to do.
		 */
		if (!pdev->sfmp_fo_determined) {
			/* This routine gets called for each path,
			 * but we only want to create one /dev file per device.
			 */
			DBG_INFO("FO enabled pfo%d\n", pdev->pfonum);
			pdev->sfmp_fo_enabled = 1;
			pdev->sfmp_fo_determined = 1;
			/* Create the /dev files now. */
			add_device_to_sg(pdev->pfonum, pdev, NULL);
			add_device_to_st(pdev->pfonum, pdev, NULL);
		} else {
			DBG_INFO("FO enabled addl path pfo%d\n", pdev->pfonum);
		}
	} else {
		/* This device has the failover setting disabled.
		 * But we still want to send SCSI commands to lin_tape.
		 * We need to adjust every path to the device
		 * described by this identifier.
		 */
		int next_id;
		for (next_id = 0; next_id <= max_pfonum; next_id++) {
			read_lock(&pfo_devices_lock);
			pdev = idr_find(&pfo_devices, next_id);
			read_unlock(&pfo_devices_lock);
			if (!pdev) {
				break; /* Done. */
			}
			if (ident_not_match(identifier, pdev)) {
				continue; /* Not this device. */
			}
			DBG_INFO("FO disabled pfo%d\n", pdev->pfonum);
			pdev->sfmp_fo_enabled = 0;
			pdev->sfmp_fo_determined = 1;
			/* Break all of the sub-paths to this device into
			 * separate devices and create the /dev files.
			 * Resolve calls back into intercept to create
			 * new devices and the /dev files as needed.
			 */
			resolve_not_fo_device(pdev, pdev->fdev_num);
			/* Now create the /dev files for the base device. */
			add_device_to_sg(pdev->pfonum, pdev, NULL);
			add_device_to_st(pdev->pfonum, pdev, NULL);
		}
	}
}
EXPORT_SYMBOL(pfo_failover_device);

void pfo_path_change(int fdev_num, int host, int channel, int target, int lun)
{
	int index;
	struct pfo_device *pdev;
	struct scsi_device *sdev;
	struct device_predicate cand;

	/* If we have a pfo path, just need to switch to that path. */
	cand.fdev_num = fdev_num;
	cand.pfonum = -1;
	cand.pathid = -1;
	cand.path_num = -1;
	cand.scsidev = NULL;
	cand.pdev = NULL;
	cand.serial[0] = '\0';

	DBG_INFO("fdev=%d %d:%d:%d:%d\n",
		fdev_num, host, channel, target, lun);

	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, find_device, &cand);
	read_unlock(&pfo_devices_lock);

	pdev = cand.pdev;
	if (pdev == NULL) {
		return;
	}

	/* Look at each path to see which one matches. */
	for (index = 0; index < pdev->subpathcount; index++) {
		sdev = to_scsi_device(pdev->subpath[index].scsidev->parent);
		if (sdev == NULL) {
			continue;
		}
		if (channel != sdev->channel) {
			continue;
		}
		if (target != sdev->id) {
			continue;
		}
		if (lun != sdev->lun) {
			continue;
		}
		if (sdev->host == NULL) {
			continue;
		}
		if (host != sdev->host->host_no) {
			continue;
		}
		/* Found it */
		/* Tell human about path status. */
		resolve_set_last_path(fdev_num,
				pdev->subpath[index].path_num);
		if (index != pdev->activepath) {
			/* Not same as last time. */
			DBG_INFO("fdev=%d path=%d", fdev_num,
					pdev->subpath[index].path_num);
			DBG_INFO("%d:%d:%d:%d\n", host, channel, target, lun);
			/* Tell sysfs about path status. */
			intercept_notify_path_change(fdev_num,
					pdev->subpath[index].path_num);
			return;
		}
	}
	/* We did not find it, so assume it is actually gone.
	 * Currently, just leave things alone and hope that is good enough.
	 * Attempt to select a valid alternative.
	 */
}
EXPORT_SYMBOL(pfo_path_change);

static void intercept_exit(void)
{
	DBG_INFO("now cleanup the APFO device residuals");
	read_lock(&pfo_devices_lock);
	idr_for_each(&pfo_devices, cleanup_device, &DELETE_PFODEV);
	read_unlock(&pfo_devices_lock);

	return;
}

static void __exit exit_pfo(void)
{
#if (!defined(SFMPENABLED))
	return;
#endif

	scsi_unregister_interface(&pfo_interface);
	upper_exit();
	resolve_exit();
	lower_exit();
	intercept_exit();

	unregister_chrdev_region(MKDEV(PFO_MAJOR, 0), 32768);

	DBG_INFO("Remove driver\n");
	stop_debug(&pfo_template.gendrv);
	driver_remove_file(&pfo_template.gendrv, &driver_attr_ctrl);
	driver_remove_file(&pfo_template.gendrv, &driver_attr_pfo_version);
	scsi_unregister_driver(&pfo_template.gendrv);
	class_destroy(pfo_sysfs_class);
}

module_init(init_pfo);
module_exit(exit_pfo);

MODULE_AUTHOR("Mike Yoknis");
MODULE_DESCRIPTION("SCSI Path Failover (pfo) driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(PFO_VERSION_STR);
MODULE_INFO(date, PFO_DATE_STR);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
module_param(netbackup, int, 0444);
MODULE_PARM_DESC(netbackup, "!=0 to use with NetBackup application");
#endif
module_param(pfo_debug_flag, int, 0644);
MODULE_PARM_DESC(pfo_debug_flag, "=0x1111 for increased verbosity, =0xffff for maximum");
#ifdef APFO
module_param(pfo_balance, int, 0644);
MODULE_PARM_DESC(pfo_balance, "=0 no load balancing, =1 limited, =2 full");
#endif
