/**
 * This file contains the code that
 * interfaces beween the APFO code and the LTO7+ failover code.
 *
 */

/*
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 * (C) Copyright 2016 Hewlett Packard Enterprise Development LP
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

char *join_version_date = "20170113A";

#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <scsi/scsi_device.h>
#include "lin_tape_tape.h"
#include "apfo.h"
#include "pfo.h"
#include "convert.h"

static struct class_interface join_class_intf;

/* Generic return codes. */
#define PFO_GOOD 0
#define PFO_BAD (-1)

/* printk stuff */
#define JOIN_PRINTK(fmt, args...)
#define JOIN_PRINT_WWNN(wwnn, fmt, args...)

#define TO_MDATA 0
#define FROM_MDATA 1

#define INVALID_MINOR (-1)
static int minor_number[MAX_FDEV_NUMBER];

/*###############################################################*/
/* Helper Functions.						 */
/*###############################################################*/
static void join_switch_path(int fdev_num, struct pfo_ID *pfo_id)
{
	pfo_path_change(fdev_num,
		pfo_id->hctl.host,
		pfo_id->hctl.channel,
		pfo_id->hctl.target,
		pfo_id->hctl.lun);
}

/*###############################################################*/
/* Functions called by the pfo module.				 */
/*###############################################################*/
/** An open() call has been made to the given device.
 * Not re-entrant per device, to avoid two simultaneous open attempts.
 * @param[in] gizmo describes the device to access
 * @return PFO_GOOD or standard Linux return error code (e.g. -ENXIO).
 */
static int join_open(struct join_device_identifier *gizmo)
{
	struct pfo_ID pfo_id;
	int result = -EMEDIUMTYPE;

	JOIN_PRINT_WWNN(gizmo->wwnn, "fdev=%d type=%d %d:%d:%d:%d %s\n",
		gizmo->fdev_num, gizmo->dev_type,
		gizmo->host, gizmo->channel, gizmo->target, gizmo->lun,
		gizmo->serial);

	if (minor_number[gizmo->fdev_num] != INVALID_MINOR) {
		JOIN_PRINTK("already open\n");
		return -EEXIST;
	}
	pfo_id.hctl.host = gizmo->host;
	pfo_id.hctl.channel = gizmo->channel;
	pfo_id.hctl.target = gizmo->target;
	pfo_id.hctl.lun = gizmo->lun;
	pfo_id.f_flags = gizmo->filp->f_flags;
	pfo_id.serial_num_len = strlen(gizmo->serial);
	memcpy(&pfo_id.wwnn, gizmo->wwnn, WWN_LEN);
	memcpy(&pfo_id.serial_num, gizmo->serial, MaxSerialNumberSize);
	if (gizmo->dev_type == 1) {
		/* Do not fail the open if there is no tape in the drive. */
		pfo_id.f_flags = pfo_id.f_flags | O_NONBLOCK;
		/* Do not do append, let the st driver decide what to do. */
		pfo_id.f_flags = pfo_id.f_flags & (~O_APPEND);
		result = pfo_open_drive(&pfo_id);
	} else if (gizmo->dev_type == 8) {
		result = pfo_open_changer(&pfo_id);
	}
	if (result) {
		/* Fail */
		printk("pfo: open FAILED %s result=%d\n",
					gizmo->serial, result);
	} else {
		/* Pass */
		minor_number[gizmo->fdev_num] = pfo_id.pfo_minor;
		JOIN_PRINTK("open passed minor=%d\n", minor_number[gizmo->fdev_num]);
		join_switch_path(gizmo->fdev_num, &pfo_id);
	}
	return result;
}

/** A close() call has been made to the given device.
 * Not re-entrant per device.
 * @param[in] gizmo describes the device to access
 * @return PFO_GOOD or standard Linux return error code (e.g. -ENXIO).
 */
static int join_close(struct join_device_identifier *gizmo)
{
	struct pfo_ID pfo_id;
	int result = -EMEDIUMTYPE;

	JOIN_PRINT_WWNN(gizmo->wwnn, "fdev=%d type=%d minor=%d %d:%d:%d:%d %s\n",
		gizmo->fdev_num, gizmo->dev_type, minor_number[gizmo->fdev_num],
		gizmo->host, gizmo->channel, gizmo->target, gizmo->lun,
		gizmo->serial);

	pfo_id.hctl.host = gizmo->host;
	pfo_id.hctl.channel = gizmo->channel;
	pfo_id.hctl.target = gizmo->target;
	pfo_id.hctl.lun = gizmo->lun;
	pfo_id.f_flags = gizmo->filp->f_flags;
	pfo_id.serial_num_len = strlen(gizmo->serial);
	pfo_id.pfo_minor = minor_number[gizmo->fdev_num];
	memcpy(&pfo_id.wwnn, gizmo->wwnn, WWN_LEN);
	memcpy(&pfo_id.serial_num, gizmo->serial, MaxSerialNumberSize);

	if (gizmo->dev_type == 1) {
		result = pfo_close_drive(&pfo_id);
	} else if (gizmo->dev_type == 8) {
		result = pfo_close_changer(&pfo_id);
	}
	/* If close() fails we must always invalidate the minor number,
	 * because the application has no recourse but to try an open()
	 * when it wants to use the device again.
	 */
	minor_number[gizmo->fdev_num] = INVALID_MINOR;
	if (result) {
		/* Fail */
		printk("lin_tape: close FAILED %s result=%d\n",
					gizmo->serial, result);
	} else {
		/* Pass */
		join_switch_path(gizmo->fdev_num, &pfo_id);
	}
	return result;
}

/** A new path to a device has been detected for use with this module.
 * @return =0 if this device is supported by this driver
 */
static int join_add(struct device *dev, struct class_interface *intf)
{
	int result;

	JOIN_PRINTK("ADD dev=%p intf=%p\n", dev, intf);
	result = join_class_intf.add_dev(dev, intf);
	/* Do not call join_switch_path here. */
	return result;
}

/** A path to a device has been removed that was used with this module.
 */
static void join_remove(struct device *dev, struct class_interface *intf)
{
	JOIN_PRINTK("REMOVE dev=%p intf=%p\n", dev, intf);
	join_class_intf.remove_dev(dev, intf);
	/* Do not call join_switch_path here. */
}

/*###############################################################*/
/* Process a command to be sent to the device.			 */
/*###############################################################*/
/** Send the given SCSI command to the given device.
 * buffer must point to kernel memory of length buffer_length
 * Not re-entrant per device, to serialize commands.
 * @param[in] gizmo describes the device to access
 * @return PFO_GOOD or standard Linux return error code (e.g. -ENXIO).
 */
static int join_send(struct join_device_identifier *gizmo,
	unchar *cdb,
	unchar cmd_length,
	unchar *buffer,
	uint buffer_length,
	uint data_direction,
	uint timeout,
	unchar *sense,
	int *sense_length,
	int *resid,
	int *errors)
{
	struct pfo_ID pfo_id;
	unchar sense_avail = 0;
	int result = -EMEDIUMTYPE;
	unchar local_sense_length = 0;
	uint local_buffer_length = buffer_length;
	unchar local_data_direction = SCSI_DATA_NONE;

	JOIN_PRINT_WWNN(gizmo->wwnn,
		"cmd=%02x fdev=%d type=%d minor=%d %d:%d:%d:%d %s\n", cdb[0],
		gizmo->fdev_num, gizmo->dev_type, minor_number[gizmo->fdev_num],
		gizmo->host, gizmo->channel, gizmo->target, gizmo->lun,
		gizmo->serial);

	if (minor_number[gizmo->fdev_num] == INVALID_MINOR) {
		JOIN_PRINTK("never opened\n");
		return -ENOENT;
	}
	pfo_id.hctl.host = gizmo->host;
	pfo_id.hctl.channel = gizmo->channel;
	pfo_id.hctl.target = gizmo->target;
	pfo_id.hctl.lun = gizmo->lun;
	pfo_id.serial_num_len = strlen(gizmo->serial);
	pfo_id.pfo_minor = minor_number[gizmo->fdev_num];
	memcpy(&pfo_id.wwnn, gizmo->wwnn, WWN_LEN);
	memcpy(&pfo_id.serial_num, gizmo->serial, MaxSerialNumberSize);

	/* Convert from Linux to lin_tape direction constants. */
	if (buffer != NULL && buffer_length != 0) {
		if (data_direction == DMA_FROM_DEVICE) {
			local_data_direction = SCSI_DATA_IN;
		} else if (data_direction == DMA_TO_DEVICE) {
			local_data_direction = SCSI_DATA_OUT;
		}
	}
	/* Convert from Linux to lin_tape timeout values. */
	timeout = (timeout + (HZ - 1)) / HZ;
	if (timeout < 10) {
		timeout = 10;
	}
	local_sense_length = (unsigned char)*sense_length;
	if (local_sense_length > 255) {
		local_sense_length = 255;
	}


	/* Send command to lin_tape. */
	if (gizmo->dev_type == 1) {
		result = pfo_tape_pass_thru(&pfo_id, cdb,
				cmd_length, buffer, &local_buffer_length,
				local_data_direction, timeout, sense,
				&local_sense_length, &sense_avail,
				resid, errors);
	} else if (gizmo->dev_type == 8) {
		result = pfo_changer_pass_thru(&pfo_id, cdb,
				cmd_length, buffer, &local_buffer_length,
				local_data_direction, timeout, sense,
				&local_sense_length, &sense_avail,
				resid, errors);
	}
	*sense_length = (int)local_sense_length;
	if (result) {
		/* Fail */
		JOIN_PRINTK("send FAILED result=%d\n", result);
	} else {
		/* Pass */
		if (*errors) {	
			JOIN_PRINTK("send errors=%X\n", *errors);
		}
		join_switch_path(gizmo->fdev_num, &pfo_id);
	}
	return result;
}

/*###############################################################*/
/* Establish or tear down the connection with the pfo module.    */
/*###############################################################*/
static struct join_class_interface join_interface = {
	.join_add    = join_add,
	.join_remove = join_remove,
	.join_open   = join_open,
	.join_close  = join_close,
	.join_send   = join_send,
};

/*###############################################################*/
/* Functions called by the lin_tape module.			 */
/*###############################################################*/
/** The module is being loaded and initialized.
 * Register with the pfo module.
 * Replaces scsi_register_interface(intf);
 */
int join_register_interface(struct class_interface *intf)
{
	int i;
	int result;

	for (i = 0; i < MAX_FDEV_NUMBER; i++) {
		minor_number[i] = INVALID_MINOR;
	}
	printk("sfmp: Version=%s\n", join_version_date);
	JOIN_PRINTK("%p\n", intf);
	join_class_intf.add_dev = intf->add_dev;
	join_class_intf.remove_dev = intf->remove_dev;
	/* Register with the pfo module. */
	result = pfo_register_sf(&join_interface);
	return result;
}

/** The module is being unloaded.  Free up data structures.
 * Unregister with the pfo module.
 * Replaces scsi_unregister_interface(intf);
 */
void join_unregister_interface(struct class_interface *intf)
{
	/* Unregister with the pfo module. */
	JOIN_PRINTK("%p\n", intf);
	pfo_unregister_sf();
	join_class_intf.add_dev = NULL;
	join_class_intf.remove_dev = NULL;
}
