/*******************************************************************************
 *   lin_tape_extra_ioctl.c
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#include "lin_tape_extra_ioctl.h"

/*******************************************************************************
*                                                                              *
*  NAME: changer_ddioc_inquiry                                                 *
*                                                                              *
*  PURPOSE: issue standard inquiry                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_ddioc_inquiry(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	struct changer_descriptor* chg = *chgp;
	dd_old_inquiry_t *old_inq = NULL;
	struct st_inquiry *inq_data = NULL;
	int rc = STATUS_SUCCESS, errno = STATUS_SUCCESS;

	old_inq = (dd_old_inquiry_t*)buffer;

	inq_data = kmalloc(sizeof(struct st_inquiry), GFP_KERNEL);
	if(!inq_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(inq_data, '\0', sizeof(struct st_inquiry));

	rc = changer_std_inquiry(&chg, (void*)inq_data);

	if(rc == STATUS_SUCCESS)
		memcpy(&old_inq->data, inq_data, sizeof(dd_inq_data_t));

EXIT_LABEL:

	if(inq_data) {
		kfree(inq_data);
		inq_data = NULL;
	} /* if */

	if(rc != STATUS_SUCCESS) errno = -rc;
	old_inq->compcode = changer_get_compcode(errno, chg);

	*chgp = chg;
	return rc;
} /* changer_ddioc_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_ddioc_ret_err                                                 *
*                                                                              *
*  PURPOSE: get the last completion code                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_ddioc_ret_err(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	struct changer_descriptor* chg = *chgp;
	dd_return_error_state_t* ret_err = NULL;
	int sense_len = 0, errno = STATUS_SUCCESS, rc = STATUS_SUCCESS;

	ret_err = (dd_return_error_state_t*)buffer;

	ret_err->last_status = chg->last_errno;
	if(chg->comp_code_valid) ret_err->last_compcode = chg->comp_code;
	else ret_err->last_compcode =
		changer_get_compcode(chg->last_errno, chg);

	if(chg->sense_valid) {
		sense_len = 8 + chg->sense_data->additional_sense_len;
		ret_err->number_sense = min(sense_len, DD_SCSI_MAX_SENSE);
		if(ret_err->number_sense)
			memcpy(&ret_err->sense_bytes, chg->sense_data,
				ret_err->number_sense);
	} else ret_err->number_sense = 0;

	*chgp = chg;
	rc = errno;

	return rc;
} /* changer_ddioc_ret_err() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_lib_info                                                *
*                                                                              *
*  PURPOSE: get library information                                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_lib_info(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	struct changer_descriptor* chg = *chgp;
	lb_library_info_t* inf = NULL;
	struct elem_assignment* elem_asn = NULL;
	int errno = STATUS_SUCCESS, rc = STATUS_SUCCESS, num_drives = 0,
		num_slots = 0;

	inf = (lb_library_info_t*)buffer;

	elem_asn = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!elem_asn) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_asn, '\0', sizeof(struct elem_assignment));

	if(chg->chg_id == IBM_3584) inf->lb_capable_flag.exchange_medium = True;
	else inf->lb_capable_flag.exchange_medium = False;

	if(chg->chg_id == IBM_3590) inf->lb_capable_flag.ier = False;
	else inf->lb_capable_flag.ier = True;

	switch(chg->chg_id) {
	case IBM_3590:
		inf->support_module = 2;
		inf->robot_can_eject = False;
		inf->inv_unreliable = False;
		inf->skip_all_audits = False;
		inf->must_position_to_elem = False;
		inf->moves_slot_to_slot = True;
		inf->uses_magazine = False;
		inf->robot_can_initiate = False;
		inf->robot_can_invert = False;
		inf->stackable_magazine = False;
		inf->mixed_devtype_capable = False;
		break;

	case IBM_3581:
	case BDT_3581:
		inf->robot_can_eject = False;
		inf->inv_unreliable = False;
		inf->skip_all_audits = False;
		inf->must_position_to_elem = False;
		inf->moves_slot_to_slot = True;
		inf->uses_magazine = False;
		inf->robot_can_initiate = False;
		inf->robot_can_invert = False;
		inf->stackable_magazine = False;
		inf->support_module = 2;
		inf->mixed_devtype_capable = False;
		break;

	case IBM_3582:
	case IBM_3583:
	case IBM_3584:
	case IBM_3576:
	case IBM_3573:
	case IBM_3572:
	case IBM_3577:
	case QTM_500:
	case QTM_6000:
	case HP_MSLG3:
	case HP_MSL6480:
	case HP_1x8G2:
		inf->robot_can_eject = False;
		inf->inv_unreliable = False;
		inf->skip_all_audits = False;
		inf->must_position_to_elem = False;
		inf->moves_slot_to_slot = True;
		inf->uses_magazine = False;
		inf->robot_can_initiate = False;
		inf->robot_can_invert = False;
		inf->stackable_magazine = False;
		inf->mixed_devtype_capable = True;

		rc = changer_modesense_page(&chg, ElementAddressPage,
			(void*)elem_asn, sizeof(struct elem_assignment));
		if(rc != STATUS_SUCCESS) break;

		num_drives = (elem_asn->data_transfer_num[0] << 8)
			+ elem_asn->data_transfer_num[1];
		num_slots = (elem_asn->storage_elem_num[0] << 8)
			+ elem_asn->storage_elem_num[1];

		if(num_drives > 3 || num_slots > 40) inf->support_module = 4;
		else inf->support_module = 2;
		break;

	default:
		rc = -EINVAL;
	} /* switch */

EXIT_LABEL:

	if(elem_asn) {
		kfree(elem_asn);
		elem_asn = NULL;
	} /* if */

	if(rc != STATUS_SUCCESS) errno = -rc;
	inf->compcode = changer_get_compcode(errno, chg);

	*chgp = chg;
	return rc;
} /* changer_lbioc_lib_info() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_ret_elem_cnt                                            *
*                                                                              *
*  PURPOSE: get library element count                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_ret_elem_cnt(struct changer_descriptor** chgp,
	void *buffer, char unused)
{
	lb_element_count_t* elem_cnt = NULL;
	struct elem_assignment* elem_asn = NULL;
	int errno = STATUS_SUCCESS;
	int rc = STATUS_SUCCESS;

	elem_cnt = (lb_element_count_t*)buffer;

	elem_asn = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!elem_asn) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_asn, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(chgp, ElementAddressPage, (void*)elem_asn,
		sizeof(struct elem_assignment));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	elem_cnt->transp_elems = (elem_asn->transport_elem_num[0]<<8)
		+ elem_asn->transport_elem_num[1];
	elem_cnt->slots = (elem_asn->storage_elem_num[0] << 8)
		+ elem_asn->storage_elem_num[1];
	elem_cnt->eeports = (elem_asn->ie_port_elem_num[0] << 8)
		+ elem_asn->ie_port_elem_num[1];
	elem_cnt->drives = (elem_asn->data_transfer_num[0] << 8)
		+ elem_asn->data_transfer_num[1];
EXIT_LABEL:

	if(elem_asn) {
		kfree(elem_asn);
		elem_asn = NULL;
	} /* if */

	if(rc != STATUS_SUCCESS) errno = -rc;
	elem_cnt->compcode = changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_ret_elem_cnt() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_move_medium                                             *
*                                                                              *
*  PURPOSE: issue move medium per the input from buffer                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: input buffer                                                      *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_move_medium(struct changer_descriptor** chgp,
	void* buffer, char unused)
{
	lb_move_medium_t* mv_elem = NULL;
	int errno = STATUS_SUCCESS;
	int rc = STATUS_SUCCESS;

	mv_elem = (lb_move_medium_t*)buffer;

	rc = changer_move_medium(chgp, mv_elem->transport_element,
		mv_elem->source_address, mv_elem->destination_address,
		mv_elem->invert);

	if(rc != STATUS_SUCCESS) errno = -rc;
	mv_elem->compcode = changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_move_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_lib_inventory                                           *
*                                                                              *
*  PURPOSE: get library inventory                                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: input buffer                                                      *
*    copy_method: to kernel or to user                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_lib_inventory(struct changer_descriptor** chgp,
	void* buffer, char copy_method)
{
	int rc = STATUS_SUCCESS;
	int errno = STATUS_SUCCESS;

	rc = changer_get_lib_inventory(chgp, buffer, copy_method);

	if(rc != STATUS_SUCCESS) errno = -rc;
	((lb_inventory_t*)buffer)->compcode =
		changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_lib_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_position                                                *
*                                                                              *
*  PURPOSE: position the robot per the input from buffer                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: input buffer                                                      *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_position(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	lb_position_t* pos_elem = NULL;
	int errno = STATUS_SUCCESS;
	int rc = STATUS_SUCCESS;

	pos_elem = (lb_position_t*)buffer;

	rc = changer_pos_element(chgp, pos_elem->transport_element,
		pos_elem->destination_address, pos_elem->invert);

	if(rc != STATUS_SUCCESS) errno = -rc;
	pos_elem->compcode = changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_position() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_audit                                                   *
*                                                                              *
*  PURPOSE: issue IES, use barcode input from buffer                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: input/output buffer                                               *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_audit(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	struct changer_descriptor* chg = *chgp;
	lb_audit_t* lib_audit = NULL;
	struct elem_assignment* elem_asn = NULL;
	ushort src = 0, eeports = 0;
	int errno = STATUS_SUCCESS;
	int rc = STATUS_SUCCESS;

	lib_audit = (lb_audit_t*)buffer;

	/* set barcode flag for inventory later */
	chg->barcode = lib_audit->barcode;

	if(chg->barcode != 2) {
		rc = changer_init_element(&chg);
		if(rc == -ENOSYS) rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	/* if barcode == 2, then only audit i/e ports */
	elem_asn = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!elem_asn) {
		rc = -ENOMEM;
		errno = -rc;
		lib_audit->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* if */
	memset(elem_asn, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(&chg, ElementAddressPage, (void*)elem_asn,
		sizeof(struct elem_assignment));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	eeports = (elem_asn->ie_port_elem_num[0] << 8)
		+ elem_asn->ie_port_elem_num[1];
	src = (elem_asn->first_ie_port_addr[0] << 8)
		+ elem_asn->first_ie_port_addr[1];

	if(eeports) {
		rc = changer_init_elem_range(&chg, src, eeports);
		if(rc == -ENOSYS) rc = STATUS_SUCCESS;
	} else {
		rc = changer_init_element(&chg);
		if(rc == -ENOSYS) rc = STATUS_SUCCESS;
	} /* else */

EXIT_LABEL:

	if(elem_asn) {
		kfree(elem_asn);
		elem_asn = NULL;
	} /* if */

	if(rc != STATUS_SUCCESS) errno = -rc;
	lib_audit->compcode = changer_get_compcode(errno, chg);

	*chgp = chg;
	return rc;
} /* changer_lbioc_audit() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_ext_ret_ioport                                          *
*                                                                              *
*  PURPOSE: unsupported function, just set success and return                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS                                                     *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_ext_ret_ioport(struct changer_descriptor** chgp,
	void* buffer, char unused)
{
	((lb_extend_retract_t*)buffer)->compcode =
		changer_get_compcode(STATUS_SUCCESS, *chgp);
	return STATUS_SUCCESS;
} /* changer_lbioc_ext_ret_ioport() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_ddioc_xinquiry                                                *
*                                                                              *
*  PURPOSE: get extended inquiry                                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_ddioc_xinquiry(struct changer_descriptor** chgp, void* buffer,
	char unused)
{
	struct changer_descriptor* chg = *chgp;
	dd_inquiry_t *ext_inq = NULL;
	struct st_inquiry *inq_data = NULL;
	unchar *inq_page = NULL;
	char id_len = 0;
	int rc = STATUS_SUCCESS, errno = STATUS_SUCCESS, offset = 0;

	ext_inq = (dd_inquiry_t*)buffer;

	inq_data = kmalloc(sizeof(struct st_inquiry), GFP_KERNEL);
	if(!inq_data) {
		rc = -ENOMEM;
		errno = -rc;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* if */
	memset(inq_data, '\0', sizeof(struct st_inquiry));

	rc = changer_std_inquiry(&chg, inq_data);
	if(rc != STATUS_SUCCESS) {
		errno = -rc;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* if */

	memcpy(&ext_inq->data, inq_data, sizeof(dd_inq_data_t));

	/* get device serial number */
	inq_page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq_page) {
		rc = -ENOMEM;
		errno = -rc;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* if */
	memset(inq_page, '\0', MAX_INQ_LEN);

	rc = changer_inquiry_page(&chg, UnitSerialPage, inq_page);
	if(rc != STATUS_SUCCESS) {
		errno = -rc;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* if */
	ext_inq->dd_valid_flag.dd_device_serial_valid = True;
	memcpy((void*)ext_inq->dd_device_serial, &inq_page[4],
		min((int)inq_page[3], DD_MAX_DEVICE_SERIAL));

	/* get world wide node name */
	memset(inq_page, '\0', MAX_INQ_LEN);
	rc = changer_inquiry_page(&chg, DeviceIdentificationPage, inq_page);

	if(rc == -EINVAL) {
		rc = errno = STATUS_SUCCESS;
		ext_inq->dd_valid_flag.ww_name_valid = False;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} else if(rc != STATUS_SUCCESS) {
		errno = -rc;
		ext_inq->compcode = changer_get_compcode(errno, chg);
		goto EXIT_LABEL;
	} /* else if */

	offset = 4;
	while(offset < inq_page[3]) {
		/* fibre channel descriptor */
		if((inq_page[offset + 1] & 0x0F) == 3) {
			offset += 4;
			id_len = inq_page[offset - 1];
			memcpy(ext_inq->ww_name, &inq_page[offset],
				min((int)id_len, DD_MAX_DEVICE_WWNAME));
			ext_inq->dd_valid_flag.ww_name_valid = True;
			offset += id_len;
			break;
		} /* if */
		offset += 4;
		id_len = inq_page[offset - 1];
		offset += id_len;
	} /* while */

	ext_inq->compcode = changer_get_compcode(errno, chg);

EXIT_LABEL:

	if(inq_page) kfree(inq_page);
	if(inq_data) kfree(inq_data);

	*chgp = chg;
	return rc;
} /* changer_ddioc_xinquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_xdrive_inventory                                        *
*                                                                              *
*  PURPOSE: get extended drive inventory                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: output buffer                                                     *
*    copy_method: to kernel space or user space                                *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_xdrive_inventory(struct changer_descriptor** chgp,
	void* buffer, char copy_method)
{
	struct changer_descriptor* chg = *chgp;
	lb_Xdrive_inventory_t* xdrive_inventory = NULL;
	lb_Xdrive_state_t* xdrive_state = NULL;
	ushort start = 0, req_drives = 0, num_drives = 0;
	struct tape_elem_status* elem_info = NULL;
	struct tape_read_devid* read_devid = NULL;
	unchar* inv_buf = NULL, * devid_buf = NULL, vol_tag = 0, bar_code = 0;
	ulong devid_order = 0, inv_order = 0, inv_buf_size = 0,
		devid_buf_size = 0;
	int rc = STATUS_SUCCESS, errno = STATUS_SUCCESS, i = 0, k = 0, done = 0;

	xdrive_inventory = (lb_Xdrive_inventory_t*)buffer;
	xdrive_state = xdrive_inventory->Xdrive_state;
	start = xdrive_inventory->drive_element;
	req_drives = xdrive_inventory->drives;

	rc = xinv_res_and_devid(&chg, &num_drives, start, req_drives,
		&devid_buf, &devid_buf_size, &devid_order, &inv_buf,
		&inv_buf_size, &inv_order, &vol_tag);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* now we put inv_buf and devid_buf together */
	elem_info = (struct tape_elem_status*)inv_buf;
	read_devid = (struct tape_read_devid*)devid_buf;
	if(vol_tag) bar_code = 1;

	for(i = 0; i < num_drives; i++) {
		done = 0;
		rc = xinv_gather_and_copy(chg, elem_info, read_devid,
			xdrive_state, start, k, req_drives, bar_code,
			copy_method, &done);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(done == 1) break;
		elem_info++;
		read_devid++;
		if(done == 2) continue;
		xdrive_state++;
		k++;
	} /* for */

EXIT_LABEL:

	if(inv_buf) {
		if(inv_buf_size <= KmallocThreshhold) kfree(inv_buf);
		else free_pages((ulong)(inv_buf), inv_order);
	} /* if */

	if(devid_buf) {
		if(devid_buf_size <= KmallocThreshhold) kfree(devid_buf);
		else free_pages((ulong)(devid_buf), devid_order);
	} /* if */

	errno = rc;
	xdrive_inventory->compcode = changer_get_compcode(errno, chg);
	*chgp = chg;
	return rc;
} /* changer_lbioc_xdrive_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: xinv_res_and_devid                                                    *
*                                                                              *
*  PURPOSE: collect data via RES and read_devid for xdrive_inventory           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    num_drives: (output) number of drives returned from report elements       *
*    start: the requested starting element address for which to get info       *
*    req_drives: number of requested drives for which to get info              *
*    devid_buf: (output) buffer for devid info                                 *
*    devid_buf_size: (output) size of the devid_buffer                         *
*    devid_order: (output) base-2 logarithm of the number of pages for devid   *
*    inv_buf: (output) buffer for res inventory                                *
*    inv_buf_size: (output) size of the inventory buffer                       *
*    inv_order:(output) base-2 logarithm of the number of pages for inventory  *
*    vol_tag: (output) barcode label supported                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int xinv_res_and_devid(struct changer_descriptor** chgp, ushort* num_drives,
	ushort start, ushort req_drives, unchar** devid_buf,
	ulong* devid_buf_size, ulong* devid_order, unchar** inv_buf,
	ulong* inv_buf_size, ulong* inv_order, unchar* vol_tag)
{
	struct elem_assignment* elem_asn = NULL;
	struct changer_descriptor* chg = *chgp;
	ulong alloc_len = 0;
	int rc = STATUS_SUCCESS;
	ushort first = 0;

	elem_asn = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);

	if(!elem_asn) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_asn, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(&chg, ElementAddressPage, (void*)elem_asn,
		sizeof(struct elem_assignment));

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	first = elem_asn->first_data_transfer_addr[0] << 8;
	first += elem_asn->first_data_transfer_addr[1];
	*num_drives = elem_asn->data_transfer_num[0] << 8;
	*num_drives += elem_asn->data_transfer_num[1];

	if(start < first || start > first + *num_drives ||
		req_drives > *num_drives) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	alloc_len = (*num_drives) * sizeof(struct tape_elem_status);
	if(alloc_len > KmallocThreshhold) {
		*inv_order = get_mem_ord(alloc_len, inv_buf_size);
		*inv_buf = (unchar*)__get_free_pages(GFP_KERNEL, *inv_order);
	} else {
		*inv_buf = kmalloc(alloc_len, GFP_KERNEL);
		*inv_buf_size = alloc_len;
	} /* else */

	if(!*inv_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(*inv_buf, '\0', *inv_buf_size);

	/* issue read element status command with DVCID 0 and vol_tag 1 */
	*vol_tag = 1;
	rc = changer_read_elem_status(&chg, *inv_buf, elem_asn,
		DataTransferReport, first, *num_drives, vol_tag, BackToKernel,
		FALSE);
	if(rc != STATUS_SUCCESS) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	alloc_len = (*num_drives) * sizeof(struct tape_read_devid);
	if(alloc_len > KmallocThreshhold) {
		*devid_order = get_mem_ord(alloc_len, devid_buf_size);
		*devid_buf = (unchar*)__get_free_pages(GFP_KERNEL,
			*devid_order);
	} else {
		*devid_buf = kmalloc(alloc_len, GFP_KERNEL);
		*devid_buf_size = alloc_len;
	} /* else */

	if(!*devid_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	rc = changer_read_devid(&chg, first, *num_drives, *devid_buf,
		BackToKernel);
	if(rc != STATUS_SUCCESS) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	*chgp = chg;
	if(elem_asn) kfree(elem_asn);
	return rc;
} /* xinv_res_and_devid() */

/*******************************************************************************
*                                                                              *
*  NAME: xinv_gather_and_copy                                                  *
*                                                                              *
*  PURPOSE: combine data from RES into lb_Xdrive_state_t struct and copy       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    elem_info: result from read element status                                *
*    read_devid: result from read devid                                        *
*    xdrive_state: output buffer                                               *
*    start: requested start element                                            *
*    k: current element to process                                             *
*    req_drives: number of drives in the report                                *
*    bar_code: if True, copy barcode info from cartridge                       *
*    copy_method: to kernel space or user space                                *
*    done: if !0, caller to break out of loop or continue                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int xinv_gather_and_copy(struct changer_descriptor* chg,
	struct tape_elem_status* elem_info, struct tape_read_devid* read_devid,
	lb_Xdrive_state_t* xdrive_state, ushort start, int k,
	ushort req_drives, unchar bar_code, char copy_method, int* done)
{
	int rc = STATUS_SUCCESS;
	lb_Xdrive_state_t* ext_drive_state = NULL;
	ushort asc_ascq = 0;

	ext_drive_state = kmalloc(sizeof(lb_Xdrive_state_t), GFP_KERNEL);
	if(!ext_drive_state) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(ext_drive_state, '\0', sizeof(lb_Xdrive_state_t));

	if(!elem_info || !read_devid || !xdrive_state) {
		printk("lin_tape: bad element state address.\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(elem_info->addr < start) {
		*done = 2;
		goto EXIT_LABEL;
	} /* if */

	if(k >= req_drives) {
		*done = 1;
		goto EXIT_LABEL;
	} /* if */

	ext_drive_state->lb_drive_state.address = elem_info->addr;
	ext_drive_state->lb_drive_state.accessible = elem_info->flags1_4;
	ext_drive_state->lb_drive_state.full = elem_info->flags1_7;
	ext_drive_state->lb_drive_state.scsi_bus_not_pres = elem_info->flags2_0;
	ext_drive_state->lb_drive_state.scsi_bus_id_flag = elem_info->flags2_2;
	ext_drive_state->lb_drive_state.scsi_bus_id = elem_info->scsi;
	ext_drive_state->lb_drive_state.lun_flag = elem_info->flags2_3;
	ext_drive_state->lb_drive_state.lun = (elem_info->flags2_7 << 16)
		+ (elem_info->flags2_6 << 8)
		+ elem_info->flags2_5;
	ext_drive_state->lb_drive_state.source_addr_flag = elem_info->flags3_0;
	ext_drive_state->lb_drive_state.source_elem_addr = elem_info->source;
	ext_drive_state->lb_drive_state.invert_media_flag = elem_info->flags3_1;

	asc_ascq = ((elem_info->asc & 0xFF) << 8) | elem_info->ascq;
	ext_drive_state->lb_drive_state.abnormal_state =
		changer_get_abnormal_state(asc_ascq);

	/* set bar code in ext_drive_state */
	if(bar_code) set_barcode(chg, ext_drive_state->lb_drive_state.barcode,
		&ext_drive_state->lb_drive_state.barcode_flag,
		&ext_drive_state->lb_drive_state.barcode_len, elem_info);

	/* the drive addresses inside elem_info and read_devid should match */
	if(elem_info->addr == read_devid->addr && read_devid->id_length > 24
		&& read_devid->identifier) {
		/* drive serial number is after 24 bytes of vendor/product id */
		memcpy(ext_drive_state->drive_serial,
			read_devid->identifier + 24,
			min(DD_MAX_DEVICE_SERIAL,
					(int)(read_devid->id_length - 24)));
		ext_drive_state->drive_serial_valid = True;
	} /* if */

	if(copy_method == BackToKernel)
		memcpy(xdrive_state, ext_drive_state,
			sizeof(lb_Xdrive_state_t));
	else if(copy_to_user((void*)xdrive_state, (void*)ext_drive_state,
		sizeof(lb_Xdrive_state_t))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
	} /* else */

EXIT_LABEL:

	if(ext_drive_state) kfree(ext_drive_state);
	return rc;
} /* xinv_gather_and_copy() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_logsensepage                                            *
*                                                                              *
*  PURPOSE: get library logsense page                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: buffer in which to copy user data                                 *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_logsensepage(struct changer_descriptor** chgp,
	void* buffer, char unused)
{
	dd_log_sense_page_t *lib_logsense = NULL;
	int rc = STATUS_SUCCESS, errno = STATUS_SUCCESS;
	uint log_length = 0;

	lib_logsense = (dd_log_sense_page_t*)buffer;

	rc = changer_logsense_page(chgp, lib_logsense->page_control >> 6,
		lib_logsense->page_code, lib_logsense->data, &log_length);

	if(rc != STATUS_SUCCESS) errno = -rc;
	lib_logsense->compcode = changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_logsensepage() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_lbioc_lib_enh_inventory                                       *
*                                                                              *
*  PURPOSE: get enhanced library inventory                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: buffer in which to copy user data                                 *
*    copy_method: kernel space or user space                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_lbioc_lib_enh_inventory(struct changer_descriptor** chgp,
	void* buffer, char copy_method)
{
	int rc = STATUS_SUCCESS;
	int errno = STATUS_SUCCESS;

	rc = changer_get_lib_enh_inventory(chgp, buffer, copy_method);

	if(rc != STATUS_SUCCESS) errno = -rc;
	((lb_enh_inventory_t*)buffer)->compcode =
		changer_get_compcode(errno, *chgp);

	return rc;
} /* changer_lbioc_lib_enh_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_lib_inventory                                             *
*                                                                              *
*  PURPOSE: get library inventory for TSM                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: buffer in which to copy user data                                 *
*    copy_method: kernel space or user space                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_get_lib_inventory(struct changer_descriptor** chgp,
	void* buffer, char copy_method)
{
	struct changer_descriptor* chg = *chgp;
	struct elem_assignment* elem_asn = NULL;
	lb_drive_state_t* drive_state = NULL;
	lb_storage_element_state_t* slot_state = NULL;
	lb_import_export_state_t* ieport_state = NULL;
	lb_medium_changer_state_t* robot_state = NULL;
	ushort req_drives, req_ieports, req_slots, req_robots;
	int rc = STATUS_SUCCESS;

	rc = init_inventory(chg, &elem_asn, &req_drives, &req_ieports,
		&req_slots, &req_robots, &drive_state, &slot_state,
		&ieport_state, &robot_state, buffer);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(req_drives && drive_state) {
		rc = inv_res(&chg, (void*)drive_state, DataTransferReport,
			req_drives, copy_method, elem_asn, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_ieports && ieport_state) {
		rc = inv_res(&chg, (void*)ieport_state, IEPortReport,
			req_ieports, copy_method, elem_asn, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_slots && slot_state) {
		rc = inv_res(&chg, (void*)slot_state, StorageElementReport,
			req_slots, copy_method, elem_asn, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_robots && robot_state) {
		rc = inv_res(&chg, (void*)robot_state, MediumTransportReport,
			req_robots, copy_method, elem_asn, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	*chgp = chg;

EXIT_LABEL:

	if(elem_asn) {
		kfree(elem_asn);
		elem_asn = NULL;
	} /* if */

	return rc;
} /* changer_get_lib_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_lib_enh_inventory                                         *
*                                                                              *
*  PURPOSE: get library enhanced inventory for TSM                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor representing the library                     *
*    buffer: buffer in which to copy user data                                 *
*    copy_method: kernel space or user space                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_get_lib_enh_inventory(struct changer_descriptor** chgp,
	void* buffer, char copy_method)
{
	struct changer_descriptor* chg = *chgp;
	struct elem_assignment* elem_asn = NULL;
	lb_enh_drive_state_t* drive_state = NULL;
	lb_enh_storage_element_state_t* storage_state = NULL;
	lb_enh_import_export_state_t* ieport_state = NULL;
	lb_medium_changer_state_t* robot_state = NULL;
	ushort req_drives, req_ieports, req_slots, req_robots;
	int rc = STATUS_SUCCESS;

	rc = init_inventory(chg, &elem_asn, &req_drives, &req_ieports,
		&req_slots, &req_robots,
		(lb_drive_state_t**)&drive_state,
		(lb_storage_element_state_t**)&storage_state,
		(lb_import_export_state_t**)&ieport_state,
		&robot_state, buffer);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(req_drives && drive_state) {
		rc = inv_res(&chg, (void*)drive_state, DataTransferReport,
			req_drives, copy_method, elem_asn, True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_ieports && ieport_state) {
		rc = inv_res(&chg, (void*)ieport_state, IEPortReport,
			req_ieports, copy_method, elem_asn, True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_slots && storage_state) {
		rc = inv_res(&chg, (void*)storage_state,
			StorageElementReport, req_slots, copy_method, elem_asn,
			True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(req_robots && robot_state) {
		/* transport uses inventory structure, not enhanced */
		rc = inv_res(&chg, (void*)robot_state,
			MediumTransportReport, req_robots, copy_method,
			elem_asn, True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	*chgp = chg;

EXIT_LABEL:

	if(elem_asn) kfree(elem_asn);

	return rc;
} /* changer_get_lib_enh_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: init_inventory                                                        *
*                                                                              *
*  PURPOSE: call mode sense on ElementAddressPage and use it to initialize inv *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which we are getting the inventory                   *
*    elem_asn: the data to be retrieved from read element status               *
*    req_drives: the requested number of drives for the report                 *
*    req_ieports: the requested number of ieports for the report               *
*    req_slots: the requested number of slots for the report                   *
*    req_robots: the requested number of robots for the report                 *
*    drive_state: user output buffer for the drive report                      *
*    slot_state: user output buffer for the storage element report             *
*    ieport_state: user output buffer for the import/export report             *
*    robot_state: user output buffer for the medium changer report             *
*    buffer: user input buffer                                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int init_inventory(struct changer_descriptor* chg,
	struct elem_assignment** elem_asn,
	ushort* req_drives, ushort* req_ieports, ushort* req_slots,
	ushort* req_robots,
	lb_drive_state_t** drive_state,
	lb_storage_element_state_t** slot_state,
	lb_import_export_state_t** ieport_state,
	lb_medium_changer_state_t** robot_state,
	void* buffer)
{
	int rc = STATUS_SUCCESS;
	ushort max_drives, max_ieports, max_slots, max_robots;
	lb_inventory_t* inventory = NULL;

	inventory = (lb_inventory_t*)buffer;

	*elem_asn = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!*elem_asn) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(*elem_asn, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(&chg, ElementAddressPage, (void*)*elem_asn,
		sizeof(struct elem_assignment));

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: changer_modesense_page failed: %d\n", rc);
		goto EXIT_LABEL;
	} /* if */

	max_robots = ((*elem_asn)->transport_elem_num[0] << 8)
		+ (*elem_asn)->transport_elem_num[1];
	max_slots = ((*elem_asn)->storage_elem_num[0] << 8)
		+ (*elem_asn)->storage_elem_num[1];
	max_ieports = ((*elem_asn)->ie_port_elem_num[0] << 8)
		+ (*elem_asn)->ie_port_elem_num[1];
	max_drives = ((*elem_asn)->data_transfer_num[0] << 8)
		+ (*elem_asn)->data_transfer_num[1];

	*req_drives = inventory->drives;
	*req_ieports = inventory->eeports;
	*req_slots = inventory->slots;
	*req_robots = inventory->transp_elems;

	if(*req_drives > max_drives || *req_slots > max_slots ||
		*req_ieports > max_ieports || *req_robots > max_robots) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	*drive_state = inventory->drive_state;
	*ieport_state = inventory->import_export_state;
	*slot_state = inventory->storage_element_state;
	*robot_state = inventory->medium_changer_state;

EXIT_LABEL:

	return rc;
} /* init_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: set_drive_state                                                       *
*                                                                              *
*  PURPOSE: copy data retrieved from RES back to user-supplied buffer          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    data_buf: the RES data                                                    *
*    req_drives: the number of drives requested by the user                    *
*    bar_code: True if barcode is valid                                        *
*    drive: temp buffer in which to organize RES data for user                 *
*    drive_state: the user buffer in which to copy the drive parameter         *
*    copy_method: kernel mode or user mode                                     *
*    enh: if True, enhanced inventory; if False, standard inventory            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_drive_state(struct changer_descriptor* chg, char* data_buf,
	ushort req_drives, unchar bar_code, void* driver_buffer,
	void** user_buffer, char copy_method, boolean enh)
{
	int offs = 0, i = 0, rc = STATUS_SUCCESS;
	ushort asc_ascq = 0;
	struct tape_elem_status* elem_info = NULL;
	lb_enh_drive_state_t* enh_drive = NULL;
	lb_drive_state_t* drive = NULL;
	unchar* drive_state = NULL;

	enh_drive = (lb_enh_drive_state_t*)driver_buffer;
	drive = (lb_drive_state_t*)driver_buffer;
	drive_state = (unchar*)*user_buffer;

	for(i = 0; i < req_drives; i++) {

		elem_info = (struct tape_elem_status*)(data_buf + offs);
		if(!drive_state || !elem_info) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		if(enh) memset(enh_drive, '\0', sizeof(lb_enh_drive_state_t));
		else memset(drive, '\0', sizeof(lb_drive_state_t));

		drive->address = elem_info->addr;
		drive->accessible = elem_info->flags1_4;
		drive->full = elem_info->flags1_7;
		drive->scsi_bus_not_pres = elem_info->flags2_0;
		drive->scsi_bus_id_flag = elem_info->flags2_2;
		drive->scsi_bus_id = elem_info->scsi;
		drive->lun_flag = elem_info->flags2_3;
#if defined(__LITTLE_ENDIAN)
		drive->lun = (elem_info->flags2_5 << 2)
			+ (elem_info->flags2_6 << 1)
			+ elem_info->flags2_7;
#elif defined(__BIG_ENDIAN)
		drive->lun = (elem_info->flags2_7 << 2)
			+ (elem_info->flags2_6 << 1)
			+ elem_info->flags2_5;
#endif
		drive->source_addr_flag = elem_info->flags3_0;
		drive->source_elem_addr = elem_info->source;
		drive->invert_media_flag = elem_info->flags3_1;

		asc_ascq = ((elem_info->asc & 0xFF) << 8) | (elem_info->ascq);
		drive->abnormal_state = changer_get_abnormal_state(asc_ascq);

		if(bar_code) set_barcode(chg, drive->barcode,
			&drive->barcode_flag, &drive->barcode_len, elem_info);

		if(enh && (chg->chg_id == IBM_3584 ||
			chg->chg_id == IBM_3583 ||
			chg->chg_id == IBM_3582 ||
			chg->chg_id == IBM_3576 ||
			chg->chg_id == IBM_3573 ||
			chg->chg_id == IBM_3572 ||
			chg->chg_id == IBM_3577 ||
			chg->chg_id == QTM_500  ||
			chg->chg_id == QTM_6000 ||
			chg->chg_id == HP_MSLG3 ||
			chg->chg_id == HP_1x8G2 ||
			chg->chg_id == HP_MSL6480)) {
			set_media_type(&enh_drive->media_devtype,
				&enh_drive->element_devtype,
				&enh_drive->spec_media_type, &enh_drive->worm,
				enh_drive->full, enh_drive->barcode_flag,
				enh_drive->barcode_len, enh_drive->barcode,
				elem_info);

			/* element_drv_type is hardcoded as 1000 */
			enh_drive->element_drv_type = 1000;
		} /* if */

		if(copy_method == BackToKernel)
			memcpy(drive_state, drive, enh ?
				sizeof(lb_enh_drive_state_t) :
				sizeof(lb_drive_state_t));

		else if(copy_to_user((void*)drive_state, (void*)drive, enh ?
				sizeof(lb_enh_drive_state_t) :
				sizeof(lb_drive_state_t))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* else if */

		if(enh) drive_state += sizeof(lb_enh_drive_state_t);
		else drive_state += sizeof(lb_drive_state_t);
		offs += sizeof(struct tape_elem_status);
	} /* for */

	*user_buffer = (void*)drive_state;

EXIT_LABEL:

	return rc;
} /* set_drive_state() */

/*******************************************************************************
*                                                                              *
*  NAME: set_ieport_state                                                      *
*                                                                              *
*  PURPOSE: copy data retrieved from RES back to user-supplied buffer          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    data_buf: the RES data                                                    *
*    req_ieports: the number of ieports requested by the user                  *
*    bar_code: True if barcode is valid                                        *
*    ieport: temp buffer in which to organize RES data for user                *
*    ieport_state: the user buffer in which to copy the ieport parameter       *
*    copy_method: kernel mode or user mode                                     *
*    enh: if True, enhanced inventory; if False, standard inventory            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_ieport_state(struct changer_descriptor* chg, char* data_buf,
	ushort req_ieports, unchar bar_code, void* driver_buffer,
	void** user_buffer, char copy_method, boolean enh)
{
	int offs = 0, i = 0, rc = STATUS_SUCCESS;
	ushort asc_ascq = 0;
	struct tape_elem_status* elem_info = NULL;
	lb_enh_import_export_state_t* enh_ieport = NULL;
	lb_import_export_state_t *ieport = NULL;
	unchar* ieport_state = NULL;

	enh_ieport = (lb_enh_import_export_state_t*)driver_buffer;
	ieport = (lb_import_export_state_t*)driver_buffer;
	ieport_state = (unchar*)*user_buffer;

	for(i = 0; i < req_ieports; i++) {
		elem_info = (struct tape_elem_status*)(data_buf + offs);
		if(!ieport_state || !elem_info) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		/* set data in ieport_state  */
		memset(ieport, '\0', enh ?
			sizeof(lb_enh_import_export_state_t) :
			sizeof(lb_import_export_state_t));
		ieport->address = elem_info->addr;
		ieport->accessible = elem_info->flags1_4;
		ieport->full = elem_info->flags1_7;
		ieport->import_enable = elem_info->flags1_2;
		ieport->export_enable = elem_info->flags1_3;
		ieport->oper_placed_media = elem_info->flags1_6;

		ieport->source_addr_flag = elem_info->flags3_0;
		ieport->source_elem_addr = elem_info->source;
		ieport->invert_media_flag = elem_info->flags3_1;

		asc_ascq = ((elem_info->asc & 0xFF) << 8)
			| (elem_info->ascq);
		ieport->abnormal_state = changer_get_abnormal_state(asc_ascq);

		if(bar_code) set_barcode(chg, ieport->barcode,
			&ieport->barcode_flag, &ieport->barcode_len, elem_info);

		if(enh && (chg->chg_id == IBM_3584 ||
			chg->chg_id == IBM_3583 ||
			chg->chg_id == IBM_3582 ||
			chg->chg_id == IBM_3576 ||
			chg->chg_id == IBM_3573 ||
			chg->chg_id == IBM_3572 ||
			chg->chg_id == IBM_3577 ||
			chg->chg_id == QTM_500  ||
			chg->chg_id == QTM_6000 ||
			chg->chg_id == HP_MSLG3 ||
			chg->chg_id == HP_1x8G2 ||
			chg->chg_id == HP_MSL6480)) {
			set_media_type(&enh_ieport->media_devtype,
				&enh_ieport->element_devtype,
				&enh_ieport->spec_media_type, &enh_ieport->worm,
				enh_ieport->full, enh_ieport->barcode_flag,
				enh_ieport->barcode_len, enh_ieport->barcode,
				elem_info);
		} /* if */

		if(copy_method == BackToKernel)
			memcpy(ieport_state, ieport, enh ?
				sizeof(lb_enh_import_export_state_t) :
				sizeof(lb_import_export_state_t));
		else if(copy_to_user((void*)ieport_state, (void*)ieport, enh ?
			sizeof(lb_enh_import_export_state_t) :
			sizeof(lb_import_export_state_t))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* else if */

		if(enh) ieport_state += sizeof(lb_enh_import_export_state_t);
		else ieport_state += sizeof(lb_import_export_state_t);
		offs += sizeof(struct tape_elem_status);
	} /* for */

	*user_buffer = (void*)ieport_state;

EXIT_LABEL:

	return rc;
} /* set_ieport_state() */

/*******************************************************************************
*                                                                              *
*  NAME: set_storage_state                                                     *
*                                                                              *
*  PURPOSE: copy data retrieved from RES back to user-supplied buffer          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    data_buf: the RES data                                                    *
*    count: the number of storage elements to transfer in this iteration       *
*    bar_code: True if barcode is valid                                        *
*    slot: temp buffer in which to organize RES data for user                  *
*    storage_state: the user buffer in which to copy the slot parameter        *
*    copy_method: kernel mode or user mode                                     *
*    enh: if True, enhanced inventory; if False, standard inventory            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_storage_state(struct changer_descriptor* chg, char* data_buf,
	ushort count, unchar bar_code, void* driver_buffer,
	void** user_buffer, char copy_method, boolean enh)
{
	int offs = 0, i = 0, rc = STATUS_SUCCESS;
	ushort asc_ascq = 0;
	struct tape_elem_status* elem_info = NULL;
	lb_enh_storage_element_state_t* enh_slot = NULL;
	lb_storage_element_state_t* slot = NULL;
	unchar* storage_state = NULL;

	enh_slot = (lb_enh_storage_element_state_t*)driver_buffer;
	slot = (lb_storage_element_state_t*)driver_buffer;
	storage_state = (unchar*)*user_buffer;

	for(i = 0, offs = 0; i < count; i++) {
		elem_info = (struct tape_elem_status*)(data_buf + offs);

		if((!storage_state) || (!elem_info)) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		/* set data in storage_state  */
		memset(slot, '\0', enh ?
			sizeof(lb_enh_storage_element_state_t) :
			sizeof(lb_storage_element_state_t));
		slot->address = elem_info->addr;
		slot->accessible = elem_info->flags1_4;
		slot->full = elem_info->flags1_7;

		slot->source_addr_flag = elem_info->flags3_0;
		slot->source_elem_addr = elem_info->source;
		slot->invert_media_flag = elem_info->flags3_1;

		asc_ascq = ((elem_info->asc & 0xFF) << 8)
			| (elem_info->ascq);
		slot->abnormal_state = changer_get_abnormal_state(asc_ascq);

		/* set bar code in storage_state */
		if(bar_code) set_barcode(chg, slot->barcode,
			&slot->barcode_flag, &slot->barcode_len, elem_info);

		/* support enhanced inventory for LTO devices */
		if(enh && (chg->chg_id == IBM_3584 ||
			chg->chg_id == IBM_3583 ||
			chg->chg_id == IBM_3582 ||
			chg->chg_id == IBM_3576 ||
			chg->chg_id == IBM_3573 ||
			chg->chg_id == IBM_3572 ||
			chg->chg_id == IBM_3577 ||
			chg->chg_id == QTM_500  ||
			chg->chg_id == QTM_6000 ||
			chg->chg_id == HP_MSLG3 ||
			chg->chg_id == HP_1x8G2 ||
			chg->chg_id == HP_MSL6480)) {

			set_media_type(&enh_slot->media_devtype,
				&enh_slot->element_devtype,
				&enh_slot->spec_media_type,
				&enh_slot->worm,
				enh_slot->full,
				enh_slot->barcode_flag,
				enh_slot->barcode_len,
				enh_slot->barcode,
				elem_info);
		} /* if */

		if(copy_method == BackToKernel)
			memcpy(storage_state, slot, enh ?
				sizeof(lb_enh_storage_element_state_t):
				sizeof(lb_storage_element_state_t));

		else if(copy_to_user((void*)storage_state, (void*)slot, enh ?
			sizeof(lb_enh_storage_element_state_t) :
			sizeof(lb_storage_element_state_t))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* else if */

		if(enh) storage_state += sizeof(lb_enh_storage_element_state_t);
		else storage_state += sizeof(lb_storage_element_state_t);
		offs += sizeof(struct tape_elem_status);
	} /* for */

	*user_buffer = (void*)storage_state;

EXIT_LABEL:

	return rc;
} /* set_storage_state() */

/*******************************************************************************
*                                                                              *
*  NAME: set_robot_state                                                       *
*                                                                              *
*  PURPOSE: copy data retrieved from RES back to user-supplied buffer          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    data_buf: the RES data                                                    *
*    req_robots: the number of robots requested by the user                    *
*    bar_code: True if barcode is valid                                        *
*    robot: temp buffer in which to organize RES data for user                 *
*    robot_state: the user buffer in which to copy the robot parameter         *
*    copy_method: kernel mode or user mode                                     *
*    enh: if True, enhanced inventory; if False, standard inventory            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_robot_state(struct changer_descriptor* chg, char* data_buf,
	ushort req_robots, unchar bar_code, void* driver_buffer,
	void** user_buffer, char copy_method, boolean enh)
{
	int offs = 0, i = 0, rc = STATUS_SUCCESS;
	ushort asc_ascq = 0;
	struct tape_elem_status* elem_info = NULL;
	lb_medium_changer_state_t* robot, * robot_state = NULL;

	robot_state = (lb_medium_changer_state_t*)*user_buffer;
	robot = (lb_medium_changer_state_t*)driver_buffer;

	for(i = 0; i < req_robots; i++) {
		elem_info = (struct tape_elem_status*)(data_buf + offs);
		if(!robot_state || !elem_info) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		/* set data in robot_state  */
		memset(robot, '\0', sizeof(lb_medium_changer_state_t));
		robot->address = elem_info->addr;
		robot->accessible = elem_info->flags1_4;
		robot->full = elem_info->flags1_7;

		robot->source_addr_flag = elem_info->flags3_0;
		robot->source_elem_addr = elem_info->source;
		robot->invert_media_flag = elem_info->flags3_1;

		asc_ascq = ((elem_info->asc & 0xFF) << 8) | (elem_info->ascq);
		robot->abnormal_state = changer_get_abnormal_state(asc_ascq);

		/* set bar code in robot_state */
		if(bar_code) set_barcode(chg, robot->barcode,
			&robot->barcode_flag, &robot->barcode_len, elem_info);

		if(copy_method == BackToKernel)
			memcpy(robot_state, robot,
				sizeof(lb_medium_changer_state_t));

		else if(copy_to_user((void*)robot_state, (void*)robot,
			sizeof(lb_medium_changer_state_t))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* else if */

		robot_state++;
		offs += sizeof(struct tape_elem_status);
	} /* for */

	*user_buffer = (void*)robot_state;

EXIT_LABEL:

	return rc;
} /* set_robot_state() */

/*******************************************************************************
*                                                                              *
*  NAME: inv_res                                                               *
*                                                                              *
*  PURPOSE: Execute the read element status for TSM enhanced inventory         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    elem_state: the user buffer in which to copy the final data               *
*    report_type: drives, slots, ie ports, or robots                           *
*    req_elems: the number of elements requested by the user                   *
*    copy_method: to kernel space or user space                                *
*    elem_asn: as returned from element address mode page                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int inv_res(struct changer_descriptor** chg, void* elem_state, int report_type,
	ushort req_elems, char copy_method, struct elem_assignment* elem_asn,
	boolean enh)
{
	int rc = STATUS_SUCCESS, alloc_len = 0, count = 0;
	uchar* saddr = NULL, * data_buf = NULL, bar_code = 0;
	char vol_tag = 0;
	ushort start_address = 0, remain = 0;
	void* elem = NULL;
	ulong buf_size = 0, order = 0;
	int (*set_elem_state)(struct changer_descriptor*, char*, ushort,
		unchar, void*, void**, char, boolean) = NULL;

	switch(report_type) {
	case DataTransferReport:
		DbgPrint(("lin_tape: inv_res, DataTransferReport\n"));
		alloc_len = enh ?
			sizeof(lb_enh_drive_state_t) :
			sizeof(lb_drive_state_t);
		saddr = elem_asn->first_data_transfer_addr;
		set_elem_state = set_drive_state;
		break;

	case StorageElementReport:
		DbgPrint(("lin_tape: inv_res, StorageElementReport\n"));
		alloc_len = enh ?
			sizeof(lb_enh_storage_element_state_t):
			sizeof(lb_storage_element_state_t);
		saddr = elem_asn->first_storage_addr;
		set_elem_state = set_storage_state;
		break;

	case IEPortReport:
		DbgPrint(("lin_tape: inv_res, IEPortReport\n"));
		alloc_len = enh ?
			sizeof(lb_enh_import_export_state_t) :
			sizeof(lb_import_export_state_t);
		saddr = elem_asn->first_ie_port_addr;
		set_elem_state = set_ieport_state;
		break;

	case MediumTransportReport:
		DbgPrint(("lin_tape: inv_res, MediumTransportReport\n"));
		alloc_len = sizeof(lb_medium_changer_state_t);
		saddr = elem_asn->medium_transport_elem_addr;
		set_elem_state = set_robot_state;
		break;

	default:
		printk("lin_tape: invalid report type %d\n", report_type);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* switch */

	elem = kmalloc(alloc_len, GFP_KERNEL);
	if(!elem) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* if number of elems requested is too large, obtain a few at a time */
	remain = req_elems;
	count = (remain < 600) ? remain : 600;

	alloc_len = count * sizeof(struct tape_elem_status);
	if(alloc_len > KmallocThreshhold) {
		order = get_mem_ord(alloc_len, &buf_size);
		data_buf = (unchar*)__get_free_pages(GFP_KERNEL, order);
	} else {
		data_buf = kmalloc(alloc_len, GFP_KERNEL);
		buf_size = alloc_len;
	} /* else */

	if(!data_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(data_buf, '\0', buf_size);
	start_address = (saddr[0] << 8) + saddr[1];

	while(remain > 0) {
		count = (remain < 600) ? remain : 600;

		rc = changer_read_elem_status(chg, data_buf, elem_asn,
			report_type, start_address, count, &vol_tag,
			BackToKernel, FALSE);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(vol_tag) bar_code = 1;

		rc = set_elem_state(*chg, data_buf, count, bar_code,
			elem, &elem_state, copy_method, enh);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		remain -= count;
		start_address += count;
	} /* while */

EXIT_LABEL:

	if(elem) kfree(elem);

	if(buf_size <= KmallocThreshhold && data_buf) kfree(data_buf);
	else if(data_buf) free_pages((ulong)data_buf, order);

	DbgPrint(("lin_tape: inv_res exit: %d\n", rc));
	return rc;
} /* inv_res() */

/*******************************************************************************
*                                                                              *
*  NAME: set_barcode                                                           *
*                                                                              *
*  PURPOSE: set barcode information for TSM enhanced library inventory         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor representing the library                      *
*    barcode: the field in which to copy the barcode                           *
*    barcode_flag: set to true if barcode valid, otherwise false               *
*    barcode_len: the length of the returned barcode                           *
*    elem_info: the raw descriptor from read_element_status                    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    return 6 char barcode for 3592 inside 3584 unless ibm3592_barcode == 8    *
*    return 8 char barcode for LTO by default                                  *
*    if lto_barcode is set to 6:                                               *
*      return 6 char barcode for LTO 1 & 2,                                    *
*      still return 8 char barcode for LTO 3 and later                         *
*                                                                              *
*******************************************************************************/

void set_barcode(struct changer_descriptor* chg, char* barcode,
	uchar* barcode_flag, int32_t* barcode_len,
	struct tape_elem_status* elem_info)
{
	int i = 0;

	memcpy(barcode, elem_info->volume, LB_BARCODE_LEN);

	if(barcode[6] == 'J') {
		i = (ibm3592_barcode == 8) ? 8 : 6;
	} else {
		if(lto_barcode != 8 && lto_barcode != 6)
			lto_barcode = 8;

		i = (barcode[6] == 'L' && (barcode[7] == '1' ||
			barcode[7] == '2')) ? lto_barcode : 8;

	} /* else */

	for(; i < LB_BARCODE_LEN; i++) barcode[i] = ' ';

	for(i = 0; i < LB_BARCODE_LEN; i++)
		if(barcode[i] == ' ' || barcode[i] == '\0') break;
	*barcode_len = i;

	if(i > 0) *barcode_flag = True;
	else *barcode_flag = False;

} /* set_barcode() */

/*******************************************************************************
*                                                                              *
*  NAME: set_media_type                                                        *
*                                                                              *
*  PURPOSE: set media type for TSM enhanced library inventory                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    media_devtype: the device type for the cartridge                          *
*    element_devtype: the device type for the element                          *
*    spec_media_type: the specific device type for the cartridge               *
*    worm: true if write-once, read-many; false otherwise                      *
*    full: true if there is a tape currently in the element                    *
*    barcode_flag: true if the barcode is valid                                *
*    barcode_len: the barcode length                                           *
*    barcode: the barcode itself                                               *
*    elem_info: the raw data from read_element_status                          *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void set_media_type(int32_t* media_devtype, int32_t* element_devtype,
	int32_t* spec_media_type, uchar* worm, uchar full, uchar barcode_flag,
	int32_t barcode_len, char* barcode, struct tape_elem_status* elem_info)
{
	*media_devtype = DEVTYPE_ANY; /* initial value */
	*element_devtype = DEVTYPE_ANY;

	if(!full || !barcode_flag || !barcode_len) {
		*spec_media_type = DD_UNKNOWN_MEDIA_TYPE;
		return;
	} /* if */

	if(elem_info->volume[6] == 'J') {
		*media_devtype = DEVTYPE_3592;

		switch(elem_info->volume[7]) {
		case 'A': case 'J':
			*spec_media_type = MT_3592_RW;
			break;

		case 'R': case 'W':
			*spec_media_type = MT_3592_WORM;
			*worm = 1;
			break;

		case 'B':
			*spec_media_type = MT_3592_RW2;
			break;

		case 'X':
			*spec_media_type = MT_3592_WORM2;
			*worm = 1;
			break;

		case 'C': case 'K':
			*spec_media_type = MT_3592_RW3;
			break;

		case 'Y':
			*spec_media_type = MT_3592_WORM3;
			*worm = 1;
			break;

		case 'D': case 'L':
			*spec_media_type = MT_3592_RW4;
			break;

		case 'Z':
			*spec_media_type = MT_3592_WORM4;
			*worm = 1;
			break;

		case 'M': case 'E':
			*spec_media_type = MT_3592_RW5;
			break;

		case 'V':
			*spec_media_type = MT_3592_WORM5;
			*worm = 1;
			break;

		default:
			*spec_media_type = DD_UNKNOWN_MEDIA_TYPE;
			break;
		} /* switch */
	} else {
		if(elem_info->volume[6] == 'L' || elem_info->volume[6] == 'M') *media_devtype = DEVTYPE_LTO;
		switch(barcode[7]) {
		case '1':
			*spec_media_type = MT_ULTRIUM;
			break;

		case '2':
			*spec_media_type = MT_ULTRIUM2;
			break;

		case 'T':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM3;
			break;

		case '3': /* fall-through for 'T' and '3' */
			*spec_media_type = MT_ULTRIUM3;
			break;

		case 'U':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM4;
			break;

		case '4': /* fall-through for 'U' and '4' */
			*spec_media_type = MT_ULTRIUM4;
			break;

		case 'V':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM5;
			break;

		case '5': /* fall-through for 'V' and '5' */
			*spec_media_type = MT_ULTRIUM5;
			break;

		case 'W':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM6;
			break;

		case '6': /* fall-through for 'W' and '6' */
			*spec_media_type = MT_ULTRIUM6;
			break;

		case 'X':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM7;
			break;

		case '7': /* fall-through for 'X' and '7' */
			*spec_media_type = MT_ULTRIUM7;
			break;

		case 'Y':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM8;
			if(elem_info->volume[6] == 'M') *spec_media_type = MT_ULTRIUM8M;
			break;

		case '8': /* fall-through for 'Y' and '8' */
			*spec_media_type = MT_ULTRIUM8;
			if(elem_info->volume[6] == 'M') *spec_media_type = MT_ULTRIUM8M;
			break;

		case 'Z':
			*worm = 1;
			*spec_media_type = MT_ULTRIUM9;
			break;

		case '9':
			*spec_media_type = MT_ULTRIUM9;
			break;

		default:
			*spec_media_type = DD_UNKNOWN_MEDIA_TYPE;
			break;
		} /* switch */

	} /* else */
	return;
} /* set_media_type() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_compcode                                                  *
*                                                                              *
*  PURPOSE: return a code based on errno and sk/asc/ascq                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    errno: the errno to translate                                             *
*    chg: changer descriptor of the changer for which the command was sent     *
*                                                                              *
*  RETURNS: compcode                                                           *
*                                                                              *
*******************************************************************************/

int changer_get_compcode(int errno, struct changer_descriptor *chg)
{
	int rc = DD_UNKNOWN_ERROR;
	struct tape_sense_data *sdata;

	sdata = chg->sense_data;

	if(errno == STATUS_SUCCESS) {
		rc = DD_NO_ERROR;
		goto EXIT_LABEL;
	} /* if */

	if(errno == EINVAL) rc = DD_ILLEGAL_REQUEST;
	if(errno == EBUSY) rc = DD_MANUAL_INTERVENTION;

	if (!chg) {
		rc = DD_UNKNOWN_ERROR;
		goto EXIT_LABEL;
	}
	if(!chg->sense_valid) {
		rc = DD_NO_SENSE;
		goto EXIT_LABEL;
	} /* if */

	switch(sdata->sense_key) {
	case NOT_READY:
		if(sdata->asc == 0x53 && sdata->ascq == 0x00)
			rc = DD_ENTRY_EXIT_FAILURE;
		break;

	case HARDWARE_ERROR:
		if(sdata->asc == 0x53 && sdata->ascq == 0x00)
			rc = DD_ENTRY_EXIT_FAILURE;
		else if(sdata->asc == 0x53 && sdata->ascq == 0x01)
			rc = DD_ENTRY_EXIT_FAILURE;
		else if(sdata->asc == 0x04 && sdata->ascq == 0x03)
			rc = DD_MANUAL_INTERVENTION;
		break;

	case ILLEGAL_REQUEST:
		if(sdata->asc == 0x3B && sdata->ascq == 0x0E)
			rc = DD_SOURCE_EMPTY;
		else if(sdata->asc == 0x3B && sdata->ascq == 0x0D)
			rc = DD_DESTINATION_FULL;
		else if(sdata->asc == 0x30 && sdata->ascq == 0x00)
			rc = DD_MEDIA_INCOMPATIBILITY;
		break;

	case UNIT_ATTENTION:
		rc = DD_UNIT_ATTENTION;
		break;

	case RECOVERED_ERROR:
		rc = DD_RECOVERED_ERROR;
		break;

	case ABORTED_COMMAND:
		rc = DD_COMMAND_ABORTED;
		break;

	} /* switch */

EXIT_LABEL:

    if (chg){
	    chg->comp_code = rc;
	    chg->comp_code_valid = TRUE;
    }

	return rc;
} /* changer_get_compcode() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_abnormal_state                                            *
*                                                                              *
*  PURPOSE: return translated asc/ascq                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    asc_ascq: from the sense data                                             *
*                                                                              *
*  RETURNS: abnormal state; default is LB_ABNORMAL_IGNORE                      *
*                                                                              *
*******************************************************************************/

int changer_get_abnormal_state(ushort asc_ascq)
{
	int rc = LB_ABNORMAL_IGNORE;

	switch(asc_ascq) {
	case 0x3003:
		rc = LB_CLEANER_INSTALLED;
		break;

	case 0x3B11:
		rc = LB_CART_HOLDER_NOT_PRESENT;
		break;

	case 0x1100:
		rc = LB_CAN_NOT_READ_LABEL;
		break;

	case 0x8002:
		rc = LB_ELEMENT_RESERVED;
		break;

	case 0x8100:
		rc = LB_BAR_CODE_LABEL_UNSURE | LB_FULL_STATUS_UNSURE;
		break;

	case 0x8200:
		rc = LB_DRIVE_NOT_PRESENT;
		break;

	case 0x8300:
		rc = LB_CAN_NOT_READ_LABEL;
		break;
	} /* switch */

	return rc;
} /* changer_get_abnormal_state() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsm_ioctl                                                    *
*                                                                              *
*  PURPOSE: Handle TSM-specific IOCTLs                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_tsm_ioctl(uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	DbgPrint(("lin_tape_tsm_ioctl entry\n"));

	switch(command) {
	case DDIOC_QUERY_PR_KEY:
		rc = lin_tape_tsm_query_pr_key(NULL, command, arg, 1);
		break;
	case DDIOC_CLEAR_RESERVE_KEY:
		rc = lin_tape_tsm_clear_reserve_key(command, arg);
		break;
	case DDIOC_PREEMP_ABORT:
		rc = lin_tape_tsm_preempt_abort(command, arg);
		break;
	default:
		DbgPrint(("lin_tape: lin_tape_tsm_ioctl unknown cmd 0x%x\n",
			command));
		rc = -EINVAL;
		break;
	} /* switch */

	DbgPrint(("lin_tape_tsm_ioctl exit: %d\n", rc));
	return rc;
} /* lin_tape_tsm_ioctl() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsm_query_pr_key                                             *
*                                                                              *
*  PURPOSE: Handle DDIOC_QUERY_PR_KEY IOCTL                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drv: Drive pointer if not pseudo device call                              *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*    pseudo: 1 if call from pseudo device                                      *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_tsm_query_pr_key(struct tape_descriptor* drv, uint command, ulong arg,
                              unchar pseudo)
{
	int rc = STATUS_SUCCESS;
	int pr_gen = 0;
	unchar type = 0;
	dd_query_prkey_t qry_pr;
	boolean tape_des_locked = False, copy_back = False;

	DbgPrint(("lin_tape_tsm_query_pr_key entry\n"));

	if(_IOC_SIZE(command) != sizeof(dd_query_prkey_t)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user(&qry_pr, (char*)arg, sizeof(dd_query_prkey_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	copy_back = True;
	if (pseudo){
	    rc = lin_tape_get_drv_by_sn(&drv, qry_pr.device_serial);
	    if(rc != STATUS_SUCCESS) {
		    qry_pr.compcode = DD_PR_OP_FAILURE;
		    goto EXIT_LABEL;
	    } /* if */
	}

	if(!drv) {
		/* From TSM's PR spec: If there is no PR, target_pr_key will
			be set to zero and function will return success */
		rc = STATUS_SUCCESS;
		memset(qry_pr.target_pr_key, '\0', PRKEY_SIZE);
		qry_pr.compcode = DD_NO_ERROR;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_read_reservation(&drv, qry_pr.target_pr_key, &pr_gen, &type);

	if(rc == -EINVAL || rc == -ENOSYS) {
		rc = STATUS_SUCCESS;
		qry_pr.compcode = DD_PR_NOT_SUPPORTED;
	} else if(rc == STATUS_SUCCESS) {
		qry_pr.compcode = DD_NO_ERROR;
	} else {
		qry_pr.compcode = DD_PR_OP_FAILURE;
	} /* else */

EXIT_LABEL:

	if(tape_des_locked) READ_UNLOCK(&tape_des_lock);

	if(copy_back) {
		if(copy_to_user((char*)arg, &qry_pr,
			sizeof(dd_query_prkey_t))) {
			if(rc == STATUS_SUCCESS) rc = -EFAULT;
		} /* if */
	} /* if */

	DbgPrint(("lin_tape_tsm_query_pr_key exit: %d\n", rc));
	return rc;
} /* lin_tape_tsm_query_pr_key() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsm_set_reserve_key                                          *
*                                                                              *
*  PURPOSE: Handle DDIOC_SET_RESERVE_KEY IOCTL                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the tape descriptor for the drive on which to perform the operation  *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_tsm_set_reserve_key(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS;
	boolean copy_back = False;
	dd_set_reserve_key_t set_pr;

	DbgPrint(("lin_tape_tsm_set_reserve_key entry\n"));

	if(_IOC_SIZE(command) != sizeof(dd_set_reserve_key_t)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user(&set_pr, (char*)arg, sizeof(dd_set_reserve_key_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	copy_back = True;

	rc = tape_register_reservation_key(&drv, set_pr.tsm_pr_key);

	if(rc == -EINVAL || rc == -ENOSYS) {
		rc = STATUS_SUCCESS;
		set_pr.compcode = DD_PR_NOT_SUPPORTED;
	} else if(rc == STATUS_SUCCESS) {
		set_pr.compcode = DD_NO_ERROR;
	} else {
		set_pr.compcode = DD_PR_OP_FAILURE;
	} /* else */

	goto EXIT_LABEL; /* avoid unreferenced label */

EXIT_LABEL:

	if(copy_back) {
		if(copy_to_user((char*)arg, &set_pr,
			sizeof(dd_set_reserve_key_t))) {
			if(rc == STATUS_SUCCESS) rc = -EFAULT;
		} /* if */
	} /* if */

	DbgPrint(("lin_tape_tsm_set_reserve_key exit: %d\n", rc));
	return rc;
} /* lin_tape_tsm_set_reserve_key() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsm_clear_reserve_key                                        *
*                                                                              *
*  PURPOSE: Handle DDIOC_CLEAR_RESERVE_KEY IOCTL                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_tsm_clear_reserve_key(uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	dd_clear_reserve_t clear_pr;
	struct tape_descriptor* drv = NULL;
	boolean copy_back = False;

	DbgPrint(("lin_tape_tsm_clear_reserve_key entry\n"));

	if(_IOC_SIZE(command) != sizeof(dd_clear_reserve_t)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user(&clear_pr, (char*)arg, sizeof(dd_clear_reserve_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	copy_back = True;

	rc = lin_tape_get_drv_by_sn(&drv, clear_pr.device_serial);
	if(rc != STATUS_SUCCESS) {
		clear_pr.compcode = DD_PR_OP_FAILURE;
		goto EXIT_LABEL;
	} /* if */

	if(!drv) {
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_clear_all_registrations(&drv);

	if(rc == -EINVAL || rc == -ENOSYS) {
		rc = STATUS_SUCCESS;
		clear_pr.compcode = DD_PR_NOT_SUPPORTED;
	} else if(rc == STATUS_SUCCESS) {
		clear_pr.compcode = DD_NO_ERROR;
	} else {
		clear_pr.compcode = DD_PR_OP_FAILURE;
	} /* else */

EXIT_LABEL:

	if(copy_back) {
		if(copy_to_user((char*)arg, &clear_pr,
			sizeof(dd_clear_reserve_t))) {
			if(rc == STATUS_SUCCESS) rc = -EFAULT;
		} /* if */
	} /* if */

	DbgPrint(("lin_tape_tsm_clear_reserve_key exit: %d\n", rc));
	return rc;
} /* lin_tape_tsm_clear_reserve_key() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsm_preempt_abort                                            *
*                                                                              *
*  PURPOSE: Handle DDIOC_PREEMP_ABORT IOCTL                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_tsm_preempt_abort(uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	dd_preempt_abort_t prm_pr;
	struct tape_descriptor* drv = NULL;
	boolean copy_back = False;

	DbgPrint(("lin_tape_tsm_preempt_abort entry\n"));

	if(_IOC_SIZE(command) != sizeof(dd_preempt_abort_t)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user(&prm_pr, (char*)arg, sizeof(dd_preempt_abort_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	copy_back = True;

	rc = lin_tape_get_drv_by_sn(&drv, prm_pr.device_serial);
	if(rc != STATUS_SUCCESS) {
		prm_pr.compcode = DD_PR_OP_FAILURE;
		goto EXIT_LABEL;
	} /* if */

	if(!drv) {
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_register_reservation_key(&drv, prm_pr.tsm_pr_key);

	if(rc == -EINVAL || rc == -ENOSYS) {
		rc = STATUS_SUCCESS;
		prm_pr.compcode = DD_PR_NOT_SUPPORTED;
		goto EXIT_LABEL;
	} else if(rc != STATUS_SUCCESS) {
		prm_pr.compcode = DD_PR_OP_FAILURE;
		goto EXIT_LABEL;
	} /* else if */

	rc = tape_preempt_reservation(&drv, prm_pr.target_pr_key,
		prm_pr.tsm_pr_key);

	if(rc == STATUS_SUCCESS) {
		prm_pr.compcode = DD_NO_ERROR;
	} else {
		prm_pr.compcode = DD_PR_OP_FAILURE;
	} /* else */

EXIT_LABEL:

	if(copy_back) {
		if(copy_to_user((char*)arg, &prm_pr,
			sizeof(dd_preempt_abort_t))) {
			if(rc == STATUS_SUCCESS) rc = -EFAULT;
		} /* if */
	} /* if */

	DbgPrint(("lin_tape_tsm_preempt_abort exit: %d\n", rc));
	return rc;
} /* lin_tape_tsm_preempt_abort() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_drv_by_sn                                                *
*                                                                              *
*  PURPOSE: Return tape_descriptor* from drive table based on serial number    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: (output) the descriptor to return                                    *
*    sn: the serial number for which to find the tape descriptor               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_get_drv_by_sn(struct tape_descriptor** drv, char* sn)
{
	int rc = STATUS_SUCCESS, i = 0;
	boolean tape_des_locked = False;

	READ_LOCK(&tape_des_lock);
	tape_des_locked = True;

	*drv = NULL;

	if(!tape_devs) {
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv ||
        !tape_devs[i]->drv->serial_num_len) continue;
		if(!memcmp(sn, tape_devs[i]->drv->serial_num,
			tape_devs[i]->drv->serial_num_len)) {
			*drv = tape_devs[i]->drv;
			break;
		} /* if */
	} /* for */

	tape_des_locked = False;
	READ_UNLOCK(&tape_des_lock);

EXIT_LABEL:

	if(tape_des_locked) READ_UNLOCK(&tape_des_lock);

	return rc;
} /* lin_tape_get_drv_by_sn() */
