/*******************************************************************************
 *   lin_tape_ioctl_changer.c
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#include "lin_tape_scsi_ioctl.h"
#include "lin_tape_timeout_changer.h"
#include "lin_tape_scsi_tape.h"

/*******************************************************************************
*                                                                              *
*  NAME: changer_inquiry_page                                                  *
*                                                                              *
*  PURPOSE: issue inquiry on specified page                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    page_code: the inquiry page for which we are retrieving data              *
*    inq_page: the buffer to receive data                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_inquiry_page(struct changer_descriptor** chgp, unchar page_code,
	unchar* inq_page)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;

	if(!(*chgp)->cdb || !inq_page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	(*chgp)->buffer = inq_page;
	(*chgp)->buf_len = MAX_INQ_LEN;

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_inquiry.op_code = INQUIRY;
	(*chgp)->cdb->cdb_inquiry.evpd = True;
	(*chgp)->cdb->cdb_inquiry.page_code = page_code;
	(*chgp)->cdb->cdb_inquiry.alloc_length = MAX_INQ_LEN;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = INQUIRY;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		InquiryTimeOut), DMA_FROM_DEVICE, False, 0);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(!(*chgp)->buffer) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data((*chgp), min((inq_page[3] & 0xFF) + 4,
			FullTraceUnitSize - 7), (*chgp)->buffer);
EXIT_LABEL:

//	*chgp = chg;
    if (chgp && *chgp){
	    (*chgp)->buf_len = 0;
	    (*chgp)->buffer = NULL;
    }
	return rc;
} /* changer_inquiry_page() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_logsense_page                                                 *
*                                                                              *
*  PURPOSE: issue logsense on specified page                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    page_ctr: value for the page control bits                                 *
*    page_code: the logsense page for which we are retrieving data             *
*    buffer: the buffer to receive data                                        *
*    length: the length of buffer in bytes                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_logsense_page(struct changer_descriptor** chgp, unchar page_ctr,
	unchar page_code, unchar *buffer, uint* length)
{
//	struct changer_descriptor* chg = *chgp;
	unchar *page = NULL;
	int rc = STATUS_SUCCESS, i = 0, to = 0;
	char ttype = lin_tape_trace_type;

	page = kmalloc(LogPageHeaderSize + LOGSENSEPAGE, GFP_KERNEL);
	if(!page || !(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', LogPageHeaderSize + LOGSENSEPAGE);

	(*chgp)->buffer = (void*)page;
	(*chgp)->buf_len = LogPageHeaderSize + LOGSENSEPAGE;

	to = compute_changer_timeout((*chgp)->chg_id, LogSenseTimeOut);
	for(i = 0; i < 2; i++) {
		memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
		(*chgp)->cdb->cdb_logsense.op_code = LOG_SENSE;
		(*chgp)->cdb->cdb_logsense.sp = False;
		(*chgp)->cdb->cdb_logsense.ppc = False;
		(*chgp)->cdb->cdb_logsense.pc = i ? True : page_ctr;
		(*chgp)->cdb->cdb_logsense.page_code = page_code;
		(*chgp)->cdb->cdb_logsense.alloc_length[0] =
			i ? (unchar)(*length >> 8 & 0xFF) : 0;
		(*chgp)->cdb->cdb_logsense.alloc_length[1] =
			i ? (unchar)(*length & 0xFF) : LogPageHeaderSize;
		(*chgp)->cdb_len = CDB_GENERIC_10;
		(*chgp)->last_op = LOG_SENSE;

		rc = changer_send_scsi_cmd(chgp, to, DMA_FROM_DEVICE, False, 0);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		page = (*chgp)->buffer;
		if(!page) {
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */

		if(ttype == LIN_TAPE_FULL_TRACE)
			changer_log_buffer_data((*chgp), i ? *length :
				min((int)((*chgp)->cdb->cdb_logsense.
					alloc_length[1]),
				FullTraceUnitSize - 7), (*chgp)->buffer);

		(*chgp)->buf_len = LogPageHeaderSize + LOGSENSEPAGE;
		*length = (page[2] << 8) + page[3] + LogPageHeaderSize;

		if(*length > LogPageHeaderSize + LOGSENSEPAGE)
			*length = LogPageHeaderSize + LOGSENSEPAGE;

	} /* for */

	memcpy(buffer, page, *length);

EXIT_LABEL:

//	*chgp = chg;
	(*chgp)->buf_len = 0;
	if(page) kfree(page);
	(*chgp)->buffer = NULL;
	return rc;
} /* changer_logsense_page() */
/*******************************************************************************
*                                                                              *
*  NAME: changer_logsense_page                                                 *
*                                                                              *
*  PURPOSE: issue logsense on specified page                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    page_ctr: value for the page control bits                                 *
*    page_code: the logsense page for which we are retrieving data             *
*    buffer: the buffer to receive data                                        *
*    length: the length of buffer in bytes                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_logsense10_page(struct changer_descriptor** chgp, unchar page_code,
	unchar subpage_code, char* buffer,  unsigned short param_pointer, uint* length)
{
	unchar *page = NULL;
	int rc = STATUS_SUCCESS, i = 0, to = 0;
	char ttype = lin_tape_trace_type;
	uint len = *length > 0? *length : LOGSENSEPAGE;

	if (!(*chgp)->cdb){
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	page = kmalloc(len, GFP_KERNEL);
	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', len);

	(*chgp)->buffer = (void*)page;
	(*chgp)->buf_len = len;

	to = compute_changer_timeout((*chgp)->chg_id, LogSenseTimeOut);
	for(i = 0; i < 2; i++) {
		memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
		(*chgp)->cdb->cdb_logsense.op_code = LOG_SENSE;
		(*chgp)->cdb->cdb_logsense.sp = False;
		(*chgp)->cdb->cdb_logsense.ppc = False;
		(*chgp)->cdb->cdb_logsense.pc = True;
		(*chgp)->cdb->cdb_logsense.page_code = page_code;
		(*chgp)->cdb->cdb_logsense.subpage_code = subpage_code;
		(*chgp)->cdb->cdb_logsense.parm_ptr[0] = (unchar)(param_pointer >> 8 & 0xFF);
		(*chgp)->cdb->cdb_logsense.parm_ptr[1] = (unchar)(param_pointer & 0xFF0);
		(*chgp)->cdb->cdb_logsense.alloc_length[0] =
			i ? (unchar)(len >> 8 & 0xFF) : 0;
		(*chgp)->cdb->cdb_logsense.alloc_length[1] =
			i ? (unchar)(len & 0xFF) : LogPageHeaderSize;
		(*chgp)->cdb_len = CDB_GENERIC_10;
		(*chgp)->last_op = LOG_SENSE;

		rc = changer_send_scsi_cmd(chgp, to, DMA_FROM_DEVICE, False, 0);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		page = (*chgp)->buffer;
		if(!page) {
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */

		if(ttype == LIN_TAPE_FULL_TRACE)
			changer_log_buffer_data((*chgp), i ? len :
				min((int)((*chgp)->cdb->cdb_logsense.
					alloc_length[1]),
				FullTraceUnitSize - 7), (*chgp)->buffer);

		(*chgp)->buf_len = len;
		len = (page[2] << 8) + page[3];

		if(len > LOGSENSEPAGE)
			len = LOGSENSEPAGE;

	} /* for */

	memcpy(buffer, page, len);
	*length = len;

EXIT_LABEL:

//	*chgp = chg;
	(*chgp)->buf_len = 0;
	if(page) kfree(page);
	(*chgp)->buffer = NULL;
	return rc;
} /* changer_logsense10_page() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_modesense_page                                                *
*                                                                              *
*  PURPOSE: issue modesense on specified page                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    page_code: the modesense page to retrieve                                 *
*    buffer: the buffer to receive data                                        *
*    length: the length of buffer in bytes                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_modesense_page(struct changer_descriptor** chgp,
	unchar page_code, unchar* buffer, uint length)
{
//	struct changer_descriptor* chg = *chgp;
	unchar *page = NULL;
	int rc = STATUS_SUCCESS;
        uint max_cpy = MAX_MDSNS_LEN - sizeof(struct mode_parm_header);
	char ttype = lin_tape_trace_type;

	page = kmalloc(MAX_MDSNS_LEN, GFP_KERNEL);
	if(!page || !(*chgp)->cdb || !buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	memset(page, '\0', MAX_MDSNS_LEN);

	(*chgp)->buffer = (void*)page;
        (*chgp)->buf_len = ((length && length <= max_cpy) ?
                length + sizeof(struct mode_parm_header) : MAX_MDSNS_LEN);

	(*chgp)->cdb->cdb_modesense.op_code = MODE_SENSE;
	(*chgp)->cdb->cdb_modesense.dbd = True;
	(*chgp)->cdb->cdb_modesense.page_code = page_code;
	(*chgp)->cdb->cdb_modesense.alloc_length = (*chgp)->buf_len;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = MODE_SENSE;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ModeSenseTimeOut), DMA_FROM_DEVICE, False, 0);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: changer_modesense_page: mode sense failed\n");
		goto EXIT_LABEL;
	} /* if */

        memcpy(buffer, &page[4], (length && length <= max_cpy) ?
                length : max_cpy);

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data((*chgp), min((int)((page[5] & 0xFF) + 2),
			FullTraceUnitSize - 7), (*chgp)->buffer);
EXIT_LABEL:

//	*chgp = chg;
	(*chgp)->buf_len = 0;
	if(page) kfree(page);
	(*chgp)->buffer = NULL;
	return rc;
} /* changer_modesense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_release                                                       *
*                                                                              *
*  PURPOSE: release changer using standard SCSI-2 release                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_release(struct changer_descriptor** chgp)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_release entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_release.op_code = RELEASE;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = RELEASE;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ReleaseTimeOut), DMA_NONE, False, 0);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_CG_RELEASED,
		NULL, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_release exit: %d\n", rc));
	return rc;
} /* changer_release() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_reserve                                                       *
*                                                                              *
*  PURPOSE: reserve changer using standard SCSI-2 reserve                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_reserve(struct changer_descriptor** chgp)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_reserve entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_reserve.op_code = RESERVE;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = RESERVE;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ReserveTimeOut), DMA_NONE, False, 0);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_CG_RESERVED,
		NULL, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_reserve exit: %d\n", rc));
	return rc;
} /* changer_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_std_inquiry                                                   *
*                                                                              *
*  PURPOSE: obtain standard inquiry (p. 0x0) data                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    buffer: output buffer                                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_std_inquiry(struct changer_descriptor** chgp, void* buffer)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: changer_std_inquiry entry\n"));

	if(!(*chgp)->cdb || !buffer) {
		printk("lin_tape changer_std_inquiry: cdb %p buffer %p\n",
			(*chgp)->cdb, buffer);
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	(*chgp)->buffer = buffer;
	(*chgp)->buf_len = sizeof(struct st_inquiry);

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_inquiry.op_code = INQUIRY;
	(*chgp)->cdb->cdb_inquiry.alloc_length = sizeof(struct st_inquiry);
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = INQUIRY;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		InquiryTimeOut), DMA_FROM_DEVICE, False, 0);

	if(!(*chgp)->buffer) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /*if */

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data((*chgp),
			min((int)(((struct st_inquiry*)buffer)->
				standard.additional_length + 5),
				FullTraceUnitSize - 7), (*chgp)->buffer);
EXIT_LABEL:

//	*chgp = chg;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	DbgPrint(("lin_tape: changer_std_inquiry ioctl exit: %d\n", rc));
	return rc;
} /* changer_std_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_tur                                                           *
*                                                                              *
*  PURPOSE: issue test unit ready to medium changer                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_tur(struct changer_descriptor** chgp)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_tur entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_tur.op_code = TEST_UNIT_READY;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->last_op = TEST_UNIT_READY;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		TestUnitReadyTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_tur exit: %d\n", rc));
	return rc;

} /* changer_tur() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_pos_element                                                   *
*                                                                              *
*  PURPOSE: position medium transport to specified element                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    robot_addr: the identifier of the medium transport element                *
*    destination: the destination element                                      *
*    invert: should always be 0                                                *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_pos_element(struct changer_descriptor** chgp, ushort robot_addr,
	ushort destination, unchar invert)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_pos_element entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_pos_elem.op_code = POSITION_TO_ELEMENT;
	(*chgp)->cdb->cdb_pos_elem.transport_elem_addr[0] =
		(unchar)(robot_addr >> 8);
	(*chgp)->cdb->cdb_pos_elem.transport_elem_addr[1] =
		(unchar)(robot_addr);
	(*chgp)->cdb->cdb_pos_elem.destination_elem_addr[0] =
		(unchar)(destination >> 8);
	(*chgp)->cdb->cdb_pos_elem.destination_elem_addr[1] =
		(unchar)(destination);
	(*chgp)->cdb->cdb_pos_elem.invert = invert;

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->last_op = POSITION_TO_ELEMENT;
	(*chgp)->cdb_len = CDB_GENERIC_10;
	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		PositionToElementTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_pos_element entry\n"));
	return rc;
} /* changer_pos_element() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_init_elem_range                                               *
*                                                                              *
*  PURPOSE: initialize element status with range                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    elem_addr: starting element of range                                      *
*    elem_num: number of elements for which to initialize status               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_init_elem_range(struct changer_descriptor** chgp,
	ushort elem_addr, ushort elem_num)
{
//	struct changer_descriptor* chg = *chgp;
	int rc;

	DbgPrint(("lin_tape: changer_init_elem_range entry\n"));

	if((*chgp)->chg_id == IBM_3590) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	if ((*chgp)->chg_id == IBM_3573 ||
		(*chgp)->chg_id == HP_MSLG3 ||
		(*chgp)->chg_id == HP_1x8G2 ||
		(*chgp)->chg_id == HP_MSL6480)
		(*chgp)->cdb->cdb_init_elem_range.op_code = INIT_ELE_W_RANGE_3573;
	else
		(*chgp)->cdb->cdb_init_elem_range.op_code = INIT_ELE_W_RANGE;
	(*chgp)->cdb->cdb_init_elem_range.range = True;
	(*chgp)->cdb->cdb_init_elem_range.start_elem_addr[0] =
		(unchar)(elem_addr >> 8);
	(*chgp)->cdb->cdb_init_elem_range.start_elem_addr[1] = (unchar)(elem_addr);
	((*chgp)->cdb)->cdb_init_elem_range.elem_num[0] = (unchar) (elem_num >> 8);
	((*chgp)->cdb)->cdb_init_elem_range.elem_num[1] = (unchar) (elem_num);
	(*chgp)->cdb_len = CDB_GENERIC_10;

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	if ((*chgp)->chg_id == IBM_3573 ||
		(*chgp)->chg_id == HP_MSLG3 ||
		(*chgp)->chg_id == HP_1x8G2 ||
		(*chgp)->chg_id == HP_MSL6480)
		(*chgp)->last_op = INIT_ELE_W_RANGE_3573;
	else (*chgp)->last_op = INIT_ELE_W_RANGE;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		InitializeElementWRangeTimeOut), DMA_NONE, 0, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_init_elem_range exit: %d\n", rc));
	return rc;
} /* changer_init_elem_range() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_init_element                                                  *
*                                                                              *
*  PURPOSE: initialize element status                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_init_element(struct changer_descriptor** chgp)
{
//	struct changer_descriptor* chg = *chgp;
	int rc;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: changer_init_element entry\n"));
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg((*chgp)->chg_log, &(*chgp)->chg_log_len, TRC_SMCIOC_IES);
	else if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_full_trace((*chgp), TRC_SMCIOC_IES, 0, NULL);

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_init_elem.op_code = INITIALIZE_ELEMENT_STATUS;
	(*chgp)->cdb_len = CDB_GENERIC_6;

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->last_op = INITIALIZE_ELEMENT_STATUS;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		InitializeElementTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_init_element exit: %d\n", rc));
	return rc;
} /* changer_init_element() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_inventory                                                     *
*                                                                              *
*  PURPOSE: Return inventory in library for specified elements                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    robot: buffer for medium transport elements                               *
*    slot: buffer for storage elements                                         *
*    ieport: buffer for import/export elements                                 *
*    drive: buffer for data transfer elements                                  *
*    copy_method: to user or kernel space                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_inventory(struct changer_descriptor** chgp, void* robot,
	void* slot, void* ieport, void* drive, unchar copy_method)
{
//	struct changer_descriptor* chg = *chgp;
	struct elem_assignment* elem_assignmnt = NULL;
	ushort robot_num = 0, slot_num = 0, ieport_num = 0, drive_num = 0;
	ushort start_address = 0, remain = 0;
	int rc = STATUS_SUCCESS, offset = 0, count = 0;
	char vol_tag = 0;

	DbgPrint(("lin_tape: changer_inventory entry\n"));

	elem_assignmnt = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!elem_assignmnt) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_assignmnt, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(chgp, ElementAddressPage,
		(void*)elem_assignmnt, sizeof(struct elem_assignment));

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s changer_inventory: mode sense failed\n",
			(*chgp)->devid);
		goto EXIT_LABEL;
	} /* if */

	robot_num = (elem_assignmnt->transport_elem_num[0] << 8)
		+ elem_assignmnt->transport_elem_num[1];
	slot_num = (elem_assignmnt->storage_elem_num[0] << 8)
		+ elem_assignmnt->storage_elem_num[1];
	ieport_num = (elem_assignmnt->ie_port_elem_num[0] << 8)
		+ elem_assignmnt->ie_port_elem_num[1];
	drive_num = (elem_assignmnt->data_transfer_num[0] << 8)
		+ elem_assignmnt->data_transfer_num[1];

	DbgPrint(("lin_tape: IBMchanger%d, robots %d, slots %d, ieports %d, "\
		"drives %d\n", (*chgp)->dev_index, robot_num, slot_num, ieport_num,
		drive_num));

	if(!(*chgp)->cdb) {
		rc =  -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* get the robot elements */
	if(robot) {
		start_address = (elem_assignmnt->medium_transport_elem_addr[0]
			<< 8) + elem_assignmnt->medium_transport_elem_addr[1];
		rc = changer_read_elem_status(chgp, robot, elem_assignmnt,
			MediumTransportReport, start_address, robot_num,
			&vol_tag, copy_method, FALSE);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(ieport) {
		start_address = (elem_assignmnt->first_ie_port_addr[0] << 8)
			+ elem_assignmnt->first_ie_port_addr[1];
		rc = changer_read_elem_status(chgp, ieport, elem_assignmnt,
			IEPortReport, start_address, ieport_num, &vol_tag,
			copy_method, FALSE);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(drive) {
		start_address =  (elem_assignmnt->first_data_transfer_addr[0]
			<< 8) + elem_assignmnt->first_data_transfer_addr[1];
		rc = changer_read_elem_status(chgp, drive, elem_assignmnt,
			DataTransferReport, start_address, drive_num, &vol_tag,
			copy_method, FALSE);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(slot) {
		remain = slot_num;
		start_address =  (elem_assignmnt->first_storage_addr[0] << 8)
			+ elem_assignmnt->first_storage_addr[1];
		while(remain > 0) {
			count = (remain < 600) ? remain : 600;
			rc = changer_read_elem_status(chgp, slot + offset,
				elem_assignmnt, StorageElementReport,
				start_address, count, &vol_tag, copy_method,
				FALSE);

			remain -= count;
			start_address += count;
			offset += count * sizeof(struct element_status);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* while */
	} /* if */

EXIT_LABEL:

//	*chgp = chg;

	if(elem_assignmnt) kfree(elem_assignmnt);
	DbgPrint(("lin_tape: changer_inventory exit: %d\n", rc));
	return rc;
} /* changer_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_read_cartridge_location                                       *
*                                                                              *
*  PURPOSE: Return cartridge location data for high density libraries          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: changer descriptor                                                   *
*    arg: input/output structure for cartridge location data                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_read_cartridge_location(struct changer_descriptor* chg, char* arg,
	unchar copy_method)
{
	int rc = STATUS_SUCCESS;
	struct read_cartridge_location* cart_loc = NULL;
	unchar vol_tag = 1;
	ushort remaining = 0;
	int count = 0;
	int elem_offset = 0;
	int start_address = 0;
	char ttype = lin_tape_trace_type;

	DbgPrint(("changer_read_cartridge_location entry\n"));
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(chg->chg_log, &chg->chg_log_len, TRC_SMCIOC_RES);

	else if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_full_trace(chg, TRC_SMCIOC_RES, 0, NULL);

	cart_loc = kmalloc(sizeof(struct read_cartridge_location), GFP_KERNEL);

	if(!cart_loc) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user((char*)cart_loc, arg,
		sizeof(struct read_cartridge_location))) {
		kfree(cart_loc);
		cart_loc = NULL;
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!cart_loc->data || !cart_loc->number_elements) {
		rc = -EINVAL;
		kfree(cart_loc);
		cart_loc = NULL;
		goto EXIT_LABEL;
	} /* if */

	remaining = cart_loc->number_elements;
	start_address = cart_loc->element_address;

	while(remaining > 0) {
		count = (remaining < 600) ? remaining : 600;
		rc = changer_read_elem_status(&chg,
			(unchar*)cart_loc->data + elem_offset,
			NULL, StorageElementReport,
			start_address, count, &vol_tag, copy_method, TRUE);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		remaining -= count;
		start_address += count;
		elem_offset += count * sizeof(struct cartridge_location_data);
	} /* while */

EXIT_LABEL:

	if(cart_loc) {
		kfree(cart_loc);
		cart_loc = NULL;
	} /* if */

	DbgPrint(("changer_read_cartridge_location exit: %d\n", rc));
	return rc;
} /* changer_read_cartridge_location() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_move_medium                                                   *
*                                                                              *
*  PURPOSE: move medium from source to destination                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    robot: medium transport element address                                   *
*    source: source element address                                            *
*    destination: destination element address                                  *
*    invert: invert tape before insertion (unsupported -- should be 0)         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_move_medium(struct changer_descriptor** chgp, ushort robot,
	ushort source, ushort destination, ushort invert)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_move_medium entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_move_medium.op_code = MOVE_MEDIUM;
	(*chgp)->cdb->cdb_move_medium.transport_elem_addr[0] = robot >> 8;
	(*chgp)->cdb->cdb_move_medium.transport_elem_addr[1] = robot;
	(*chgp)->cdb->cdb_move_medium.src_addr[0] = source >> 8;
	(*chgp)->cdb->cdb_move_medium.src_addr[1] = source;
	(*chgp)->cdb->cdb_move_medium.dest_addr[0] = destination >> 8;
	(*chgp)->cdb->cdb_move_medium.dest_addr[1] = destination;
	(*chgp)->cdb->cdb_move_medium.invert = invert;
	(*chgp)->cdb_len = CDB_GENERIC_12;
	(*chgp)->last_op = MOVE_MEDIUM;

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		MoveMediumTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_move_medium exit: %d\n", rc));
	return rc;
} /* changer_move_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_prevent_allow_remove                                          *
*                                                                              *
*  PURPOSE: lock or unlock the library door                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    prevent: if True, prevent removal; if False, allow removal                *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_prevent_allow_remove(struct changer_descriptor **chgp,
	unchar prevent)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: changer_prevent_allow_remove entry\n"));
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg((*chgp)->chg_log, &(*chgp)->chg_log_len,
			prevent ? TRC_SMCIOC_PREVENT : TRC_SMCIOC_ALLOW);
	else if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_full_trace((*chgp),
			prevent ? TRC_SMCIOC_PREVENT : TRC_SMCIOC_ALLOW, 0,
			NULL);

	if((*chgp)->chg_id == IBM_3590) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_medium_removal.op_code = ALLOW_MEDIUM_REMOVAL;
	(*chgp)->cdb->cdb_medium_removal.prevent = prevent;
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->last_op = ALLOW_MEDIUM_REMOVAL;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		PreventAllowMediaTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_prevent_allow_remove exit: %d\n", rc));
	return rc;
} /* changer_prevent_allow_remove() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_exchange_medium                                               *
*                                                                              *
*  PURPOSE: exchange media within library                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    robot: identifier of the medium transport element                         *
*    source: source element                                                    *
*    destination1: final location of the cartridge that started at source      *
*    destination2: final location of the cartridge that started at dest. 1     *
*    invert1: invert dest1 tape before insertion (unsupported -- should be 0)  *
*    invert2: invert dest2 tape before insertion (unsupported -- should be 0)  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_exchange_medium(struct changer_descriptor** chgp, ushort robot,
	ushort source, ushort destination1, ushort destination2, char invert1,
	char invert2)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_exchange_medium entry\n"));

	if((*chgp)->chg_id != IBM_3584) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_exchange_medium.op_code = EXCHANGE_MEDIUM;
	(*chgp)->cdb->cdb_exchange_medium.transport_elem_addr[0] = robot >> 8;
	(*chgp)->cdb->cdb_exchange_medium.transport_elem_addr[1] = robot & 0xFF;
	(*chgp)->cdb->cdb_exchange_medium.src_addr[0] = source >> 8;
	(*chgp)->cdb->cdb_exchange_medium.src_addr[1] = source & 0xFF;
	(*chgp)->cdb->cdb_exchange_medium.first_dest_addr[0] = destination1 >> 8;
	(*chgp)->cdb->cdb_exchange_medium.first_dest_addr[1] = destination1 & 0xFF;
	(*chgp)->cdb->cdb_exchange_medium.second_dest_addr[0] = destination2 >> 8;
	(*chgp)->cdb->cdb_exchange_medium.second_dest_addr[1] = destination2 & 0xFF;
	(*chgp)->cdb->cdb_exchange_medium.invert1 = invert1;
	(*chgp)->cdb->cdb_exchange_medium.invert2 = invert2;
	(*chgp)->cdb_len = CDB_GENERIC_12;

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->last_op = EXCHANGE_MEDIUM;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ExchangeMediumTimeOut), DMA_NONE, False, 0);

EXIT_LABEL:

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_exchange_medium exit: %d\n", rc));
	return rc;
} /* changer_exchange_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_read_devid                                                    *
*                                                                              *
*  PURPOSE: get device IDs for all drives in this changer                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    start: starting element address for the read element status               *
*    elem_num: number of elements to retrieve                                  *
*    output: user buffer into which to copy the data                           *
*    copy_method: to kernel or to user                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_read_devid(struct changer_descriptor** chgp, ushort start,
	ushort elem_num, void* output, unchar copy_method)
{
//	struct changer_descriptor* chg = *chgp;
	struct tape_read_devid* elem_info = NULL;
	uint drive_num = 0, drive_info_size = 0, offset = 0;
	uint elem_desc_length = 0, byte_offset = 0;
	ulong order = 0;
	unchar* buffer = NULL;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_read_devid entry\n"));

	rc = changer_get_num_drives(chgp, &drive_num);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = chg_devid_get_drive_report(chgp, drive_num, start, elem_num,
		&order);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	buffer = (*chgp)->buffer;
	offset = sizeof(struct elem_status_data);
	if(buffer[offset] != DataTransferReport) {
		printk("lin_tape: %s changer_read_devid unknown "\
			"report format\n", (*chgp)->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	drive_info_size = (buffer[offset + 5] << 16)
		+ (buffer[offset + 6] << 8) + buffer[offset + 7];

	if(!output || !drive_info_size) {
		printk("lin_tape: %s changer_read_devid cannot get "\
			"drive information\n", (*chgp)->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	elem_desc_length = (buffer[offset + 2] << 8) + buffer[offset + 3];

	if(((drive_info_size % elem_desc_length) ||
		elem_desc_length < MinDeviceID) && (*chgp)->chg_id != IBM_3581) {
		printk("lin_tape: %s changer_read_devid unknown element "\
			"report format\n", (*chgp)->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	elem_info = kmalloc(sizeof(struct tape_read_devid), GFP_KERNEL);
	if(!elem_info) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_info, '\0', sizeof(struct tape_read_devid));

	byte_offset = offset + sizeof(struct elem_status_page);
	while(byte_offset < (offset + sizeof(struct elem_status_page) +
			drive_info_size)) {
		copy_res_to_devid(elem_info, buffer, byte_offset,
			elem_desc_length);
		if(copy_method == BackToKernel) {
			memcpy(output, elem_info,
				sizeof(struct tape_read_devid));
		} else if(copy_to_user(output, elem_info,
			sizeof(struct tape_read_devid))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* else */

		output += sizeof(struct tape_elem_status);
		byte_offset += elem_desc_length;
		if((*chgp)->chg_id == IBM_3581) byte_offset += 4;
	} /* while */

EXIT_LABEL:

//	*chgp = chg;

	if(elem_info) kfree(elem_info);

	if((*chgp)->buffer) {
		if((*chgp)->buf_len <= KmallocThreshhold) kfree((*chgp)->buffer);
		else free_pages((ulong)((*chgp)->buffer), order);
	} /* if */

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;

	DbgPrint(("lin_tape: changer_read_devid exit: %d\n", rc));
	return rc;
} /* changer_read_devid() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_num_drives                                                *
*                                                                              *
*  PURPOSE: retrieve number of drives in this changer from modesense           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    drive_num: (output) the number of drives                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_get_num_drives(struct changer_descriptor** chgp, uint* drive_num)
{
	struct elem_assignment* elem_assignmnt = NULL;
	int rc = STATUS_SUCCESS;

	elem_assignmnt = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!elem_assignmnt) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(elem_assignmnt, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(chgp, ElementAddressPage,
		(void*)elem_assignmnt, sizeof(struct elem_assignment));

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	*drive_num = (elem_assignmnt->data_transfer_num[0] << 8)
		+ elem_assignmnt->data_transfer_num[1];

	if(!*drive_num) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(elem_assignmnt) kfree(elem_assignmnt);
	return rc;
} /* changer_get_num_drives() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_devid_get_drive_report                                            *
*                                                                              *
*  PURPOSE: issue RES on drive elements for the changer_read_devid IOCTL       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    drive_num: the number of drives                                           *
*    start: starting element address                                           *
*    elem_num: number of elements to retrieve                                  *
*    order: (output) the base-2 logarithm of the number of pages allocated     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chg_devid_get_drive_report(struct changer_descriptor** chgp, uint drive_num,
	ushort start, ushort elem_num, ulong* order)
{
	int rc = STATUS_SUCCESS;
	uint total_alloc = 0;
	ulong buf_size = 0, total_report_size = 0;
//	struct changer_descriptor* chg = *chgp;
	char ttype = lin_tape_trace_type;
	unchar* buffer = NULL;

	total_alloc = drive_num * MinElementDescriptionWithIDSize
		+ sizeof(struct elem_status_data)
		+ sizeof(struct elem_status_page);

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);

	(*chgp)->cdb->cdb_read_elem_status.op_code = READ_ELEMENT_STATUS;
	(*chgp)->cdb->cdb_read_elem_status.elem_type_code = DataTransferReport;
	(*chgp)->cdb->cdb_read_elem_status.vol_tag = False;
	(*chgp)->cdb->cdb_read_elem_status.start_elem_addr[0] = (unchar)(start >> 8);
	(*chgp)->cdb->cdb_read_elem_status.start_elem_addr[1] =
		(unchar)(start & 0xFF);
	(*chgp)->cdb->cdb_read_elem_status.elem_num[0] = (unchar)(elem_num >> 8);
	(*chgp)->cdb->cdb_read_elem_status.elem_num[1] = (unchar)(elem_num & 0xFF);
	(*chgp)->cdb->cdb_read_elem_status.dvcid = True;
	(*chgp)->cdb->cdb_read_elem_status.cur_data = False;
	(*chgp)->cdb->cdb_read_elem_status.alloc_length[0] =
		(unchar)(total_alloc >> 16);
	(*chgp)->cdb->cdb_read_elem_status.alloc_length[1] =
		(unchar)(total_alloc >> 8);
	(*chgp)->cdb->cdb_read_elem_status.alloc_length[2] =
		(unchar)(total_alloc & 0xFF);

	(*chgp)->cdb_len = CDB_GENERIC_12;
	(*chgp)->last_op = READ_ELEMENT_STATUS;

	if(total_alloc > KmallocThreshhold) {
		*order = get_mem_ord(total_alloc, &buf_size);
		(*chgp)->buffer = (unchar*)__get_free_pages(GFP_KERNEL, *order);
		(*chgp)->buf_len = buf_size;
	} else {
		(*chgp)->buffer = kmalloc(total_alloc, GFP_KERNEL);
		(*chgp)->buf_len = total_alloc;
	} /* else */

	if(!(*chgp)->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset((*chgp)->buffer, '\0', (*chgp)->buf_len);

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ReadElementStatusTimeOut), DMA_FROM_DEVICE, False, 0);

	buffer = (*chgp)->buffer;
	if(!buffer) {
		printk("lin_tape: changer_read_devid: no buffer\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	total_report_size = buffer[5] << 16;
	total_report_size += buffer[6] << 8;
	total_report_size += buffer[7];

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data((*chgp), min((int)total_report_size,
			(int)(FullTraceUnitSize - 7)), (*chgp)->buffer);

	if(total_report_size + sizeof(struct elem_status_data) > total_alloc) {
		printk("lin_tape: %s changer_read_devid: underflow\n",
			(*chgp)->devid);
		if(rc >= 0) rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

//	*chgp = chg;
	return rc;
} /* chg_devid_get_drive_report() */

/*******************************************************************************
*                                                                              *
*  NAME: copy_res_to_devid                                                     *
*                                                                              *
*  PURPOSE: Copy read element status data to read_devid struct                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    elem_info: destination buffer                                             *
*    buffer: source buffer                                                     *
*    byte_offset: starting byte in buffer from which to copy                   *
*    elem_desc_length: element descriptor length                               *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void copy_res_to_devid(struct tape_read_devid* elem_info, unchar* buffer,
	uint byte_offset, uint elem_desc_length)
{
	memset(elem_info, '\0', sizeof(struct tape_read_devid));
	elem_info->addr = buffer[byte_offset] << 8;
	elem_info->addr += buffer[byte_offset + 1];
	elem_info->flags1_0 = (buffer[byte_offset+2] & 0x80) >> 7;
	elem_info->flags1_1 = (buffer[byte_offset+2] & 0x40) >> 6;
	elem_info->flags1_2 = (buffer[byte_offset + 2] & 0x20) >> 5;
	elem_info->flags1_3 = (buffer[byte_offset + 2] & 0x10) >> 4;
	elem_info->flags1_4 = (buffer[byte_offset + 2] & 0x08) >> 3;
	elem_info->flags1_5 = (buffer[byte_offset + 2] & 0x04) >> 2;
	elem_info->flags1_6 = (buffer[byte_offset + 2] & 0x02) >> 1;
	elem_info->flags1_7 = (buffer[byte_offset + 2] & 0x01);
	elem_info->reserved1 = buffer[byte_offset + 3];
	elem_info->asc = buffer[byte_offset + 4];
	elem_info->ascq = buffer[byte_offset + 5];

	elem_info->flags2_0 = (buffer[byte_offset+6] & 0x80) >> 7;
	elem_info->flags2_1 = (buffer[byte_offset+6] & 0x40) >> 6;
	elem_info->flags2_2 = (buffer[byte_offset + 6] & 0x20) >> 5;
	elem_info->flags2_3 = (buffer[byte_offset + 6] & 0x10) >> 4;
	elem_info->flags2_4 = (buffer[byte_offset + 6] & 0x08) >> 3;
#if defined(__LITTLE_ENDIAN)
	elem_info->flags2_5 = (buffer[byte_offset+6] & 0x01);
	elem_info->flags2_6 = (buffer[byte_offset+6] & 0x02) >> 1;
	elem_info->flags2_7 = (buffer[byte_offset + 6] & 0x04) >> 2;
#elif defined(__BIG_ENDIAN)
	elem_info->flags2_5 = (buffer[byte_offset+6] & 0x04) >> 2;
	elem_info->flags2_6 = (buffer[byte_offset+6] & 0x02) >> 1;
	elem_info->flags2_7 = (buffer[byte_offset+6] & 0x01);
#else
	error
#endif

	elem_info->scsi = buffer[byte_offset+7];
	elem_info->reserved2 = buffer[byte_offset+8];
	elem_info->flags3_0 = (buffer[byte_offset+9] & 0x80) >> 7;
	elem_info->flags3_1 = (buffer[byte_offset+9] & 0x40) >> 6;
	elem_info->flags3_2 = (buffer[byte_offset+9] & 0x20) >> 5;
	elem_info->flags3_3 = (buffer[byte_offset + 9] & 0x10) >> 4;
	elem_info->flags3_4 = (buffer[byte_offset + 9] & 0x08) >> 3;
	elem_info->flags3_5 = (buffer[byte_offset + 9] & 0x04) >> 2;
	elem_info->flags3_6 = (buffer[byte_offset + 9] & 0x02) >> 1;
	elem_info->flags3_7 = (buffer[byte_offset + 9] & 0x01);
	elem_info->source = (buffer[byte_offset + 10]<<8)
		+ buffer[byte_offset + 11];

	/* Since PVolTag is 0, Volume Tag info is omitted here for LTO device */

	/* flags4_0 - 4 and flags5_0-4 are 4 bit fields
	* -7-6-5-4-3-2-1-0-  => -7-6-5-4-3-2-1-0-
	* |a|b|c|d|e|f|g|h|  => |e|f|g|h|d|c|b|a|
	*/
	elem_info->flags4_0 = (buffer[byte_offset+12] & 0x80) >> 7;
	elem_info->flags4_1 = (buffer[byte_offset+12] & 0x40) >> 6;
	elem_info->flags4_2 = (buffer[byte_offset+12] & 0x20) >> 5;
	elem_info->flags4_3 = (buffer[byte_offset+12] & 0x10) >> 4;
	elem_info->flags4_7 = (buffer[byte_offset+12] & 0x08) >> 3;
	elem_info->flags4_6 = (buffer[byte_offset+12] & 0x04) >> 2;
	elem_info->flags4_5 = (buffer[byte_offset+12] & 0x02) >> 1;
	elem_info->flags4_4 = (buffer[byte_offset+12] & 0x01);
	elem_info->flags5_0 = (buffer[byte_offset+13] & 0x80) >> 7;
	elem_info->flags5_1 = (buffer[byte_offset+13] & 0x40) >> 6;
	elem_info->flags5_2 = (buffer[byte_offset+13] & 0x20) >> 5;
	elem_info->flags5_3 = (buffer[byte_offset+13] & 0x10) >> 4;
	elem_info->flags5_7 = (buffer[byte_offset+13] & 0x08) >> 3;
	elem_info->flags5_6 = (buffer[byte_offset+13] & 0x04) >> 2;
	elem_info->flags5_5 = (buffer[byte_offset+13] & 0x02) >> 1;
	elem_info->flags5_4 = (buffer[byte_offset+13] & 0x01);

	elem_info->reserved3 = buffer[byte_offset+14];
	elem_info->id_length = buffer[byte_offset+15];

	/* If we still have some bytes, this is the identifier. Copy them.
		If we still have some bytes after identifier, this is a vendor-
		specific field for 3583. Ignore them. */
	if(elem_desc_length - MinDeviceID)
		memcpy(elem_info->identifier, &buffer[byte_offset + 16],
			elem_info->id_length);

	return;
} /* copy_res_to_devid() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_request_sense                                                 *
*                                                                              *
*  PURPOSE: retrieve most recent sense data                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    length: (output) number of bytes sent back                                *
*    resid: (output) sense buffer information field                            *
*    sense: (output) sense data itself                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_request_sense(struct changer_descriptor** chgp,
	int* length, int* resid, unchar* sense)
{
//	struct changer_descriptor* chg = *chgp;
	struct tape_sense_data* sense_buf = NULL;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_request_sense entry\n"));

	if(!(*chgp)->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	sense_buf = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);

	if(!sense_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(sense_buf, '\0', sizeof(struct tape_sense_data));

	(*chgp)->buffer = (void*)sense_buf;
	(*chgp)->buf_len = sizeof(struct tape_sense_data);

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	(*chgp)->cdb->cdb_req_sense.op_code = REQUEST_SENSE;
	(*chgp)->cdb->cdb_req_sense.alloc_length = sizeof(struct tape_sense_data);
	(*chgp)->cdb_len = CDB_GENERIC_6;
	(*chgp)->last_op = REQUEST_SENSE;

	rc = changer_send_scsi_cmd(chgp, compute_changer_timeout((*chgp)->chg_id,
		ModeSenseTimeOut), DMA_FROM_DEVICE, False, 0);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	sense_buf = (*chgp)->buffer;
	if(!sense_buf) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	*resid = sense_buf->valid ?
		(uint)((sense_buf->information[0] << 24)
			+ (sense_buf->information[1] << 16)
			+ (sense_buf->information[2] << 8)
			+ sense_buf->information[3])
		: 0;

	*length = (int)sense_buf->additional_sense_len + 8;
	memcpy(sense, sense_buf, sizeof(struct tape_sense_data));

EXIT_LABEL:

//	*chgp = chg;

	(*chgp)->buf_len = 0;
	if(sense_buf) kfree(sense_buf);
	(*chgp)->buffer = NULL;

	DbgPrint(("lin_tape: changer_request_sense exit: %d\n", rc));
	return rc;
} /* changer_request_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_read_elem_status                                              *
*                                                                              *
*  PURPOSE: Issue read element status to libary and return data                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor                                                  *
*    output: buffer to receive RES data                                        *
*    elem_assignment: start addr and num of elms of each type (mode pg 0x1D)   *
*    elem_type: the type of element requested                                  *
*    start_address: the starting address requested                             *
*    num_elems: the number of elements requested                               *
*    vol_tag: a boolean representing whether voltag has been retrieved         *
*    copy_method: to user space or kernel space                                *
*    se_dvcid: for get_cartridge_location (slot elems with dvcid == True)      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_read_elem_status(struct changer_descriptor** chgp,
	unchar *output, struct elem_assignment *elem_assignmnt,
	unchar elem_type, ushort start_address, int num_elems, char *vol_tag,
	unchar copy_method, boolean se_dvc_id)
{
//	struct changer_descriptor* chg = *chgp;
	ulong count;
	ulong report_size, status_page_offset;
	unchar *res_buffer;
	uint desc_report_size = 0;
	ulong order = 0;
	int rc = STATUS_SUCCESS;
	int timeout = 0;

	DbgPrint(("changer_read_elem_status entry\n"));
	if(!output) {
		printk("lin_tape: changer_read_elem_status: output buffer "\
			"invalid.\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if((elem_type != DataTransferReport) &&
		(elem_type != MediumTransportReport) &&
		(elem_type != IEPortReport) &&
		(elem_type != StorageElementReport)) {
		printk("lin_tape: changer_read_elem_status: bad element type "\
			"request. \n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(se_dvc_id && (elem_type != StorageElementReport
		|| !SupportVolTag((*chgp)->chg_id))) {
		printk("lin_tape: changer_read_elem_status: caller requested "\
			"read_cartridge_location but elements not storage\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = alloc_res_out_buf((*chgp), &count, &order, num_elems, se_dvc_id);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	create_res_command((*chgp), elem_type, start_address, num_elems, se_dvc_id,
				count);

	timeout = compute_changer_timeout((*chgp)->chg_id,
		ReadElementStatusTimeOut);

	rc = changer_send_scsi_cmd(chgp, timeout, DMA_FROM_DEVICE, False, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = validate_return_res_data((*chgp), &res_buffer, &report_size,
		&status_page_offset, count, elem_type);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	*vol_tag = res_buffer[status_page_offset + 1] >> 7;

	desc_report_size = (res_buffer[status_page_offset + 5] << 16)
		+ (res_buffer[status_page_offset + 6] << 8)
		+ res_buffer[status_page_offset + 7];

	if(desc_report_size) {
		rc = copy_res_data(output, elem_type, num_elems, res_buffer,
			status_page_offset, desc_report_size,
			vol_tag, se_dvc_id, copy_method);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

//	*chgp = chg;

EXIT_LABEL:

	if(((*chgp)->buf_len <= KmallocThreshhold) && ((*chgp)->buffer))
		kfree((*chgp)->buffer);
	else if((*chgp)->buffer)
		free_pages((ulong) ((*chgp)->buffer), order);

	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;

	DbgPrint(("changer_read_elem_status exit: %d\n", rc));
	return rc;
} /* changer_read_elem_status() */

/*******************************************************************************
*                                                                              *
*  NAME: alloc_res_out_buf                                                     *
*                                                                              *
*  PURPOSE: compute size of RES buffer, allocate and zero                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: changer descriptor                                                   *
*    count: (output) the number of bytes in the return buffer                  *
*    order: (output) if buffer is larger than threshhold, the order of the buf *
*    num_elems: number of elements for which to return data                    *
*    se_dvc_id: request is for get_cart_loc (SE with DvcId = True)             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int alloc_res_out_buf(struct changer_descriptor* chg, ulong* count,
	ulong* order, int num_elems, boolean se_dvc_id)
{
	int rc = STATUS_SUCCESS;
	ulong buf_size = 0;

	DbgPrint(("alloc_res_out_buf entry\n"));
	if(!SupportVolTag(chg->chg_id))
		*count = num_elems * MinElementDescriptionSize
			+ sizeof(struct elem_status_data)
			+ sizeof(struct elem_status_page);

	else if(!se_dvc_id)
		*count = num_elems * MaxElementDescriptionSize
			+ sizeof(struct elem_status_data)
			+ sizeof(struct elem_status_page);

	else /* obtain data for cartridge location reporting */
		*count = num_elems * ElDescriptorWithSEDvcIdSize
			+ sizeof(struct elem_status_data)
			+ sizeof(struct elem_status_page);

	if(*count > KmallocThreshhold) {
		*order = get_mem_ord(*count, &buf_size);
		chg->buffer = (unchar *) __get_free_pages(GFP_KERNEL, *order);
		chg->buf_len = buf_size;
	} else {
		chg->buffer = kmalloc(*count, GFP_KERNEL);
		chg->buf_len = *count;
	} /* else */

	if(!chg->buffer) {
		chg->buf_len = 0;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(chg->buffer, '\0', chg->buf_len);

EXIT_LABEL:

	DbgPrint(("alloc_res_out_buf exit: %d\n", rc));
	return rc;

} /* alloc_res_out_buf() */

/*******************************************************************************
*                                                                              *
*  NAME: create_res_command                                                    *
*                                                                              *
*  PURPOSE: fill out RES CDB and do other work preliminary to sending to HBA   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: changer descriptor                                                   *
*    elem_type: the type of the element for which to return data               *
*    start_address: the starting element address for which to return data      *
*    num_elems: number of elements for which to return data                    *
*    se_dvc_id: request is for get_cart_loc (SE with DvcId = True)             *
*    count: number of bytes in chg buffer                                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void create_res_command(struct changer_descriptor* chg, unchar elem_type,
	ushort start_address, int num_elems, boolean se_dvc_id, ulong count)
{
	DbgPrint(("create_res_command entry\n"));
	chg->cdb->cdb_read_elem_status.op_code = READ_ELEMENT_STATUS;
	chg->cdb->cdb_read_elem_status.elem_type_code = elem_type;

	if(SupportVolTag(chg->chg_id))
		chg->cdb->cdb_read_elem_status.vol_tag = True;
	else
		chg->cdb->cdb_read_elem_status.vol_tag = False;

	chg->cdb->cdb_read_elem_status.start_elem_addr[0]
		= start_address >> 8;
	chg->cdb->cdb_read_elem_status.start_elem_addr[1]
		= start_address & 0xFF;

	chg->cdb->cdb_read_elem_status.elem_num[0] = (unchar)(num_elems >> 8);
	chg->cdb->cdb_read_elem_status.elem_num[1]
		= (unchar)(num_elems & 0xFF);

	chg->cdb->cdb_read_elem_status.dvcid = se_dvc_id;
	chg->cdb->cdb_read_elem_status.cur_data = False;

	chg->cdb->cdb_read_elem_status.alloc_length[0]
		= (unchar)(count >> 16);
	chg->cdb->cdb_read_elem_status.alloc_length[1] = (unchar)(count >> 8);
	chg->cdb->cdb_read_elem_status.alloc_length[2]
		= (unchar)(count & 0xFF);

	chg->cdb_len = CDB_GENERIC_12;
	chg->last_op = READ_ELEMENT_STATUS;

	DbgPrint(("create_res_command exit: void\n"));
	return;

} /* create_res_command() */

/*******************************************************************************
*                                                                              *
*  NAME: validate_return_res_data                                              *
*                                                                              *
*  PURPOSE: make sure data returned from read elem status call is valid        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: changer descriptor                                                   *
*    buffer: (output) points to buffer returned by RES (chg->buffer)           *
*    report_size: (output) length of data returned by RES                      *
*    offset: (output) offset of RES data (past the header)                     *
*    count: the amount of data expected from the call                          *
*    elem_type: the type of data expected from the call                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int validate_return_res_data(struct changer_descriptor* chg, unchar** buffer,
	ulong* report_size, ulong* offset, ulong count, unchar elem_type)
{
	int rc = STATUS_SUCCESS;

	DbgPrint(("validate_return_res_data entry\n"));
	*buffer = chg->buffer;
	if(!*buffer) {
		printk("lin_tape: validate_return_res_data: "\
			"returned bad address for IBMchanger%d.\n",
			chg->dev_index);
		chg->buf_len = 0;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	*report_size = ((*buffer)[5] << 16) + ((*buffer)[6] << 8)
		+ (*buffer)[7];

	if(*report_size > count) {
		printk("lin_tape: validate_return_res_data: "\
			"report_size > count.\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* examine the requested Element Status Page */
	*offset = sizeof(struct elem_status_data);
	if((*buffer)[*offset] != elem_type) {
		printk("lin_tape: validate_res_data: Invalid element "\
			"type returned.\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("validate_return_res_data exit: %d\n", rc));
	return rc;

} /* validate_return_res_data() */

/*******************************************************************************
*                                                                              *
*  NAME: copy_res_data                                                         *
*                                                                              *
*  PURPOSE: copy the data returned from RES into the output buffer             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    output: the buffer to receive data                                        *
*    elem_type: type of changer element for which information is requested     *
*    num_elems: number of elements for which information is requested          *
*    res_buffer: buffer containing the returned RES data                       *
*    status_page_offset: offset into res_buffer where elem status pg begins    *
*    desc_report_size: number of bytes of complete descriptor data             *
*    vol_tag: boolean to indicate voltag was returned, use large enough buffer *
*    se_dvc_id: get_cartridge_location, slot elements with Dvcid = True        *
*    copy_method: to kernel space or to user space                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int copy_res_data(unchar* output, unchar elem_type, int num_elems,
	unchar* res_buffer, ulong status_page_offset, uint desc_report_size,
	char* vol_tag, boolean se_dvc_id, unchar copy_method)
{
	uint desc_length = 0;
	uint output_elem_size = 0;
	int rc = 0;
	struct tape_elem_status* elem_buffer = NULL;
	ulong desc_offset = 0;

	DbgPrint(("copy_res_data entry\n"));

	desc_length = (res_buffer[status_page_offset + 2] << 8)
		+ res_buffer[status_page_offset + 3];

	if((desc_report_size % desc_length) ||
		(desc_length < MinElementDescriptionSize)) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if((desc_report_size / desc_length != num_elems)) {
		printk("lin_tape: copy_res_data: Incorrect "\
			"number of elements returned.\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	output_elem_size = se_dvc_id ? sizeof(struct cartridge_location_data)
		: sizeof(struct tape_elem_status);

	elem_buffer = kmalloc(output_elem_size, GFP_KERNEL);
	if(!elem_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* examine each element descriptor
	* number of descriptors = drive_info_size / desc_length
	*/
	desc_offset = status_page_offset + sizeof(struct elem_status_page);

	DbgPrint(("copy_res_data: copying %d descriptors\n", num_elems));
	while(desc_offset < (status_page_offset
		+ sizeof(struct elem_status_page) + desc_report_size)) {

		rc = copy_descriptor_data(output, elem_buffer,
			output_elem_size, elem_type, res_buffer, desc_offset,
			desc_length, vol_tag, se_dvc_id, copy_method);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		output += output_elem_size;
		desc_offset += desc_length;

	} /* while */

	kfree(elem_buffer);
	elem_buffer = NULL;

EXIT_LABEL:

	DbgPrint(("copy_res_data exit: %d\n", rc));
	return rc;

} /* copy_res_data() */

/*******************************************************************************
*                                                                              *
*  NAME: copy_descriptor_data                                                  *
*                                                                              *
*  PURPOSE: copy a single descriptor returned from RES into the output buffer  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    output: the buffer to receive data                                        *
*    elem_buffer: secondary buffer used to hold descriptor data                *
*    output_elem_size: size of elem_buffer, corresp. to one output descriptor  *
*    elem_type: type of changer element for which information is requested     *
*    res_buffer: buffer containing the returned RES data                       *
*    desc_offset: offset from start of res_buffer to descriptor                *
*    desc_length: length of descriptor in res_buffer                           *
*    vol_tag: boolean to indicate voltag was returned, use large enough buffer *
*    se_dvc_id: get_cartridge_location (slot elements with Dvcid = True)       *
*    copy_method: to kernel space or to user space                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int copy_descriptor_data(unchar* output, struct tape_elem_status* elem_buffer,
	uint output_elem_size, unchar elem_type, unchar* res_buffer,
	ulong desc_offset, uint desc_length, char* vol_tag, boolean se_dvc_id,
	unchar copy_method)
{
	int rc = STATUS_SUCCESS;
	uint extra_length = 0;

	memset(elem_buffer, '\0', output_elem_size);

	if(elem_type == DataTransferReport
		|| elem_type == IEPortReport
		|| elem_type == StorageElementReport
		|| elem_type == MediumTransportReport) {

		elem_buffer->addr = (res_buffer[desc_offset] << 8)
			+ res_buffer[desc_offset + 1];

		elem_buffer->flags1_0 = (res_buffer[desc_offset+2] & 0x80) >> 7;
		elem_buffer->flags1_1 = (res_buffer[desc_offset+2] & 0x40) >> 6;
		elem_buffer->flags1_2 = (res_buffer[desc_offset+2] & 0x20) >> 5;
		elem_buffer->flags1_3 = (res_buffer[desc_offset+2] & 0x10) >> 4;
		elem_buffer->flags1_4 = (res_buffer[desc_offset+2] & 0x08) >> 3;
		elem_buffer->flags1_5 = (res_buffer[desc_offset+2] & 0x04) >> 2;
		elem_buffer->flags1_6 = (res_buffer[desc_offset+2] & 0x02) >> 1;
		elem_buffer->flags1_7 = (res_buffer[desc_offset+2] & 0x01);

		elem_buffer->reserved1 = res_buffer[desc_offset+3];

		elem_buffer->asc = res_buffer[desc_offset+4];
		elem_buffer->ascq = res_buffer[desc_offset+5];

		/* flags2_5 - flags2_7 are bits for LUN */
		elem_buffer->flags2_0 = (res_buffer[desc_offset+6] & 0x80) >> 7;
		elem_buffer->flags2_1 = (res_buffer[desc_offset+6] & 0x40) >> 6;
		elem_buffer->flags2_2 = (res_buffer[desc_offset+6] & 0x20) >> 5;
		elem_buffer->flags2_3 = (res_buffer[desc_offset+6] & 0x10) >> 4;
		elem_buffer->flags2_4 = (res_buffer[desc_offset+6] & 0x08) >> 3;

#if defined(__LITTLE_ENDIAN)

		elem_buffer->flags2_5=(res_buffer[desc_offset+6] & 0x01);
		elem_buffer->flags2_6=(res_buffer[desc_offset+6] & 0x02) >> 1;
		elem_buffer->flags2_7=(res_buffer[desc_offset+6] & 0x04) >> 2;

#elif defined(__BIG_ENDIAN)

		elem_buffer->flags2_5=(res_buffer[desc_offset+6] & 0x04) >> 2;
		elem_buffer->flags2_6=(res_buffer[desc_offset+6] & 0x02) >> 1;
		elem_buffer->flags2_7=(res_buffer[desc_offset+6] & 0x01);

#else

		rc = -EINVAL;
		goto EXIT_LABEL;

#endif

		/* scsi bus address */
		elem_buffer->scsi = res_buffer[desc_offset + 7];

		elem_buffer->reserved2 = res_buffer[desc_offset + 8];

		elem_buffer->flags3_0 = (res_buffer[desc_offset+9] & 0x80) >> 7;
		elem_buffer->flags3_1 = (res_buffer[desc_offset+9] & 0x40) >> 6;
		elem_buffer->flags3_2 = (res_buffer[desc_offset+9] & 0x20) >> 5;
		elem_buffer->flags3_3 = (res_buffer[desc_offset+9] & 0x10) >> 4;
		elem_buffer->flags3_4 = (res_buffer[desc_offset+9] & 0x08) >> 3;
		elem_buffer->flags3_5 = (res_buffer[desc_offset+9] & 0x04) >> 2;
		elem_buffer->flags3_6 = (res_buffer[desc_offset+9] & 0x02) >> 1;
		elem_buffer->flags3_7 = (res_buffer[desc_offset+9] & 0x01);

		/* source storage element address */
		elem_buffer->source = (res_buffer[desc_offset+10] << 8)
			+ res_buffer[desc_offset+11];

		/* extra_length is for voltag and cartridge location */
		extra_length = desc_length - MinElementDescriptionSize;

		if(*vol_tag && extra_length)
			memcpy(elem_buffer->volume,
				&res_buffer[desc_offset + VOL_TAG_OFFSET],
				(extra_length < VOL_TAG_SIZE) ?
				extra_length : VOL_TAG_SIZE);

		/* remaining extra_length is for cartridge location */
		extra_length -= VOL_TAG_SIZE;
		if(((signed)extra_length)
			> sizeof(struct cartridge_location_data)
				- VOL_TAG_OFFSET - VOL_TAG_SIZE) {
			extra_length = sizeof(struct cartridge_location_data)
				- VOL_TAG_OFFSET - VOL_TAG_SIZE;
		} /* if */

		if(se_dvc_id && ((signed)extra_length) > 0) {
			memcpy(&elem_buffer->volume[VOL_TAG_SIZE],
				&res_buffer[desc_offset + VOL_TAG_OFFSET
					+ VOL_TAG_SIZE],
				extra_length);
		} /* if */
	} /* if */

	/* copy current element to output */
	if(copy_method == BackToKernel) {
		memcpy(output, elem_buffer, output_elem_size);
	} else {
		if(copy_to_user(output, elem_buffer, output_elem_size)) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* else */

EXIT_LABEL:

	return rc;

} /* copy_descriptor_data() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_read_buffer                                                   *
*                                                                              *
*  PURPOSE: entry function for changer passthrough                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp  the changer for which the command is requested                      *
*    buf_id: the ID of the buffer to read                                      *
*    buf_len: the length allocated to receive read buffer data                 *
*    buffer: data buffer                                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_read_buffer(struct changer_descriptor** chgp, char buf_id,
	int buf_len, unchar* buffer)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: changer_read_buffer entry\n"));

	if(!(*chgp) || !(*chgp)->cdb || !buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);

	(*chgp)->buffer = kmalloc(buf_len, GFP_KERNEL);
	if(!(*chgp)->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((*chgp)->buffer, '\0', buf_len);
	(*chgp)->buf_len = buf_len;

	(*chgp)->cdb->cdb_read_buffer.op_code = READ_BUFFER;
	(*chgp)->cdb->cdb_read_buffer.mode = 0x01; /* data only */
	(*chgp)->cdb->cdb_read_buffer.buf_id = buf_id;
	(*chgp)->cdb->cdb_read_buffer.alloc_length[0] = (unchar)(buf_len >> 16);
	(*chgp)->cdb->cdb_read_buffer.alloc_length[1] = (unchar)(buf_len >> 8);
	(*chgp)->cdb->cdb_read_buffer.alloc_length[2] = (unchar)(buf_len & 0xFF);
	(*chgp)->cdb_len = CDB_GENERIC_10;
	(*chgp)->last_op = READ_BUFFER;

	rc = changer_send_scsi_cmd(chgp, 30 * HZ, DMA_FROM_DEVICE, False, 0);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: changer_read_buffer unable to read buffer\n");
		goto EXIT_LABEL;
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data((*chgp),
			min((int)((*chgp)->buf_len), (int)(FullTraceUnitSize - 7)),
			(*chgp)->buffer);

	memcpy(buffer, (*chgp)->buffer, buf_len);

EXIT_LABEL:

	if((*chgp)->buffer) kfree((*chgp)->buffer);
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_read_buffer exit: %d\n", rc));
	return rc;
} /* changer_read_buffer() */

/***********************************************************************
 *                                                                     *
 * Medium Changer Pass Through Functions                               *
 *                                                                     *
 ***********************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: changer_pass_thru                                                     *
*                                                                              *
*  PURPOSE: entry function for changer passthrough                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp  the changer for which the command is requested                      *
*    cdb: the command                                                          *
*    cmd_length: number of valid bytes in the CDB                              *
*    buffer: data buffer, if applicable                                        *
*    buffer_length: length of buffer                                           *
*    data_direction: to device or from device                                  *
*    timeout: seconds between command issue and command timeout                *
*    sense: sense data buffer                                                  *
*    sense_length: length of sense data buffer                                 *
*    sense_avail: (output) True if sense is available, False if not available  *
*    resid: the residual count returned in the sense data, if applicable       *
*    result: the result of the call to tape_execute_scsi_command               *
*    attempt_number: how many times this command has been tried so far         *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_pass_thru(struct changer_descriptor** chgp, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint* buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar* sense_length,
	unchar* sense_avail, int* resid, int* result, int attempt_number, boolean fpfo)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: changer_pass_thru entry\n"));

	rc = changer_pass_thru_send_cmd(chgp, cdb, cmd_length, buffer,
		buffer_length, data_direction, timeout, sense, sense_length,
		sense_avail, resid, result, fpfo);

	if(!(*chgp)->do_failover) goto EXIT_LABEL;

	if(attempt_number >= MaxChangerFailoverAttempts) goto EXIT_LABEL;

	attempt_number++;
	rc = changer_perform_pass_thru_failover(chgp, cdb, cmd_length, buffer,
		buffer_length, data_direction, timeout, sense, sense_length,
		sense_avail, resid, result, attempt_number, fpfo);

EXIT_LABEL:

//	*chgp = chg;
	return rc;
} /* changer_pass_thru() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_pass_thru_send_cmd                                            *
*                                                                              *
*  PURPOSE: set up buffers, send pt command, and handle errors                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the pointer to changer for which the command is requested           *
*    cdb: the command                                                          *
*    cmd_length: number of valid bytes in the CDB                              *
*    buffer: data buffer, if applicable                                        *
*    buffer_length: length of buffer                                           *
*    data_direction: to device or from device                                  *
*    timeout: seconds between command issue and command timeout                *
*    sense: sense data buffer                                                  *
*    sense_length: length of sense data buffer                                 *
*    sense_avail: (output) True if sense is available, False if not available  *
*    resid: the residual count returned in the sense data, if applicable       *
*    result: the result of the call to tape_execute_scsi_command               *
*    fpfo: True if request comes from pfo                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_pass_thru_send_cmd(struct changer_descriptor** chgp, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint* buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar* sense_length,
	unchar* sense_avail, int* resid, int* result, boolean fpfo)
{
	ulong order = 0;
	int temp_resid = 0, rc = STATUS_SUCCESS;
	unchar* sense_buffer = NULL;
	uint local_buffer_length = *buffer_length;
	unchar orig_sense_length = *sense_length;

	DbgPrint(("lin_tape: changer_pass_thru_send_cmd entry\n"));

	*result = 0;
	*sense_length = 0;
	*sense_avail = 0;

	rc = changer_pass_thru_check_parms((*chgp), cdb, cmd_length, buffer,
		local_buffer_length, data_direction, timeout);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	memset((*chgp)->cdb, '\0', MAX_CDB_SIZE);
	memcpy((*chgp)->cdb, cdb, cmd_length);

	rc = changer_pass_thru_alloc_bufs((*chgp), buffer, &sense_buffer,
		data_direction, local_buffer_length, &order, fpfo);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	*result = tape_execute_scsi_command((*chgp)->dev_obj, (unchar*)(*chgp)->cdb,
		data_direction, (*chgp)->buffer, local_buffer_length, 0,
		timeout * HZ, sense_buffer, &temp_resid);

	*resid = temp_resid;
	if(temp_resid > 0)
		local_buffer_length = local_buffer_length - temp_resid;
	//*buffer_length = local_buffer_length;

	(*chgp)->sense_valid = False;

	if(*result == STATUS_SUCCESS) {
		rc = STATUS_SUCCESS;
		*sense_length = 0;
		if(data_direction == SCSI_DATA_IN) {
			if(!fpfo && copy_to_user(buffer, (*chgp)->buffer,
				local_buffer_length)) {
				rc = -EFAULT;
			} /* if */
			else if (fpfo) memcpy(buffer, (*chgp)->buffer,
									local_buffer_length);
		} /* if */
		goto EXIT_LABEL;
	} /* if */

	if (*result==-ENOMEM){
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}

	if((driver_byte(*result) & DRIVER_SENSE) && sense_buffer) {
		rc = changer_pass_thru_handle_sense((*chgp), sense, sense_buffer,
			sense_length, orig_sense_length, sense_avail, result, fpfo);
		goto EXIT_LABEL;
	} /* if */

	*sense_length = 0;
	if(host_byte(*result) == DID_ERROR || host_byte(*result) == 0x10)
		printk("lin_tape: driver_byte %02X, host_byte %02X, "\
			"msg_byte %02X, status_byte %02X\n",
			driver_byte(*result), host_byte(*result),
			msg_byte(*result), status_byte(*result));

	if((*chgp)->alt_pathing && (*chgp)->next_path!=(*chgp) &&
		( (host_byte(*result) != DID_OK && host_byte(*result) != DID_TARGET_FAILURE)
		|| status_byte(*result) > 0x01
		|| driver_byte(*result) == DRIVER_TIMEOUT
        || driver_byte(*result) == DRIVER_ERROR)){
			(*chgp)->do_failover = True;
			*sense_length = orig_sense_length;
	}

EXIT_LABEL:

	if((*chgp)->buffer) {
		if((*chgp)->buf_len <= KmallocThreshhold)
			kfree((*chgp)->buffer);
		else
			free_pages((ulong)(*chgp)->buffer, order);
	} /* if */
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;

	if(sense_buffer) kfree(sense_buffer);

	DbgPrint(("lin_tape: changer_pass_thru_send_cmd exit: %d\n", rc));
	return rc;
} /* changer_pass_thru_send_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_pass_thru_check_parms                                         *
*                                                                              *
*  PURPOSE: validate parameters passed in to the pass through function         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which the command is requested                       *
*    cdb: the command                                                          *
*    cmd_length: number of valid bytes in the CDB                              *
*    buffer: data buffer, if applicable                                        *
*    local_buffer_length: length of buffer                                     *
*    data_direction: to device or from device                                  *
*    timeout: seconds between command issue and command timeout                *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_pass_thru_check_parms(struct changer_descriptor* chg, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint local_buffer_length,
	uint data_direction, uint timeout)
{
	int rc = STATUS_SUCCESS;

	if(!cdb || !timeout || !cmd_length || (data_direction != SCSI_DATA_IN
		&& data_direction != SCSI_DATA_OUT
		&& data_direction != SCSI_DATA_NONE)) {
		printk("lin_tape: changer passthru invalid argument\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(data_direction != SCSI_DATA_NONE && (!buffer
		|| !local_buffer_length)) {
		printk("lin_tape: changer_pass_thru_send_cmd no buffer\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(!chg->cdb) {
		printk("lin_tape: changer passthru: no cdb\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(cmd_length > MAX_CDB_SIZE) {
		printk("lin_tape: changer passthru cdb size: %d\n", cmd_length);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;
} /* changer_pass_thru_check_parms() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_pass_thru_alloc_bufs                                          *
*                                                                              *
*  PURPOSE: validate parameters passed in to the pass through function         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which the command is requested                       *
*    buffer: the user buffer                                                   *
*    sense_buffer: (output) receives the newly allocated sense buffer          *
*    data_direction: to device or from device                                  *
*    local_buffer_length: length required for data buffer                      *
*    order: base-2 logarithm of the number of pages required for data buffer   *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_pass_thru_alloc_bufs(struct changer_descriptor* chg,
	unchar* buffer, unchar** sense_buffer, uint data_direction,
	uint local_buffer_length, ulong* order, boolean fpfo)
{
	ulong buf_size = 0;
	int rc = STATUS_SUCCESS;

	if (!*sense_buffer){
		*sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
		if(!*sense_buffer) {
			printk("lin_tape: changer passthru OUT OF MEMORY\n");
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	}
	memset(*sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	if(data_direction != SCSI_DATA_NONE) {
		if(local_buffer_length > KmallocThreshhold) {
			*order = get_mem_ord(local_buffer_length, &buf_size);

			chg->buffer = (unchar*)__get_free_pages(GFP_KERNEL,
				*order);
			chg->buf_len = buf_size;
		} else {
			chg->buffer = kmalloc(local_buffer_length, GFP_KERNEL);
			chg->buf_len = local_buffer_length;
		} /* else */

		if(!chg->buffer) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		if(data_direction == SCSI_DATA_OUT) {
			if(!fpfo && copy_from_user(chg->buffer, buffer,
				local_buffer_length)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			else if (fpfo) memcpy(chg->buffer, buffer,
									local_buffer_length);
		} /* if */
	} /* if */

EXIT_LABEL:

	return rc;
} /* changer_pass_thru_alloc_bufs() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_pass_thru_handle_sense                                        *
*                                                                              *
*  PURPOSE: interpret sense information returned from changer passthru command *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which the command is requested                       *
*    sense: sense buffer passed in by caller of passthru                       *
*    sense_buffer: sense buffer allocated by passthru                          *
*    sense_length: length of sense                                             *
*    orig_sense_length: sense length as input by user before beign modified    *
*    sense_avail: output parm notifying user that sense is valid               *
*    result: result from passthrough command                                   *
*    fpfo: True if request comes fro pfo                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_pass_thru_handle_sense(struct changer_descriptor* chg,
	unchar* sense, unchar* sense_buffer, unchar* sense_length,
	unchar orig_sense_length, unchar* sense_avail, int* result, boolean fpfo)
{
	char sk = 0, asc = 0, ascq = 0;
	int rc = STATUS_SUCCESS;

	if(sense && orig_sense_length) {
		if(orig_sense_length > SCSI_SENSE_BUFFERSIZE)
			orig_sense_length = SCSI_SENSE_BUFFERSIZE;
		if(!fpfo && copy_to_user(sense, sense_buffer, orig_sense_length)) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
		else {
			if (fpfo) memcpy(sense, sense_buffer, orig_sense_length);
			*sense_length = orig_sense_length;
		}
	} /* if */

	*sense_avail = TRUE;
	if(!chg->sense_data) goto EXIT_LABEL;

	memcpy(chg->sense_data, sense_buffer, min(SCSI_SENSE_BUFFERSIZE,
			(int)sizeof(struct tape_sense_data)));
	chg->sense_valid = True;

	sk = chg->sense_data->sense_key;
	asc = chg->sense_data->asc;
	ascq = chg->sense_data->ascq;

	if (sk!=0 || asc!=0 || ascq!=0)
		DbgPrint(("SK 0x%x ASC 0x%x ASCQ 0x%x\n", sk, asc, ascq));
	/* check failover error conditions */
	if(!chg->alt_pathing || chg->next_path==chg) goto EXIT_LABEL;
	if(sk == HARDWARE_ERROR
		||
		sk == ABORTED_COMMAND
		||
		(sk == NOT_READY && asc == 0x04 && ascq == 0x0)
		||
		(host_byte(*result) != DID_OK
		&&
		host_byte(*result) != DID_TARGET_FAILURE)
        ||
        driver_byte(*result) == DRIVER_ERROR )
            chg->do_failover = True;

EXIT_LABEL:

	return rc;
} /* changer_pass_thru_handle_sense() */

/*******************************************************************
 *                                                                 *
 * SCSI Core Functions                                             *
 *                                                                 *
 *******************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: changer_send_scsi_cmd                                                 *
*                                                                              *
*  PURPOSE: send command and handle errors                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which the command is requested                      *
*    timeout: seconds between issue and timeout of the command                 *
*    direction: to device or from device                                       *
*    retries: number of retries requested                                      *
*    attempt_number: number of tries so far                                    *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_send_scsi_cmd(struct changer_descriptor** chgp, int timeout,
	unchar direction, int retries, int attempt_number)
{
//	struct changer_descriptor* chg = *chgp;
	int rc, temp_rc;

	rc = changer_send_scsi_io((*chgp), timeout, direction, retries);

	if(rc == STATUS_SUCCESS) goto EXIT_LABEL;

	if(!(*chgp)->cmd_wo_ua_erp && (*chgp)->sense_valid &&
		(*chgp)->sense_data->sense_key == UNIT_ATTENTION &&
		(*chgp)->sense_data->asc == 0x29 && (*chgp)->sense_data->ascq == 0) {

		if((*chgp)->curr_reserve_type != LinTapeNoReserve) {
			DbgPrint(("lin_tape UNIT ATTN erp: re-reserve\n"));
			temp_rc = changer_send_erp_cmd((*chgp), RESERVE, 0, 0);
			if(temp_rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		DbgPrint(("lin_tape UNIT ATTN erp: reissue inquiry p 0x80.\n"));
		temp_rc = changer_send_erp_cmd((*chgp), INQUIRY, 0, UnitSerialPage);
		goto EXIT_LABEL;
	} /* if */

	if(!(*chgp)->do_failover) goto EXIT_LABEL;
	if(attempt_number >= MaxChangerFailoverAttempts) goto EXIT_LABEL;

	/* Perform failover */
	(*chgp)->result = rc;
	attempt_number++;
	rc = changer_perform_failover(chgp, timeout, direction, retries,
		attempt_number);

EXIT_LABEL:

//	*chgp = chg;
	return rc;
} /* changer_send_scsi_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_send_scsi_io                                                  *
*                                                                              *
*  PURPOSE: make sure it is legal to send request, set up sense buf & send     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which the command is requested                       *
*    timeout: seconds between issue and timeout of the command                 *
*    direction: to device or from device                                       *
*    retries: unused                                                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_send_scsi_io(struct changer_descriptor* chg, int timeout,
	unchar direction, int retries)
{
	int retry_deferred = -1, rc = STATUS_SUCCESS, result = 0, resid = 0,
		trc_len = 0;
	unchar* cdb = (unchar*)chg->cdb, * trc = NULL, * sense_buffer = NULL;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: changer_send_scsi_io entry\n"));

	if(!chg || !chg->dev_obj) {
		if(chg) printk("lin_tape: %s changer_send_scsi_io: no device "\
			"object\n", chg->devid);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(chg->ghost_device) {
		printk("lin_tape: IBMchanger%d rmv pending\n", chg->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = 1 + chg->cdb_len;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(trc) {
			trc[0] = chg->cdb_len;
			memcpy(&trc[1], chg->cdb, chg->cdb_len);
		} /* if */
	} /* if */

	if(!sense_buffer) {
		sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
		if(!sense_buffer) {
			printk("lin_tape: OUT OF MEMORY\n");
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

    do {
		retry_deferred++;
		if(trc && ttype == LIN_TAPE_FULL_TRACE)
			changer_setup_and_log_full(chg, chg->cur_path, TRC_CDB,
				trc, trc_len);

		result = tape_execute_scsi_command(chg->dev_obj, cdb, direction,
			chg->buffer, chg->buf_len, 0, timeout, sense_buffer,
			&resid);

		if (result == -ENOMEM) rc = -ENOMEM;
		else rc = changer_check_result(chg, result, sense_buffer, resid);

	} while(rc == -EAGAIN && retry_deferred < MAX_RETRIES);

EXIT_LABEL:

	if(sense_buffer) kfree(sense_buffer);
	if(trc) kfree(trc);
	DbgPrint(("lin_tape: changer_send_scsi_io exit: %d\n", rc));
	return rc;
} /* changer_send_scsi_io() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_check_result                                                  *
*                                                                              *
*  PURPOSE: check result from changer command and take appropriate action      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which the command was executed                       *
*    result: returned from the command                                         *
*    resid: unused                                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_check_result(struct changer_descriptor* chg, int result,
	unchar* sense_buffer, int resid)
{
	int rc = -EIO, trc_len = 0;
	char ttype = lin_tape_trace_type;
	unchar* trc = NULL;

	if(!chg) {
		printk("lin_tape: %s changer_check_result: device removed\n",
			chg->devid);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(!sense_buffer) {
		printk("lin_tape: %s changer_check_result: sense buffer\n",
			chg->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = 4;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(trc) {
			trc[0] = driver_byte(result);
			trc[1] = host_byte(result);
			trc[2] = msg_byte(result);
			trc[3] = status_byte(result);

			changer_setup_and_log_full(chg, chg->cur_path,
				TRC_STATUS, trc, trc_len);
		} /* if */
	} else if((ttype == LIN_TAPE_NO_TRACE
		|| ttype == LIN_TAPE_REGULAR_TRACE)
		&& result != STATUS_SUCCESS) {
		changer_log_cdb_and_status(chg, result, sense_buffer);
	} /* else if */

	chg->result = result;

	/* if it is a scsi command sent from ERP, ignore the sense data */
	if(chg->in_erp) {
		if(result == STATUS_SUCCESS) rc = STATUS_SUCCESS;
		else rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* if the results were sucessful, then reset sense data */
	if(result == STATUS_SUCCESS) {
		chg->sense_valid = False;
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	if(chg->devid && host_byte(result) == DID_ERROR)
		printk("lin_tape: %s driver_byte %02X, host_byte %02X, "\
			"msg_byte %02X, status_byte %02X\n", chg->devid,
			driver_byte(result), host_byte(result),
			msg_byte(result), status_byte(result));

	/* if we have sense data, save it */
	if((driver_byte(result) & DRIVER_SENSE)
		&& sense_buffer && (host_byte(result) == DID_OK
		|| host_byte(result) == DID_TARGET_FAILURE)) {
		rc = changer_save_sense_data(chg, result, sense_buffer, rc);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} else {
		chg->sense_valid = False;
		if(chg->sense_data)
			memset((void*)chg->sense_data, '\0',
				sizeof(struct tape_sense_data));

		if(status_byte(result) == RESERVATION_CONFLICT ||
			status_byte(result) == SAM_STAT_RESERVATION_CONFLICT)
			rc = -EBUSY;

		chg->do_failover = (chg->alt_pathing && chg->next_path!=chg &&
			((host_byte(result) != DID_OK && host_byte(result) != DID_TARGET_FAILURE)
			|| status_byte(result) != SAM_STAT_GOOD
			|| driver_byte(result) == DRIVER_TIMEOUT
            || driver_byte(result) == DRIVER_ERROR));
	} /* else */

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;
} /* changer_check_result() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_log_cdb_and_status                                            *
*                                                                              *
*  PURPOSE: Post errors to log unless SK 6 or 2 with certain asc/q values      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to log the data                                *
*    result: returned from the command just exectued                           *
*    sense_buffer: if non-NULL, contains sense data                            *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void changer_log_cdb_and_status(struct changer_descriptor* chg, int result,
	unchar* sense_buffer)
{
	char sk = 0, asc = 0, ascq = 0, * trc = NULL;
	int trc_len = 0;

	if((driver_byte(result) & DRIVER_SENSE) && sense_buffer) {
		sk = sense_buffer[2] & 0x0F;
		asc = sense_buffer[12];
		ascq = sense_buffer[13];

		if((sk == 0x02 && asc == 0x04 && ascq == 0x01) ||
			sk == 0x06) goto EXIT_LABEL;
	} /* if */

	trc_len = 1 + chg->cdb_len;
	trc = kmalloc(trc_len, GFP_KERNEL);
	if(!trc) goto EXIT_LABEL;

	trc[0] = chg->cdb_len;

	memcpy(&trc[1], chg->cdb, chg->cdb_len);
	changer_setup_and_log_full(chg, chg->cur_path, TRC_CDB, trc, trc_len);

	memset(trc, '\0', trc_len);

	trc_len = 4;
	trc[0] = driver_byte(result);
	trc[1] = host_byte(result);
	trc[2] = msg_byte(result);
	trc[3] = status_byte(result);

	changer_setup_and_log_full(chg, chg->cur_path, TRC_STATUS, trc,
		trc_len);

EXIT_LABEL:

	if(trc) kfree(trc);
	return;
} /* changer_log_cdb_and_status() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_save_sense_data                                               *
*                                                                              *
*  PURPOSE: copy sense to changer descriptor and log if appropriate            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to save the sense data                         *
*    result: returned from the command just exectued                           *
*    sense_buffer: if non-NULL, contains sense data                            *
*    rc: default return code                                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_save_sense_data(struct changer_descriptor* chg, int result,
	unchar* sense_buffer, int rc)
{
	char ttype = lin_tape_trace_type, sk = 0, asc = 0, ascq = 0,
		* trc = NULL;
	int trc_len = 0;

	if(!chg->sense_data) {
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(chg->chg_log, &chg->chg_log_len,
				TRC_MSG_CHG_NO_SENSE);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	memset((void*)chg->sense_data, '\0', sizeof(struct tape_sense_data));
	memcpy((void*)chg->sense_data, sense_buffer, min(SCSI_SENSE_BUFFERSIZE,
			(int)sizeof(struct tape_sense_data)));

	sk = chg->sense_data->sense_key;
	asc = chg->sense_data->asc;
	ascq = chg->sense_data->ascq;

	printk("lin_tape: %s changer_check_result sensekey: %X "\
		"asc: %X ascq: %X\n", chg->devid, chg->sense_data->sense_key,
		chg->sense_data->asc, chg->sense_data->ascq);

	if(ttype == LIN_TAPE_REGULAR_TRACE) {
		tape_log_msg4(chg->chg_log, &chg->chg_log_len,
			TRC_MSG_CHG_SENSE, chg->sense_data->sense_key,
			chg->sense_data->asc, chg->sense_data->ascq);
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE ||
		((sk != 0x02 || asc != 0x04 || ascq != 0x01)
		&&
		sk != 0x06)) {
		trc_len = min(SCSI_SENSE_BUFFERSIZE,
			(int)sizeof(struct tape_sense_data));

		trc = kmalloc(trc_len, GFP_KERNEL);
		if(trc) {
			memcpy(trc, chg->sense_data, trc_len);
			changer_setup_and_log_full(chg, chg->cur_path,
				TRC_SENSE, trc, trc_len);
		} /* if */
	} /* if */

	chg->sense_valid = True;
	chg->resid_valid = chg->sense_data->valid;

	if(sk == ILLEGAL_REQUEST) rc = -EINVAL;

	chg->do_failover = False;
	if(chg->alt_pathing && chg->next_path!=chg) {
		if(sk == HARDWARE_ERROR || sk == ABORTED_COMMAND
			|| (sk == NOT_READY && asc == 0x04 && ascq == 0x0))
			chg->do_failover = True;

		/* some of HBA drivers set driver_byte 0x08
		even though host_byte is set to 0x01 */
		if(host_byte(result) && host_byte(result) != DID_TARGET_FAILURE)
			chg->do_failover = True;
	} /* if */

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;
} /* changer_save_sense_data() */

/*****************************************************
 *                                                   *
 * Error Recovery Functions                          *
 *                                                   *
 *****************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: changer_send_erp_cmd                                                  *
*                                                                              *
*  PURPOSE: issue error recovery procedure                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to issue the command                           *
*    cmd: defined for INQUIRY, TEST_UNIT_READY, and RESERVE                    *
*    update: if data has changed, update the descriptor                        *
*    count: nquiry page for which to issue the ERP                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_send_erp_cmd(struct changer_descriptor* chg, unchar cmd,
	unchar update, unsigned int count)
{
	int rc = -EIO;

	DbgPrint(("lin_tape: changer_send_erp_cmd: cmd 0x%x\n", cmd));

	switch(cmd) {
	case INQUIRY:
		rc = changer_erp_inquiry(chg, update, count);
		break;

	case TEST_UNIT_READY:
		rc = changer_erp_tur(chg);
		break;

	case RESERVE:
		rc = changer_erp_reserve(chg);
		break;

	default:
		rc = -EIO;
	} /* switch */

	DbgPrint(("lin_tape: changer_send_erp_cmd exit: %d\n", rc));
	return rc;
} /* changer_send_erp_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_erp_inquiry                                                   *
*                                                                              *
*  PURPOSE: issue error recovery procedure for inquiry                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to issue the command                           *
*    update: if data has changed, update the descriptor                        *
*    page: inquiry page for which to issue the ERP                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_erp_inquiry(struct changer_descriptor* chg, unchar update,
	unsigned int page)
{
	unchar* inq_page = NULL;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL, * sense_buffer = NULL;
	char ttype = lin_tape_trace_type;
	int rc = -EIO, page_len = 0, offset = 0, result = 0, resid = 0;

	sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_buffer) {
		printk("lin_tape: cannot allocate memory for erp command\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	if(page != UnitSerialPage) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	inq_page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq_page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(inq_page, '\0', MAX_INQ_LEN);

	tmp_cdb = (union cdb*)kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset((void*)tmp_cdb, '\0', MAX_CDB_SIZE);

	tmp_cdb->cdb_inquiry.op_code = INQUIRY;
	tmp_cdb->cdb_inquiry.evpd = True;
	tmp_cdb->cdb_inquiry.page_code = UnitSerialPage;
	tmp_cdb->cdb_inquiry.alloc_length = MAX_INQ_LEN;

	tcdb = (unchar*)tmp_cdb;

	rc = changer_exe_scsi_cmd(chg, tcdb, 6, (void*)inq_page, MAX_INQ_LEN,
		compute_changer_timeout(chg->chg_id, InquiryTimeOut),
		DMA_FROM_DEVICE, &result, sense_buffer, &resid);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: inquiry page 0x80 on IBMchanger%d failed, "\
			"rc %d\n", chg->dev_index, rc);
		if(update) {
			memset(chg->devid, '\0', MAX_VOLUME_LABEL_SIZE);
			sprintf(&chg->devid[0], "IBMChgr%d------",
				chg->dev_index);
			memcpy(&chg->devid[13], "NoSer", strlen("NoSer"));
		} /* if */
		goto EXIT_LABEL;
	} /* if */

	page_len = inq_page[3];
	if(update) {
		chg->serial_num_len = min(page_len, MaxSerialNumberSize);
		memcpy(chg->serial_num, &inq_page[4], chg->serial_num_len);

		offset = page_len + 3;
		while(inq_page[offset] == ' ' && offset >= 9) offset--;

		memset(chg->devid, '\0', MAX_VOLUME_LABEL_SIZE + 1);
		sprintf(&chg->devid[0], "IBMChgr%d------", chg->dev_index);
		if(offset > 9)
			memcpy(&chg->devid[13], &inq_page[offset - 4], 5);
		else
			memcpy(&chg->devid[13], "NoSer", strlen("NoSer"));
	} else {
		rc = memcmp(chg->serial_num, &inq_page[4], chg->serial_num_len);
		if(rc) {
			printk("lin_tape: serial number mismatch\n");
			rc = -EINVAL;
		} /* if */
	} /* else */

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_buffer_data(chg,
			min((int)((inq_page[3] & 0xFF) + 4),
			(FullTraceUnitSize - 7)), inq_page);
EXIT_LABEL:

	if(sense_buffer) kfree(sense_buffer);
	if(inq_page) kfree(inq_page);
	if(tmp_cdb) kfree(tmp_cdb);

	return rc;
} /* changer_erp_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_erp_tur                                                       *
*                                                                              *
*  PURPOSE: issue error recovery procedure for test unit ready                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to issue the command                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_erp_tur(struct changer_descriptor* chg)
{
	union cdb *tmp_cdb = NULL;
	unchar *tcdb = NULL;
	unchar* sense_buffer = NULL;
	int rc = -EIO, result = 0, resid = 0;

	sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_buffer) {
		printk("lin_tape: cannot allocate memory for erp command\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	tmp_cdb = (union cdb*)kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset((void*)tmp_cdb, '\0', MAX_CDB_SIZE);

	tmp_cdb->cdb_tur.op_code = TEST_UNIT_READY;

	tcdb = (unchar*)tmp_cdb;

	rc = changer_exe_scsi_cmd(chg, tcdb, 6, NULL, 0,
		compute_changer_timeout(chg->chg_id,
		TestUnitReadyTimeOut), DMA_NONE, &result, sense_buffer,
		&resid);
EXIT_LABEL:

	if(sense_buffer) kfree(sense_buffer);
	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* changer_erp_tur() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_erp_reserve                                                   *
*                                                                              *
*  PURPOSE: issue error recovery procedure for reserve                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to issue the command                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_erp_reserve(struct changer_descriptor* chg)
{
	unchar* sense_buffer = NULL;
	unchar *tcdb = NULL;
	union cdb *tmp_cdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_buffer) {
		printk("lin_tape: cannot allocate memory for erp command\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	tmp_cdb = (union cdb*)kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset((void*)tmp_cdb, '\0', MAX_CDB_SIZE);

	tmp_cdb->cdb_reserve.op_code = RESERVE;

	tcdb = (unchar*)tmp_cdb;

	rc = changer_exe_scsi_cmd(chg, tcdb, 6, NULL, 0,
		compute_changer_timeout(chg->chg_id, ReserveTimeOut),
		DMA_NONE, &result, sense_buffer, &resid);

EXIT_LABEL:

	if(sense_buffer) kfree(sense_buffer);
	if(tmp_cdb) kfree(tmp_cdb);

	return rc;
} /* changer_erp_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_exe_scsi_cmd                                                  *
*                                                                              *
*  PURPOSE: send the given SCSI command to the SCSI layer for an ERP           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to issue the command                           *
*    cdb: command descriptor block                                             *
*    cdb_len: length of command descriptor block in bytes                      *
*    buffer: data buffer                                                       *
*    buf_len: the length of the data buffer                                    *
*    timeout: seconds between command issue and timeout                        *
*    direction: to device or from device                                       *
*    result: result from passthrough command                                   *
*    sense_buffer: buffer for sense data                                       *
*    resid: amount of residual data                                            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_exe_scsi_cmd(struct changer_descriptor* chg, unchar* cdb,
	unchar cdb_len, void* buffer, int buf_len, int timeout,
	unchar direction, int* result, unchar* sense_buffer, int* resid)
{
	int rc = STATUS_SUCCESS, trc_len = 0;
	char ttype = lin_tape_trace_type;
	unchar* trc = NULL;

	if(!sense_buffer || !cdb || (buf_len > 0 && !buffer) ||
		(buffer && !buf_len)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(!chg || !chg->dev_obj) {
		if(chg) printk("lin_tape: %s changer_send_scsi_cmd: no device "\
			"object\n", chg->devid);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(chg->ghost_device) {
		printk("lin_tape: IBMchanger%d rmv pending\n", chg->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	/* make sure the HBA is not in error recovery */
	if (!scsi_block_when_processing_errors(chg->dev_obj)){
		printk("lin_tape: IBMchanger%d scsi errors offline or deleted\n", chg->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	if(chg->trace && ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = 1 + cdb_len;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(trc) {
			trc[0] = cdb_len;
			memcpy(&trc[1], cdb, cdb_len);
			changer_setup_and_log_full(chg, chg->cur_path, TRC_CDB,
				trc, trc_len);
			kfree(trc);
			trc = NULL;
		} /* if */

		if(buf_len && (cdb[0] == MODE_SELECT ||
			cdb[0] == DISPLAY_MESSAGE || cdb[0] == SEND_DIAGNOSTIC))
			changer_log_buffer_data(chg, min(buf_len,
				FullTraceUnitSize - 7), buffer);
	} /* if */

	*result = tape_execute_scsi_command(chg->dev_obj, cdb, direction,
		buffer, buf_len, 0, timeout, sense_buffer, resid);
	rc = *result;
	if (rc==-ENOMEM) goto EXIT_LABEL;

	if(ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = 4;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(!trc) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		trc[0] = driver_byte(rc);
		trc[1] = host_byte(rc);
		trc[2] = msg_byte(rc);
		trc[3] = status_byte(rc);

		changer_setup_and_log_full(chg, chg->cur_path, TRC_STATUS, trc,
			trc_len);

		kfree(trc);
		trc = NULL;

		if(rc != STATUS_SUCCESS && (driver_byte(rc) &
			DRIVER_SENSE) && sense_buffer) {
			trc_len = min(SCSI_SENSE_BUFFERSIZE,
				(int)sizeof(struct tape_sense_data));

			trc = kmalloc(trc_len, GFP_KERNEL);
			if(!trc) {
				rc = -ENOMEM;
				goto EXIT_LABEL;
			} /* if */

			memcpy(trc, (unchar*)sense_buffer, trc_len);
			changer_setup_and_log_full(chg, chg->cur_path,
				TRC_SENSE, trc, trc_len);
			kfree(trc);
			trc = NULL;
		} /* if */
	} /* if */

	if(sense_buffer && (sense_buffer[2] & 0x0F) == UNIT_ATTENTION)
		rc = UNIT_ATTENTION;
	else if(host_byte(rc) != DID_OK &&
		host_byte(rc) != DID_TARGET_FAILURE &&
		status_byte(rc) != SAM_STAT_RESERVATION_CONFLICT &&
		status_byte(rc) != RESERVATION_CONFLICT)
		rc = -EAGAIN;
	else if(status_byte(rc) == SAM_STAT_RESERVATION_CONFLICT ||
		status_byte(rc) == RESERVATION_CONFLICT)
		rc = -EBUSY;
	else if(rc != STATUS_SUCCESS)
		rc = -EIO;

EXIT_LABEL:

	return rc;
} /* changer_exe_scsi_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_perform_failover                                              *
*                                                                              *
*  PURPOSE: handle failover                                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer on which to perform failover                            *
*    timeout: seconds between command issue and timeout                        *
*    direction: to device or from device                                       *
*    retries: how many times to reissue command on failure                     *
*    attempt_number: how many retries have happened so far                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_perform_failover(struct changer_descriptor** chgp, int timeout,
	unchar direction, int retries, int attempt_number)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = -EIO;
	unchar opcode = (*chgp)->cdb->cdb_6.op_code;

	printk("lin_tape: changer_perform_failover entry.\n");

	rc = changer_failover_ops(chgp, 10 * timeout, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	if(!(*chgp) || !(*chgp)->cdb || (*chgp)->cdb->cdb_6.op_code != opcode) {
		DbgPrint(("lin_tape chg %p: corrupt CDB.  Possible chg swap.",
			(*chgp)));
		if((*chgp)) {
			DbgPrint(("  cdb: %p", (*chgp)->cdb));
			if((*chgp)->cdb)
				DbgPrint(("  old op 0x%x new op 0x%x", opcode,
					(*chgp)->cdb->cdb_6.op_code));
		} /* if */
		DbgPrint(("\n"));
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	rc = changer_send_scsi_cmd(chgp, timeout, direction, retries,
		attempt_number);

EXIT_LABEL:

//	*chgp = chg;
	return rc;
} /* changer_perform_failover() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_perform_pass_thru_failover                                    *
*                                                                              *
*  PURPOSE: handle failover on passthrough commands                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer on which to perform failover                            *
*    cdb: the command being attempted                                          *
*    cmd_length: the number of valid bytes in the CDB                          *
*    buffer: data buffer                                                       *
*    buffer_length: size of buffer in bytes                                    *
*    data_direction: to device or from device                                  *
*    timeout: seconds between command issue and timeout                        *
*    sense: buffer for sense data                                              *
*    sense_length: size of sense in bytes                                      *
*    sense_avail: True if sense is valid                                       *
*    resid: amount of residual data                                            *
*    result: result from passthrough command                                   *
*    attempt_number: how many retries have happened so far                     *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_perform_pass_thru_failover(struct changer_descriptor** chgp,
	unchar* cdb, unchar cmd_length, unchar* buffer, uint* buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar* sense_length,
	unchar* sense_avail, int* resid, int* result, int attempt_number, boolean fpfo)
{
//	struct changer_descriptor* chg = *chgp;
	int rc = -EIO;

	printk("lin_tape: changer_perform_pass_thru_failover entry\n");

	rc = changer_failover_ops(chgp, 4 * timeout, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	if(!(*chgp)) {
		DbgPrint(("lin_tape chg %p: corrupt possible chg swap\n",
			(*chgp)));
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	rc = changer_pass_thru(chgp, cdb, cmd_length, buffer, buffer_length,
		data_direction, timeout, sense, sense_length, sense_avail,
		resid, result, attempt_number, fpfo);
EXIT_LABEL:

//	*chgp = chg;
	return rc;
} /* changer_perform_pass_thru_failover() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_failover_ops                                                  *
*                                                                              *
*  PURPOSE: handle common tasks for passthru and non-passthru failover         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_failover_ops(struct changer_descriptor** chgp, int timeout, int in_open)
{
	int rc = STATUS_SUCCESS, i = 0;
	char ttype = lin_tape_trace_type, * trc = NULL;

	(*chgp)->do_failover = False;

	if(ttype == LIN_TAPE_INVALID_TRACE) {
		printk("lin_tape: lin_taped daemon is not running, "\
			"IBMchanger%d cannot failover.\n", (*chgp)->dev_index);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	rc = changer_setup_and_log_full((*chgp), (*chgp)->cur_path,
		TRC_CG_ATMPT_FAILOVER, NULL, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	DbgPrint(("lin_tape: IBMchanger%d is to wait.\n", (*chgp)->dev_index));
	rc = wait_event_interruptible_timeout((*chgp)->wait_queue,
		!chgp || !(*chgp) || (*chgp)->ready_wakeup != False, timeout);
	if(!(*chgp) ) {
		printk("changer removed while waiting\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if(rc < 0) {
		printk("lin_tape: IBMchanger%d failover failed due to "\
			"unexpected interrupt.\n", (*chgp)->dev_index);
		rc = (*chgp)->result;
		goto EXIT_LABEL;
	} else if(!rc) {
		printk("lin_tape: IBMchanger%d failover failed due to "\
			"timeout.\n", (*chgp)->dev_index);
		rc = (*chgp)->result;
		goto EXIT_LABEL;
	} else {
		DbgPrint(("lin_tape: IBMchanger%d is woken up on a new path "\
			"IBMchanger%d\n", (*chgp)->dev_index, (*chgp)->cur_path));
		WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		(*chgp)->ready_wakeup = False;
        WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		rc = STATUS_SUCCESS;
	} /* else */

	if((*chgp)->wakeup_result != STATUS_SUCCESS) {
		rc = (*chgp)->result;
		if (!rc) rc = -EIO;
		printk("lin_tape: IBMchanger%d failover failed %x\n",
			(*chgp)->dev_index,rc);
		goto EXIT_LABEL;
	} /* if */

	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg5((*chgp)->chg_log, &(*chgp)->chg_log_len,
			TRC_CG_CPF_SUCCESS,
			(*chgp)->dev_index >> 8, (*chgp)->dev_index & 0xFF,
			(*chgp)->cur_path  >> 8, (*chgp)->cur_path  & 0xFF);
	else if(ttype == LIN_TAPE_FULL_TRACE) {
		trc = kmalloc(4, GFP_KERNEL);
		if(trc) {
			trc[0] = (*chgp)->dev_index >> 8;
			trc[1] = (*chgp)->dev_index & 0xFF;
			trc[2] = (*chgp)->cur_path >> 8;
			trc[3] = (*chgp)->cur_path & 0xFF;
			changer_log_full_trace((*chgp), TRC_CG_CPF_SUCCESS, 4, trc);
			kfree(trc);
		}
	} /* else if */

	printk("lin_tape: set current path to IBMchanger%d\n", (*chgp)->cur_path);

	changer_update_to_new_current_path(chgp);

	for(i = 0; i < 5; i++)
		/* necessary on RS-422 cable pull failback */
		if(changer_send_erp_cmd((*chgp), TEST_UNIT_READY, 0, 0)
			!= UNIT_ATTENTION) break;

EXIT_LABEL:

	return rc;
} /* changer_failover_ops() */

/*******************************************************************************
*                                                                              *
*  NAME: get_mem_ord                                                           *
*                                                                              *
*  PURPOSE: compute order of 2 based on requested size                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    req_size: the number of bytes from which to take the base-2 logarithm     *
*    malloc_size: (output) the total amount of memory to allocate              *
*                                                                              *
*  RETURNS: base-2 logarithm of the number of pages to allocate                *
*                                                                              *
*******************************************************************************/

unsigned long get_mem_ord(ulong req_size, ulong* malloc_size)
{
	uint whole_pgs = 0, partial = 0;
	ulong ord = 0;

	if(req_size <= PAGE_SIZE) {
		*malloc_size = PAGE_SIZE;
		ord = 0;
		goto EXIT_LABEL;
	} /* if */

	partial = (req_size & (PPGMASK)) > 0;
	whole_pgs = (req_size >> PAGE_SHIFT);

	for(ord = 0; whole_pgs > 1;
		partial |= whole_pgs & 0x01, whole_pgs >>= 1, ord++) continue;

	ord += partial;
	*malloc_size = (ulong)(1 << (ord + PAGE_SHIFT));

EXIT_LABEL:

	return ord;
} /* get_mem_ord() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_get_best_path                                                 *
*                                                                              *
*  PURPOSE: consult daemon for best path and update descriptor accordingly     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer we are attempting to open                               *
*    buf: a caller-allocated scsi_path structure                               *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

void changer_get_best_path(struct changer_descriptor** chgp)
{
	int rc = -EIO;
//	struct changer_descriptor* chg = *chgp;
	char ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: checking for the best path on IBMchanger%d\n",
		(*chgp)->dev_index));

	if(ttype == LIN_TAPE_INVALID_TRACE) {
		DbgPrint(("lin_tape: lin_taped daemon is not running, "\
			"IBMchanger%d will use the current path.\n",
			(*chgp)->dev_index));
		goto EXIT_LABEL;
	} /* if */

	rc = changer_setup_and_log_full((*chgp), (*chgp)->dev_index, TRC_CG_PATH_CHECK,
		NULL, 0);
	if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape: IBMchanger%d path check failure %d, "\
			"using IBMchanger%d.\n", (*chgp)->dev_index, rc,
			(*chgp)->cur_path));
		goto EXIT_LABEL;
	} /* if */

	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	(*chgp)->ready_wakeup = False;
	WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	DbgPrint(("lin_tape: IBMchanger%d is to wait.\n", (*chgp)->dev_index));
	rc = wait_event_interruptible_timeout((*chgp)->wait_queue,
		!chgp || !(*chgp) || (*chgp)->ready_wakeup != False, 300 * HZ);
	if(!(*chgp)) {
		printk("IBMhanger removed while waiting\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if(rc < 0) {
		printk("lin_tape: IBMchanger%d path checking failed due to "\
			"unexpected interrupt. rc = %d\n", (*chgp)->dev_index, rc);
		goto EXIT_LABEL;
	} else if(!rc) {
		if((*chgp)->wakeup_result)
			printk("lin_tape: IBMchanger%d path checking failed, "\
			"wakeup result: %d\n", (*chgp)->dev_index,
			(*chgp)->wakeup_result);
		else printk("lin_tape: IBMchanger%d path checking failed due "\
			"to timeout.\n", (*chgp)->dev_index);
		goto EXIT_LABEL;
	} else {
		DbgPrint(("lin_tape: IBMchanger%d will use the IBMchanger%d "\
			"path\n", (*chgp)->dev_index, (*chgp)->cur_path));

		WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		(*chgp)->ready_wakeup = False;
		WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	} /* else */

	DbgPrint(("lin_tape: set current path to: IBMchanger%d\n",
		(*chgp)->cur_path));

	changer_update_to_new_current_path(chgp);

EXIT_LABEL:

//	*chgp = chg;
	return;
} /* changer_get_best_path() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_update_to_new_current_path                                    *
*                                                                              *
*  PURPOSE: for change in path, update new path's variables                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer we are attempting to open                               *
*    buf: a caller-allocated scsi_path structure                               *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

void changer_update_to_new_current_path(struct changer_descriptor** chgp)
{
//	struct changer_descriptor* chg = *chgp, * trav = *chgp;
	struct changer_descriptor* trav = (*chgp)->next_path;

	WRITE_LOCK(&changer_des_lock);
	do {
		trav->cur_path = (*chgp)->cur_path;
		trav = trav->next_path;
	} while(trav != (*chgp));
	WRITE_UNLOCK(&changer_des_lock);

	WRITE_LOCK(&changer_devs[(*chgp)->cur_path]->entry_lock);
	changer_devs[(*chgp)->cur_path]->chg->busy = (*chgp)->busy & 0xEC;
	changer_devs[(*chgp)->cur_path]->chg->chg_open = (*chgp)->chg_open;
	changer_devs[(*chgp)->cur_path]->chg->open_count = (*chgp)->open_count;
	changer_devs[(*chgp)->cur_path]->chg->cur_path = (*chgp)->cur_path;
	changer_devs[(*chgp)->cur_path]->chg->cdb = (*chgp)->cdb;
	changer_devs[(*chgp)->cur_path]->chg->cdb_len = (*chgp)->cdb_len;
	changer_devs[(*chgp)->cur_path]->chg->buffer = (*chgp)->buffer;
	changer_devs[(*chgp)->cur_path]->chg->buf_len = (*chgp)->buf_len;
	changer_devs[(*chgp)->cur_path]->chg->last_op = (*chgp)->last_op;

	changer_devs[(*chgp)->cur_path]->chg->result = (*chgp)->result;
	changer_devs[(*chgp)->cur_path]->chg->sense_data = (*chgp)->sense_data;
	changer_devs[(*chgp)->cur_path]->chg->sense_valid = (*chgp)->sense_valid;
	changer_devs[(*chgp)->cur_path]->chg->in_erp = (*chgp)->in_erp;

	changer_devs[(*chgp)->cur_path]->chg->comp_code = (*chgp)->comp_code;
	changer_devs[(*chgp)->cur_path]->chg->comp_code_valid = (*chgp)->comp_code_valid;
	changer_devs[(*chgp)->cur_path]->chg->last_errno = (*chgp)->last_errno;
	changer_devs[(*chgp)->cur_path]->chg->barcode = (*chgp)->barcode;

	changer_devs[(*chgp)->cur_path]->chg->curr_reserve_type = (*chgp)->curr_reserve_type;
	changer_devs[(*chgp)->cur_path]->chg->explicit_reserved = (*chgp)->explicit_reserved;

	changer_devs[(*chgp)->cur_path]->chg->trace = (*chgp)->trace;
	changer_devs[(*chgp)->cur_path]->chg->chg_log = (*chgp)->chg_log;
	changer_devs[(*chgp)->cur_path]->chg->chg_log_len = (*chgp)->chg_log_len;
	changer_devs[(*chgp)->cur_path]->chg->full_trc_head = (*chgp)->full_trc_head;
	changer_devs[(*chgp)->cur_path]->chg->full_trc_tail = (*chgp)->full_trc_tail;

	WRITE_UNLOCK(&changer_devs[(*chgp)->cur_path]->entry_lock);

	/* only clear these out if the changer path changed */
	if((*chgp)->cur_path != (*chgp)->dev_index)
	{
        WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		(*chgp)->cdb = NULL;
		(*chgp)->cdb_len = 0;
		(*chgp)->buffer = NULL;
		(*chgp)->buf_len = 0;
		(*chgp)->result = 0;
		(*chgp)->sense_data = NULL;
		(*chgp)->sense_valid = False;
		(*chgp)->in_erp = False;
		(*chgp)->comp_code = 0;
		(*chgp)->comp_code_valid = False;
		(*chgp)->last_errno = 0;
		(*chgp)->barcode = 0;
		(*chgp)->curr_reserve_type = LinTapeNoReserve;
		(*chgp)->explicit_reserved = False;
		(*chgp)->trace = False;
		(*chgp)->chg_log = NULL;
		(*chgp)->chg_log_len = 0;
		(*chgp)->full_trc_head = NULL;
		(*chgp)->full_trc_tail = NULL;
		WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	}
// this is where the path changes at the chg level in array
	(*chgp) = changer_devs[(*chgp)->cur_path]->chg;

//	*chgp = chg;
	return;
} /* changer_update_to_new_current_path() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_query_path                                                    *
*                                                                              *
*  PURPOSE: display primary and first alternate path for this changer          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer we are attempting to open                               *
*    buf: a caller-allocated scsi_path structure                               *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_query_path(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	int name_length = 0, i = 0, count = 0, rc = STATUS_SUCCESS;
//	struct changer_descriptor* chg = *chgp;
	struct scsi_path* buffer = (struct scsi_path*)buf;
	boolean des_lockd = False;

	DbgPrint(("lin_tape: changer_query_path entry\n"));

	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buffer, '\0', sizeof(struct scsi_path));

	name_length = min(30, MaxLogicName);

	READ_LOCK(&changer_des_lock);
	des_lockd = True;
	if(!(*chgp)->alt_pathing || !(*chgp)->serial_num_len) {
		memcpy(buffer->primary_name, (*chgp)->lname,
			name_length);
		memcpy(buffer->primary_parent, (*chgp)->pname,
			name_length);
		buffer->primary_host = (*chgp)->host;
		buffer->primary_id = (*chgp)->target;
		buffer->primary_lun = (*chgp)->lun;
		buffer->primary_bus = (*chgp)->channel;
		buffer->primary_enabled = (*chgp)->enabled;
		buffer->primary_id_valid = 1;
		buffer->primary_fenced = (*chgp)->fenced;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg ||
			changer_devs[i]->chg->serial_num_len
				!= (*chgp)->serial_num_len ||
			memcmp(changer_devs[i]->chg->serial_num,
				(*chgp)->serial_num, (*chgp)->serial_num_len)) continue;
		if(!count) {
			memcpy(buffer->primary_name, changer_devs[i]->chg->lname,
				name_length);
			memcpy(buffer->primary_parent, changer_devs[i]->chg->pname,
				name_length);
			buffer->primary_host = changer_devs[i]->chg->host;
			buffer->primary_id = changer_devs[i]->chg->target;
			buffer->primary_lun = changer_devs[i]->chg->lun;
			buffer->primary_bus = changer_devs[i]->chg->channel;
			buffer->primary_enabled = changer_devs[i]->chg->enabled;
			buffer->primary_id_valid = 1;
			buffer->primary_fenced = changer_devs[i]->chg->fenced;
		} else {
			buffer->alternate_configured = True;

			memcpy(buffer->alternate_name,
				changer_devs[i]->chg->lname, name_length);
			memcpy(buffer->alternate_parent,
				changer_devs[i]->chg->pname, name_length);
			buffer->alternate_host = changer_devs[i]->chg->host;
			buffer->alternate_id = changer_devs[i]->chg->target;
			buffer->alternate_lun = changer_devs[i]->chg->lun;
			buffer->alternate_bus = changer_devs[i]->chg->channel;
			buffer->alternate_enabled = changer_devs[i]->chg->enabled;
			buffer->alternate_id_valid = 1;
			buffer->alternate_fenced = changer_devs[i]->chg->fenced;
		} /* else */

		count++;
		if(count >= 2) break;
	} /* for */

EXIT_LABEL:

	if(des_lockd) READ_UNLOCK(&changer_des_lock);

//	*chgp = chg;
	DbgPrint(("lin_tape: changer_query_path exit: %d\n", rc));
	return rc;
} /* changer_query_path() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_show_paths                                                    *
*                                                                              *
*  PURPOSE: display primary and alternate paths for this changer               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer we are attempting to open                               *
*    buf: a caller-allocated device_paths structure                            *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t changer_show_paths(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	int i = 0, name_length = 0, count = 0, rc = STATUS_SUCCESS;
	struct device_paths* buffer = (struct device_paths*)buf;
//	struct changer_descriptor* chg = *chgp;
	boolean des_lockd = False;

	DbgPrint(("changer_show_paths entry\n"));

	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buffer, '\0', sizeof(struct device_paths));

	name_length = min(30, MaxLogicName);

	READ_LOCK(&changer_des_lock);
	des_lockd = True;

	if(!(*chgp)->alt_pathing || !(*chgp)->serial_num_len) {
		buffer->number_paths = 1;
		memcpy(buffer->path[0].name, (*chgp)->lname, name_length);
		memcpy(buffer->path[0].parent, (*chgp)->pname, name_length);
		buffer->path[0].host = (*chgp)->host;
		buffer->path[0].id = (*chgp)->target;
		buffer->path[0].lun = (*chgp)->lun;
		buffer->path[0].bus = (*chgp)->channel;
		buffer->path[0].enabled = (*chgp)->enabled;
		buffer->path[0].fenced = (*chgp)->fenced;
		buffer->path[0].id_valid = 1;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg ||
			changer_devs[i]->chg->serial_num_len
				!= (*chgp)->serial_num_len ||
			memcmp(changer_devs[i]->chg->serial_num, (*chgp)->serial_num,
				(*chgp)->serial_num_len)) continue;

		memcpy(buffer->path[count].name, changer_devs[i]->chg->lname,
			name_length);
		memcpy(buffer->path[count].parent, changer_devs[i]->chg->pname,
			name_length);
		buffer->path[count].host = changer_devs[i]->chg->host;
		buffer->path[count].id = changer_devs[i]->chg->target;
		buffer->path[count].lun = changer_devs[i]->chg->lun;
		buffer->path[count].bus = changer_devs[i]->chg->channel;
		buffer->path[count].enabled = changer_devs[i]->chg->enabled;
		buffer->path[count].fenced = changer_devs[i]->chg->fenced;
		buffer->path[count].id_valid = 1;

		count++;
		if(count >= MAX_SCSI_PATH) break;
	} /* for */

	buffer->number_paths = count;

EXIT_LABEL:

	if(des_lockd) READ_UNLOCK(&changer_des_lock);
//	*chgp = chg;
	DbgPrint(("changer_show_paths exit: %d\n", rc));
	return rc;
} /* changer_show_paths() */
