/*******************************************************************************
 *   lin_tape_scsi_ioctl.h
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#ifndef _LIN_TAPE_SCSI_IOCTL_H_
#define _LIN_TAPE_SCSI_IOCTL_H_

#include <linux/kernel.h>
#include <linux/blkdev.h>       /* RQ_SCSI_BUSY                    */
#include <linux/dma-mapping.h>  /* DMA_FROM_DEVICE, etc            */
#include <linux/completion.h>   /* struct completion               */
#include <linux/random.h>       /* get_random_bytes()              */
#include <linux/ctype.h>        /* isxdigit(), tolower()           */
#include <linux/version.h>      /* kernel version number           */

#include <asm/page.h>           /* PAGE_SIZE, PAGE_SHIFT, etc.     */
#include <asm/uaccess.h>        /* copy user space                 */
#include <asm/dma.h>

#include <scsi/scsi_device.h>   /* struct scsi_device;             */
                                /* to_scsi_device                  */
#include <scsi/scsi_host.h>     /* struct scsi_host_template;      */
                                /* struct Scsi_Host                */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && (!defined(CONFIG_SLE_SP))
#include <scsi/scsi_request.h>  /* struct scsi_request, etc.       */
#endif

#include <scsi/scsi_eh.h>       /* error recovery functions        */
#include <scsi/scsi.h>          /* scsi command definition, etc.   */
#include <scsi/scsi_ioctl.h>    /* support generial scsi ioctl call*/
#include <scsi/scsi_cmnd.h>

#include "lin_tape_tape.h"

extern int lin_tape_debug;
extern int petro;
extern int alternate_pathing;
extern int dynamic_attributes;
extern char *dpf_keys;
extern int disable_auto_dump_drive;
extern int busy_retry;
extern int lin_tape_as_sfmp;

extern char lin_tape_trace_type;
extern rwlock_t lin_tape_trace_lock;
extern unchar *lin_tape_log;
extern uint lin_tape_log_len;

extern rwlock_t hba_lock;
extern struct adapter_descriptor *hba_tab;
extern struct descriptor_changer **changer_devs;
extern struct descriptor_tape **tape_devs;
extern rwlock_t tape_des_lock;
extern rwlock_t changer_des_lock;

/* ioctl data structure fields and definitations */
#define ResetPageCode   0x80
#define ResetPageLength 0x04
#define ResetDiagID1    0x20
#define ResetDiagID2    0x02

struct pos_info {
	unchar bop;
	unchar eop;
	ulong first_blk_pos;
	ulong last_blk_pos;
	ulong blks_in_buf;
	ulong bytes_in_buf;
};

/* 7 byte diagnostic page */
struct diag_page {
	unchar page_code;
	unchar reserved1;
	unchar page_length[2];
	unchar diag_id[2];
	unchar flags;
	unchar reserved2;
};

/* standard inquiry data field */
struct inquiry {
#if defined(__LITTLE_ENDIAN)
	unchar dev_type:5;
	unchar dev_type_qualifier:3;
	unchar dev_type_modifier:7;
	unchar rm:1;
	unchar ansi:3;
	unchar ecma:3;
	unchar iso:2;
	unchar response_data_format:4;
	unchar hisup:1;
	unchar naca:1;
	unchar trmiop:1;
	unchar aenc:1;
	unchar additional_length;
	unchar reserved;
	unchar reserved1:3;
	unchar mchngr:1;
	unchar reserved2:4;
	unchar soft_reset:1;
	unchar cmd_queue:1;
	unchar reserved3:1;
	unchar linked_cmds:1;
	unchar sync:1;
	unchar wide_16bit:1;
	unchar wide_32bit:1;
	unchar relative_addressing:1;
#elif defined(__BIG_ENDIAN)
	unchar dev_type_qualifier:3;
	unchar dev_type:5;
	unchar rm:1;
	unchar dev_type_modifier:7;
	unchar iso:2;
	unchar ecma:3;
	unchar ansi:3;
	unchar aenc:1;
	unchar trmiop:1;
	unchar naca:1;
	unchar hisup:1;
	unchar response_data_format:4;
	unchar additional_length;
	unchar reserved;
	unchar reserved2:4;
	unchar mchngr:1;
	unchar reserved1:3;
	unchar relative_addressing:1;
	unchar wide_32bit:1;
	unchar wide_16bit:1;
	unchar sync:1;
	unchar linked_cmds:1;
	unchar reserved3:1;
	unchar cmd_queue:1;
	unchar soft_reset:1;
#else
	unchar unsupported[8];
#endif

	unchar vendor_id[8];
	unchar product_id[16];
	unchar revision[4];
	unchar vendor_specific[20];
	unchar reserved4[40];
};

struct inquiry_extended_data {
#if defined(__LITTLE_ENDIAN)
	unchar dev_type           : 5;
	unchar dev_type_qualifier : 3;
#elif defined(__BIG_ENDIAN)
	unchar dev_type_qualifier : 3;
	unchar dev_type           : 5;

#else
	error
#endif
	unchar page_code;
	unchar reserved1;
	unchar page_length;

#if defined(__LITTLE_ENDIAN)
	unchar ref_chk            : 1;
	unchar app_chk            : 1;
	unchar grd_chk            : 1;
	unchar spt                : 3;
	unchar actv_ucode         : 2;
#elif defined(__BIG_ENDIAN)
	unchar actv_ucode         : 2;
	unchar spt                : 3;
	unchar grd_chk            : 1;
	unchar app_chk            : 1;
	unchar ref_chk            : 1;
#else
	error
#endif

	unchar reserved2[59];
};

/* 4 byte mode parameter header for mode sense/mode select 6*/
struct mode_parm_header {
	unchar mode_data_length;
	unchar medium_type;
#if defined(__LITTLE_ENDIAN)
	unchar speed:4;
	unchar buffered_mode:3;
	unchar wp:1;
#elif defined(__BIG_ENDIAN)
	unchar wp:1;
	unchar buffered_mode:3;
	unchar speed:4;
#else
	unchar unsupported;
#endif
	unchar blk_descriptor_length;
};

#define MAX_MODE_PAGE_10_LENGTH    10240
struct mode_parm_header_10
{
	unchar mode_data_length[2];
	unchar medium_type;
#if defined(__LITTLE_ENDIAN)
	unchar speed:4;
	unchar buffered_mode:3;
	unchar wp:1;
#elif defined(__BIG_ENDIAN)
	unchar wp:1;
	unchar buffered_mode:3;
	unchar speed:4;
#else
	unchar unsupported;
#endif
	unchar reserved[2];
	unchar blk_descriptor_length[2];
};

struct generic_mode_page_10
{
	struct mode_parm_header_10 header;
	unchar mode_page[MAX_MODE_PAGE_10_LENGTH];
};


/* mode block descriptor for mode sense / mode select 6 and 10*/
struct mode_block_descriptor {
	unchar density_code;
	unchar blks_num[3];
	unchar reserved;
	unchar blk_length[3];
};

/* set block size mode select data */
struct media_parm {
	struct mode_parm_header header;
	struct mode_block_descriptor block;
};

/* rw error recovery page */
#define RWErrorRecoveryPageNumber    0x01
struct rw_err_recovery {
	struct mode_parm_header header;
	unchar page[12];
};

/* sequential access device configuration page */
#define DeviceConfigurationPageNumber 0x10
#define DvcCfgExtSubpage              0x01
#define DataCompressionPageNumber     0x0F
/* device configuration mode buffer */
struct dev_conf_parm {
	struct mode_parm_header header;
	unchar page[16];
};

/* compression parameter */
struct compression_parm {
	struct mode_parm_header header;
	unchar page[16];
};

/* read position data */
struct read_pos {
#if defined(__LITTLE_ENDIAN)
	unchar reserved1:2;
	unchar bpu:1;
	unchar reserved2:1;
	unchar bycu:1;
	unchar bcu:1;
	unchar eop:1;
	unchar bop:1;
#elif defined(__BIG_ENDIAN)
	unchar bop:1;
	unchar eop:1;
	unchar bcu:1;
	unchar bycu:1;
	unchar reserved2:1;
	unchar bpu:1;
	unchar reserved1:2;
#else
	unchar unsupported;
#endif
	unchar partition;
	unchar reserved3[2];
	unchar first_blk_pos[4];
	unchar last_blk_pos[4];
	unchar reserved4;
	unchar blks_in_buf[3];
	unchar bytes_in_buf[4];
};

struct _short_data_format {
#if defined __LITTLE_ENDIAN
	unchar bpew : 1;
	unchar perr : 1;
	unchar lolu : 1;
	unchar rsvd : 1;
	unchar bycu : 1;
	unchar locu : 1;
	unchar eop : 1;
	unchar bop : 1;
#elif defined __BIG_ENDIAN
	unchar bop : 1;
	unchar eop : 1;
	unchar locu : 1;
	unchar bycu : 1;
	unchar rsvd : 1;
	unchar lolu : 1;
	unchar perr : 1;
	unchar bpew : 1;
#else
		error
#endif
	unchar active_partition;
	char reserved[2];
	unchar first_logical_obj_position[4];
	unchar last_logical_obj_position[4];
	unchar num_buffer_logical_obj[4];
	unchar num_buffer_bytes[4];
	char reserved1;
};

struct _long_data_format {
#if defined __LITTLE_ENDIAN
	unchar bpew : 1;
	unchar rsvd2 : 1;
	unchar lonu : 1;
	unchar mpu : 1;
	unchar rsvd1 : 2;
	unchar eop : 1;
	unchar bop : 1;
#elif defined __BIG_ENDIAN
	unchar bop : 1;
	unchar eop : 1;
	unchar rsvd1 : 2;
	unchar mpu : 1;
	unchar lonu : 1;
	unchar rsvd2 : 1;
	unchar bpew : 1;
#else
		error
#endif
	char reserved[6];
	unchar active_partition;
	unchar logical_obj_number[8];
	unchar logical_file_id[8];
	unchar obsolete[8];
};

struct _extended_data_format {
#if defined __LITTLE_ENDIAN
	unchar bpew : 1;
	unchar perr : 1;
	unchar lolu : 1;
	unchar rsvd : 1;
	unchar bycu : 1;
	unchar locu : 1;
	unchar eop : 1;
	unchar bop : 1;
#elif defined __BIG_ENDIAN
	unchar bop : 1;
	unchar eop : 1;
	unchar locu : 1;
	unchar bycu : 1;
	unchar rsvd : 1;
	unchar lolu : 1;
	unchar perr : 1;
	unchar bpew : 1;
#else
		error
#endif
	unchar active_partition;
	unchar additional_length[2];
	unchar num_buffer_logical_obj[4];
	unchar first_logical_obj_position[8];
	unchar last_logical_obj_position[8];
	unchar num_buffer_bytes[8];
	unchar reserved;
};

struct _read_tape_position {
	unchar data_format;
	union {
		struct short_data_format rp_short;
		struct long_data_format rp_long;
		struct extended_data_format rp_extended;
	} rp_data;
};

/* medium sense page (359x only) */
#define MediumSensePageNumber 0x23
#define Capacity100 0x00
#define Capacity75  0xC0
#define Capacity50  0x80
#define Capacity25  0x40
#define Capacity100RangeHigh 0xff
#define Capacity100RangeLow  0xC1
#define Capacity75RangeHigh  0xC0
#define Capacity75RangeLow   0x81
#define Capacity50RangeHigh  0x80
#define Capacity50RangeLow   0x41
#define Capacity25RangeHigh  0x40
#define Capacity25RangeLow   0x01

struct medium_sense_page {
#if defined(__LITTLE_ENDIAN)
	unchar page_code:6;
	unchar reserved1:1;
	unchar page_same:1;
	unchar page_length;
	unchar reserved2[2];
	unchar medium_id[2];
	unchar format_id;
	unchar reserved3[2];
	unchar worm_mode:4;
	unchar worm_locked:1;
	unchar reserved4:3;
	unchar permanent_wp:1;
	unchar reserved5:2;
	unchar reset_persistent_wp:1;
	unchar persistent_wp:1;
	unchar reset_associate_wp:1;
	unchar associated_wp:1;
	unchar physical_wp:1;
	unchar capacity_scaling_valid:1;
	unchar seg_scaling_valid:1;
	unchar reserved6:6;
	unchar capacity_scaling;
	unchar medium_capacity[4];
	unchar partition_num;
	unchar reserved7[2];
	unchar edge_track:1;
	unchar reserved:7;
	unchar internal_vid_source:3;
	unchar reserved8:3;
	unchar dbm_valid:1;
	unchar internal_vid_valid:1;
#elif defined(__BIG_ENDIAN)
	unchar page_save:1;
	unchar reserved1:1;
	unchar page_code:6;
	unchar page_length;
	unchar reserved2[2];
	unchar medium_id[2];
	unchar format_id;
	unchar reserved3[2];
	unchar reserved4:3;
	unchar worm_locked:1;
	unchar worm_mode:4;
	unchar physical_wp:1;
	unchar associated_wp:1;
	unchar reset_associate_wp:1;
	unchar persistent_wp:1;
	unchar reset_persistent_wp:1;
	unchar reserved5:2;
	unchar permanent_wp:1;
	unchar reserved6:7;
	unchar capacity_scaling_valid:1;
	unchar capacity_scaling;
	unchar medium_capacity[4];
	unchar partition_num;
	unchar reserved7[2];
	unchar reserved:7;
	unchar edge_track:1;
	unchar internal_vid_valid:1;
	unchar dbm_valid:1;
	unchar reserved8:3;
	unchar internal_vid_source:3;
#else
	unchar unsupported[22];
#endif
	unchar internal_vid[6];
	unchar partition_capacity[4];
	unchar kb_transfered[4];
	unchar blks_written[4];
	unchar curr_file[4];
	unchar files_written[4];
	unchar ww_cartridge_id[12];
};

/* medium sense page buffer */
struct medium_sense {
	struct mode_parm_header header;
	struct medium_sense_page page;
};

/* time-of-day control page for 3592 */
#define TODControlPageNumber 0x21
struct tod_control {
	struct mode_parm_header header;
	unchar page[12];
};

#define InitiatorSpecPageNumber 0x24

/* read/write control for 3592 */
#define RWControlPageNumber 0x25
struct rw_control_page{
#if defined(__LITTLE_ENDIAN)
	unchar page_code:6;
	unchar reserved:1;
	unchar page_save:1;
	unchar page_length;
	unchar ignore_sequence_check;
	unchar ignore_data_check;
	unchar reserved1[2];
	unchar leop[2];
	unchar disable_fastsync:1;
	unchar reserved2:5;
	unchar early_warning:1;
	unchar housekeeping_errors:1;
	unchar default_density;
	unchar pending_density;
	unchar reserved3[21];
#elif defined(__BIG_ENDIAN)
	unchar page_save:1;
	unchar reserved:1;
	unchar page_code:6;
	unchar page_length;
	unchar ignore_sequence_check;
	unchar ignore_data_check;
	unchar reserved1[2];
	unchar leop[2];
	unchar housekeeping_errors:1;
	unchar early_warning:1;
	unchar reserved2:5;
	unchar disable_fastsync:1;
	unchar default_density;
	unchar pending_density;
	unchar reserved3[21];
#else
	unchar unsupported[32];
#endif
};
struct rw_control{
	struct mode_parm_header header;
	struct mode_block_descriptor block;
	struct rw_control_page page;
};

/* read block limits data buffer */
struct blk_limits {
	unchar reserved;
	unchar max_blk_size[3];
	unchar min_blk_size[2];
};

/* log sense buffer */
#define MaxLogSense   4096    /* for full log sense */
#define SimMimLogPage 0x31
#define TapeAlertLogPage 0xE2
#define LogPageHeaderSize 4

/* read dump */
struct read_buffer_header {
	unchar reserved;
	unchar buf_length[3];
};

/* report density support */
#define MaxDensityReports 8
struct report_density_header {
	unchar avail_length[2];
	unchar reserved0[2];
};

struct report_density_descriptor {
	unchar primary_density;
	unchar second_density;
#if defined(__LITTLE_ENDIAN)
	unchar reserved1:5;
	unchar deflt:1;
	unchar dup:1;
	unchar wrtok:1;
#elif defined(__BIG_ENDIAN)
	unchar wrtok:1;
	unchar dup:1;
	unchar deflt:1;
	unchar reserved1:5;
#else
	unchar unsupported;
#endif
	unchar reserved2;
	unchar reserved3;
	unchar bits_per_mm[3];
	unchar media_width[2];
	unchar tracks[2];
	unchar capacity[4];
	unchar assign_org[8];
	unchar density_name[8];
	unchar description[20];
};

struct report_density {
	struct report_density_header header;
	struct report_density_descriptor descriptor[MaxDensityReports];
};

/* display message (359x only) */
struct display_msg {
	unchar type;
	unchar function;
	unchar reserved[4];
	unchar length[2];
	unchar msg0[MAXMSGLEN];
	unchar msg1[MAXMSGLEN];
};

/* persistent reservation */
#define DPF_KEY_LENGTH 16
#define QUERY_RESERVED_KEYS     0       /* query all registered keys         */
#define QUERY_RESERVATIONS      1       /* query current persistent reserves */
#define QUERY_FULL_STATUS       3       /* query full reservation status     */

/* reservation types */
#define EXCLUSIVE_ACCESS          3
#define REGISTRANTS_ONLY          6

/* service action codes */
#define REGISTER_KEY              0
#define CREATE_PERSIST_RESERVE    1
#define RELEASE_PERSIST_RESERVE   2
#define CLEAR_KEYS_RESERVES       3
#define PREEMPT_PERSIST_RESERVE   4
#define PREEMPT_AND_ABORT         5
#define REGISTER_IGNORE_KEY       6

#define PERSIST_PAGE_HEAD_SIZE    8

struct persist_read_keys {
	unchar generation[4];
	unchar additional_length[4];
	char key_list[1];
};

struct persist_reserve {
	char reserve_key[RESERVE_KEY_LENGTH];
	unchar scope_addr[4];
	unchar reserved;
#if defined(__LITTLE_ENDIAN)
	unchar type:4;
	unchar scope:4;
#elif defined(__BIG_ENDIAN)
	unchar scope:4;
	unchar type:4;
#else
	unchar unsupported;
#endif
	unchar extra_length[2];
};

struct persist_read_reservs {
	unchar generation[4];
	unchar additional_length[4];
	struct persist_reserve key_descriptor_list[1];
};

#define PROTO_ID_FCP   (0x00)
#define PROTO_ID_SCSI  (0x01)
#define PROTO_ID_SAS   (0x06)

struct pr_transport_id {
#if defined(__LITTLE_ENDIAN)
	unchar proto:4;
	unchar reserved1:2;
	unchar format:2;
#elif defined(__BIG_ENDIAN)
	unchar format:2;
	unchar reserved1:2;
	unchar proto:4;
#else
	unchar unsupported;
#endif
};

struct pr_fc_transport_id {
	struct pr_transport_id hdr;
	unchar reserved1[7];
	unchar n_port_name[8];
	unchar reserved2[8];
};

struct pr_scsi_transport_id {
	struct pr_transport_id hdr;
	unchar reserved1;
	unchar scsi_address[2];
	unchar obsolete[2];
	unchar rel_tg_id[2];
	unchar reserved2[16];
};

struct pr_sas_transport_id {
	struct pr_transport_id hdr;
	unchar reserved1[3];
	unchar sas_address[8];
	unchar reserved2[12];
};

struct pr_full_status_descriptor {

	unchar reservation_key[8];   /* 0-7 */
	unchar reserved1[4];         /* 8-11 */
#if defined(__LITTLE_ENDIAN)         /* 12-13 */
	unchar r_holder:1;
	unchar all_tg_pt:1;
	unchar reserved2:6;
	unchar type:4;
	unchar scope:4;
#elif defined(__BIG_ENDIAN)
	unchar reserved2:6;
	unchar all_tg_pt:1;
	unchar r_holder:1;
	unchar scope:4;
	unchar type:4;
#else
	unchar unsupported[2];
#endif
	unchar reserved3[4];         /* 14-17 */
	unchar rel_tg_port_id[2];    /* 18-19 */
	unchar additional_length[4]; /* 20-23 */
	unchar additional_data[1];
};

struct pr_full_status {
	unchar generation[4];
	unchar additional_length[4];
	unchar full_status_descriptor[1];
};

struct persist_reserve_out {
	char reserve_key[8];
	char service_key[8];
	unchar scope_spec_addr[4];
#if defined(__LITTLE_ENDIAN)
	unchar aptpl:1;
	unchar reserved1:7;
#elif defined(__BIG_ENDIAN)
	unchar reserved1:7;
	unchar aptpl:1;
#else
	unchar unsupported;
#endif
	unchar reserved;
	unchar ext_length[2];
};


#define ElementAddressPage        0x1D
#define MaxElementDescriptionSize 52    /* if volume tag available */
#define MinElementDescriptionSize 12    /* without Volume Tag      */
#define MaxElementDescriptionWithIDSize 116
#define MinElementDescriptionWithIDSize 80
#define ElDescriptorWithSEDvcIdSize (sizeof(struct cartridge_location_data))
#define MinDeviceID 16
#define MediumTransportReport 1
#define StorageElementReport  2
#define IEPortReport          3
#define DataTransferReport    4

/* element address assignment page */
struct elem_assignment {
#if defined(__LITTLE_ENDIAN)
	unchar page_code:6;
	unchar reserved:1;
	unchar ps:1;
#elif defined(__BIG_ENDIAN)
	unchar ps:1;
	unchar reserved:1;
	unchar page_code:6;
#else
	unchar unsupported;
#endif
	unchar parm_length;
	unchar medium_transport_elem_addr[2];
	unchar transport_elem_num[2];
	unchar first_storage_addr[2];
	unchar storage_elem_num[2];
	unchar first_ie_port_addr[2];
	unchar ie_port_elem_num[2];
	unchar first_data_transfer_addr[2];
	unchar data_transfer_num[2];
	unchar reserved2[2];
};

struct elem_status_data {
	unchar first_elem_addr_reported[2];
	unchar elem_num[2];
	unchar reserved;
	unchar byte_cnt[3];
};

struct elem_status_page {
	unchar type_code;
#if defined(__LITTLE_ENDIAN)
	unchar reserved:6;
	unchar avol_tag:1;
	unchar pvol_tag:1;
#elif defined(__BIG_ENDIAN)
	unchar pvol_tag:1;
	unchar avol_tag:1;
	unchar reserved:6;
#else
	unchar unsupported;
#endif
	unchar elem_descriptor_length[2];
	unchar reserved1;
	unchar byte_cnt[3];
};

#define VOL_TAG_SIZE            (36)
#define TAPE_IDENTIFIER_SIZE    (36)
#define VOL_TAG_OFFSET          (12)

struct tape_elem_status {
	ushort addr;              /* 0-1 */
	unchar flags1_0:1;        /* 2 */
	unchar flags1_1:1;
	unchar flags1_2:1;
	unchar flags1_3:1;
	unchar flags1_4:1;
	unchar flags1_5:1;
	unchar flags1_6:1;
	unchar flags1_7:1;
	unchar reserved1;         /* 3 */
	unchar asc;               /* 4 */
	unchar ascq;              /* 5 */
	unchar flags2_0:1;        /* 6 */
	unchar flags2_1:1;
	unchar flags2_2:1;
	unchar flags2_3:1;
	unchar flags2_4:1;
	unchar flags2_5:1;
	unchar flags2_6:1;
	unchar flags2_7:1;
	unchar scsi;              /* 7 */
	unchar reserved2;         /* 8 */
	unchar flags3_0:1;        /* 9 */
	unchar flags3_1:1;
	unchar flags3_2:1;
	unchar flags3_3:1;
	unchar flags3_4:1;
	unchar flags3_5:1;
	unchar flags3_6:1;
	unchar flags3_7:1;
	ushort source;            /* 10-11 */
	unchar volume[VOL_TAG_SIZE];
	unchar reserved3[4];
};

struct tape_read_devid {
	ushort addr;
	unchar flags1_0:1;
	unchar flags1_1:1;
	unchar flags1_2:1;
	unchar flags1_3:1;
	unchar flags1_4:1;
	unchar flags1_5:1;
	unchar flags1_6:1;
	unchar flags1_7:1;
	unchar reserved1;
	unchar asc;
	unchar ascq;
	unchar flags2_0:1;
	unchar flags2_1:1;
	unchar flags2_2:1;
	unchar flags2_3:1;
	unchar flags2_4:1;
	unchar flags2_5:1;
	unchar flags2_6:1;
	unchar flags2_7:1;
	unchar scsi;
	unchar reserved2;
	unchar flags3_0:1;
	unchar flags3_1:1;
	unchar flags3_2:1;
	unchar flags3_3:1;
	unchar flags3_4:1;
	unchar flags3_5:1;
	unchar flags3_6:1;
	unchar flags3_7:1;
	ushort source;
	/* volume tag info is omitted here for LTO device */
	unchar flags4_0:1;
	unchar flags4_1:1;
	unchar flags4_2:1;
	unchar flags4_3:1;
	unchar flags4_4:1;
	unchar flags4_5:1;
	unchar flags4_6:1;
	unchar flags4_7:1;
	unchar flags5_0:1;
	unchar flags5_1:1;
	unchar flags5_2:1;
	unchar flags5_3:1;
	unchar flags5_4:1;
	unchar flags5_5:1;
	unchar flags5_6:1;
	unchar flags5_7:1;
	unchar reserved3;
	unchar id_length;
	unchar identifier[TAPE_IDENTIFIER_SIZE];
};

struct device_config_page
{
#if defined __LITTLE_ENDIAN
	unchar page_code : 6;                    /* 0 */
	unchar spf : 1;
	unchar ps : 1;
#elif defined __BIG_ENDIAN
	unchar ps : 1;
	unchar spf : 1;
	unchar page_code : 6;
#else
	error
#endif

	unchar page_length;                      /* 1 */

#if defined __LITTLE_ENDIAN
	unchar active_format : 5;                /* 2 */
	unchar caf : 1;
	unchar obs_1 : 1;
	unchar reserved : 1;
#elif defined __BIG_ENDIAN
	unchar reserved : 1;
	unchar obs_1 : 1;
	unchar caf : 1;
	unchar active_format : 5;
#else
	error
#endif

	unchar active_partition;                 /* 3 */
	unchar wt_buf_full;                      /* 4 */
	unchar rd_buf_empty;                     /* 5 */
	unchar wt_delay_time[2];                 /* 6-7 */

#if defined __LITTLE_ENDIAN
	unchar rew : 1;                          /* 8 */
	unchar robo : 1;
	unchar socf : 2;
	unchar avc : 1;
	unchar obs_2 : 1;
	unchar lois : 1;
	unchar obr : 1;
#elif defined __BIG_ENDIAN
	unchar obr : 1;
	unchar lois : 1;
	unchar obs_2 : 1;
	unchar avc : 1;
	unchar socf : 2;
	unchar robo : 1;
	unchar rew : 1;
#else
	error
#endif

	unchar obs_3;                            /* 9 */

#if defined __LITTLE_ENDIAN
	unchar bam : 1;                          /* 10 */
	unchar baml : 1;
	unchar swp : 1;
	unchar sew : 1;
	unchar eeg : 1;
	unchar eod_defd : 3;
#elif defined __BIG_ENDIAN
	unchar eod_defd : 3;
	unchar eeg : 1;
	unchar sew : 1;
	unchar swp : 1;
	unchar baml : 1;
	unchar bam : 1;
#else
	error
#endif

	unchar buf_size_at_warn[3];              /* 11-13 */
	unchar sdc_algorithm;                    /* 14 */

#if defined __LITTLE_ENDIAN
	unchar prmwp : 1;                        /* 15 */
	unchar perswp : 1;
	unchar asocwp : 1;
	unchar rew_on_reset : 2;
	unchar oir : 1;
	unchar wtre : 2;
#elif defined __BIG_ENDIAN
	unchar wtre : 2;
	unchar oir : 1;
	unchar rew_on_reset : 2;
	unchar asocwp : 1;
	unchar perswp : 1;
	unchar prmwp : 1;
#else
	error
#endif
};

struct device_config_ext_subpage {
#if defined __LITTLE_ENDIAN
	unchar page_code : 6;                    /* 0 */
	unchar spf : 1;
	unchar ps : 1;
#elif defined __BIG_ENDIAN
	unchar ps : 1;
	unchar spf : 1;
	unchar page_code : 6;
#else
	error
#endif
	unchar sub_page_code;                    /* 1 */
	unchar page_length[2];                   /* 2-3 */
#if defined __LITTLE_ENDIAN
	unchar taplsd : 1;                       /* 4 */
	unchar tarpc : 1;
	unchar taser : 1;
	unchar tarpf : 1;
	unchar reserved1 : 4;

	unchar short_erase_mode : 4;             /* 5 */
	unchar write_mode : 4;
#elif defined __BIG_ENDIAN
	unchar reserved1 : 4;
	unchar tarpf : 1;
	unchar taser : 1;
	unchar tarpc : 1;
	unchar taplsd : 1;

	unchar write_mode : 4;
	unchar short_erase_mode : 4;
#else
	error
#endif

	unchar pews[2];                          /* 6-7 */
#if defined __LITTLE_ENDIAN
	unchar vcelbre : 1;                      /* 8 */
	unchar reserved2 : 7;
#elif defined __BIG_ENDIAN
	unchar reserved2 : 7;
	unchar vcelbre : 1;
#else
	error
#endif
	unchar reserved3[23];                    /* 9-31 */
};

struct dvc_cfg_ext_subpg_with_hdr
{
	struct mode_parm_header_10 header;
	struct device_config_ext_subpage page;
};

struct medium_partition_page
{
#if defined __LITTLE_ENDIAN
	unchar page_code : 6;                    /* 0 */
	unchar spf : 1;
	unchar ps : 1;
#elif defined __BIG_ENDIAN
	unchar ps : 1;
	unchar spf : 1;
	unchar page_code : 6;
#else
	error
#endif

	unchar page_length;                      /* 1 */
	unchar max_addtl_parts;                  /* 2 */
	unchar addtl_parts_defd;                 /* 3 */

#if defined __LITTLE_ENDIAN
	unchar addp : 1;                         /* 4 */
	unchar clear : 1;
	unchar pofm : 1;
	unchar psum : 2;
	unchar idp : 1;
	unchar sdp : 1;
	unchar fdp : 1;
#elif defined __BIG_ENDIAN
	unchar fdp : 1;
	unchar sdp : 1;
	unchar idp : 1;
	unchar psum : 2;
	unchar pofm : 1;
	unchar clear : 1;
	unchar addp : 1;
#else
	error
#endif

	unchar medium_fmt_recg;                  /* 5 */

#if defined __LITTLE_ENDIAN
	unchar partition_units : 4;              /* 6 */
	unchar partition_type : 4;
#elif defined __BIG_ENDIAN
	unchar partition_type : 4;
	unchar partition_units : 4;
#else
	error
#endif

	unchar reserved_2;                       /* 7 */
	unchar first_part_size[2];               /* 8-9 */
	unchar last_part_size[2];                /* 10-11 */
};

#define MODE_SENSE_CONTROL_PG (0x0A)
#define CONTROL_PG_DP_SUBPG   (0xF0)

struct control_mode_page_dp_subp {
#if defined __LITTLE_ENDIAN
	unchar page_code : 6;
	unchar spf : 1;
	unchar ps : 1;
#elif defined __BIG_ENDIAN
	unchar ps : 1;
	unchar spf : 1;
	unchar page_code : 6;
#else
	error
#endif
	unchar subpg_code;
	unchar page_length[2];
	unchar lb_prot_method;

#if defined __LITTLE_ENDIAN
	unchar lb_prot_info_length : 6;
	unchar reserved1 : 2;

	unchar reserved2 : 5;
	unchar rbdp : 1;
	unchar lbp_r : 1;
	unchar lbp_w : 1;
#elif defined __BIG_ENDIAN
	unchar reserved1 : 2;
	unchar lb_prot_info_length : 6;

	unchar lbp_w : 1;
	unchar lbp_r : 1;
	unchar rbdp : 1;
	unchar reserved2 : 5;
#else
	error
#endif

	unchar reserved3[25];
};

struct control_pg_dp_subp_with_hdr
{
	struct mode_parm_header_10 header;
	struct control_mode_page_dp_subp page;
};

struct scsi_execute_callback
{
	struct completion scsi_execute_lock;
	int resid;
	int status_code;
	char sense_data[SCSI_SENSE_BUFFERSIZE];
};

/* Busy Retry */
#define BUSY_RETRY (480)

/* Dynamic host attributes */
#define DRA_DSFN             (0x1800) /* device special filename */
#define DRA_DPF_ACTIVE_PATH  (0x1801)
#define DRA_HOST_NAME        (0x1802)
#define DRA_OS               (0x1803)
#define DRA_OS_VER           (0x1804)
#define DRA_DD_NAME          (0x1805)
#define DRA_DD_VER           (0x1806)
#define DRA_PID              (0x1807)

#define DRA_DSFN_LEN         (32)
#define DRA_ACTIVE_PATH_LEN  (4)
#define DRA_HOST_NAME_LEN    (32)
#define DRA_OS_LEN           (16)
#define DRA_OS_VER_LEN       (32)
#define DRA_DD_NAME_LEN      (16)
#define DRA_DD_VER_LEN       (16)
#define DRA_PID_LEN          (8)

#define DRA_LIST_HDR_LEN  (8)

struct dyn_attr {
	unsigned char i_t_nexus[2];
	unsigned char attr_id[2];
#if defined __LITTLE_ENDIAN
	unsigned char format :2;
	unsigned char reserved : 5;
	unsigned char read_only : 1;
#elif defined __BIG_ENDIAN
	unsigned char read_only : 1;
	unsigned char reserved : 5;
	unsigned char format :2;
#endif
	unsigned char attr_length[4];
	unsigned char attr_value[1];
};

struct dyn_attr_parm_list {
	unsigned char reserved[DRA_LIST_HDR_LEN];
	unsigned char attrs[1];
};

struct dyn_attrs {
	unsigned int id;
	unsigned int len;
	char* val_str;
};

extern struct dyn_attrs dyn_attrs[];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#define utsname() &system_utsname
#endif

#define SCSIOP_WRITE_DYNAMIC_RUNTIME (0xD2)

struct feature_switches_subpage {
#if defined __LITTLE_ENDIAN
	unchar page_code : 6;                    /* 0x30 */
	unchar spf : 1;                          /* 1b */
	unchar ps : 1;
#elif defined __BIG_ENDIAN
	unchar ps : 1;
	unchar spf : 1;
	unchar page_code : 6;
#else
	error
#endif
	unchar sub_page_code;                    /* 0x43 */
	unchar page_length[2];                   /* 0x10 */
#if defined __LITTLE_ENDIAN
    unchar dbopc : 1;
    unchar earchive : 1;
    unchar reserved : 6;
#elif defined __BIG_ENDIAN
    unchar reserved : 6;
    unchar earchive : 1;
    unchar dbopc : 1;
#else
    error
#endif
    unchar reserved2[15];
};

/********************* Function Prototypes **********************************/

/* tape ioctls */
ssize_t tape_capacity_scaling(struct tape_descriptor**, unchar, unchar);
ssize_t tape_set_compression(struct tape_descriptor**, unchar);
int tape_data_safe_mode_and_pew(struct tape_descriptor** drvp,
	struct stchgp_s* parm_buf);
ssize_t tape_display_msg(struct tape_descriptor**, char, char*, char*);
ssize_t tape_erase(struct tape_descriptor**, unchar, unchar);
ssize_t tape_get_compression(struct tape_descriptor**);
ssize_t tape_get_media_parm(struct tape_descriptor**);
ssize_t tape_wp_capscal(struct tape_descriptor** drvp);
ssize_t tape_get_pos(struct tape_descriptor**, struct pos_info*, unchar,
	boolean);
ssize_t tape_inquiry_page(struct tape_descriptor**, unchar, unchar*);
ssize_t tape_load_unload(struct tape_descriptor**, unchar, unchar);
ssize_t tape_get_timeouts(struct tape_descriptor* drv);
ssize_t tape_full_logsense(struct tape_descriptor**);
ssize_t tape_logsense_page(struct tape_descriptor**, unchar, unchar, unchar*,
	uint*);
ssize_t tape_modesense_page(struct tape_descriptor**, unchar, unchar*, int,
	boolean, boolean);
ssize_t tape_modeselect(struct tape_descriptor** drvp, unchar* mode_buf,
	uint allocation_length);
ssize_t tape_modesense10_page(struct tape_descriptor **drvp, unchar page_code,
	unchar *modesense_buf, int alloc_len, int subpg, boolean dbd,
	boolean return_header);
ssize_t tape_modeselect10(struct tape_descriptor** drvp, unchar* mode_buf,
	boolean sp, uint alloc_len);
ssize_t tape_prevent_medium_remove(struct tape_descriptor**, unchar);
ssize_t tape_release(struct tape_descriptor**);
ssize_t tape_report_density(struct tape_descriptor**, unchar, ushort*,
	struct report_density*);
ssize_t tape_request_sense(struct tape_descriptor**, int*, int*, unchar*);
ssize_t tape_reserve(struct tape_descriptor**);
ssize_t tape_reset_drive(struct tape_descriptor**);
ssize_t tape_rewind(struct tape_descriptor**, unchar);
ssize_t tape_self_test(struct tape_descriptor**);
ssize_t tape_set_all_parms(struct tape_descriptor**);
ssize_t tape_set_media_parm(struct tape_descriptor**, unchar, unchar, unchar,
	ulong);
ssize_t tape_set_medium_sense(struct tape_descriptor**, unchar);
ssize_t tape_set_tod(struct tape_descriptor**);
ssize_t tape_set_pos(struct tape_descriptor**, int, long long);
ssize_t tape_std_inquiry(struct tape_descriptor** drvp, void* buf, char unused);
ssize_t tape_tur(struct tape_descriptor**);
ssize_t tape_update_all_parms(struct tape_descriptor**);
ssize_t tape_update_block_limits(struct tape_descriptor**);
ssize_t tape_wfm(struct tape_descriptor**, unchar, ulong);
ssize_t tape_set_rw_error_recovery_page(struct tape_descriptor**, unchar,
	unchar);
ssize_t tape_get_density(struct tape_descriptor **drvp, void* buf, char unused);
ssize_t tape_set_density(struct tape_descriptor** drvp, void* buf, char unused);

ssize_t tape_get_encryption_capability(struct tape_descriptor**);
ssize_t tape_get_encryption_enablement(struct tape_descriptor**);
ssize_t tape_get_encryption_state(struct tape_descriptor**,
	struct encryption_status*);
ssize_t tape_set_encryption_state(struct tape_descriptor**,
	struct encryption_status*);
ssize_t tape_set_data_key(struct tape_descriptor**, struct data_key*);
ssize_t tape_discard_data_key (struct tape_descriptor**);
int encryption_diag(struct tape_descriptor**,
	struct encryption_diagnostics* diag);
int tape_perform_encryption_generic(struct tape_descriptor **drvp, int timeout,
	int* proxy_action, int* return_code);
int tape_encryption_write_setting(struct tape_descriptor **drvp);
int tape_perform_encryption(struct tape_descriptor **drvp, int timeout,
	unchar direction, size_t xfer_cnt, int retries);
int lin_tape_query_partition(struct tape_descriptor* drv, uint command,
	ulong arg);
int lin_tape_create_partition(struct tape_descriptor* drv, uint command,
	ulong arg);
int lin_tape_set_active_partition(struct tape_descriptor** drvp, uint command,
	ulong arg);
int lin_tape_allow_data_overwrite(struct tape_descriptor* drv, uint command,
	ulong arg);
ssize_t tape_get_pos_ex(struct tape_descriptor** drvp, uint command, ulong arg);
int lin_tape_dyn_attrs_parm_list(struct tape_descriptor* drv, char* parm_list,
	int len_minus_hdr, boolean clear);
int lin_tape_query_amu(struct tape_descriptor* drv, uint command,ulong arg);
int lin_tape_set_amu(struct tape_descriptor* drv, uint command,	ulong arg);
int lin_tape_load_unload(struct tape_descriptor **drvp, uint command,ulong arg);
int tape_pass_thru(struct tape_descriptor **drvp, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar *sense_length,
	unchar *sense_avail, int *resid, int *result, boolean fpfo);
int tape_pass_thru_persistent_reserve(struct tape_descriptor** drvp,
	unchar* sense, unchar orig_sense_length, unchar* sense_avail,
	unchar* new_slen, boolean fpfo);
int tape_pass_thru_persistent_release(struct tape_descriptor** drvp,
	unchar* sense, unchar orig_sense_length, unchar* sense_avail,
	unchar* new_slen, boolean fpfo);
int tape_pass_thru_handle_buffers(struct tape_descriptor* drv, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, ulong* order, boolean fpfo);
int tape_pass_thru_unsuccessful(struct tape_descriptor** drvp, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, unchar* sense,
	unchar orig_sense_length, unchar* sense_avail, int* resid,
	unchar* new_slen, boolean fpfo, int rc);
int tape_pass_thru_successful(struct tape_descriptor** drvp, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, unchar* sense,
	unchar orig_sense_length, unchar* sense_avail, int* resid,
	unchar* new_slen, boolean fpfo);
ssize_t tape_query_path(struct tape_descriptor** drvp, void* buf, char unused);
ssize_t tape_show_paths(struct tape_descriptor** drvp, void* buf, char unused);
ssize_t tape_qry_blk_prot(struct tape_descriptor** drvp, void* buf,
	char unused);
ssize_t tape_set_blk_prot(struct tape_descriptor** drvp, void* buf,
	char unused);
void store_lbp(struct tape_descriptor* drv,
	struct logical_block_protection* lbp);
int stioc_query_eot_warn(struct tape_descriptor** drvp, uint command, ulong arg);
int stioc_set_eot_warn(struct tape_descriptor** drvp, uint command, ulong arg);
ssize_t stioc_verify_tape_data(struct tape_descriptor** drvp, void* buf,
	char unused);
int handle_verify_on_failover(struct tape_descriptor* drv);
ssize_t stioc_query_rao(struct tape_descriptor** drvp, void* buf, char unused);
int stioc_generate_rao(struct tape_descriptor** drvp, ulong arg);
int execute_grao_sg(struct tape_descriptor** drvp, uint length);
ssize_t stioc_receive_rao(struct tape_descriptor** drvp, ulong arg);
int lin_tape_perform_rrao_sg(struct tape_descriptor*, uint, uint);
int copy_sg_to_user(struct tape_descriptor*, void *, uint);


/* functions for data path failover support */
ssize_t persistent_reserve_in(char* buf, struct tape_descriptor *drv,
	unchar service_action, int drv_buf_len);
ssize_t persistent_reserve_out(struct tape_descriptor *drv,
	unchar service_action, struct persist_reserve_out* parm_list_buf);
ssize_t tape_read_reservation(struct tape_descriptor**, unchar*,
	unsigned int*, unchar*);
ssize_t tape_register_reservation_key(struct tape_descriptor**, unchar*);
ssize_t tape_create_persist_reserve(struct tape_descriptor**);
ssize_t tape_clear_all_registrations(struct tape_descriptor**);
ssize_t tape_read_keys(struct tape_descriptor**, void*, ulong);
ssize_t tape_read_reserves(struct tape_descriptor**, void*, ulong);
int tape_preempt_reservation(struct tape_descriptor** drvp, unchar* old_key,
	unchar* new_key);
int erp_read_buffer(struct tape_descriptor* drv, char* sense_data);
int lin_tape_perform_dump_sg(struct tape_descriptor** drvp, size_t xfer_len,
		union cdb* tmp_cdb);
int tape_del_ops_sglist(struct tape_descriptor *drv);
int lin_tape_perform_ops_sg(struct tape_descriptor** drvp, unchar* tmp_cdb,
	size_t xfer_cnt, unchar direction);
int setup_sg_for_ops(struct tape_descriptor* drv, size_t xfer_cnt);
int tape_copy_ops_sg_buffer(struct tape_descriptor* drv, char* user_location,
	size_t buf_length);
int tape_init_op_sg(struct tape_descriptor*, size_t);
int tape_build_op_sg(struct tape_descriptor*, size_t);

/* tape scsi core functions */
ssize_t tape_send_scsi_cmd(struct tape_descriptor**, int, unchar, size_t, int);
ssize_t tape_send_scsi_io(struct tape_descriptor*, int, unchar, size_t, int);
int tape_log_cdb(struct tape_descriptor* drv);
int tape_send_rw_indirect(struct tape_descriptor* drv, unchar* cdb,
	unchar direction, size_t xfer_cnt, int timeout, unchar* sense_data,
	int* resid, int* status_code);
int tape_execute_scsi_command(struct scsi_device* device, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int *resid);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && (!defined(CONFIG_SLE_SP))
int lin_tape_do_req(struct scsi_device* sdev, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid);

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
int lin_tape_execute_async(struct scsi_device* sdev, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid);

#else
int lin_tape_blk_rq(struct scsi_device* sdev, unchar* cdb, int direction,
	void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && !defined(CONFIG_SLE_SP)
void tape_callback(struct scsi_cmnd*);
#else
void tape_execute_callback(void* data, char* sense_data, int status_code,
	int resid);
#endif

int tape_check_result(struct tape_descriptor *drv, int result,
	unchar *sense_buffer, int resid);
void tape_check_log_cdb_and_status(struct tape_descriptor* drv, int result,
	char ttype, char* trc, unchar* sense_buffer);
void tape_check_reset_on_success(struct tape_descriptor* drv, int resid);
void tape_check_handle_sense(struct tape_descriptor* drv, int result, int* rc,
	unchar* sense_buffer, char* trc, char ttype);
void tape_check_handle_nonsense(struct tape_descriptor* drv, int result,
	int* rc);
void tape_check_write_sense_to_log(struct tape_descriptor* drv, char ttype,
	unchar* trc);
void tape_check_save_sense_information(struct tape_descriptor* drv);
void tape_check_sense_motion_commands(struct tape_descriptor* drv, char ttype);
void tape_check_sense_key_specific(struct tape_descriptor* drv, int* rc);
void tape_check_daemon_functions(struct tape_descriptor* drv, int* rc,
	char ttype);
void tape_check_sk0(struct tape_descriptor* drv, int* rc);
void tape_check_sk1(struct tape_descriptor* drv, int* rc);
void tape_check_sk2(struct tape_descriptor* drv, int* rc);
void tape_check_sk5(struct tape_descriptor* drv, int* rc);
void tape_check_sk6(struct tape_descriptor* drv, int* rc);
void tape_check_skB(struct tape_descriptor* drv, int* rc);
void log_reservation_conflict(struct tape_descriptor* drv);
void log_rsv_process_pr_descriptor(struct tape_descriptor* drv,
	struct pr_full_status_descriptor* ds);
void log_rsv_fcp(struct tape_descriptor* drv, struct pr_fc_transport_id* fc,
	unchar* wwpn);
void log_rsv_scsi(struct tape_descriptor* drv,
	struct pr_scsi_transport_id* scsi, unchar* wwpn);
void log_rsv_sas(struct tape_descriptor* drv, struct pr_sas_transport_id* sas,
	unchar* wwpn);

/* tape error recovery functions */
ssize_t tape_send_erp_cmd(struct tape_descriptor*, int, int, unsigned int);
int erp_inquiry(struct tape_descriptor* drv, int last_timeout,
	unsigned int count, char* sense_data);
int erp_tur_or_reserve(struct tape_descriptor* drv, unchar command,
	char* sense_data);
int erp_locate(struct tape_descriptor* drv, unsigned int count,
	char* sense_data);
int erp_read_position(struct tape_descriptor* drv, char* sense_data);
int erp_read_buffer(struct tape_descriptor* drv, char* sense_data);
int erp_log_sense(struct tape_descriptor* drv, unsigned int count);
int erp_mode_select(struct tape_descriptor* drv, char* sense_data);
int erp_set_block_size(struct tape_descriptor* drv,
	union cdb* tmp_cdb, char* sense_data);
int erp_set_compression(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data);
int erp_logical_write_protect(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data);
int erp_tod(struct tape_descriptor* drv, union cdb* tmp_cdb, char* sense_data);
int erp_density(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data);
int erp_change_definition(struct tape_descriptor* drv, char* sense_data);
int erp_persistent_reserve_in(struct tape_descriptor* drv, unsigned int count,
	unchar* sense_data);
int erp_persistent_reserve_out(struct tape_descriptor* drv, int last_timeout,
	unsigned int count, unchar* sense_data);
int changer_erp_inquiry(struct changer_descriptor* chg, unchar update,
	unsigned int count);
int changer_erp_tur(struct changer_descriptor* chg);
int changer_erp_reserve(struct changer_descriptor* chg);
ssize_t tape_exe_scsi_cmd(struct tape_descriptor *drv, unchar * cdb,
	unchar cdb_len, void *buffer, int buf_len, int timeout,
	unchar direction, int *result, unchar* sense_data, int *resid);
void tape_check_simmim_dump_logsense(struct tape_descriptor*);
void read_simmim(struct tape_descriptor* drv);
void read_dump(struct tape_descriptor* drv);
void read_full_logsense(struct tape_descriptor* drv);
void tape_unit_attent_erp(struct tape_descriptor*);
extern int lin_tape_write_dynamic_attrs(struct tape_descriptor**, boolean);
void tape_post_check1_erp(struct tape_descriptor*);
ssize_t tape_logsense_erp(struct tape_descriptor*, unchar, unchar*, uint*);
ssize_t tape_preempt_erp(struct tape_descriptor*, unchar*, unchar*);
int erp_lbp(struct tape_descriptor* drv, unchar* sense_data);
int erp_prevent_medium_removal(struct tape_descriptor* drv, unchar* sense_data);
int erp_request_sense(struct tape_descriptor* drv, unchar* sense_data);
int erp_write_dynamic_runtime(struct tape_descriptor* drv, unchar* sense_data);
int erp_amu(struct tape_descriptor** drv,unchar* sense_data);
int tape_perform_failover(struct tape_descriptor **drvp, int timeout,
	unchar direction, size_t xfer_cnt, int retries);
int tape_perform_failover_generic(struct tape_descriptor **drvp, int timeout);
int post_dpf_msgs_to_log(struct tape_descriptor* drv);
int adjust_dpf_write_ptr(struct tape_descriptor* drv, size_t* xfer_cnt);
int adjust_dpf_sgl(struct tape_descriptor* drv, struct scatterlist* sgl,
	ulong bytes_to_advance);
void tape_update_to_new_current_path(struct tape_descriptor **drvp);
void tape_get_best_path(struct tape_descriptor **drvp);
void copy_parms_to_new_device(int old_dev_index, int new_dev_index);
void delete_parms_on_device(int dev_index);
void tape_update_descriptor_on_erp(int old_dev_index, int new_dev_index);

#define BUF_RW(drv) (drv->rw_buf && drv->rw_buf_order != -1 && \
	drv->rw_buf_size && !drv->sg_list)
#define SGL_RW(drv) (drv->sg_list && drv->sg_list->total_sgl && \
	drv->sg_list->total_sg_size && drv->sg_list->use_sg && !drv->rw_buf && \
	!drv->rw_buf_size && drv->rw_buf_order == -1)

/* medium changer ioctls */
#define BackToKernel 1
#define SupportVolTag(X) ((X != IBM_3590) && (X != IBM_3592))

ssize_t changer_inquiry_page(struct changer_descriptor**, unchar, unchar*);
ssize_t changer_logsense_page(struct changer_descriptor**, unchar, unchar,
	unchar*, uint*);
ssize_t changer_logsense10_page(struct changer_descriptor**, unchar, unchar,
    char*, unsigned short, uint*);
ssize_t changer_modesense_page(struct changer_descriptor**, unchar, unchar*,
	uint);
ssize_t changer_release(struct changer_descriptor**);
ssize_t changer_request_sense(struct changer_descriptor**, int*, int*, unchar*);
ssize_t changer_reserve(struct changer_descriptor**);
ssize_t changer_std_inquiry(struct changer_descriptor**, void*);
ssize_t changer_tur(struct changer_descriptor**);
ssize_t changer_pos_element(struct changer_descriptor**, ushort, ushort,
	unchar);
ssize_t changer_init_elem_range(struct changer_descriptor**, ushort, ushort);
ssize_t changer_init_element(struct changer_descriptor**);
ssize_t changer_inventory(struct changer_descriptor**, void*, void*, void*,
	void*, unchar);
int changer_read_cartridge_location(struct changer_descriptor* chg, char* arg,
	unchar copy_method);
ssize_t changer_prevent_allow_remove(struct changer_descriptor**, unchar);
ssize_t changer_move_medium(struct changer_descriptor**, ushort, ushort,
	ushort, ushort);
ssize_t changer_exchange_medium(struct changer_descriptor**, ushort, ushort,
	ushort, ushort, char, char);
ssize_t changer_read_devid(struct changer_descriptor**, ushort, ushort, void*,
	unchar);
int changer_get_num_drives(struct changer_descriptor** chgp, uint* drive_num);
int chg_devid_get_drive_report(struct changer_descriptor** chgp, uint drive_num,
	ushort start, ushort elem_num, ulong* order);
void copy_res_to_devid(struct tape_read_devid* elem_info, unchar* buffer,
	uint byte_offset, uint elem_desc_length);

/* Read element status and supporting functions */
ssize_t changer_read_elem_status(struct changer_descriptor**, unchar*,
	struct elem_assignment*, unchar,ushort, int, char*, unchar, boolean);
int alloc_res_out_buf(struct changer_descriptor* chg, ulong* count,
	ulong* order, int elem_num, boolean se_dvc_id);
void create_res_command(struct changer_descriptor* chg, unchar elem_type,
	ushort start_address, int elem_num, boolean se_dvc_id, ulong count);
int validate_return_res_data(struct changer_descriptor* chg, unchar** buffer,
	ulong* report_size, ulong* offset, ulong count, unchar elem_type);
int copy_res_data(unchar* output, unchar elem_type, int num_elems,
	unchar* res_buffer, ulong status_page_offset, uint desc_report_size,
	char* vol_tag, boolean se_dvc_id, unchar copy_method);
int copy_descriptor_data(unchar* output, struct tape_elem_status* elem_buffer,
	uint output_elem_size, unchar elem_type, unchar* res_buffer,
	ulong desc_offset, uint desc_length, char* vol_tag, boolean se_dvc_id,
	unchar copy_method);

ssize_t changer_read_buffer(struct changer_descriptor**, char, int, unchar*);
int changer_pass_thru(struct changer_descriptor **chgr, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar *sense_length,
	unchar *sense_avail, int *resid, int *result, int attempt_number, boolean fpfo);
int changer_pass_thru_send_cmd(struct changer_descriptor **chg, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar *sense_length,
	unchar *sense_avail, int *resid, int *result, boolean fpfo);
int changer_pass_thru_check_parms(struct changer_descriptor* chg, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint local_buffer_length,
	uint data_direction, uint timeout);
int changer_pass_thru_alloc_bufs(struct changer_descriptor* chg,
	unchar* buffer, unchar** sense_buffer, uint data_direction,
	uint local_buffer_length, ulong* order, boolean fpfo);
int changer_pass_thru_handle_sense(struct changer_descriptor* chg,
	unchar* sense, unchar* sense_buffer, unchar* sense_length,
	unchar orig_sense_length, unchar* sense_avail, int* result, boolean fpfo);
ssize_t changer_query_path(struct changer_descriptor** chgp, void* buf,
	char unused);
ssize_t changer_show_paths(struct changer_descriptor** chgp, void* buffer,
	char unused);

/* changer scsi core functions*/
ssize_t changer_send_scsi_cmd(struct changer_descriptor**, int, unchar, int,
	int);
ssize_t changer_send_scsi_io(struct changer_descriptor*, int, unchar, int);
void changer_call_back(struct scsi_cmnd*);
void changer_log_cdb_and_status(struct changer_descriptor* chg, int result,
	unchar* sense_buffer);
int changer_save_sense_data(struct changer_descriptor* chg, int result,
	unchar* sense_buffer, int rc);
int changer_check_result(struct changer_descriptor *chg, int result,
	unchar *sense_buffer, int resid);

/* changer scsi error recovery functions*/
ssize_t changer_send_erp_cmd(struct changer_descriptor*, unchar, unchar,
	unsigned int);

ssize_t changer_exe_scsi_cmd(struct changer_descriptor *chg,
	unchar *cdb, unchar cdb_len, void *buffer, int buf_len,
	int timeout, unchar direction, int* result, unchar *sense_buffer,
	int* resid);
int changer_perform_failover(struct changer_descriptor**,int,unchar,int,int);
int changer_perform_pass_thru_failover(struct changer_descriptor **chgr,
	unchar* cdb, unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar *sense_length,
	unchar *sense_avail, int *resid, int *result, int attempt_number, boolean fpfo);
int changer_failover_ops(struct changer_descriptor** chgp, int timeout, int);
void changer_update_to_new_current_path(struct changer_descriptor **chgr);
void changer_get_best_path(struct changer_descriptor **chgr);
int lin_tape_tsm_ioctl(uint command, ulong arg);

#ifndef DID_TARGET_FAILURE
#define DID_TARGET_FAILURE (0x10)
#endif

/* Additional erp cases: must not conflict with SCSI opcodes, which occupy the
lower eight bits.  So begin at 0x101 */
#define ERP_LBP (0x101)
#define PREVENT_MEDIUM_REMOVAL ALLOW_MEDIUM_REMOVAL

/* help functions */
#define PPGMASK    PAGE_SIZE - 1
#define MAX_PAGE_ORDER (8)
unsigned long get_mem_ord(ulong, ulong*);

extern int tape_rw_buffer(struct tape_descriptor*, size_t);
extern int tape_copy_write_buffer(struct tape_descriptor*, const char*, size_t, boolean fpfo);
extern int tape_copy_read_buffer(struct tape_descriptor*, char*, size_t, boolean fpfo);
extern int tape_build_sg(struct tape_descriptor*, size_t);


#endif
