/*******************************************************************************
 *   lin_tape_scsi_trace.h
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License version 2.1
 *   as published by the Free Software Foundation.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#ifndef _LIN_TAPE_SCSI_TRACE_H_
#define _LIN_TAPE_SCSI_TRACE_H_


/*  lin_tape takes two types of traces:
 *  1. regular trace: trace ioctl entry, not every scsi command
 *  2. full trace: trace every scsi command, status, full sense data
 *
 *  The value of trace type is default to regular trace,
 *  it can be changed in /etc/IBMtaped.conf
 *
 */
#define LIN_TAPE_NO_TRACE      0
#define LIN_TAPE_REGULAR_TRACE 1
#define LIN_TAPE_FULL_TRACE    2
#define LIN_TAPE_INVALID_TRACE 3

/*****************************************************************************
 *                                                                           *
 *          ioctl for setting trace type                                     *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_SET_TRACE_TYPE _IOW('z', 0x85, struct tape_set_trace_type)

struct tape_set_trace_type {
	char trace_type;
};

/*****************************************************************************
 *                                                                           *
 *          ioctls for regular traces (used by IBMtaped daemon)              *
 *                                                                           *
 *****************************************************************************/

  /***************************************************************************
   * POLL_TRACE poll driver for trace notices                                *
   ***************************************************************************/
#define LIN_TAPE_POLL_TRACE _IOR('z', 0x80, struct tape_poll)

#define FullTraceUnitSize 256
struct tape_trace {
	unchar devid[20];
	unchar scsi_path[4];
	unchar msg[255];                      /* for regular trace */
	uint msg_len;                         /* for regular trace */
	unchar full_trc[FullTraceUnitSize];   /* for full trace    */
	uint trc_len;                         /* for full trace    */
	uint dmp_len;
	uint op_sg;
	uint simmim_len;
	uint logsense_len;
};

struct tape_poll {
	uint dev_cnt;
	struct tape_trace *log_trace;
};

  /***************************************************************************
   * RETRIEVE_DRIVE_DUMP retrieve driver dump for trace notices              *
   ***************************************************************************/
#define RETRIEVE_DRIVE_DUMP _IOR('z', 0x81, struct tape_retrv_dump)
struct tape_retrv_dump {
	unchar devid[20];
	uint dmp_len;
	unchar *dmp;
};

  /***************************************************************************
   * RETRIEVE_SIMMIM_DATA retrieve sim mim date from the drive
   ***************************************************************************/
#define RETRIEVE_SIMMIM_DATA _IOR('z', 0x82, struct tape_retrv_simmim)

struct tape_retrv_simmim {
	unchar devid[20];
	uint simmim_len;
	unchar *simmim;
};

  /***************************************************************************
   * RETRIEVE_LOG_DATA retrieve log data form the drive
   ***************************************************************************/
#define RETRIEVE_LOG_DATA _IOR('z', 0x83, struct tape_retrv_logsense)

struct tape_retrv_logsense {
	unchar devid[20];
	uint logsense_len;
	unchar *logsense;
};


/*****************************************************************************
 *                                                                           *
 *          structure for full trace                                         *
 *                                                                           *
 *****************************************************************************/

struct tape_full_trace_unit {
	unchar unit_data[FullTraceUnitSize];
	ushort cur_len;
	struct tape_full_trace_unit *next;
};

/*****************************************************************************
 *                                                                           *
 *          ioctl for getting failover parameters                            *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_GET_FO_PARMS _IOR('z', 0x84, struct tape_get_fo_parms)

#define MAX_DPF_KEYS 64
struct tape_get_fo_parms{
	char alternate_pathing;
	char dpf_keys_avail;
	unchar dpf_keys[MAX_DPF_KEYS * 17];
};


/*****************************************************************************
 *                                                                           *
 *          ioctls for getting device number                                  *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_GET_DRIVE_NUM   _IOR('z', 0x86, int)
#define LIN_TAPE_GET_CHANGER_NUM _IOR('z', 0x87, int)

/*****************************************************************************
 *                                                                           *
 *          ioctl for setting failover                                       *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_SET_FO _IOW('z', 0x88, int)

/*****************************************************************************
 *                                                                           *
 *          ioctl for pass through for failover/encryption, etc              *
 *                                                                           *
 *****************************************************************************/
struct device_command_t{
	/* inputs */
	unsigned char command_length; /* Length of SCSI command              */
	unsigned char scsi_cdb[16];   /* SCSI command descriptor block       */
	uint timeout_value;           /* Timeout in seconds                  */
	uint buffer_length;           /* Length of data buffer or 0          */
	unsigned char *buffer;        /* Pointer to data buffer or NULL      */

	/* outputs */
	uint number_bytes;            /* Number of bytes transfered (unused) */
	unsigned char sense_length;   /* Number of valid sense bytes         */
	unsigned char sense[255];     /* Sense data when sense length > 0    */

	/* add inputs for Linux */
	int logical_dev_num;          /* logical device number               */
	unsigned char dev_type;       /* tape drive or medium changer        */
	unsigned char data_direction; /* data-in or data-out                 */
	unsigned char serial_num[16]; /* serial number, first 16 bytes       */

	/* add outputs for Linux */
	int32_t result;               /* driver, host, message, status bytes */
	unsigned char reserved[9];
};
#define LIN_TAPE_SEND_COMMAND _IOWR('z', 0x89, struct device_command_t)

/*****************************************************************************
 *                                                                           *
 *          ioctl for process wakeup                                         *
 *                                                                           *
 *****************************************************************************/
struct device_wakeup_t {
	char dev_type;
	unsigned short prv_path;
	unsigned short new_path;
	int result;
	int result2;
	unsigned long information;
};
#define LIN_TAPE_WAKE_UP _IOW('z', 0x90, struct device_wakeup_t)

/*****************************************************************************
 *                                                                           *
 *          ioctl for ERP handling                                           *
 *                                                                           *
 *****************************************************************************/
struct device_erp_t{
	unsigned char dev_type;
	unsigned short logical_dev_num;
	unsigned char command;
	unsigned char serial_num[16]; /* serial number, first 16 bytes       */
};
#define LIN_TAPE_HANDLE_ERP _IOWR('z', 0x91, struct device_erp_t)

/*****************************************************************************
 *                                                                           *
 *          IOCTL for getting driver parameter (petro) setting               *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_GET_PETRO _IOR('z', 0x92, int)

/*****************************************************************************
 *                                                                           *
 *          ioctl for setting encryption allowable                           *
 *                                                                           *
 *****************************************************************************/
#define LIN_TAPE_SET_EA _IOW('z', 0x92, int)

/*****************************************************************************
 *                                                                           *
 *          ioctl for setting sfmp                                           *
 *                                                                           *
 *****************************************************************************/
struct device_sfmp{
	unsigned char fo_enabled;
	int id;
	unsigned char type;
};
#define LIN_TAPE_SET_SFMP _IOW('z', 0x93, struct device_sfmp)


#define STATUS_COMPLETE   1
#define STATUS_NO_RETRY   2

#ifdef __KERNEL__
struct tape_descriptor;
struct changer_descriptor;

/* functions for regular traces */
void tape_log_msg(unchar*, uint*, unchar);
void tape_log_msg2(unchar*, uint*, unchar, unchar);
void tape_log_msg3(unchar*, uint*, unchar, unchar, unchar);
void tape_log_msg4(unchar*, uint*, unchar, unchar, unchar, unchar);
void tape_log_msg5(unchar*, uint*, unchar, unchar, unchar, unchar, unchar);
void log_multibyte_msg(unchar* log, uint* log_len, int num_msgs, unchar msg1,
	unchar msg2, unchar msg3, unchar msg4, unchar msg5);

/* functions for full traces */
int tape_setup_and_log_full(struct tape_descriptor* drv, int path, int code,
	char* addtl_buf, int addtl_len);
void tape_log_full_trace(struct tape_descriptor*, unchar, unchar, unchar*);
void tape_log_buffer_data(struct tape_descriptor*, int, unchar*);
void tape_log_rc(struct tape_descriptor*, unchar);

int changer_setup_and_log_full(struct changer_descriptor* chg, int path,
	int code, char* addtl_buf, int addtl_len);
void changer_log_full_trace(struct changer_descriptor*, unchar, unchar,
	unchar*);
void changer_log_buffer_data(struct changer_descriptor*, int, unchar*);
void changer_log_rc(struct changer_descriptor*, unchar);

#endif

#define TRC_SIOC_INQUIRY                (1)
#define TRC_SIOC_INQUIRY_PG             (2)
#define TRC_SIOC_REQSENSE               (3)
#define TRC_SIOC_RESERVE                (4)
#define TRC_SIOC_RELEASE                (5)
#define TRC_SIOC_CHG_TUR                (6)
#define TRC_SIOC_LG_SNS_PG              (7)
#define TRC_SIOC_MD_SNS_PG              (8)
#define TRC_SMCIOC_READ_EL_INFO         (9)
#define TRC_SMCIOC_POS                  (10)
#define TRC_SMCIOC_IES_RANGE            (11)
#define TRC_SMCIOC_RES                  (12)
#define TRC_SMCIOC_MVMED                (13)
#define TRC_SMCIOC_IES                  (14)
#define TRC_SMCIOC_PREVENT              (15)
#define TRC_SMCIOC_ALLOW                (16)
#define TRC_SMCIOC_EXMED                (17)
#define TRC_SMCIOC_DEVIDS               (18)
#define TRC_SMCIOC_UNLOAD               (19)
#define TRC_SMCIOC_LOAD                 (20)
#define TRC_SMCIOC_CMD                  (21)
#define TRC_SMCIOC_INV                  (22)
#define TRC_SMCIOC_ERR                  (23)
#define TRC_MSG_CHG_NO_SENSE            (24)
#define TRC_MSG_CHG_SENSE               (25)
#define TRC_CHG_CLOSED                  (26)

#define TRC_CHG_DVC_OPEN                (29)
#define TRC_QRY_PATH                    (30)
#define TRC_DVC_PATHS                   (31)
#define TRC_ENABLE_IOCTL                (32)
#define TRC_DISABLE_IOCTL               (33)
#define TRC_CG_CPF_SUCCESS              (34)
#define TRC_TP_NEW_SCSI_ADDR            (35)
#define TRC_CG_NEW_SCSI_ADDR            (36)

#define TRC_LOCATE                      (40)
#define TRC_TP_DPF_SUCCESS              (41)
#define TRC_TP_ENCRYPT                  (42)

#define TRC_MSG_ERR_NO_SENSE            (50)
#define TRC_MSG_SENSE                   (51)
#define TRC_MSG_INFO                    (52)
#define TRC_MSG_EOM                     (53)
#define TRC_MSG_EOD                     (54)
#define TRC_MSG_FM                      (55)
#define TRC_TP_CLOSED                   (56)

#define TRC_WT_ENT                      (60)
#define TRC_DINQ_ENT                    (61)
#define TRC_RERR_ENT                    (62)
#define TRC_RELC_ENT                    (63)
#define TRC_LINV_ENT                    (64)
#define TRC_LMVM_ENT                    (65)
#define TRC_LAUD_ENT                    (66)
#define TRC_LERI_ENT                    (67)
#define TRC_LPOS_ENT                    (68)
#define TRC_LINF_ENT                    (69)
#define TRC_XINQ_ENT                    (70)
#define TRC_XINV_ENT                    (71)
#define TRC_LSNS_ENT                    (72)
#define TRC_EIVT_ENT                    (73)

#define TRC_TP_GHOST                    (78)
#define TRC_CG_GHOST                    (79)
#define TRC_MSG_TP_RM                   (80)
#define TRC_MSG_CG_RM                   (81)
#define TRC_TP_ATMPT_FAILOVER           (82)
#define TRC_CG_ATMPT_FAILOVER           (83)

#define TRC_TP_OPEN                     (85)
#define TRC_CG_OPEN                     (86)
#define TRC_TP_RM_HBA_CT                (87)
#define TRC_CHG_RM_HBA_CT               (88)
#define TRC_CG_RESERVED                 (89)
#define TRC_CG_RELEASED                 (90)
#define TRC_TP_RESERVED                 (91)
#define TRC_TP_RELEASED                 (92)
#define TRC_TP_RPT_POS                  (93)
#define TRC_TP_ENABLED                  (94)
#define TRC_CHG_ENABLED                 (95)
#define TRC_TP_DISABLED                 (96)
#define TRC_CHG_DISABLED                (97)
#define TRC_TP_PATH_CHECK               (98)
#define TRC_CG_PATH_CHECK               (99)
#define TRC_TP_DVC_OPEN                 (100)

#define TRC_WT_NO_CDB                   (104)

#define TRC_WT_FAIL                     (110)
#define TRC_RD_BLK_LEN                  (111)

#define TRC_RD_INV_FXD                  (113)
#define TRC_RD_ODD_FXD                  (114)
#define TRC_RD_EXC_DEF                  (115)
#define TRC_RD_INV_VAR                  (116)
#define TRC_RD_CRPT_SG                  (117)
#define TRC_RD_OVERFLOW                 (118)
#define TRC_RD_NO_CDB                   (119)

#define TRC_RD_SG_OVRRN                 (122)

#define TRC_MT_RESET                    (124)
#define TRC_MT_FSF                      (125)
#define TRC_MT_BSF                      (126)
#define TRC_MT_FSR                      (127)
#define TRC_MT_BSR                      (128)
#define TRC_MT_WEOF                     (129)
#define TRC_MT_REW                      (130)
#define TRC_MT_OFFL                     (131)
#define TRC_MT_NOP                      (132)
#define TRC_MT_RETEN                    (133)
#define TRC_MT_BSFM                     (134)
#define TRC_MT_FSFM                     (135)
#define TRC_MT_EOM                      (136)
#define TRC_MT_ERASE                    (137)
#define TRC_MT_RAS1                     (138)
#define TRC_MT_SETBLK                   (139)
#define TRC_MT_SEEK                     (140)
#define TRC_MT_TELL                     (141)
#define TRC_MT_SETDRVBUF                (142)
#define TRC_MT_LOCK                     (143)
#define TRC_MT_UNLOCK                   (144)
#define TRC_MT_LOAD                     (145)
#define TRC_MT_UNLOAD                   (146)
#define TRC_MT_COMPRESS                 (147)
#define TRC_INV_MT                      (148)
#define TRC_MTIOCGET                    (149)
#define TRC_MTIOCPOS                    (150)
#define TRC_SCSI_TUR                    (151)
#define TRC_SCSI_IOC                    (152)
#define TRC_DISPLAY_MSG                 (153)
#define TRC_STOP                        (154)
#define TRC_STOFFL                      (155)
#define TRC_STREW                       (156)
#define TRC_STERASE                     (157)
#define TRC_STRETEN                     (158)
#define TRC_STWEOF                      (159)
#define TRC_STFSF                       (160)
#define TRC_STRSF                       (161)
#define TRC_STFSR                       (162)
#define TRC_STRSR                       (163)
#define TRC_STINSRT                     (164)
#define TRC_STEJECT                     (165)
#define TRC_STTUR                       (166)
#define TRC_STLOAD                      (167)
#define TRC_STSEOD                      (168)
#define TRC_STFSSF                      (169)
#define TRC_STRSSF                      (170)
#define TRC_CHGP                        (171)
#define TRC_INQ                         (172)
#define TRC_REQSENSE                    (173)
#define TRC_RESERVE                     (174)
#define TRC_RELEASE                     (175)
#define TRC_SIOC_TUR                    (176)
#define TRC_LOG_SNS_PG                  (177)
#define TRC_MODE_SNS_PG                 (178)
#define TRC_INQ_PG                      (179)
#define TRC_SETP                        (180)
#define TRC_QRYP                        (181)
#define TRC_SYNC                        (182)
#define TRC_QRYPOS                      (183)
#define TRC_SETPOS                      (184)
#define TRC_QRYSENSE                    (185)
#define TRC_QRYINQ                      (186)
#define TRC_READ_POS                    (187)
#define TRC_RESET_DRV                   (188)
#define TRC_PRV_MEDIUM                  (189)
#define TRC_ALLOW_MEDIUM                (190)
#define TRC_RPT_DENSITY                 (191)
#define TRC_QVER_CMD                    (192)
#define TRC_PTCMD                       (193)

#define TRC_STIOC_READ_KEYS             (195)
#define TRC_STIOC_READ_RSVS             (196)
#define TRC_STIOC_CLEAR_REGS            (197)
#define TRC_MTDVC                       (198)
#define TRC_INV_IOC                     (199)
#define TRC_LOG_SNS10_PG                (200)
#define TRC_ENH_LOG_SNS                 (201)

#define TRC_TPS_NOALLOC                 (204)
#define TRC_RAO                         (205)

#define TRC_TP_NOALLOC                  (210)

#define TRC_TP_SERIAL_NUM               (212)
#define TRC_CGS_NOALLOC                 (213)
#define TRC_CG_NOALLOC                  (214)

#define TRC_CG_SERIAL_NUM               (216)

#define TRC_RM_CLASS_DVC                (218)
#define TRC_BAD_DEVS                    (219)
#define TRC_TP_ATMPT_OPEN               (220)
#define TRC_TP_OUT_RANGE                (221)
#define TRC_BAD_RW_BUF                  (222)
#define TRC_BAD_RW_CTLS                 (223)
#define TRC_BAD_SG                      (224)
#define TRC_OPEN_NO_CDB                 (225)
#define TRC_OPEN_NO_SENSE               (226)

#define TRC_TAPEX_OPENED                (228)

#define TRC_CG_ATMPT_OPEN               (230)

#define TRC_CG_NO_RESERVE               (232)
#define TRC_CHG_OPENED                  (233)

#define TRC_TBL_CORRUPT                 (235)
#define TRC_TP_ATMPT_CLOSE              (236)
#define TRC_CLOSE_BSY                   (237)
#define TRC_RELEASE_FAILED              (238)
#define TRC_RW_BUF_CORRUPT              (239)
#define TRC_SG_STUCK                    (240)
#define TRC_CG_ATMPT_CLOSE              (241)
#define TRC_FLUSH_FM                    (242)
#define TRC_FLUSH_FWD                   (243)
#define TRC_SIMMIM_NOT_SAVED            (244)

#define TRC_DMP_NOT_SAVED               (246)

#define TRC_SCSI_PATH                   (248)
#define TRC_CDB                         (249)
#define TRC_BFF                         (250)
#define TRC_STATUS                      (251)
#define TRC_SENSE                       (252)
#define TRC_RC                          (253)

#define TRC_TRUNC                       (255)
#endif
