/*******************************************************************************
 *   lin_tape_tape.h
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#ifndef _LIN_TAPE_TAPE_H_
#define _LIN_TAPE_TAPE_H_

#include <linux/kernel.h>       /* general kernel support       */
#include <linux/mm.h>           /* memory                       */
#include <linux/errno.h>        /* errno                        */

#include <linux/version.h>
#include <linux/slab.h>
#include <linux/utsname.h>      /* uname info                   */

#include <linux/mtio.h>		    /* mt commands                  */
#include "IBM_tape.h"
#include "lin_tape_srb_tape.h"
#include "lin_tape_scsi_trace.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
#include <linux/blkdev.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)

#include <asm/scatterlist.h>


/* scatterlist.h functions for use in earlier kernels */
#define sg_page(sg) ((sg) ? ((sg)->page) : NULL)

#define sg_virt(sg) ((sg) ? (page_address(sg_page(sg)) + (sg)->offset) : NULL)

#define sg_set_page(sg, pg, len, off) {\
	if(sg) {                       \
		(sg)->page = pg;       \
		(sg)->length = len;    \
		(sg)->offset = off;    \
	}                              \
}
#else

#include <linux/scatterlist.h> /* scatter/gather list          */

#endif

#define TRC_LOCKING (0)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#define max_hw_segs(sdev) sdev->request_queue->max_hw_segments
#define max_ph_segs(sdev) sdev->request_queue->max_phys_segments
#define max_seg_size(sdev) sdev->request_queue->max_segment_size
#elif defined RHEL_MAJOR && RHEL_MAJOR == 6
#define max_hw_segs(sdev) sdev->request_queue->limits.max_segments
#define max_ph_segs(sdev) sdev->request_queue->limits.max_segments
#define max_seg_size(sdev) sdev->request_queue->limits.max_segment_size
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
#define max_hw_segs(sdev) sdev->request_queue->limits.max_hw_segments
#define max_ph_segs(sdev) sdev->request_queue->limits.max_phys_segments
#define max_seg_size(sdev) sdev->request_queue->limits.max_segment_size
#else
#define max_hw_segs(sdev) queue_max_segments(sdev->request_queue)
#define max_ph_segs(sdev) queue_max_segments(sdev->request_queue)
#define max_seg_size(sdev) queue_max_segment_size(sdev->request_queue)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#define KOBJ_CT(kobj) ((kobj) ? atomic_read(&(kobj)->refcount) : 0)
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,11,0)
#define KOBJ_CT(kobj) ((kobj) ? atomic_read(&(kobj)->kref.refcount) : 0)
#else
#define KOBJ_CT(kobj) ((kobj) ? atomic_read(&(kobj)->kref.refcount.refs) : 0)
#endif
#endif

/* global definitions */
#ifndef LIN_TAPE_VERSION
/* need to change MODULE_VERSION as well */
#define LIN_TAPE_VERSION "3.0.56"
#endif

#ifndef True
#define True 1
#endif
#ifndef False
#define False 0
#endif

#define ONE_K (1024)
#ifndef BITS_PER_BYTE
#define BITS_PER_BYTE (8)
#endif

#define STATUS_SUCCESS 0

#define LIN_TAPEDebugLevel   KERN_INFO
#define Debug(x)            if(lin_tape_debug) printk(LIN_TAPEDebugLevel x)
#define Debug1(x,y)         if(lin_tape_debug) printk(LIN_TAPEDebugLevel x,y)
#define Debug2(x,y,z)       if(lin_tape_debug) printk(LIN_TAPEDebugLevel x,y,z)
#define Debug3(x,y,z,w)     if(lin_tape_debug) \
	printk(LIN_TAPEDebugLevel x,y,z,w)
#define Debug4(x,y,z,w,v)   if(lin_tape_debug) \
	printk(LIN_TAPEDebugLevel x,y,z,w,v)
#define Debug5(x,y,z,w,v,u) if(lin_tape_debug) \
	printk(LIN_TAPEDebugLevel x,y,z,w,v,u)

/* generic DbgPrint */
#define DbgPrint(x) if(lin_tape_debug > 0 && lin_tape_debug < 3) printk x;

/* special DbgPrint */
#define DbgPrint2(x) if(lin_tape_debug > 0 && lin_tape_debug <= 3) printk x;

/* special DbgPrint */
#define DbgPrint3(x) if(lin_tape_debug == 3) printk x;

/* threshhold */
#define MAX_RETRIES           3
#define MaxSerialNumberSize   30
#define KmallocThreshhold    (1<<16)
#define MaxMessageLogSize    255
#define MaxLogicName 30
#define MaxChangerFailoverAttempts 5

/* minor numbers */
#define MaxLinTapeDevices ((petro==1)?512:1024)
#define MaxLinTapeChangerDevices 1023
#define LinTapeNonRewindOffset ((petro==1)?512:1024) /* if petro is 0,
						0-1023 for rewind,
						1024-2047 for non-rewind */
#define LinTapeNonReserveOffset 1024 /* for petro support:
					0-511 rewind,
					512-1023 for non-rewind,
					1024-1535 for non-reserve;
					1536-2047 is unused*/

#define LinTapeChangerOffset 2048

#define LinTapeBaseMinor 0 /* base minor number for the driver module */
#define LinTapeMaxCount  1024*3

#define PRKEY_SIZE (8)

/* tape drive parameters */
#define AIXMode 0
#define SUNMode 1

#define ReadOnlyMode  0
#define WriteOnlyMode 1
#define ReadAndWrite  2

#define SpecialFileNamelength 16
#define RewindDevice           1
#define NonRewindDevice        0

#define LinTapeDefaultTransferLength 0x40000

/* inquiry page codes*/
#define DeviceIdentificationPage 0x83
#define UnitSerialPage           0x80
#define DriveRevisionLevelsPage  0xC0

/* reservation type */
#define LinTapeNoReserve         0
#define LinTapeReserve6          1
#define LinTapeReserve10         2
#define LinTapePersistentReserve 3

extern char tape_reserve_type[];
#define TP_RSV_PERSISTENT(drv) ((drv)->alt_pathing || \
	!memcmp(tape_reserve_type, "persistent", strlen("persistent")))

#define SpaceFileMark            1
#define SpaceRecord              2
#define SpaceEOM                 3
#define LocateBlock              4
#define LocateBlockImmediate     5
#define SpaceSequentialFM        6
#define Locate16Block            7
#define Locate16BlockImmediate   8
#define Locate16File             9
#define Locate16FileImmediate    10

#define AllModifications         0
#define BlockSizeOnly            1
#define BufferModeOnly           2
#define DensityCodeOnly          3

#define ReadAccess(X)  ( (X == O_RDONLY) || (X == O_RDWR) )
#define WriteAccess(X) ( (X == O_WRONLY) || (X == O_RDWR) )

/*
 * vender-defined SCSI Commands
 * see <scsi/scsi.h> for the other SCSI commands
 */
#define REPORT_DENSITY       0x44
#define DISPLAY_MESSAGE      0xC0
#define INIT_ELE_W_RANGE     0xE7
#define INIT_ELE_W_RANGE_3573 0x37

#define WRITE_FILEMARK_ZERO  0xFE
#define NO_COMMAND           0XFF

#define REWIND               REZERO_UNIT
#define LOAD_UNLOAD          START_STOP
#define LOCATE               SEEK_10

#ifndef POSITION_TO_ELEMENT
#define POSITION_TO_ELEMENT   0x2b
#endif

#ifndef INITIALIZE_ELEMENT_STATUS
#define INITIALIZE_ELEMENT_STATUS 0x07
#endif

#ifndef EXCHANGE_MEDIUM
#define EXCHANGE_MEDIUM       0xa6
#endif

#define InvalidEOMFlag(X) (     (X == REWIND) || \
				(X == WRITE_6) || \
				(X == WRITE_FILEMARKS) || \
				(X == WRITE_FILEMARK_ZERO) || \
				(X == SPACE) ||  \
				(X == LOAD_UNLOAD) ||  \
				(X == LOCATE))

#define InvalidEODFlag(X) (     (X == REWIND) || \
				(X == WRITE_6) ||  \
				(X == WRITE_FILEMARKS) || \
				(X == WRITE_FILEMARK_ZERO) || \
				(X == SPACE) ||  \
				(X == READ) || \
				(X == LOCATE))

#define TapeMotion(X) (         (X == REWIND) || \
				(X == READ_6)   || \
				(X == READ_REVERSE) || \
				(X == SPACE)  || \
				(X == LOCATE) || \
				(X == LOAD_UNLOAD)  || \
				(X == ERASE)  || \
				(X == WRITE_6)  || \
				(X == WRITE_FILEMARKS) )


/*
 * rw scatter-gather buffer
 */
#define LinTapeSingleRWBufferSize 1<<20 /* 1MB */
#define LinTapeMinScatterList  64
#define LinTapeMaxScatterList 256
struct tape_sgl {
	uint total_sgl;
	uint total_sg_size;
	uint use_sg;
	struct scatterlist gather[LinTapeMaxScatterList];
	struct scatterlist sgl_for_io[LinTapeMaxScatterList];
};
#define DUMP_CHUNK_SIZE	(256*ONE_K)
#define DUMP 1   /* scatter/gather for operations is used by dump */
#define DUMP_READY 11 /* successfully captured sg dump, indicator for daemon */
#define RAO  2  /* scatter/gather for RAO operations  */

#define WWN_LEN 8
#define R_HOLDER_LEN 32
#define PSEUDO_MINOR         (LinTapeMaxCount - 1)
#define TSM_MINOR            (LinTapeMaxCount - 2)
#define SPECIAL_MINOR        (LinTapeMaxCount - 3)
#define PSEUDO_TAPE(dev_idx) (dev_idx == PSEUDO_MINOR)
#define TSM_TAPE(dev_idx)    (dev_idx == TSM_MINOR)
#define SPECIAL_DEV(dev_idx)    (dev_idx == SPECIAL_MINOR)

/* tape drive descriptor */
struct tape_descriptor {
	/* internal structure */
	union cdb *cdb;
	unchar cdb_len;
	char lname[MaxLogicName]; /* logical name of device */
	char pname[MaxLogicName]; /* logical name of parent device */
	struct scsi_device *dev_obj;
	int sdev_ct;
	unsigned short dev_index;

	struct cdev *rew_dev;
	struct cdev *nonrew_dev;
	struct cdev *nonresv_dev;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	struct class_device *rew_class_dev;
	struct class_device *nonrew_class_dev;
	struct class_device *nonresv_class_dev;
#else
	struct device *rew_class_dev;
	struct device *nonrew_class_dev;
	struct device *nonresv_class_dev;
#endif

	ulong rewflags;
	ulong nonrewflags;
	ulong nonrsvflags;

	struct new_utsname uname_info;
	struct completion srb_lock;

	/* variables for scsi channel path */
	unchar host;
	unchar channel;
	unchar target;
	unchar lun;

	/* variables tracking buffers */
	unchar direct_io;
	void *buffer;
	ulong buf_len;
	struct tape_sgl *sg_list;
	int supported_sg_tablesize; /* supported max length of sg list */
	void *rw_buf;
	long rw_buf_order;
	ulong rw_buf_size;

	/* variables tracking the operation state of the device */
	unchar busy;
	unchar eom;
	ulong eom_blk_id;
	unchar eod;
	unchar drv_id;
	unchar last_op;
	unchar last_tape_motion;
	int result;
	unchar drv_open; /* currently open */
	unchar opened;   /* ever been open */
	unchar cmd_in_open;
	unchar prevent_removal;

	/* variables tracking the medium information */
	ulong blk_size;
	ulong blk_num;
	unchar buffered_mode;
	unchar capacity_scaling; /* 0, 1, 2, 3, or 4 */
	unchar capacity_scaling_hex_val;
	unchar compression;
	unchar density;
	unchar ending_fm;
	unchar logic_write_protect;
	unchar medium_type;
	unchar write_protect;
	unchar wwnn[WWN_LEN];
	unchar wwpn[WWN_LEN];
	unchar set_density_valid;
	unchar set_default_density;
	unchar set_pending_density;
	unchar data_safe_mode;
	unchar pews[2];
	unchar busy_retry;

	/* variables tracking the medium position */
	unsigned long long curr_pos;
	unchar pos_valid;

	/* variables tracking device setup parameters */
	unchar access_mode;
	unchar acf_mode;
	unchar autoload;
	unchar autoload_mode;
	unchar v2; /* True if drive is "low-cost" v2 rather than "legacy" v1 */

	ulong max_blk_size;
	ulong max_xfer_len;
	ulong min_blk_size;
	unchar read_mode;
	unchar sili;
	unchar record_space_mode;
	unchar retain_reservation;
	unchar rew_immed;
	unchar special_file_type;
	unchar trailer_labels;
	unchar wfm_immed;
	unchar limit_read_recov;
	unchar limit_write_recov;

    unchar lbp_set;
	struct logical_block_protection lbp;

	unchar amu_enable;

	/* variables for reserve/release on open/close */
	unchar curr_reserve_type;
	unchar explicit_reserved;
	unchar reservation_key[RESERVE_KEY_LENGTH];
	unsigned int pr_generation;

	/* variables used for erp */
	boolean eot_warn;
	unchar deferred_eom;
	unchar fm_detected;
	unchar ili;
	unchar resid_valid;
	int resid;
	struct tape_sense_data *sense_data;
	unchar sense_valid;
	unchar in_erp;
	unchar wfm_on_close_safe;
	unchar cmd_wo_ua_erp;
	unchar last_errno;
	unchar ghost_device;
	unchar remove_pending;

	/* variables tracking logging parameters */
	unchar devid[MAX_VOLUME_LABEL_SIZE + 1];
	unchar trace;
	unchar disable_simmim_log;
	unchar disable_auto_drive_dump;
	unchar auto_logsense;
	unchar trc_hook;
	unchar simmim_avail;
	unchar *simmim;
	uint simmim_len;
	rwlock_t simmim_lock;
	unchar full_logsense_avail;
	unchar *logsense;
	uint logsense_len;
	rwlock_t logsense_lock;
	unchar dmp_avail;
	unchar *dmp;
	ulong dmp_len;
	ulong dmp_size_order;
	rwlock_t dmp_lock;
	unchar *tape_log;
	uint tape_log_len;
	unchar doing;

	unchar op_sg;  /* operations scatter/gather indicator */
	struct tape_sgl *op_sglist;

	/* variables used for full trace */
	struct tape_full_trace_unit *full_trc_head;
	struct tape_full_trace_unit *full_trc_tail;
	rwlock_t full_trc_lock;

	/* variables used for encryption */
	unchar encryption_capable; /* is the drive capable of encrypting data */
	unchar encryption_state;   /* is the drive set to encrypt data        */
	unchar application_set_encryption; /* is the drive set to AME         */
	int sys_encryption_write;          /* should the drive encrypt data   */
	int sys_encryption_proxy; /* should the driver handle enc key request */
	int encryption_keypath_configured; /* drive configured to request key */
	int encryption_proxy_entitled; /* is drive authorized for encryption  */
	int server_test; /* EKM test to be performed, or a key request if 0   */
	int encryption_return_code;
	unchar do_encryption; /* encryption has been requested for this op    */

	/* path failover variables */
	unchar serial_num_len;
	char serial_num[MaxSerialNumberSize];
	unchar do_failover;
	unchar alt_pathing;
	unsigned short cur_path;
	int reserved_path;
	unchar isPrimary;
	unchar checked_multipath;
	struct tape_descriptor* next_path;
	unchar fenced;
	unchar enabled;
	unchar proc_read;
	wait_queue_head_t wait_queue;
	char ready_wakeup;
	int wakeup_result;
	ulong wakeup_info;
	int failover_msgs_queued;
	unchar reserve_holder[R_HOLDER_LEN];

	/* auto timeout retrieval  */
	uint timeout[256]; /* Array for command timeout values, the index is same as command opcode */
    unchar autotimeout;
};
struct descriptor_tape {
    struct tape_descriptor *drv;
    rwlock_t drv_entry_lock;
};
/* medium changer descriptor */
struct changer_descriptor {
	/* internal structure */
	union cdb *cdb;
	unchar cdb_len;

	char lname[MaxLogicName]; /* logical name of device                   */
	char pname[MaxLogicName]; /* logical name of parent device            */
	struct scsi_device *dev_obj;
	int sdev_ct;
	unsigned short dev_index;

	struct cdev *chg_dev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	struct class_device *chg_class_dev;
#else
	struct device *chg_class_dev;
#endif
	ulong chgflags;

	struct completion srb_lock;

	struct new_utsname uname_info;
	/* variables for scsi channel path */
	unchar host;
	unchar channel;
	unchar target;
	unchar lun;

	/* variables tracking buffers */
	void *buffer;
	ulong buf_len;

	/* variables tracking the operation state of the device */
	unchar busy;
	unchar chg_id;
	unchar chg_open;
	unchar open_count;
	unchar last_op;
	int result;
	unchar attempting_open;

	/* variables for reserve/release on open/close */
	unchar curr_reserve_type;
	unchar explicit_reserved;

	/* variables used for extra ioctls */
	int comp_code;
	unchar comp_code_valid;
	int last_errno;
	int barcode;

	/* variables used for erp */
	struct tape_sense_data *sense_data;
	unchar sense_valid;
	unchar resid;
	unchar resid_valid;
	unchar in_erp;
	unchar cmd_wo_ua_erp;
	unchar ghost_device;
	unchar remove_pending;

	/* variables tracking logging parameters */
	unchar trace;
	unchar *chg_log;
	uint chg_log_len;
	unchar devid[MAX_VOLUME_LABEL_SIZE + 1];
	unchar doing;

	/* variables used for full trace */
	struct tape_full_trace_unit *full_trc_head;
	struct tape_full_trace_unit *full_trc_tail;
	rwlock_t full_trc_lock;

	/* path failover variables */
	unchar serial_num_len; /* length of serial number                     */
	char serial_num[MaxSerialNumberSize]; /* serial number                */
	unchar do_failover;  /* If True, fail over                            */
	unchar alt_pathing;  /* alternate pathing active                      */
	int cur_path;        /* current path index                            */
	unchar isPrimary;    /* indicates that this device is a primary device*/
	unchar checked_multipath;
	struct changer_descriptor* next_path;
	unchar fenced;       /* indicates that this device is disabled        */
	unchar enabled;
	unchar proc_read;    /* flag indicates if /proc file has been read    */
	wait_queue_head_t wait_queue;
	char ready_wakeup;
	int wakeup_result;
	int failover_msgs_queued;
	unchar wwnn[WWN_LEN];
	unchar wwpn[WWN_LEN];

};
struct descriptor_changer{
    struct changer_descriptor *chg;
	rwlock_t entry_lock;
};

struct pfo_descriptor {
	unchar serial_num_len; /* length of serial number                     */
	char serial_num[MaxSerialNumberSize]; /* serial number                */
	unchar wwnn[WWN_LEN];
	unsigned int pfo_minor;
	struct pfo_descriptor *next;
};

#if TRC_LOCKING
#define WRITE_LOCK(x) {                                                    \
	printk("LINE %d, write locking %p\n", __LINE__, x);                    \
	write_lock(x);                                                         \
}

#define WRITE_UNLOCK(x) {                                                  \
	printk("LINE %d, write unlocking %p\n", __LINE__, x);                  \
	write_unlock(x);                                                       \
}

#define READ_LOCK(x) {                                                     \
	printk("LINE %d, read locking %p\n", __LINE__, x);                     \
	read_lock(x);                                                          \
}

#define READ_UNLOCK(x) {                                                   \
	printk("LINE %d, read unlocking %p\n", __LINE__, x);                   \
	read_unlock(x);                                                        \
}
#else
#define WRITE_LOCK write_lock
#define WRITE_UNLOCK write_unlock
#define READ_LOCK read_lock
#define READ_UNLOCK read_unlock
#endif
#endif
