/**
 * This file contains the code that
 * interfaces with the SCSI mid-layer.
 * It sends a SCSI command over a given path to a device controlled
 * by APFO code via the SCSI mid-layer.
 * This file knows nothing about fail-over devices.
 * This file handles the OS dependent SCSI command interface.
 *
 */
/*
 * (c) Copyright International Business Machines  Corp., 2020
 *
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 * (C) Copyright 2018 Hewlett Packard Enterprise Development LP
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

#include <linux/version.h>
#include <linux/mutex.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
#include <linux/smp_lock.h>
#endif
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_eh.h>
#include <linux/uio.h>
#include "apfo.h"
#include "pfo.h"

#define TO_MDATA 0
#define FROM_MDATA 1

char *lower_version_date = "20180322A";

struct the_path_struct {
	void *path_sdev;
	unsigned int host_no;
	unsigned int channel;
	unsigned int id;
	unsigned int lun;
};
static struct the_path_struct the_path[MAX_PATH_NUMBER];
static DEFINE_MUTEX(add_lock);

#if __SIZEOF_POINTER__ == 4
/* 32-bit */
#define BAD_PTR(ptr) (!((unsigned long int)ptr & (unsigned long int)(1LL<<((8*sizeof(void*))-1))))
#elif __SIZEOF_POINTER__ == 8
/* 64-bit */
#define BAD_PTR(ptr) (!((unsigned long long int)ptr & (unsigned long long int)(1LL<<((8*sizeof(void*))-1))))
#else
/* error */
#endif

/* Local function prototypes */
static int lower_scsi_execute(struct scsi_device *sdev,
				struct apfo_scsi_command *pfo_cmd,
				unsigned char *local_raw_sense);

/** Look at the errors from the SCSI command,
 * translate them to OS independent values.
 * @return
 * PFO_CMD_ERROR status was received from device, but shows not-good status,
 * PFO_PATH_DOWN status not received from device, path may be gone,
 * PFO_CMD_TIMEOUT status not received from device, command may have timed out,
 * PFO_OTHER_ERROR assume status not received from device, pass to user space.
 * PFO_FATAL_ERROR status was not received from device, pass to user space.
 */
static int translate_error(struct apfo_scsi_command *pfo_cmd, int path_num)
{
	int result;
	int host_status;
	int driver_status;

	/* Try most likely first. Did we get valid status? */
	if (likely((pfo_cmd->response_code & 0x70) == 0x70)) {
		/* Got a status response */
		return PFO_CMD_ERROR;
	}
	host_status = host_byte(pfo_cmd->errors);
	driver_status = driver_byte(pfo_cmd->errors);
	if (driver_status == DRIVER_ERROR) {
		/* Must cause an E___ error at the application level. */
		return PFO_FATAL_ERROR;
	}

	if (host_status == DID_ERROR) {
		int seconds;
		if (pfo_cmd->source == PFO_SOURCE_PFO) {
			return PFO_PATH_DOWN;
		}
		/* User space command.  May be path down, or user error. */
		msleep(2000);
		print_log("Sleeping up to 60 Seconds"
				" while checking for path loss.\n");
		for (seconds = 0; seconds < 56; seconds += 2) {
			msleep(2000);
			result = lower_path_status(path_num);
			if (result != PFO_GOOD) {
				DBG_INFO("Path down path_num=%d\n", path_num);
				msleep(2000);
				return PFO_PATH_DOWN;
			}
		}
		return PFO_OTHER_ERROR;
	}
	/* Detect command timeout. */
	if (host_status == DID_RESET ||
	    host_status == DID_ABORT ||
	    host_status == DID_REQUEUE ||
	    driver_status == DRIVER_TIMEOUT) {
		return PFO_CMD_TIMEOUT;
	}
	/* These are error codes that we suspect mean the path died. */
	if ((host_status == DID_NO_CONNECT) ||
	    (host_status == DID_TIME_OUT) ||
	    (host_status == DID_BAD_TARGET) ||
	    (host_status == DID_TRANSPORT_DISRUPTED)) {
		/* These may indicate that the path
		 * is in the process of going down.
		 * So we sleep to give time for the
		 * process to finish on every path.
		 */
		msleep(200);
		DBG_INFO("Path down path_num=%d\n", path_num);
		msleep(200);
		return PFO_PATH_DOWN;
	}
	if (host_status == DID_OK) {
		/* The HBA thinks things are fine. */
		switch (pfo_cmd->scsi_status_code) {
		case SAM_STAT_BUSY:
		case SAM_STAT_RESERVATION_CONFLICT:
		case SAM_STAT_TASK_SET_FULL:
		case SAM_STAT_ACA_ACTIVE:
		case SAM_STAT_TASK_ABORTED:
			return PFO_CMD_ERROR;
		default:
			break;
		}
	}
	DBG_WARNING("Unsupported result=0x%08X, path_num=%d\n",
			pfo_cmd->errors, path_num);
	/* Check to see if we already know that the path went away. */
	msleep(200);
	result = lower_path_status(path_num);
	if (result != PFO_GOOD) {
		DBG_INFO("Path down path_num=%d\n", path_num);
		msleep(200);
		return PFO_PATH_DOWN;
	}
	/*
	 * This is an odd thing.  May be application code bug.
	 * There was no status received from the device.
	 * Just pass this error back to user space.
	 */
	return PFO_OTHER_ERROR;
}

/** Send the given command over the given path
 * @return PFO_GOOD, PFO_PATH_DOWN, PFO_CMD_TIMEOUT,
 * PFO_CMD_ERROR, PFO_OTHER_ERROR, or PFO_FATAL_ERROR
 */
int lower_send(int path_num, struct apfo_scsi_command *pfo_cmd)
{
	int result;
	int valid_sense;
	struct scsi_device *sdev;
	struct scsi_sense_hdr local_sense_data;
	unsigned char *local_raw_sense;
	unsigned char raw_sense[SCSI_SENSE_BUFFERSIZE]={0};

	/* Check path_status before and after
	 * waiting for the path to be available.
	 */
	result = lower_path_status(path_num);
	if (unlikely(result != PFO_GOOD)) {
		DBG_INFO("lower_path_status is down\n");
		trace_event(PFO_TRACE_ATTEMPT, pfo_cmd->source, -1,
				path_num, (void *)pfo_cmd->scsi_cmd);
		return PFO_PATH_DOWN;
	}
	/* Only one command per path at a time */
	result = lock_apfo_path(path_num);
	if (unlikely(result)) {
		DBG_ERROR("lock failed returned %d\n", result);
		return PFO_FATAL_ERROR;
	}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	lock_kernel();
#endif
	sdev = the_path[path_num].path_sdev;
	/* Confirm path is still good */
	result = lower_path_status(path_num);
	if (unlikely(result != PFO_GOOD)) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		unlock_kernel();
#endif
		unlock_apfo_path(path_num);
		DBG_INFO("lower_path_status down\n");
		trace_event(PFO_TRACE_ATTEMPT, pfo_cmd->source, -1,
				path_num, (void *)pfo_cmd->scsi_cmd);
		return PFO_PATH_DOWN;
	}
	/* Reserve this path. */
	result = scsi_device_get(sdev);
	if (unlikely(result)) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		unlock_kernel();
#endif
		unlock_apfo_path(path_num);
		DBG_WARNING("scsi_device_get returned %d\n", result);
		trace_event(PFO_TRACE_ATTEMPT, pfo_cmd->source, -1,
				path_num, (void *)pfo_cmd->scsi_cmd);
		return PFO_PATH_DOWN;
	}
	/* Check for mid-layer issues */
	result = scsi_block_when_processing_errors(sdev);
	if (unlikely(!result)) {
		/* In error recovery for this device */
		scsi_device_put(sdev);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		unlock_kernel();
#endif
		unlock_apfo_path(path_num);
		DBG_WARNING("block_when_processing_errors ret=%d\n", result);
		trace_event(PFO_TRACE_ATTEMPT, pfo_cmd->source, -1,
				path_num, (void *)pfo_cmd->scsi_cmd);
		return PFO_PATH_DOWN;
	}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	unlock_kernel();
#endif

	trace_event(PFO_TRACE_COMMAND, pfo_cmd->source, -1,
			path_num, (void *)pfo_cmd->scsi_cmd);

	/* Due to a Qlogic firmware bug every command must
	 * have a minimum timeout of 10 seconds.
	 */
	if (pfo_cmd->timeout < (HZ * 10)) {
		pfo_cmd->timeout = HZ * 10;
	}
	//if (pfo_cmd->raw_sense) {
	//	local_raw_sense = pfo_cmd->raw_sense;
	//} else {
		local_raw_sense = raw_sense;
	//}
	/* Send the command */
	if (pfo_cmd->source == PFO_SOURCE_PFO) {
		/* SCSI Command from decide code */
		result = scsi_execute(sdev, pfo_cmd->scsi_cmd,
			pfo_cmd->data_direction, pfo_cmd->buffer,
			pfo_cmd->bufflen,
			raw_sense,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
			NULL,
#endif
			pfo_cmd->timeout,
			pfo_cmd->retries, pfo_cmd->flags,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
			0,
#endif
			&pfo_cmd->resid_len);
			if (pfo_cmd->raw_sense)
                memcpy(pfo_cmd->raw_sense, raw_sense, SCSI_SENSE_BUFFERSIZE);
	} else {
		/* SCSI Command from user space */
		result = lower_scsi_execute(sdev, pfo_cmd, local_raw_sense);
	}
	/* Release this path */
	scsi_device_put(sdev);
	unlock_apfo_path(path_num);

	valid_sense = scsi_normalize_sense(local_raw_sense,
				SCSI_SENSE_BUFFERSIZE, &local_sense_data);
	/* Put the SCSI command results in the structure */
	pfo_cmd->errors = result;
	pfo_cmd->scsi_status_code = result & 0xff;
	pfo_cmd->sense_key = local_sense_data.sense_key;
	pfo_cmd->asc = local_sense_data.asc;
	pfo_cmd->ascq = local_sense_data.ascq;
	pfo_cmd->response_code = local_sense_data.response_code;

	if (unlikely(result)) {
		int error_result;
		trace_event(PFO_TRACE_ERROR, pfo_cmd->source, -1,
				path_num, (void *) &result);
		/* Translate result into OS independent return value. */
		error_result = translate_error(pfo_cmd, path_num);
		if (valid_sense &&
		    (local_sense_data.sense_key ||
		     local_sense_data.asc ||
		     local_sense_data.ascq)) {
			trace_event(PFO_TRACE_SENSE, pfo_cmd->source, -1,
					path_num, (void *)&local_sense_data);
		}
		return error_result;
	}
	return PFO_GOOD;
}

/** Associate the given scsi_device structure with a new path number.
 * @return the new path number, or PFO_NOT_APFO_PATH if table is full
 */
int lower_add_new_path(void *real_sdev)
{
	int path_num;
	char string[50];
	static int next_path_num;
	struct scsi_device *sdev;

	mutex_lock(&add_lock);
	/* Assign each path a new number, as a failover tracking tool */
	for (path_num = next_path_num; path_num < MAX_PATH_NUMBER; path_num++) {
		if (the_path[path_num].path_sdev == NULL) {
			goto found_path;
		}
	}
	for (path_num = 0; path_num < MAX_PATH_NUMBER; path_num++) {
		if (the_path[path_num].path_sdev == NULL) {
			goto found_path;
		}
	}
	mutex_unlock(&add_lock);
	print_log("Unable to add a new path. Exceeded maximum count.\n");
	return PFO_NOT_APFO_PATH;

found_path:
	/* Create an entry to put into the list of FO devices */
	the_path[path_num].path_sdev = real_sdev;
	sdev = real_sdev;
	the_path[path_num].host_no = sdev->host->host_no;
	the_path[path_num].channel = sdev->channel;
	the_path[path_num].id = sdev->id;
	the_path[path_num].lun = sdev->lun;
	next_path_num = path_num + 1;
	mutex_unlock(&add_lock);
	DBG_DEBUG("set path %d to 0x%p\n", path_num, real_sdev);
	strcpy(string, "Add ");
	lower_print_path(&string[strlen(string)], path_num);
	strcat(string, "\n");
	print_log(string);
	return path_num;
}

/** Check path status.
 * @return PFO_GOOD or PFO_PATH_DOWN
 */
int lower_path_status(int path_num)
{
	struct scsi_device *sdev;

	if (path_num < 0 ||
	    path_num >= MAX_PATH_NUMBER) {
		DBG_ERROR("Bad path number %d\n", path_num);
		return PFO_PATH_DOWN;
	}

	/* Must confirm path_num is valid above,
	 * before reading the pointer from the array here. */
	sdev = the_path[path_num].path_sdev;

	if (
        !sdev ||
	    !sdev->host ||
	    !sdev->request_queue ||
	    sdev->sdev_state != SDEV_RUNNING) {
		return PFO_PATH_DOWN;
	}
	return PFO_GOOD;
}

/** Mark this path as permanently removed.
 * Path numbers are reused again after they have been removed,
 * so this routine must be called after decide_remove_path.
 */
void lower_remove_path(int path_num)
{
	char string[100];

	strcpy(string, "Remove ");
	lower_print_path(&string[strlen(string)], path_num);
	strcat(string, "\n");
	print_log(string);
	if (strchr(string, '?') != NULL) {
		DBG_WARNING("path_num=%d\n",path_num);
	}

	mutex_lock(&add_lock);
	if (the_path[path_num].path_sdev == NULL) {
		DBG_WARNING("DOUBLE removed path=%d\n", path_num);
	} else {
		DBG_DEBUG("path=%d to 0x%p\n",
			path_num, the_path[path_num].path_sdev);
	}
	the_path[path_num].path_sdev = NULL;
	mutex_unlock(&add_lock);
}

void lower_exit(void)
{
	/* No memory is allocated by this file */
	return;
}

/** Sends a user initiated command.
 * @return the error code from the SCSI mid-layer
 * or DRIVER_ERROR << 24
 */
static int lower_scsi_execute(struct scsi_device *sdev,
				struct apfo_scsi_command *pfo_cmd,
				unsigned char *local_raw_sense) {

	int err = 0;
	struct request *rq;
	struct bio *bio = NULL;
	int write = (pfo_cmd->data_direction == DMA_TO_DEVICE);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
    struct scsi_request *srq;
#endif
	rq = blk_get_request(sdev->request_queue, write, GFP_KERNEL);
	if (!rq) {
		return DRIVER_ERROR << 24;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
    srq = scsi_req(rq);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,11,0)
	rq->cmd_type = REQ_TYPE_BLOCK_PC;
#endif
	rq->cmd_flags |= pfo_cmd->flags;

	if ((pfo_cmd->bufflen > 0) && (pfo_cmd->data_direction != DMA_NONE)) {
		if (pfo_cmd->mdata != NULL) {
			/* The memory offset is defined by sg or st.
			 * We need to make sure that it retains the
			 * original value when a retry happens.
			 */
			struct rq_map_data *m_data;
			m_data = pfo_cmd->mdata;
			m_data->offset = pfo_cmd->offset;
		}
		/* Map the user memory */
		if (pfo_cmd->source == PFO_SOURCE_ST) {
			err = blk_rq_map_user(sdev->request_queue,
					rq,
					pfo_cmd->mdata,
					NULL,
					pfo_cmd->bufflen,
					GFP_ATOMIC);
		} else if (pfo_cmd->source == PFO_SOURCE_SG ||
			   pfo_cmd->source == PFO_SOURCE_IOCTL) {
			if (pfo_cmd->iov_count) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0)
				struct iovec *iov = NULL;
				struct iov_iter i;

				err = import_iovec(write, pfo_cmd->buffer, pfo_cmd->iov_count, 0, &iov, &i);
				if (err < 0)
					return err;

				iov_iter_truncate(&i, pfo_cmd->bufflen);
				err = blk_rq_map_user_iov(sdev->request_queue,
									rq,
									pfo_cmd->mdata,
									&i,
									GFP_KERNEL);
#else
				err = blk_rq_map_user_iov(sdev->request_queue,
					rq,
					pfo_cmd->mdata,
					(struct sg_iovec *)pfo_cmd->buffer,
					pfo_cmd->iov_count,
					pfo_cmd->bufflen,
					GFP_KERNEL);
#endif
			}
			else{
				err = blk_rq_map_user(sdev->request_queue,
					rq,
					pfo_cmd->mdata,
					pfo_cmd->buffer,
					pfo_cmd->bufflen,
					GFP_KERNEL);
			}
		} else if (pfo_cmd->source == PFO_SOURCE_CMD_1 ||
			   pfo_cmd->source == PFO_SOURCE_EXEC) {
			err = blk_rq_map_kern(sdev->request_queue,
					rq,
					pfo_cmd->buffer,
					pfo_cmd->bufflen,
					GFP_KERNEL);
		} else {
			DBG_WARNING("Unrecognized source value=%d\n",
				pfo_cmd->source);
		}

		if (err) {
			DBG_WARNING("memory map failed "
				"bufflen=%d iov_count=%d err=%d\n",
				pfo_cmd->bufflen, pfo_cmd->iov_count, err);
			DBG_WARNING("mdata=%p buf=%p\n",
				pfo_cmd->mdata, pfo_cmd->buffer);
			blk_put_request(rq);
			return DRIVER_ERROR << 24;
		} else {
			bio = rq->bio;
		}
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
	srq->cmd_len = pfo_cmd->cmd_len;
	memset(srq->cmd, 0, BLK_MAX_CDB);
	memcpy(srq->cmd, pfo_cmd->scsi_cmd, pfo_cmd->cmd_len);
	srq->sense = local_raw_sense;
	srq->sense_len = 0;
	srq->retries = pfo_cmd->retries;
#else
	rq->cmd_len = pfo_cmd->cmd_len;
	memset(rq->cmd, 0, BLK_MAX_CDB);
	memcpy(rq->cmd, pfo_cmd->scsi_cmd, pfo_cmd->cmd_len);
	rq->sense = local_raw_sense;
	rq->sense_len = 0;
	rq->retries = pfo_cmd->retries;
#endif
	rq->timeout = pfo_cmd->timeout;
	rq->end_io_data = pfo_cmd;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,12,0)
	blk_execute_rq(rq->q, NULL, rq, 1);
#else
	blk_execute_rq(NULL, rq, 1);
#endif

	if (bio && (pfo_cmd->source != PFO_SOURCE_CMD_1) &&
		   (pfo_cmd->source != PFO_SOURCE_EXEC)) {
		err = blk_rq_unmap_user(bio);
		if (err) {
			DBG_ERROR("unmap returns: %d\n", err);
		}
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
	pfo_cmd->resid_len = srq->resid_len;
	pfo_cmd->sense_len = srq->sense_len;
	pfo_cmd->errors = srq->result;
#else
	pfo_cmd->resid_len = rq->resid_len;
	pfo_cmd->sense_len = rq->sense_len;
	pfo_cmd->errors = rq->errors;
#endif
	blk_put_request(rq);
	return pfo_cmd->errors;
}

int lower_physical_path_number(int path_num)
{
	if (path_num >= 0 &&
	    path_num < MAX_PATH_NUMBER &&
	    the_path[path_num].path_sdev != NULL) {
		return the_path[path_num].host_no;
	}
	return -1;
}

/** Format information about this path for the user interface.
 * Fits on one line, ends with nul.
 * This will cat to the string, so you can call it
 * to place the data after anything else,
 * so it must have a nul already in the string.
 * @return PFO_GOOD or PFO_NOT_APFO_PATH
 */
int lower_print_path(char *string, int path_num)
{
	char tmp[100];

	if (path_num >= 0 &&
	    path_num < MAX_PATH_NUMBER &&
	    the_path[path_num].path_sdev != NULL) {
		sprintf (tmp, "%d:%d:%d:%d ",
			the_path[path_num].host_no,
			the_path[path_num].channel,
			the_path[path_num].id,
			the_path[path_num].lun);
		if (strlen(tmp) < 9) {
			strcat(tmp, " ");
		}
		if (strlen(tmp) < MAX_PATH_LEN) {
			strcat(string, tmp);
			return PFO_GOOD;
		}
	}
	/* Something is not kosher */
	strcat(string, "?:?:?:?  ");
	return PFO_NOT_APFO_PATH;
}

/** Copy data to/from a kernel buffer
 * from/to memory that has been mapped in an iov struct.
 */
static int copy_iov(const struct iovec *iov_struct, void *kernel_buffer,
			int iov_count, int count, int direction) {
	int index;
	int length;
	int result;
	int remain = count;
	unsigned char *iov_buffer;
	unsigned char *target_buffer = kernel_buffer;

	DBG_INFO("iov_count=%d %s\n", iov_count,
			direction == TO_MDATA ? "to iov": "from iov");
	for (index = 0; index < iov_count; index++) {
		length = iov_struct[index].iov_len;
		if (length > remain) {
			length = remain;
		}
		if (length <= 0) {
			continue;
		}
		iov_buffer = iov_struct[index].iov_base;
		if (iov_buffer == NULL) {
			DBG_ERROR("NULL pointer so copy failed\n");
			return 1;
		}
		if (direction == TO_MDATA) {
			result = copy_to_user(iov_buffer, target_buffer,
						length);
			if (result) {
				DBG_ERROR("copy to user failed\n");
				return 1;
			}
		} else {
			result = copy_from_user(target_buffer, iov_buffer,
						length);
			if (result) {
				DBG_ERROR("copy from user failed\n");
				return 1;
			}
		}
		target_buffer += length;
		remain -= length;
		if (remain == 0) {
			/* The only successful outcome. */
			return 0;
		}
		if (remain < 0 ) {
			DBG_ERROR("too many bytes moved\n");
			return 1;
		}
	}
	DBG_ERROR("not enough memory so copy to user failed\n");
	return 1;
}

/** Copy data to/from a kernel buffer
 * from/to memory that has been mapped in an mdata struct.
 */
static int copy_mdata(struct rq_map_data *mdata, void *kernel_buffer,
			int count, int direction) {
	int len;
	int index = 0;
	int remain = count;
	char *mdata_buffer;
	int offset = mdata->offset;

	DBG_INFO("count=%d %s\n", count,
			direction == TO_MDATA ? "to mdata": "from mdata");
	if (offset) {
		DBG_INFO("offset = %d\n", offset);
		len = PAGE_SIZE - offset;
		mdata_buffer = page_address(mdata->pages[index]) + offset;
		if (direction == TO_MDATA) {
			memcpy(mdata_buffer, kernel_buffer, len);
		} else {
			memcpy(kernel_buffer, mdata_buffer, len);
		}
		index++;
		kernel_buffer += len;
		remain = remain - len;
	}
	len = PAGE_SIZE;
	for (remain = remain; remain > 0; remain -= PAGE_SIZE) {
		mdata_buffer = page_address(mdata->pages[index]);
		if (remain < PAGE_SIZE) {
			len = remain;
		}
		if (direction == TO_MDATA) {
			memcpy(mdata_buffer, kernel_buffer, len);
		} else {
			memcpy(kernel_buffer, mdata_buffer, len);
		}
		index++;
		kernel_buffer += PAGE_SIZE;
	}
	return 0;
}

/** Send the given SCSI command over the given sfmp device.
 * Handles memory mapping issues.
 * @return PFO_GOOD or PFO_FATAL_ERROR or return from lin_ code.
 */
int lower_sfmp_send(int fdev_num, struct apfo_scsi_command *pfo_cmd)
{
	int result;
	int errors = 0;
	int alloced = 0;
	char *bounce = NULL;
	int valid_sense = 0;
	struct scsi_sense_hdr local_sense_data;
	int sense_length = SCSI_SENSE_BUFFERSIZE;

	trace_event(PFO_TRACE_COMMAND, pfo_cmd->source, fdev_num,
			-1, (void *)pfo_cmd->scsi_cmd);

	/* Send proper memory buffer based upon source of the command. */
	if (pfo_cmd->bufflen > 0) {
		switch (pfo_cmd->source) {
		case PFO_SOURCE_ST:
			/* st always at mdata, so copy. */
			bounce = kzalloc(pfo_cmd->bufflen, GFP_KERNEL);
			if (bounce == NULL) {
				DBG_ERROR("st bounce allocation failed\n");
				result = PFO_FATAL_ERROR;
				goto exit;
			}
			alloced = 1;
			if (pfo_cmd->data_direction == DMA_TO_DEVICE) {
				result = copy_mdata(pfo_cmd->mdata, bounce,
						pfo_cmd->bufflen, FROM_MDATA);
				if (result) {
					DBG_ERROR("copy from mdata failed\n");
					result = PFO_FATAL_ERROR;
					goto exit;
				}
			}
			break;
		case PFO_SOURCE_SG:
		case PFO_SOURCE_IOCTL:
			bounce = kzalloc(pfo_cmd->bufflen, GFP_KERNEL);
			if (bounce == NULL) {
				DBG_ERROR("sg bounce allocation failed\n");
				result = PFO_FATAL_ERROR;
				goto exit;
			}
			alloced = 1;
			if (pfo_cmd->data_direction == DMA_TO_DEVICE) {
				/* Data coming from user space, so copy. */
				if (pfo_cmd->iov_count) {
					result = copy_iov(pfo_cmd->buffer,
							bounce,
							pfo_cmd->iov_count,
							pfo_cmd->bufflen,
							FROM_MDATA);
				} else {
					result = copy_from_user(bounce,
							pfo_cmd->buffer,
							pfo_cmd->bufflen);
				}
				if (result) {
					DBG_ERROR("copy from user failed\n");
					result = PFO_FATAL_ERROR;
					goto exit;
				}
			}
			break;
		case PFO_SOURCE_CMD_1:
		case PFO_SOURCE_PFO:
		case PFO_SOURCE_EXEC:
			/* The buffer pointer is valid kernel space, so ok. */
			bounce = pfo_cmd->buffer;
			break;
		default:
			DBG_ERROR("Unrecognized source value.\n");
			result = PFO_FATAL_ERROR;
			goto exit;
		}
	} else {
		/* No data to transfer. */
		if (pfo_cmd->source == PFO_SOURCE_CMD_1) {
			/* Upper level code does not handle this right. */
			pfo_cmd->data_direction = DMA_NONE;
		}
	}

	result = intercept_sfmp_send(fdev_num,
			pfo_cmd->scsi_cmd,
			COMMAND_SIZE(pfo_cmd->scsi_cmd[0]),
			bounce,
			pfo_cmd->bufflen,
			pfo_cmd->data_direction,
			pfo_cmd->timeout,
			pfo_cmd->raw_sense,
			&sense_length,
			&pfo_cmd->resid_len,
			&errors);

	if (result == 0 && pfo_cmd->bufflen > 0) {
		if (pfo_cmd->source == PFO_SOURCE_ST) {
			/* st always at mdata, so copy. */
			if (pfo_cmd->data_direction == DMA_FROM_DEVICE) {
				result = copy_mdata(pfo_cmd->mdata, bounce,
						pfo_cmd->bufflen, TO_MDATA);
				if (result) {
					DBG_ERROR("copy to mdata failed\n");
					result = PFO_FATAL_ERROR;
					goto exit;
				}
			}
		} else if (pfo_cmd->source == PFO_SOURCE_SG ||
			   pfo_cmd->source == PFO_SOURCE_IOCTL) {
			if (pfo_cmd->data_direction == DMA_FROM_DEVICE) {
				/* Data needs sent to user space, so copy. */
				if (pfo_cmd->iov_count) {
					result = copy_iov(pfo_cmd->buffer,
							bounce,
							pfo_cmd->iov_count,
							pfo_cmd->bufflen,
							TO_MDATA);
				} else {
					result = copy_to_user(pfo_cmd->buffer,
						bounce, pfo_cmd->bufflen);
				}
				if (result) {
					DBG_ERROR("copy to user failed\n");
					result = PFO_FATAL_ERROR;
					goto exit;
				}
			}
		}
	}

	if (errors) {
		DBG_DEBUG("Raw=%02x %02x %02x %02x %02x %02x %02x %02x\n",
			pfo_cmd->raw_sense[0],
			pfo_cmd->raw_sense[1],
			pfo_cmd->raw_sense[2],
			pfo_cmd->raw_sense[12],
			pfo_cmd->raw_sense[13],
			pfo_cmd->raw_sense[15],
			pfo_cmd->raw_sense[16],
			pfo_cmd->raw_sense[17]);
	}

	valid_sense = scsi_normalize_sense(pfo_cmd->raw_sense,
				SCSI_SENSE_BUFFERSIZE, &local_sense_data);
	/* Put the SCSI command results in the structure */
	pfo_cmd->errors = errors;
	pfo_cmd->scsi_status_code = errors & 0xff;
	pfo_cmd->sense_key = local_sense_data.sense_key;
	pfo_cmd->asc = local_sense_data.asc;
	pfo_cmd->ascq = local_sense_data.ascq;
	pfo_cmd->response_code = local_sense_data.response_code;
	pfo_cmd->sense_len = sense_length;

	if (unlikely(errors)) {
		trace_event(PFO_TRACE_ERROR, pfo_cmd->source, fdev_num,
				-1, (void *) &errors);
		if (valid_sense &&
		    (local_sense_data.sense_key ||
		     local_sense_data.asc ||
		     local_sense_data.ascq)) {
			trace_event(PFO_TRACE_SENSE, pfo_cmd->source, fdev_num,
					-1, (void *)&local_sense_data);
		} else {
			DBG_WARNING("No valid sense seen.\n");
		}
	}
exit:
	if (alloced) {
		kfree(bounce);
	}
	return result;
}
