/**
 * This file contains the code that receives a SCSI command
 * that is intended to be sent to a device controlled by either
 * HP's LTO-6 APFO or IBM's failover methods.
 * All calls to the APFO decide.c file must come through this file first.
 *
 * @author Mike Yoknis
 */

/*
 * (c) Copyright 2015 Hewlett-Packard Development Company, L.P
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

char *resolve_version_date = "20151120D";

#include <linux/string.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include "resolve_ibm_hp.h"

int resolve_print_device(int fdev_num, char *string, int size);

/*###############################################################*/
/* Constants used only in this file.				 */
/*###############################################################*/

/*###############################################################*/
/* Variables used only in this file.				 */
/*###############################################################*/

int apfo_nr_dev;
struct failover_dev **apfo_dev;
int print_debug;
struct gen_path_entry *path_info[MAX_PATH_NUMBER];

/* Function prototypes */

/*###############################################################*/
/* Low level utilities for sending specific commands over a path */
/*###############################################################*/

/** Send Inquiry command over the path given.
 * @return result from lower_send
 */
static int pfo_inquiry(int path_num, int page_num, int vpd,
				char *data_buf, int buf_len)
{
#define PFO_INQUIRY_ATTEMPTS 4
	int count;
	int ret_val;
	unsigned char fo_cmd[6];
	struct apfo_scsi_command pfo_cmd;

	for (count = 0; count < PFO_INQUIRY_ATTEMPTS; count++) {
		memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
		fo_cmd[0] = INQUIRY;
		if (vpd) {
			fo_cmd[1] = 1;
		} else {
			fo_cmd[1] = 0;
		}
		fo_cmd[2] = page_num;
		fo_cmd[3] = buf_len >> 8;
		fo_cmd[4] = buf_len & 0xFF;
		fo_cmd[5] = 0;

		pfo_cmd.scsi_cmd = fo_cmd;
		pfo_cmd.data_direction = DMA_FROM_DEVICE;
		pfo_cmd.timeout = PFO_TIMEOUT_60SEC;
		pfo_cmd.buffer = data_buf;
		pfo_cmd.bufflen = buf_len;
		pfo_cmd.source = PFO_SOURCE_PFO;
		pfo_cmd.retries = PFO_COMMAND_RETRIES;
		ret_val = lower_send(path_num, &pfo_cmd);
		if (ret_val == PFO_GOOD) {
			break;
		}
		if (ret_val == PFO_PATH_DOWN) {
			DBG_INFO("Path down at inquiry attempt\n");
			sleep_half_second();
		}
		if (pfo_cmd.sense_key == 0x2 &&
			pfo_cmd.asc == 0x3E &&
			pfo_cmd.ascq == 0x0)
			sleep_six_seconds();
	}
	return ret_val;
}

static void resolve_clear_sfmp_reservations(int fdev_num)
{
#define RESERVATION_LENGTH 24
	int ret_val, result;
	int path_num;
	unsigned char fo_cmd[16];
	struct apfo_scsi_command pfo_cmd;
	unsigned char data_buf[RESERVATION_LENGTH];
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	if (cur_dev->num_paths == 0) {
		printk("No paths available for the clear function\n");
		return;
	}
	/* Find a good path */
	path_num = cur_dev->last_known_good_path;
	result=lower_path_status(path_num);
	if ( result != PFO_GOOD) {
		path_num = cur_dev->paths[0]->path_number;
		result=lower_path_status(path_num);
		if ( result != PFO_GOOD) {
			printk("A path is not available for clear.\n");
			return;
		}
	}
	/* Send the command that sets a key */
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	memset(fo_cmd, 0, 16);
	memset(data_buf, 0, RESERVATION_LENGTH);
	data_buf[8] = 0;
	data_buf[9] = 1;
	data_buf[10] = 2;
	data_buf[11] = 3;
	data_buf[12] = 4;
	data_buf[13] = 5;
	data_buf[14] = 6;
	data_buf[15] = 7;
	fo_cmd[0] = 0x5f;
	fo_cmd[1] = 6;
	fo_cmd[8] = RESERVATION_LENGTH;
	pfo_cmd.scsi_cmd = fo_cmd;
	pfo_cmd.data_direction = DMA_TO_DEVICE;
	pfo_cmd.timeout = PFO_TIMEOUT_60SEC;
	pfo_cmd.buffer = data_buf;
	pfo_cmd.bufflen = RESERVATION_LENGTH;
	pfo_cmd.source = PFO_SOURCE_PFO;
	pfo_cmd.retries = 0;
	ret_val = lower_send(path_num, &pfo_cmd);
	if (ret_val != PFO_GOOD) {
		printk("Clear Reservations - Set Key failed.\n");
		return;
	}
	/* Send the command that clears all reservations */
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	memset(fo_cmd, 0, 16);
	fo_cmd[0] = 0x5f;
	fo_cmd[1] = 3;
	fo_cmd[8] = RESERVATION_LENGTH;
	memset(data_buf, 0, RESERVATION_LENGTH);
	data_buf[0] = 0;
	data_buf[1] = 1;
	data_buf[2] = 2;
	data_buf[3] = 3;
	data_buf[4] = 4;
	data_buf[5] = 5;
	data_buf[6] = 6;
	data_buf[7] = 7;
	pfo_cmd.scsi_cmd = fo_cmd;
	pfo_cmd.data_direction = DMA_TO_DEVICE;
	pfo_cmd.timeout = PFO_TIMEOUT_60SEC;
	pfo_cmd.buffer = data_buf;
	pfo_cmd.bufflen = RESERVATION_LENGTH;
	pfo_cmd.source = PFO_SOURCE_PFO;
	pfo_cmd.retries = 0;
	ret_val = lower_send(path_num, &pfo_cmd);
	if (ret_val != PFO_GOOD) {
		printk("Clear Reservations - Clear failed.\n");
		return;
	}
}

/*###############################################################*/
/* Misc. helper functions					 */
/*###############################################################*/

/** Extract WWNN from the inquiry data.
 * @param[in] inq_buffer From Inquiry 0x83 command.
 * @param[out] wwnn The device's world wide node name.
 * @param[out] protocol SAS or FC.
 * @param[out] wwpn The device's world wide port name.
 */
static int pfo_find_wwnn(unsigned char *inq_buffer, unsigned char *wwnn,
			 int *protocol, unsigned char *wwpn)
{
	int i;
	int found = 0;
	int page_len = (inq_buffer[2] << 8) + inq_buffer[3];
	int descriptor_len;
	unsigned char letter;
	int index, j;

	if (protocol != NULL) {
		*protocol = inq_buffer[4] >> 4;
	}
	if (wwpn != NULL) {
		memset(wwpn, '\0', WWN_SIZE);
	}

	/* go through all identifiers,
	 * and find the one that contains the WWNN: */
	i = 4;
	while (i < page_len) {
		descriptor_len = inq_buffer[i+3];
		/* if the identifier type is NAA (0x03),
		 * and the association is addressed unit (00b), that
		 * is what we're after. */
		if ((inq_buffer[i+1] & 0x3F) == 0x03) {
			memcpy(wwnn, &inq_buffer[i+4], WWN_SIZE);
			found = 1;
		}
		/* for when we're interested, also pull out the port's WWN: */
		if ((wwpn != NULL) && ((inq_buffer[i+1] & 0x3F) == 0x13)) {
			memcpy(wwpn, &inq_buffer[i+4], WWN_SIZE);
		}
		i += (descriptor_len + 4);
	}

	if (!found) {
		/* Did not find a NAA WWNN;
		 * see if we can find an EUI-64 identifier: */
		i = 4;
		while (i < page_len) {
			descriptor_len = inq_buffer[i+3];
			/* if the identifier type is EUI-64 (0x02),
			 * and the association is addressed unit (00b), that
			 * is what we're after. */
			if (((inq_buffer[i+1] & 0x0F) == 0x02) &&
			    ((inq_buffer[i+1] & 0x30) == 0x00)) {
				memcpy(wwnn, &inq_buffer[i+4], WWN_SIZE);
				found = 1;
				break;
			}
			i += (descriptor_len + 4);
		}
	}

	if (!found) {
		/* Did not find a NAA WWNN; see if we can find serial number: */
		i = 4;
		while (i < page_len) {
			descriptor_len = inq_buffer[i+3];
			/* The identifier type (0x01) is T10 vendor ID. */
			if ((inq_buffer[i+1] & 0x0F) == 0x01) {
				found = 1;
				for (index = 0; index < WWN_SIZE; index++)
					wwnn[index] = 0;
				/* pack the serial number into
				 * the 8 bytes we have for wwnn */
				index = 15;
				for (j = descriptor_len + 3; j >= 0; j--) {
					letter = inq_buffer[i+j];
					if (letter >= '0') {
						found = 1;
						letter = letter & 0xf;
						letter = index&1 ? letter : (letter<<4);
						wwnn[index>>1] = letter | wwnn[index>>1];
						index--;
						if (index < 0) {
							break;
						}
					}
				}
				break;
			}
			i += (descriptor_len + 4);
		}
	}

	return found;
}

/** Populate the path information structure.
 */
static int populate_path_struct(int path_num, int fdev_num,
				unsigned char *wwpn,
				int active_mc_path, int speed)
{
	struct gen_path_entry *tmp_entry;

	if (path_info[path_num] == NULL) {
		tmp_entry = kzalloc(sizeof(struct gen_path_entry), GFP_ATOMIC);
		if (tmp_entry == NULL) {
			return -ENOMEM;
		}
		path_info[path_num] = tmp_entry;
	}
	path_info[path_num]->path_number = path_num;
	path_info[path_num]->fdev_number = fdev_num;
	path_info[path_num]->is_active_mc_path = active_mc_path;
	path_info[path_num]->link_speed = speed;
	memcpy(path_info[path_num]->wwpn, wwpn, WWN_SIZE);
	return 0;
}

/** Get the device serial number with an inquiry command.
 */
static int get_serial_number(int path_num, unsigned char *data_buf,
				unsigned char *serial)
{
	int i;
	int size;
	int start;
	int result;

	result = pfo_inquiry(path_num, 0x80, 1, data_buf, INQ_LEN);
	if (result == PFO_GOOD) {
		start = 4;
		size = data_buf[3];
		if (data_buf[3] > (SN_SIZE - 1)) {
			size = SN_SIZE - 1;
			start = 4 + data_buf[3] - size;
			for (i = 4 + data_buf[3] -1; i > 5; i--) {
				if (data_buf[i] == '\0' ||
				    data_buf[i] == ' ' ) {
					start--;
				} else {
					break;
				}
			}
		}
		memcpy(serial, &data_buf[start], size);
		serial[size] = '\0';
	}
	return result;
}

/* malloc and default initialize the memory for a new device.
 * Needs to have the structure locked before entry.
 */
static int resolve_create_new_device(void)
{
	int error = 0;
	struct failover_dev **tmp_list;
	struct failover_dev *tmp_entry;

	DBG_INFO("Enter\n");
	if (apfo_nr_dev > MAX_FDEV_NUMBER) {
		/* Attempts to use this device will fail,
		 * but this is better than panic.
		 */
		print_log("Device not usable. Too many pfo devices.\n");
		error = PFO_NOT_APFO_DEVICE;
		goto done;
	}

	/* Expand the list with more memory for the new entry */
	tmp_list = kzalloc(
		(apfo_nr_dev+1) * sizeof(struct failover_dev *),
		GFP_ATOMIC);
	if (tmp_list == NULL) {
		error = -ENOMEM;
		goto done;
	}
	if (apfo_dev != NULL) {
		memcpy(tmp_list, apfo_dev,
			apfo_nr_dev * sizeof(struct failover_dev *));
	}
	apfo_dev = tmp_list;
	/* Create an entry to put into the list of FO devices */
	tmp_entry = kzalloc(sizeof(struct failover_dev), GFP_ATOMIC);
	if (tmp_entry == NULL) {
		error = -ENOMEM;
		goto done;
	}
	apfo_dev[apfo_nr_dev] = tmp_entry;

	apfo_dev[apfo_nr_dev]->fo_dev_number = apfo_nr_dev;
	apfo_dev[apfo_nr_dev]->path_with_session = PFO_NO_PATH_SELECTED;
	apfo_dev[apfo_nr_dev]->fsc_sequence_num = 0;
	apfo_dev[apfo_nr_dev]->pref_path_str[0] = '\0';
	apfo_dev[apfo_nr_dev]->ua_count = -1;
	apfo_dev[apfo_nr_dev]->previous_path_num = PFO_NO_PATH_SELECTED;
	apfo_dev[apfo_nr_dev]->num_paths = 1;
	apfo_dev[apfo_nr_dev]->cached_smc_data = 0;
	apfo_dev[apfo_nr_dev]->fo_session_is_open = 0;
	apfo_dev[apfo_nr_dev]->open_count = 0;
	apfo_dev[apfo_nr_dev]->cmd_retry = 0;
	apfo_dev[apfo_nr_dev]->fo_enabled = 1;
	apfo_dev[apfo_nr_dev]->open_fail_sent = 0;
	apfo_dev[apfo_nr_dev]->ncn_just_executed = 0;
	apfo_dev[apfo_nr_dev]->paths_lost_on_fo_command_retry = 0;
	apfo_dev[apfo_nr_dev]->marker1 = 0xefbeadde;
	apfo_dev[apfo_nr_dev]->marker2 = 0xefbeadde;
	apfo_dev[apfo_nr_dev]->marker3 = 0xefbeadde;
	apfo_dev[apfo_nr_dev]->marker4 = 0xefbeadde;

	apfo_dev[apfo_nr_dev]->paths =
		kzalloc(sizeof(struct gen_path_entry *), GFP_ATOMIC);
	if (apfo_dev[apfo_nr_dev]->paths == NULL) {
		kfree(apfo_dev[apfo_nr_dev]);
		error = -ENOMEM;
		goto done;
	}
done:
	return error;
}

/** Gets the WWNN of the device, and builds up the list of unique WWNs.
 * @param[out] fdev_num The failover device number,
 * iff return is not PFO_NOT_APFO_DEVICE
 * @param[out] wwnn The device's world wide node name,
 * iff return is PFO_NEW_APFO_DEVICE or PFO_OLD_APFO_DEVICE.
 * @return PFO_NOT_APFO_DEVICE, PFO_OLD_APFO_DEVICE, or PFO_NEW_APFO_DEVICE
 * or negative number indicating fatal error
 * or PFO_PATH_DOWN if communication failed
 */
static int pfo_probe_device(int path_num, int *fdev_num,
				unsigned char *wwnn, unsigned char *serial)
{
	int i;
	int result;
	int device_type;
	int ret_val = 0;
	int error = 0;
	char found = 0;
	int active_mc_path = 0;
	int failover_disabled = 1;
	int sfmp_device = 0;
	unsigned char vendor_id[10];
	unsigned char product_id[18];
	unsigned char wwpn[WWN_SIZE];
	int protocol_type;
	int speed = -1;
	struct failover_dev *cur_dev;
	unsigned char data_buf[INQ_LEN];

	*fdev_num = -1;
	/* First, get the basic Inquiry data; this will determine if
	 * it's a device type that supports failover.
	 */
	result = pfo_inquiry(path_num, 0x00, 0, data_buf, INQ_LEN);
	if (result == PFO_PATH_DOWN || result == PFO_CMD_TIMEOUT) {
		DBG_ERROR("Inquiry 0x00 failed, bad path=%d\n", path_num);
		return PFO_PATH_DOWN;
	}
	if (result != PFO_GOOD) {
		DBG_ERROR("Inquiry failed, path=%d\n", path_num);
		return PFO_NOT_APFO_DEVICE;
	}
	device_type = data_buf[0] & 0x1F;
	memcpy(vendor_id, &data_buf[8], 8);
	vendor_id[8] = 0;
	memcpy(product_id, &data_buf[16], 16);
	product_id[16] = 0;

	/* Get serial number. */
	result = get_serial_number(path_num, data_buf, serial);
	if (result == PFO_PATH_DOWN || result == PFO_CMD_TIMEOUT) {
		DBG_ERROR("Inquiry 0x80 failed, bad path=%d\n", path_num);
		DBG_INFO("device %s %s\n", vendor_id, product_id);
		return PFO_PATH_DOWN;
	}

	if ((device_type != TAPE_DEVICE) && (device_type != MEDIA_CHANGER)) {
		DBG_INFO("failover not supported for device type %d\n",
				device_type);
		return PFO_NOT_APFO_DEVICE;
	}

#ifdef APFO
	result = pfo_inquiry(path_num, 0xC8, 1, data_buf, INQ_LEN);
	if (result == PFO_PATH_DOWN || result == PFO_CMD_TIMEOUT) {
		DBG_ERROR("Inquiry 0xC8 failed, bad path=%d\n", path_num);
		DBG_INFO("device %s %s\n", vendor_id, product_id);
		return PFO_PATH_DOWN;
	}
	if (result != PFO_GOOD ||
	    data_buf[1] != 0xC8 || (data_buf[4] & 0x04) == 0) {
		/* Does not support APFO as is. */
	}

	if (decide_failover_is_enabled(path_num, device_type, &active_mc_path)) {
		DBG_INFO("device %s %s has failover enabled\n",
				vendor_id, product_id);
		failover_disabled = 0;
	}
#else
	/* We are not yet sure whether lin_tape will support
	 * this device, so assume that it will, for now.
	 */
	sfmp_device = 1;
	failover_disabled = 0;
#endif

	result = pfo_inquiry(path_num, 0x83, 1, data_buf, INQ_LEN);
	if (result == PFO_PATH_DOWN || result == PFO_CMD_TIMEOUT) {
		DBG_ERROR("Inquiry 0x83 failed, bad path=%d\n", path_num);
		DBG_INFO("device %s %s\n", vendor_id, product_id);
		return PFO_PATH_DOWN;
	}
	if (result != PFO_GOOD) {
		DBG_ERROR("device %s %s Inquiry command failed\n",
				vendor_id, product_id);
		return PFO_NOT_APFO_DEVICE;
	}
	if (!pfo_find_wwnn(data_buf, wwnn, &protocol_type, wwpn)) {
		DBG_WARNING("device %s %s "
				"unable to find WWNN so "
				"cannot support fail-over.\n",
				vendor_id, product_id);
		return PFO_NOT_APFO_DEVICE;
	}
	result = lock_apfo_device(STRUCT_MUTEX);
	if (result) {
		return -EINTR;
	}
	for (i = 0; i < apfo_nr_dev; i++) {
		if (memcmp(apfo_dev[i]->wwnn, wwnn, WWN_SIZE) == 0) {
			/* Found a new path to an existing device;
			 * add to list of paths: */
			struct gen_path_entry **tmp_list;
			void *ptr;
			char string[100];

			cur_dev = apfo_dev[i];
			if (!cur_dev->sfmp_device) {
				DBG_INFO("sfmp already rejected.\n");
				ret_val = PFO_NOT_APFO_DEVICE;
				goto done;
			}
			if (!cur_dev->fo_enabled) {
				DBG_INFO("sfmp already said not fo\n");
				sfmp_device = 1;
				failover_disabled = 1;
				found = 0; /* Create new device. */
				break;
			}
			found = 1;
			if (failover_disabled) {
				/*
				 * Remember that we once
				 * saw it as disabled so
				 * we can report it as a
				 * new device when it is
				 * seen again, enabled.
				 */
				cur_dev->fo_enabled = 0;
				ret_val = PFO_NOT_APFO_DEVICE;
				goto done;
			}
			*fdev_num = i;
			tmp_list = kzalloc((cur_dev->num_paths+1) *
				sizeof(struct gen_path_entry *), GFP_ATOMIC);
			if (tmp_list == NULL) {
				error = -ENOMEM;
				goto done;
			}
			memcpy(tmp_list, cur_dev->paths,
			  cur_dev->num_paths * sizeof(struct gen_path_entry *));
			ptr = cur_dev->paths;
			cur_dev->paths = tmp_list;
			kfree(ptr);
			error = populate_path_struct(path_num, i, wwpn,
							active_mc_path, speed);
			if (error) {
				goto done;
			}
			cur_dev->paths[cur_dev->num_paths] = path_info[path_num];
			cur_dev->num_paths++;
			DBG_INFO("New failover path fdev=%d path=%d\n",
				*fdev_num, path_num);
			strcpy(string,"See ");
			lower_print_path(&string[strlen(string)], path_num);
			sprintf(&string[strlen(string)],
			    "old failover %s device wwnn="
			    "%02x%02x%02x%02x%02x%02x%02x%02x",
			    (device_type == MEDIA_CHANGER) ? "changer" : "tape",
			    wwnn[0], wwnn[1], wwnn[2], wwnn[3],
			    wwnn[4], wwnn[5], wwnn[6], wwnn[7]);
			if (device_type == TAPE_DEVICE) {
				sprintf(&string[strlen(string)],
				   " wwpn=%02x%02x%02x%02x%02x%02x%02x%02x",
				   wwpn[0], wwpn[1], wwpn[2], wwpn[3],
				   wwpn[4], wwpn[5], wwpn[6], wwpn[7]);
			}
			strcat(string,"\n");
			print_log(string);
			break;
		}
	}

	/* Create a whole new device even if we see that the path
	 * is to the same equipment because we already know that
	 * it will not be a failover candidate.
	 */
	if (!found) {
		/* Found a new device: */
		char string[100];

		DBG_INFO("New failover device path=%d\n", path_num);
		strcpy(string,"See ");
		lower_print_path(&string[strlen(string)], path_num);
		sprintf(&string[strlen(string)],
			"new failover %s device wwnn="
			"%02x%02x%02x%02x%02x%02x%02x%02x",
			(device_type == MEDIA_CHANGER) ? "changer" : "tape",
			wwnn[0], wwnn[1], wwnn[2], wwnn[3],
			wwnn[4], wwnn[5], wwnn[6], wwnn[7]);
		if (device_type == TAPE_DEVICE) {
			sprintf(&string[strlen(string)],
				" wwpn=%02x%02x%02x%02x%02x%02x%02x%02x",
				wwpn[0], wwpn[1], wwpn[2], wwpn[3],
				wwpn[4], wwpn[5], wwpn[6], wwpn[7]);
		}
		strcat(string,"\n");
		print_log(string);
		/* malloc and default initialize the memory */
		error = resolve_create_new_device();
		if (error) {
			goto done;
		}
		apfo_dev[apfo_nr_dev]->last_known_good_path = path_num;
		apfo_dev[apfo_nr_dev]->sfmp_device = sfmp_device;
		memcpy(apfo_dev[apfo_nr_dev]->wwnn, wwnn, WWN_SIZE);
		apfo_dev[apfo_nr_dev]->protocol_type = protocol_type;
		apfo_dev[apfo_nr_dev]->dev_type = device_type;
		apfo_dev[apfo_nr_dev]->fo_enabled = !failover_disabled;
		error = populate_path_struct(path_num, apfo_nr_dev, wwpn,
						active_mc_path, speed);
		if (error) {
			kfree(apfo_dev[apfo_nr_dev]->paths);
			kfree(apfo_dev[apfo_nr_dev]);
			goto done;
		}
		apfo_dev[apfo_nr_dev]->paths[0] = path_info[path_num];
		*fdev_num = apfo_nr_dev;
		apfo_nr_dev++;
	}

	cur_dev = apfo_dev[*fdev_num];
	if (found && cur_dev->fo_enabled) {
		/*
		 * If we found it already in our list, and
		 * we have not previously seen it as having apfo disabled
		 * then we report it as being an old device.
		 */
		cur_dev->fo_enabled = 1;
		ret_val = PFO_OLD_APFO_DEVICE;
	} else {
		/*
		 * If we could not find it already in our list, or
		 * we previously saw it as having apfo disabled
		 * then we report it as being a new device.
		 * intercept.c wants to see a disabled->enabled
		 * transition as a new device.
		 */
		ret_val = PFO_NEW_APFO_DEVICE;
	}

done:
	unlock_apfo_device(STRUCT_MUTEX);

	if (error != 0) {
		return error;
	} else {
		return ret_val;
	}
}

/** This can be called under 2 condidions -
 * 1) lin_tape does not support it, so should use std /dev/sg.
 * 2) lin_tape has already said it supports it,
 *    but failover disabled at the device.
 *
 * Split out the paths into separate devices.
 * This must be done before a /dev file is created
 * so the open() call will go to the right place.
 */
void resolve_not_fo_device(void *pdev, int fdev_num)
{
	int path_count;
	int fdev_num_new;
	struct failover_dev *cur_dev;

	if (fdev_num < 0) {
		DBG_ERROR("BAD fdev=%d\n", fdev_num);
		return;
	}
	cur_dev = apfo_dev[fdev_num];

	DBG_INFO("Enter fdev=%d\n", fdev_num);
	DBG_DEBUG("num paths=%d\n", cur_dev->num_paths);
	lock_apfo_device(STRUCT_MUTEX);
	/* The first path already has the cur dev struct,
	 * so create new devices for any additional paths.
	 */
	for (path_count = 1; path_count < cur_dev->num_paths; path_count++) {
		int error;
		int path_num;
		/* Create the structure. */
		error = resolve_create_new_device();
		if (error) {
			print_log("Unable to create device structure.\n");
			break;
		}
		path_num = cur_dev->paths[path_count]->path_number;
		/* Fill the structure. */
		apfo_dev[apfo_nr_dev]->last_known_good_path = path_num;
		memcpy(apfo_dev[apfo_nr_dev]->wwnn, cur_dev->wwnn, WWN_SIZE);
		apfo_dev[apfo_nr_dev]->protocol_type = cur_dev->protocol_type;
		apfo_dev[apfo_nr_dev]->dev_type = cur_dev->dev_type;
		apfo_dev[apfo_nr_dev]->sfmp_device = cur_dev->sfmp_device;
		apfo_dev[apfo_nr_dev]->fo_enabled = 0;
		/* The path struct for this path is already created
		 * all we need to do is point it to the new device.
		 */
		apfo_dev[apfo_nr_dev]->paths[0] = path_info[path_num];
		path_info[path_num]->fdev_number = apfo_nr_dev;
		DBG_INFO("Created new device fdev=%d\n", apfo_nr_dev);
		fdev_num_new = apfo_nr_dev;
		apfo_nr_dev++;
		/* Now that the device is ready to be used tell the
		 * intercept code about it so it can create /dev files.
		 */
		if (intercept_not_fo_device(pdev, fdev_num_new, path_num)) {
			print_log("Unable to create /dev file.\n");
			break;
		}
	}
	/* All new devices are in place, now adjust the initial one. */
	cur_dev->last_known_good_path = cur_dev->paths[0]->path_number;
	cur_dev->num_paths = 1;
	cur_dev->fo_enabled = 0;
	cur_dev->pref_path_str[0] = '\0';
	unlock_apfo_device(STRUCT_MUTEX);
}

/** This is not an IBM drive.
 * This must be determined before a /dev file is created
 * so the open() call goes to the right place.
 */
void resolve_set_is_not_sfmp_device(int fdev_num)
{
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	DBG_INFO("fdev=%d\n", fdev_num);
	cur_dev->sfmp_device = 0;
}

/*###############################################################*/
/* Functions that interface to other files in the module.	 */
/*###############################################################*/
/** An open() call has been made to the given device.
 * @return PFO_GOOD if there is a functional path to the device, or
 * PFO_PATH_DOWN if there is any problem opening this device.
 */
int resolve_open(int fdev_num, struct file *filp)
{
	int ret_val = PFO_PATH_DOWN;
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	if (cur_dev->sfmp_device) {
		lock_apfo_device(fdev_num);
		if (cur_dev->open_count++) {
			/* Already open. */
			DBG_INFO("not first open\n");
			unlock_apfo_device(fdev_num);
			return PFO_GOOD;
		}
		if (intercept_sfmp_open(fdev_num, filp)) {
			/* Failed */
			cur_dev->open_count = 0;
			ret_val = PFO_PATH_DOWN;
		} else {
			/* Passed */
			cur_dev->open_count = 1;
			ret_val = PFO_GOOD;
		}
		unlock_apfo_device(fdev_num);
	}
#ifdef APFO
	else {
		ret_val = decide_open(fdev_num);
	}
#endif
	return ret_val;
}

/** A close() call has been made to the given device.
 */
int resolve_close(int fdev_num, struct file *filp)
{
	int ret_val = PFO_PATH_DOWN;
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	if (cur_dev->sfmp_device) {
		lock_apfo_device(fdev_num);
		if (--cur_dev->open_count) {
			/* Leave open. */
			DBG_INFO("not last close\n");
			unlock_apfo_device(fdev_num);
			return PFO_GOOD;
		}
		if (intercept_sfmp_close(fdev_num, filp)) {
			ret_val = PFO_PATH_DOWN;
		} else {
			ret_val = PFO_GOOD;
		}
		unlock_apfo_device(fdev_num);
	}
#ifdef APFO
	else {
		ret_val = decide_close(fdev_num);
	}
#endif
	return ret_val;
}

/** Determines whether this path is to an APFO device or not.
 * If it is, then it decides whether it is to a known device or a new one.
 * It updates all data structures in this file accordingly.
 * It is the caller's responsibility
 * to remove the path from lower level if it is not APFO.
 * It is the caller's responsibility
 * to inform upper of this device if it is a new APFO device.
 * @param[in] path_num The new path number to add.
 * @param[out] fdev_num The new failover device number,
 * iff return is PFO_NEW_APFO_DEVICE.
 * @param[out] wwnn The device's world wide node name,
 * iff return is PFO_NEW_APFO_DEVICE or PFO_OLD_APFO_DEVICE.
 * @param[out] serial The device's serial number in ASCII,
 * iff return is PFO_NEW_APFO_DEVICE or PFO_OLD_APFO_DEVICE.
 * @return -ENOMEM, PFO_NOT_APFO_DEVICE, PFO_OLD_APFO_DEVICE, or
 * PFO_NEW_APFO_DEVICE or PFO_PATH_DOWN if communication failed
 */
int resolve_add_new_path(int path_num, int *fdev_num,
			unsigned char *wwnn, unsigned char *serial)
{
	int ret_val;

	wwnn[0] = wwnn[1] = wwnn[2] = wwnn[3] = 0xff;
	wwnn[4] = wwnn[5] = wwnn[6] = wwnn[7] = 0xff;
	serial[0]='\0';

	ret_val = pfo_probe_device(path_num, fdev_num, wwnn, serial);
	if (ret_val == PFO_NEW_APFO_DEVICE) {
		trace_event(PFO_TRACE_DEVICE_NEW, -1,
				*fdev_num, path_num, NULL);
	}
	if (ret_val == PFO_OLD_APFO_DEVICE) {
		trace_event(PFO_TRACE_PATH_NEW, -1,
				*fdev_num, path_num, NULL);
	}
	DBG_INFO("wwnn=%02x%02x%02x%02x%02x%02x%02x%02x\n",
			wwnn[0], wwnn[1], wwnn[2], wwnn[3],
			wwnn[4], wwnn[5], wwnn[6], wwnn[7]);
	DBG_INFO("serial=%s\n", serial);
	return ret_val;
}

/** Mark this path as permanently removed.
 * This must be called before lower_remove_path() */
int resolve_remove_path(int path_num)
{
	int index;
	int fdev_num;
	char string[100];
	struct failover_dev *cur_dev;

	if (path_num < 0) {
		DBG_WARNING("negative path=%d\n", path_num);
		return PFO_GOOD;
	}
	if (path_num >= MAX_PATH_NUMBER) {
		DBG_WARNING("too big path=%d\n", path_num);
		return PFO_GOOD;
	}
	/* Do not allow another task to change the data structures
	 * while we are adjusting the data structures. */
	lock_apfo_device(STRUCT_MUTEX);
	if (path_info[path_num] == NULL) {
		unlock_apfo_device(STRUCT_MUTEX);
		DBG_WARNING("null ptr path=%d\n", path_num);
		return PFO_GOOD;
	}
	fdev_num = path_info[path_num]->fdev_number;
	if (fdev_num < 0) {
		unlock_apfo_device(STRUCT_MUTEX);
		DBG_WARNING("fdev=%d\n", fdev_num);
		return PFO_GOOD;
	}
	cur_dev = apfo_dev[fdev_num];
	if (cur_dev->num_paths < 1) {
		unlock_apfo_device(STRUCT_MUTEX);
		DBG_WARNING("num_paths=%d\n", cur_dev->num_paths);
		return PFO_GOOD;
	}
	trace_event(PFO_TRACE_PATH_BROKEN, -1, fdev_num, path_num, NULL);
	/* Path numbers are re-used,
	 * so we must remove every reference to bad numbers. */
	if (path_num == cur_dev->path_with_session) {
		cur_dev->path_with_session = PFO_NO_PATH_SELECTED;
	}
	if (path_num == cur_dev->last_known_good_path) {
		cur_dev->last_known_good_path = PFO_NO_PATH_SELECTED;
	}
	/* Find which item to remove from this device's list of paths */
	if (path_num != cur_dev->paths[cur_dev->num_paths-1]->path_number) {
		for (index = 0; index < cur_dev->num_paths; index++) {
			if (path_num == cur_dev->paths[index]->path_number) {
				cur_dev->paths[index] =
					cur_dev->paths[cur_dev->num_paths-1];
				break;
			}
		}
		/* There is a short window here where a good path
		 * will show up twice on the device's list. */
	}
	/* Reduce the count last, so all potential
	 * paths are still available during remove.
	 */
	cur_dev->num_paths--;
	unlock_apfo_device(STRUCT_MUTEX);
	DBG_INFO("Removing path failover device fdev=%d path_num=%d\n",
			fdev_num, path_num);
	strcpy(string,"Detach ");
	lower_print_path(&string[strlen(string)], path_num);
	sprintf(&string[strlen(string)], "failover %s device wwnn="
		"%02x%02x%02x%02x%02x%02x%02x%02x",
		(cur_dev->dev_type == MEDIA_CHANGER) ? "changer" : "tape",
		cur_dev->wwnn[0],
		cur_dev->wwnn[1],
		cur_dev->wwnn[2],
		cur_dev->wwnn[3],
		cur_dev->wwnn[4],
		cur_dev->wwnn[5],
		cur_dev->wwnn[6],
		cur_dev->wwnn[7]);
	if (cur_dev->dev_type == TAPE_DEVICE) {
		sprintf(&string[strlen(string)],
			" wwpn=%02x%02x%02x%02x%02x%02x%02x%02x",
			path_info[path_num]->wwpn[0],
			path_info[path_num]->wwpn[1],
			path_info[path_num]->wwpn[2],
			path_info[path_num]->wwpn[3],
			path_info[path_num]->wwpn[4],
			path_info[path_num]->wwpn[5],
			path_info[path_num]->wwpn[6],
			path_info[path_num]->wwpn[7]);
	}
	strcat(string, "\n");
	print_log(string);
	return PFO_GOOD;
}

/** Creates the text string that tells a user
 * about the status of control actions for the given device.
 * It applies to the whole driver.
 * @return string length
 */
int resolve_ctrl_get(char *out_string, size_t size)
{
#ifdef APFO
	return decide_ctrl_get(out_string, size);
#else
	return 0;
#endif
}

/** Interprets a text string that indicates a user
 * invoked control action for the given device
 * via /sys/devices/virtual/pfo/pfo?/ctrl
 * fdev==-1 means it applies to the whole driver
 * via /sys/bus/scsi/drivers/pfo/ctrl
 * @return PFO_GOOD or PFO_EINVAL
 */
int resolve_ctrl(int fdev_num, const char *in_string, size_t size)
{
#define MAX_SIZE 32
	int fdev;
	char *buffer;
	char string[MAX_SIZE];
	char err_string[2*MAX_SIZE];
	struct failover_dev *cur_dev;

	/* This is a user interface, so check for bad entries */
	if (size >= MAX_SIZE) {
		size = MAX_SIZE-1;
	}
	strncpy(string, in_string, size);
	string[size] = '\0';
	if (string[size-1] == '\n') {
		string[size-1] = '\0';
	}

	if (fdev_num == -1) {
		/* Driver wide command */
		if (strncmp(string, "dump", 4) == 0) {
			buffer = kzalloc(4096, GFP_KERNEL);
			if (buffer == NULL) {
				return PFO_GOOD;
			}
			for (fdev = 0; fdev < apfo_nr_dev; fdev++) {
				buffer[0] = '\0';
				print_debug = 1;
				resolve_print_device(fdev, buffer, 4095);
				print_debug = 0;
				sleep_half_second();
				print_log(buffer);
			}
			return PFO_GOOD;
		}
		goto fail_out;
	}

	if (fdev_num < 0 || fdev_num >= apfo_nr_dev) {
		DBG_ERROR("Device number %d out of range.\n", fdev_num);
		return PFO_EINVAL;
	}

	/* Specific device related command */
	cur_dev = apfo_dev[fdev_num];

	if (strncmp(string, "clear", 5) == 0) {
		resolve_clear_sfmp_reservations(fdev_num);
		return PFO_GOOD;
	}
	if (strncmp(string, "fdev", 4) == 0) {
		int size = (unsigned long)(&cur_dev->marker4) -
				(unsigned long)cur_dev + 8;
		print_hex_dump(KERN_ERR, "fdev: ", DUMP_PREFIX_NONE,
			16, 1, cur_dev, size, true);
		return PFO_GOOD;
	}
	if (strncmp(string, "dump", 4) == 0) {
		buffer = kzalloc(4096, GFP_KERNEL);
		if (buffer == NULL) {
			return PFO_GOOD;
		}
		buffer[0] = '\0';
		print_debug = 1;
		resolve_print_device(fdev_num, buffer, 4095);
		print_debug = 0;
		sleep_half_second();
		print_log(buffer);
		return PFO_GOOD;
	}
fail_out:
#ifdef APFO
	return decide_ctrl(fdev_num, in_string, size);
#endif
	strcpy(err_string, "Unrecognized ctrl string=");
	strcat(err_string, string);
	strcat(err_string, "\n");
	print_log(err_string);
	return PFO_EINVAL;
}

/** Creates a text string that holds all info for the given device.
 * It is terminated with cr and nul.
 * This will cat to the string, so you can call it
 * to place the data after anything else, so it must
 * have a nul already in the string.
 */
int resolve_print_device(int fdev_num, char *string, int size)
{
	int len, result;
	int index;
	char tmp[100];
	struct failover_dev *cur_dev;

	if (fdev_num < 0 || fdev_num >= apfo_nr_dev) {
		strcat(string, "Not failover capable\n");
		return PFO_GOOD;
	}

	cur_dev = apfo_dev[fdev_num];
	len = strlen(string);
	/* Single line description of the device */
	len += snprintf(string+len, size-len,
			"wwnn=%02x%02x%02x%02x%02x%02x%02x%02x type=",
			cur_dev->wwnn[0],
			cur_dev->wwnn[1],
			cur_dev->wwnn[2],
			cur_dev->wwnn[3],
			cur_dev->wwnn[4],
			cur_dev->wwnn[5],
			cur_dev->wwnn[6],
			cur_dev->wwnn[7]);
	if (cur_dev->dev_type == MEDIA_CHANGER) {
		len += snprintf(string+len, size-len, "changer");
	}
	if (cur_dev->dev_type == TAPE_DEVICE) {
		len += snprintf(string+len, size-len, "tape");
	}
	if (cur_dev->fo_session_is_open) {
		len += snprintf(string+len, size-len,
			" session_key=%02x%02x%02x%02x",
			cur_dev->session_key[0],
			cur_dev->session_key[1],
			cur_dev->session_key[2],
			cur_dev->session_key[3]);
	}
	len += snprintf(string+len, size-len, "\n");

	if (print_debug) {
		/* The failover device's number */
		len += snprintf(string+len, size-len,
			"fdev=%d ", cur_dev->fo_dev_number);
		/* The path on which the session has been established */
		len += snprintf(string+len, size-len,
			"sesspath=%d ", cur_dev->path_with_session);
		/* The two bit number passed in APFO scsi commands */
		len += snprintf(string+len, size-len,
			"fsc=%d ", cur_dev->fsc_sequence_num);
		/* The path number of the last successful command */
		len += snprintf(string+len, size-len,
			"lastgood=%d ", cur_dev->last_known_good_path);
		/* The preferred path to the device */
		len += snprintf(string+len, size-len,
			"pref=%s ", cur_dev->pref_path_str);
		/* Unit Attention count */
		len += snprintf(string+len, size-len,
			"ua=%d ", cur_dev->ua_count);
		/* How many paths to this physical device */
		len += snprintf(string+len, size-len,
			"num=%d ", cur_dev->num_paths);
		/* Command byte of the last user space command sent */
		len += snprintf(string+len, size-len,
			"cmd=%02x ", cur_dev->cmnd_byte);
		/* Timeout of the last user space command sent */
		len += snprintf(string+len, size-len,
			"to=%d ", cur_dev->cmnd_timeout);
		/* Counter of how many open /dev files using this session */
		len += snprintf(string+len, size-len,
			"session=%d ", cur_dev->fo_session_is_open);
		/* Counter of how many open /dev files using this session */
		len += snprintf(string+len, size-len,
			"opencnt=%d ", cur_dev->open_count);
		/* 0= When there is no command being processed on this device */
		len += snprintf(string+len, size-len,
			"retry=%d ", cur_dev->cmd_retry);
		/* =1 if Nexus Change Notification was the last thing we did */
		len += snprintf(string+len, size-len,
			"ncn=%d ", cur_dev->ncn_just_executed);
		/* Has SMC data been colected? */
		len += snprintf(string+len, size-len,
			"smc=%d ", cur_dev->cached_smc_data);
		/* 0 = Fibre, 6 = SAS */
		/* 1 = Tape, 8 = Library */
		len += snprintf(string+len, size-len, "\n");
	}

	if (cur_dev->num_paths == 0) {
		len += snprintf(string+len, size-len, "No paths\n");
	}
	/* Single line description for each path. */
	for (index = 0; index < cur_dev->num_paths; index++) {
		int path_num;
		path_num = cur_dev->paths[index]->path_number;
		tmp[0] = '\0';
		lower_print_path(tmp, path_num);
		len += snprintf(string+len, size-len, tmp);
		result=lower_path_status(path_num);
		if ( result == PFO_GOOD) {
			len += snprintf(string+len, size-len, "up   ");
		} else {
			len += snprintf(string+len, size-len, "down ");
		}
		if (path_num != cur_dev->last_known_good_path) {
			len += snprintf(string+len, size-len, "-    ");
		} else {
			if (cur_dev->open_count) {
				len += snprintf(string+len, size-len,
								"open ");
			} else {
				len += snprintf(string+len, size-len,
								"last ");
			}
		}
		/* Place holder for preferred path in APFO */
		len += snprintf(string+len, size-len, "-      ");
		if (cur_dev->dev_type == MEDIA_CHANGER) {
			if (cur_dev->paths[index]->is_active_mc_path) {
				len += snprintf(string+len, size-len,
								"active_drive");
			} else {
				len += snprintf(string+len, size-len, "-");
			}
		} else {
			len += snprintf(string+len, size-len,
				"wwpn=%02x%02x%02x%02x%02x%02x%02x%02x",
				cur_dev->paths[index]->wwpn[0],
				cur_dev->paths[index]->wwpn[1],
				cur_dev->paths[index]->wwpn[2],
				cur_dev->paths[index]->wwpn[3],
				cur_dev->paths[index]->wwpn[4],
				cur_dev->paths[index]->wwpn[5],
				cur_dev->paths[index]->wwpn[6],
				cur_dev->paths[index]->wwpn[7]);
		}
		if (print_debug) {
			len += snprintf(string+len, size-len,
				" path=%d", path_num);
			len += snprintf(string+len, size-len,
				" fdev=%d", cur_dev->paths[index]->fdev_number);
		}
		len += snprintf(string+len, size-len, "\n");
		if (len > size-100) {
			/*
			 * We use snprintf to protect from buffer overflow.
			 * But, the Linux kernel does not accept this as
			 * a valid method.  If it sees a negative value
			 * it spits a stack dump warning to the console.
			 * This is not acceptable, so we still have to
			 * break out of the loop before we hit the warning.
			 */
			if (len < size-6) {
				strcat(string, "...\n");
			}
			break;
		}
	}
	if (len >= size-6) {
		snprintf(string+size-6, 5, "...\n");
	}
	return PFO_GOOD;
}

void resolve_exit(void)
{
#ifdef APFO
	decide_exit();
#else
	int fdev_num;
	struct failover_dev *cur_dev;

	/* Free memory. */
	for (fdev_num = 0; fdev_num < apfo_nr_dev; fdev_num++) {
		DBG_INFO("Free fdev=%d\n", fdev_num);
		cur_dev = apfo_dev[fdev_num];
		kfree(cur_dev);
	}
	apfo_nr_dev = 0;
	/* Chosen to not free structs for all previously allocated paths. */
#endif
}

/** Get a copy of the wwnn directly from the device via the path specified.
 * An Inquiry command is executed on the path to obtain the wwnn.
 * @return PFO_GOOD if there is now a wwnn in the buffer supplied or PFO_ERROR
 */
int resolve_get_wwnn_from_path(int path_num, unsigned char wwnn[])
{
	int result;
	int protocol_type;
	unsigned char wwpn[WWN_SIZE];
	unsigned char data_buf[INQ_LEN];

	result = pfo_inquiry(path_num, 0x83, 1, data_buf, INQ_LEN);
	if (result != PFO_GOOD) {
		DBG_INFO("Inquiry command failed. Path=%d\n", path_num);
		return PFO_ERROR;
	}
	if (!pfo_find_wwnn(data_buf, wwnn, &protocol_type, wwpn)) {
		DBG_INFO("Unable to find WWNN. Path=%d\n", path_num);
		return PFO_ERROR;
	}
	return PFO_GOOD;
}

/*###############################################################*/
/* Functions that interface to other modules.			 */
/*###############################################################*/
/** Is this an IBM drive?
 * @return 0 if this is not an IBM drive.
 */
int pfo_is_sfmp_device(int fdev_num)
{
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	return cur_dev->sfmp_device;
}
EXPORT_SYMBOL(pfo_is_sfmp_device);

void resolve_set_last_path(int fdev_num, int path_num)
{
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	DBG_INFO("fdev=%d path=%d\n", fdev_num, path_num);
	cur_dev->last_known_good_path = path_num;
}

/** Remove a device that does not have failover enabled.
 * Clean up any related data structures used in this file.
 */
void resolve_remove_non_fo_device(int fdev_num)
{
	int path_num;
	struct failover_dev *cur_dev;

	DBG_INFO("Enter fdev=%d\n", fdev_num);
	if (fdev_num < 0) {
		return;
	}
	cur_dev = apfo_dev[fdev_num];

	/* And remove the associated path.
	 * Non-failover devices should only have one path.
	 */
	path_num = cur_dev->paths[0]->path_number;
	if (path_num >= 0) {
		resolve_remove_path(path_num);
		lower_remove_path(path_num);
	}
	/* There is no provision for taking a device out,
	 * so we simply make it unavailable.
	 */
	cur_dev->wwnn[0] = 0xba;
	cur_dev->wwnn[1] = 0xdf;
	cur_dev->wwnn[2] = 0xad;
	cur_dev->wwnn[3] = 0x1c;
	cur_dev->wwnn[4] = 0xed;
	cur_dev->wwnn[5] = 0xc0;
	cur_dev->wwnn[6] = 0xff;
	cur_dev->wwnn[7] = 0xee;
}

/*###############################################################*/
/* Process a command to be sent to the failover device.		 */
/*###############################################################*/
/** Send the given user space command to the given failover device
 * @return PFO_GOOD or PFO_FATAL_ERROR
 */
int resolve_send(int fdev_num, struct apfo_scsi_command *pfo_cmd)
{
	int ret_val = PFO_FATAL_ERROR;
	struct failover_dev *cur_dev = apfo_dev[fdev_num];

	if (cur_dev->sfmp_device) {
		/* Send this command to the sfmp driver. */
		unsigned char users_raw_sense[SCSI_SENSE_BUFFERSIZE];

		if (pfo_cmd->raw_sense == NULL) {
			/* If caller does not provide a buffer */
			memset(users_raw_sense, 0, SCSI_SENSE_BUFFERSIZE);
			pfo_cmd->raw_sense = users_raw_sense;
		}
		lock_apfo_device(fdev_num);
		ret_val = lower_sfmp_send(fdev_num, pfo_cmd);
		if (ret_val) {
			DBG_INFO("sfmp ret_val=%d\n", ret_val);
			ret_val = PFO_FATAL_ERROR;
		}
		unlock_apfo_device(fdev_num);
	}
#ifdef APFO
	else {
		/* Send this command to the APFO driver. */
		ret_val = decide_send(fdev_num, pfo_cmd);
	}
#endif
	return ret_val;
}
