/**
 * This file contains the code that
 * records events.  It holds them in a buffer
 * that a user can read.  It also sends them
 * to the console.
 *
 * November 25, 2014 - Do not print duplicate reads or writes.
 *
 */
/*
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/mutex.h>
#include <linux/jiffies.h>
#include <linux/mm.h>
#include <linux/unistd.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>

#include "apfo.h"

DEFINE_MUTEX(trace_lock);

static unsigned char *trace_buffer;

static void print_event(unsigned char type, unsigned char source, int fdev_num,
			int path_num, void *data,
			int the_pid, long int the_time, int count) {
	int i;
	int strlenbuf;
	char tmp[100];
	char buf[100];
	static long int start;
	static int index;
	unsigned char *cmd = (unsigned char *)data;
	unsigned int *error = (unsigned int *)data;
	struct scsi_sense_hdr *sense_data = (struct scsi_sense_hdr *) data;

	buf[0] = '\0';

	if (start == 0)  {
		trace_buffer = kzalloc(PFO_TRACE_BUF_SIZE, GFP_ATOMIC);
		if (trace_buffer == NULL) {
			/* just try again next time */
			return;
		}
		start = jiffies;
	}

	sprintf(tmp, "%6ld ", (the_time-start) & 0xffffff);
	strcat(buf, tmp);
	sprintf(tmp, "%5d ", the_pid);
	strcat(buf, tmp);
	if (fdev_num != -1) {
		sprintf(tmp, "fdev=%d ", fdev_num);
		strcat(buf, tmp);
	}
	switch (source) {
	case PFO_SOURCE_SG:
		strcat(buf, "g:");
		break;
	case PFO_SOURCE_ST:
		strcat(buf, "t:");
		break;
	case PFO_SOURCE_PFO:
		strcat(buf, "p:");
		break;
	case PFO_SOURCE_IOCTL:
		strcat(buf, "i:");
		break;
	case PFO_SOURCE_CMD_1:
		strcat(buf, "1:");
		break;
	case PFO_SOURCE_EXEC:
		strcat(buf, "e:");
		break;
	default:
		break;
	}
	switch (type) {
	case PFO_TRACE_EMPTY:
		break;
	case PFO_TRACE_ERROR:
		strcat(buf, "ERROR=");
		sprintf(tmp, "0x%02X 0x%02X 0x%02X",
			(*error >> 24) & 0xffff,
			(*error >> 16) & 0xff,
			(*error) & 0xffff);
		strcat(buf, tmp);
		break;
	case PFO_TRACE_ATTEMPT:
	case PFO_TRACE_COMMAND:
		if (cmd != NULL) {
			sprintf(tmp, "_%02x%02x%02x%02x%02x%02x",
					cmd[0], cmd[1], cmd[2],
					cmd[3], cmd[4], cmd[5]);
			for (i = 6; i < COMMAND_SIZE(cmd[0]); i++) {
				sprintf(tmp + strlen(tmp), "%02x", cmd[i]);
			}
			strcat(buf, tmp);
		}
		if (type == PFO_TRACE_ATTEMPT)
			strcat(buf, " Attempt");
		break;
	case PFO_TRACE_UNACCEPTABLE:
		strcat(buf, "Unacceptable");
		break;
	case PFO_TRACE_PATH_BROKEN:
		strcat(buf, "Remove");
		break;
	case PFO_TRACE_FATAL:
		strcat(buf, "Fatal");
		break;
	case PFO_TRACE_BREAK_TIMEOUT:
		strcat(buf, "Timeout");
		break;
	case PFO_TRACE_DEVICE_NEW:
		strcat(buf, "New Device");
		break;
	case PFO_TRACE_PATH_NEW:
		strcat(buf, "Added ");
		break;
	case PFO_TRACE_SENSE:
		sprintf(tmp, "Key_%02x %02x/%02x ",
				sense_data->sense_key,
				sense_data->asc,
				sense_data->ascq);
		strcat(buf, tmp);
		break;
	case PFO_TRACE_PATH_BAD:
		strcat(buf, "Bad");
		break;
	}
	if (path_num != -1) {
		sprintf(tmp, " path=%d", path_num);
		strcat(buf, tmp);
	}
	if (count > 1) {
		sprintf(tmp, " times %d", count);
		strcat(buf, tmp);
	}
	strcat(buf, "\n");
	/* Only log if not production (or error) verbosity */
	if (pfo_debug_flag & ~0x1111) {
		printk(KERN_NOTICE "PFO:%s", buf);
	}
	strlenbuf = strlen(buf);
	if (index + strlenbuf + 3 > PFO_TRACE_BUF_SIZE) {
		/* wrap around bottom */
		*trace_buffer = '\0';
		index = 0;
	}
	strcat(trace_buffer, buf);
	index += strlenbuf;
}

int trace_get(char *buf)
{
	int ignore;

	ignore = mutex_lock_interruptible(&trace_lock);
	if (trace_buffer != NULL) {
		memcpy(buf, trace_buffer, 4096);
	}
	buf[4094] = '\0';
	mutex_unlock(&trace_lock);
	return 0;
}

void trace_exit(void)
{
	int ignore;

	ignore = mutex_lock_interruptible(&trace_lock);
	if (trace_buffer != NULL) {
		kfree(trace_buffer);
	}
	trace_buffer = NULL;
	mutex_unlock(&trace_lock);
	return;
}

/** Skip duplicates.
 * The trace of reads and writes are time delayed,
 * but still have the correct time stamps.
 * Print the first one in the set, then keep count of all beyond it.
 * Print the FSC of the last command in the set.
 */
static int unique_event(unsigned char type, unsigned char source, int fdev_num,
			int path_num, void *data, int the_pid)
{
	unsigned char this_command;
	static int the_count = 0;
	static long int last_time;
	static unsigned char last_cdb[6];
	static unsigned char last_source;
	static int last_the_pid;
	static int last_fdev_num;
	static int last_path_num;

	/* Eliminate the obvious */
	if (data == NULL || type != PFO_TRACE_COMMAND) {
		goto not_duplicate;
	}
	this_command = *(unsigned char *)data;
	if (this_command != 0x08 && this_command != 0x0a) {
		goto not_duplicate;
	}
	/* We now have a read or write.
	 * Can we confirm a match?
	 */
	if (memcmp(data, last_cdb, 5) == 0 &&
	    source   == last_source &&
	    fdev_num == last_fdev_num &&
	    path_num == last_path_num &&
	    the_pid  == last_the_pid) {
		/* We have an exact match, so just count. */
		memcpy(last_cdb, data, 6);
		the_count++;
		last_time = jiffies;
		return 0;
	}
	/* Not a match. */
	if (the_count > 1) {
		/* Print the remembered commands,
		 * count=1 indicates we already printed the first one.
		 */
		print_event(PFO_TRACE_COMMAND, last_source,
			last_fdev_num, last_path_num, last_cdb,
			last_the_pid, last_time, the_count - 1);
		the_count = 0;
	}
	/* Remember this new command */
  	memcpy(last_cdb, data, 6);
   	last_source = source;
   	last_fdev_num = fdev_num;
    	last_path_num = path_num;
    	last_the_pid = the_pid;
	the_count = 1;
	last_time = jiffies;
	/* Always print the first one,
	 * to mark the start time of the set, and
	 * in case it is the only one.
	 */
	return 1;
	
not_duplicate:
	if (the_count > 1) {
		/* Print the remembered commands,
		 * count=1 indicates we already printed the first one.
		 */
		print_event(PFO_TRACE_COMMAND, last_source,
			last_fdev_num, last_path_num, last_cdb,
			last_the_pid, last_time, the_count - 1);
		the_count = 0;
	}
	last_source = 255;
	last_fdev_num = 255;
	last_path_num = 255;
	last_the_pid = 255;
	return 1;
}

void trace_event(unsigned char type, unsigned char source, int fdev_num,
			int path_num, void *data) {
	int ignore;
	int the_pid;

	the_pid = task_tgid_vnr(current);
	ignore = mutex_lock_interruptible(&trace_lock);
	if (unique_event(type, source, fdev_num, path_num, data, the_pid)) {
		print_event(type, source, fdev_num, path_num, data, the_pid,
				jiffies, 1);
	}
	mutex_unlock(&trace_lock);
}
