/**
 * This file contains the code that
 * interfaces with the sgmp and stmp drivers.
 * It determines whether a SCSI command should be controlled
 * by APFO code, or whether it should just be sent directly
 * to the SCSI mid-layer.
 * It knows nothing about the concept of a path.
 *
 */
/*
 * (c) Copyright International Business Machines  Corp., 2020
 *
 * (c) Copyright 2014, 2015 Hewlett-Packard Development Company, L.P
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/mm.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
#include <linux/slab.h>
#else
#include <linux/smp_lock.h>
#endif
#include <linux/tty.h>
#include <linux/uio.h>

#include <scsi/scsi_host.h>
#include <scsi/scsi_driver.h>
#include <scsi/scsi_ioctl.h>
#include <scsi/sg.h>

#include "scsi.h"
#include "pfo.h"
#include "apfo.h"

char *upper_version_date = "20150722D";

/* Constants used only in this file. */
#define FDEV_LIST_EMPTY (-1)

/* Static variables */
static int highest_fdev_num = FDEV_LIST_EMPTY;
static void *apfo_devices[MAX_FDEV_NUMBER+1];
static DEFINE_MUTEX(add_lock);

/* Local functions */
static int local_sg_scsi_ioctl(int fdev_num, fmode_t mode,
		struct scsi_ioctl_command __user *sic);

/*###############################################################*/
/* Functions that are only called from within this file.         */
/*###############################################################*/

/** Insert request and wait for the result.
 * buffer must be kernel memory, if bufflen>0
 */
static int pfo_scsi_execute(int fdev_num, unsigned char *cmd,
		 int data_direction, void *buffer, unsigned bufflen,
		 unsigned char *sense, int timeout, int retries, int flags,
		 int *resid)
{
	int retval;
	struct apfo_scsi_command pfo_cmd;

	DBG_INFO("enter fdev %d\n", fdev_num);
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = cmd;
	pfo_cmd.cmd_len = COMMAND_SIZE(cmd[0]);
	pfo_cmd.buffer = buffer;
	pfo_cmd.bufflen = bufflen;
	pfo_cmd.source = PFO_SOURCE_EXEC;
	pfo_cmd.data_direction = data_direction;
	pfo_cmd.timeout = timeout;
	pfo_cmd.retries = retries;
	pfo_cmd.iov_count = 0;
#ifdef REQ_QUIET
	pfo_cmd.flags = flags | REQ_QUIET;
#else
	pfo_cmd.flags = flags | RQF_QUIET;
#endif


	pfo_cmd.mdata = NULL;
	pfo_cmd.raw_sense = sense;
	pfo_cmd.sense_len = 0;
	pfo_cmd.response_code = 0;
	/* Send the command */
	retval = resolve_send(fdev_num, &pfo_cmd);
	if (retval != PFO_GOOD) {
		return DID_BAD_TARGET << 16;
	}
	if (resid) {
		*resid = pfo_cmd.resid_len;
	}
	return pfo_cmd.errors;
}

/** Same as scsi_execute_req
 * buffer must be kernel memory, if bufflen>0
 */
static int pfo_scsi_execute_req(int fdev_num, unsigned char *cmd,
			int data_direction, void *buffer, unsigned bufflen,
			struct scsi_sense_hdr *sshdr, int timeout, int retries,
			int *resid)
{
	char *sense = NULL;
	int result;

	if (sshdr) {
		sense = kzalloc(SCSI_SENSE_BUFFERSIZE, GFP_NOIO);
		if (!sense)
			return DRIVER_ERROR << 24;
	}
	result = pfo_scsi_execute(fdev_num, cmd, data_direction, buffer,
			bufflen, sense, timeout, retries, 0, resid);
	if (sshdr)
		scsi_normalize_sense(sense, SCSI_SENSE_BUFFERSIZE, sshdr);

	kfree(sense);
	return result;
}

/** Same as ioctl_internal_command
 */
static int pfo_internal_command(struct scsi_device *sdev, int fdev_num,
					char *cmd, int timeout, int retries)
{
	int result;
	struct scsi_sense_hdr sshdr;


	result = pfo_scsi_execute_req(fdev_num, cmd, DMA_NONE, NULL, 0,
				  &sshdr, timeout, retries, NULL);


	if ((driver_byte(result) & DRIVER_SENSE) &&
	    (scsi_sense_valid(&sshdr))) {
		switch (sshdr.sense_key) {
		case ILLEGAL_REQUEST:
			if (cmd[0] == ALLOW_MEDIUM_REMOVAL)
				sdev->lockable = 0;
			else
				DBG_WARNING("pfo_internal_command: "
				       "ILLEGAL REQUEST asc=0x%x ascq=0x%x\n",
				       sshdr.asc, sshdr.ascq);
			break;
		case NOT_READY:	/* This happens if there is no disc in drive */
			if (sdev->removable)
				break;
			DBG_WARNING("pfo_internal_command return code = %x\n",
				    result);
			break;
		case UNIT_ATTENTION:
			if (sdev->removable) {
				sdev->changed = 1;
				result = 0; /* No longer considered an error */
				break;
			}
			DBG_WARNING("pfo_internal_command return code = %x\n",
				    result);
			break;
		default:
			DBG_WARNING("pfo_internal_command return code = %x\n",
				    result);
			break;
		}
	}

	return result;
}

/** Same as scsi_set_medium_removal
 */
static int pfo_medium_removal(struct scsi_device *sdev,
				int fdev_num, char state)
{
	char scsi_cmd[MAX_COMMAND_SIZE];
	int ret;

	if (!sdev->removable || !sdev->lockable)
	       return 0;

	scsi_cmd[0] = ALLOW_MEDIUM_REMOVAL;
	scsi_cmd[1] = 0;
	scsi_cmd[2] = 0;
	scsi_cmd[3] = 0;
	scsi_cmd[4] = state;
	scsi_cmd[5] = 0;

	ret = pfo_internal_command(sdev, fdev_num, scsi_cmd,
			10 * HZ, 5);
	if (ret == 0)
		sdev->locked = (state == SCSI_REMOVAL_PREVENT);
	return ret;
}

/**
 * scsi_ioctl - Dispatch ioctl to scsi device
 * @sdev: scsi device receiving ioctl
 * @cmd: which ioctl is it
 * @arg: data associated with ioctl
 *
 * Description: The scsi_ioctl() function differs from most ioctls in that it
 * does not take a major/minor number as the dev field.  Rather, it takes
 * a pointer to a &struct scsi_device.
 */
int pfo_scsi_ioctl(struct scsi_device *sdev, int ioctl_cmd, void __user *arg)
{
	char tmp[50];
	int fdev_num = pfo_device_number(sdev);

	/* No idea how this happens.... */
	if (!sdev)
		return -ENXIO;

	if (fdev_num != NOT_PFO_DEVICE) {
		/* There are 3 categories of ioctl commands -
		 * We know we should return a failure (e.g. -ENOTTY).
		 * We know it is acceptable to pass to the Linux kernel.
		 * We must support with our own code.
		 */
		switch (ioctl_cmd) {
		/* We know we can pass these down to Linux kernel */
		case SCSI_IOCTL_GET_IDLUN:
		case SCSI_IOCTL_GET_BUS_NUMBER:
		case SCSI_IOCTL_PROBE_HOST:
		case SCSI_IOCTL_GET_PCI:
		case SG_GET_TRANSFORM:
			DBG_INFO("Sent to kernel cmd=0x%04X\n", ioctl_cmd);
			return scsi_ioctl(sdev, ioctl_cmd, arg);
		case SCSI_IOCTL_SEND_COMMAND:
			return local_sg_scsi_ioctl(fdev_num, 0, arg);
		case SCSI_IOCTL_DOORLOCK:
			return pfo_medium_removal(sdev, fdev_num,
						SCSI_REMOVAL_PREVENT);
		case SCSI_IOCTL_DOORUNLOCK:
			return pfo_medium_removal(sdev, fdev_num,
						SCSI_REMOVAL_ALLOW);
		/* We know that we do not support */
		case SCSI_IOCTL_TEST_UNIT_READY:
		case SCSI_IOCTL_START_UNIT:
		case SCSI_IOCTL_STOP_UNIT:
		default:
			break;
		}
	} else {
		return scsi_ioctl(sdev, ioctl_cmd, arg);
	}
	/* Fail all unrecognized commands */
	DBG_INFO("");
	sprintf(tmp, "scsi ioctl cmd=0x%04X not supported by APFO.\n",
				ioctl_cmd);
	print_log(tmp);
	return -ENOSYS;
}
EXPORT_SYMBOL(pfo_scsi_ioctl);

/** Same as blk_fill_sghdr_rq
 */
static int pfo_blk_fill_sghdr_rq(struct request_queue *q, struct apfo_scsi_command *pfo_cmd,
				struct sg_io_hdr *hdr, fmode_t mode)
{
	if (copy_from_user(pfo_cmd->scsi_cmd, hdr->cmdp, hdr->cmd_len))
		return -EFAULT;

#if LINUX_VERSION_CODE > KERNEL_VERSION(3,12,0)
	if (blk_verify_command(pfo_cmd->scsi_cmd, mode & FMODE_WRITE))
#else
	/* RHEL 6.4 and beyond */
	if (blk_verify_command(q, pfo_cmd->scsi_cmd, mode & FMODE_WRITE))
#endif
//6.3 	if (blk_verify_command(pfo_cmd->scsi_cmd, mode & FMODE_WRITE))

		return -EPERM;

	/*
	 * fill in request structure
	 */

	pfo_cmd->timeout = msecs_to_jiffies(hdr->timeout);
	if (!pfo_cmd->timeout)
		pfo_cmd->timeout = BLK_DEFAULT_SG_TIMEOUT;
	if (pfo_cmd->timeout < BLK_MIN_SG_TIMEOUT)
		pfo_cmd->timeout = BLK_MIN_SG_TIMEOUT;

	return 0;
}

/** Same as blk_complete_sghdr_rq
 */
static int pfo_blk_complete_sghdr_rq(struct apfo_scsi_command *pfo_cmd,
				struct sg_io_hdr *hdr)
{
	int ret = 0;

	/*
	 * fill in all the output members
	 */
	hdr->status = pfo_cmd->errors & 0xff;
	hdr->masked_status = status_byte(pfo_cmd->errors);
	hdr->msg_status = msg_byte(pfo_cmd->errors);
	hdr->host_status = host_byte(pfo_cmd->errors);
	hdr->driver_status = driver_byte(pfo_cmd->errors);
	hdr->info = 0;
	if (hdr->masked_status || hdr->host_status || hdr->driver_status)
		hdr->info |= SG_INFO_CHECK;
	hdr->resid = pfo_cmd->resid_len;
	hdr->sb_len_wr = 0;

	if (pfo_cmd->sense_len && hdr->sbp) {
		int len = pfo_cmd->sense_len;
		if (len > hdr->mx_sb_len)
			len = hdr->mx_sb_len;

		if (!copy_to_user(hdr->sbp, pfo_cmd->raw_sense, len))
			hdr->sb_len_wr = len;
		else
			ret = -EFAULT;
	}


	return ret;
}

/** Same as sg_io
 */
static int pfo_sg_io(struct request_queue *q, int fdev_num,
		struct sg_io_hdr *hdr, fmode_t mode)
{
	unsigned long start_time;
	int ret = 0;
	char sense[SCSI_SENSE_BUFFERSIZE];

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
	struct sg_iovec *sg_iov = NULL;
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	struct apfo_scsi_command pfo_cmd;

	DBG_INFO("enter fdev %d\n", fdev_num);
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = scsi_cmd;
	pfo_cmd.cmd_len = hdr->cmd_len;
	pfo_cmd.buffer = NULL;
	pfo_cmd.bufflen = 0;
#ifdef REQ_QUIET
	pfo_cmd.flags = REQ_QUIET;
#else
	pfo_cmd.flags = RQF_QUIET;
#endif
	pfo_cmd.source = PFO_SOURCE_IOCTL;
	pfo_cmd.iov_count = 0;
	pfo_cmd.mdata = NULL;
	pfo_cmd.data_direction = DMA_NONE;

	if (hdr->interface_id != 'S')
		return -EINVAL;
	if (hdr->cmd_len > BLK_MAX_CDB)
		return -EINVAL;


	if (hdr->dxfer_len) {
		switch (hdr->dxfer_direction) {
		default:
			return -EINVAL;
		case SG_DXFER_TO_DEV:
			pfo_cmd.data_direction = DMA_TO_DEVICE;
			break;
		case SG_DXFER_TO_FROM_DEV:
		case SG_DXFER_FROM_DEV:
			pfo_cmd.data_direction = DMA_FROM_DEVICE;
			break;
		}
	}


	if (pfo_blk_fill_sghdr_rq(q, &pfo_cmd, hdr, mode)) {
		return -EFAULT;
	}

	if (hdr->iovec_count) {
		const int size = sizeof(struct sg_iovec) * hdr->iovec_count;
		size_t iov_data_len;
		struct iovec *iov;
		int i;

		sg_iov = kmalloc(size, GFP_KERNEL);
		if (!sg_iov) {
			ret = -ENOMEM;
			goto out;
		}

		if (copy_from_user(sg_iov, hdr->dxferp, size)) {
			kfree(sg_iov);
			ret = -EFAULT;
			goto out;
		}

		/*
		 * Sum up the vecs, making sure they don't overflow
		 */
		iov = (struct iovec *) sg_iov;
		iov_data_len = 0;
		for (i = 0; i < hdr->iovec_count; i++) {
			if (iov_data_len + iov[i].iov_len < iov_data_len) {
				kfree(sg_iov);
				ret = -EINVAL;
				goto out;
			}
			iov_data_len += iov[i].iov_len;
		}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,14,0)
		/* SG_IO howto says that the shorter of the two wins */
		if (hdr->dxfer_len < iov_data_len) {
			hdr->iovec_count = iov_shorten(iov,
							hdr->iovec_count,
							hdr->dxfer_len);
			iov_data_len = hdr->dxfer_len;
		}
#endif
		pfo_cmd.buffer = sg_iov;
		pfo_cmd.bufflen = iov_data_len;
		pfo_cmd.iov_count = hdr->iovec_count;
	} else if (hdr->dxfer_len) {
		pfo_cmd.buffer = hdr->dxferp;
		pfo_cmd.bufflen = hdr->dxfer_len;
	}

	if (ret)
		goto out;

	memset(sense, 0, sizeof(sense));
	pfo_cmd.raw_sense = sense;
	pfo_cmd.sense_len = 0;
	pfo_cmd.retries = 0;

	start_time = jiffies;

	/* ignore return value. All information is passed back to caller
	 * (if he doesn't check that is his problem).
	 * N.B. a non-zero SCSI status is _not_ necessarily an error.
	 */
	ret = resolve_send(fdev_num, &pfo_cmd);

	hdr->duration = jiffies_to_msecs(jiffies - start_time);

	if (sg_iov) {
		kfree(sg_iov);
	}
	if (ret) {
		ret = -ENOMEM;
		goto out;
	}
#else
	struct sg_iovec *iov = NULL;
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	struct apfo_scsi_command pfo_cmd;

	DBG_INFO("enter fdev %d\n", fdev_num);
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = scsi_cmd;
	pfo_cmd.cmd_len = hdr->cmd_len;
	pfo_cmd.buffer = NULL;
	pfo_cmd.bufflen = 0;
#ifdef REQ_QUIET
	pfo_cmd.flags = REQ_QUIET;
#else
	pfo_cmd.flags = RQF_QUIET;
#endif
	pfo_cmd.source = PFO_SOURCE_IOCTL;
	pfo_cmd.iov_count = 0;
	pfo_cmd.mdata = NULL;

	if (hdr->interface_id != 'S')
		return -EINVAL;
	if (hdr->cmd_len > BLK_MAX_CDB)
		return -EINVAL;


	pfo_cmd.data_direction = DMA_NONE;
	if (hdr->dxfer_len) {
		switch (hdr->dxfer_direction) {
		default:
			return -EINVAL;
		case SG_DXFER_TO_DEV:
			pfo_cmd.data_direction = DMA_TO_DEVICE;
			break;
		case SG_DXFER_TO_FROM_DEV:
		case SG_DXFER_FROM_DEV:
			pfo_cmd.data_direction = DMA_FROM_DEVICE;
			break;
		}
	}


	if (pfo_blk_fill_sghdr_rq(q, &pfo_cmd, hdr, mode)) {
		return -EFAULT;
	}

	if (hdr->iovec_count) {
		const int size = sizeof(struct sg_iovec) * hdr->iovec_count;
		size_t iov_data_len;

		iov = kmalloc(size, GFP_KERNEL);
		if (!iov) {
			ret = -ENOMEM;
			goto out;
		}

		if (copy_from_user(iov, hdr->dxferp, size)) {
			kfree(iov);
			ret = -EFAULT;
			goto out;
		}

		/* SG_IO howto says that the shorter of the two wins */
		iov_data_len = iov_length((struct iovec *)iov,
					hdr->iovec_count);
		if (hdr->dxfer_len < iov_data_len) {
			hdr->iovec_count = iov_shorten((struct iovec *)iov,
							hdr->iovec_count,
							hdr->dxfer_len);
			iov_data_len = hdr->dxfer_len;
		}
		pfo_cmd.buffer = iov;
		pfo_cmd.bufflen = iov_data_len;
		pfo_cmd.iov_count = hdr->iovec_count;
	} else
	if (hdr->dxfer_len) {
		pfo_cmd.buffer = hdr->dxferp;
		pfo_cmd.bufflen = hdr->dxfer_len;
	}

	if (ret)
		goto out;

	memset(sense, 0, sizeof(sense));
	pfo_cmd.raw_sense = sense;
	pfo_cmd.sense_len = 0;
	pfo_cmd.retries = 0;

	start_time = jiffies;

	ret = resolve_send(fdev_num, &pfo_cmd);
	if (iov)
		kfree(iov);
	if (ret) {
		ret = -ENOMEM;
		goto out;
	}

	hdr->duration = jiffies_to_msecs(jiffies - start_time);
#endif

	return pfo_blk_complete_sghdr_rq(&pfo_cmd, hdr);
out:
	return ret;
}

/** Same as sg_scsi_ioctl.
 */
static int local_sg_scsi_ioctl(int fdev_num, fmode_t mode,
		struct scsi_ioctl_command __user *sic)
{
#define OMAX_SB_LEN 16	  /* For backward compatibility */
	int err;
	unsigned int in_len, out_len, bytes, opcode, cmdlen;
	char *buffer = NULL, sense[SCSI_SENSE_BUFFERSIZE];

	int retval, timeout, retries, data_direction;
	struct apfo_scsi_command pfo_cmd;
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];

	DBG_INFO("deprecated\n");

	if (!sic)
		return -EINVAL;

	/*
	 * get in an out lengths, verify they don't exceed a page worth of data
	 */
	if (get_user(in_len, &sic->inlen))
		return -EFAULT;
	if (get_user(out_len, &sic->outlen))
		return -EFAULT;
	if (in_len > PAGE_SIZE || out_len > PAGE_SIZE)
		return -EINVAL;
	if (get_user(opcode, sic->data))
		return -EFAULT;

	bytes = max(in_len, out_len);
	if (bytes) {
		buffer = kzalloc(bytes, /* q->bounce_gfp | */
				GFP_USER | __GFP_NOWARN);
		if (!buffer)
			return -ENOMEM;

	}

	data_direction = in_len ? DMA_TO_DEVICE : DMA_FROM_DEVICE;

	cmdlen = COMMAND_SIZE(opcode);

	/*
	 * get command and data to send to device, if any
	 */
	err = -EFAULT;
	if (copy_from_user(scsi_cmd, sic->data, cmdlen))
		goto error;

	if (in_len && copy_from_user(buffer, sic->data + cmdlen, in_len))
		goto error;

#if LINUX_VERSION_CODE > KERNEL_VERSION(3,12,0)
	err = blk_verify_command(scsi_cmd, mode & FMODE_WRITE);
#else
	/* RHEL 6.4 and beyond */
	err = 0;
#endif
//RHEL 6.3 err = blk_verify_command(scsi_cmd, mode & FMODE_WRITE);

	if (err)
		goto error;

	/* default.  possible overriden later */
	retries = 5;

	switch (opcode) {
	case SEND_DIAGNOSTIC:
	case FORMAT_UNIT:
		timeout = FORMAT_UNIT_TIMEOUT;
		retries = 1;
		break;
	case START_STOP:
		timeout = START_STOP_TIMEOUT;
		break;
	case MOVE_MEDIUM:
		timeout = MOVE_MEDIUM_TIMEOUT;
		break;
	case READ_ELEMENT_STATUS:
		timeout = READ_ELEMENT_STATUS_TIMEOUT;
		break;
	case READ_DEFECT_DATA:
		timeout = READ_DEFECT_DATA_TIMEOUT;
		retries = 1;
		break;
	default:
		timeout = BLK_DEFAULT_SG_TIMEOUT;
		break;
	}


	memset(sense, 0, sizeof(sense));


	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = scsi_cmd;
	pfo_cmd.cmd_len = cmdlen;
	pfo_cmd.buffer = buffer;
	pfo_cmd.bufflen = bytes;
	pfo_cmd.source = PFO_SOURCE_CMD_1;
	pfo_cmd.data_direction = data_direction;
	pfo_cmd.timeout = timeout;
	pfo_cmd.retries = retries;
	pfo_cmd.iov_count = 0;
#ifdef REQ_QUIET
	pfo_cmd.flags = REQ_QUIET;
#else
	pfo_cmd.flags = RQF_QUIET;
#endif
	pfo_cmd.mdata = NULL;
	pfo_cmd.raw_sense = sense;
	pfo_cmd.sense_len = 0;
	pfo_cmd.response_code = 0;
	/* Send the command */
	retval = resolve_send(fdev_num, &pfo_cmd);
	if (retval != PFO_GOOD) {
		err = -ENOMEM;
		goto error;
	}
	err = pfo_cmd.errors & 0xff;	/* only 8 bit SCSI status */
	if (err) {
		if (pfo_cmd.sense_len && pfo_cmd.raw_sense) {
			bytes = (OMAX_SB_LEN > pfo_cmd.sense_len) ?
				pfo_cmd.sense_len : OMAX_SB_LEN;
			if (copy_to_user(sic->data, pfo_cmd.raw_sense, bytes))
				err = -EFAULT;
		}
	} else {
		if (copy_to_user(sic->data, buffer, out_len))
			err = -EFAULT;
	}

error:
	kfree(buffer);
	return err;
}

/** Handle scsi_cmd_ioctl() call.
 */
int pfo_scsi_cmd_ioctl(struct scsi_device *sdev, struct request_queue *q,
			struct gendisk *disk, fmode_t mode,
			unsigned int ioctl_cmd, void __user *arg)
{
	char tmp[50];
	int fdev_num = pfo_device_number(sdev);

	if (q == NULL) {
		return -ENXIO;
	}
	if (fdev_num != NOT_PFO_DEVICE) {
		/* There are 3 categories of ioctl commands -
		 * We know we should return a failure (e.g. -ENOTTY).
		 * We know it is acceptable to pass to the Linux kernel.
		 * We must support with our own code.
		 */
		switch (ioctl_cmd) {
		/* The following list contains the items from
		 * scsi_cmd_ioctl().  The CDROM items deleted.
		 */
		case SCSI_IOCTL_SEND_COMMAND:
			return local_sg_scsi_ioctl(fdev_num, mode, arg);
		case SG_GET_VERSION_NUM:
		case SCSI_IOCTL_GET_IDLUN:
		case SCSI_IOCTL_GET_BUS_NUMBER:
		case SG_SET_TIMEOUT:
		case SG_GET_TIMEOUT:
		case SG_GET_RESERVED_SIZE:
		case SG_SET_RESERVED_SIZE:
		case SG_EMULATED_HOST:
			DBG_INFO("Sent to kernel cmd=0x%04X\n", ioctl_cmd);
			return scsi_cmd_ioctl(q, disk, mode, ioctl_cmd, arg);
		case SG_IO:
			{
			struct sg_io_hdr hdr;
			int err;
			if (copy_from_user(&hdr, arg, sizeof(hdr)))
				return -EFAULT;
			err = pfo_sg_io(q, fdev_num, &hdr, mode);
			if (err)
				return err;
			if (copy_to_user(arg, &hdr, sizeof(hdr)))
				return -EFAULT;
			return err;
			}
		default:
			break;
		}
	} else {
		return scsi_cmd_ioctl(q, disk, mode, ioctl_cmd, arg);
	}
	if ((ioctl_cmd & TCGETS & 0xff00) == (TCGETS & 0xff00)) {
		DBG_WARNING("ioctl cmd=0x%04X not supported by APFO.\n", ioctl_cmd);
		return -ENOIOCTLCMD;
	}
	/* Fail all unrecognized commands */
	DBG_INFO("");
	sprintf(tmp, "scsi cmd ioctl cmd=0x%04X not supported by APFO.\n",
				ioctl_cmd);
	print_log(tmp);
	return -ENOSYS;
}
EXPORT_SYMBOL(pfo_scsi_cmd_ioctl);

/*###############################################################*/
/* Functions that interface to other files in the module.        */
/*###############################################################*/

int upper_add_new_device(void *replicate_sdev, int fdev_num)
{
	int index;

	DBG_INFO("fdev=%d sdev=%p\n", fdev_num, replicate_sdev);
	mutex_lock(&add_lock);
	if (highest_fdev_num == FDEV_LIST_EMPTY) {
		for (index = 0; index <= MAX_FDEV_NUMBER; index++) {
			apfo_devices[index] = NULL;
		}
	}
	apfo_devices[fdev_num] = replicate_sdev;
	if (fdev_num > highest_fdev_num) {
		highest_fdev_num = fdev_num;
	}
	mutex_unlock(&add_lock);
	return 0;
}

void upper_not_fo_device(int fdev_num)
{
	DBG_INFO("fdev=%d was sdev=%p\n", fdev_num, apfo_devices[fdev_num]);
	apfo_devices[fdev_num] = NULL;
}

void upper_exit(void)
{
	/* No memory is allocated by this file */
	highest_fdev_num = FDEV_LIST_EMPTY;
}

int pfo_open(struct scsi_device *sdev, struct file *filp)
{
	int fdev_num = pfo_device_number(sdev);

	if (fdev_num != NOT_PFO_DEVICE) {
		if (resolve_open(fdev_num, filp) != PFO_GOOD) {
			DBG_INFO("return failure fdev=%d\n", fdev_num);
			return -ENODEV;
		}
	}
	return 0;
}
EXPORT_SYMBOL(pfo_open);

int pfo_close(struct scsi_device *sdev, struct file *filp)
{
	int fdev_num = pfo_device_number(sdev);

	if (fdev_num != NOT_PFO_DEVICE) {
		if (resolve_close(fdev_num, filp) != PFO_GOOD) {
			DBG_INFO("return failure fdev=%d\n", fdev_num);
			return -EIO;
		}
	}
	return 0;
}
EXPORT_SYMBOL(pfo_close);

/** Translate scsi_device pointer to a failover deivce number.
 * @param[in] sdev scsi_device pointer
 * @return the pfo device number or NOT_PFO_DEVICE.
 */
int pfo_device_number(struct scsi_device *sdev)
{
	int index;

	if (highest_fdev_num == FDEV_LIST_EMPTY) {
		return NOT_PFO_DEVICE;
	}
	for (index = 0; index <= highest_fdev_num; index++) {
		if (sdev == apfo_devices[index]) {
			return index;
		}
	}
	return NOT_PFO_DEVICE;
}
EXPORT_SYMBOL(pfo_device_number);

/** Handle this SCSI command request.
 * buffer must be kernel memory, if bufflen>0
 */
int pfo_execute_req(struct scsi_device *sdev, unsigned char *cmd,
		int data_direction, void *buffer, unsigned bufflen,
		struct scsi_sense_hdr *sshdr, int timeout, int retries,
		int *resid)
{
	int fdev_num = pfo_device_number(sdev);

	if (fdev_num != NOT_PFO_DEVICE) {
		return pfo_scsi_execute_req(fdev_num, cmd, data_direction,
			buffer, bufflen, sshdr, timeout, retries, resid);
	} else {
		return scsi_execute_req(sdev, cmd, data_direction,
			buffer, bufflen, sshdr, timeout, retries, resid);
	}
}
EXPORT_SYMBOL(pfo_execute_req);

int pfo_st_compat_ioctl(int fdev_num, int ioctl_cmd, void __user *arg)
{
	/* Not supported by APFO driver. */
	return -ENOIOCTLCMD;
}
EXPORT_SYMBOL(pfo_st_compat_ioctl);

int pfo_sg_compat_ioctl(int fdev_num, int ioctl_cmd, void __user *arg)
{
	/* Not supported by APFO driver. */
	return -ENOIOCTLCMD;
}
EXPORT_SYMBOL(pfo_sg_compat_ioctl);


int pfo_sg_scsi_ioctl(struct scsi_device *sdev, struct request_queue *q,
		struct gendisk *disk, fmode_t mode,
		struct scsi_ioctl_command __user *sic)
{
	char tmp[50];
	int fdev_num = pfo_device_number(sdev);

	if (fdev_num != NOT_PFO_DEVICE) {
		return local_sg_scsi_ioctl(fdev_num, mode, sic);
	} else {
		return sg_scsi_ioctl(q, disk, mode, sic);
	}
	/* Fail all unrecognized commands */
	DBG_INFO("");
	sprintf(tmp, "sg scsi ioctl cmd=0x%04X not supported by APFO.\n",
				SCSI_IOCTL_SEND_COMMAND);
	print_log(tmp);
	return -ENOSYS;
}
EXPORT_SYMBOL(pfo_sg_scsi_ioctl);

/** Handle this SCSI command request from the generic driver.
 * @return 0=pass
 */
int pfo_sg_scsi_execute(int fdev_num, const unsigned char *cmd, int cmd_len,
		int data_direction, void *buffer, unsigned bufflen,
		unsigned char *raw_sense, int timeout, int retries,
		int *resid_len, struct rq_map_data *mdata,
		int *errors, unsigned int iov_count)
{
	struct apfo_scsi_command pfo_cmd;
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	int retval;

	memcpy(scsi_cmd, cmd, cmd_len);
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = scsi_cmd;
	pfo_cmd.cmd_len = cmd_len;
	pfo_cmd.raw_sense = raw_sense;
	pfo_cmd.data_direction = data_direction;
	pfo_cmd.timeout = timeout;
	pfo_cmd.buffer = buffer;
	pfo_cmd.bufflen = bufflen;
	pfo_cmd.retries = retries;
#ifdef REQ_QUIET
	pfo_cmd.flags = REQ_QUIET;
#else
	pfo_cmd.flags = RQF_QUIET;
#endif
	pfo_cmd.source = PFO_SOURCE_SG;
	pfo_cmd.iov_count = iov_count;
	pfo_cmd.mdata = mdata;
	if (mdata != NULL) {
		pfo_cmd.offset = mdata->offset;
	}
	/* Send the command */
	retval = resolve_send(fdev_num, &pfo_cmd);
	/* Put the results back into the request */
	*errors = pfo_cmd.errors;
	*resid_len = pfo_cmd.resid_len;
	return retval;
}
EXPORT_SYMBOL(pfo_sg_scsi_execute);

/** Handle this SCSI command request from the tape driver.
 * @return 0=pass
 */
int pfo_st_scsi_execute(int fdev_num, const unsigned char *cmd,
		int data_direction, void *buffer, unsigned bufflen,
		unsigned char *raw_sense, int timeout, int retries,
		int *resid_len, struct rq_map_data *mdata,
		int *errors)
{
	struct apfo_scsi_command pfo_cmd;
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	int retval;

	memcpy(scsi_cmd, cmd, COMMAND_SIZE(cmd[0]));
	memset(&pfo_cmd, 0, sizeof(struct apfo_scsi_command));
	pfo_cmd.scsi_cmd = scsi_cmd;
	pfo_cmd.cmd_len = COMMAND_SIZE(cmd[0]);
	pfo_cmd.raw_sense = raw_sense;
	pfo_cmd.data_direction = data_direction;
	pfo_cmd.timeout = timeout;
	pfo_cmd.buffer = buffer;
	pfo_cmd.bufflen = bufflen;
	pfo_cmd.retries = retries;
#ifdef REQ_QUIET
	pfo_cmd.flags = REQ_QUIET;
#else
	pfo_cmd.flags = RQF_QUIET;
#endif
	pfo_cmd.source = PFO_SOURCE_ST;
	pfo_cmd.iov_count = 0;
	pfo_cmd.mdata = mdata;
	if (mdata != NULL) {
		pfo_cmd.offset = mdata->offset;
	}
	/* Send the command */
	retval = resolve_send(fdev_num, &pfo_cmd);
	/* Put the results back into the request */
	*errors = pfo_cmd.errors;
	*resid_len = pfo_cmd.resid_len;
	return retval;
}
EXPORT_SYMBOL(pfo_st_scsi_execute);
