/*******************************************************************************
 *   lin_tape_ioctl_tape.c
 *
 *   (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ******************************************************************************/

#include "lin_tape_scsi_ioctl.h"
#include "lin_tape_timeout_tape.h"
#include "ibmekm/inc/dd_event.h"
#include "lin_tape_scsi_tape.h"
#include <linux/delay.h>

/*****************************************************************************
 *                                                                           *
 *              Tape Drive Ioctls                                            *
 *                                                                           *
 *****************************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: tape_capacity_scaling                                                 *
*                                                                              *
*  PURPOSE: Set capacity of medium                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    capacity_scaling: value to place in drive descriptor after operation      *
*    capacity_scaling_hex_val: value to place in mode page for mode select     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_capacity_scaling(struct tape_descriptor** drvp,
	unchar capacity_scaling, unchar capacity_scaling_hex_val)
{
	struct medium_sense* mode_buf = NULL;
	int rc = STATUS_SUCCESS, buf_len = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_capacity_scaling entry\n"));

	if(!WriteAccess(drv->access_mode)) {
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buf = kmalloc(sizeof(struct medium_sense), GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', sizeof(struct medium_sense));

	rc = tape_modesense_page(&drv, MediumSensePageNumber, (unchar*)mode_buf,
		sizeof(struct medium_sense), True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(mode_buf->page.capacity_scaling == capacity_scaling_hex_val) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	mode_buf->header.blk_descriptor_length = 0;
	mode_buf->page.capacity_scaling_valid = True;
	mode_buf->page.capacity_scaling = capacity_scaling_hex_val;

	buf_len = mode_buf->page.page_length + 2 +
		sizeof(struct mode_parm_header);
	drv->buffer = kmalloc(buf_len, GFP_KERNEL);
	if(!drv->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->buffer, '\0', buf_len);
	memcpy(drv->buffer, mode_buf,
		min((int)buf_len, (int)sizeof(struct medium_sense)));

	kfree(mode_buf);
	mode_buf = drv->buffer;
	drv->buffer = NULL;

	rc = tape_modeselect(&drv, (unchar*)mode_buf, buf_len);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(rc == STATUS_SUCCESS) {
		drv->capacity_scaling = capacity_scaling;
		drv->capacity_scaling_hex_val = capacity_scaling_hex_val;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	if(mode_buf) kfree(mode_buf);

	if(drv) {
		if(drv->buffer) kfree(drv->buffer);
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	return rc;
} /* tape_capacity_scaling() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_compression                                                  *
*                                                                              *
*  PURPOSE: Set compression on/off                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    compression: if True, set compression on, else set compression off        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_compression(struct tape_descriptor** drvp, unchar compression)
{
	struct dev_conf_parm* dev_conf_buf = NULL;
	struct compression_parm* comp_buf = NULL;
	int rc = 0;
	int set_comp = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_set_compression entry\n"));

	if(!drv->cdb) {
		printk("lin_tape: tape_set_compression: no CDB\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	set_comp = compression ? 1 : 0;
	if(drv->drv_id == IBM_3580) {
		comp_buf = kmalloc(sizeof(struct compression_parm), GFP_KERNEL);
		if(!comp_buf) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(comp_buf, '\0', sizeof(struct compression_parm));

		rc = tape_modesense_page(&drv, DataCompressionPageNumber,
			(unchar*)comp_buf, sizeof(struct compression_parm),
			True, True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		comp_buf->header.mode_data_length = False;
		comp_buf->page[2] = set_comp ?
			(comp_buf->page[2] | 0x80) : (comp_buf->page[2] & 0x7E);

		rc = tape_modeselect(&drv, (unchar*)comp_buf,
			sizeof(struct compression_parm));
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	dev_conf_buf = kmalloc(sizeof(struct dev_conf_parm), GFP_KERNEL);
	if(!dev_conf_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(dev_conf_buf, '\0', sizeof(struct dev_conf_parm));
	rc = tape_modesense_page(&drv, DeviceConfigurationPageNumber,
		(unchar*)dev_conf_buf, sizeof(struct dev_conf_parm), True,
		True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	dev_conf_buf->header.mode_data_length = False;
	dev_conf_buf->page[14] = set_comp;

	rc = tape_modeselect(&drv, (unchar*)dev_conf_buf,
		sizeof(struct dev_conf_parm));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	drv->compression = compression;

EXIT_LABEL:

	*drvp = drv;
	if(comp_buf) kfree(comp_buf);
	if(dev_conf_buf) kfree(dev_conf_buf);
	if(drv) {
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */

	return rc;
} /* tape_set_compression() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_data_safe_mode_and_pew                                           *
*                                                                              *
*  PURPOSE: Query/set data_safe_mode and pew with mode page 0x10 subpage 0x01  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    parm_buf: the change parameters struct for this request                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    parm_buf->data_safe_mode: -1 to get only, 0 to set False, 1 to set True   *
*                                                                              *
*                                                                              *
*******************************************************************************/

int tape_data_safe_mode_and_pew(struct tape_descriptor** drvp,
	struct stchgp_s* parm_buf)
{
	int rc = STATUS_SUCCESS, size = 0;
	struct dvc_cfg_ext_subpg_with_hdr* subpage = NULL;
	struct tape_descriptor* drv = *drvp;
	int set = parm_buf ? (int)parm_buf->data_safe_mode : -1;

	if((drv->drv_id != IBM_3580 && drv->drv_id != IBM_3592) ||
		(drv->drv_id == IBM_3580 && drv_gen(drv) < 5) ||
		(drv->drv_id == IBM_3592 && drv_gen(drv) < 4)) {
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	size = sizeof(struct dvc_cfg_ext_subpg_with_hdr);

	subpage = kmalloc(size, GFP_KERNEL);
	if(!subpage) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(subpage, '\0', size);

	rc = tape_modesense10_page(&drv, DeviceConfigurationPageNumber,
		(unchar*)subpage, size, DvcCfgExtSubpage, True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->data_safe_mode = subpage->page.write_mode;
	memcpy(drv->pews, subpage->page.pews, 2);

	if(!parm_buf) goto EXIT_LABEL;

	/* if neither parameter needs to be set, skip mode select */
	if((set < 0 || set == drv->data_safe_mode) &&
		(!memcmp(drv->pews, parm_buf->pews, 2))) goto EXIT_LABEL;

	if(set >= 0) subpage->page.write_mode = set ? 0x1 : 0x0;
	memcpy(subpage->page.pews, parm_buf->pews, 2);
	subpage->page.ps = 0;
	subpage->page.spf = 1;

	rc = tape_modeselect10(&drv, (unchar*)subpage, True, size);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	drv->data_safe_mode = subpage->page.write_mode;
	memcpy(drv->pews, subpage->page.pews, 2);

EXIT_LABEL:

	*drvp = drv;
	if(subpage) kfree(subpage);
	return rc;
} /* tape_data_safe_mode_and_pew() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_display_msg                                                      *
*                                                                              *
*  PURPOSE: Display message on the 359x panel                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    msg_func: type of messages and how to process them (see SCSI spec)        *
*    msg0: eight-byte ASCII message to display on panel                        *
*    msg1: if msg0 and msg1 are blank, clear the message from the panel        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_display_msg(struct tape_descriptor** drvp, char msg_func,
	char* msg0, char* msg1)
{
	struct display_msg* display_buf = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_display_msg entry\n"));

	if(drv->drv_id != IBM_3590 && drv->drv_id != IBM_3592) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	display_buf = kmalloc(sizeof(struct display_msg), GFP_KERNEL);
	if(!display_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(display_buf, '\0', sizeof(struct display_msg));

	display_buf->function = msg_func;
	memcpy(&display_buf->msg0[0], msg0, 8);
	memcpy(&display_buf->msg1[0], msg1, 8);
	display_buf->length[1] = 0x10;

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_display_msg.op_code = DISPLAY_MESSAGE;
	drv->cdb->cdb_display_msg.parm_list_length = sizeof(struct display_msg);
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = DISPLAY_MESSAGE;

	drv->buffer = display_buf;
	drv->buf_len = sizeof(struct display_msg);

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, DisplayMessageTimeOut),
		DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	if(display_buf) kfree(display_buf);
	drv->buffer = NULL;
	*drvp = drv;
	return rc;
} /* tape_display_msg() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_erase                                                            *
*                                                                              *
*  PURPOSE: Erase the tape                                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    long_bit: if True, perform long erase, else perform short erase           *
*    immediate: if True, return before function completes, else wait           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_erase(struct tape_descriptor** drvp, unchar long_bit,
	unchar immediate)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_erase entry\n"));

	if(!drv->cdb) {
		printk("lin_tape: tape_erase() failed to locate CDB space\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_erase.op_code = ERASE;
	drv->cdb->cdb_erase.long_bit = long_bit;
	drv->cdb->cdb_erase.immediate = immediate;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = ERASE;
	drv->buffer = NULL;
	drv->buf_len = 0;

	drv->pos_valid = False;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, EraseTimeOut), DMA_NONE, 0, False);

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_erase() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_inquiry_page                                                     *
*                                                                              *
*  PURPOSE: Retrieve inquiry data from page indicated by page code             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    page_code: inquiry page code                                              *
*    page: (output) returned data                                              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
ssize_t tape_inquiry_page(struct tape_descriptor** drvp, unchar page_code,
	unchar* page)
{
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_inquiry_page entry\n"));

	if(!drv->cdb || !page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = (void*)page;
	drv->buf_len = MAX_INQ_LEN;
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_inquiry.op_code = INQUIRY;
	drv->cdb->cdb_inquiry.evpd = True;
	drv->cdb->cdb_inquiry.page_code = page_code;
	drv->cdb->cdb_inquiry.alloc_length = MAX_INQ_LEN;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = INQUIRY;
	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, InquiryTimeOut), DMA_FROM_DEVICE,
		0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s unable to get inquiry page %x\n",
			drv->devid, page_code);
		goto EXIT_LABEL;
	} /* if */

	if(!drv->buffer) {
		printk("lin_tape: %s unable to get page %x\n", drv->devid,
			page_code);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)((page[3] & 0xFF) + 4),
			(int) (FullTraceUnitSize - 7)), drv->buffer);

EXIT_LABEL:

	*drvp = drv;
	drv->buf_len = 0;
	drv->buffer = NULL;

	return rc;
} /* tape_inquiry_page() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_load_unload                                                      *
*                                                                              *
*  PURPOSE: Load or unload cartridge from drive                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    immediate: if True, return before command completes, else wait            *
*    load: if True load, else unload                                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_load_unload(struct tape_descriptor **drvp, unchar immediate,
	unchar load)
{
	int rc = STATUS_SUCCESS, erp_rc = 0;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_load_unload entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(!load && drv->auto_logsense && ttype != LIN_TAPE_INVALID_TRACE) {
		if(!drv->logsense)
			drv->logsense = kmalloc(MaxLogSense, GFP_KERNEL);

		WRITE_LOCK(&drv->logsense_lock);
		if(drv->logsense) {
			drv->logsense_len = 0;
			WRITE_UNLOCK(&drv->logsense_lock);

			drv->in_erp = True;
			erp_rc = tape_full_logsense(&drv);
			drv->in_erp = False;
			if(erp_rc != STATUS_SUCCESS) {
				DbgPrint(("lin_tape: %s log sense failed "\
					"error: %d\n", drv->devid, erp_rc));
				WRITE_LOCK(&drv->logsense_lock);
				if(drv->logsense) kfree(drv->logsense);
				drv->logsense = NULL;
				drv->logsense_len = 0;
				WRITE_UNLOCK(&drv->logsense_lock);
			} /* if */
		} else WRITE_UNLOCK(&drv->logsense_lock);
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_load_unload.op_code = LOAD_UNLOAD;
	drv->cdb->cdb_load_unload.immediate = immediate;
	drv->cdb->cdb_load_unload.load = load;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = LOAD_UNLOAD;
	drv->buffer = NULL;
	drv->buf_len = 0;
	drv->pos_valid = False;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		load ?  LoadTimeOut : UnloadTimeOut), DMA_NONE, 0, False);

	if(drv && drv->sense_valid && drv->sense_data->sense_key == NOT_READY && !load)
		rc = STATUS_SUCCESS;

	if(rc == STATUS_SUCCESS) {
		drv->eom = False;
		drv->eom_blk_id = 0;
		drv->deferred_eom = False;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_load_unload() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_full_logsense                                                    *
*                                                                              *
*  PURPOSE: Retrieve all log sense pages                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_full_logsense(struct tape_descriptor** drvp)
{
	unchar* logsense_buf = NULL, * pages = NULL, * page_buf = NULL;
	uint page_num = 0, logsense_len = 0, total_len = 0;
	int rc = 0, i = 0;
	struct tape_descriptor* drv = *drvp;

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	READ_LOCK(&drv->logsense_lock);
	logsense_buf = drv->logsense;
	READ_UNLOCK(&drv->logsense_lock);

	if(!logsense_buf) return -ENOMEM;

	WRITE_LOCK(&drv->logsense_lock);
	drv->logsense_len = 0;
	WRITE_UNLOCK(&drv->logsense_lock);

	page_buf = kmalloc(LogPageHeaderSize + LOGSENSEPAGE, GFP_KERNEL);
	if(!page_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	pages = kmalloc(255, GFP_KERNEL);
	if(!pages) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(page_buf, '\0', LogPageHeaderSize + LOGSENSEPAGE);
	memset(pages, '\0', 255);

	rc = tape_logsense_page(&drv, 0, 0, pages, &page_num);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	for(i = 4; i < page_num; i++) {
		rc = tape_logsense_page(&drv, pages[i], 0, page_buf,
			&logsense_len);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if((logsense_len + total_len) <= MaxLogSense) {
			memcpy(logsense_buf, page_buf, logsense_len);
			logsense_buf += logsense_len;
			total_len += logsense_len;
		} else {
			rc = -ENOMEM;
			break;
		} /* else */

	} /* for */

EXIT_LABEL:

	*drvp = drv;

	WRITE_LOCK(&drv->logsense_lock);
	drv->logsense_len = total_len;
	WRITE_UNLOCK(&drv->logsense_lock);

	if(page_buf) kfree(page_buf);
	if(pages) kfree(pages);

	return rc;
} /* tape_full_logsense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_logsense_page                                                    *
*                                                                              *
*  PURPOSE: Get log sense page                                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    page_code: the log page to query                                          *
*    subpage_code: the log subpage to query (0 if N/A)                         *
*    logsense_buf: data buffer into which to copy log sense page               *
*    logsense_len: size of logsense_buf desired by calling function            *
* 	     If zero default to allowed size           							   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
ssize_t tape_logsense_page(struct tape_descriptor** drvp, unchar page_code,
	unchar subpage_code, unchar* logsense_buf, uint* logsense_len)
{
	unchar* page_buf = NULL;
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;
	uint length = 0;
	uint requested_length = *logsense_len;

	DbgPrint(("lin_tape: tape_logsense_page entry: %X\n", requested_length));

	if(!requested_length)
		*logsense_len = LOGSENSEPAGE;
	page_buf = kmalloc(LogPageHeaderSize, GFP_KERNEL);
	if(!drv || !drv->cdb || !page_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(page_buf, '\0', LogPageHeaderSize);
	drv->buffer = (void*)page_buf;
	drv->buf_len = LogPageHeaderSize;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_logsense.op_code = LOG_SENSE;
	drv->cdb->cdb_logsense.sp = False;
	drv->cdb->cdb_logsense.ppc = False;
	drv->cdb->cdb_logsense.pc = True;
	drv->cdb->cdb_logsense.page_code = page_code;
	drv->cdb->cdb_logsense.subpage_code = subpage_code;
	drv->cdb->cdb_logsense.alloc_length[0] = False;
	drv->cdb->cdb_logsense.alloc_length[1] = LogPageHeaderSize;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = LOG_SENSE;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, LogSenseTimeOut),
		DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s can't get log sense\n", drv->devid);
		goto EXIT_LABEL;
	} /* if */

	page_buf = drv->buffer;
	if(!page_buf) {
		printk("lin_tape: %s tape_logsense_page no buf\n", drv->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min(LogPageHeaderSize,
			(int) (FullTraceUnitSize - 7)), drv->buffer);

	length = (page_buf[2] << 8) + page_buf[3] + LogPageHeaderSize;
	if(*logsense_len != LogPageHeaderSize)
		*logsense_len += LogPageHeaderSize;

	if(length > *logsense_len) length = *logsense_len;
	if(length < *logsense_len) *logsense_len = length;

	kfree(page_buf);

	page_buf = kmalloc(*logsense_len, GFP_KERNEL);
	if(!page_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page_buf, '\0', *logsense_len);

	drv->buffer = (void*)page_buf;
	drv->buf_len = *logsense_len;
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_logsense.op_code = LOG_SENSE;
	drv->cdb->cdb_logsense.sp = False;
	drv->cdb->cdb_logsense.ppc = False;
	drv->cdb->cdb_logsense.pc = True;
	drv->cdb->cdb_logsense.page_code = page_code;
	drv->cdb->cdb_logsense.subpage_code = subpage_code;
	drv->cdb->cdb_logsense.alloc_length[0] =
		(unchar)((*logsense_len) >> 8 & 0xFF);
	drv->cdb->cdb_logsense.alloc_length[1] =
		(unchar)((*logsense_len) & 0xFF);
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = LOG_SENSE;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, LogSenseTimeOut),
		DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	page_buf = drv->buffer;
	if(!page_buf) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)*logsense_len,
			(int) (FullTraceUnitSize - 7)), drv->buffer);

	length = (page_buf[2] << 8) + page_buf[3] + LogPageHeaderSize;

	if(*logsense_len > LOGSENSEPAGE) *logsense_len = LOGSENSEPAGE;

	memcpy(logsense_buf, page_buf, *logsense_len);

EXIT_LABEL:

	*drvp = drv;
	if(drv) {
		drv->buf_len = 0;
		drv->buffer = NULL;
	} /* if */
	if(page_buf) kfree(page_buf);

	return rc;
} /* tape_logsense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_modesense_page                                                   *
*                                                                              *
*  PURPOSE: Issue mode sense on the specified page to the specified device     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    page_code: the mode page to query                                         *
*    modesense_buf: data buffer into which to copy mode sense page             *
*    alloc_len: size of modesense_buf                                          *
*    dbd: True if block descriptor not requested, False if it is requested     *
*    return_header: return entire page including four-byte header              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_modesense_page(struct tape_descriptor** drvp, unchar page_code,
	unchar* modesense_buf, int alloc_len, boolean dbd,
	boolean return_header)
{
	unchar *mode_buf = NULL;
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS, timeout = 0, drv_alloc_len = 0;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_modesense_page entry: 0x%x\n", page_code));

	if(!drv->cdb || !modesense_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(alloc_len > MAX_MDSNS_LEN) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	drv_alloc_len = return_header ? alloc_len
		: (alloc_len > MAX_MDSNS_LEN - sizeof(struct mode_parm_header))
		? MAX_MDSNS_LEN
		: alloc_len + sizeof(struct mode_parm_header);
	mode_buf = kmalloc(drv_alloc_len, GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', drv_alloc_len);

	drv->buffer = (void*)mode_buf;
	drv->buf_len = drv_alloc_len;
	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	drv->cdb->cdb_modesense.op_code = MODE_SENSE;
	drv->cdb->cdb_modesense.dbd = dbd;
	drv->cdb->cdb_modesense.page_code = page_code;
	drv->cdb->cdb_modesense.alloc_length = drv_alloc_len;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = MODE_SENSE;

	timeout = compute_timeout(drv->drv_id, ModeSenseTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s tape_modesense_page failed: %d\n",
			drv->devid, rc);
		goto EXIT_LABEL;
	} /* if */

	if(!drv->buffer) {
		printk("lin_tape: %s tape_modesense_page unable to receive "\
			"mode sense data\n", drv->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)(drv->buf_len),
			(int)(FullTraceUnitSize - 7)), drv->buffer);

	if(return_header) memcpy(modesense_buf, mode_buf, drv_alloc_len);
	else memcpy(modesense_buf, &mode_buf[sizeof(struct mode_parm_header)],
		drv_alloc_len - sizeof(struct mode_parm_header));

EXIT_LABEL:

	*drvp = drv;
	if(mode_buf) kfree(mode_buf);
	drv->buf_len = 0;
	drv->buffer = NULL;

	DbgPrint(("lin_tape: tape_modesense_page exit: %d\n", rc));
	return rc;
} /* tape_modesense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_modeselect                                                       *
*                                                                              *
*  PURPOSE: Issue mode select with the given buffer to the specified device    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    modesense_buf: the mode page                                              *
*    alloc_len: number of bytes within the mode page to set                    *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_modeselect(struct tape_descriptor** drvp, unchar* mode_buf,
	uint alloc_len)
{
	int rc = STATUS_SUCCESS, timeout = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_modeselect entry\n"));

	if(!drv->cdb || !mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(alloc_len > MAX_MDSNS_LEN) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* per the 3580 spec, this value is 0 on mode select */
	((struct mode_parm_header*)mode_buf)->mode_data_length = 0;

	drv->buf_len = alloc_len;
	drv->buffer = kmalloc(drv->buf_len, GFP_KERNEL);
	if(!drv->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->buffer, '\0', drv->buf_len);
	memcpy(drv->buffer, mode_buf, drv->buf_len);
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_modeselect.op_code = MODE_SELECT;
	drv->cdb->cdb_modeselect.pf = True;
	drv->cdb->cdb_modeselect.parm_list_length = drv->buf_len;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = MODE_SELECT;

	timeout = compute_timeout(drv->drv_id, ModeSelectTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	*drvp = drv;

	if(drv->buffer) kfree(drv->buffer);
	drv->buffer = NULL;
	drv->buf_len = 0;
	if(drv->cdb) memset(drv->cdb, '\0', MAX_CDB_SIZE);

	DbgPrint(("lin_tape: tape_modeselect exit: %d\n", rc));
	return rc;
} /* tape_modeselect() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_modesense10_page                                                 *
*                                                                              *
*  PURPOSE: Issue mode sense on the specified page to the specified device     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    page_code: the mode page to query                                         *
*    modesense_buf: data buffer into which to copy mode sense page             *
*    alloc_len: the allocation length given for the return data                *
*    subpg: 0 if no subpage requested, otherwise the subpage number            *
*    dbd: True if block descriptor not requested, False if it is requested     *
*    return_header: return entire page including four-byte header              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_modesense10_page(struct tape_descriptor** drvp, unchar page_code,
	unchar* modesense_buf, int alloc_len, int subpg, boolean dbd,
	boolean return_header)
{
	unchar* mode_buf = NULL;
	int rc = STATUS_SUCCESS, timeout = 0, drv_alloc_len = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_modesense10_page entry\n"));

	if(!drv->cdb || !modesense_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv_alloc_len = return_header ? alloc_len
		: alloc_len + sizeof(struct mode_parm_header_10);
	mode_buf = kmalloc(drv_alloc_len, GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', drv_alloc_len);

	drv->buffer = (void*)mode_buf;
	drv->buf_len = drv_alloc_len;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_modesense_10.op_code = MODE_SENSE_10;
	drv->cdb->cdb_modesense_10.dbd = dbd;
	drv->cdb->cdb_modesense_10.page_code = page_code;
	if(subpg) drv->cdb->cdb_modesense_10.sub_page_code = subpg;
	drv->cdb->cdb_modesense_10.alloc_length[0] =
		(unchar)((drv_alloc_len >> 8) & 0xFF);
	drv->cdb->cdb_modesense_10.alloc_length[1] =
		(unchar)(drv_alloc_len & 0xFF);
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = MODE_SENSE_10;

	timeout = compute_timeout(drv->drv_id, ModeSenseTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s tape_modesense10_page failed: %d\n",
			drv->devid, rc);
		goto EXIT_LABEL;
	} /* if */

	if(!drv->buffer) {
		printk("lin_tape: %s tape_modesense10_page unable to receive "\
			"mode sense data\n", drv->devid);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	if(return_header) memcpy(modesense_buf, mode_buf, drv_alloc_len);
	else memcpy(modesense_buf,
		&mode_buf[sizeof(struct mode_parm_header_10)],
		drv_alloc_len - sizeof(struct mode_parm_header_10));

EXIT_LABEL:

	*drvp = drv;

	if(mode_buf) kfree(mode_buf);
	drv->buf_len = 0;
	drv->buffer = NULL;

	DbgPrint(("lin_tape: tape_modesense10_page exit: %d\n", rc));
	return rc;
} /* tape_modesense10_page() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_modeselect10                                                     *
*                                                                              *
*  PURPOSE: Issue mode select with the given buffer to the specified device    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    modesense_buf: the mode page                                              *
*    sp: value of the save pages parameter                                     *
*    alloc_len: number of bytes within the mode page to set                    *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_modeselect10(struct tape_descriptor** drvp, unchar* mode_buf,
	boolean sp, uint alloc_len)
{
	int rc = STATUS_SUCCESS, timeout = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_modeselect10 entry\n"));

	if(!drv->cdb || !mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(alloc_len > MAX_MODE_PAGE_10_LENGTH) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	mode_buf[0] = 0;
	mode_buf[1] = 0;

	drv->buf_len = alloc_len;
	drv->buffer = kmalloc(drv->buf_len, GFP_KERNEL);
	if(!drv->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->buffer, '\0', drv->buf_len);

	memcpy(drv->buffer, mode_buf, drv->buf_len);
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	drv->cdb->cdb_modeselect_10.pf = True;
	drv->cdb->cdb_modeselect_10.sp = sp ? 1 : 0;

	/* include header in alloc_len */
	drv->cdb->cdb_modeselect_10.parm_list_length[0] =
		(alloc_len >> 8) & 0xFF;
	drv->cdb->cdb_modeselect_10.parm_list_length[1] = alloc_len & 0xFF;
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = MODE_SELECT_10;
	timeout = compute_timeout(drv->drv_id, ModeSelectTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	*drvp = drv;

	if(drv->buffer) kfree(drv->buffer);
	drv->buffer = NULL;
	drv->buf_len = 0;
	if(drv->cdb) memset(drv->cdb, '\0', MAX_CDB_SIZE);

	DbgPrint(("lin_tape: tape_modeselect10 exit: %d\n", rc));
	return rc;
} /* tape_modeselect10() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_compression                                                  *
*                                                                              *
*  PURPOSE: get the compression mode of drive                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_compression(struct tape_descriptor** drvp)
{
	struct dev_conf_parm* dev_conf_buf = NULL;
	struct compression_parm* comp_buf = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_get_compression entry\n"));

	if(drv->drv_id == IBM_3580) {
		comp_buf = kmalloc(sizeof(struct compression_parm), GFP_KERNEL);
		if(!comp_buf) {
			printk("lin_tape: can't allocate compression buffer\n");
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(comp_buf, '\0', sizeof(struct compression_parm));

		rc = tape_modesense_page(&drv, DataCompressionPageNumber,
			(unchar*)comp_buf, sizeof(struct compression_parm),
			True, True);

		if(rc == STATUS_SUCCESS)
			drv->compression = comp_buf->page[2] >> 7;
		goto EXIT_LABEL;
	} /* if */

	dev_conf_buf = kmalloc(sizeof(struct dev_conf_parm), GFP_KERNEL);
	if(!dev_conf_buf) {
		printk("lin_tape: cannot allocate compression buffer\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(dev_conf_buf, '\0', sizeof(struct dev_conf_parm));

	rc = tape_modesense_page(&drv, DeviceConfigurationPageNumber,
		(unchar*)dev_conf_buf, sizeof(struct dev_conf_parm), True,
		True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	drv->compression = dev_conf_buf->page[14];

EXIT_LABEL:

	*drvp = drv;

	drv->buf_len = 0;
	if(dev_conf_buf) kfree(dev_conf_buf);
	if(comp_buf) kfree(comp_buf);
	drv->buffer = NULL;
	dev_conf_buf = NULL;
	comp_buf = NULL;

	return rc;
} /* tape_get_compression() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_media_parm                                                   *
*                                                                              *
*  PURPOSE: issue mode select p. 0x10 to get medium parameters                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_media_parm(struct tape_descriptor **drvp)
{
	struct media_parm *mode_buf = NULL;
	int rc = STATUS_SUCCESS, timeout = 0;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_get_media_parm entry\n"));

	timeout = compute_timeout(drv->drv_id, ModeSenseTimeOut);

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buf = kmalloc(MAX_MDSNS_LEN, GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', MAX_MDSNS_LEN);

	rc = tape_modesense_page(&drv, DeviceConfigurationPageNumber,
		(unchar*)mode_buf, MAX_MDSNS_LEN, False, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->blk_size = (mode_buf->block.blk_length[0] << 16)
		+ (mode_buf->block.blk_length[1] << 8)
		+ mode_buf->block.blk_length[2];

	if(drv->sili && drv->blk_size == 4) drv->blk_size = 0;

	drv->density = mode_buf->block.density_code;
	drv->buffered_mode = mode_buf->header.buffered_mode;
	drv->medium_type = mode_buf->header.medium_type;

	if(drv->drv_id == IBM_3580) drv->write_protect = mode_buf->header.wp;

	if(mode_buf) kfree(mode_buf);
	mode_buf = NULL;

	/* 359x mode page 0x23 to report physical write protect flag */
	if(drv->drv_id == IBM_3590 || drv->drv_id == IBM_3592) {
		rc = tape_wp_capscal(&drv);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if (!drv->autotimeout) tape_get_timeouts(*drvp);
	*drvp = drv;

EXIT_LABEL:

	if(mode_buf) kfree(mode_buf);
	drv->buffer = NULL;
	drv->buf_len = 0;

	DbgPrint(("lin_tape: tape_get_media_parm exit: %d\n", rc));
	return rc;

} /* tape_get_media_parm() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_wp_capscal                                                       *
*                                                                              *
*  PURPOSE: set write protection and capacity scaling parameters for drive     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_wp_capscal(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS, timeout = 0;
	struct tape_descriptor* drv = *drvp;
	struct medium_sense *medium_buf = NULL;
	unchar cap = 0, ttype = 0;

	timeout = compute_timeout(drv->drv_id, ModeSenseTimeOut);
	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	medium_buf = kmalloc(sizeof(struct medium_sense), GFP_KERNEL);
	if(!medium_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(medium_buf, '\0', sizeof(struct medium_sense));

	rc = tape_modesense_page(&drv, MediumSensePageNumber,
		(unchar*)medium_buf, sizeof(struct medium_sense), True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* if medium_id is equal to 0x0000 then there is no medium present */
	if(!medium_buf->page.medium_id[0] && !medium_buf->page.medium_id[1]) {
		drv->write_protect = False;
		drv->logic_write_protect = False;
		drv->capacity_scaling = SCALE_100;
		goto EXIT_LABEL;
	} /* if */

	/* if there are any forms of write protect on the medium */
	drv->write_protect = medium_buf->page.physical_wp;
	if(medium_buf->page.permanent_wp)
		drv->logic_write_protect = WORM_PROTECT;
	else if(medium_buf->page.persistent_wp)
		drv->logic_write_protect = PERSISTENT_PROTECT;
	else if(medium_buf->page.associated_wp)
		drv->logic_write_protect = ASSOCIATED_PROTECT;
	else drv->logic_write_protect = False;

	cap = medium_buf->page.capacity_scaling;
	drv->capacity_scaling_hex_val = cap;
	if(drv->drv_id != IBM_3592) {
		if(cap == Capacity100 || cap >= Capacity100RangeLow)
			drv->capacity_scaling = SCALE_100;
		else if(cap == Capacity75 || (cap <= Capacity75RangeHigh
			&& cap >= Capacity75RangeLow))
			drv->capacity_scaling = SCALE_75;
		else if(cap == Capacity50 || (cap <= Capacity50RangeHigh
			&& cap >= Capacity50RangeLow))
			drv->capacity_scaling = SCALE_50;
		else if(cap == Capacity25 || (cap <= Capacity25RangeHigh
			&& cap >= Capacity25RangeLow))
			drv->capacity_scaling = SCALE_25;
		else drv->capacity_scaling = SCALE_100;
	} else drv->capacity_scaling = SCALE_VALUE;

EXIT_LABEL:

	*drvp = drv;
	if(medium_buf) kfree(medium_buf);
	return rc;
} /* tape_wp_capscal() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_pos_ex                                                       *
*                                                                              *
*  PURPOSE: get current position of the tape                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    command: the IOCTL code                                                   *
*    arg: the argument from the IOCTL call                                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_pos_ex(struct tape_descriptor** drvp, uint command, ulong arg)
{
	ssize_t rc = STATUS_SUCCESS;
	struct read_tape_position *read_pos_buf = NULL;

	read_pos_buf = kmalloc(sizeof(struct read_tape_position), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(read_pos_buf, '\0', sizeof(struct read_tape_position));

	if(copy_from_user((void*)read_pos_buf, (void*)arg,
		sizeof(struct read_tape_position))) {
		printk("lin_tape_get_pos_ex: cannot copy from user\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_get_pos(drvp, (struct pos_info*)&read_pos_buf->rp_data,
		read_pos_buf->data_format, True);

	if(rc == STATUS_SUCCESS) {
		if(copy_to_user((void*)arg, (void*)read_pos_buf,
			sizeof(struct read_tape_position))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

EXIT_LABEL:

	if(read_pos_buf) {
		kfree(read_pos_buf);
		read_pos_buf = NULL;
	} /* if */

	return rc;

} /* tape_get_pos_ex() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_pos                                                          *
*                                                                              *
*  PURPOSE: get current position of the tape                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    read_pos_info: output buffer                                              *
*    data_format: short, long, or extended                                     *
*    use_ex: use the extended versions of the output structures                *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_pos(struct tape_descriptor** drvp,
	struct pos_info* read_pos_info, unchar data_format, boolean use_ex)
{
	struct read_pos *buf = NULL;
	char ttype = 0, * flo = NULL;
	int rc = 0, cpy_len = 8, i = 0;
	ulong timeout = 0, alloc_len = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_get_pos entry: fmt %d ex %d\n", data_format,
		use_ex));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	alloc_len =
		data_format == RP_SHORT_FORM
			? sizeof(struct short_data_format) :
		data_format == RP_LONG_FORM
			? sizeof(struct long_data_format) :
		data_format == RP_EXTENDED_FORM
			? sizeof(struct extended_data_format) : 0;

	if(!alloc_len) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(alloc_len, GFP_KERNEL);
	if(!buf) {
		printk("lin_tape: tape_get_pos cannot allocate data buffer\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', alloc_len);

	/* reset the position flag */
	drv->eom = False;
	drv->pos_valid = False;

	drv->buffer = (void*)buf;
	drv->buf_len = alloc_len;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_read_pos.op_code = READ_POSITION;
	drv->cdb->cdb_read_pos.service_action = data_format;
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = READ_POSITION;

	timeout = compute_timeout(drv->drv_id, ReadPositionTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_FROM_DEVICE, 0, False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	buf = drv->buffer;
	if(buf->bpu) {
		drv->pos_valid = False;
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE &&
			drv->buf_len)
			tape_log_buffer_data(drv, min((int)drv->buf_len,
				(int)(FullTraceUnitSize - 7)),
				drv->buffer);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(use_ex) {
		memcpy(read_pos_info, buf, alloc_len);
		drv->pos_valid = True;
		drv->eom = read_pos_info->eop;

		switch(data_format) {
		case RP_SHORT_FORM:
			flo = ((struct short_data_format*)
				read_pos_info)->
				first_logical_obj_position;
			cpy_len = 4;
			break;
		case RP_LONG_FORM:
			flo = ((struct long_data_format*)
				read_pos_info)->logical_obj_number;
			break;
		case RP_EXTENDED_FORM:
			flo = ((struct extended_data_format*)
				read_pos_info)->
				first_logical_obj_position;
			break;

		} /* switch */

		drv->curr_pos = 0;
		for(i = 0; i < cpy_len; i++)
			drv->curr_pos |= ((long long)(flo[i] & 0xFF) <<
				(cpy_len * BITS_PER_BYTE -
				(i + 1) * BITS_PER_BYTE));

		DbgPrint(("lin_tape: current position %llu\n", drv->curr_pos));

		goto EXIT_LABEL;
	} /* if */

	read_pos_info->bop = (unchar)buf->bop;
	read_pos_info->eop = (unchar)buf->eop;

	read_pos_info->first_blk_pos = (ulong)((buf->first_blk_pos[0] << 24)
		+ (buf->first_blk_pos[1] << 16) + (buf->first_blk_pos[2] << 8)
		+ (buf->first_blk_pos[3]));

	read_pos_info->last_blk_pos = (ulong)((buf->last_blk_pos[0] << 24)
		+ (buf->last_blk_pos[1] << 16) + (buf->last_blk_pos[2] << 8)
		+ (buf->last_blk_pos[3]));

	if(!buf->bcu) {
		read_pos_info->blks_in_buf = (ulong)((buf->blks_in_buf[0] << 16)
			+ (buf->blks_in_buf[1] << 8) + (buf->blks_in_buf[2]));
	} else read_pos_info->blks_in_buf = 0;

	if(!buf->bycu) {
		read_pos_info->bytes_in_buf =
			(ulong)((buf->bytes_in_buf[0] << 24)
			+ (buf->bytes_in_buf[1] << 16)
			+ (buf->bytes_in_buf[2] << 8)
			+ (buf->bytes_in_buf[3]));
	} else read_pos_info->bytes_in_buf = 0;

	drv->pos_valid = True;
	drv->eom = read_pos_info->eop;
	drv->curr_pos = read_pos_info->first_blk_pos;

	DbgPrint(("lin_tape: current position %llu\n", drv->curr_pos));

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)drv->buf_len,
			(int)(FullTraceUnitSize-7)), drv->buffer);

EXIT_LABEL:

	if(buf) kfree(buf);
	drv->buffer = NULL;
	drv->buf_len = 0;

	*drvp = drv;

	DbgPrint(("lin_tape: tape_get_pos exit: %d\n", rc));
	return rc;
} /* tape_get_pos() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_prevent_medium_remove                                            *
*                                                                              *
*  PURPOSE: lock/unlock drive door                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    prevent: If true, prevent medium removal, else allow                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_prevent_medium_remove(struct tape_descriptor** drvp,
	unchar prevent)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_prevent_medium_remove entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_medium_removal.op_code = ALLOW_MEDIUM_REMOVAL;
	drv->cdb->cdb_medium_removal.prevent = prevent;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = ALLOW_MEDIUM_REMOVAL;
	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, PreventAllowMediaTimeOut),
		DMA_NONE, 0, False);

EXIT_LABEL:

	if(rc == STATUS_SUCCESS) drv->prevent_removal = prevent;
	DbgPrint(("lin_tape: tape_prevent_medium_remove: prevent = %d\n",
		drv->prevent_removal));
	*drvp = drv;
	return rc;
} /* tape_prevent_medium_remove() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_release                                                          *
*                                                                              *
*  PURPOSE: release the reservation                                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    media: If true, return medium descriptor, else return density descriptor  *
*    report_cnt: (output) number of report density descriptors returned        *
*    report_buf: buffer to receive descriptors                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_release(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_release entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_release.op_code = RELEASE;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = RELEASE;

	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, ReleaseTimeOut), DMA_NONE, 0,
		False);

EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* tape_release() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_report_density                                                   *
*                                                                              *
*  PURPOSE: report current density for writing data to tape                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    media: If true, return medium descriptor, else return density descriptor  *
*    report_cnt: (output) number of report density descriptors returned        *
*    report_buf: buffer to receive descriptors                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_report_density(struct tape_descriptor** drvp, unchar media,
	ushort* report_cnt, struct report_density* report_buf)
{
	ushort buf_len = 0;
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_report_density entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(report_buf, '\0', sizeof(struct report_density));

	drv->buffer = (void*)report_buf;
	drv->buf_len = sizeof(struct report_density);

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_report_density.op_code = REPORT_DENSITY;
	drv->cdb->cdb_report_density.media = media;
	drv->cdb->cdb_logsense.alloc_length[0] =
		(unchar)(sizeof(struct report_density) >> 8);
	drv->cdb->cdb_logsense.alloc_length[1] =
		(unchar)(sizeof(struct report_density) & 0xFF);
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = REPORT_DENSITY;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, ReportDensityTimeOut),
		DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	report_buf = drv->buffer;
	if(!report_buf) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)(drv->buf_len),
			(int)(FullTraceUnitSize - 7)), drv->buffer);

	buf_len = report_buf->header.avail_length[0] << 8;
	buf_len += report_buf->header.avail_length[1];

	/* buf_len includes 2 bytes of reserved fields; we need take them out */
	*report_cnt = (ushort)((buf_len - 2) /
		sizeof(struct report_density_descriptor));

	/* 3580 reports back 54 length instead of 56 */
	if(!*report_cnt && buf_len > sizeof(struct report_density_header))
		*report_cnt = 1;

EXIT_LABEL:
	*drvp = drv;
	drv->buf_len = 0;
	drv->buffer = NULL;

	return rc;
} /* tape_report_density() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_query_partition                                              *
*                                                                              *
*  PURPOSE: retrieve partition information from a currently loaded tape        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive for which we want to query partition   *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_query_partition(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, i = 0;
	struct mode_sense_page* page = NULL;
	struct query_partition* buf = NULL;
	struct medium_partition_page* part = NULL;
	struct device_config_page* cfg = NULL;
	char *partsizes = NULL;

	DbgPrint(("lin_tape_query_partition entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct query_partition)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_tur(&drv);
	if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape_query_partition: tur returned %d, "\
			"operation not permitted\n", rc));
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(sizeof(struct query_partition), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buf, '\0', sizeof(struct query_partition));

	page = kmalloc(sizeof(struct mode_sense_page), GFP_KERNEL);
	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', sizeof(struct mode_sense_page));

	page->page_code = MEDIUM_PARTITION_MODE_PAGE;
	part = (struct medium_partition_page*)page->data;
	rc = tape_modesense_page(&drv, MEDIUM_PARTITION_MODE_PAGE, (char*)part,
		MAX_MDSNS_LEN, True, False);

	if(rc != STATUS_SUCCESS) {
		DbgPrint(("tape_modesense_page on p. 0x11 failed: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	buf->max_partitions = part->max_addtl_parts + 1;
	buf->number_of_partitions = part->addtl_parts_defd + 1;
	buf->size_unit = part->partition_units;

	partsizes = (char *)&part->first_part_size[0];
	for(i = 0; i < buf->number_of_partitions; i++)
		buf->size[i] = (partsizes[i * 2] << 8)
			| (partsizes[i * 2 + 1] & 0xFF);

	buf->partition_method = part->partition_type;

	memset(page, '\0', sizeof(struct mode_sense_page));

	page->page_code = DEVICE_CONFIG_MODE_PAGE;
	cfg = (struct device_config_page*)page->data;
	rc = tape_modesense_page(&drv, DEVICE_CONFIG_MODE_PAGE, (char*)cfg,
		MAX_MDSNS_LEN, True, False);

	if(rc != STATUS_SUCCESS) {
		DbgPrint(("tape_modesense_page on p. 0x10 failed: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	buf->active_partition = cfg->active_partition;

	if(copy_to_user((char*)arg, buf, sizeof(struct query_partition))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(page) kfree(page);
	if(buf) kfree(buf);

	DbgPrint(("lin_tape_query_partition exit: %d\n", rc));
	return rc;
} /* lin_tape_query_partition() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_create_partition                                             *
*                                                                              *
*  PURPOSE: reformat tape with specified partition information                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive for which we want to create partition  *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_create_partition(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, timeout = 0, i = 0, alloc_len = 0,
		hdr_len = sizeof(struct mode_parm_header);
	struct mode_sense_page* page = NULL;
	struct tape_partition* usr_part = NULL;
	struct medium_partition_page* md_part = NULL;
	char *partsizes = NULL;

	DbgPrint(("lin_tape_create_partition entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct tape_partition)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	usr_part = kmalloc(sizeof(struct tape_partition), GFP_KERNEL);
	if(!usr_part) {
		printk("lin_tape_create_partition: kmalloc failed to "\
			"allocate tape_partition!!!\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;

	if(copy_from_user((void*)usr_part, (void*)arg,
		sizeof(struct tape_partition))) {
		printk("lin_tape_create_partition: cannot copy from user\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(usr_part->number_of_partitions > MAX_SUPPORTED_PARTITIONS) {
		DbgPrint(("lin_tape: too many partitions %d\n",
			usr_part->number_of_partitions));
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(usr_part->number_of_partitions == 1) {
		memset((void*)drv->cdb, '\0', MAX_CDB_SIZE);
		drv->cdb->cdb_format_medium.op_code = FORMAT_UNIT;
		drv->cdb_len = CDB_GENERIC_6;
		drv->last_op = FORMAT_UNIT;
		timeout = compute_timeout(drv->drv_id, FormatMediumTimeOut);
		rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

		if(rc != STATUS_SUCCESS)
			printk("lin_tape_create_partition: FORMAT_UNIT "\
				"failed: %d\n", rc);
		goto EXIT_LABEL;
	} /* if */

	page = kmalloc(sizeof(struct mode_sense_page), GFP_KERNEL);
	if(!page) {
		printk("lin_tape_create_partition: kmalloc failed to "\
			"allocate mode sense page!!!\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', sizeof(struct mode_sense_page));

	page->page_code = MEDIUM_PARTITION_MODE_PAGE;
	md_part = (struct medium_partition_page*)
		&page->data[hdr_len];
	rc = tape_modesense_page(&drv, MEDIUM_PARTITION_MODE_PAGE,
		page->data, MAX_MDSNS_LEN, True, True);

	if(rc != STATUS_SUCCESS) {
		DbgPrint(("tape_modesense_page on p. 0x11 failed: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	if(usr_part->number_of_partitions > md_part->max_addtl_parts + 1) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	md_part->addtl_parts_defd = usr_part->number_of_partitions - 1;
	if(usr_part->type == FDP_PARTITION) {
		md_part->fdp = 1;
		md_part->sdp = 0;
		md_part->idp = 0;
	} else if(usr_part->type == SDP_PARTITION) {
		md_part->sdp = 1;
		md_part->fdp = 0;
		md_part->idp = 0;
	} else if(usr_part->type == IDP_PARTITION) {
		md_part->idp = 1;
		md_part->fdp = 0;
		md_part->sdp = 0;

		md_part->partition_units = usr_part->size_unit;

		partsizes = (char *)&md_part->first_part_size[0];
		for(i = 0; i < usr_part->number_of_partitions; i++) {
			partsizes[i * 2]
				= usr_part->size[i] >> 8;
			partsizes[i * 2 + 1]
				= usr_part->size[i] & 0xFF;
		} /* for */
	} else {
		printk("lin_tape_create_partition: invalid partition type\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* else */

	//md_part->pofm = 1;
	md_part->partition_type =
		drv->drv_id == IBM_3592 ? usr_part->partition_method : 0;

	alloc_len = hdr_len
		+ (1 + offsetof(struct medium_partition_page, page_length))
		+ ((struct medium_partition_page*)&page->data[hdr_len])
			->page_length;
	rc = tape_modeselect(&drv, page->data, alloc_len);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape_create_partition: cannot set mode page "\
			"0x11: %d\n", rc);
		goto EXIT_LABEL;
	} /* if */

	memset((void*)drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_format_medium.op_code = FORMAT_UNIT;
	drv->cdb->cdb_format_medium.format = 0x01;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = FORMAT_UNIT;
	timeout = compute_timeout(drv->drv_id, FormatMediumTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS)
		printk("lin_tape_create_partition: FORMAT_UNIT failed: %d\n",
			rc);

EXIT_LABEL:

	if(usr_part) kfree(usr_part);
	if(page) kfree(page);

	DbgPrint(("lin_tape_create_partition exit: %d\n", rc));
	return rc;
} /* lin_tape_create_partition() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_active_partition                                         *
*                                                                              *
*  PURPOSE: move tape head to specified partition                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive for which we want to set partition    *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_set_active_partition(struct tape_descriptor** drvp, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, i = 0, sft = 0, timeout = 0;
	unsigned long long len = 0;
	struct set_active_partition* set_part = NULL;
	struct read_tape_position* pos_buf = NULL;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape_set_active_partition entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct set_active_partition)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	set_part = kmalloc(sizeof(struct set_active_partition), GFP_KERNEL);
	if(!set_part) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;

	if(copy_from_user((void*)set_part, (void*)arg,
		sizeof(struct set_active_partition))) {
		printk("lin_tape_set_active_partition: can't copy from user\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	DbgPrint(("lin_tape_set_active_partition: partition %d\n",
		set_part->partition_number));
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_locate_16.op_code = LOCATE_16;
	drv->cdb->cdb_locate_16.cp = 1;
	drv->cdb->cdb_locate_16.partition = set_part->partition_number;

	len = set_part->logical_block_id;
	for(i = 0; i < sizeof(long long); i++) {
		sft = (sizeof(long long) - 1 - i) * 8;
		drv->cdb->cdb_locate_16.block_id[i] = (len >> sft) & 0xFF;
	} /* for */

	drv->cdb_len = CDB_GENERIC_16;
	drv->last_op = LOCATE_16;
	timeout = compute_timeout(drv->drv_id, LocateTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS){
		printk("lin_tape_set_active_partition: LOCATE_16 failed: %d\n", rc);
		goto EXIT_LABEL;
	}

	pos_buf = kmalloc(sizeof(struct read_tape_position), GFP_KERNEL);
	if(!pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(pos_buf, '\0', sizeof(struct read_tape_position));

	rc = tape_get_pos(&drv, (struct pos_info*)&pos_buf->rp_data,
		RP_LONG_FORM, True);
	if(rc != STATUS_SUCCESS)
		printk("lin_tape_set_active_partition: LOCATE_16 failed to "\
			"read_pos: %d\n", rc);

EXIT_LABEL:

	if(pos_buf) kfree(pos_buf);
	if(set_part) kfree(set_part);
	DbgPrint(("lin_tape_set_active_partition exit: %d\n", rc));
	*drvp = drv;
	return rc;
} /* lin_tape_set_active_partition() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_allow_data_overwrite                                         *
*                                                                              *
*  PURPOSE: allow for writing and partition create when in data safe mode      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive for which we want to allow overwrite   *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_allow_data_overwrite(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, i = 0, sft = 0, timeout = 0;
	unsigned long long id = 0;
	struct allow_data_overwrite* allow_wrt = NULL;

	DbgPrint(("lin_tape_allow_data_overwrite entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct allow_data_overwrite)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	allow_wrt = kmalloc(sizeof(struct allow_data_overwrite), GFP_KERNEL);
	if(!allow_wrt) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;

	if(copy_from_user((void*)allow_wrt, (void*)arg,
		sizeof(struct allow_data_overwrite))) {
		printk("lin_tape_allow_data_overwrite: can't copy from user\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	drv->cdb->cdb_allow_data_overwrite.op_code = ALLOW_DATA_OVERWRITE;
	drv->last_op = ALLOW_DATA_OVERWRITE;
	drv->cdb_len = CDB_GENERIC_16;
	switch(allow_wrt->allow_format_overwrite) {
	case 0:
		drv->cdb->cdb_allow_data_overwrite.allow_overwrite = 0x01;
		drv->cdb->cdb_allow_data_overwrite.partition =
			allow_wrt->partition_number;
		id = allow_wrt->logical_block_id;
		for(i = 0; i < sizeof(long long); i++) {
			sft = (sizeof(long long) - 1 - i) * 8;
			drv->cdb->cdb_allow_data_overwrite.logical_id[i] =
				(id >> sft) & 0xFF;
		} /* for */

		break;
	case 1:
		drv->cdb->cdb_allow_data_overwrite.allow_overwrite = 0x02;
		break;
	default:
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* switch */

	timeout = compute_timeout(drv->drv_id, AllowOverwriteTimeout);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	DbgPrint(("lin_tape_allow_data_overwrite exit: %d\n", rc));
	return rc;

} /* lin_tape_allow_data_overwrite() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_request_sense                                                    *
*                                                                              *
*  PURPOSE: get drive sense data                                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    sense_len: (output) amount of sense data returned                         *
*    resid: (output) residual ID field from sense data                         *
*    sense_buf: (output) buffer into which to copy the return data             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_request_sense(struct tape_descriptor** drvp, int* sense_len,
	int* resid, unchar* sense_buf)
{
	struct tape_sense_data* buffer = NULL;
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_request_sense entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	buffer = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
	if(!buffer) {
		printk("lin_tape: tape_request_sense: no sense data pointer\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buffer, '\0', sizeof(struct tape_sense_data));

	drv->buffer = (void*)buffer;
	drv->buf_len = sizeof(struct tape_sense_data);

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_req_sense.op_code = REQUEST_SENSE;
	drv->cdb->cdb_req_sense.alloc_length = sizeof(struct tape_sense_data);
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = REQUEST_SENSE;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, ModeSenseTimeOut),
		DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: tape_request_sense: cannot request sense\n");
		goto EXIT_LABEL;
	} /* if */

	buffer = drv->buffer;
	if(!buffer) {
		printk("lin_tape: tape_request_sense: cannot receive sense\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int)(drv->buf_len),
			(int)(FullTraceUnitSize - 7)), drv->buffer);

	if(buffer->valid) {
		*resid = (uint) ((buffer->information[0] << 24)
			+ (buffer->information[1] << 16)
			+ (buffer->information[2] << 8)
			+ buffer->information[3]);
	} else *resid = 0;
	*sense_len = (int) buffer->additional_sense_len + 8;

	memcpy(sense_buf, buffer, sizeof(struct tape_sense_data));

EXIT_LABEL:

	*drvp = drv;
	if(buffer) kfree(buffer);
	drv->buffer = NULL;
	drv->buf_len = 0;

	return rc;
} /* tape_request_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_reserve                                                          *
*                                                                              *
*  PURPOSE: Issue SCSI-2 RESERVE, opcode 0x16                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_reserve(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_reserve entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_reserve.op_code = RESERVE;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = RESERVE;
	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, ReserveTimeOut), DMA_NONE, 0,
		False);
EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* tape_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_reset_drive                                                      *
*                                                                              *
*  PURPOSE: Reset drive using send diagnostic command; keep current settings   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_reset_drive(struct tape_descriptor** drvp)
{
	struct diag_page *page = NULL;
	char ttype = lin_tape_trace_type;
	int rc;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_reset_drive entry\n"));

	if(drv->auto_logsense && ttype != LIN_TAPE_INVALID_TRACE) {
		WRITE_LOCK(&drv->logsense_lock);
		if(drv->logsense == NULL)
			drv->logsense = kmalloc(MaxLogSense, GFP_KERNEL);

		if(drv->logsense) {
			int erp_rc = 0;

			drv->logsense_len = 0;
			WRITE_UNLOCK(&drv->logsense_lock);

			drv->in_erp = True;
			erp_rc = tape_full_logsense(&drv);
			drv->in_erp = False;
			if(erp_rc != STATUS_SUCCESS) {
				DbgPrint(("lin_tape: %s log sense failed "\
					"error: %d\n", drv->devid, erp_rc));
				WRITE_LOCK(&drv->logsense_lock);
				kfree(drv->logsense);
				drv->logsense = NULL;
				drv->logsense_len = 0;
				WRITE_UNLOCK(&drv->logsense_lock);
			} /* if */
		} else WRITE_UNLOCK(&drv->logsense_lock);
	} /* if */

	page = kmalloc(sizeof(struct diag_page), GFP_KERNEL);
	if(!page) {
		printk("lin_tape: reset drive: cannot allocate data buffer\n");
		return -ENOMEM;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_send_diagnostic.op_code = SEND_DIAGNOSTIC;
	drv->cdb->cdb_send_diagnostic.pf = 1;
	drv->cdb->cdb_send_diagnostic.device_offline = 1;
	drv->cdb->cdb_send_diagnostic.parm_list_length[0] = 0x00;
	drv->cdb->cdb_send_diagnostic.parm_list_length[1] = 0x08;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = SEND_DIAGNOSTIC;

	memset(page, '\0', sizeof(struct diag_page));
	page->page_code = ResetPageCode;
	page->page_length[0] = False;
	page->page_length[1] = ResetPageLength;
	page->diag_id[0] = ResetDiagID1;
	page->diag_id[1] = ResetDiagID2;
	page->flags = False;

	drv->buffer = (void*)page;
	drv->buf_len = sizeof(struct diag_page);

	drv->pos_valid = False;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, SendDiagnosticTimeOut),
		DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	*drvp = drv;
	if(page) kfree(page);
	drv->buffer = NULL;
	drv->buf_len = 0;

	return rc;
} /* tape_reset_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_rewind                                                           *
*                                                                              *
*  PURPOSE: Issue drive rewind                                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    immediate: return immediately even if rewind hasn't completed             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_rewind(struct tape_descriptor** drvp, unchar immediate)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_rewind entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_rewind.op_code = REWIND;
	drv->cdb->cdb_rewind.immediate = immediate;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = REWIND;
	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv,
		compute_timeout(drv->drv_id, RewindTimeOut), DMA_NONE, 0,
		False);

	if(rc == STATUS_SUCCESS) {
		drv->eom = False;
		drv->deferred_eom = False;
		drv->eom_blk_id = 0;
		drv->pos_valid = True;
		drv->curr_pos = 0;
	} else drv->pos_valid = False;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_rewind() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_self_test                                                        *
*                                                                              *
*  PURPOSE: self test of drive using SEND_DIAGNOSTIC command                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_self_test(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_self_test entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_send_diagnostic.op_code = SEND_DIAGNOSTIC;
	drv->cdb->cdb_send_diagnostic.pf = True;
	drv->cdb->cdb_send_diagnostic.self_test = True;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = SEND_DIAGNOSTIC;
	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		SendDiagnosticTimeOut), DMA_NONE, 0, False);

EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* tape_self_test() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_all_parms                                                    *
*                                                                              *
*  PURPOSE: set all of the drive parameters based on settings in "drv"         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_all_parms(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_set_all_parms entry\n"));
	rc = tape_set_media_parm(&drv, AllModifications, drv->buffered_mode,
		drv->density, drv->blk_size);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s cannot set medium settings\n", drv->devid);
		goto EXIT_LABEL;
	} /* if */

	rc = tape_set_compression(&drv, drv->compression);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s cannot change compression\n", drv->devid);
		goto EXIT_LABEL;
	} /* if */

	rc = tape_set_rw_error_recovery_page(&drv, drv->limit_read_recov,
		drv->limit_write_recov);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s can't set recovery t'outs\n", drv->devid);
		goto EXIT_LABEL;
	} /* if */

	if(drv->drv_id != IBM_3590 && drv->drv_id != IBM_3592) goto EXIT_LABEL;

	rc = tape_tur(&drv);

	if(rc != STATUS_SUCCESS) {
		rc = STATUS_SUCCESS;
		if(drv->drv_id == IBM_3592) tape_set_tod(&drv);
		goto EXIT_LABEL;
	} /* if */

	rc = tape_set_medium_sense(&drv, drv->logic_write_protect);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: logic_write_protect %d\n",
			drv->logic_write_protect);
		DbgPrint(("lin_tape: %s cannot set wrt-protect\n", drv->devid));
		goto EXIT_LABEL;
	} /* if */
	if(drv->drv_id == IBM_3592) tape_set_tod(&drv);

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_set_all_parms() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_medium_sense                                                 *
*                                                                              *
*  PURPOSE: set logic write protect on/off                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    logic_write_protect: value to use for type of write protection            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_medium_sense(struct tape_descriptor** drvp,
	unchar logic_write_protect)
{
	struct medium_sense* mode_buf = NULL;
	int rc = STATUS_SUCCESS, buf_len = 0;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_set_medium_sense entry\n"));

	if(drv->in_erp) goto EXIT_LABEL;

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buf = kmalloc(sizeof(struct medium_sense), GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', sizeof(struct medium_sense));

	rc = tape_modesense_page(&drv, MediumSensePageNumber,
		(unchar*)mode_buf, sizeof(struct medium_sense), True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->buffer = NULL;
	mode_buf->header.blk_descriptor_length = False;

	switch(logic_write_protect) {
	case NO_PROTECT:
		if(drv->cmd_in_open && mode_buf->page.physical_wp)
			break;
		mode_buf->page.associated_wp = False;
		mode_buf->page.persistent_wp = False;
		mode_buf->page.reset_associate_wp = True;
		mode_buf->page.reset_persistent_wp = True;
		break;
	case ASSOCIATED_PROTECT:
		mode_buf->page.associated_wp = True;
		break;
	case PERSISTENT_PROTECT:
		mode_buf->page.persistent_wp = True;
		break;
	case WORM_PROTECT:
		mode_buf->page.permanent_wp = True;
		break;
	default:
		rc = -EINVAL;
		goto EXIT_LABEL;
		break;
	} /* switch */

	buf_len = mode_buf->page.page_length + 2 +
		sizeof(struct mode_parm_header);
	drv->buffer = kmalloc(buf_len, GFP_KERNEL);
	memset(drv->buffer, '\0', buf_len);
	memcpy(drv->buffer, mode_buf, min((int)buf_len,
		(int)sizeof(struct medium_sense)));
	kfree(mode_buf);
	mode_buf = drv->buffer;
	drv->buffer = NULL;

	rc = tape_modeselect(&drv, (unchar*)mode_buf, buf_len);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:
	*drvp = drv;

	if(drv && !drv->in_erp) {
		if(drv->buffer) kfree(drv->buffer);
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	if(mode_buf) kfree(mode_buf);

	return rc;
} /* tape_set_medium_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_media_parm                                                   *
*                                                                              *
*  PURPOSE: set medium parameters in drive                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    modification: which of the parameters to change (block size, etc.)        *
*    buffered_mode: boolean describing whether buffered mode is to be set      *
*    density_code: tape density, if it is to be set                            *
*    blk_size: block size, if it is to be set                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_media_parm(struct tape_descriptor** drvp, unchar modification,
	unchar buffered_mode, unchar density_code, ulong blk_size)
{
	struct media_parm* mode_buf = NULL;
	ulong tmp_blk_size = 0;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_set_media_parm entry\n"));

	if(modification != DensityCodeOnly && modification != BufferModeOnly &&
		blk_size != 0 && (blk_size > drv->max_blk_size ||
		blk_size < drv->min_blk_size || blk_size > drv->max_xfer_len)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buf = kmalloc(sizeof(struct media_parm), GFP_KERNEL);
	if(mode_buf == NULL) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', sizeof(struct media_parm));

	rc = tape_modesense_page(&drv, DeviceConfigurationPageNumber,
		(unchar*)mode_buf, sizeof(struct media_parm), False, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* check if sili is ON.  If it is ON and drv->blk_size is 0,
	* we want to set blk len field in mode sense header to be non-zero (4)
	* in order to force a check condition on overlength read
	*/
	tmp_blk_size = (drv->sili && !blk_size) ? 4 : blk_size;

	switch(modification) {
	case BlockSizeOnly:
		mode_buf->block.blk_length[0] =
			(unchar)(tmp_blk_size >> 16 & 0xFF);
		mode_buf->block.blk_length[1] =
			(unchar)(tmp_blk_size >> 8 & 0xFF);
		mode_buf->block.blk_length[2] = (unchar)(tmp_blk_size & 0xFF);
		break;
	case BufferModeOnly:
		mode_buf->header.buffered_mode = buffered_mode ? True : False;
		break;
	case DensityCodeOnly:
		mode_buf->block.density_code = density_code;
		break;
	default:
		mode_buf->block.density_code = density_code;
		mode_buf->block.blk_length[0] =
			(unchar)(tmp_blk_size >> 16 & 0xFF);
		mode_buf->block.blk_length[1] =
			(unchar)(tmp_blk_size >> 8 & 0xFF);
		mode_buf->block.blk_length[2] = (unchar)(tmp_blk_size & 0xFF);
		mode_buf->header.buffered_mode = buffered_mode ? True : False;
		break;
	} /* switch */

	rc = tape_modeselect(&drv, (unchar*)mode_buf,
		sizeof(struct media_parm));
	if(rc == STATUS_SUCCESS && modification != BufferModeOnly)
		drv->blk_size = blk_size;

EXIT_LABEL:

	*drvp = drv;
	if(drv) {
		drv->buf_len = 0;
		drv->buffer = NULL;
	} /* if */
	if(mode_buf) kfree(mode_buf);
	return rc;
} /* tape_set_media_parm() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_tod                                                          *
*                                                                              *
*  PURPOSE: set time of day in drive                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_tod(struct tape_descriptor** drvp)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	struct timeval* tv = NULL;
#else
    struct timespec64 timespec;
#endif // LINUX_VERSION_CODE
	struct tod_control* buffer = NULL;
	struct tape_descriptor *drv = *drvp;
	int now = 0, rc = STATUS_SUCCESS;

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	buffer = kmalloc(sizeof(struct tod_control), GFP_KERNEL);
	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	tv = kmalloc(sizeof(struct timeval), GFP_KERNEL);
	if(!tv) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
#endif // LINUX_VERSION_CODE
	memset(buffer, '\0', sizeof(struct tod_control));

	rc = tape_modesense_page(&drv, TODControlPageNumber, (unchar*)buffer,
		sizeof(struct tod_control), True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	memset(tv, '\0', sizeof(struct timeval));
	do_gettimeofday(tv);
	/* tv.sec in structure timeval is a "long" integer
	* it has 4 bytes in 32-bit kernel, but 8 bytes in 64-bit kernel
	* we truncate it to 4 byte integer
	*/
	now = (int)tv->tv_sec;
#else
	memset(&timespec, '\0', sizeof(struct timespec64));
    ktime_get_real_ts64(&timespec);
	now = (int)timespec.tv_sec;
#endif // LINUX_VERSION_CODE

#if defined(__LITTLE_ENDIAN)
	buffer->page[2] = ((char*)&now)[3];
	buffer->page[3] = ((char*)&now)[2];
	buffer->page[4] = ((char*)&now)[1];
	buffer->page[5] = ((char*)&now)[0];
#elif defined(__BIG_ENDIAN)
	buffer->page[2] = ((char*)&now)[0];
	buffer->page[3] = ((char*)&now)[1];
	buffer->page[4] = ((char*)&now)[2];
	buffer->page[5] = ((char*)&now)[3];
#endif

	buffer->page[10] |= 0x04;

	rc = tape_modeselect(&drv, (unchar*)buffer, sizeof(struct tod_control));

EXIT_LABEL:

	*drvp = drv;
	if(drv) {
		if(drv->buffer) kfree(drv->buffer);
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	if(buffer) kfree(buffer);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	if(tv) kfree(tv);
#endif


	return rc;
} /* tape_set_tod() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_rw_error_recovery_page                                       *
*                                                                              *
*  PURPOSE: set read/write error recovery time limit                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    limit_read_recov: time limit for read                                     *
*    limit_write_recov: time limite for write                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_rw_error_recovery_page(struct tape_descriptor** drvp,
	unchar limit_read_recov, unchar limit_write_recov)
{
	struct rw_err_recovery* buffer = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	buffer = kmalloc(sizeof(struct rw_err_recovery), GFP_KERNEL);
	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buffer, '\0', sizeof(struct rw_err_recovery));

	rc = tape_modesense_page(&drv, RWErrorRecoveryPageNumber,
		(unchar*)buffer, sizeof(struct rw_err_recovery), True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if((limit_read_recov == (buffer->page[3] == 0x05)) &&
		(limit_write_recov == (buffer->page[8] == 0x05))) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	buffer->page[3] = limit_read_recov ? 0x05 : 0xff;
	buffer->page[8] = limit_write_recov ? 0x05 : 0xff;
	rc = tape_modeselect(&drv, (unchar*)buffer,
		sizeof(struct rw_err_recovery));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	*drvp = drv;

	if(drv && drv->buffer) {
		kfree(drv->buffer);
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	return rc;
} /* tape_set_rw_error_recovery_page() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_density                                                      *
*                                                                              *
*  PURPOSE: Get density code for 3592 drives                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_density(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct rw_control *buffer = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;
	struct density_data_t* density_info = (struct density_data_t*)buf;

	if(!drv || !density_info || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(strncmp(drv->dev_obj->model, "03592", strlen("03592"))
		|| drv_gen(drv) == 1) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* issue mode sense to get the current settings */
	buffer = kmalloc(sizeof(struct rw_control), GFP_KERNEL);
	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buffer, '\0', sizeof(struct rw_control));

	rc = tape_modesense_page(&drv, RWControlPageNumber, (unchar*)buffer,
		sizeof(struct rw_control), False, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	density_info->density_code = (unchar)buffer->block.density_code;
	density_info->default_density = (unchar)buffer->page.default_density;
	density_info->pending_density = (unchar)buffer->page.pending_density;

EXIT_LABEL:

	*drvp = drv;
	if(buffer) kfree(buffer);
	buffer = NULL;
	if(drv) {
		drv->buf_len = 0;
		drv->buffer = NULL;
	} /* if */
	return rc;
} /* tape_get_density() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_density                                                      *
*                                                                              *
*  PURPOSE: Set density code for 3592 drives                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_density(struct tape_descriptor** drvp, void* buf, char unused)
{
	int old_density_valid, old_default_density, old_pending_density;
	struct rw_control *buffer = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;
	struct density_data_t* density_info = (struct density_data_t*)buf;

	if(!drv || !density_info || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(strncmp(drv->dev_obj->model, "03592", strlen("03592"))
		|| drv_gen(drv) == 1) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* save the current density in case of problems */
	old_density_valid = drv->set_density_valid;
	old_default_density = drv->set_default_density;
	old_pending_density = drv->set_pending_density;

	/* issue mode sense to get current settings */
	buffer = kmalloc(sizeof(struct rw_control), GFP_KERNEL);
	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buffer, '\0', sizeof(struct rw_control));
	rc = tape_modesense_page(&drv, RWControlPageNumber, (unchar*)buffer,
		sizeof(struct rw_control), False, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* modify the change requested */
	drv->set_density_valid = True;
	drv->set_default_density = density_info->default_density;
	drv->set_pending_density = density_info->pending_density;
	buffer->block.density_code = 0x7F;
	buffer->page.default_density = density_info->default_density;
	buffer->page.pending_density = density_info->pending_density;

	rc = tape_modeselect(&drv, (unchar*)buffer, sizeof(struct rw_control));

	/* if mode select fails, restore previous settings */
	if(rc != STATUS_SUCCESS){
		drv->set_density_valid = old_density_valid;
		drv->set_default_density = old_default_density;
		drv->set_pending_density = old_pending_density;
		buffer->block.density_code = 0x7F;

		if(old_density_valid){
			buffer->page.default_density = old_default_density;
			buffer->page.pending_density = old_pending_density;
		} else {
			buffer->page.default_density = 0;
			buffer->page.pending_density = 0;
		} /* else */

		drv->in_erp = True;
		rc = tape_modeselect(&drv, (unchar*)buffer,
			sizeof(struct rw_control));
		drv->in_erp = False;
	} /* if */

EXIT_LABEL:

	if(buffer) kfree(buffer);
	if(drv) {
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */

	return rc;
} /* tape_set_density() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_encryption_capability                                        *
*                                                                              *
*  PURPOSE: Find out whether drive is capable of doing encryption              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_encryption_capability(struct tape_descriptor** drvp)
{
	int rc = -EIO;
	struct generic_mode_page_10* mode_buffer = NULL;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_get_encryption_capability entry.\n"));

	if(!drv) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer = kmalloc(sizeof(struct generic_mode_page_10), GFP_KERNEL);
	if(!mode_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buffer, '\0', sizeof(struct generic_mode_page_10));

	rc = tape_modesense10_page(&drv, InitiatorSpecPageNumber,
		(unchar*)mode_buffer, sizeof(struct generic_mode_page_10), 0,
		True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->encryption_capable = (mode_buffer->mode_page[7]) & 0x01;

EXIT_LABEL:

	*drvp = drv;
	if(mode_buffer) kfree(mode_buffer);
	DbgPrint(("lin_tape: tape_get_encryption_capability exit: %d\n", rc));
	return rc;
} /* tape_get_encryption_capability() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_encryption_enablement                                        *
*                                                                              *
*  PURPOSE: Obtain encryption state for SME                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_encryption_enablement(struct tape_descriptor** drvp)
{
	int rc = -EIO;
	unchar* mode_buffer = NULL;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_get_encryption_enablement entry.\n"));

	if(!drv) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* mode sense to get current settings */
	if(!drv->cdb) {
		drv->encryption_state = STATE_UNKNOWN;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer = kmalloc(sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page), GFP_KERNEL);
	if(!mode_buffer) {
		drv->encryption_state = STATE_UNKNOWN;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buffer, '\0', sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page));

	rc = tape_modesense10_page(&drv, RWControlPageNumber, mode_buffer,
		sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page), 0, True, True);
	if(rc != STATUS_SUCCESS) {
		drv->encryption_state = STATE_UNKNOWN;
		goto EXIT_LABEL;
	} /* if */

	drv->encryption_state = mode_buffer[28] & 0x03;
	drv->encryption_keypath_configured =
		(((mode_buffer[29] & 0xE0) == 0x20 &&
		(mode_buffer[29] & 0x1C) == 0x00) ||
		(mode_buffer[29] & 0x1C) == 0x04);

EXIT_LABEL:

	if(mode_buffer) kfree(mode_buffer);
	if(drv) {
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	*drvp = drv;

	return rc;
} /* tape_get_encryption_enablement() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_encryption_state                                             *
*                                                                              *
*  PURPOSE: Obtain encryption state for SME                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    status: the structure containing the state to get                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_encryption_state(struct tape_descriptor** drvp,
	struct encryption_status* status)
{
	int rc = -EIO;
	unchar* mode_buffer = NULL;
	unchar encryption_method = 0;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_get_encryption_state entry.\n"));

	if(!drv || !status || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer = kmalloc(sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page), GFP_KERNEL);
	if(!mode_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buffer, '\0', sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page));

	rc = tape_modesense10_page(&drv, RWControlPageNumber, mode_buffer,
		sizeof(struct generic_mode_page_10) +
		sizeof(struct rw_control_page), 0, True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	encryption_method = (unchar)mode_buffer[35];
	switch(encryption_method) {
	case 0:
		status->encryption_method = METHOD_NONE;
		status->encryption_state = STATE_OFF;
		break;
	case 0x10:
		status->encryption_method = METHOD_SYSTEM;
		status->encryption_state =
		(drv->sys_encryption_write && drv->encryption_state) ?
			STATE_ON : STATE_OFF;
		break;
	case 0x50:
		status->encryption_method = METHOD_APPLICATION;
		status->encryption_state =
		(drv->application_set_encryption && drv->encryption_state) ?
			STATE_ON : STATE_OFF;
		break;
	case 0x60:
		status->encryption_method = METHOD_LIBRARY;
		status->encryption_state = STATE_NA;
		break;
	case 0xFF:
		status->encryption_method = METHOD_CUSTOM;
		status->encryption_state = STATE_NA;
		break;
	default:
		status->encryption_method = METHOD_UNKNOWN;
		status->encryption_state = STATE_NA;
		break;
	} /* switch */

EXIT_LABEL:

	if(mode_buffer) kfree(mode_buffer);
	if(drv) {
		drv->buffer = NULL;
		drv->buf_len = 0;
	} /* if */
	*drvp = drv;

	return rc;
} /* tape_get_encryption_state() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_encryption_state                                             *
*                                                                              *
*  PURPOSE: Set the encryption state on the drive                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    status: the structure containing the state to set                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_encryption_state(struct tape_descriptor** drvp,
	struct encryption_status* status)
{
	int rc = -EIO, alloc_len = 0;
	unchar* mode_buffer = NULL;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_set_encryption_state entry.\n"));

	if(!drv || !status)
		return -ENOMEM;

	if(status->encryption_state != STATE_OFF
		&& status->encryption_state != STATE_ON) {
		printk("Invalid encryption state setting: %d\n",
			status->encryption_state);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	alloc_len = sizeof(struct mode_parm_header_10)
		+ sizeof(struct rw_control_page);
	mode_buffer = kmalloc(alloc_len, GFP_KERNEL);
	if(!mode_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buffer, '\0', alloc_len);

	rc = tape_modesense10_page(&drv, RWControlPageNumber, mode_buffer,
		alloc_len, 0, True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* issue mode select command */
	mode_buffer[26] |= 0x01;
	if(status->encryption_state == STATE_OFF){
		mode_buffer[28] = (unchar)(mode_buffer[28] & 0xFC);
	} else {

		mode_buffer[28] = (unchar)((mode_buffer[28] & 0x8C) | 0x01);
		if(drv->drv_id == IBM_3592) mode_buffer[30] = 0x02;
	} /* else */

	rc = tape_modeselect10(&drv, mode_buffer, False, alloc_len);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	if(mode_buffer) kfree(mode_buffer);
	*drvp = drv;

	DbgPrint(("lin_tape: tape_set_encryption_state exit: %d\n", rc));
	return rc;
} /* tape_set_encryption_state() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_data_key                                                     *
*                                                                              *
*  PURPOSE: set the encryption data key for a drive                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    data_key: the key to set                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_data_key(struct tape_descriptor** drvp,
	struct data_key* data_key)
{
	int rc = -EIO;
	struct generic_mode_page_10* mode_buffer = NULL;
	int sub_page_length = 0;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_set_data_key entry.\n"));

	if(!drv || !data_key || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer = kmalloc(sizeof(struct generic_mode_page_10), GFP_KERNEL);
	if(!mode_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(mode_buffer, '\0', sizeof(struct generic_mode_page_10));

	rc = tape_modesense10_page(&drv, RWControlPageNumber,
		(void*)mode_buffer, sizeof(struct generic_mode_page_10),
		0xC0, True, True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	sub_page_length = (data_key->data_key_index_length) ? 144 : 132;

	mode_buffer->header.mode_data_length[0] = 0;
	mode_buffer->header.mode_data_length[1] = 0;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	drv->cdb->cdb_modeselect_10.pf = True;
	drv->cdb->cdb_modeselect_10.parm_list_length[0] = 0;
	drv->cdb->cdb_modeselect_10.parm_list_length[1] =
		sizeof(struct mode_parm_header_10) + sub_page_length;
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = MODE_SELECT_10;
	drv->buf_len = sizeof(struct mode_parm_header_10) + sub_page_length;
	drv->buffer = (void*)mode_buffer;

	mode_buffer->mode_page[0] = 0x65;
	mode_buffer->mode_page[1] = 0xE0;
	mode_buffer->mode_page[2] = 0;
	mode_buffer->mode_page[3] = sub_page_length - 4;

	mode_buffer->mode_page[4] = 0;
	mode_buffer->mode_page[5] = 0;
	mode_buffer->mode_page[6] = 0;
	mode_buffer->mode_page[7] = sub_page_length - 8;

	mode_buffer->mode_page[8] = 0;
	mode_buffer->mode_page[9] = 0;
	mode_buffer->mode_page[10] = 0;
	mode_buffer->mode_page[11] = sub_page_length - 16;

	mode_buffer->mode_page[68] = 0x21;
	mode_buffer->mode_page[69] = 0xE0;

	mode_buffer->mode_page[70] = 0;
	mode_buffer->mode_page[71] = sub_page_length - 76;

	mode_buffer->mode_page[72] &= 0x78;
	mode_buffer->mode_page[73] = 0x00;
	mode_buffer->mode_page[74] = 0x00;
	mode_buffer->mode_page[75] = 0x01;
	mode_buffer->mode_page[76] = 0x11;
	mode_buffer->mode_page[77] = 0x00;
	mode_buffer->mode_page[78] = 0x00;
	mode_buffer->mode_page[79] = 0x20;

	memcpy(&mode_buffer->mode_page[80], data_key->data_key, 32);
	mode_buffer->mode_page[112] = 0x18;
	mode_buffer->mode_page[113] = 0x00;
	mode_buffer->mode_page[114] = 0x00;
	mode_buffer->mode_page[115] = (data_key->data_key_index_length) ?
		0x14 : 0x08;
	mode_buffer->mode_page[116] = 0x0;
	mode_buffer->mode_page[117] = 0x02;
	mode_buffer->mode_page[118] = 0x00;
	mode_buffer->mode_page[119] = 0x00;
	mode_buffer->mode_page[120] = 0x1A;
	mode_buffer->mode_page[121] = 0x00;
	mode_buffer->mode_page[122] = 0x00;
	if(data_key->data_key_index_length) {
		mode_buffer->mode_page[123] = 0x0C;

		memcpy(&mode_buffer->mode_page[124],
			data_key->data_key_index,
			data_key->data_key_index_length);
		mode_buffer->mode_page[136] = 0x00;
		mode_buffer->mode_page[137] = 0x00;
		mode_buffer->mode_page[138] = 0x00;
		mode_buffer->mode_page[139] = 0x00;

		mode_buffer->mode_page[140] = 0x04;
		mode_buffer->mode_page[141] = 0x00;
		mode_buffer->mode_page[142] = 0x00;
		mode_buffer->mode_page[143] = 0x00;
	} else {
		mode_buffer->mode_page[123] = 0x00;
		mode_buffer->mode_page[124] = 0x00;
		mode_buffer->mode_page[125] = 0x00;
		mode_buffer->mode_page[126] = 0x00;
		mode_buffer->mode_page[127] = 0x00;

		mode_buffer->mode_page[128] = 0x04;
		mode_buffer->mode_page[129] = 0x00;
		mode_buffer->mode_page[130] = 0x00;
		mode_buffer->mode_page[131] = 0x00;
	} /* else */

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		ModeSelectTimeOut), DMA_TO_DEVICE, 0, False);
	if(rc != STATUS_SUCCESS)
		DbgPrint(("lin_tape: set data key failed, rc %d\n", rc));

	drv->buf_len = 0;
	drv->buffer = NULL;


EXIT_LABEL:

	*drvp = drv;
	if(mode_buffer) kfree(mode_buffer);
	return rc;
} /* tape_set_data_key() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_discard_data_key                                                 *
*                                                                              *
*  PURPOSE: discard the encryption data key                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_discard_data_key(struct tape_descriptor** drvp)
{
	int rc = -EIO;
	struct generic_mode_page_10* mode_buffer = NULL;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_discard_data_key entry.\n"));

	if(!drv || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer = kmalloc(sizeof(struct generic_mode_page_10), GFP_KERNEL);
	if(!mode_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_modesense10_page(&drv, RWControlPageNumber,
		(unchar*)mode_buffer, sizeof(struct generic_mode_page_10), 0,
		True, True);

	if(rc == STATUS_SUCCESS && (mode_buffer->mode_page[22] & 0x1F)) {
		mode_buffer->mode_page[22] &= 0xE0;
		rc = tape_modeselect10(&drv, (unchar*)mode_buffer, False,
			sizeof(struct mode_parm_header_10) +
			sizeof(struct rw_control_page));
		if(rc != STATUS_SUCCESS)
			DbgPrint(("lin_tape: discard key failed: %d\n", rc));
	} /* if */

EXIT_LABEL:

	if(mode_buffer) kfree(mode_buffer);
	if(drv) {
		drv->buf_len = 0;
		drv->buffer = NULL;
	} /* if */
	*drvp = drv;

	return rc;
} /* tape_discard_data_key() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_pos                                                          *
*                                                                              *
*  PURPOSE: set tape position to the specified block id                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    method: Space or Locate                                                   *
*    count: number of objects on tape to traverse                              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_pos(struct tape_descriptor ** drvp, int method,
	long long count)
{
	int rc = STATUS_SUCCESS, timeout = 0, alloc_size = 0;
	boolean ex = False;
	struct tape_descriptor *drv = *drvp;
	struct pos_info* read_pos_buf = NULL;
	struct read_tape_position *read_pos_ex = NULL;
	void** read_pos_gen = NULL;

	DbgPrint(("lin_tape: tape_set_pos entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	switch(method) {
	case SpaceFileMark:
	case SpaceRecord:
	case SpaceSequentialFM:
	case SpaceEOM:
		drv->cdb->cdb_space.op_code = SPACE;
		drv->cdb->cdb_space.code =
			(method == SpaceFileMark) ? 0x1 :
			(method == SpaceSequentialFM) ? 0x2 :
			(method == SpaceEOM) ? 0x3 : 0x0;
		drv->cdb_len = CDB_GENERIC_6;
		drv->last_op = SPACE;
		if(method == SpaceEOM) break;
		drv->cdb->cdb_space.count[0] = (char)(count >> 16 & 0xFF);
		drv->cdb->cdb_space.count[1] = (char)(count >> 8 & 0xFF);
		drv->cdb->cdb_space.count[2] = (char)(count & 0xFF);
		break;

	case LocateBlock:
	case LocateBlockImmediate:
		drv->cdb->cdb_locate.op_code = LOCATE;
		drv->cdb->cdb_locate.immediate =
			(method == LocateBlockImmediate) ? 0x1 : 0x0;
		drv->cdb->cdb_locate.blk_addr[0] = (char)(count >> 24 & 0xFF);
		drv->cdb->cdb_locate.blk_addr[1] = (char)(count >> 16 & 0xFF);
		drv->cdb->cdb_locate.blk_addr[2] = (char)(count >> 8 & 0xFF);
		drv->cdb->cdb_locate.blk_addr[3] = (char)(count & 0xFF);
		drv->cdb_len = CDB_GENERIC_10;
		drv->last_op = LOCATE;
		break;

	case Locate16Block:
	case Locate16BlockImmediate:
	case Locate16File:
	case Locate16FileImmediate:
		drv->cdb->cdb_locate_16.op_code = LOCATE_16;
		drv->cdb->cdb_locate_16.immediate =
			(method == Locate16BlockImmediate ||
			method == Locate16FileImmediate) ? 0x1 : 0x0;
		drv->cdb->cdb_locate_16.dest_type =
			(method == Locate16Block ||
			method == Locate16BlockImmediate) ? 0x0 : 0x1;

		drv->cdb->cdb_locate_16.block_id[0] = (count >> 56 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[1] = (count >> 48 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[2] = (count >> 40 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[3] = (count >> 32 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[4] = (count >> 24 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[5] = (count >> 16 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[6] = (count >> 8 & 0xFF);
		drv->cdb->cdb_locate_16.block_id[7] = (count & 0xFF);
		break;
	default:
		printk("lin_tape: tape_set_pos unknown set position method\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* switch */

	drv->buffer = NULL;
	drv->buf_len = 0;
	timeout = compute_timeout(drv->drv_id, SpaceTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_NONE, 0, False);

	/* successful tape movement clears the end of medium flag */
	if(rc == STATUS_SUCCESS) {
		if((drv->drv_id == IBM_3580 && drv_gen(drv) >= 6) ||
			(drv->drv_id == IBM_3592 && drv_gen(drv) >= 4)) {
			ex = True;
			alloc_size = sizeof(struct read_tape_position);
			read_pos_gen = (void**)&read_pos_ex;
		} else {
			ex = False;
			alloc_size = sizeof(struct pos_info);
			read_pos_gen = (void**)&read_pos_buf;
		} /* else */

		*read_pos_gen = kmalloc(alloc_size, GFP_KERNEL);
		if(!*read_pos_gen) {
			drv->pos_valid = False;
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(*read_pos_gen, '\0', alloc_size);
		if(ex) read_pos_buf = (struct pos_info*)&read_pos_ex->rp_data;

//		tape_wfm(&drv, False, False);
		rc = tape_get_pos(&drv, read_pos_buf,
			ex ? RP_LONG_FORM : RP_SHORT_FORM, ex);

		if(rc == STATUS_SUCCESS && drv->eom_blk_id) {
			if(drv->curr_pos < drv->eom_blk_id) {
				drv->deferred_eom = False;
				drv->eom_blk_id = 0;
			} /* if */
		} /* if */
		drv->eom = False;

		if(*read_pos_gen) {
			kfree(*read_pos_gen);
			*read_pos_gen = NULL;
		} /* if */
	} else drv->pos_valid = False;

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_set_pos exit: %d\n", rc));
	*drvp = drv;
	return rc;
} /* tape_set_pos() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_std_inquiry                                                      *
*                                                                              *
*  PURPOSE: Retrieve inquiry p. 0x0 from drive                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_std_inquiry(struct tape_descriptor** drvp, void* buf, char unused)
{
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_std_inquiry entry\n"));

	if(!drv->cdb || !buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = (void*)buf;
	drv->buf_len = sizeof(struct st_inquiry);
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_inquiry.op_code = INQUIRY;
	drv->cdb->cdb_inquiry.alloc_length = sizeof(struct st_inquiry);
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = INQUIRY;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		InquiryTimeOut), DMA_FROM_DEVICE, 0, False);

	if(!drv->buffer) {
		printk("lin_tape: tape_std_inquiry no data returned\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(rc == STATUS_SUCCESS && drv->trace && ttype == LIN_TAPE_FULL_TRACE
		&& drv->buf_len)
		tape_log_buffer_data(drv, min((int)(((struct st_inquiry*)buf)->
			standard.additional_length+5),
			(int)(FullTraceUnitSize-7)), drv->buffer);

EXIT_LABEL:

	*drvp = drv;
	drv->buffer = NULL;
	drv->buf_len = 0;
	return rc;
} /* tape_std_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_tur                                                              *
*                                                                              *
*  PURPOSE: Test unit ready on the drive                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_tur(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_tur entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_tur.op_code = TEST_UNIT_READY;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = TEST_UNIT_READY;
	drv->buffer = NULL;
	drv->buf_len = 0;
	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		TestUnitReadyTimeOut), DMA_NONE, 0, False);

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_tur() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_update_all_parms                                                 *
*                                                                              *
*  PURPOSE: Sync the parameters in the descriptor with the actual drive        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_update_all_parms(struct tape_descriptor **drvp)
{
	int rc;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("tape_update_all_parms entry\n"));

	rc = tape_get_media_parm(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = tape_get_compression(&drv);
	if(rc != STATUS_SUCCESS) printk("lin_tape: tape_get_compression: "\
		"%d\n", rc);

	if((drv->drv_id == IBM_3580 && drv_gen(drv) >= 5) ||
		(drv->drv_id == IBM_3592 && drv_gen(drv) >= 4)) {
		rc = tape_data_safe_mode_and_pew(&drv, NULL);
		if(rc != STATUS_SUCCESS) printk("lin_tape: "\
			"tape_data_safe_mode_and_pew: %d\n", rc);
	} /* if */

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	*drvp = drv;
	DbgPrint(("tape_update_all_parms exit: %d\n", rc));
	return rc;
} /* tape_update_all_parms() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_update_block_limits                                              *
*                                                                              *
*  PURPOSE: Update block limits                                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_update_block_limits(struct tape_descriptor** drvp)
{
	struct blk_limits* blk_limit_buf = NULL;
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_update_block_limits entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	blk_limit_buf = kmalloc(sizeof(struct blk_limits), GFP_KERNEL);
	if(!blk_limit_buf) {
		printk("lin_tape: cannot allocate mode buffer\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(blk_limit_buf, '\0', sizeof(struct blk_limits));

	drv->buffer = (void*)blk_limit_buf;
	drv->buf_len = sizeof(struct blk_limits);

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_read_blk_limits.op_code = READ_BLOCK_LIMITS;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = READ_BLOCK_LIMITS;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		ReadBlockLimitsTimeOut), DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: unable to do mode sense\n");
		goto EXIT_LABEL;
	} /* if */

	blk_limit_buf = drv->buffer;
	if(!blk_limit_buf) {
		printk("lin_tape: unable to receive block limits data\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
		tape_log_buffer_data(drv, min((int) (drv->buf_len),
			(int)(FullTraceUnitSize - 7)), drv->buffer);

	drv->max_blk_size = (blk_limit_buf->max_blk_size[0] << 16)
		+ (blk_limit_buf->max_blk_size[1] << 8)
		+ blk_limit_buf->max_blk_size[2];

	drv->min_blk_size = (blk_limit_buf->min_blk_size[0] << 8)
		+ blk_limit_buf->min_blk_size[1];

EXIT_LABEL:

	*drvp = drv;

	if(blk_limit_buf) kfree(blk_limit_buf);
	drv->buffer = NULL;
	drv->buf_len = 0;

	return rc;
} /* tape_update_block_limits() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_query_rao                                                       *
*                                                                              *
*  PURPOSE: Handle QUERY_RAO command to the drive                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    buf: input/output buffer                                                  *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
ssize_t stioc_query_rao(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct tape_descriptor* drv = *drvp;
	struct query_rao_info* qri = (struct query_rao_info*)buf;
	int rc = STATUS_SUCCESS;
	unchar *uds_limits_page = NULL;

	if(!drv->cdb || !buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	uds_limits_page = kmalloc(UDS_LIMITS_PAGE, GFP_KERNEL);
	if(!uds_limits_page){
		printk("lin_tape: stioc_query_rao unable to alloc 4 bytes in kernel\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}

	memset(uds_limits_page, '\0', UDS_LIMITS_PAGE);
	drv->buffer = (void *)uds_limits_page;
	drv->buf_len = UDS_LIMITS_PAGE;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	drv->cdb_len = CDB_GENERIC_12;
	drv->cdb->cdb_rrao.op_code = MAINTENANCE_IN;
	drv->cdb->cdb_rrao.uds_limits = 1;
	drv->cdb->cdb_rrao.service_action = RAO_SERVICE_ACTION;
	drv->cdb->cdb_rrao.uds_type = (unchar)(qri->uds_type & UDS_TYPE_MASK);
	drv->cdb->cdb_rrao.alloc_length[3]= UDS_LIMITS_PAGE;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
			RAOTimeout), DMA_FROM_DEVICE, 0, False);
	qri->max_uds_number = (uds_limits_page[0]<< 8) + uds_limits_page[1];
	qri->max_uds_size = (uds_limits_page[2]<< 8) + uds_limits_page[3];
	qri->max_host_uds_number = qri->max_uds_number;

EXIT_LABEL:
	if(uds_limits_page) {
		kfree(uds_limits_page);
		drv->buffer = NULL;
		drv->buf_len = 0;
	}
	return rc;

} /* stioc_query_rao() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_generate_rao                                                    *
*                                                                              *
*  PURPOSE: send list of requests to drive in order for it to generate order   *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    buf: input/output buffer                                                  *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int stioc_generate_rao(struct tape_descriptor** drvp, ulong arg)
{
	struct tape_descriptor* drv = *drvp;
	struct generate_rao* grao = NULL;
	int rc = STATUS_SUCCESS, i = 0, remain = 0, copy_size;
	ulong alloc_size = 0, alloc_order = 0;
	uint length;
	void *tmp = NULL;
	char* user_location;
	ulong tmp_len = 0;

    #if defined(__LITTLE_ENDIAN)
	int f, p;
	struct grao_uds_desc {
	  ushort desc_leng; /* descriptor length */
	  char reserved[3];
	  char uds_name[10]; /* uds name given by application */
	  char partition; /* Partition number 0-n to overwrite */
	  unsigned long long beginning_loi; /* Beginning logical object ID */
	  unsigned long long ending_loi; /* Ending logical object ID */
	}uds;
	struct grao_list_header{
      unchar reserved[4];
      char addl_data[4]; /* additional data */
	}dl;
    #endif
	int grao_strct_size = sizeof(struct generate_rao);

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	grao = kmalloc( grao_strct_size, GFP_KERNEL );

	if(!grao){
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(grao, '\0', grao_strct_size);

	if(copy_from_user((char*)grao, (char*)arg, grao_strct_size)){
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	length = grao->grao_list_leng;
	user_location = grao->grao_list;

	if(user_location == NULL){
		printk("lin_tape: passed pointer for grao is NULL\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	}
	#if defined(__LITTLE_ENDIAN)
	    memcpy ((unchar *)&dl, (unchar *)user_location, sizeof(struct grao_list_header));
	    user_location[4] = dl.addl_data[3];
	    user_location[5] = dl.addl_data[2];
	    user_location[6] = dl.addl_data[1];
	    user_location[7] = dl.addl_data[0];

	    memcpy ((unchar *)&uds,(unchar *)(user_location+sizeof(struct grao_list_header)),
	    		     sizeof(struct grao_uds_desc));

	    for (i=sizeof(struct grao_list_header); i<length; i+=sizeof(struct grao_uds_desc))
	    {
	    	memcpy ((unchar *)&uds,(unchar *)(user_location+i),sizeof(struct grao_uds_desc));

	    	user_location[i] = (unchar) ((uds.desc_leng >> 8) & 0xff);
	    	user_location[i+1] = (unchar) (uds.desc_leng & 0xff);

	    	for (p=i+16, f=56; p<i+24; p++, f-=8)
	    		user_location[p] = (unchar) ((uds.beginning_loi >> f) & 0xff);
	    	for (f=56; p<i+32; p++, f-=8)
	    		user_location[p] = (unchar) ((uds.ending_loi >> f) & 0xff);
	    }
	#endif
	alloc_order = get_mem_ord(length, &alloc_size);

	if(drv->buffer != NULL){
		printk("lin_tape: buffer != null\n");
		tmp = drv->buffer;
		tmp_len = drv->buf_len;
		drv->buffer = NULL;
	}

	if(alloc_order <= (MAX_PAGE_ORDER-1)){
		drv->buffer = (unchar*)__get_free_pages(GFP_KERNEL, alloc_order);
		if(drv->buffer != NULL){
			if(copy_from_user((char*)drv->buffer, (char*)user_location, length)){
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
		}
	}
	if(drv->buffer == NULL){

		DbgPrint(("lin_tape: using s/g for rao generation\n"));
		drv->op_sg = RAO;
		rc = setup_sg_for_ops(drv, length);
		if(rc == STATUS_SUCCESS){
			//copy data into sg list
			remain = length;
			while(remain > 0 && i < drv->supported_sg_tablesize){
				copy_size = (drv->op_sglist->gather[i].length < remain) ?
					drv->op_sglist->gather[i].length : remain;

				memcpy(&drv->op_sglist->sgl_for_io[i],
					&drv->op_sglist->gather[i],
					sizeof(struct scatterlist));

				drv->op_sglist->sgl_for_io[i].length = copy_size;

				if(copy_from_user(sg_virt(&drv->op_sglist->sgl_for_io[i]),
					user_location, copy_size)) {
					printk("lin_tape: IBMtape%d scatter gather "\
						"copy_from_user failed rao on gather %d, "\
						"returned EFAULT.\n", drv->dev_index,
						i);
					rc = -EFAULT;
					goto EXIT_LABEL;
				} /* if */

				user_location += copy_size;
				remain -= copy_size;
				drv->op_sglist->use_sg++;
				i++;
			} /* while */
		} /* if success */
	} /* if buff NULL */

	drv->buf_len = length;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	drv->cdb_len = CDB_GENERIC_12;
	drv->cdb->cdb_grao.op_code = MAINTENANCE_OUT;
	drv->cdb->cdb_grao.service_action = RAO_SERVICE_ACTION;
	drv->cdb->cdb_grao.process = grao->process;
	drv->cdb->cdb_grao.uds_type = grao->uds_type;
	drv->cdb->cdb_grao.parm_list_length[0] = (unchar)(length >> 24);
	drv->cdb->cdb_grao.parm_list_length[1] = (unchar)(length >> 16);
	drv->cdb->cdb_grao.parm_list_length[2] = (unchar)(length >> 8);
	drv->cdb->cdb_grao.parm_list_length[3] = (unchar)(length & 0xFF);


	if(drv->op_sg == RAO){
		rc = execute_grao_sg(&drv, length);
	}else
	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
			RAOTimeout), DMA_TO_DEVICE, 0, False);

EXIT_LABEL:

	if(drv->op_sg == RAO){
		if(tape_del_ops_sglist(drv))
			printk("lin_tape: unable to delete SG list for GRAO\n");
		drv->op_sg = 0;
	}

	if(tmp != NULL){  //this should not happen
		drv->buffer = tmp;
		drv->buf_len = tmp_len;
	}else {
		if(drv->buffer){
			//free pages
			free_pages((ulong)drv->buffer, alloc_order);
			drv->buffer = NULL;
			drv->buf_len = 0;
		}
	}
	return rc;

} /* stioc_generate_rao() */

/*******************************************************************************
*                                                                              *
*  NAME: execute_grao_sg                                                       *
*                                                                              *
*  PURPOSE: sending list to drive when scatter/gather is used when op system   *
*           cannot allocate enough space currently shouldn't have to use this  *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    length: Size of the structure sent by calling program with grao info      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int execute_grao_sg(struct tape_descriptor** drvp, uint length){

	struct tape_descriptor *drv = *drvp;
    int rc=STATUS_SUCCESS, resid =0;
	unchar *sense_data;

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data){
		printk("lin_tape: failed to allocate sense_data\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset((void*)sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

	if(length == drv->op_sglist->sgl_for_io[0].length){
		rc = tape_execute_scsi_command((*drvp)->dev_obj, (unchar*)(*drvp)->cdb,
			DMA_TO_DEVICE, (char*)&(drv->op_sglist->sgl_for_io[0]), length,
			drv->op_sglist->use_sg,	compute_timeout(drv->drv_id, RAOTimeout),
			sense_data, &resid);
	}else{
		printk("lin_tape: grao length %x != sg length %x",
				length, drv->op_sglist->sgl_for_io[0].length);
	}

	if (rc==-ENOMEM) goto EXIT_LABEL;
	if(rc) 	rc = tape_check_result(drv, rc, sense_data, resid);


	EXIT_LABEL:
	return rc;
}  /* execute_grao_sg */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_receive_rao                                                     *
*                                                                              *
*  PURPOSE: Receive the Recommended Access Order generated from the drive      *
* 			from the generate rao command                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
ssize_t stioc_receive_rao(struct tape_descriptor** drvp, ulong arg)
{
	struct tape_descriptor* drv = *drvp;
	struct receive_rao_list* rrao = NULL;
	int rc = STATUS_SUCCESS;
	uint offset = 0, length = 0;
	int alloc_order=0;
	ulong alloc_size;
	int rao_ssize = sizeof(struct receive_rao_list);
    #if defined(__LITTLE_ENDIAN)
	int p,f,i;
	struct rao_uds_desc {
	  ushort desc_leng; /* descriptor length */
	  char reserved[3];
	  char uds_name[10]; /* uds name given by application */
	  char partition; /* Partition number 0-n to overwrite */
	  unsigned long long beginning_loi; /* Beginning logical object ID */
	  unsigned long long ending_loi; /* Ending logical object ID */
	}uds;
	struct rao_list_header{
      unchar others[4];
      char addl_data[4]; /* additional data */
	}dl;
    #endif

	DbgPrint(("stioc_receive_rao entry"));

	rrao = kmalloc( rao_ssize, GFP_KERNEL );

	if(!rrao){
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(rrao, '\0', rao_ssize);

	if(copy_from_user((char*)rrao, (char*)arg, rao_ssize)){
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	length = rrao->rrao_list_leng;
	alloc_size = (ulong)length;
	offset = rrao->rrao_list_offset;

	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	drv->cdb_len = CDB_GENERIC_12;
	drv->cdb->cdb_rrao.op_code = MAINTENANCE_IN;
	drv->cdb->cdb_rrao.uds_limits = 0;
	drv->cdb->cdb_rrao.service_action = RAO_SERVICE_ACTION;

	if(length <= KmallocThreshhold) {
	  drv->buffer = (unchar *)kmalloc(length, GFP_KERNEL);
	  drv->buf_len = length;
	}
	else {
      alloc_order = get_mem_ord(length, &alloc_size);

      if(alloc_order <= (MAX_PAGE_ORDER-1)){
		drv->buffer = (unchar*)__get_free_pages(GFP_KERNEL, alloc_order);
		drv->buf_len = length;
	  }
	}

	if(drv->buffer == NULL){
		drv->op_sg = RAO;
		rc = setup_sg_for_ops(drv, length);
		if(rc == STATUS_SUCCESS){
			rc = lin_tape_perform_rrao_sg(drv, offset, length);
			if(rc == STATUS_SUCCESS){  	//copy to user
				rc = copy_sg_to_user(drv, rrao->rrao_list, length);
				if(!rc){
					printk("lin_tape: unable to copy sg to user\n");
				}
				else{
                  if(rrao->rrao_list == NULL){
                    printk("lin_tape: After copy to user buffer is NULL\n");
                    rc = -EFAULT;
                    goto EXIT_LABEL;
                  }
               }
			}
		}
		if(tape_del_ops_sglist(drv))
			DbgPrint(("lin_tape: ops_sglist for IBMtape%d not "\
   				      "deleted\n", drv->dev_index));
		goto EXIT_LABEL;
	}

	drv->cdb->cdb_rrao.list_offset[0] = (unchar)(offset >> 24);
	drv->cdb->cdb_rrao.list_offset[1] = (unchar)(offset >> 16);
	drv->cdb->cdb_rrao.list_offset[2] = (unchar)(offset >> 8);
	drv->cdb->cdb_rrao.list_offset[3] = (unchar)(offset & 0xFF);
	drv->cdb->cdb_rrao.alloc_length[0] = (unchar)(length >> 24);
	drv->cdb->cdb_rrao.alloc_length[1] = (unchar)(length >> 16);
	drv->cdb->cdb_rrao.alloc_length[2] = (unchar)(length >> 8);
	drv->cdb->cdb_rrao.alloc_length[3] = (unchar)(length & 0xFF);

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id, RAOTimeout),
			DMA_FROM_DEVICE, 0, False);

	if(rc == STATUS_SUCCESS){
		if(copy_to_user( (char*)rrao->rrao_list,
				(char *)drv->buffer, length)){
			printk("Unable to copy rrao to user\n");
			rc = -EFAULT;
		}
	}

EXIT_LABEL:

    #if defined(__LITTLE_ENDIAN)
	if (rc == STATUS_SUCCESS)
	{
     memcpy ((unchar *)&dl, (unchar *)rrao->rrao_list, sizeof(struct rao_list_header));
     length = (uint) dl.addl_data[3] +
                            ((uint) dl.addl_data[2] << 8)+
                            ((uint) dl.addl_data[1] << 16)+
                            ((uint) dl.addl_data[0] << 24);
     length+=sizeof(struct rao_list_header);
     rrao->rrao_list_leng = min(rrao->rrao_list_leng,length);

     ((unchar *)rrao->rrao_list)[4]=  dl.addl_data[3];
     ((unchar *)rrao->rrao_list)[5] = dl.addl_data[2];
     ((unchar *)rrao->rrao_list)[6] = dl.addl_data[1];
     ((unchar *)rrao->rrao_list)[7] = dl.addl_data[0];

     for (i=sizeof(struct rao_list_header); i<length; i+=sizeof(struct rao_uds_desc))
     {
       memcpy ((unchar *)&uds,(unchar *)(rrao->rrao_list+i),sizeof(struct rao_uds_desc));
       rrao->rrao_list[i] = (unchar) ((uds.desc_leng >> 8) & 0xff);
       rrao->rrao_list[i+1] = (unchar) (uds.desc_leng & 0xff);

       for (p=i+16, f=56; p<i+24; p++, f-=8)
         rrao->rrao_list[p] = (unchar) ((uds.beginning_loi >> f) & 0xff);
       for (f=56; p<i+32; p++, f-=8)
         rrao->rrao_list[p] = (unchar) ((uds.ending_loi >> f) & 0xff);
       if (uds.desc_leng > sizeof(struct rao_uds_desc)){ //with geometry
       	   memcpy ((unchar *)&dl.others, (unchar *)rrao->rrao_list+i+32, 2);
           rrao->rrao_list[i+32]=dl.others[1];
           rrao->rrao_list[i+33]=dl.others[0];
           memcpy ((unchar *)&dl.addl_data, (unchar *)rrao->rrao_list+i+36, 4);
           rrao->rrao_list[i+36]=dl.addl_data[3];
           rrao->rrao_list[i+37]=dl.addl_data[2];
           rrao->rrao_list[i+38]=dl.addl_data[1];
           rrao->rrao_list[i+39]=dl.addl_data[0];
           i+=10;
           memcpy ((unchar *)&dl.others, (unchar *)rrao->rrao_list+i+32, 2);
           rrao->rrao_list[i+32]=dl.others[1];
           rrao->rrao_list[i+33]=dl.others[0];
           memcpy ((unchar *)&dl.addl_data, (unchar *)rrao->rrao_list+i+36, 4);
           rrao->rrao_list[i+36]=dl.addl_data[3];
           rrao->rrao_list[i+37]=dl.addl_data[2];
           rrao->rrao_list[i+38]=dl.addl_data[1];
           rrao->rrao_list[i+39]=dl.addl_data[0];
          i+=10;
       }
     }
	}
    #endif

	if (drv->buffer){
	  if(alloc_size <= KmallocThreshhold) kfree(drv->buffer);
	  else if (alloc_order <= (MAX_PAGE_ORDER-1))
		  free_pages((ulong)drv->buffer, alloc_order);
	}
	DbgPrint(("stioc_receive_rao exit: %d\n", rc));

	return rc;

} /* stioc_receive_rao() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_perform_rrao_sg                                              *
*                                                                              *
*  PURPOSE: When scatter/gather is used in rrao the execute needs to be set    *
* 			differently and ran.  This function does that.                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    user_offset: offset passed in my user for the parameters                  *
*    length:  the length in bytes of data to transfer                          *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int lin_tape_perform_rrao_sg(struct tape_descriptor* drv, uint user_offset,
		uint length)
{
	int rc = STATUS_SUCCESS, i = 0;
	uint remain = length;
	uint offset = user_offset;
	int xfer_size;
	int id = drv->drv_id;
	int resid = 0;
	int temp_use_sg;
	unchar *sense_data;

	temp_use_sg = drv->op_sglist->use_sg;
	drv->op_sglist->use_sg = 1;
	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data){
		printk("lin_tape: failed to allocate sense_data\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset((void*)sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

	while(remain > 0 && i < temp_use_sg){

		xfer_size = drv->op_sglist->sgl_for_io[i].length;

		drv->cdb->cdb_rrao.list_offset[0] = (unchar)(offset >> 24);
		drv->cdb->cdb_rrao.list_offset[1] = (unchar)(offset >> 16);
		drv->cdb->cdb_rrao.list_offset[2] = (unchar)(offset >> 8);
		drv->cdb->cdb_rrao.list_offset[3] = (unchar)(offset & 0xFF);
		drv->cdb->cdb_rrao.alloc_length[0] = (unchar)(xfer_size >> 24);
		drv->cdb->cdb_rrao.alloc_length[1] = (unchar)(xfer_size >> 16);
		drv->cdb->cdb_rrao.alloc_length[2] = (unchar)(xfer_size >> 8);
		drv->cdb->cdb_rrao.alloc_length[3] = (unchar)(xfer_size & 0xFF);

		rc = tape_execute_scsi_command(drv->dev_obj, (char*)drv->cdb,
			DMA_FROM_DEVICE, (char*)&(drv->op_sglist->sgl_for_io[i]),
			xfer_size, drv->op_sglist->use_sg,
			compute_timeout(id, ReadBufferTimeOut), sense_data, &resid);

		remain -= xfer_size;
		offset += xfer_size;
		i++;
		if(rc != STATUS_SUCCESS) break;
	}

	if(remain && rc != STATUS_SUCCESS){
		rc = -EFAULT;
	}

	if (rc == -ENOMEM) goto EXIT_LABEL;
	if(rc) 	rc = tape_check_result(drv, rc, sense_data, resid);


	EXIT_LABEL:

	return rc;
} /* lin_tape_perform_rrao_sg() */

/*******************************************************************************
*                                                                              *
*  NAME: copy_sg_to_user                                                       *
*                                                                              *
*  PURPOSE: Returning scatter/gather to user from op_sglist                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*    drvp: tape descriptor for the drive                                       *
*    user_location: user location to return the data to         I              *
*    length:  length of the data in bytes that needs returning                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int copy_sg_to_user(struct tape_descriptor* drv, void *user_location,
		uint length)
{
	int rc = STATUS_SUCCESS;
	int i = 0;
	uint remain = 0;
	size_t copy_size = 0;
	char* temp_location = user_location;

	i = 0;
	remain = length;

	while(remain > 0 && i < drv->supported_sg_tablesize) {
		if(!drv->op_sglist || !drv->op_sglist->gather[i].length
			|| !sg_page(&drv->op_sglist->gather[i])) {
			printk("lin_tape: IBMtape%d "\
				"lin_tape_copy_read_buffer corrupted "\
				"scatter gather, returned EFAULT.\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		/* copying either scatter list size or the remaining */
		copy_size = (drv->op_sglist->gather[i].length < remain) ?
			drv->op_sglist->gather[i].length : remain;

		if(copy_to_user(temp_location,
				sg_virt(&drv->op_sglist->gather[i]),
				copy_size)) {
			printk("lin_tape: IBMtape%d sg copy failed\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		temp_location += copy_size;
		remain -= copy_size;
		drv->op_sglist->use_sg--;
		i++;
	} /* while */

	EXIT_LABEL:

	return rc;

}  /* copy_sg_to_user() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_wfm                                                              *
*                                                                              *
*  PURPOSE: Write file mark(s)                                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    immediate: return from function before verifying that data is committed   *
*    count: number of marks to write                                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_wfm(struct tape_descriptor** drvp, unchar immediate, ulong count)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_wfm entry\n"));

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->wakeup_info = 0;

	drv->cdb->cdb_wfm.op_code = WRITE_FILEMARKS;
	drv->cdb->cdb_wfm.immediate = immediate & 0x01;
	drv->cdb->cdb_wfm.write_setmarks = False;
	drv->cdb->cdb_wfm.count[0] = (unchar)(count >> 16 & 0xFF);
	drv->cdb->cdb_wfm.count[1] = (unchar)(count >> 8 & 0xFF);
	drv->cdb->cdb_wfm.count[2] = (unchar)(count & 0xFF);
	drv->cdb_len = CDB_GENERIC_10;
	drv->last_op = count ? WRITE_FILEMARKS : WRITE_FILEMARK_ZERO;
	drv->wfm_on_close_safe = True;
	drv->buffer = NULL;
	drv->buf_len = 0;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		WriteFileMarkTimeOut), DMA_NONE, 0, False);

	if(rc != STATUS_SUCCESS) {
		if(count) drv->pos_valid = False;

		if(drv->eom && drv->resid_valid) {
			if(drv->sense_data->sense_key == VOLUME_OVERFLOW)
				rc = -EIO;

			else if(drv->resid == 0) {
				if(drv->sense_data->sense_key == 0x00
					&& drv->sense_data->asc == 0x00
					&& drv->sense_data->ascq == 0x07)
					/* programmable early warning zone */
					rc = -ENOSPC;

				else if(drv->deferred_eom == False)
					/* set so next write will notify eom */
					drv->deferred_eom = True;

				if(count) drv->ending_fm = count;
				if(drv->eot_warn) rc = -ENOSPC;
				if(rc != -ENOSPC) rc = STATUS_SUCCESS;
			} else {
				rc = -EIO;
				goto EXIT_LABEL;
			} /* else */
		} /* if */
	} /* if */

	/* keep track of how many filemarks we have at eod (for tape_flush) */
	if(rc == STATUS_SUCCESS && count) {
		drv->ending_fm = count;

		if(drv->pos_valid)
			drv->curr_pos += count;

		DbgPrint(("lin_tape: tape_wfm current position  %llu\n",
			drv->curr_pos));
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_wfm() */

/****************************************************************
 *                                                              *
 *        SCSI Pass Through Functions                           *
 *                                                              *
 ****************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru                                                        *
*                                                                              *
*  PURPOSE: Perform tape passthrough operations                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    cdb: the command we are about to send via passthrough                     *
*    cmd_length: number of valid bytes in CDB                                  *
*    buffer: data buffer                                                       *
*    buffer_length: number of bytes required for op                            *
*    data_direction: to device or from device                                  *
*    timeout: seconds between command issue and time out                       *
*    sense: buffer to receive sense data                                       *
*    sense_length: length of sense in bytes                                    *
*    sense_avail: True if sense data is available for retrieval                *
*    resid: residual bytes returned in sense data                              *
*    result: as returned from SCSI layer                                       *
*    fpfo: flag set True only if request comes from pfo                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru(struct tape_descriptor** drvp, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint* buffer_length,
	uint data_direction, uint timeout, unchar* sense, unchar* sense_length,
	unchar* sense_avail, int* resid, int* result, boolean fpfo)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;
	ulong order = 0;
	unchar orig_sense_length = *sense_length;

	DbgPrint(("lin_tape: tape_pass_thru entry\n"));

	*result = 0;
	*sense_length = 0;
	*sense_avail = 0;

	if(!cdb || !timeout || !cmd_length || (data_direction != SCSI_DATA_IN
		&& data_direction != SCSI_DATA_OUT
		&& data_direction != SCSI_DATA_NONE)) {
		printk("lin_tape: tape_pass_thru: invalid argument\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(data_direction != SCSI_DATA_NONE && (!buffer || !*buffer_length)) {
		DbgPrint(("lin_tape: tape_pass_through: no buffer\n"));
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(drv) drv->wakeup_info = 0;

	/* Change regular reserve to persistent reserve if necessary */
	if(cdb[0] == RESERVE && TP_RSV_PERSISTENT(drv)) {
		rc = tape_pass_thru_persistent_reserve(&drv, sense,
			orig_sense_length, sense_avail, sense_length, fpfo);
		goto EXIT_LABEL;
	} else if(cdb[0] == RELEASE && TP_RSV_PERSISTENT(drv)) {
		rc = tape_pass_thru_persistent_release(&drv, sense,
			orig_sense_length, sense_avail, sense_length, fpfo);
		goto EXIT_LABEL;
	} /* else if */

	if(data_direction != SCSI_DATA_NONE) {
		rc = tape_pass_thru_handle_buffers(drv, cdb, buffer,
			buffer_length, data_direction, &order, fpfo);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	memcpy(drv->cdb, cdb, cmd_length);
	drv->cdb_len = cmd_length;
	drv->last_op = cdb[0];
	if(cdb[0] == READ_6 || cdb[0] == WRITE_6) {
	    drv->sili = (drv->cdb)->cdb_read_write.sili;
	    rc = tape_send_scsi_cmd(&drv, timeout * HZ, data_direction,
		                        *buffer_length, 0);
	} else	rc = tape_send_scsi_cmd(&drv, timeout * HZ, data_direction,
			                        0, 0);
    if (rc == -ENXIO || rc == -ENOMEM) goto EXIT_LABEL;

	if(rc != STATUS_SUCCESS) {
		rc = tape_pass_thru_unsuccessful(&drv, cdb, buffer,
			buffer_length, data_direction, sense,
			orig_sense_length, sense_avail, resid, sense_length, fpfo, rc);
	} else {
		rc = tape_pass_thru_successful(&drv, cdb, buffer, buffer_length,
			data_direction, sense, orig_sense_length, sense_avail,
			resid, sense_length, fpfo);
	} /* else */


EXIT_LABEL:

	if(drv->result)	*result = drv->result;

	if(drv->buffer && drv->buf_len <= KmallocThreshhold) kfree(drv->buffer);
	else if(drv->buffer) free_pages((ulong)drv->buffer, order);
	drv->buffer = NULL;
	drv->buf_len = 0;

	*drvp = drv;

	DbgPrint(("lin_tape: tape_pass_thru exit 0x%x %x\n", rc, *result));
	return rc;
} /* tape_pass_thru() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru_persistent_reserve                                     *
*                                                                              *
*  PURPOSE: Handle persistent reserve on pass through                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    sense: buffer to receive sense data                                       *
*    orig_sense_length: length of sense in bytes                               *
*    sense_avail: (output) True if sense data is available for retrieval       *
*    new_slen: (output) new sense length if applicable                         *
*    fpfo: set True if request comes from pfo                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru_persistent_reserve(struct tape_descriptor** drvp,
	unchar* sense, unchar orig_sense_length, unchar* sense_avail,
	unchar* new_slen, boolean fpfo)
{
	struct tape_descriptor* drv = *drvp;
	unchar* key = NULL, * rkey = NULL;
	unsigned int pr_generation = 0;
	int rc = STATUS_SUCCESS;
	char pr_type = 0;

	key = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	rkey = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!key || !rkey) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(key, '\0', RESERVE_KEY_LENGTH);
	memset(rkey, '\0', RESERVE_KEY_LENGTH);

	if(drv->curr_reserve_type != LinTapePersistentReserve) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_read_reservation(&drv, key, &pr_generation, &pr_type);
	if(rc != STATUS_SUCCESS) {
		if(drv->sense_valid && drv->sense_data) goto SENSE_LABEL;
		else goto EXIT_LABEL;
	} /* if */

	if(!memcmp(drv->reservation_key, key, RESERVE_KEY_LENGTH)
		&& drv->pr_generation == pr_generation
		&& pr_type == EXCLUSIVE_ACCESS) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	get_random_bytes(rkey, RESERVE_KEY_LENGTH);
	rc = tape_register_reservation_key(&drv, rkey);
	if(rc != STATUS_SUCCESS) {
		if(drv->sense_valid && drv->sense_data) goto SENSE_LABEL;
		else goto EXIT_LABEL;
	} /* if */

	rc = tape_create_persist_reserve(&drv);
	if(rc == STATUS_SUCCESS) {
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
		drv->curr_reserve_type = LinTapePersistentReserve;
		goto EXIT_LABEL;
	} /* if */

	/* failure: remove the key just registered  and update pr_generation */
	rc = tape_register_reservation_key(&drv, NULL);
	rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
	if(rc != STATUS_SUCCESS) {
		if(drv->sense_valid && drv->sense_data) goto SENSE_LABEL;
		else goto EXIT_LABEL;
	} /* if */

	drv->pr_generation = pr_generation;
	goto EXIT_LABEL;

SENSE_LABEL:

	*sense_avail = TRUE;
	if(!sense || orig_sense_length <= 0) goto EXIT_LABEL;
	if(orig_sense_length > SCSI_SENSE_BUFFERSIZE)
		orig_sense_length = SCSI_SENSE_BUFFERSIZE;
	if(!fpfo && copy_to_user(sense, drv->sense_data, orig_sense_length))
		rc = -EFAULT;
	else{
		if (fpfo) memcpy(sense, drv->sense_data, orig_sense_length);
		*new_slen = orig_sense_length;
	}

EXIT_LABEL:

	*drvp = drv;
	if(key) kfree(key);
	if(rkey) kfree(rkey);
	return rc;
} /* tape_pass_thru_persistent_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru_persistent_release                                     *
*                                                                              *
*  PURPOSE: Handle persistent release on pass through                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    sense: buffer to receive sense data                                       *
*    orig_sense_length: length of sense in bytes                               *
*    sense_avail: (output) True if sense data is available for retrieval       *
*    new_slen: (output) new sense length if applicable                         *
*    fpfo: set to True if request comes from pfo                               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru_persistent_release(struct tape_descriptor** drvp,
	unchar* sense, unchar orig_sense_length, unchar* sense_avail,
	unchar* new_slen, boolean fpfo)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;
	unsigned int pr_generation = 0;

	if(drv->curr_reserve_type != LinTapePersistentReserve) goto EXIT_LABEL;

	/* register key 0 to clear the reservation and registration */
	rc = tape_register_reservation_key(&drv, NULL);
	if(rc == STATUS_SUCCESS) {
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
		drv->curr_reserve_type = LinTapeNoReserve;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->sense_valid || !drv->sense_data) goto EXIT_LABEL;

	*sense_avail = TRUE;
	if(!sense || orig_sense_length <= 0) goto EXIT_LABEL;

	if(orig_sense_length > SCSI_SENSE_BUFFERSIZE)
		orig_sense_length = SCSI_SENSE_BUFFERSIZE;
	if(!fpfo && copy_to_user(sense, drv->sense_data, orig_sense_length)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	}
	else {
		if (fpfo) memcpy(sense,drv->sense_data, orig_sense_length);
		*new_slen = orig_sense_length;
	}

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* tape_pass_thru_persistent_release() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru_handle_buffers                                         *
*                                                                              *
*  PURPOSE: Check and if necessary allocate buffers for pass through operation *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    cdb: the command we are about to send via passthrough                     *
*    buffer: data buffer                                                       *
*    buffer_length: number of bytes required for op                            *
*    data_direction: to device or from device                                  *
*    order: (output) base-2 log of the number of pages allocated               *
*    fpfo: set True if request comes from pfo                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru_handle_buffers(struct tape_descriptor* drv, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, ulong* order, boolean fpfo)
{
	ulong buf_size = 0;
	int rc = STATUS_SUCCESS;

	if(cdb[0] == READ_6 || cdb[0] == WRITE_6) {
		if(!drv->sg_list && (drv->rw_buf_size < *buffer_length
			|| !drv->rw_buf))
			rc = tape_rw_buffer(drv, (ulong)*buffer_length);

		if((drv->sg_list || rc != STATUS_SUCCESS) && (!drv->sg_list ||
			*buffer_length > drv->sg_list->total_sg_size)) {
			rc = tape_build_sg(drv, (ulong)*buffer_length);
			if(rc != STATUS_SUCCESS) {
				printk("lin_tape: IBMtape%d sg of length %i "\
					"failed.\n", drv->dev_index,
					*buffer_length);
				rc = -ENOMEM;
				goto EXIT_LABEL;
			} /* if */
		} /* if */

		if(data_direction == SCSI_DATA_OUT)
			rc = tape_copy_write_buffer(drv, buffer,
				*buffer_length, fpfo);
		goto EXIT_LABEL;
	} /* if */

	if(*buffer_length > KmallocThreshhold) {
		*order = get_mem_ord(*buffer_length, &buf_size);
		drv->buffer = (unchar*)__get_free_pages(GFP_KERNEL, *order);
		drv->buf_len = buf_size;
	} else {
		drv->buffer = kmalloc(*buffer_length, GFP_KERNEL | GFP_DMA);
		drv->buf_len = *buffer_length;
	} /* else */

	if(!drv->buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(!fpfo && copy_from_user(drv->buffer, buffer, *buffer_length)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	else if (fpfo) memcpy(drv->buffer, buffer, *buffer_length);

EXIT_LABEL:

	return rc;
} /* tape_pass_thru_handle_buffers() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru_unsuccessful                                           *
*                                                                              *
*  PURPOSE: Handle cleanup after unsuccessful tape pass through command        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    cdb: the command we are about to send via passthrough                     *
*    buffer: data buffer                                                       *
*    buffer_length: number of bytes required for op                            *
*    data_direction: to device or from device                                  *
*    sense: buffer to receive sense data                                       *
*    orig_sense_length: length of sense in bytes                               *
*    sense_avail: True if sense data is available for retrieval                *
*    resid: residual bytes returned in sense data                              *
*    new_slen: (output) new sense length if applicable                         *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru_unsuccessful(struct tape_descriptor** drvp, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, unchar* sense,
	unchar orig_sense_length, unchar* sense_avail, int* resid,
	unchar* new_slen, boolean fpfo, int uc)
{
	struct tape_descriptor* drv = *drvp;
	int rc = uc;

	if(cdb[0] == READ_6 || cdb[0] == WRITE_6 ||
		cdb[0] == WRITE_FILEMARKS || cdb[0] == SPACE ||
		cdb[0] == LOCATE || cdb[0] == READ_REVERSE ||
		cdb[0] == 0x14 || cdb[0] == 0x34) drv->pos_valid = False;

	if(drv->sense_valid) {
		if(sense && orig_sense_length > 0) {
			if(orig_sense_length > SCSI_SENSE_BUFFERSIZE)
				orig_sense_length = SCSI_SENSE_BUFFERSIZE;
			if(!fpfo && copy_to_user(sense, drv->sense_data,
				orig_sense_length)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} else {
				if (fpfo) memcpy(sense, drv->sense_data, orig_sense_length);
				*new_slen = orig_sense_length;
			}
		} /* if */
		*sense_avail = TRUE;
		if(drv->sense_data->sense_key) {
			/* return success and let application handle */
			rc = STATUS_SUCCESS;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	if(data_direction != SCSI_DATA_IN || cdb[0] != READ_6) goto EXIT_LABEL;
	if(!drv->sense_valid || !drv->resid_valid) goto EXIT_LABEL;

	*resid = drv->resid;
	if(drv->blk_size) {
		if(drv->read_mode == AIXMode && drv->fm_detected) {
			drv->in_erp = TRUE;
			rc = tape_set_pos(&drv, SpaceFileMark, -1);
			drv->in_erp = FALSE;

			if(rc != STATUS_SUCCESS) {
				printk("lin_tape: IBMtape%d unable to RSF\n",
					drv->dev_index);
				goto EXIT_LABEL;
			} /* if */
		} /* if */
		goto RESID_LABEL;
	} /* if */

	if(drv->resid < 0) {
		tape_copy_read_buffer(drv, buffer, *buffer_length, fpfo);
		goto EXIT_LABEL;
	} /* if */

	if(drv->fm_detected && drv->sense_data->asc == 0x00 &&
		drv->sense_data->ascq == 0x01) {
		DbgPrint(("lin_tape: FileMark detected.\n"));
		/* return success and let application handle */
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

RESID_LABEL:
	*buffer_length = *buffer_length - drv->resid;
	if(*buffer_length > 0)
		rc = tape_copy_read_buffer(drv, buffer, *buffer_length, fpfo);
	else rc = -EIO;

EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* tape_pass_thru_unsuccessful() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_pass_thru_successful                                             *
*                                                                              *
*  PURPOSE: Handle cleanup after successful tape pass through command          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    cdb: the command we are about to send via passthrough                     *
*    buffer: data buffer                                                       *
*    buffer_length: number of bytes required for op                            *
*    data_direction: to device or from device                                  *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_pass_thru_successful(struct tape_descriptor** drvp, unchar* cdb,
	unchar* buffer, uint* buffer_length, uint data_direction, unchar* sense,
	unchar orig_sense_length, unchar* sense_avail, int* resid,
	unchar* new_slen, boolean fpfo)
{
	struct tape_descriptor* drv = *drvp;
	long read_length = 0;
	int rc = STATUS_SUCCESS, cpy_len = 8, i;
	uint per_xfer_length = 0;
	struct read_pos *buf = NULL;
	struct pos_info* read_pos_info;
	char *flo = NULL;
	char ttype = 0, data_format = 0;

	if(drv->sense_valid) {
		if(sense && orig_sense_length > 0) {
			if(orig_sense_length > SCSI_SENSE_BUFFERSIZE)
				orig_sense_length = SCSI_SENSE_BUFFERSIZE;
			if(!fpfo && copy_to_user(sense, drv->sense_data,
				orig_sense_length)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} else {
				if (fpfo) memcpy(sense, drv->sense_data, orig_sense_length);
				*new_slen = orig_sense_length;
			}
		} /* if */
		*sense_avail = TRUE;
	} /* if */

	if(data_direction == SCSI_DATA_IN) {
		if(cdb[0] == READ_6) {
			if(drv->resid_valid && drv->resid > 0)
				read_length = *buffer_length - drv->resid;
			else read_length = *buffer_length;
			*buffer_length = read_length;
            rc = tape_copy_read_buffer(drv, buffer, read_length, fpfo);
		} else {
			if (fpfo) memcpy(buffer, drv->buffer, drv->buf_len);
			else if(copy_to_user(buffer, drv->buffer, drv->buf_len))
				     rc = -EFAULT;
		}
	} /* if */

	/* update block size, compression, etc */
	if(cdb[0] == MODE_SELECT || cdb[0] == MODE_SENSE) rc = tape_update_all_parms(&drv);

	if(cdb[0] == ALLOW_MEDIUM_REMOVAL)
		drv->prevent_removal =
			((union cdb*)cdb)->cdb_medium_removal.prevent;

	if (cdb[0] == REWIND) {
		drv->eom = False;
		drv->deferred_eom = False;
		drv->eom_blk_id = 0;
		drv->pos_valid = True;
		drv->curr_pos = 0;
	}else
	if (cdb[0] == 0X34){
		ttype = lin_tape_trace_type;
		buf = (struct read_pos *) drv->buffer;
		if(buf->bpu) {
			drv->pos_valid = False;
			if(drv->trace && ttype == LIN_TAPE_FULL_TRACE &&
				drv->buf_len)
				tape_log_buffer_data(drv, min((int)drv->buf_len,
					(int)(FullTraceUnitSize - 7)),
					drv->buffer);
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */

		data_format = cdb[1] & 0x1f;
        drv->pos_valid = True;
		if(data_format == RP_LONG_FORM || data_format == RP_EXTENDED_FORM) {
			read_pos_info = (struct pos_info*) buf;
			switch(data_format) {
			case RP_LONG_FORM:
				flo = ((struct long_data_format*)
					read_pos_info)->logical_obj_number;
				break;
			case RP_EXTENDED_FORM:
				flo = ((struct extended_data_format*)
					read_pos_info)->
					first_logical_obj_position;
				break;

			} /* switch */

			drv->curr_pos = 0;
			for(i = 0; i < cpy_len; i++)
				drv->curr_pos |= ((long long)(flo[i] & 0xFF) <<
					(cpy_len * BITS_PER_BYTE -
					(i + 1) * BITS_PER_BYTE));
		} /* if */
		else {
            drv->curr_pos = (ulong)((buf->first_blk_pos[0] << 24)
            	+ (buf->first_blk_pos[1] << 16) + (buf->first_blk_pos[2] << 8)
				+ (buf->first_blk_pos[3]));
		}

		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE && drv->buf_len)
			tape_log_buffer_data(drv, min((int)drv->buf_len,
				(int)(FullTraceUnitSize-7)), drv->buffer);
	}

	/* update current tape position */
	if(drv->eod || drv->fm_detected || drv->eom) {
    	drv->pos_valid = False;
	}

	if(rc != STATUS_SUCCESS ) goto EXIT_LABEL;
	if(cdb[0] == READ_6 || cdb[0] == WRITE_6){
	    if (drv->pos_valid) {
	    	per_xfer_length =
	    			(drv->cdb->cdb_read_write.xfer_length[0] << 16)
					+ (drv->cdb->cdb_read_write.xfer_length[1] << 8)
					+ (drv->cdb->cdb_read_write.xfer_length[2] & 0xFF);

			if((drv->cdb)->cdb_read_write.fixed) drv->curr_pos += per_xfer_length;
			else drv->curr_pos++;
	    }
	    else	rc = tape_send_erp_cmd(drv, READ_POSITION, 0, 0);
	} else if(cdb[0] == WRITE_FILEMARKS || cdb[0] == SPACE ||
		cdb[0] == LOCATE || cdb[0] == READ_REVERSE ||
		cdb[0] == 0x14){
		rc = tape_send_erp_cmd(drv, READ_POSITION, 0, 0);
		if (fpfo && !rc && drv->eom && drv->pos_valid && drv->curr_pos == 0)
			drv->eom = False;
	}

EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* tape_pass_thru_successful() */

/*******************************************************************************
*                                                                              *
*        Data Path Failover Functions                                          *
*                                                                              *
*******************************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: persistent_reserve_in                                                 *
*                                                                              *
*  PURPOSE: Issue persistent reservation in                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    buf: output buffer                                                        *
*    drv: the descriptor representing the drive on which to perform operation  *
*    service_action: for the persistent in CDB                                 *
*    buf_len: bytes allocated to buf                                           *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t persistent_reserve_in(char* buf, struct tape_descriptor *drv,
	unchar service_action, int buf_len)
{
	int rc = STATUS_SUCCESS, timeout = 0, large_size = 0, order = 0;
	struct _cdb_persist_reserve_in* pr_in_cdb = NULL;
	boolean cdb_allocated = False;

	DbgPrint(("lin_tape: persistent_reserve_in entry, action %d\n",
		service_action));

	if(!drv->cdb) {
		drv->cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
		if(!drv->cdb) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		cdb_allocated = True;
	} /* if */

	if(buf_len > KmallocThreshhold) {
		order = get_mem_ord(buf_len, (ulong*)&large_size);
		drv->buffer = (char*)__get_free_pages(GFP_KERNEL, order);
		drv->buf_len = large_size;
	} else {
		drv->buffer = kmalloc(buf_len, GFP_KERNEL);
		drv->buf_len = buf_len;
	} /* else */

	if(!drv->buffer) {
		printk("lin_tape: persistent_reserve_in memory allocation "\
				"failed.\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->buffer, '\0', drv->buf_len);

	memset(drv->cdb, 0, MAX_CDB_SIZE);
	pr_in_cdb = &(drv->cdb)->cdb_persist_reserve_in;

	pr_in_cdb->op_code = PERSISTENT_RESERVE_IN;
	pr_in_cdb->serv_action = service_action;
	pr_in_cdb->alloc_length[0] = (unchar)((drv->buf_len >> 16) & 0xFF);
	pr_in_cdb->alloc_length[1] = (unchar)((drv->buf_len >> 8) & 0xFF);
	pr_in_cdb->alloc_length[2] = (unchar)(drv->buf_len & 0xFF);

	drv->last_op = PERSISTENT_RESERVE_IN;
	drv->cdb_len = CDB_GENERIC_10;

	timeout = compute_timeout(drv->drv_id, PersistReserveInTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_FROM_DEVICE, 0, False);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	memcpy(buf, drv->buffer, buf_len);

EXIT_LABEL:

	if(cdb_allocated) {
		kfree(drv->cdb);
		drv->cdb = NULL;
	} /* if */

	if(drv->buffer) {
		if(drv->buf_len <= KmallocThreshhold)
			kfree(drv->buffer);
		else free_pages((ulong)drv->buffer, order);

	} /* if */

	drv->buf_len = 0;
	drv->buffer = NULL;

	DbgPrint(("lin_tape: persistent_reserve_in exit: %d\n", rc));
	return rc;

} /* persistent_reserve_in() */

/*******************************************************************************
*                                                                              *
*  NAME: persistent_reserve_out                                                *
*                                                                              *
*  PURPOSE: Issue persistent reservation out                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape drive for which PR-out is being issued                          *
*    service_action: the specific action for the PR-out command                *
*    parm_list_buf: the parameter list for the PR-out command                  *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t persistent_reserve_out(struct tape_descriptor* drv,
	unchar service_action, struct persist_reserve_out* parm_list_buf)
{
	int rc = 0, timeout = 0;
	struct _cdb_persist_reserve_out* pr_out_cdb = NULL;
	boolean cdb_allocated = False;

	DbgPrint(("lin_tape: persistent_reserve_out entry\n"));

	if(!drv->cdb) {
		drv->cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
		if(!drv->cdb) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		cdb_allocated = True;
	} /* if */

	drv->buffer = parm_list_buf;
	drv->buf_len = sizeof(struct persist_reserve_out);

	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	pr_out_cdb = &drv->cdb->cdb_persist_reserve_out;
	pr_out_cdb->op_code = PERSISTENT_RESERVE_OUT;
	pr_out_cdb->serv_action = service_action;
	pr_out_cdb->parm_list_length[0] = 0;
	pr_out_cdb->parm_list_length[1] = sizeof(struct persist_reserve_out);

	switch(service_action) {
		case CREATE_PERSIST_RESERVE: case PREEMPT_PERSIST_RESERVE:
		case PREEMPT_AND_ABORT:
			pr_out_cdb->type = EXCLUSIVE_ACCESS;
			break;
		default:
			pr_out_cdb->type = 0;
	} /* switch */

	drv->last_op = PERSISTENT_RESERVE_OUT;
	drv->cdb_len = CDB_GENERIC_10;

	timeout = compute_timeout(drv->drv_id, PersistReserveOutTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	if(cdb_allocated) {
		kfree(drv->cdb);
		drv->cdb = NULL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;

	DbgPrint(("lin_tape: persistent_reserve_out exit: %d\n", rc));
	return rc;

} /* persistent_reserve_out() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_read_reservation                                                 *
*                                                                              *
*  PURPOSE: Issue persistent reservation in to obtain reservation key,         *
*           PR generation and reservation type                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    key: (output) reservation key                                             *
*    pr_generation: (output) number of persistent reserves issued              *
*    type: (output) reservation type                                           *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_read_reservation(struct tape_descriptor **drvp, unchar *key,
	unsigned int *pr_generation, unchar *type)
{
	struct persist_read_reservs *buf = NULL;
	int rc = STATUS_SUCCESS;
	int length = 0;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_read_reservation entry.\n"));

	buf = kmalloc(sizeof(struct persist_read_reservs), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buf, '\0', sizeof(struct persist_read_reservs));

	/* get the length of key list first */
	rc = persistent_reserve_in((char*)buf, drv, QUERY_RESERVATIONS,
		PERSIST_PAGE_HEAD_SIZE);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	*pr_generation = (buf->generation[0] << 24)
		+ (buf->generation[1] << 16)
		+ (buf->generation[2] << 8)
		+ (buf->generation[3]);

	if(key == NULL) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	length = (buf->additional_length[0] << 24)
		+ (buf->additional_length[1] << 16)
		+ (buf->additional_length[2] << 8)
		+ buf->additional_length[3];

	/* Since we always use Exclusive Access type reservation, if there is
		a reservation, the key list should only have 1 key.  */

	if(length / sizeof(struct persist_reserve) != 1) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* Length of key list == 1, so buf should be large enough;
		no reallocation necessary */

	rc = persistent_reserve_in((char*)buf, drv, QUERY_RESERVATIONS,
		sizeof(struct persist_read_reservs));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	memcpy(key, buf->key_descriptor_list[0].reserve_key,
		RESERVE_KEY_LENGTH);
	*type = buf->key_descriptor_list[0].type;

	*drvp = drv;

EXIT_LABEL:

	if(buf) {
		kfree(buf);
		buf = NULL;
	} /* if */

	DbgPrint(("lin_tape: tape_read_reservation exit: %d\n", rc));
	return rc;

} /* tape_read_reservation() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_register_reservation_key                                         *
*                                                                              *
*  PURPOSE: Register a key with a drive to use in persistent reservations      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    key: key to register                                                      *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_register_reservation_key(struct tape_descriptor **drvp,
	unchar *key)
{
	struct persist_reserve_out *parm_list_buf = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_register_reservation_key entry.\n"));

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);

	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));

	if(key) {
		/* Register a key */
		/* reserve_key = 0, service_key = key passed in.  */
		memcpy(parm_list_buf->service_key, key, RESERVE_KEY_LENGTH);
	} else {
		/* Unregister an existing registration. */
		/* service_key = 0, reserve_key = key of current initiator */
		memcpy(parm_list_buf->reserve_key, drv->reservation_key,
			RESERVE_KEY_LENGTH);
	} /* else */
	rc = persistent_reserve_out(drv,
		key ? REGISTER_IGNORE_KEY : REGISTER_KEY, parm_list_buf);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	memcpy(drv->reservation_key, parm_list_buf->service_key,
		RESERVE_KEY_LENGTH);

	/* report that drive was released */
	if(!key) tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_RELEASED,
					NULL, 0);

EXIT_LABEL:

	*drvp = drv;

	if(parm_list_buf) kfree(parm_list_buf);
	parm_list_buf = NULL;

	DbgPrint(("lin_tape: tape_register_reservation_key exit: %d\n", rc));
	return rc;

} /* tape_register_reservation_key() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_create_persist_reserve                                           *
*                                                                              *
*  PURPOSE: Create a persistent reservation                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_create_persist_reserve(struct tape_descriptor **drvp)
{
	struct persist_reserve_out* parm_list_buf = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_create_persist_reserve entry.\n"));

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);

	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));
	memcpy(parm_list_buf->reserve_key, drv->reservation_key,
		RESERVE_KEY_LENGTH);

	rc = persistent_reserve_out(drv, CREATE_PERSIST_RESERVE, parm_list_buf);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* report that drive was reserved */
	tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_RESERVED, NULL, 0);

EXIT_LABEL:

	if(parm_list_buf) kfree(parm_list_buf);
	drv->buffer = NULL;
	drv->buf_len = 0;

	*drvp = drv;

	DbgPrint(("lin_tape: tape_create_persist_reserve exit: %d\n", rc));

	return rc;
} /* tape_create_persist_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_preempt_reservation                                              *
*                                                                              *
*  PURPOSE: Preempt reservation and reestablish a new one                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    old_key: the key to preempt                                               *
*    new_key: the key with with to establish the new reservation               *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int tape_preempt_reservation(struct tape_descriptor** drvp, unchar* old_key,
	unchar* new_key)
{
	int rc = STATUS_SUCCESS;
	struct persist_reserve_out* parm_list_buf = NULL;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: tape_preempt_reservation entry\n"));

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);
	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));
	memcpy(parm_list_buf->reserve_key, new_key, RESERVE_KEY_LENGTH);
	memcpy(parm_list_buf->service_key, old_key, RESERVE_KEY_LENGTH);

	rc = persistent_reserve_out(drv, PREEMPT_AND_ABORT, parm_list_buf);

EXIT_LABEL:

	if(parm_list_buf) kfree(parm_list_buf);
	drv->buffer = NULL;
	drv->buf_len = 0;

	*drvp = drv;

	DbgPrint(("lin_tape: tape_preempt_reservation exit: %d\n", rc));
	return rc;
} /* tape_preempt_reservation() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_clear_all_registrations                                          *
*                                                                              *
*  PURPOSE: remove all persistent registrations on the device                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_clear_all_registrations(struct tape_descriptor **drvp)
{
	struct persist_reserve_out *parm_list_buf = NULL;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;
	unchar rkey[RESERVE_KEY_LENGTH] = {0};

	DbgPrint(("lin_tape: tape_clear_all_registrations entry.\n"));

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);

	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));

	memcpy(parm_list_buf->reserve_key, drv->reservation_key,
		RESERVE_KEY_LENGTH);

	rc = persistent_reserve_out(drv, CLEAR_KEYS_RESERVES, parm_list_buf);

	/* check sense; if it is op_code err, reset rc to -ENOSYS */
	if(rc == STATUS_SUCCESS) {
		goto EXIT_LABEL;
	} else if(drv->sense_valid &&
		drv->sense_data->sense_key == 0x05 &&
		drv->sense_data->asc == 0x20 &&
		drv->sense_data->ascq == 0x00) {
		DbgPrint(("lin_tape: microcode level does not support "\
			"this function.\n"));
		drv->sense_valid = False;
		rc = -ENOSYS;
	} else if(rc == -EBUSY) {
		DbgPrint(("lin_tape: tape_clear_all_registrations failed "\
			"possibly because of no registration. Establishing\n"));

		get_random_bytes(rkey, RESERVE_KEY_LENGTH);
		rc = tape_register_reservation_key(&drv, rkey);

		if(rc != STATUS_SUCCESS) {
			rc = -EBUSY; /* return original code */
			goto EXIT_LABEL;
		} /* if */

		memcpy(parm_list_buf->reserve_key, rkey, RESERVE_KEY_LENGTH);

		rc = persistent_reserve_out(drv, CLEAR_KEYS_RESERVES,
			parm_list_buf);

		if(rc == STATUS_SUCCESS) goto EXIT_LABEL;

		rc = -EBUSY; /* return original code */
	} /* if */

EXIT_LABEL:

	if(rc == STATUS_SUCCESS) {
		/* report that drive was released */
		tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_RELEASED,
			NULL, 0);
	} /* if */

	*drvp = drv;

	if(parm_list_buf) kfree(parm_list_buf);
	parm_list_buf = NULL;

	DbgPrint(("lin_tape: tape_clear_all_registrations exit: %d\n", rc));
	return rc;

} /* tape_clear_all_registrations() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_read_keys                                                        *
*                                                                              *
*  PURPOSE: read reservation keys that exist on a device                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: (output) the buffer to receive the key data                          *
*    buf_len: length allocated for keys including PR-in eight-byte header      *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_read_keys(struct tape_descriptor ** drvp, void *buf, ulong buf_len)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_read_keys entry.\n"));

	rc = persistent_reserve_in(buf, drv, QUERY_RESERVED_KEYS, buf_len);

	if(rc != STATUS_SUCCESS) {

		/* check sense data; if op_code error, reset rc to -ENOSYS */
		if(drv->sense_valid &&
			drv->sense_data->sense_key == 0x05 &&
			drv->sense_data->asc == 0x20 &&
			drv->sense_data->ascq == 0x00) {
			DbgPrint(("lin_tape: microcode level does not support "\
				"this function.\n"));
			drv->sense_valid = False;
			rc = -ENOSYS;
		} /* if */
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	*drvp = drv;

	DbgPrint(("lin_tape: tape_read_keys exit: %d\n", rc));
	return rc;
} /* tape_read_keys() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_read_reserves                                                    *
*                                                                              *
*  PURPOSE: read persistent reservations from a device                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: (output) the buffer to receive the reservation data                  *
*    buf_len: length allocated for keys including PR-in eight-byte header      *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_read_reserves(struct tape_descriptor **drvp, void *buf,
	ulong buf_len)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;

	DbgPrint(("lin_tape: tape_read_reserves entry.\n"));

	rc = persistent_reserve_in(buf, drv, QUERY_RESERVATIONS, buf_len);

	if(rc != STATUS_SUCCESS) {
		/* check sense data; if op_code error, reset rc to -ENOSYS */
		if(drv->sense_valid &&
			drv->sense_data->sense_key == 0x05 &&
			drv->sense_data->asc == 0x20 &&
			drv->sense_data->ascq == 0x00) {
			DbgPrint(("lin_tape: microcode level does not support "\
				"this function.\n"));

			drv->sense_valid = False;
			rc = -ENOSYS;
		} /* if */
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_read_reserves exit: %d.\n", rc));
	return rc;
} /* tape_read_reserves() */

/*******************************************************************************
*                                                                              *
*        SCSI Core Functions                                                   *
*                                                                              *
*******************************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: tape_send_scsi_cmd                                                    *
*                                                                              *
*  PURPOSE: pass command to tape_send_scsi_io and handle error recovery        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    timeout: of the command being processed                                   *
*    direction: data to device or data from device                             *
*    xfer_cnt: bytes transferred                                               *
*    retries: number of times this function should retry on failure            *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_send_scsi_cmd(struct tape_descriptor **drvp, int timeout,
	unchar direction, size_t xfer_cnt, int retries)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: tape_send_scsi_cmd entry.\n"));

	rc = tape_send_scsi_io(drv, timeout, direction, xfer_cnt, retries);
	if(!drv) {
		printk("lin_tape: tape_send_scsi_cmd device removed!\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(rc == STATUS_SUCCESS) goto EXIT_LABEL;

	if(!drv->cmd_in_open && !drv->cmd_wo_ua_erp) {
		if(drv->simmim_avail || drv->dmp_avail
			|| drv->full_logsense_avail)
			tape_check_simmim_dump_logsense(drv);
	} /* if */

	if(!drv->cmd_in_open && !drv->cmd_wo_ua_erp &&
		(drv->last_op != RELEASE) &&
		(drv->sense_valid && drv->sense_data
		&& drv->sense_data->sense_key == 6))
		tape_unit_attent_erp(drv);

	if(!drv->cmd_in_open && !drv->cmd_wo_ua_erp) {
		if((drv->drv_id == IBM_3590 || drv->drv_id == IBM_3592) &&
			(drv->sense_valid && drv->sense_data)) {
			if((drv->sense_data->sense_key == ABORTED_COMMAND)
				&& (drv->sense_data->asc == 0x44)
				&& (drv->sense_data->ascq == 0x0))
				tape_post_check1_erp(drv);
		} /* if */
	} /* if */

	/* handle fail over and encryption error recovery */
	if(drv->do_encryption) {
		drv->result = rc;
		rc = tape_perform_encryption(&drv, timeout, direction,
			xfer_cnt, retries);
		*drvp = drv;
	} else if(drv->do_failover) {
		drv->result = rc;
		rc = tape_perform_failover(&drv, timeout, direction,
			xfer_cnt, retries);

		*drvp = drv;
	} /* else if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_send_scsi_cmd exit: %d\n", rc));
	return rc;
} /* tape_send_scsi_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_send_scsi_io                                                     *
*                                                                              *
*  PURPOSE: build CDB and SCSI request, and send command to scsi_mod module    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: descriptor representing the drive on which to perform the operation  *
*    timeout: of the command being processed                                   *
*    direction: data to device or data from device                             *
*    xfer_cnt: bytes transferred                                               *
*    retries: number of times this function should retry on failure            *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

ssize_t tape_send_scsi_io(struct tape_descriptor* drv, int timeout,
	unchar direction, size_t xfer_cnt, int retries)
{
	int deferred = -1;
	unchar* cdb = (unchar*)drv->cdb;
	int rc = STATUS_SUCCESS;
	unchar* sense_data = NULL;
	int status_code;
	int resid = 0;
	int busy_count = 0;
	wait_queue_head_t wq;

	init_waitqueue_head(&wq);


	DbgPrint(("lin_tape: tape_send_scsi_io entry\n"));

	if(!drv || !drv->dev_obj) {
		printk("lin_tape: %s tape_send_scsi_io no device object!\n",
			drv->devid);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->remove_pending || drv->ghost_device) {
		printk("lin_tape: IBMtape%d remove pending.\n", drv->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data) {
		printk("lin_tape: failed to allocate sense_data\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	do {
		deferred++;

		memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

		rc = tape_log_cdb(drv);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		/* send the command down to scsi_mod */
		if(cdb[0] == WRITE_6 || cdb[0] == READ_6) {

			if(drv->direct_io) {
				status_code = tape_execute_scsi_command(
					drv->dev_obj, cdb, direction,
					(char*)&(drv->sg_list->gather[0]),
					xfer_cnt, -drv->sg_list->use_sg,
					timeout, sense_data, &resid);
			} else {
				rc = tape_send_rw_indirect(drv, cdb, direction,
					xfer_cnt, timeout, sense_data, &resid,
					&status_code);
				if(rc != STATUS_SUCCESS && !drv->busy_retry) goto EXIT_LABEL;

			} /* else */

		} else {
			status_code = tape_execute_scsi_command(drv->dev_obj,
				cdb, direction, drv->buffer, drv->buf_len, 0,
				timeout, sense_data, &resid);
		} /* else */

		if(status_code == -ENOMEM){
			rc = status_code;
			goto EXIT_LABEL;
		}

		if(drv->busy_retry){
			if( status_byte(status_code) == SAM_STAT_BUSY
					|| status_byte(status_code) == BUSY){
				busy_count++;
				if(busy_count < min ((int)drv->busy_retry,(int)BUSY_RETRY))
					wait_event_interruptible_timeout(wq,
						FALSE, HZ/2);
				else
					busy_count = 0;
			}else busy_count = 0;

		}

		rc = tape_check_result(drv, status_code, sense_data, resid);


	} while((rc == -EAGAIN && deferred < MAX_RETRIES) || (busy_count) );

EXIT_LABEL:

	if(sense_data) kfree(sense_data);

	DbgPrint(("lin_tape: tape_send_scsi_io exit: %d\n", rc));
	return rc;
} /* tape_send_scsi_io() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_log_cdb                                                          *
*                                                                              *
*  PURPOSE: Write drive's current CDB to the trace log                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: descriptor representing the drive for which to log the command       *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int tape_log_cdb(struct tape_descriptor* drv)
{
	unchar *trc = NULL;
	int trc_len = 0, rc = 0;
	unchar* cdb = (unchar*)drv->cdb;
	char ttype;

	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {

		trc_len = 1 + drv->cdb_len;
		trc = kmalloc(trc_len, GFP_KERNEL);

		if(!trc) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		trc[0] = drv->cdb_len;
		memcpy(&trc[1], drv->cdb, drv->cdb_len);
		tape_setup_and_log_full(drv, drv->cur_path, TRC_CDB, trc,
			trc_len);

		if(drv->buf_len &&
			(cdb[0] == MODE_SELECT || cdb[0] == DISPLAY_MESSAGE ||
			cdb[0] == SEND_DIAGNOSTIC))
			tape_log_buffer_data(drv, min((int) (drv->buf_len),
				(int)(FullTraceUnitSize-7)), drv->buffer);
	} /* if */

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;
} /* tape_log_cdb() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_send_rw_indirect                                                 *
*                                                                              *
*  PURPOSE: Send SCSI IO to lower layers when read/write, indirect IO          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: descriptor representing the drive for which to log the command       *
*    cdb: the command to send                                                  *
*    direction: to or from device                                              *
*    xfer_cnt: length of transfer                                              *
*    timeout: time allocated to the transfer                                   *
*    sense_data: buffer to receive sense data from the device                  *
*    resid: residual count from the transfer                                   *
*    status_code: return code from the execute command to analyze              *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    There are three methods for tape_execute_scsi_command:                    *
*                                                                              *
*      1) Direct DMA buffer                                                    *
*      2) Single ScatterGather                                                 *
*      3) Multiple ScatterGather                                               *
*                                                                              *
*******************************************************************************/

int tape_send_rw_indirect(struct tape_descriptor* drv, unchar* cdb,
	unchar direction, size_t xfer_cnt, int timeout, unchar* sense_data,
	int* resid, int* status_code)
{
	int rc = STATUS_SUCCESS;
	char* xfer_data = NULL;
	uint pgs = 0;

	DbgPrint(("lin_tape: tape_send_rw_indirect entry.\n"));

	if(BUF_RW(drv)) {
		*status_code = tape_execute_scsi_command(drv->dev_obj, cdb,
			direction, drv->rw_buf, xfer_cnt, 0, timeout,
			sense_data, resid);
		if (*status_code == -ENOMEM) rc=-ENOMEM;
	} else if(SGL_RW(drv)) {
		if(drv->sg_list->use_sg == 1) {
			if(xfer_cnt > drv->sg_list->gather[0].length) {
				printk("lin_tape: corrupted single sg\n");
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			xfer_data = sg_virt(&drv->sg_list->sgl_for_io[0]);
			pgs = 0;
		} else {
			if(drv->sg_list->use_sg > drv->sg_list->total_sgl) {
				printk("lin_tape: scatter list usage "\
					"overflow\n");
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			xfer_data = (char*)&drv->sg_list->sgl_for_io[0];
			pgs = drv->sg_list->use_sg;
		} /* else */
		*status_code = tape_execute_scsi_command(drv->dev_obj, cdb,
			direction, xfer_data, xfer_cnt, pgs, timeout,
			sense_data, resid);
		if (*status_code == -ENOMEM) rc=-ENOMEM;
	} else {
		printk("lin_tape: fatal error: corrupted buffer!\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* else */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_send_rw_indirect exit: %d\n", rc));
	return rc;
} /* tape_send_rw_indirect() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_execute_scsi_command                                             *
*                                                                              *
*  PURPOSE: Abstracts the interface for sending SCSI commands.                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    scsi_device: the device to receive the command                            *
*    cdb: the command to send                                                  *
*    direction: to or from device                                              *
*    buffer: used for input/output                                             *
*    buf_len: length of buffer                                                 *
*    use_sg: use vector (scatter-gather) operations in the transfer            *
*    timeout: time allocated to this command in seconds                        *
*    sense_data: sense data from the device                                    *
*    resid: residual count from the transfer                                   *
*                                                                              *
*  RETURNS: status code (driver byte, host byte, msg byte, status byte)        *
*                                                                              *
*******************************************************************************/

int tape_execute_scsi_command(struct scsi_device* sdev, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int *resid)
{
	int status_code = STATUS_SUCCESS;
	DbgPrint(("lin_tape: tape_execute_scsi_command entry: 0x%x \n",
		cdb[0] & 0xFF));

    if(!sdev) {
		printk("lin_tape: remove pending.\n");
		status_code = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && (!defined(CONFIG_SLE_SP))
	status_code = lin_tape_do_req(sdev, cdb, direction, buffer, buf_len,
		use_sg, timeout, sense_data, resid);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
	status_code = lin_tape_execute_async(sdev, cdb, direction, buffer,
		buf_len, use_sg, timeout, sense_data, resid);
#else
	status_code = lin_tape_blk_rq(sdev, cdb, direction, buffer, buf_len,
		use_sg, timeout, sense_data, resid);
#endif

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_execute_scsi_command exit: 0x%x\n",
		status_code));
	return status_code;
} /* tape_execute_scsi_command() */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && (!defined(CONFIG_SLE_SP))
/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_do_req                                                       *
*                                                                              *
*  PURPOSE: Issue SCSI command via scsi_do_req                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: scsi_device structure representing target device for the request    *
*    cdb: the command                                                          *
*    direction: to device or from device                                       *
*    buffer: data to transfer, if any                                          *
*    buf_len: length of buffer in bytes                                        *
*    use_sg: if not 0, contains number of scatter-gather pages to use          *
*    timeout: maximum time taken to complete the command                       *
*    sense_data: output buffer for sense data on completion                    *
*    resid: residual count for read-write transfers                            *
*                                                                              *
*  RETURNS: status code (driver byte, host byte, msg byte, status byte)        *
*                                                                              *
*******************************************************************************/

int lin_tape_do_req(struct scsi_device* sdev, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid)
{
	int status_code = 0;
	struct scsi_request *srb = NULL;
	struct completion srb_complete;

	if(use_sg < 0) use_sg = -use_sg;
	srb = scsi_allocate_request(sdev, GFP_KERNEL);
	if(!srb) {
		printk("lin_tape: unable to allocate srb\n");
		status_code = (DRIVER_ERROR << 24);
		goto EXIT_LABEL;
	} /* if */

	init_completion(&srb_complete);
	srb->sr_data_direction = direction;
	srb->sr_cmd_len = (cdb[0] == DISPLAY_MESSAGE) ? CDB_GENERIC_6 : 0;
	srb->sr_request->waiting = &srb_complete;
	srb->sr_request->rq_status = RQ_SCSI_BUSY;

	srb->sr_use_sg = use_sg;
	scsi_do_req(srb, cdb, (char*)buffer, buf_len, tape_callback,
		timeout, False);

	wait_for_completion(&srb_complete);

	srb->sr_request->waiting = NULL;

	memcpy(sense_data, srb->sr_sense_buffer, SCSI_SENSE_BUFFERSIZE);
	status_code = srb->sr_result;
	*resid = srb->sr_command->resid;

	if(srb) {
		scsi_release_request(srb);
		srb = NULL;
	} /* if */

EXIT_LABEL:

	return status_code;

} /* lin_tape_do_req() */

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_execute_async                                                *
*                                                                              *
*  PURPOSE: Issue SCSI command via scsi_execute_async                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: scsi_device structure representing target device for the request    *
*    cdb: the command                                                          *
*    direction: to device or from device                                       *
*    buffer: data to transfer, if any                                          *
*    buf_len: length of buffer in bytes                                        *
*    use_sg: if not 0, contains number of scatter-gather pages to use          *
*    timeout: maximum time taken to complete the command                       *
*    sense_data: output buffer for sense data on completion                    *
*    resid: residual count for read-write transfers                            *
*                                                                              *
*  RETURNS: status code (driver byte, host byte, msg byte, status byte)        *
*                                                                              *
*******************************************************************************/

int lin_tape_execute_async(struct scsi_device* sdev, unchar* cdb,
	int direction, void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid)
{
	int status_code = 0;
	struct scsi_execute_callback *callback = NULL;

	if(use_sg < 0) use_sg = -use_sg;
	callback = kmalloc(sizeof(struct scsi_execute_callback), GFP_KERNEL);
	if(!callback) {
		printk("lin_tape: kmalloc for callback failed\n");
		status_code = (DRIVER_ERROR << 24);
		goto EXIT_LABEL;
	} /* if */

	memset(callback, '\0', sizeof(struct scsi_execute_callback));

	init_completion(&callback->scsi_execute_lock);
	status_code = scsi_execute_async(sdev, cdb, COMMAND_SIZE(cdb[0]),
		direction, buffer, buf_len, use_sg, timeout, 0,
		(void*)callback, tape_execute_callback, GFP_KERNEL);

	if(STATUS_SUCCESS == status_code) {
		wait_for_completion(&callback->scsi_execute_lock);

		status_code = callback->status_code;
		memcpy(sense_data, callback->sense_data, SCSI_SENSE_BUFFERSIZE);
		*resid = callback->resid;
	} /* if */

	if(callback) {
		kfree(callback);
		callback = NULL;
	} /* if */
EXIT_LABEL:

	return status_code;

} /* lin_tape_execute_async() */

#else
/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_blk_rq                                                       *
*                                                                              *
*  PURPOSE: Issue SCSI command via blk_execute_rq                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: scsi_device structure representing target device for the request    *
*    cdb: the command                                                          *
*    direction: to device or from device                                       *
*    buffer: data to transfer, if any                                          *
*    buf_len: length of buffer in bytes                                        *
*    use_sg: if > 0, scatter-gather is allocated.  Number of sg pages to use   *
*    timeout: maximum time taken to complete the command                       *
*    sense_data: output buffer for sense data on completion                    *
*    resid: residual count for read-write transfers                            *
*                                                                              *
*  RETURNS: status code (driver byte, host byte, msg byte, status byte)        *
*                                                                              *
*******************************************************************************/

int lin_tape_blk_rq(struct scsi_device* sdev, unchar* cdb, int direction,
	void* buffer, int buf_len, int use_sg, int timeout,
	unchar* sense_data, int* resid)
{
	int status_code = 0, rc = 0, rw = (direction == DMA_TO_DEVICE), i = 0,
		alloc_len = 0, nr_ents = 0, order = 0;
	struct request* req = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
    struct scsi_request *srq = NULL;
#endif
	struct rq_map_data* mdata = NULL;
	struct page* pg = NULL;
	struct bio* bio = NULL;
	void* tmp_rq = NULL;
	int direct_io = use_sg < 0, remain_len = buf_len, copy_size = 0;
	boolean allocd = False;
	char* buf_ptr = (char*)buffer;

	if(use_sg < 0) use_sg = -use_sg;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,11,0)
	req = blk_get_request(sdev->request_queue, rw, GFP_KERNEL);
#else
	req = blk_get_request(sdev->request_queue, rw? REQ_OP_SCSI_OUT : REQ_OP_SCSI_IN, GFP_KERNEL);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,0,0)
	if(!req) {
#else
    if(IS_ERR(req)) {
#endif
		printk("lin_tape: cannot get request for SCSI command\n");
		status_code =  DRIVER_ERROR << 24;
		goto EXIT_LABEL;
	} /* if */
	tmp_rq = (void*)sdev->request_queue;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
	srq = scsi_req(req);
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,11,0)
	req->cmd_type = REQ_TYPE_BLOCK_PC;
#endif
#ifdef REQ_QUIET
	req->cmd_flags |= REQ_QUIET;
#else
	req->rq_flags |= RQF_QUIET;
#endif

	if(buf_len) {
		mdata = kmalloc(sizeof(struct rq_map_data), GFP_KERNEL);
		if(!mdata) {
			status_code = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(mdata, '\0', sizeof(struct rq_map_data));

		mdata->page_order = 0;

		for(i = PAGE_SIZE; i < buf_len; i *= 2) order++;

		if(!direct_io) mdata->page_order = order;

		nr_ents = use_sg ? use_sg :
			DIV_ROUND_UP(buf_len, PAGE_SIZE << mdata->page_order);
		alloc_len = nr_ents * sizeof(struct page*);

		mdata->pages = kmalloc(alloc_len, GFP_KERNEL);
		if(!mdata->pages) {
			status_code = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(mdata->pages, '\0', alloc_len);

		mdata->offset = use_sg
			? ((struct scatterlist*)buffer)->offset : 0;

		for(i = 0; i < nr_ents; i++) {
			if(use_sg) {
				mdata->pages[i] = sg_page(
					&((struct scatterlist*)buffer)[i]);
			} else {
				pg = alloc_pages(GFP_KERNEL, mdata->page_order);
				if(!pg) {
					status_code = -ENOMEM;
					goto EXIT_LABEL;
				} /* if */
				allocd = True;
				mdata->pages[i] = pg;

				copy_size = min(remain_len,
					(int)PAGE_SIZE << mdata->page_order);

				memcpy(page_address(pg), buf_ptr, copy_size);
				buf_ptr += copy_size;
				remain_len -= copy_size;
			} /* else */
			mdata->nr_entries++;
		} /* for */

		mdata->null_mapped = 1;

		if (mdata->page_order > MAX_PAGE_ORDER)
				mdata->page_order = MAX_PAGE_ORDER;
		rc = blk_rq_map_user(req->q, req, mdata, NULL, buf_len,
			GFP_KERNEL);

		if(rc) {
			printk("lin_tape: blk_rq_map_user rc %d cmd_flags %d\n",-rc,(int) req->cmd_flags & 1);
			status_code = DRIVER_ERROR << 24;
			goto EXIT_LABEL;
		} /* if */
		bio = req->bio;
	} /* if */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,11,0)
	req->cmd_len = COMMAND_SIZE(cdb[0]);
	memset(req->cmd, '\0', MAX_CDB_SIZE);
	memcpy(req->cmd, cdb, req->cmd_len);
	req->sense = sense_data;
	req->sense_len = 0;
	req->retries = 0;
#else
    srq->cmd_len = COMMAND_SIZE(cdb[0]);
	memset(srq->cmd, '\0', MAX_CDB_SIZE);
	memcpy(srq->cmd, cdb, srq->cmd_len);
	srq->retries = 0;
#endif
	req->timeout = timeout;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,12,0)
	blk_execute_rq(req->q, NULL, req, 1);
#else
	blk_execute_rq(NULL, req, 1);
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4,11,0)
	*resid = req->resid_len;
	status_code = req->errors;
#else
	*resid = srq->resid_len;
	status_code = srq->result;
	if (srq->sense_len)
		memcpy(sense_data, srq->sense, SCSI_SENSE_BUFFERSIZE);
#endif
    if(!scsi_block_when_processing_errors(sdev)) {
            printk("scsi block errors\n");
        status_code = DRIVER_ERROR << 24;
        goto EXIT_LABEL;
    }
	if(buf_len && !use_sg) {
		buf_ptr = (char*)buffer;
		remain_len = buf_len;
		for(i = 0; i < mdata->nr_entries; i++) {
			copy_size = min(remain_len,
				(int)PAGE_SIZE << mdata->page_order);
			memcpy(buf_ptr, page_address(mdata->pages[i]),
				copy_size);
			buf_ptr += copy_size;
			remain_len -= copy_size;
		} /* for */
	} /* if */

EXIT_LABEL:
	if(bio) blk_rq_unmap_user(bio);

	if(mdata) {
		if(mdata->pages) {
			if(allocd) {
				for(i = 0; i < mdata->nr_entries; i++) {
					__free_pages(mdata->pages[i], order);
				} /* for */
			} /* if */
			kfree(mdata->pages);
		} /* if */
		kfree(mdata);
	} /* if */
	if(req && sdev && tmp_rq == sdev->request_queue) {
		blk_put_request(req);
	}
	return status_code;
} /* lin_tape_blk_rq() */
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17) && !defined(CONFIG_SLE_SP)

/*******************************************************************************
*                                                                              *
*  NAME: tape_callback                                                         *
*                                                                              *
*  PURPOSE: Completion routine for scsi_execute_async                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    scsi_cmd: represents the command that we are completing                   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES: Used in kernels prior to RHEL 5                                      *
*                                                                              *
*******************************************************************************/

void tape_callback(struct scsi_cmnd *scsi_cmd)
{
	if(!scsi_cmd) {
		printk("lin_tape: fatal error: no scsi_cmd passed in\n");
		return;
	}

	scsi_cmd->request->rq_status = RQ_SCSI_DONE;

	if(scsi_cmd->request->waiting)
		complete(scsi_cmd->request->waiting);

	return;

}/* tape_callback() */

#else

/*******************************************************************************
*                                                                              *
*  NAME: tape_execute_callback                                                 *
*                                                                              *
*  PURPOSE: Completion routine for scsi_execute_async                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    data: data context (the callback data structure)                          *
*    sense_data: sense data from the device                                    *
*    status_code: status from the operation that just completed                *
*    resid: residual count from the transfer                                   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES: Used in kernels starting with RHEL 5                                 *
*                                                                              *
*******************************************************************************/

void tape_execute_callback(void* data, char* sense_data, int status_code,
	int resid)
{
	struct scsi_execute_callback* callback
		= (struct scsi_execute_callback*)data;

	memcpy(callback->sense_data, sense_data, SCSI_SENSE_BUFFERSIZE);
	callback->status_code = status_code;
	callback->resid = resid;
	complete(&callback->scsi_execute_lock);
} /* tape_execute_callback() */

#endif

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_result                                                     *
*                                                                              *
*  PURPOSE: Check the result from scsi_mod module                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    result: the result sent back by scsi_mod                                  *
*    sense_buffer: sense data from the drive                                   *
*    resid: residual count from the transfer                                   *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int tape_check_result(struct tape_descriptor *drv, int result,
	unchar *sense_buffer, int resid)
{
	int rc = -EIO;
	char ttype;
	unchar* trc = NULL;

	DbgPrint(("lin_tape: tape_check_result entry.\n"));

	trc = kmalloc(ONE_K, GFP_KERNEL);
	if(!trc) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(trc, '\0', ONE_K);

	if(!drv || !sense_buffer) goto EXIT_LABEL;

	drv->result = result;

	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	tape_check_log_cdb_and_status(drv, result, ttype, trc, sense_buffer);

	if(drv->in_erp) {
		if(result == STATUS_SUCCESS) rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	if(TapeMotion(drv->last_op)) drv->last_tape_motion = drv->last_op;

	/* If result is STATUS_SUCCESS, reset sense data and residual count */
	if(result == STATUS_SUCCESS) {
		tape_check_reset_on_success(drv, resid);
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */
	if(drv->devid && host_byte(result) == DID_ERROR)
		printk("lin_tape: %s driver_byte %02X, host_byte %02X, "\
			"msg_byte %02X, status_byte %02X\n",
			drv->devid, driver_byte(result),
			host_byte(result), msg_byte(result),
			status_byte(result));
	/* If we have sense data, save it */
	if(sense_buffer &&
                    (driver_byte(result) & DRIVER_SENSE
                      ||
                     status_byte(result) == CHECK_CONDITION))
	{
		tape_check_handle_sense(drv, result, &rc, sense_buffer, trc,
			ttype);
	} else {
		tape_check_handle_nonsense(drv, result, &rc);
	} /* else */

EXIT_LABEL:

	if(trc) kfree(trc);
	DbgPrint(("lin_tape: tape_check_result exit: %d\n", rc));
	return rc;

} /* tape_check_result() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_log_cdb_and_status                                         *
*                                                                              *
*  PURPOSE: Log CDB and sense data to system log if conditions permit          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    result: the result sent back by scsi_mod                                  *
*    ttype: the tracing policiy currently in effect for the driver             *
*    trc: working buffer (passed in to avoid multiple large stack allocations) *
*    sense_buffer: sense data from the drive                                   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_log_cdb_and_status(struct tape_descriptor* drv, int result,
	char ttype, char* trc, unchar* sense_buffer)
{
	int trc_len = 0;
	char SK, ASC, ASCQ, ILI, EOM, FM;

	if(!drv || !drv->trace) return;

	if(ttype == LIN_TAPE_FULL_TRACE) {

		trc_len = 4;
		trc[0] = driver_byte(result);
		trc[1] = host_byte(result);
		trc[2] = msg_byte(result);
		trc[3] = status_byte(result);

		tape_setup_and_log_full(drv, drv->cur_path, TRC_STATUS, trc,
			trc_len);
		return;
	} /* if */

	if(result == STATUS_SUCCESS) return;

	if(ttype == LIN_TAPE_NO_TRACE || ttype == LIN_TAPE_REGULAR_TRACE) {

		if((driver_byte(result) & DRIVER_SENSE) && sense_buffer) {
			SK = sense_buffer[2] & 0x0F;
			ILI = sense_buffer[2] & 0x20;
			EOM = sense_buffer[2] & 0x40;
			FM = sense_buffer[2] & 0x80;
			ASC = sense_buffer[12];
			ASCQ = sense_buffer[13];
			if((SK == 0x00 && (ILI || EOM || FM
				|| (ASC == 0x00 && ASCQ == 0x01)
				|| (ASC == 0x00 && ASCQ == 0x05)))

			|| (SK == 0x02 && ASC == 0x04 && ASCQ == 0x01)

			|| (SK == 0x06)

			|| (SK == 0x08 && ASC == 0x00 && ASCQ == 0x05))
				return;
		} /* if */

		trc_len = 1 + drv->cdb_len;

		trc[0] = drv->cdb_len;

		if (!drv || !&trc[1] || drv->cdb_len<=0 || !drv->cdb ) return;
		memcpy(&trc[1], drv->cdb, drv->cdb_len);

		tape_setup_and_log_full(drv, drv->cur_path, TRC_CDB, trc,
			trc_len);

		memset(trc, '\0', trc_len);
		trc_len = 4;

		trc[0] = driver_byte(result);
		trc[1] = host_byte(result);
		trc[2] = msg_byte(result);
		trc[3] = status_byte(result);

		tape_setup_and_log_full(drv, drv->cur_path, TRC_STATUS, trc,
			trc_len);

	} /* if */

} /* tape_check_log_cdb_and_status() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_reset_on_success                                           *
*                                                                              *
*  PURPOSE: Log CDB and sense data to system log if conditions permit          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    resid: residual count returned from drive                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_reset_on_success(struct tape_descriptor* drv, int resid)
{
	drv->sense_valid = False;
	drv->resid_valid = False;
	drv->ili = False;

	if(TapeMotion(drv->last_op)) drv->fm_detected = False;

	if(InvalidEOMFlag(drv->last_op)) drv->eom = False;

	if(InvalidEODFlag(drv->last_op)) drv->eod = False;

	/* sili bit support for underlength read */
	if(drv->last_op == READ_6 && drv->sili) {
		drv->resid_valid = True;
		if(resid > 0) {
			drv->resid = resid;
			DbgPrint(("lin_tape: Residual count returned "\
				"from HBA: %d\n", resid));
		} else drv->resid = 0;
	} /* if */

} /* tape_check_reset_on_success() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_handle_sense                                               *
*                                                                              *
*  PURPOSE: Interpret sense data returned from device                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    result: the status of the returned command                                *
*    rc: (input/output) the return code from this portion of the result check  *
*    sense_buffer: buffer containing sense data returned from the device       *
*    trc: working buffer (passed in to avoid multiple large stack allocations) *
*    ttype: trace type                                                         *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_handle_sense(struct tape_descriptor* drv, int result, int* rc,
	unchar* sense_buffer, char* trc, char ttype)
{
#ifdef CONFIG_PPC64
	if((status_byte(result) == STATUS_SUCCESS) &&
		(drv->last_op == REQUEST_SENSE)) {
		*rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */
#endif

	if(!drv->sense_data) {
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &(drv->tape_log_len),
				TRC_MSG_ERR_NO_SENSE);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_MSG_ERR_NO_SENSE, 0, NULL);

		*rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* save the sense data */
	memset(drv->sense_data, 0, sizeof(struct tape_sense_data));
	memcpy((void*)drv->sense_data, sense_buffer, min(SCSI_SENSE_BUFFERSIZE,
		(int)sizeof(struct tape_sense_data)));

	if(drv->trace && (ttype != LIN_TAPE_INVALID_TRACE))
		tape_check_write_sense_to_log(drv, ttype, trc);

	drv->sense_valid = True;
	drv->resid_valid = drv->sense_data->valid;

	if(drv->sense_data->valid) tape_check_save_sense_information(drv);

	tape_check_sense_motion_commands(drv, ttype);

	tape_check_sense_key_specific(drv, rc);

	tape_check_daemon_functions(drv, rc, ttype);

EXIT_LABEL:

	return;

} /* tape_check_handle_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_handle_nonsense                                            *
*                                                                              *
*  PURPOSE: Interpret non-sense error data returned from device                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    result: of the command                                                    *
*    rc: (input/output) the return code from this portion of the result check  *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_handle_nonsense(struct tape_descriptor* drv, int result,
	int* rc)
{

	drv->sense_valid = False;
	if(drv->sense_data)
		memset(drv->sense_data, '\0', sizeof(struct tape_sense_data));

	/* reservation conflict */
	if(status_byte(result) == SAM_STAT_RESERVATION_CONFLICT
		|| status_byte(result) == RESERVATION_CONFLICT) {
		if (TP_RSV_PERSISTENT(drv))	log_reservation_conflict(drv);
		*rc = -EBUSY;
	} /* if */

	/* check failover error condition */
	if(drv->alt_pathing && drv->next_path!=drv && !drv->cmd_wo_ua_erp &&
		drv->last_op != PERSISTENT_RESERVE_IN &&
		drv->last_op != PERSISTENT_RESERVE_OUT) {
		/* under following conditions we do failover
		* see drivers/scsi/scsi.h for these macro definitions
		*/
		if(host_byte(result) != DID_OK
			&& host_byte(result) != DID_TARGET_FAILURE)
			drv->do_failover = True;

		if(driver_byte(result) == DRIVER_TIMEOUT)
			drv->do_failover = True;
		if(driver_byte(result) == DRIVER_ERROR)
			drv->do_failover = True;
	} /* if */

} /* tape_check_handle_nonsense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_write_sense_to_log                                         *
*                                                                              *
*  PURPOSE: Log sense data to error log                                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    ttype: trace type                                                         *
*    trc: working buffer (passed in to avoid multiple large stack allocations) *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_write_sense_to_log(struct tape_descriptor* drv, char ttype,
	unchar* trc)
{
	char SK = 0, ASC = 0, ASCQ = 0, EOM = 0, FM = 0, ILI = 0;
	int trc_len;

	if(ttype == LIN_TAPE_REGULAR_TRACE) {
		tape_log_msg4(drv->tape_log, &(drv->tape_log_len),
			TRC_MSG_SENSE, drv->sense_data->sense_key,
			drv->sense_data->asc, drv->sense_data->ascq);
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = min(SCSI_SENSE_BUFFERSIZE,
			(int)sizeof(struct tape_sense_data));
		memcpy(trc, (unchar*)drv->sense_data, trc_len);
		tape_setup_and_log_full(drv, drv->cur_path, TRC_SENSE, trc,
			trc_len);
		return;
	} /* if */

	SK = drv->sense_data->sense_key;
	ASC = drv->sense_data->asc;
	ASCQ = drv->sense_data->ascq;
	EOM = drv->sense_data->eom;
	FM = drv->sense_data->fm;
	ILI = drv->sense_data->ili;

	/* Do not log for "normal" operations */
	if((SK == 0x00 && (ILI || EOM || FM
		|| (ASC == 0x00 && ASCQ == 0x01)
		|| (ASC == 0x00 && ASCQ == 0x05)))

	|| (SK == 0x02
		&& ((ASC == 0x04 && ASCQ == 0x01)
		|| (ASC == 0x3A && ASCQ == 0x00)))

	|| (SK == 0x06)

	|| (SK == 0x08 && ASC == 0x00 && ASCQ == 0x05)

	/* AME */
	|| (((drv->sense_data->asc == 0xEF)
		|| (drv->sense_data->asc == 0xEE)) &&
		((drv->application_set_encryption)
			&& (drv->drv_id == IBM_3580)))) return;

	trc_len = min(SCSI_SENSE_BUFFERSIZE,
		(int)sizeof(struct tape_sense_data));

	memcpy(trc, (unchar*)drv->sense_data, trc_len);
	tape_setup_and_log_full(drv, drv->cur_path, TRC_SENSE, trc, trc_len);

} /* tape_check_write_sense_to_log() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_save_sense_information                                     *
*                                                                              *
*  PURPOSE: Copy information field from sense data to drive descriptor         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    If the valid sense is set, we save the information field                  *
*    If the information field is unsigned int and we shift so that             *
*    1's won't over lap each other then the final residual value               *
*    should still be two's comp number, example:                               *
*                  11110000  (240  unsigned)                                   *
*            + 100100000000  (2304 unsigned)                                   *
*            --------------                                                    *
*            = 100111110000  (2544 unsigned)                                   *
*                                                                              *
*******************************************************************************/

void tape_check_save_sense_information(struct tape_descriptor* drv)
{
	int information;

	information = (uint) (drv->sense_data->information[0] << 24)
		+ (drv->sense_data->information[1] << 16)
		+ (drv->sense_data->information[2] << 8)
		+ drv->sense_data->information[3];

	/*
	* If command was fixed block write, the actual residual
	* value is the information field times block size
	*/
	if(!drv->sili) {
		if((drv->last_op == READ_6 || drv->last_op == WRITE_6)
			&& drv->blk_size)
			drv->resid = information * drv->blk_size;
		else
			drv->resid = information;
	} else {
		/* for overlength read forced check condition */
		drv->resid = information;
	} /* else */

	tape_log_msg5(drv->tape_log, &(drv->tape_log_len),
		TRC_MSG_INFO,
		drv->sense_data->information[0],
		drv->sense_data->information[1],
		drv->sense_data->information[2],
		drv->sense_data->information[3]);

} /* tape_check_save_sense_information() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sense_motion_commands                                      *
*                                                                              *
*  PURPOSE: Update drive descriptor for read/write and other motion commands   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    ttype: trace type                                                         *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sense_motion_commands(struct tape_descriptor* drv, char ttype)
{
	/* update the EOM flag, only valid on WRITE or WRITE_FILEMARK */
	if(((drv->last_op == WRITE_6) || (drv->last_op == WRITE_FILEMARKS) ||
		(drv->last_op == WRITE_FILEMARK_ZERO)) &&
		((drv->sense_data->asc != 0x00) ||
		(drv->sense_data->ascq != 0x04))) {
		drv->eom = drv->sense_data->eom;

		if(drv->eom) {
			/* encountered eom */
			if(drv->trace && ttype==LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(drv->tape_log,
						&(drv->tape_log_len),
						TRC_MSG_EOM);
		} /* if */

	} /* if */

	if(drv->last_op == READ_6) {
		drv->eod = drv->sense_data->eom;
		if((drv->sense_data->sense_key == 0x08) &&
			(drv->sense_data->asc == 0x00) &&
			(drv->sense_data->ascq == 0x05))
			drv->eod = True;

		if(drv->eod == True) {
			/* encountered eod */
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(drv->tape_log,
						&(drv->tape_log_len),
						TRC_MSG_EOD);
		} /* if */

	} /* if */

	if((drv->sense_data->valid) && (TapeMotion(drv->last_op)))
		drv->fm_detected = drv->sense_data->fm;

	if(drv->fm_detected) {
		/* encountered file mark */
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &(drv->tape_log_len),
					TRC_MSG_FM);
	} /* if */

	/* update illegal length indicator field (mainly used for RW) */
	drv->ili = drv->sense_data->ili;

} /* tape_check_sense_motion_commands() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sense_key_specific                                         *
*                                                                              *
*  PURPOSE: Sense-key specific processing                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sense_key_specific(struct tape_descriptor* drv, int* rc)
{
	if(drv->sense_data->sense_key != NO_SENSE) {
		DbgPrint(("lin_tape: %s sense key: %x asc: %x ascq: %x\n",
			drv->devid,
			drv->sense_data->sense_key,
			drv->sense_data->asc,
			drv->sense_data->ascq));
	} /* if */

	switch(drv->sense_data->sense_key) {
		case NO_SENSE:
			tape_check_sk0(drv, rc);
			break;
		case RECOVERED_ERROR:
			tape_check_sk1(drv, rc);
			break;
		case NOT_READY:
			tape_check_sk2(drv, rc);
			break;
		case ILLEGAL_REQUEST:
			tape_check_sk5(drv, rc);
			break;
		case UNIT_ATTENTION:
			tape_check_sk6(drv, rc);
			break;
		case DATA_PROTECT:
			*rc = -EPERM;
			break;
		case MISCOMPARE:
			*rc = -EINVAL;
			break;
		case ABORTED_COMMAND:
			tape_check_skB(drv, rc);
			break;
		default:
			*rc = -EIO;
			break;
	} /* switch */

} /* tape_check_sense_key_specific() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_daemon functions                                           *
*                                                                              *
*  PURPOSE: Miscellaneous IBM proprietary post-processing handled by daemon    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: return code from this command                                         *
*    ttype: trace type                                                         *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_daemon_functions(struct tape_descriptor* drv, int* rc,
	char ttype)
{
	if((drv->sense_data->asc == 0xEF) &&
		(drv->encryption_keypath_configured) &&
		(drv->encryption_capable) &&
		(drv->encryption_proxy_entitled) &&
		(drv->sys_encryption_proxy)) {
		*rc = -EIO;
		drv->do_encryption = True;
		drv->do_failover = False;
	} /* if */

	/* Get the simmim for 3590s
	* To avoid nested scsi requests, we just set a flag here
	*/
	if(((drv->drv_id == IBM_3590) || (drv->drv_id == IBM_3592))
		&& (drv->disable_simmim_log == False)
		&& (drv->sense_data->sense_key != ILLEGAL_REQUEST)
		&& (drv->sense_data->sksv == False)
		&& ((drv->sense_data->bit_ptr) & 0x01)
		&& (ttype != LIN_TAPE_INVALID_TRACE)) {
		drv->simmim_avail = True;
	}

	/* Get the Drive Dumps
	* To avoid nested scsi requests, we just set a flag here
	*/
	if((!drv->disable_auto_drive_dump)
		&& (ttype != LIN_TAPE_INVALID_TRACE)) {
		if(((drv->drv_id == IBM_3580)
			&& (drv->sense_data->vendor_specific4 & 0x02))
			|| (((drv->drv_id == IBM_3590)
			|| (drv->drv_id == IBM_3592))
			&& ((drv->sense_data->vendor_specific1 & 0x20)
			|| (drv->sense_data->vendor_specific_sense[46] & 0x20))))
		drv->dmp_avail = True;
	} /* if */
} /* tape_check_daemon_functions() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sk0                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key 0 (no sense data) conditions                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sk0(struct tape_descriptor* drv, int* rc)
{
	/* ili and eom conditions */
	if(drv->sense_data->ili) {
		if(drv->resid_valid && drv->resid < 0)
			*rc = -EOVERFLOW;

		if(drv->last_op == VERIFY_TAPE_DATA)
			*rc = -EIO;
	} /* if */
} /* tape_check_sk0() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sk1                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key 1 (recovered error) conditions               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sk1(struct tape_descriptor* drv, int* rc)
{
	/* deferred condition, log max condition */
	if(drv->sense_data->asc == 0x5B &&
		drv->sense_data->ascq == 0x02 &&
		drv->auto_logsense == True) {
		/* set the flag to avoid nested scsi requests.
		* we issue full log sense command later
		*/
		drv->full_logsense_avail = True;
	}

	if(drv->sense_data->err_code == CommandDeferred)
		*rc = -EAGAIN;
	else
		*rc = STATUS_SUCCESS;
} /* tape_check_sk1() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sk2                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key 2 (not ready) conditions                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sk2(struct tape_descriptor* drv, int* rc)
{
	if(drv->sense_data->asc == 0x3A && drv->sense_data->ascq == 0x00)
	*rc = -ENOMEDIUM;
	else
    if (drv->sense_data->asc == 0x3E && drv->sense_data->ascq == 0x00){
        msleep(5000);
        *rc = -EAGAIN;
    }

} /* tape_check_sk2() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sk5                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key 5 (illegal request) conditions               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sk5(struct tape_descriptor* drv, int* rc)
{
	/* Some HBA series change sense key from 2 (Not Ready)
	* to 5 (Illegal Request) if the device has a LUN > 0
	* we need change the sense key back to 2
	*/
#ifdef CONFIG_PPC64
	char ASC = drv->sense_data->asc;
	char ASCQ = drv->sense_data->ascq;

	if((ASC == 0x04 && ASCQ == 0x00)
		|| (ASC == 0x04 && ASCQ == 0x01)
		|| (ASC == 0x04 && ASCQ == 0x02)
		|| (ASC == 0x04 && ASCQ == 0x03)
		|| (ASC == 0x04 && ASCQ == 0x04)
		|| (ASC == 0x04 && ASCQ == 0x07)
		|| (ASC == 0x30 && ASCQ == 0x03)
		|| (ASC == 0x30 && ASCQ == 0x07)
		|| (ASC == 0x3A && ASCQ == 0x00)
		|| (ASC == 0x3E && ASCQ == 0x00)
		|| (ASC == 0x53 && ASCQ == 0x00)) {
		drv->sense_data->sense_key = NOT_READY;
		*rc = -EIO;
		return;
	} /* if */
#endif

	*rc = -EINVAL;
} /* tape_check_sk5() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_sk6                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key 6 (unit attention) conditions                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_sk6(struct tape_descriptor* drv, int* rc)
{
	char ASC = drv->sense_data->asc;
	char ASCQ = drv->sense_data->ascq;

	if(ASC == 0x30 && ASCQ == 0x00)
		*rc = -EMEDIUMTYPE;

	if(ASC == 0x28 && ASCQ == 0x00) {
		/* Medium may have changed therefore all we know
		* about the the medium may no longer be correct
		*/
		drv->resid_valid = False;
		drv->fm_detected = False;
		drv->ili = False;
		drv->deferred_eom = False;
		drv->eom = False;
		drv->eom_blk_id = 0;

		drv->pos_valid = False;

		*rc = -EIO;
	} /* if */

	/*
	* Process the following asc/q in Unit Attn sense data
	* If the sense key is 6, and asc/ascq is the following,
	* the drive may lose the reservation and media may have
	* been changed.  We should not write 2 FMs on close
	* nor rewind the tape for rewindable device on close.
	* 28/00: Not Ready to Ready, Medium may have changed
	* 29/00: Power on, reset, or Bus reset occurred
	* 2A/01: Mode parameters changed
	* 2A/02: Log parameters changed
	* 2F/00: Commands cleared by another initiator
	* 3F/01: Microcode has been changed
	* 3F/02: Change operating definition
	* 3F/03: Inquiry data has changed
	*/
	if((drv->last_op == WRITE_6) || (drv->last_op == WRITE_FILEMARKS)) {
		if((ASC == 0x28 && ASCQ == 0x00)
			|| (ASC == 0x29 && ASCQ == 0x00)
			|| (ASC == 0x2A && ASCQ == 0x01)
			|| (ASC == 0x2A && ASCQ == 0x02)
			|| (ASC == 0x2F && ASCQ == 0x00)
			|| (ASC == 0x3F && ASCQ == 0x01)
			|| (ASC == 0x3F && ASCQ == 0x02)
			|| (ASC == 0x3F && ASCQ == 0x03)) {
			drv->wfm_on_close_safe = False;
		} /* if */
	} /* if */

} /* tape_check_sk6() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_skB                                                        *
*                                                                              *
*  PURPOSE: Handler for sense key B (ABORTED_COMMAND) conditions               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    rc: (input/output) the return code from this command                      *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/
void tape_check_skB(struct tape_descriptor* drv, int* rc)
{
    if(drv->sense_data->asc == 0x4b && drv->sense_data->ascq == 0x06){
        *rc = -EAGAIN;
    }
    else
    if  (drv->sense_data->asc == 0x29 && drv->sense_data->ascq == 0x07){
        msleep(5000);
        *rc = -EAGAIN;
    }

} /* tape_check_skB() */

/*******************************************************************************
*                                                                              *
*  NAME: log_reservation_conflict                                              *
*                                                                              *
*  PURPOSE: Record reservation handler for detected reservation conflict       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_reservation_conflict(struct tape_descriptor* drv)
{
	static int in_recursion = 0;
	int rc = 0, length = 0, i = 0, ds_len = 0;
	struct pr_full_status* buf = NULL;
	struct pr_full_status_descriptor* ds = NULL;

	DbgPrint(("lin_tape: log_reservation_conflict entry\n"));
	/* there is a danger of infinite recursion; block further calls here */
	if(in_recursion) goto EXIT_LABEL;

	in_recursion = 1;

	buf = kmalloc(PERSIST_PAGE_HEAD_SIZE, GFP_KERNEL);
	if(!buf) {
		printk("lin_tape: ENOMEM while logging reserve holder\n");
		in_recursion = 0;
		goto EXIT_LABEL;
	} /* if */

	memset(buf, '\0', PERSIST_PAGE_HEAD_SIZE);

	/* get the length of available data */
	rc = persistent_reserve_in((char*)buf, drv, QUERY_FULL_STATUS,
		PERSIST_PAGE_HEAD_SIZE);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: pr-in error while logging reserve holder\n");
		in_recursion = 0;
		goto EXIT_LABEL;
	} /* if */

	length = (buf->additional_length[0] << 24)
		+ (buf->additional_length[1] << 16)
		+ (buf->additional_length[2] << 8)
		+ buf->additional_length[3];

	length += PERSIST_PAGE_HEAD_SIZE;
	kfree(buf);

	buf = kmalloc(length, GFP_KERNEL);
	if(!buf) {
		printk("lin_tape: ENOMEM while logging reserve holder\n");
		in_recursion = 0;
		goto EXIT_LABEL;
	} /* if */

	memset(buf, '\0', length);

	rc = persistent_reserve_in((char*)buf, drv, QUERY_FULL_STATUS, length);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: pr-in error while logging reserve holder\n");
		in_recursion = 0;
		goto EXIT_LABEL;
	} /* if */

	i = PERSIST_PAGE_HEAD_SIZE;
	ds = (struct pr_full_status_descriptor*)&((char*)buf)[i];

	while(i < length) {
		log_rsv_process_pr_descriptor(drv, ds);

		ds_len = (ds->additional_length[0] << 24)
			+ (ds->additional_length[1] << 16)
			+ (ds->additional_length[2] << 8)
			+ ds->additional_length[3];

		i += (ulong)(&((struct pr_full_status_descriptor*)
			NULL)->additional_data) + ds_len;
		ds = (struct pr_full_status_descriptor*)&((char*)buf)[i];
	} /* while */

	in_recursion = 0;

EXIT_LABEL:

	if(buf) kfree(buf);
	DbgPrint(("lin_tape: log_reservation_conflict exit: void\n"));
	return;
} /* log_reservation_conflict() */

/*******************************************************************************
*                                                                              *
*  NAME: log_rsv_process_pr_descriptor                                         *
*                                                                              *
*  PURPOSE: Process individual transcript ID descriptor from PR-in             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    ds: the descriptor to process                                             *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_rsv_process_pr_descriptor(struct tape_descriptor* drv,
	struct pr_full_status_descriptor* ds)
{
	unchar proto = 0;
	struct pr_fc_transport_id* fc = NULL;
	struct pr_scsi_transport_id* scsi = NULL;
	struct pr_sas_transport_id* sas = NULL;
	unchar wwpn[R_HOLDER_LEN] = {0};

	memset(wwpn, '\0', R_HOLDER_LEN);

	proto = ((struct pr_transport_id*)ds->additional_data)->proto;
	fc = (struct pr_fc_transport_id*)ds->additional_data;
	scsi = (struct pr_scsi_transport_id*)ds->additional_data;
	sas = (struct pr_sas_transport_id*)ds->additional_data;

	if(ds->r_holder) {
		switch(proto) {
			case PROTO_ID_FCP:
				log_rsv_fcp(drv, fc, wwpn);
				break;
			case PROTO_ID_SCSI:
				log_rsv_scsi(drv, scsi, wwpn);
				break;
			case PROTO_ID_SAS:
				log_rsv_sas(drv, sas, wwpn);
				break;
			default:
				break;
		} /* switch */

		if(wwpn[0]) {
			memcpy(drv->reserve_holder, wwpn, R_HOLDER_LEN);

			/* Perform the actual logging */
			printk("lin_tape: reserve held by %s\n", wwpn);

		} /* if */
	} /* if */

} /* log_rsv_process_pr_descriptor() */

/*******************************************************************************
*                                                                              *
*  NAME: log_rsv_fcp                                                           *
*                                                                              *
*  PURPOSE: Handle FCP-type reservation logging                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    fc: the fibre-channel transport ID                                        *
*    wwpn: port identifier                                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_rsv_fcp(struct tape_descriptor* drv, struct pr_fc_transport_id* fc,
	unchar* wwpn)
{
	int i;

	for(i = 0; i < 8; i++)
		sprintf(&wwpn[strlen(wwpn)], "%02X",
			fc->n_port_name[i] & 0xFF);

	if(!memcmp(wwpn, drv->reserve_holder, R_HOLDER_LEN))
		memset(wwpn, '\0', R_HOLDER_LEN);

} /* log_rsv_fcp() */

/*******************************************************************************
*                                                                              *
*  NAME: log_rsv_scsi                                                          *
*                                                                              *
*  PURPOSE: Handle SCSI-type reservation logging                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    scsi: the SCSI address                                                    *
*    wwpn: port identifier                                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_rsv_scsi(struct tape_descriptor* drv,
	struct pr_scsi_transport_id* scsi, unchar* wwpn)
{
	int i;

	for(i = 0; i < 2; i++)
		sprintf(&wwpn[strlen(wwpn)], "%02X",
			scsi->scsi_address[i] & 0xFF);

	if(!memcmp(wwpn, drv->reserve_holder, R_HOLDER_LEN))
		memset(wwpn, '\0', R_HOLDER_LEN);

} /* log_rsv_scsi() */

/*******************************************************************************
*                                                                              *
*  NAME: log_rsv_sas                                                           *
*                                                                              *
*  PURPOSE: Handle SAS-type reservation logging                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the recipient of the returned command                                *
*    sas: the SAS address                                                      *
*    wwpn: port identifier                                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_rsv_sas(struct tape_descriptor* drv, struct pr_sas_transport_id* sas,
	unchar* wwpn)
{
	int i;

	for(i = 0; i < 8; i++)
		sprintf(&wwpn[strlen(wwpn)], "%02X",
			sas->sas_address[i] & 0xFF);

	if(!memcmp(wwpn, drv->reserve_holder, R_HOLDER_LEN))
		memset(wwpn, '\0', R_HOLDER_LEN);

} /* log_rsv_sas() */

/****************************************************************
 *                                                              *
 *        Error Recovery Functions                              *
 *                                                              *
 ****************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: tape_send_erp_cmd                                                     *
*                                                                              *
*  PURPOSE: Handle error recovery                                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    command: type of SCSI command for which recovery is to be attempted       *
*    last_timeout: last tape op timeout or update flag for inquiry             *
*    count: page code for inquiry/log sense, svc act. for PR-out, locate count *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_send_erp_cmd(struct tape_descriptor *drv, int command,
	int last_timeout, unsigned int count)
{
	unchar* sense_data = NULL;
	int rc = -EIO;

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

	switch(command) {
	case INQUIRY:
		rc = erp_inquiry(drv, last_timeout, count, sense_data);
		break;

	case TEST_UNIT_READY: case RESERVE:
		rc = erp_tur_or_reserve(drv, command, sense_data);
		break;

	case LOCATE:
		rc = erp_locate(drv, count, sense_data);
		break;

	case READ_POSITION:
		rc = erp_read_position(drv, sense_data);
		break;

	case READ_BUFFER:
		rc = erp_read_buffer(drv, sense_data);
		break;

	case LOG_SENSE:
		rc = erp_log_sense(drv, count);
		break;

	case MODE_SELECT:
		rc = erp_mode_select(drv, sense_data);
		break;

	case CHANGE_DEFINITION:
		rc = erp_change_definition(drv, sense_data);
		break;

	case PERSISTENT_RESERVE_IN:
		rc = erp_persistent_reserve_in(drv, count, sense_data);
		break;

	case PERSISTENT_RESERVE_OUT:
		rc = erp_persistent_reserve_out(drv, last_timeout, count,
			sense_data);
		break;

	case ERP_LBP:
		rc = erp_lbp(drv, sense_data);
		break;

	case PREVENT_MEDIUM_REMOVAL:
		rc = erp_prevent_medium_removal(drv, sense_data);
		break;

	case REQUEST_SENSE:
		rc = erp_request_sense(drv, sense_data);
		break;

	case SCSIOP_WRITE_DYNAMIC_RUNTIME:
		rc = erp_write_dynamic_runtime(drv, sense_data);
		break;

	default:
		rc = -EIO;
	} /* switch */

EXIT_LABEL:

	if(sense_data) kfree(sense_data);
	return rc;
} /* tape_send_erp_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_inquiry                                                           *
*                                                                              *
*  PURPOSE: Issue INQUIRY as part of ERP processing                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    last_timeout: seconds between command issue and timeout                   *
*    count: block address                                                      *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_inquiry(struct tape_descriptor* drv, int last_timeout,
	unsigned int count, char* sense_data)
{
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0, page_len = 0, offset = 0;
	unchar* page = NULL;
	unchar ttype = lin_tape_trace_type;
	boolean log_data = False;

	page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', MAX_INQ_LEN);

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		kfree(page);
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_inquiry.op_code = INQUIRY;
	tmp_cdb->cdb_inquiry.evpd = True;
	tmp_cdb->cdb_inquiry.page_code = count;
	tmp_cdb->cdb_inquiry.alloc_length = MAX_INQ_LEN;

	tcdb = (unchar*)tmp_cdb;
	rc = tape_exe_scsi_cmd(drv, tcdb, 6, (void*)page, MAX_INQ_LEN,
		compute_timeout(drv->drv_id, InquiryTimeOut), DMA_FROM_DEVICE,
		&result, sense_data, &resid);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	log_data = (drv->trace && ttype == LIN_TAPE_FULL_TRACE);

	page_len = page[3];
	if(last_timeout != 1) {
		if(memcmp(drv->serial_num, &page[4], drv->serial_num_len))
			printk("lin_tape: serial number mismatch\n");
		goto EXIT_LABEL;
	} /* if */

	drv->serial_num_len = min((int)page[3], MaxSerialNumberSize);
	memcpy(drv->serial_num, &page[4], drv->serial_num_len);

	/* set devid: Use the last 5 digit of the serial number */
	offset = page_len + 3;
	while(page[offset] == ' ' && offset >= 9) offset--;

	memset(drv->devid, '\0', MAX_VOLUME_LABEL_SIZE + 1);
	sprintf(&drv->devid[0], "IBMtape%d------", drv->dev_index);

	if(offset > 9) memcpy(&drv->devid[13], &page[offset - 4], 5);
	else memcpy(&drv->devid[13], "NoSer", strlen("NoSer"));

EXIT_LABEL:

	if(log_data) {
		tape_log_buffer_data(drv, min((int)((page[3] & 0xFF) + 4),
			(int)(FullTraceUnitSize - 7)), page);
	} /* if */
	if(page) kfree(page);
	if(tmp_cdb) kfree(tmp_cdb);

	return rc;
} /* erp_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_tur_or_reserve                                                    *
*                                                                              *
*  PURPOSE: Issue TUR or RESERVE as part of ERP processing                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    command: TUR or RESERVE                                                   *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_tur_or_reserve(struct tape_descriptor* drv, unchar command,
	char* sense_data)
{
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	if(command == TEST_UNIT_READY)
		tmp_cdb->cdb_tur.op_code = TEST_UNIT_READY;
	else tmp_cdb->cdb_reserve.op_code = RESERVE;

	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 6, NULL, 0,
		compute_timeout(drv->drv_id, ReserveTimeOut), DMA_NONE,
		&result, sense_data, &resid);
EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_tur_or_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_locate                                                            *
*                                                                              *
*  PURPOSE: Issue locate as part of ERP processing                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    count: block address                                                      *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_locate(struct tape_descriptor* drv, unsigned int count,
	char* sense_data)
{
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_locate.op_code = LOCATE;
	tmp_cdb->cdb_locate.blk_addr[0] = (char)(count >> 24 & 0xFF);
	tmp_cdb->cdb_locate.blk_addr[1] = (char)(count >> 16 & 0xFF);
	tmp_cdb->cdb_locate.blk_addr[2] = (char)(count >> 8 & 0xFF);
	tmp_cdb->cdb_locate.blk_addr[3] = (char)(count & 0xFF);

	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 10, NULL, 0,
		compute_timeout(drv->drv_id, LocateTimeOut),
		DMA_NONE, &result, sense_data, &resid);

EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_locate() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_read_position                                                     *
*                                                                              *
*  PURPOSE: Issue read position as part of ERP processing                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_read_position(struct tape_descriptor* drv, char* sense_data)
{
	struct read_pos *buf = NULL;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	buf = kmalloc(sizeof(struct read_pos), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* reset the position flag first */
	drv->pos_valid = False;

	memset(buf, '\0', sizeof(struct read_pos));

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_read_pos.op_code = READ_POSITION;

	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 10, buf, sizeof(struct read_pos),
		compute_timeout(drv->drv_id, ReadPositionTimeOut),
		DMA_FROM_DEVICE, &result, sense_data, &resid);

	if(rc == STATUS_SUCCESS) {
		if(buf->bpu) {
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */
		drv->pos_valid = True;
		drv->eom = (unchar)buf->bop;;
		drv->curr_pos = (ulong)(buf->first_blk_pos[0] << 24)
			+ (ulong)(buf->first_blk_pos[1] << 16)
			+ (ulong)(buf->first_blk_pos[2] << 8)
			+ (ulong)(buf->first_blk_pos[3]);

		DbgPrint(("lin_tape: current position %llu\n",
				drv->curr_pos));
	} /* if */

EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);
	if(buf) kfree(buf);
	return rc;
} /* erp_read_position() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_read_buffer                                                       *
*                                                                              *
*  PURPOSE: Issue read buffer as part of ERP processing                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_read_buffer(struct tape_descriptor* drv, char* sense_data)
{
	struct read_buffer_header *dmp_header = NULL;
	union cdb* tmp_cdb = NULL;
	ulong alloc_size = 0, alloc_order = 0;
	uint dmp_size = 0;
	unchar* tcdb = NULL;
	int xfer_size = 256 * ONE_K, xfer_nums = 0, excess_xfer = 0,
		buf_offset = 0, rc = -EIO, result = 0, resid = 0, id = 0;

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	dmp_header = kmalloc(sizeof(struct read_buffer_header), GFP_KERNEL);
	if(!dmp_header) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(dmp_header, '\0', sizeof(struct read_buffer_header));
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	id = drv->drv_id;
	tmp_cdb->cdb_read_buffer.op_code = READ_BUFFER;
	tmp_cdb->cdb_read_buffer.mode = 3; /* just asking for the descriptor */
	tmp_cdb->cdb_read_buffer.buf_id =
		(id == IBM_3590 || id == IBM_3592) ?  0 : 1; /* dump (r-only) */
	tmp_cdb->cdb_read_buffer.alloc_length[0] = 0;
	tmp_cdb->cdb_read_buffer.alloc_length[1] = 0;
	tmp_cdb->cdb_read_buffer.alloc_length[2] =
		sizeof(struct read_buffer_header);

	tcdb = (unchar*)tmp_cdb;
	rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)dmp_header,
		sizeof(struct read_buffer_header),
		compute_timeout(id, ReadBufferTimeOut), DMA_FROM_DEVICE,
		&result, sense_data, &resid);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	dmp_size = (dmp_header->buf_length[0] << 16)
		+ (dmp_header->buf_length[1] << 8)
		+ dmp_header->buf_length[2];

	if(!dmp_size) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* if there exist a dump field already
	* then get ready to overwrite it */
	alloc_order = get_mem_ord(dmp_size, &alloc_size);

	WRITE_LOCK(&drv->dmp_lock);
	if(drv->dmp) {
		if(drv->dmp_size_order >= 0)
			free_pages((ulong)drv->dmp, drv->dmp_size_order);
		drv->dmp_len = 0;
		drv->dmp_size_order = 0;
		drv->dmp = NULL;
	} /* if */

	if(alloc_order <= (MAX_PAGE_ORDER-1)){
		drv->dmp = (unchar*)__get_free_pages(GFP_KERNEL, alloc_order);
		drv->dmp_size_order = alloc_order;
	}else{
		drv->op_sg = DUMP;
		WRITE_UNLOCK(&drv->dmp_lock);
		rc = setup_sg_for_ops(drv, dmp_size);
		if(rc == STATUS_SUCCESS){
			rc = lin_tape_perform_dump_sg(&drv, dmp_size, tmp_cdb);
			if(rc != STATUS_SUCCESS){
				if(drv->op_sglist){
				if(tape_del_ops_sglist(drv))
						DbgPrint(("lin_tape: ops_sglist for IBMtape%d not "\
								"deleted\n", drv->dev_index));
				}
				drv->op_sg = 0; // tape_del_ops_sglist failed and didn't off
				printk("lin_tape: Failure to capture dump using sg\n");
				goto EXIT_LABEL;
			}
			else goto EXIT_LABEL;
		}else{
			if( tape_del_ops_sglist(drv))
				DbgPrint(("lin_tape: Unable free dmp ops_sglist for IBMtape%d",
						drv->dev_index));
			printk("lin_tape: Unable to collect dump for IBMtape%d\n",
					drv->dev_index);
			drv->op_sg = 0; // if tape_del_ops_sglist fails could not set it
			goto EXIT_LABEL;
		}
	}

	if(!drv->dmp) {
		drv->dmp_size_order = 0;
		drv->dmp_len = 0;
		WRITE_UNLOCK(&drv->dmp_lock);
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->dmp, '\0', alloc_size);
	drv->dmp_len = 0;
	WRITE_UNLOCK(&drv->dmp_lock);

	xfer_nums = dmp_size / xfer_size;
	excess_xfer = dmp_size % xfer_size;
	if(excess_xfer) xfer_nums++;

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_read_buffer.op_code = READ_BUFFER;
	tmp_cdb->cdb_read_buffer.mode =
		(id == IBM_3590 || id == IBM_3592) ?  1 : 2; /* data */
	tmp_cdb->cdb_read_buffer.buf_id =
		(id == IBM_3590 || id == IBM_3592) ?  0 : 1; /* dump (r-only) */

	buf_offset = 0;
	while(xfer_nums) {
		tmp_cdb->cdb_read_buffer.buf_offset[0] =
			(unchar)(buf_offset >> 16);
		tmp_cdb->cdb_read_buffer.buf_offset[1] =
			(unchar)(buf_offset >> 8);
		tmp_cdb->cdb_read_buffer.buf_offset[2] =
			(unchar)(buf_offset % 0xFF);

		if(excess_xfer && xfer_nums == 1) {

			tmp_cdb->cdb_read_buffer.alloc_length[0] =
				(unchar)(excess_xfer >> 16);
			tmp_cdb->cdb_read_buffer.alloc_length[1] =
				(unchar)(excess_xfer >> 8);
			tmp_cdb->cdb_read_buffer.alloc_length[2] =
				(unchar)(excess_xfer & 0xFF);

			tcdb = (unchar*)tmp_cdb;

			rc = tape_exe_scsi_cmd(drv, tcdb, 10,
				(void*)drv->dmp+buf_offset, excess_xfer,
				compute_timeout(id, ReadBufferTimeOut),
				DMA_FROM_DEVICE, &result, sense_data, &resid);
		} else {

			tmp_cdb->cdb_read_buffer.alloc_length[0] =
				(unchar)(xfer_size >> 16);
			tmp_cdb->cdb_read_buffer.alloc_length[1] =
				(unchar)(xfer_size >> 8);
			tmp_cdb->cdb_read_buffer.alloc_length[2] =
				(unchar)(xfer_size & 0xFF);
				tcdb = (unchar *) tmp_cdb;

			rc = tape_exe_scsi_cmd(drv, tcdb, CDB_GENERIC_10,
				(void*)drv->dmp + buf_offset, xfer_size,
				compute_timeout(id, ReadBufferTimeOut),
				DMA_FROM_DEVICE, &result, sense_data, &resid);
		} /* else */

		buf_offset += xfer_size;
		xfer_nums -= 1;

		if(rc != STATUS_SUCCESS) break;
	} /* while */

	if(rc == STATUS_SUCCESS) {
		WRITE_LOCK(&(drv->dmp_lock));
		drv->dmp_len = dmp_size;
		WRITE_UNLOCK(&(drv->dmp_lock));
	} else {
		WRITE_LOCK(&(drv->dmp_lock));
		if(drv->dmp != NULL && drv->dmp_size_order >= 0)
			free_pages((ulong)drv->dmp, drv->dmp_size_order);
		drv->dmp = NULL;
		drv->dmp_len = 0;
		drv->dmp_size_order = 0;
		WRITE_UNLOCK(&(drv->dmp_lock));
	} /* else */

EXIT_LABEL:

	if(dmp_header) kfree(dmp_header);
	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_read_buffer() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_log_sense                                                         *
*                                                                              *
*  PURPOSE: Get simmim page or full log sense as part of ERP processing        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    count: indicates whether simmim page or full log sense is requested       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_log_sense(struct tape_descriptor* drv, unsigned int count)
{
	unchar* pages = NULL;
	unchar* page_buf = NULL;
	uint page_num = 0, logsense_len = 0, total_len = 0;
	unchar* logsense_buf = NULL;
	int i = 0, rc = -EIO;
	boolean free_logsense = False;

	if(count == SimMimLogPage) {
		rc = tape_logsense_erp(drv, SimMimLogPage, drv->simmim,
			&logsense_len);
		if(rc == STATUS_SUCCESS) {
			WRITE_LOCK(&drv->simmim_lock);
			drv->simmim_len = logsense_len;
			WRITE_UNLOCK(&drv->simmim_lock);
		} /* if */
		goto EXIT_LABEL;
	} /* if */
	if(count != 0xFF) goto EXIT_LABEL;

	WRITE_LOCK(&drv->logsense_lock);
	if(!drv->logsense) drv->logsense = kmalloc(MaxLogSense, GFP_KERNEL);
	drv->logsense_len = 0;
	WRITE_UNLOCK(&drv->logsense_lock);

	if(!drv->logsense) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	free_logsense = True;

	READ_LOCK(&drv->logsense_lock);
	logsense_buf = drv->logsense;
	READ_UNLOCK(&drv->logsense_lock);

	page_buf = kmalloc(LogPageHeaderSize + LOGSENSEPAGE, GFP_KERNEL);
	if(!page_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	pages = kmalloc(255, GFP_KERNEL);
	if(!pages) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(page_buf, '\0', LogPageHeaderSize + LOGSENSEPAGE);
	memset(pages, '\0', 255);

	/* issue log sense page 0x00 to get the page_num and each page code */
	rc = tape_logsense_erp(drv, 0x00, pages, &page_num);
	if(rc != STATUS_SUCCESS) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	free_logsense = False;

	/* issue log sense page for each supported page code */
	for(i = 4; i < page_num; i++) {
		rc = tape_logsense_erp(drv, pages[i], page_buf, &logsense_len);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(logsense_len + total_len <= MaxLogSense) {
			memcpy(logsense_buf, page_buf, logsense_len);
			logsense_buf += logsense_len;
			total_len += logsense_len;
		} else {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* else */
	} /* for */

	if(rc == STATUS_SUCCESS) {
		WRITE_LOCK(&drv->logsense_lock);
		drv->logsense_len = total_len;
		if(!drv->logsense_len) free_logsense = True;
		WRITE_UNLOCK(&drv->logsense_lock);
	} /* if */

EXIT_LABEL:

	if(free_logsense) {
		WRITE_LOCK(&drv->logsense_lock);
		kfree(drv->logsense);
		drv->logsense = NULL;
		WRITE_UNLOCK(&drv->logsense_lock);
	} /* if */
	if(pages) kfree(pages);
	if(page_buf) kfree(page_buf);
	return rc;
} /* erp_log_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_mode_select                                                       *
*                                                                              *
*  PURPOSE: Handle MODE_SELECT ERP request                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_mode_select(struct tape_descriptor* drv, char* sense_data)
{
	union cdb* tmp_cdb = NULL;
	int rc = -EIO;

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	rc = erp_set_block_size(drv, tmp_cdb, sense_data);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = erp_set_compression(drv, tmp_cdb, sense_data);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(drv->drv_id == IBM_3590 || drv->drv_id == IBM_3592) {
		rc = erp_logical_write_protect(drv, tmp_cdb, sense_data);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if(drv->drv_id == IBM_3592) {
		erp_tod(drv, tmp_cdb, sense_data);
	} /* if */

	if(drv->drv_id == IBM_3592 && drv_gen(drv) > 1) {
		rc = erp_density(drv, tmp_cdb, sense_data);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	tape_encryption_write_setting(&drv);

	if(drv->lbp_set && ((drv->drv_id == IBM_3592 && drv_gen(drv) >= 2)
		|| (drv->drv_id == IBM_3580 && drv_gen(drv) >= 5))) {
		rc = erp_lbp(drv, sense_data);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	if (((drv->drv_id == IBM_3592 && drv_gen(drv) >4) ||
        (drv->drv_id == IBM_3580 && drv_gen(drv) >6)) && drv->amu_enable != 0xFF){
		rc = erp_amu(&drv, sense_data);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	}

EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);

	return rc;
} /* erp_mode_select() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_issue_mode_sense                                                  *
*                                                                              *
*  PURPOSE: Construct and send mode sense (6) for error recovery               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tcdb: the CDB to use to avoid overwriting the drive's CDB                 *
*    mode_buf: buffer to receive return data                                   *
*    sense_data: buffer to receive sense data                                  *
*    page: mode sense page number                                              *
*    size: size allocated for the mode sense                                   *
*    dbd: if True, disable block descriptors; if False, allow                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_issue_mode_sense(struct tape_descriptor* drv, union cdb* tcdb,
	char* mode_buf, char* sense_data, int page, int size, unchar dbd)
{
	int rc = -EIO, result = 0, resid = 0;

	memset(tcdb, '\0', MAX_CDB_SIZE);
	tcdb->cdb_modesense.op_code = MODE_SENSE;
	tcdb->cdb_modesense.dbd = dbd;
	tcdb->cdb_modesense.page_code = page;
	tcdb->cdb_modesense.alloc_length = size;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_6,
		(void*)mode_buf, size,
		compute_timeout(drv->drv_id, ModeSenseTimeOut),
		DMA_FROM_DEVICE, &result, sense_data, &resid);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	return rc;
} /* erp_issue_mode_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_issue_mode_sense10                                                *
*                                                                              *
*  PURPOSE: Construct and send mode sense (10) for error recovery              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int erp_issue_mode_sense10(struct tape_descriptor** drvp, unchar page_code,
                           unchar* modesense_buf, int alloc_len, int subpg,
                           boolean dbd, boolean return_header, char* sense_data)
{
	int rc = -EIO, result = 0, resid = 0, drv_alloc_len;
	union cdb* tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	unchar *mode_buf;
struct tape_descriptor* drv=*drvp;
	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv_alloc_len = return_header ? alloc_len
		: alloc_len + sizeof(struct mode_parm_header_10);
	mode_buf = kmalloc(drv_alloc_len, GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', drv_alloc_len);

	memset(tcdb, '\0', MAX_CDB_SIZE);
	tcdb->cdb_modesense_10.op_code = MODE_SENSE_10;
	tcdb->cdb_modesense_10.dbd = dbd;
	tcdb->cdb_modesense_10.page_code = page_code;
	tcdb->cdb_modesense_10.sub_page_code = subpg;
	tcdb->cdb_modesense_10.alloc_length[0] =
		(unchar)((drv_alloc_len >> 8) & 0xFF);
	tcdb->cdb_modesense_10.alloc_length[1] =
		(unchar)(drv_alloc_len & 0xFF);

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_10,
		(void*)mode_buf, drv_alloc_len,
		compute_timeout(drv->drv_id, ModeSenseTimeOut),
		DMA_FROM_DEVICE, &result, sense_data, &resid);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:
	if(tcdb) kfree(tcdb);

	return rc;
} /* erp_issue_mode_sense10() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_issue_mode_select                                                 *
*                                                                              *
*  PURPOSE: Construct and send mode select (6) for error recovery              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tcdb: the CDB to use to avoid overwriting the drive's CDB                 *
*    mode_buf: buffer to receive return data                                   *
*    sense_data: buffer to receive sense data                                  *
*    size: size allocated for the mode sense                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_issue_mode_select(struct tape_descriptor* drv, union cdb* tcdb,
	char* mode_buf, char* sense_data, int size)
{
	int rc = -EIO, result = 0, resid = 0;

	((struct mode_parm_header*)mode_buf)->mode_data_length = 0;

	memset(tcdb, '\0', MAX_CDB_SIZE);
	tcdb->cdb_modeselect.op_code = MODE_SELECT;
	tcdb->cdb_modeselect.pf = True;
	tcdb->cdb_modeselect.parm_list_length = size;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_6,
		(void*)mode_buf, size,
		compute_timeout(drv->drv_id, ModeSelectTimeOut),
		DMA_TO_DEVICE, &result, sense_data, &resid);

	return rc;
} /* erp_issue_mode_select() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_issue_mode_select10                                               *
*                                                                              *
*  PURPOSE: Construct and send mode select (10) for error recovery             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int erp_issue_mode_select10(struct tape_descriptor* drv, unchar* mode_buf,
                            boolean sp, uint alloc_len, char* sense_data)
{
	int rc = -EIO, result = 0, resid = 0;
	union cdb* tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);

	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(alloc_len > MAX_MODE_PAGE_10_LENGTH) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	mode_buf[0] = 0;
	mode_buf[1] = 0;

	memset(tcdb, '\0', MAX_CDB_SIZE);
	tcdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	tcdb->cdb_modeselect_10.pf = True;
	tcdb->cdb_modeselect_10.sp = sp ? 1 : 0;
	/* include header in alloc_len */
	tcdb->cdb_modeselect_10.parm_list_length[0] = (alloc_len >> 8) & 0xFF;
	tcdb->cdb_modeselect_10.parm_list_length[1] = alloc_len & 0xFF;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_10,
		(void*)mode_buf, alloc_len,
		compute_timeout(drv->drv_id, ModeSelectTimeOut),
		DMA_TO_DEVICE, &result, sense_data, &resid);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:
	if(tcdb) kfree(tcdb);
	return rc;
} /* erp_issue_mode_select10() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_set_block_size                                                    *
*                                                                              *
*  PURPOSE: Set block size as part of MODE_SELECT ERP processing               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tmp_cdb: caller-allocated CDB to fill out and use                         *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_set_block_size(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data)
{
	struct media_parm* mode_buf = NULL;
	ulong tmp_blk_size = 0;
	int rc = -EIO;

	if(drv->blk_size && (drv->blk_size > drv->max_blk_size ||
		drv->blk_size < drv->min_blk_size ||
		drv->blk_size > drv->max_xfer_len)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	mode_buf = kmalloc(sizeof(struct media_parm), GFP_KERNEL);
	if(!mode_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mode_buf, '\0', sizeof(struct media_parm));

	rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)mode_buf, sense_data,
		DeviceConfigurationPageNumber, sizeof(struct media_parm),
		False);
	if(rc) goto EXIT_LABEL;

	/* We need check if sili is ON.  If it is ON and drv->blk_size is 0,
	* we want to set BLOCK LENGTH field in mode sense header to be
	* non-zero (4) in order to force a check condition on overlength read
	*/
	tmp_blk_size = (drv->sili && !drv->blk_size) ? 4 : drv->blk_size;

	if(mode_buf->block.blk_length[0] != (unchar)(tmp_blk_size >> 16 & 0xFF)
		|| mode_buf->block.blk_length[1] != (unchar)
			(tmp_blk_size >> 8 & 0xFF)
		|| mode_buf->block.blk_length[2] != (unchar)
			(tmp_blk_size & 0xFF)) {
		mode_buf->block.blk_length[0] =
			(unchar)(tmp_blk_size >> 16 & 0xFF);
		mode_buf->block.blk_length[1] = (unchar)
			(tmp_blk_size >> 8 & 0xFF);
		mode_buf->block.blk_length[2] = (unchar)
			(tmp_blk_size & 0xFF);
		mode_buf->block.density_code = drv->density;
		mode_buf->header.buffered_mode =
			drv->buffered_mode ? True : False;
		rc = erp_issue_mode_select(drv, tmp_cdb, (char*)mode_buf,
			sense_data, sizeof(struct media_parm));
	} /* if */

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	if(mode_buf) kfree(mode_buf);
	return rc;

} /* erp_set_block_size() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_set_compression                                                   *
*                                                                              *
*  PURPOSE: Set compression as part of MODE_SELECT ERP processing              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tmp_cdb: caller-allocated CDB to fill out and use                         *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_set_compression(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data)
{
	struct compression_parm* comp_buf = NULL;
	struct dev_conf_parm* dev_conf_buf = NULL;
	int rc = -EIO;

	if(drv->drv_id == IBM_3580) {
		comp_buf = kmalloc(sizeof(struct compression_parm), GFP_KERNEL);
		if(!comp_buf) goto EXIT_LABEL;

		memset(comp_buf, '\0', sizeof(struct compression_parm));
		rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)comp_buf,
			sense_data, DataCompressionPageNumber,
			sizeof(struct compression_parm), True);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if((comp_buf->page[2] >> 7) != drv->compression) {
			if(drv->compression)
				comp_buf->page[2] = comp_buf->page[2] | 0x80;
			else comp_buf->page[2] = comp_buf->page[2] & 0x7F;

			rc = erp_issue_mode_select(drv, tmp_cdb,
				(char*)comp_buf, sense_data,
				sizeof(struct compression_parm));
		} /* if */
	} /* if */

	dev_conf_buf = kmalloc(sizeof(struct dev_conf_parm), GFP_KERNEL);
	if(!dev_conf_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(dev_conf_buf, '\0', sizeof(struct dev_conf_parm));

	rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)dev_conf_buf,
		sense_data, DeviceConfigurationPageNumber,
		sizeof(struct dev_conf_parm), True);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: tape_send_erp_cmd: cannot get mode sense\n");
		goto EXIT_LABEL;
	} /* if */

	if(dev_conf_buf && (dev_conf_buf->page[14] != drv->compression)) {
		dev_conf_buf->header.mode_data_length = False;
		dev_conf_buf->page[14] = drv->compression;

		rc = erp_issue_mode_select(drv, tmp_cdb, (char*)dev_conf_buf,
			sense_data, sizeof(struct dev_conf_parm));
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(comp_buf) kfree(comp_buf);
	if(dev_conf_buf) kfree(dev_conf_buf);
	return rc;
} /* erp_set_compression() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_logical_write_protect                                             *
*                                                                              *
*  PURPOSE: Set write protection as part of MODE_SELECT ERP processing         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tmp_cdb: caller-allocated CDB to fill out and use                         *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_logical_write_protect(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data)
{
	struct medium_sense* medium_buf = NULL;
	unchar* tmp_buf = NULL;
	int rc = -EIO, size = 0;

	medium_buf = kmalloc(sizeof(struct medium_sense), GFP_KERNEL);
	if(!medium_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(medium_buf, '\0', sizeof(struct medium_sense));

	rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)medium_buf, sense_data,
		MediumSensePageNumber, sizeof(struct medium_sense), True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if((drv->logic_write_protect != NO_PROTECT ||
		(!medium_buf->page.associated_wp &&
		!medium_buf->page.persistent_wp))
		&&
		(drv->logic_write_protect != ASSOCIATED_PROTECT ||
		medium_buf->page.associated_wp)
		&&
		(drv->logic_write_protect != PERSISTENT_PROTECT
		|| medium_buf->page.persistent_wp)) goto EXIT_LABEL;

	switch(drv->logic_write_protect) {
	case NO_PROTECT:
		medium_buf->page.associated_wp = False;
		medium_buf->page.persistent_wp = False;
		medium_buf->page.reset_associate_wp = True;
		medium_buf->page.reset_persistent_wp = True;
		break;
	case ASSOCIATED_PROTECT:
		medium_buf->page.associated_wp = True;
		break;
	case PERSISTENT_PROTECT:
		medium_buf->page.persistent_wp = True;
		break;
	case WORM_PROTECT:
		medium_buf->page.permanent_wp = True;
		break;
	} /* switch */

	/* 3590 and 3592 have different page length on 0x23 */
	size = medium_buf->page.page_length + 2
		+ sizeof(struct mode_parm_header);

	tmp_buf = kmalloc(size, GFP_KERNEL);
	if(!tmp_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_buf, '\0', size);

	memcpy(tmp_buf, medium_buf, min(size,
		(int)sizeof(struct medium_sense)));
	rc = erp_issue_mode_select(drv, tmp_cdb, (char*)tmp_buf, sense_data,
		size);

EXIT_LABEL:

	if(tmp_buf) kfree(tmp_buf);
	if(medium_buf) kfree(medium_buf);
	return rc;
} /* erp_logical_write_protect() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_tod                                                               *
*                                                                              *
*  PURPOSE: Set time of day as part of MODE_SELECT ERP processing              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tmp_cdb: caller-allocated CDB to fill out and use                         *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_tod(struct tape_descriptor* drv, union cdb* tmp_cdb, char* sense_data)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	struct timeval* tv = NULL;
#else
    struct timespec64 timespec;
#endif // LINUX_VERSION_CODE

	struct tod_control* tod_buf = NULL;
	int now = 0, rc = -EIO;

	tod_buf = kmalloc(sizeof(struct tod_control), GFP_KERNEL);
	if(!tod_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	tv = kmalloc(sizeof(struct timeval), GFP_KERNEL);
	if(!tv) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
#endif // LINUX_VERSION_CODE
	rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)tod_buf, sense_data,
		TODControlPageNumber, sizeof(struct tod_control), True);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	/* get time info to set TOD */
	memset(tv, '\0', sizeof(struct timeval));

	do_gettimeofday(tv);
	/* tv.sec in structure timeval is a "long" integer
	* it has 4 bytes in 32-bit kernel, but 8 bytes in 64-bit kernel
	* we truncate it to 4 byte integer
	*/
	now = (int) tv->tv_sec;
#else
	/* get time info to set TOD */
	memset(&timespec, '\0', sizeof(struct timespec64));

    ktime_get_real_ts64(&timespec);
    now = (int) timespec.tv_sec;
#endif // LINUX_VERSION_CODE

#if defined(__LITTLE_ENDIAN)
	tod_buf->page[2] = ((char*)&now)[3];
	tod_buf->page[3] = ((char*)&now)[2];
	tod_buf->page[4] = ((char*)&now)[1];
	tod_buf->page[5] = ((char*)&now)[0];
#elif defined(__BIG_ENDIAN)
	tod_buf->page[2] = ((char*)&now)[0];
	tod_buf->page[3] = ((char*)&now)[1];
	tod_buf->page[4] = ((char*)&now)[2];
	tod_buf->page[5] = ((char*)&now)[3];
#endif

	tod_buf->page[10] |= 0x04;	/* reset valid */

	rc = erp_issue_mode_select(drv, tmp_cdb, (char*)tod_buf, sense_data,
		sizeof(struct tod_control));

EXIT_LABEL:
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	if(tv) kfree(tv);
#endif
	if(tod_buf) kfree(tod_buf);
	return rc;
} /* erp_tod() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_density                                                           *
*                                                                              *
*  PURPOSE: Set density as part of MODE_SELECT ERP processing                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    tmp_cdb: caller-allocated CDB to fill out and use                         *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_density(struct tape_descriptor* drv, union cdb* tmp_cdb,
	char* sense_data)
{
	struct rw_control* rw_buffer = NULL;
	int rc = -EIO;

	rw_buffer = kmalloc(sizeof(struct rw_control), GFP_KERNEL);
	if(!rw_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(rw_buffer, '\0', sizeof(struct rw_control));

	rc = erp_issue_mode_sense(drv, tmp_cdb, (char*)rw_buffer, sense_data,
		RWControlPageNumber, sizeof(struct rw_control), False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* modify changes if requested */
	if(!drv->set_density_valid ||
		(rw_buffer->page.default_density == drv->set_default_density &&
		rw_buffer->page.pending_density == drv->set_pending_density))
		goto EXIT_LABEL;

	rw_buffer->page.default_density = drv->set_default_density;
	rw_buffer->page.pending_density = drv->set_pending_density;
	rw_buffer->block.density_code = 0x7F;

	rc = erp_issue_mode_select(drv, tmp_cdb, (char*)rw_buffer, sense_data,
		sizeof(struct rw_control));

EXIT_LABEL:

	if(rw_buffer) kfree(rw_buffer);
	return rc;
} /* erp_density() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_change_definition                                                 *
*                                                                              *
*  PURPOSE: Post check-1 error recovery                                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_change_definition(struct tape_descriptor* drv, char* sense_data)
{
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	if(drv->drv_id != IBM_3590 && drv->drv_id != IBM_3592) goto EXIT_LABEL;

	if(drv->cmd_in_open || drv->cmd_wo_ua_erp) goto EXIT_LABEL;
	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->pos_valid = False;

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_chg_definition.op_code = CHANGE_DEFINITION;
	tmp_cdb->cdb_chg_definition.def_parm = 0x60;

	tcdb = (unchar*)tmp_cdb;
	rc = tape_exe_scsi_cmd(drv, tcdb, 10, NULL, 0,
		compute_timeout(drv->drv_id, ChangeDefinitionTimeOut),
			DMA_NONE, &result, sense_data, &resid);
EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_change_defition() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_persistent_reserve_in                                             *
*                                                                              *
*  PURPOSE: Send persistent reserve as part of ERP processing                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    count: type of action to perform in PR-out                                *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_persistent_reserve_in(struct tape_descriptor* drv, unsigned int count,
	unchar* sense_data)
{
	union cdb* tmp_cdb = NULL;
	unchar* buffer = NULL, * tcdb = NULL;
	int length = 0, rc = -EIO, i = 0, result = 0, resid = 0;

	DbgPrint(("lin_tape: read_reservation_erp\n"));

	if(count != QUERY_RESERVATIONS) goto EXIT_LABEL;

	tmp_cdb = (union cdb*)kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < 2; i++) {
		buffer = kmalloc(length + PERSIST_PAGE_HEAD_SIZE, GFP_KERNEL);
		if(!buffer) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		memset(buffer, '\0', length + PERSIST_PAGE_HEAD_SIZE);
		memset(tmp_cdb, '\0', MAX_CDB_SIZE);

		tmp_cdb->cdb_persist_reserve_in.op_code = PERSISTENT_RESERVE_IN;
		tmp_cdb->cdb_persist_reserve_in.serv_action =
			QUERY_RESERVATIONS;
		tmp_cdb->cdb_persist_reserve_in.alloc_length[0] = (unchar)
			((length + PERSIST_PAGE_HEAD_SIZE) >> 16) & 0xFF;
		tmp_cdb->cdb_persist_reserve_in.alloc_length[1] = (unchar)
			((length + PERSIST_PAGE_HEAD_SIZE) >> 8) & 0xFF;
		tmp_cdb->cdb_persist_reserve_in.alloc_length[2] = (unchar)
			(length + PERSIST_PAGE_HEAD_SIZE) & 0xFF;

		tcdb = (unchar*)tmp_cdb;
		rc = tape_exe_scsi_cmd(drv, tcdb, 10, buffer,
			length + PERSIST_PAGE_HEAD_SIZE,
			compute_timeout(drv->drv_id, PersistReserveInTimeOut),
			DMA_FROM_DEVICE, &result, sense_data, &resid);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(i) break;

		drv->pr_generation = (buffer[0] << 24) + (buffer[1] << 16) +
			(buffer[2] << 8) + (buffer[3]);

		length = (buffer[4] << 24) + (buffer[5] << 16) +
			(buffer[6] << 8) + buffer[7];

		if(length / sizeof(struct persist_reserve) != 1) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		kfree(buffer);
		buffer = NULL;
	} /* for */

	if(memcmp(drv->reservation_key, buffer + PERSIST_PAGE_HEAD_SIZE,
		RESERVE_KEY_LENGTH)) {
		printk("lin_tape: mismatched reservation key.\n");
		rc = -EINVAL;
	} /* if */

EXIT_LABEL:

	if(buffer) kfree(buffer);
	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_persistent_reserve_in() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_persistent_reserve_out                                            *
*                                                                              *
*  PURPOSE: Send persistent reserve as part of ERP processing                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    last_timeout: seconds between command issue and timeout                   *
*    count: type of action to perform in PR-out                                *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_persistent_reserve_out(struct tape_descriptor* drv, int last_timeout,
	unsigned int count, unchar* sense_data)
{
	struct persist_reserve_out *parm_list_buf = NULL;
	union cdb *tcdb = NULL;
	int rc = -EIO, to = 0, result = 0, resid = 0;

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);
	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));

	tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tcdb, '\0', MAX_CDB_SIZE);

	tcdb->cdb_persist_reserve_out.op_code = PERSISTENT_RESERVE_OUT;
	tcdb->cdb_persist_reserve_out.parm_list_length[0] = 0;
	tcdb->cdb_persist_reserve_out.parm_list_length[1] =
		sizeof(struct persist_reserve_out);

	if(count == REGISTER_IGNORE_KEY) {
		DbgPrint(("lin_tape: register_reservation_key_erp\n"));
		get_random_bytes(parm_list_buf->service_key,
			RESERVE_KEY_LENGTH);
		tcdb->cdb_persist_reserve_out.serv_action =
			REGISTER_IGNORE_KEY;
		to = last_timeout;
	} else if(count == CREATE_PERSIST_RESERVE) {
		DbgPrint(("lin_tape: create_persistent_reservation_erp.\n"));
		memcpy(parm_list_buf->reserve_key, drv->reservation_key,
			RESERVE_KEY_LENGTH);
		tcdb->cdb_persist_reserve_out.serv_action =
			CREATE_PERSIST_RESERVE;
		tcdb->cdb_persist_reserve_out.type = EXCLUSIVE_ACCESS;
	} else goto EXIT_LABEL;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, 10, parm_list_buf,
		sizeof(struct persist_reserve_out),
		compute_timeout(drv->drv_id, PersistReserveOutTimeOut) + to,
		DMA_TO_DEVICE, &result, sense_data, &resid);

	if(count == REGISTER_IGNORE_KEY && rc == STATUS_SUCCESS)
		memcpy(drv->reservation_key, parm_list_buf->service_key,
			RESERVE_KEY_LENGTH);
EXIT_LABEL:

	if(parm_list_buf) kfree(parm_list_buf);
	if(tcdb) kfree(tcdb);
	return rc;
} /* erp_persistent_reserve_out() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_lbp                                                               *
*                                                                              *
*  PURPOSE: Restore logical block protection as part of ERP processing         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: caller-allocated buffer to receive sense data                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_lbp(struct tape_descriptor* drv, unchar* sense_data)
{
	union cdb *tcdb = NULL;
	struct logical_block_protection* lbp =
		(struct logical_block_protection*)&drv->lbp;
	struct control_pg_dp_subp_with_hdr* subpage = NULL;
	int rc = -EIO, size = 0, result = 0, resid = 0;

	size = sizeof(struct control_pg_dp_subp_with_hdr);
	subpage = kmalloc(size, GFP_KERNEL);
	if(!subpage) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(subpage, '\0', size);

	tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tcdb, '\0', MAX_CDB_SIZE);

	tcdb->cdb_modesense_10.op_code = MODE_SENSE_10;
	tcdb->cdb_modesense_10.dbd = True;
	tcdb->cdb_modesense_10.page_code = MODE_SENSE_CONTROL_PG;
	tcdb->cdb_modesense_10.sub_page_code = CONTROL_PG_DP_SUBPG;
	tcdb->cdb_modesense_10.alloc_length[0] =
		(unchar)((size >> 8) & 0xFF);
	tcdb->cdb_modesense_10.alloc_length[1] = (unchar)(size & 0xFF);

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_10,
		(void*)subpage, size,
		compute_timeout(drv->drv_id, ModeSenseTimeOut),
		DMA_FROM_DEVICE, &result, sense_data, &resid);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: %s tape_modesense10_page failed: %d\n",
			drv->devid, rc);
		goto EXIT_LABEL;
	} /* if */

	subpage->page.ps = 0;
	subpage->page.spf = 1;
	subpage->page.lb_prot_method = lbp->lbp_method;

	if(lbp->lbp_method) {
		subpage->page.lb_prot_info_length = lbp->lbp_info_length;
		subpage->page.rbdp = lbp->rbdp;
		subpage->page.lbp_r = lbp->lbp_r;
		subpage->page.lbp_w = lbp->lbp_w;
	} else {
		subpage->page.lb_prot_info_length =
		subpage->page.rbdp =
		subpage->page.lbp_r =
		subpage->page.lbp_w = 0;
	} /* else */

	((unchar*)subpage)[0] = 0;
	((unchar*)subpage)[1] = 0;

	memset(tcdb, '\0', MAX_CDB_SIZE);
	tcdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	tcdb->cdb_modeselect_10.pf = True;
	tcdb->cdb_modeselect_10.parm_list_length[0] = (size >> 8) & 0xFF;
	tcdb->cdb_modeselect_10.parm_list_length[1] = size & 0xFF;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_10,
		(void*)subpage, size,
		compute_timeout(drv->drv_id, ModeSelectTimeOut),
		DMA_TO_DEVICE, &result, sense_data, &resid);

	if(rc) {
		printk("lin_tape: set_blk_prot cannot set mode pg 0x0A "\
			"sub 0xF0\n");
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(subpage) kfree(subpage);
	if(tcdb) kfree(tcdb);
	return rc;
} /* erp_lbp() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_prevent_medium_removal                                            *
*                                                                              *
*  PURPOSE: Restore medium removal prevention on this drive                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: buffer for sense data                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_prevent_medium_removal(struct tape_descriptor* drv, unchar* sense_data)
{
	int rc = STATUS_SUCCESS, result = 0, resid = 0;
	union cdb *tcdb = NULL;

	DbgPrint(("lin_tape: erp_prevent_medium_removal entry\n"));

	tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tcdb, '\0', MAX_CDB_SIZE);

	tcdb->cdb_medium_removal.op_code = ALLOW_MEDIUM_REMOVAL;
	tcdb->cdb_medium_removal.prevent = 1;

	rc = tape_exe_scsi_cmd(drv, (unchar*)tcdb, CDB_GENERIC_6,
		NULL, 0, compute_timeout(drv->drv_id, PreventAllowMediaTimeOut),
		DMA_NONE, &result, sense_data, &resid);

EXIT_LABEL:

	if(rc == STATUS_SUCCESS) drv->prevent_removal = 1;
	DbgPrint(("lin_tape: erp_prevent_medium_removal.  prevent = %d\n",
		drv->prevent_removal));
	if(tcdb) kfree(tcdb);
	return rc;
} /* erp_prevent_medium_removal() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_request_sense                                                     *
*                                                                              *
*  PURPOSE: Request sense for error recovery                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: buffer for sense data                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_request_sense(struct tape_descriptor* drv, unchar* sense_data)
{
	struct tape_sense_data* buffer = NULL;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	int rc = -EIO, result = 0, resid = 0;

	DbgPrint(("lin_tape: erp_request_sense entry\n"));

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	buffer = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
	if(!buffer) {
		printk("lin_tape: erp_request_sense: no sense data pointer\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buffer, '\0', sizeof(struct tape_sense_data));

	tmp_cdb->cdb_req_sense.op_code = REQUEST_SENSE;
	tmp_cdb->cdb_req_sense.alloc_length = sizeof(struct tape_sense_data);

	tcdb = (unchar*)tmp_cdb;
	rc = tape_exe_scsi_cmd(drv, tcdb, CDB_GENERIC_10, buffer,
		sizeof(struct tape_sense_data),
		compute_timeout(drv->drv_id, ModeSenseTimeOut),
		DMA_FROM_DEVICE, &result, sense_data, &resid);

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: tape_request_sense: cannot request sense\n");
		goto EXIT_LABEL;
	} /* if */

	if(buffer->valid) {
		if(!drv->sense_data)
			drv->sense_data =
				kmalloc(sizeof(struct tape_sense_data),
					GFP_KERNEL);

		if(!drv->sense_data) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		memcpy(drv->sense_data, buffer, sizeof(struct tape_sense_data));
	} /* if */

EXIT_LABEL:

	if(buffer) kfree(buffer);
	if(tmp_cdb) kfree(tmp_cdb);

	return rc;
} /* erp_request_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_write_dynamic_runtime                                             *
*                                                                              *
*  PURPOSE: Write dynamic runtime attributes for error recovery                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    sense_data: buffer for sense data                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_write_dynamic_runtime(struct tape_descriptor* drv, unchar* sense_data)
{
	int rc = STATUS_SUCCESS, tot_len = 0, i = 0, result = 0, resid = 0,
		timeout = 0;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL;
	struct dyn_attr_parm_list* parm_list = NULL;

	if(!dynamic_attributes) goto EXIT_LABEL;

	DbgPrint(("lin_tape: erp_write_dynamic_runtime entry\n"));

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	/* length calculated as the size of the containing structure, minus the
		array-of-one placeholder byte, plus the size of the data */
	tot_len = sizeof(struct dyn_attr_parm_list) - sizeof(char);
	for(i = 0; dyn_attrs[i].val_str; i++)
		tot_len += sizeof(struct dyn_attr) - sizeof(char) +
			dyn_attrs[i].len;

	parm_list = kmalloc(tot_len, GFP_KERNEL);
	if(!parm_list) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(parm_list, '\0', tot_len);

	rc = lin_tape_dyn_attrs_parm_list(drv, (char*)parm_list, tot_len,
		False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	tmp_cdb->cdb_write_dynamic_attribute.op_code =
		SCSIOP_WRITE_DYNAMIC_RUNTIME;
	for(i = 0; i < 4; i++)
		tmp_cdb->cdb_write_dynamic_attribute.parameter_list_length[i] =
			(tot_len >> (32 - (i + 1) * BITS_PER_BYTE)) & 0xFF;

	tcdb = (unchar*)tmp_cdb;
	timeout = compute_timeout(drv->drv_id, WriteDynamicRTTimeout);
	rc = tape_exe_scsi_cmd(drv, tcdb, CDB_GENERIC_12, parm_list,
		tot_len, timeout, DMA_TO_DEVICE, &result, sense_data, &resid);
	if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape: cannot Write Dynamic RT attrs\n"));
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(parm_list) kfree(parm_list);
	if(tmp_cdb) kfree(tmp_cdb);
	return rc;
} /* erp_write_dynamic_runtime() */

/*******************************************************************************
*                                                                              *
*  NAME: erp_amu                                                               *
*                                                                              *
*  PURPOSE: Restore Archive Mode Unthread                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int erp_amu(struct tape_descriptor** drvp, unchar *sense)
{
    struct tape_descriptor* drv = *drvp;
	union cdb* tmp_cdb = NULL;
    int rc = -EIO, alloc_len, result = 0, resid, sensefree = False;
	unchar *page = NULL, *tcdb = NULL, *sense_data = NULL, earchive;
    DbgPrint(("lin_tape: amu entry\n"));

    if (!drv){
        rc = -ENOMEM;
        goto EXIT_LABEL;
    }

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		drv->amu_enable = 0xFF;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	alloc_len =  sizeof(struct feature_switches_subpage) +
                 sizeof(struct mode_parm_header_10);
	page = kmalloc(alloc_len, GFP_KERNEL);
	if(!page) {
		drv->amu_enable = 0xFF;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', alloc_len);

	if (sense) sense_data = sense;
	else	{
            sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
            sensefree = True;
	}
	if(!sense_data) {
        drv->amu_enable = 0xFF;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

	tmp_cdb->cdb_modesense_10.op_code = MODE_SENSE_10;
	tmp_cdb->cdb_modesense_10.dbd = True;
	tmp_cdb->cdb_modesense_10.page_code = DEVICE_ATTRIBUTES_PAGE;
	tmp_cdb->cdb_modesense_10.sub_page_code = FEATURE_SWITCHES;
	tmp_cdb->cdb_modesense_10.alloc_length[0] = (unchar)((alloc_len >> 8) & 0xFF);
	tmp_cdb->cdb_modesense_10.alloc_length[1] = (unchar)(alloc_len & 0xFF);

	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)page,
			alloc_len, compute_timeout(drv->drv_id,
			ModeSenseTimeOut), DMA_FROM_DEVICE, &result,
			sense_data, &resid);

	if(rc != STATUS_SUCCESS) {
        if (drv->amu_enable == 0xFF)
          rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	earchive = ((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->earchive;
	if (drv->amu_enable == 0xFF){
        drv->amu_enable = earchive;
        rc = STATUS_SUCCESS;
        goto EXIT_LABEL;
	}
    if (drv->amu_enable == earchive){
        rc = STATUS_SUCCESS;
        goto EXIT_LABEL;
    }
    page[0]=0;
    page[1]=0;
	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->earchive = drv->amu_enable;
	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->spf = 1;
	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->ps = 1;

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	tmp_cdb->cdb_modeselect_10.pf = True;
	tmp_cdb->cdb_modeselect_10.sp = 1;
	tmp_cdb->cdb_modeselect_10.parm_list_length[0] = (alloc_len >> 8) & 0xFF;;
	tmp_cdb->cdb_modeselect_10.parm_list_length[1] = alloc_len & 0xFF;
	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)page,
		alloc_len, compute_timeout(drv->drv_id,
		ModeSelectTimeOut), DMA_TO_DEVICE, &result,
		sense_data, &resid);
	if (rc != STATUS_SUCCESS){
        goto EXIT_LABEL;
	}

EXIT_LABEL:
    if (tmp_cdb) kfree(tmp_cdb);
    if (sensefree) kfree(sense_data);
	if (page) kfree(page);
	DbgPrint(("lin_tape: amu exit: %d amu is %d\n", rc, drv->amu_enable));
	return rc;
} /* erp_amu() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_perform_dump_sg                                              *
*                                                                              *
*  PURPOSE: set up the CDB, execute the dump command and handle errors         *
*      			utilizing scatter/gather                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape_descriptor on which to perform the dump                        *
*    xfer_length: total bytes requested                                        *
*    tmp_cdb: cdb used to call the dump                                        *
*                                                                              *
*  RETURNS: bytes read or error code                                           *
*                                                                              *
*******************************************************************************/
int lin_tape_perform_dump_sg(struct tape_descriptor** drvp, size_t xfer_len,
		union cdb* tmp_cdb)
{
	int rc = 0, i = 0;
	unsigned long buf_offset = 0;
	struct tape_descriptor* drv = *drvp;
	int id = drv->drv_id;
	int resid = 0;
	int temp_use_sg;
	unchar* tcdb;
	unsigned int xfer_size;
	unchar *sense_data;

	temp_use_sg = drv->op_sglist->use_sg;
	drv->op_sglist->use_sg = 1;

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data){
		printk("lin_tape: failed to allocate sense_data\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset((void*)sense_data, '\0', SCSI_SENSE_BUFFERSIZE);
	memset((void*)tmp_cdb, '\0', MAX_CDB_SIZE);

	tmp_cdb->cdb_read_buffer.op_code = READ_BUFFER;
	tmp_cdb->cdb_read_buffer.mode =
		(id == IBM_3590 || id == IBM_3592) ?  1 : 2; /* data */
	tmp_cdb->cdb_read_buffer.buf_id =
		(id == IBM_3590 || id == IBM_3592) ?  0 : 1; /* dump (r-only) */

	for(i = 0; i < temp_use_sg; i++){

		xfer_size = drv->op_sglist->sgl_for_io[i].length;

		tmp_cdb->cdb_read_buffer.alloc_length[0] =
			(unchar)(xfer_size >> 16);
		tmp_cdb->cdb_read_buffer.alloc_length[1] =
			(unchar)(xfer_size >> 8);
		tmp_cdb->cdb_read_buffer.alloc_length[2] =
			(unchar)(xfer_size & 0xFF);

		tmp_cdb->cdb_read_buffer.buf_offset[0] =
			(unchar)(buf_offset >> 16);
		tmp_cdb->cdb_read_buffer.buf_offset[1] =
			(unchar)(buf_offset >> 8);
		tmp_cdb->cdb_read_buffer.buf_offset[2] =
			(unchar)(buf_offset & 0xFF);

		tcdb = (unchar*)tmp_cdb;

		rc = tape_execute_scsi_command(drv->dev_obj, tcdb, DMA_FROM_DEVICE,
			(char*)&(drv->op_sglist->sgl_for_io[i]), xfer_size,
			drv->op_sglist->use_sg,	compute_timeout(id, ReadBufferTimeOut),
			sense_data, &resid);

		buf_offset += xfer_size;

		if(rc != STATUS_SUCCESS) break;
	}

	EXIT_LABEL:
	drv->op_sglist->use_sg = temp_use_sg;

	if(rc != STATUS_SUCCESS)
	{
		WRITE_LOCK(&(drv->dmp_lock));
		if(drv->op_sglist){
			if(tape_del_ops_sglist(drv))
				printk("lin_tape: unable to delete SG list for dump\n");
			drv->dmp_len = 0;
			drv->dmp_size_order = 0;
			drv->op_sg = 0;
		}
		WRITE_UNLOCK(&(drv->dmp_lock));
	}else {
		WRITE_LOCK(&(drv->dmp_lock));
		drv->op_sg = DUMP_READY;
		drv->dmp_len = (ulong)xfer_len;
		WRITE_UNLOCK(&(drv->dmp_lock));
	}
	if(sense_data) kfree(sense_data);
	*drvp = drv;
	return rc;
} /* lin_tape_perform_dump_sg() */


/*******************************************************************************
*                                                                              *
*  NAME: tape_exe_scsi_cmd                                                     *
*                                                                              *
*  PURPOSE: Build CDB and send request                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    cdb: command descriptor block                                             *
*    cdb_len: number of bytes in CDB                                           *
*    buffer: I/O buffer                                                        *
*    buf_len: length of buffer                                                 *
*    timeout: seconds between command issue and timeout                        *
*    direction: to device or from device                                       *
*    result: return code from tape_execute_scsi_command                        *
*    sense_data: buffer into which to copy sense data from returned command    *
*    resid: residual count from sense data, if applicable                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_exe_scsi_cmd(struct tape_descriptor* drv, unchar* cdb,
	unchar cdb_len, void* buffer, int buf_len, int timeout,
	unchar direction, int* result, unchar* sense_data, int* resid)
{
	char ttype = lin_tape_trace_type;
	unchar *trc = NULL;
	int rc = STATUS_SUCCESS, trc_len = 0;

	if(!drv || !drv->dev_obj || drv->ghost_device) {
		printk("lin_tape: IBMtape%d being removed.\n", drv->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(!cdb || (buf_len > 0 && !buffer) || (!buf_len && buffer)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
		trc_len = 1 + cdb_len;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(trc) {
			trc[0] = cdb_len;
			memcpy(&trc[1], cdb, cdb_len);
			tape_setup_and_log_full(drv, drv->cur_path, TRC_CDB,
				trc, trc_len);
			kfree(trc);
			trc = NULL;
		} /* if */

		if(buf_len && (cdb[0] == MODE_SELECT ||
			cdb[0] == DISPLAY_MESSAGE || cdb[0] == SEND_DIAGNOSTIC))
			tape_log_buffer_data(drv, min(buf_len,
				(int)(FullTraceUnitSize - 7)), buffer);
	} /* if */

	/* make sure the HBA is not in its error recovery status */
	if (!scsi_block_when_processing_errors(drv->dev_obj)){
		printk("lin_tape: IBMtape%d scsi errors offline or deleted\n", drv->dev_index);
		rc = -EIO;
	}
    else{
        *result = tape_execute_scsi_command(drv->dev_obj, cdb, direction,
            buffer, buf_len, 0, timeout, sense_data, resid);
        rc = *result;
    }

	if (rc==-ENOMEM) goto CHECK_RC;

	if(!drv->trace || ttype != LIN_TAPE_FULL_TRACE) goto CHECK_RC;

	trc_len = 4;
	trc = kmalloc(trc_len, GFP_KERNEL);
	if(!trc) goto CHECK_RC;

	trc[0] = driver_byte(rc);
	trc[1] = host_byte(rc);
	trc[2] = msg_byte(rc);
	trc[3] = status_byte(rc);

	tape_setup_and_log_full(drv, drv->cur_path, TRC_STATUS, trc, trc_len);
	kfree(trc);
	trc = NULL;

	if(rc == STATUS_SUCCESS) goto CHECK_RC;

	if(!(driver_byte(rc) & DRIVER_SENSE) || !sense_data) goto CHECK_RC;

	trc_len = min(SCSI_SENSE_BUFFERSIZE,
		(int)sizeof(struct tape_sense_data));
	tape_setup_and_log_full(drv, drv->cur_path, TRC_SENSE, sense_data,
		trc_len);

CHECK_RC:

	if(sense_data && (sense_data[2] & 0x0F) == UNIT_ATTENTION)
		rc = UNIT_ATTENTION;
	else if(host_byte(rc) != DID_OK &&
		host_byte(rc) != DID_TARGET_FAILURE &&
		status_byte(rc) != RESERVATION_CONFLICT &&
		status_byte(rc) != SAM_STAT_RESERVATION_CONFLICT)
		rc = -EAGAIN; /* failover */
	else if(rc != STATUS_SUCCESS && rc != -ENOMEM) rc = -EIO;

EXIT_LABEL:

	return rc;
} /* tape_exe_scsi_cmd() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_check_simmim_dump_logsense                                       *
*                                                                              *
*  PURPOSE: Check and retrieve simmim, drive dump, and full logsense           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_check_simmim_dump_logsense(struct tape_descriptor* drv)
{
	char ttype = lin_tape_trace_type;
	if(ttype == LIN_TAPE_INVALID_TRACE) goto EXIT_LABEL;

	if(drv->simmim_avail) read_simmim(drv);
	if(drv->dmp_avail) read_dump(drv);
	if(drv->full_logsense_avail) read_full_logsense(drv);

EXIT_LABEL:

	return;
} /* tape_check_simmim_dump_logsense() */

/*******************************************************************************
*                                                                              *
*  NAME: read_simmim                                                           *
*                                                                              *
*  PURPOSE: retrieve simmim as part of ERP handling                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void read_simmim(struct tape_descriptor* drv)
{
	int erp_rc = 0;
	boolean lockd = False;
	unchar ttype = lin_tape_trace_type;
	unchar* buf = NULL;

	READ_LOCK(&drv->simmim_lock);
	lockd = True;
	if(drv->simmim_len) goto EXIT_LABEL;
	lockd = False;
	READ_UNLOCK(&drv->simmim_lock);

	WRITE_LOCK(&drv->simmim_lock);
	drv->simmim = kmalloc(LogPageHeaderSize + LOGSENSEPAGE, GFP_KERNEL);
	WRITE_UNLOCK(&drv->simmim_lock);

	if(!drv->simmim) {
		DbgPrint(("lin_tape: %s failed to allocate simmim space\n",
			drv->devid));
		erp_rc = ENOMEM;
		goto ERROR_LABEL;
	} /* if */

	drv->in_erp = True;
	erp_rc = tape_send_erp_cmd(drv, LOG_SENSE, 0, SimMimLogPage);
	drv->in_erp = False;

	if(erp_rc == STATUS_SUCCESS) goto EXIT_LABEL;

	WRITE_LOCK(&drv->simmim_lock);
	if(drv->simmim) kfree(drv->simmim);
	drv->simmim = NULL;
	drv->simmim_len = 0;
	WRITE_UNLOCK(&drv->simmim_lock);

	DbgPrint(("lin_tape: %s failed to save simmim, errno %d\n", drv->devid,
		-erp_rc));

ERROR_LABEL:

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg2(drv->tape_log, &drv->tape_log_len,
			TRC_SIMMIM_NOT_SAVED, -erp_rc);
	else if(drv->trace && ttype != LIN_TAPE_INVALID_TRACE) {
		buf = kmalloc(1, GFP_KERNEL);
		if(!buf) goto EXIT_LABEL;

		buf[0] = -erp_rc;
		tape_log_full_trace(drv, TRC_SIMMIM_NOT_SAVED, 1, buf);
		kfree(buf);
		buf = NULL;
	} /* else if */

EXIT_LABEL:

	drv->simmim_avail = False;
	if(lockd) READ_UNLOCK(&drv->simmim_lock);
	return;
} /* read_simmim() */

/*******************************************************************************
*                                                                              *
*  NAME: read_dump                                                             *
*                                                                              *
*  PURPOSE: retrieve drive dump as part of ERP handling                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void read_dump(struct tape_descriptor* drv)
{
	int erp_rc = 0;
	unchar* buf = NULL;
	unchar ttype = lin_tape_trace_type;

	drv->in_erp = True;
	erp_rc = tape_send_erp_cmd(drv, READ_BUFFER, 0, 0);
	drv->in_erp = False;

	if(erp_rc == STATUS_SUCCESS) goto EXIT_LABEL;

	DbgPrint(("lin_tape: %s failed to save dump, errno %d\n", drv->devid,
		-erp_rc));
	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg2(drv->tape_log, &drv->tape_log_len,
			TRC_DMP_NOT_SAVED, -erp_rc);
	else if(drv->trace && ttype != LIN_TAPE_INVALID_TRACE) {
		buf = kmalloc(1, GFP_KERNEL);
		if(!buf) goto EXIT_LABEL;
		buf[0] = -erp_rc;
		tape_log_full_trace(drv, TRC_DMP_NOT_SAVED, 1, buf);
		kfree(buf);
		buf = NULL;
	} /* else if */

EXIT_LABEL:

	drv->dmp_avail = False;
	return;
} /* read_dump() */

/*******************************************************************************
*                                                                              *
*  NAME: read_full_logsense                                                    *
*                                                                              *
*  PURPOSE: retrieve full log sense as part of ERP handling                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void read_full_logsense(struct tape_descriptor* drv)
{
	int erp_rc = 0;

	drv->in_erp = True;
	erp_rc = tape_send_erp_cmd(drv, LOG_SENSE, 0, 0);
	drv->in_erp = False;

	if(erp_rc == STATUS_SUCCESS) goto EXIT_LABEL;

	DbgPrint(("lin_tape: %s log sense failed: %d\n", drv->devid, erp_rc));
	WRITE_LOCK(&drv->logsense_lock);
	if(drv->logsense) kfree(drv->logsense);
	drv->logsense = NULL;
	drv->logsense_len = 0;
	WRITE_UNLOCK(&drv->logsense_lock);

EXIT_LABEL:
	drv->full_logsense_avail = False;
	return;
} /* read_full_logsense() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_unit_attent_erp                                                  *
*                                                                              *
*  PURPOSE: ERP handling for unit attention                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_unit_attent_erp(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS, retry = 0;
	unchar asc, ascq;

	DbgPrint(("lin_tape: tape_unit_attent_erp entry\n"));

	if(!drv) goto EXIT_LABEL;
	if(drv->cmd_in_open && drv->last_op == TEST_UNIT_READY) goto EXIT_LABEL;
	if(drv->cmd_wo_ua_erp) goto EXIT_LABEL;

	drv->in_erp = True;

	if(!drv->sense_valid || !drv->sense_data ||
		drv->sense_data->sense_key != UNIT_ATTENTION) goto EXIT_LABEL;
	asc = drv->sense_data->asc;
	ascq = drv->sense_data->ascq;
	if(asc == 0x28 && ascq == 0) {
		/* for SME, manage write setting, as this may be cleared on */
		/* a not-ready-to-ready transition */
		tape_encryption_write_setting(&drv);
		goto EXIT_LABEL;
	} /* if */

	if((asc != 0x29 || ascq != 0x00) &&
		(asc != 0x2A || ascq != 0x01) &&
		(asc != 0x2A || ascq != 0x02) &&
		(asc != 0x2A || ascq != 0x03) &&
		(asc != 0x2A || ascq != 0x04) &&
		(asc != 0x2A || ascq != 0x05) &&
		(asc != 0x2F || ascq != 0x00) &&
		(asc != 0x3F || ascq != 0x01) &&
		(asc != 0x3F || ascq != 0x02) &&
		(asc != 0x3F || ascq != 0x03)) goto EXIT_LABEL;

	/* reset dynamic attributes if applicable */
	if((drv->drv_id == IBM_3580 && drv_gen(drv) >= 5)
		|| (drv->drv_id == IBM_3592 && drv_gen(drv) >= 4))
		tape_send_erp_cmd(drv, SCSIOP_WRITE_DYNAMIC_RUNTIME, 0, 0);
	drv->pos_valid = False;

	/* make sure the device is unchanged */
	rc = tape_send_erp_cmd(drv, INQUIRY, 0, UnitSerialPage);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(drv->curr_reserve_type == LinTapeReserve6) {
		rc = tape_send_erp_cmd(drv, RESERVE, 0, 0);
		while(rc == UNIT_ATTENTION)
			rc = tape_send_erp_cmd(drv, RESERVE, 0, 0);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} else if(drv->curr_reserve_type == LinTapePersistentReserve
		&& (asc == 0x29 && ascq == 0)) {
		rc = tape_send_erp_cmd(drv, PERSISTENT_RESERVE_IN, 0,
			QUERY_RESERVATIONS);
		while(rc == UNIT_ATTENTION)
			rc = tape_send_erp_cmd(drv, PERSISTENT_RESERVE_IN, 0,
				QUERY_RESERVATIONS);
		if(drv->pr_generation == 0) {
			printk("lin_tape: power-on reset detected.\n");
			rc = tape_send_erp_cmd(drv, PERSISTENT_RESERVE_OUT, 0,
				REGISTER_IGNORE_KEY);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
			rc = tape_send_erp_cmd(drv, PERSISTENT_RESERVE_OUT, 0,
				CREATE_PERSIST_RESERVE);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

			/* report that drive was reserved */
			tape_setup_and_log_full(drv, drv->cur_path,
				TRC_TP_RESERVED, NULL, 0);

			/* update pr_generation */
			rc = tape_send_erp_cmd(drv, PERSISTENT_RESERVE_IN, 0,
				QUERY_RESERVATIONS);
		} /* if */
	} /* else if */

	/* set tape paramters to the drive, includ. TOD and density for 3592 */
	do {
		rc = tape_send_erp_cmd(drv, MODE_SELECT, 0, 0);
	} while(rc == UNIT_ATTENTION);

	if(drv->prevent_removal) {
		retry = 0;
		do {
			rc = tape_send_erp_cmd(drv, PREVENT_MEDIUM_REMOVAL, 0,
				0);
			retry++;
		} while(rc == UNIT_ATTENTION && retry < 5);
	} /* if */

EXIT_LABEL:

	drv->in_erp = False;
	return;
} /* tape_unit_attent_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_post_check1_erp                                                  *
*                                                                              *
*  PURPOSE: ERP handling for 359x post check 1 condition                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_post_check1_erp(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: tape_post_check1_erp entry.\n"));

	if(!drv) return;
	if(drv->drv_id != IBM_3590 && drv->drv_id != IBM_3592) return;
	if(drv->cmd_in_open) return;

	drv->in_erp = True;
	rc = tape_send_erp_cmd(drv, CHANGE_DEFINITION, 0, 0x60);
	drv->pos_valid = False;
	drv->in_erp = False;

	return;
} /* tape_post_check1_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_logsense_erp                                                     *
*                                                                              *
*  PURPOSE: Retrieve log sense page as part of ERP processing                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive                                        *
*    page_code: number of the log page to retrieve                             *
*    logsense_buf: output buffer                                               *
*    logsense_len: output buffer length                                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_logsense_erp(struct tape_descriptor* drv, unchar page_code,
	unchar* logsense_buf, uint* logsense_len)
{
	unchar* page_buf = NULL, * tcdb = NULL, * sense_data = NULL;
	char ttype = lin_tape_trace_type;
	union cdb* tmp_cdb = NULL;
	int rc = 0, result = 0, resid = 0, i = 0, len = 0;

	DbgPrint(("lin_tape: tape_logsense_erp entry.\n"));

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);

	for(i = 0; i < 2; i++) {
		len = i ? *logsense_len : LogPageHeaderSize;

		if(page_buf) kfree(page_buf);
		page_buf = kmalloc(len, GFP_KERNEL);
		if(!page_buf) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(page_buf, '\0', len);

		memset(tmp_cdb, '\0', MAX_CDB_SIZE);
		tmp_cdb->cdb_logsense.op_code = LOG_SENSE;
		tmp_cdb->cdb_logsense.sp = False;
		tmp_cdb->cdb_logsense.ppc = False;
		tmp_cdb->cdb_logsense.pc = True;
		tmp_cdb->cdb_logsense.page_code = page_code;
		tmp_cdb->cdb_logsense.alloc_length[0] = (unchar)(len >> 8);
		tmp_cdb->cdb_logsense.alloc_length[1] = (unchar)(len & 0xFF);

		tcdb = (unchar*)tmp_cdb;
		rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)page_buf,
			len, compute_timeout(drv->drv_id, LogSenseTimeOut),
			DMA_FROM_DEVICE, &result, sense_data, &resid);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_buffer_data(drv, min(len,
				(int)(FullTraceUnitSize - 7)), (void*)page_buf);
		*logsense_len = (page_buf[2] << 8) +
			(page_buf[3]) + LogPageHeaderSize;

		if(*logsense_len > LogPageHeaderSize + LOGSENSEPAGE)
			*logsense_len = LogPageHeaderSize + LOGSENSEPAGE;
	} /* for */

	memcpy(logsense_buf, page_buf, *logsense_len);

EXIT_LABEL:

	if(page_buf) kfree(page_buf);
	if(tmp_cdb) kfree(tmp_cdb);
	if(sense_data) kfree(sense_data);

	return rc;
} /* tape_logsense_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_preempt_erp                                                      *
*                                                                              *
*  PURPOSE: Preempt persistent reservation                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive on which we want to do encryption      *
*    prev_key: key of the currently-held reservation                           *
*    new_key: key of the reservation we are establishing                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_preempt_erp(struct tape_descriptor* drv, unchar* prev_key,
	unchar* new_key)
{
	struct persist_reserve_out* parm_list_buf = NULL;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL, * sense_data = NULL;
	int rc = STATUS_SUCCESS, result = 0, resid = 0;

	DbgPrint(("lin_tape: tape_preempt_erp entry.\n"));

	parm_list_buf = kmalloc(sizeof(struct persist_reserve_out), GFP_KERNEL);
	if(!parm_list_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(parm_list_buf, '\0', sizeof(struct persist_reserve_out));
	memcpy(parm_list_buf->reserve_key, new_key, RESERVE_KEY_LENGTH);
	memcpy(parm_list_buf->service_key, prev_key, RESERVE_KEY_LENGTH);

	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;

	} /* if */
	memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);

	tmp_cdb->cdb_persist_reserve_out.op_code = PERSISTENT_RESERVE_OUT;
	tmp_cdb->cdb_persist_reserve_out.serv_action = PREEMPT_AND_ABORT;
	tmp_cdb->cdb_persist_reserve_out.type = EXCLUSIVE_ACCESS;
	tmp_cdb->cdb_persist_reserve_out.parm_list_length[0] = 0;
	tmp_cdb->cdb_persist_reserve_out.parm_list_length[1] =
		sizeof(struct persist_reserve_out);
	tcdb = (unchar*)tmp_cdb;
	rc = tape_exe_scsi_cmd(drv, tcdb, 10, parm_list_buf,
		sizeof(struct persist_reserve_out),
		compute_timeout(drv->drv_id, PersistReserveOutTimeOut),
		DMA_TO_DEVICE, &result, sense_data, &resid);

	/* remember the sense key if it is a unit attention */
	if(sense_data && ((sense_data[2] & 0x0F) == UNIT_ATTENTION))
		rc = UNIT_ATTENTION;

EXIT_LABEL:

	if(tmp_cdb) kfree(tmp_cdb);
	if(parm_list_buf) kfree(parm_list_buf);
	if(sense_data) kfree(sense_data);
	return rc;
} /* tape_preempt_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_perform_encryption                                               *
*                                                                              *
*  PURPOSE: Handle encryption                                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    timeout: of the command to retry                                          *
*    direction: input or output                                                *
*    xfer_cnt: size of buffer to receive data                                  *
*    retries: number of times this function should retry on failure            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_perform_encryption(struct tape_descriptor** drvp, int timeout,
	unchar direction, size_t xfer_cnt, int retries)
{
	int rc = STATUS_SUCCESS, action_code = 0, return_code = 0;

	DbgPrint(("lin_tape: entering tape_perform_encryption\n"));
	rc = tape_perform_encryption_generic(drvp, 10*timeout, &action_code,
		&return_code);
	if(rc == STATUS_SUCCESS && action_code == DD_RETRY_COMMAND) {
		DbgPrint(("lin_tape: retrying command\n"));
		rc = tape_send_scsi_cmd(drvp, timeout, direction, xfer_cnt,
			retries);
	} else rc = (*drvp)->result;

	return rc;
} /* tape_perform_encryption() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_perform_failover                                                 *
*                                                                              *
*  PURPOSE: Handle failover for drive operations                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    timeout: of the command to retry                                          *
*    direction: input or output                                                *
*    xfer_cnt: size of buffer to receive data                                  *
*    retries: number of times this function should retry on failure            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_perform_failover(struct tape_descriptor **drvp, int timeout,
	unchar direction, size_t xfer_cnt, int retries)
{
	int rc;
	struct tape_descriptor* drv = *drvp;
	struct tape_sgl* sgl_store = NULL;
	void* rw_store = NULL;
	unchar opcode = drv->cdb->cdb_6.op_code;

	rc = tape_perform_failover_generic(&drv, 2 * timeout);

	if(rc == STATUS_COMPLETE) {
		DbgPrint(("lin_tape: \"failed\" command succeeded\n"));
		rc = STATUS_SUCCESS;
		drv->result = 0;
		goto EXIT_LABEL;
	} else if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape: failed command cannot be retried: "\
			"rc = %d\n", rc));
		if (rc != -ENXIO) rc = drv->result;
		goto EXIT_LABEL;
	} else if(!drv || !drv->cdb || drv->cdb->cdb_6.op_code != opcode) {
		/* Drive may have been removed and come back */
		DbgPrint(("lin_tape drv %p: corrupt CDB.  Possible drive "\
			"swap.", drv));
		if(drv) {
			DbgPrint(("  cdb %p", drv->cdb));
			if(drv->cdb) {
				DbgPrint(("  old op 0x%x new op 0x%x", opcode,
				drv->cdb->cdb_6.op_code));
			} /* if */
		} /* if */
		DbgPrint(("\n"));
		rc = -EIO;
		goto EXIT_LABEL;
	} /* else if */

	sgl_store = drv->sg_list;
	rw_store = drv->rw_buf;

	if(sgl_store) {
		drv->sg_list = kmalloc(sizeof(struct tape_sgl), GFP_KERNEL);
		if(!drv->sg_list) {
			drv->sg_list = sgl_store;
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memcpy(drv->sg_list, sgl_store, sizeof(struct tape_sgl));
	} /* if */

	rc = adjust_dpf_write_ptr(drv, &xfer_cnt);
	if(rc != STATUS_SUCCESS) {
		if(drv->sg_list) kfree(drv->sg_list);
		drv->sg_list = sgl_store;
		drv->rw_buf = rw_store;
		goto EXIT_LABEL;
	} /* if */

	DbgPrint(("lin_tape: retrying command\n"));
	rc = tape_send_scsi_cmd(&drv, timeout, direction, xfer_cnt, retries);

	if(drv->sg_list) kfree(drv->sg_list);
	drv->sg_list = sgl_store;
	drv->rw_buf = rw_store;
	sgl_store = rw_store = NULL;
	goto EXIT_LABEL;

EXIT_LABEL:

	*drvp = drv;

	return rc;
} /* tape_perform_failover() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_perform_failover_generic                                         *
*                                                                              *
*  PURPOSE: Perform generic operations associated with failover                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    timeout: of the command that triggered the failover                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_perform_failover_generic(struct tape_descriptor **drvp, int timeout)
{
	struct tape_descriptor* drv = *drvp;
	char* trc = NULL;
	int rc = -EIO;
	unchar ttype = 0;

	DbgPrint(("lin_tape: tape_perform_failover_generic entry.\n"));

	drv->do_failover = False;

	/* if daemon is not running, return failure */
	READ_LOCK(&lin_tape_trace_lock);
	if(lin_tape_trace_type == LIN_TAPE_INVALID_TRACE) {
		printk("lin_tape: lin_taped daemon is not running, IBMtape%d "\
			"cannot failover.\n", drv->dev_index);
		READ_UNLOCK(&lin_tape_trace_lock);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&lin_tape_trace_lock);
	rc = post_dpf_msgs_to_log(drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	DbgPrint(("lin_tape: IBMtape%d is to wait.\n", drv->dev_index));

	rc = wait_event_interruptible_timeout(drv->wait_queue,
		!drv || !drvp || drv->ready_wakeup, timeout);
	if(!drv || !drvp) {
		printk("Tape removed while waiting\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(rc < 0) {
		printk("lin_tape: IBMtape%d failover failed due to "\
			"unexpected interrupt. rc = %d\n", drv->dev_index, rc);
		rc = drv->result;
		goto EXIT_LABEL;
	} else if(rc == 0) {
		printk("lin_tape: IBMtape%d failover failed due to timeout.\n",
			drv->dev_index);
		rc = drv->result;
		goto EXIT_LABEL;
	} /* else */

	DbgPrint(("lin_tape: IBMtape%d is woken up on a new path "\
		"IBMtape%d\n", drv->dev_index, drv->cur_path));

	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->ready_wakeup = False;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	if(drv->wakeup_result < STATUS_SUCCESS) {
		printk("lin_tape: IBMtape%d failover failed.\n",
			drv->dev_index);
		rc = drv->result;
		goto EXIT_LABEL;
	} /* if */

	DbgPrint(("lin_tape: set current path to: IBMtape%d\n", drv->cur_path));

	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg5(drv->tape_log, &(drv->tape_log_len),
			TRC_TP_DPF_SUCCESS, (drv->dev_index >> 8),
			drv->dev_index & 0xFF, (drv->cur_path >> 8),
			drv->cur_path & 0xFF);
	else if(ttype == LIN_TAPE_FULL_TRACE) {
		trc = kmalloc(4, GFP_KERNEL);
		if(trc) {
			trc[0] = drv->dev_index >> 8;
			trc[1] = drv->dev_index & 0xFF;
			trc[2] = drv->cur_path >> 8;
			trc[3] = drv->cur_path & 0xFF;
			tape_log_full_trace(drv, TRC_TP_DPF_SUCCESS, 4, trc);
		} /* if */
	} /* else if */

	tape_update_to_new_current_path(&drv);

	if(drv->prevent_removal) {
		rc = tape_send_erp_cmd(drv, PREVENT_MEDIUM_REMOVAL, 0, 0);
		if(rc) {
			printk("lin_tape_perform_failover_generic: "\
				"cannot re-establish prevent med. removal.\n");
		} /* if */
	} /* if */

	if((drv->drv_id == IBM_3580 && drv_gen(drv) >= 5)
		|| (drv->drv_id == IBM_3592 && drv_gen(drv) >= 4))
		tape_send_erp_cmd(drv, SCSIOP_WRITE_DYNAMIC_RUNTIME, 0, 0);

	if(drv->last_op == VERIFY_TAPE_DATA) {
		rc = handle_verify_on_failover(drv);
		if(rc != STATUS_SUCCESS) {
			printk("VERIFY failover ops: no retry. rc %d.\n", rc);
			printk("\tSetting prv wakeup result %d to NO_RETRY\n",
				drv->wakeup_result);
			drv->wakeup_result = STATUS_NO_RETRY;
		} /* if */
	} /* if */

	*drvp = drv;

	rc = drv->wakeup_result;

EXIT_LABEL:

	if(trc) kfree(trc);
	DbgPrint(("lin_tape: tape_perform_failover_generic exit: %d\n", rc));
	return rc;
} /* tape_perform_failover_generic() */

/*******************************************************************************
*                                                                              *
*  NAME: post_dpf_msgs_to_log                                                  *
*                                                                              *
*  PURPOSE: Notify the daemon to failover, via the message log                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int post_dpf_msgs_to_log(struct tape_descriptor* drv)
{
	char* trc = NULL;
	int rc = STATUS_SUCCESS, trc_len = 0;

	/* update tape-motion status */
	trc_len = drv->cdb_len + 10;
	trc = kmalloc(trc_len, GFP_KERNEL);
	if(!trc) {
		printk("lin_tape: IBMtape%d cannot failover due to memory "\
			"allocation failure.\n", drv->dev_index);
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	trc[0] = drv->cdb_len;
	memcpy(&trc[1], drv->cdb, drv->cdb_len);
	trc[drv->cdb_len + 1] = drv->pos_valid;
	trc[drv->cdb_len + 2] = (drv->curr_pos >> 24) & 0xFF;
	trc[drv->cdb_len + 3] = (drv->curr_pos >> 16) & 0xFF;
	trc[drv->cdb_len + 4] = (drv->curr_pos >> 8) & 0xFF;
	trc[drv->cdb_len + 5] = (drv->curr_pos) & 0xFF;
	trc[drv->cdb_len + 6] = (drv->blk_size >> 24) & 0xFF;
	trc[drv->cdb_len + 7] = (drv->blk_size >> 16) & 0xFF;
	trc[drv->cdb_len + 8] = (drv->blk_size >> 8) & 0xFF;
	trc[drv->cdb_len + 9] = (drv->blk_size) & 0xFF;

	rc = tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_RPT_POS, trc,
		trc_len);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_ATMPT_FAILOVER,
		NULL, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;

} /* post_dpf_msgs_to_log() */

/*******************************************************************************
*                                                                              *
*  NAME: adjust_dpf_write_ptr                                                  *
*                                                                              *
*  PURPOSE: Advance the data pointer in the write buffer for partial retry     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive on which we want to fail over          *
*    xfer_cnt: transfer count to update                                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    If part of the write succeeded before failover we may need to perform a   *
*    partial write.  We cannot try the entire command if this is a WORM        *
*    cartridge; this algorithm is designed to work with WORM and non-WORM.     *
*                                                                              *
*******************************************************************************/

int adjust_dpf_write_ptr(struct tape_descriptor* drv, size_t* xfer_cnt)
{
	int rc = STATUS_SUCCESS;
	ulong orig_xfer_len = 0, blks_to_advance = 0, bytes_to_advance = 0;
	unchar* cdb_xlen = NULL, op = 0;

	DbgPrint(("lin_tape: adjust_dpf_write_ptr entry\n"));

	op = (drv->cdb)->cdb_6.op_code;
	if(op != WRITE_6 && op != WRITE_FILEMARKS) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->wakeup_info) {
		/* This condition should have returned STATUS_COMPLETE */
		DbgPrint(("lin_tape: remainder == 0.  EINVAL\n"));
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	switch(op) {
	case WRITE_6:
		if(!drv->blk_size) break; /* retry entire operation */

		cdb_xlen = (drv->cdb)->cdb_read_write.xfer_length;
		orig_xfer_len = (cdb_xlen[0] << 16) + (cdb_xlen[1] << 8) +
			(cdb_xlen[2]);
		cdb_xlen[0] = (unchar)(drv->wakeup_info >> 16 & 0xFF);
		cdb_xlen[1] = (unchar)(drv->wakeup_info >> 8 & 0xFF);
		cdb_xlen[2] = (unchar)(drv->wakeup_info & 0xFF);
		blks_to_advance = orig_xfer_len - drv->wakeup_info;
		bytes_to_advance = blks_to_advance * drv->blk_size;

		if(drv->direct_io) {
			rc = adjust_dpf_sgl(drv, drv->sg_list->gather,
				bytes_to_advance);
		} else if(BUF_RW(drv)) {
			drv->rw_buf += bytes_to_advance;
			rc = STATUS_SUCCESS;
		} else if(SGL_RW(drv)) {
			rc = adjust_dpf_sgl(drv, drv->sg_list->sgl_for_io,
				bytes_to_advance);
		} else {
			DbgPrint(("lin_tape: corrupt rw buf\n"));
			rc = -EFAULT;
		} /* else */
		if(rc == STATUS_SUCCESS) *xfer_cnt -= bytes_to_advance;
		break;

	case WRITE_FILEMARKS:
		(drv->cdb)->cdb_wfm.count[0]
			= (unchar)(drv->wakeup_info >> 16 & 0xFF);
		(drv->cdb)->cdb_wfm.count[1]
			= (unchar)(drv->wakeup_info >> 8 & 0xFF);
		(drv->cdb)->cdb_wfm.count[2]
			= (unchar)(drv->wakeup_info & 0xFF);
		break;

	default:
		break;
	} /* switch */

EXIT_LABEL:

	DbgPrint(("lin_tape: adjust_dpf_write_ptr exit: %d\n", rc));
	return rc;

} /* adjust_dpf_write_ptr() */

/*******************************************************************************
*                                                                              *
*  NAME: adjust_dpf_sgl                                                        *
*                                                                              *
*  PURPOSE: Reorganize scatter-gather list used for partial write              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for which to retry a partial write                   *
*    sgl: the scatterlist to adjust                                            *
*    bytes_to_advance: place to begin next write                               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS                                                     *
*                                                                              *
*******************************************************************************/

int adjust_dpf_sgl(struct tape_descriptor* drv, struct scatterlist* sgl,
	ulong bytes_to_advance)
{
	int remainder = bytes_to_advance, i = 0, start = 0, new_len = 0,
		new_offs = 0;
	char* address = NULL;

	for(i = 0; remainder; i++) {
		if(sgl[i].length <= remainder) {
			remainder -= sgl[i].length;
			drv->sg_list->use_sg--;
			if(sgl == drv->sg_list->gather) {
				drv->sg_list->total_sgl--;
				drv->sg_list->total_sg_size -= sgl[i].length;
			} /* if */
			sg_set_page(&sgl[i], NULL, 0, 0);
			if(!remainder) start = i + 1;
		} else {
			address = sg_virt(&sgl[i]) + remainder;
			new_len = sgl[i].length - remainder;
			new_offs = sgl[i].offset + remainder;
			remainder = 0;
			sg_set_page(&sgl[i], virt_to_page(address), new_len,
				new_offs);
			start = i;
		} /* else */
	} /* for */

	if(!start) return STATUS_SUCCESS;

	for(i = start; i < start + drv->sg_list->use_sg; i++) {
		sg_set_page(&sgl[i - start], sg_page(&sgl[i]),
			sgl[i].length, sgl[i].offset);
		sg_set_page(&sgl[i], NULL, 0, 0);
	} /* for */

	return STATUS_SUCCESS;
} /* adjust_dpf_sgl() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_update_to_new_current_path                                       *
*                                                                              *
*  PURPOSE: Copy old parms to new path and clean up                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

void tape_update_to_new_current_path(struct tape_descriptor **drvp)
{
	struct tape_descriptor* drv = *drvp, * trav = drv->next_path;

	WRITE_LOCK(&tape_des_lock);
	while(trav != drv){
		trav->cur_path = drv->cur_path;
		trav = trav->next_path;

	}
	WRITE_UNLOCK(&tape_des_lock);

	copy_parms_to_new_device(drv->dev_index, drv->cur_path);

	/* only clear these out if the drive path truly changed */
	if(drv->cur_path != drv->dev_index)
		delete_parms_on_device(drv->dev_index);

	*drvp = tape_devs[drv->cur_path]->drv;

} /* tape_update_to_new_current_path() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_best_path                                                    *
*                                                                              *
*  PURPOSE: if daemon is running, post request for best path                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_get_best_path(struct tape_descriptor** drvp)
{
	int rc = -EIO, trc_len = 0, cur=(*drvp)->cur_path, dev_index = 0;
	struct tape_descriptor* drv = *drvp;
	char* trc = NULL;

	DbgPrint(("lin_tape: checking for the best path on IBMtape%d\n",
		drv->dev_index));

	/* if daemon is not running, do nothing */
	READ_LOCK(&lin_tape_trace_lock);
	if(lin_tape_trace_type == LIN_TAPE_INVALID_TRACE){
		DbgPrint(("lin_tape: lin_taped daemon is not running, "\
			"IBMtape%d will use the current path.\n",
			drv->dev_index));
		READ_UNLOCK(&lin_tape_trace_lock);
		goto EXIT_LABEL;
	} /* if */

	READ_UNLOCK(&lin_tape_trace_lock);

	trc_len = 5;
	trc = kmalloc(trc_len, GFP_KERNEL);

	if(!trc) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(trc, '\0', trc_len);

	rc = tape_setup_and_log_full(drv, drv->dev_index, TRC_TP_PATH_CHECK,
		trc, trc_len);

	if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape: IBMtape%d cannot check for better path "\
			"due to memory allocation failure\n",drv->dev_index));
		goto EXIT_LABEL;
	} /* if */

	/* wait */
	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->ready_wakeup = False;
	dev_index = drv->dev_index;
	DbgPrint(("lin_tape: IBMtape%d is to wait.\n", drv->dev_index));
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	rc = wait_event_interruptible_timeout(drv->wait_queue,
		(!drvp || !drv || drv->ready_wakeup || drv->remove_pending), 1800 * HZ);
	if(!drvp || !drv || drv->remove_pending) {
		printk("drive removed while waiting\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if(rc < 0) {
		printk("lin_tape: IBMtape%d path checking failed due to "\
			"unexpected interrupt. rc = %d\n", drv->dev_index, rc);
		goto EXIT_LABEL;
	} else if(rc == 0) {
		if(drv->wakeup_result)
			printk("lin_tape: IBMtape%d path checking failed, "\
				"wakeup_result = %d\n", drv->dev_index,
				drv->wakeup_result);
		else printk("lin_tape: IBMtape%d path checking failed due to "\
			"timeout.\n", drv->dev_index);
		goto EXIT_LABEL;
	} else {
		DbgPrint(("lin_tape: IBMtape%d will use the IBMtape%d path\n",
			drv->dev_index, drv->cur_path));

        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		drv->ready_wakeup = False;
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	} /* else */

	DbgPrint(("lin_tape: set current path to: IBMtape%d\n", drv->cur_path));
    if (cur != drv->cur_path)
	tape_update_to_new_current_path(&drv);

EXIT_LABEL:

	if(trc) kfree(trc);
	*drvp = drv;
	return;
} /* tape_get_best_path() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_query_path                                                       *
*                                                                              *
*  PURPOSE: Display primary path and first alternate path for this drive       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_query_path(struct tape_descriptor** drvp, void* buf, char unused)
{
	int name_length = 0, i = 0, count = 0, rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;
	struct scsi_path* buffer = (struct scsi_path*)buf;
	boolean lockd = False;

	DbgPrint(("lin_tape: tape_query_path entry\n"));

	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buffer, '\0', sizeof(struct scsi_path));

	name_length = min(30, MaxLogicName);

	READ_LOCK(&tape_des_lock);
	lockd = True;
	if(!drv->alt_pathing || !drv->serial_num_len) {
		memcpy(buffer->primary_name, drv->lname, name_length);
		memcpy(buffer->primary_parent, drv->pname, name_length);
		buffer->primary_host = drv->host;
		buffer->primary_id = drv->target;
		buffer->primary_lun = drv->lun;
		buffer->primary_bus = drv->channel;
		buffer->primary_enabled = drv->enabled;
		buffer->primary_id_valid = 1;
		buffer->primary_fenced = drv->fenced;
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv
		|| tape_devs[i]->drv->serial_num_len != drv->serial_num_len ||
		memcmp(tape_devs[i]->drv->serial_num, drv->serial_num,
			drv->serial_num_len)) continue;
		if(!count) {
			memcpy(buffer->primary_name,
				tape_devs[i]->drv->lname, name_length);
			memcpy(buffer->primary_parent,
				tape_devs[i]->drv->pname, name_length);
			buffer->primary_host = tape_devs[i]->drv->host;
			buffer->primary_id = tape_devs[i]->drv->target;
			buffer->primary_lun = tape_devs[i]->drv->lun;
			buffer->primary_bus = tape_devs[i]->drv->channel;
			buffer->primary_enabled = tape_devs[i]->drv->enabled;
			buffer->primary_id_valid = 1;
			buffer->primary_fenced = tape_devs[i]->drv->fenced;
			count++;
			continue;
		} /* if */

		buffer->alternate_configured = True;
		memcpy(buffer->alternate_name, tape_devs[i]->drv->lname,
			name_length);
		memcpy(buffer->alternate_parent, tape_devs[i]->drv->pname,
			name_length);
		buffer->alternate_host = tape_devs[i]->drv->host;
		buffer->alternate_id = tape_devs[i]->drv->target;
		buffer->alternate_lun = tape_devs[i]->drv->lun;
		buffer->alternate_bus = tape_devs[i]->drv->channel;
		buffer->alternate_enabled = tape_devs[i]->drv->enabled;
		buffer->alternate_id_valid = 1;
		buffer->alternate_fenced = tape_devs[i]->drv->fenced;
		count++;
		if(count >= 2) break;
	} /* for */

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	if(lockd) READ_UNLOCK(&tape_des_lock);
	*drvp = drv;
	return rc;
} /* tape_query_path() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_show_path                                                        *
*                                                                              *
*  PURPOSE: Display primary path and all alternate paths for this drive        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    buf: output buffer                                                        *
*    unused: so we can pass this function as a pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_show_paths(struct tape_descriptor** drvp, void* buf, char unused)
{
	int i = 0, name_length = 0, count = 0, rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;
	struct device_paths* buffer = (struct device_paths*)buf;
	boolean lockd = False;

	DbgPrint(("tape_show_paths entry\n"));

	if(!buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buffer, '\0', sizeof(struct device_paths));

	name_length = min(30, MaxLogicName);

	READ_LOCK(&tape_des_lock);
	lockd = True;
	if(!drv->alt_pathing || !drv->serial_num_len) {
		buffer->number_paths = 1;
		memcpy(buffer->path[0].name, drv->lname, name_length);
		memcpy(buffer->path[0].parent, drv->pname, name_length);
		buffer->path[0].host = drv->host;
		buffer->path[0].id = drv->target;
		buffer->path[0].lun = drv->lun;
		buffer->path[0].bus = drv->channel;
		buffer->path[0].enabled = drv->enabled;
		buffer->path[0].fenced = drv->fenced;
		buffer->path[0].id_valid = 1;
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv
			|| tape_devs[i]->drv->serial_num_len != drv->serial_num_len ||
			memcmp(tape_devs[i]->drv->serial_num,
				drv->serial_num, drv->serial_num_len)) continue;

		memcpy(buffer->path[count].name, tape_devs[i]->drv->lname,
			name_length);
		memcpy(buffer->path[count].parent, tape_devs[i]->drv->pname,
			name_length);
		buffer->path[count].host = tape_devs[i]->drv->host;
		buffer->path[count].id = tape_devs[i]->drv->target;
		buffer->path[count].lun = tape_devs[i]->drv->lun;
		buffer->path[count].bus = tape_devs[i]->drv->channel;
		buffer->path[count].enabled = tape_devs[i]->drv->enabled;
		buffer->path[count].fenced = tape_devs[i]->drv->fenced;
		buffer->path[count].id_valid = 1;

		count++;
		if(count >= MAX_SCSI_PATH) break;
	} /* for */
	buffer->number_paths = count;

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	if(lockd) READ_UNLOCK(&tape_des_lock);
	*drvp = drv;
	return rc;
} /* tape_show_paths() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_qry_blk_prot                                                     *
*                                                                              *
*  PURPOSE: Query for logical block protection parameters                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: the struct logical_block_protection                                  *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_qry_blk_prot(struct tape_descriptor** drvp, void* buf, char unused)
{
	int rc = STATUS_SUCCESS, size = 0;
	struct logical_block_protection* lbp =
		(struct logical_block_protection*)buf;
	struct inquiry_extended_data* inq = NULL;
	struct tape_descriptor* drv = *drvp;
	struct control_pg_dp_subp_with_hdr* subpage = NULL;

	DbgPrint(("lin_tape_qry_blk_prot entry \n"));
	inq = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(inq, '\0', MAX_INQ_LEN);

	rc = tape_inquiry_page(&drv, EXTENDED_INQUIRY_PAGE, (char*)inq);
	if(rc) {
		DbgPrint(("lin_tape: qry_blk_prot cannot get inquiry 0x86\n"));
		goto EXIT_LABEL;
	} /* if */

	if(inq->spt != 0x1) {
		DbgPrint(("lin_tape: logical block protection not "\
			"supported. spt 0x%x\n", inq->spt));
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	lbp->lbp_capable = True;

	size = sizeof(struct control_pg_dp_subp_with_hdr);
	subpage = kmalloc(size, GFP_KERNEL);
	if(!subpage) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(subpage, '\0', size);

	rc = tape_modesense10_page(&drv, MODE_SENSE_CONTROL_PG, (char*)subpage,
		size, CONTROL_PG_DP_SUBPG, True, True);
	if(rc) {
		printk("lin_tape: qry_blk_prot cannot get mode pg 0x0A "\
			"sub 0xF0\n");
		goto EXIT_LABEL;
	} /* if */

	lbp->lbp_method = subpage->page.lb_prot_method;
	lbp->lbp_info_length = subpage->page.lb_prot_info_length;
	lbp->lbp_w = subpage->page.lbp_w;
	lbp->lbp_r = subpage->page.lbp_r;
	lbp->rbdp = subpage->page.rbdp;

	store_lbp(drv, lbp);

EXIT_LABEL:

	*drvp = drv;
	if(inq) kfree(inq);
	if(subpage) kfree(subpage);
	DbgPrint(("lin_tape: qry_blk_prot exit: %d\n", rc));
	return rc;
} /* tape_qry_blk_prot() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_set_blk_prot                                                     *
*                                                                              *
*  PURPOSE: Set logical block protection parameters                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: the struct logical_block_protection                                  *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_set_blk_prot(struct tape_descriptor** drvp, void* buf, char unused)
{
	int rc = STATUS_SUCCESS, size = 0;
	struct logical_block_protection* lbp =
		(struct logical_block_protection*)buf;
	struct tape_descriptor* drv = *drvp;
	struct control_pg_dp_subp_with_hdr* subpage = NULL;

	DbgPrint(("lin_tape_set_blk_prot entry\n"));
	store_lbp(drv, NULL);
	size = sizeof(struct control_pg_dp_subp_with_hdr);
	subpage = kmalloc(size, GFP_KERNEL);
	if(!subpage) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(subpage, '\0', size);

	rc = tape_modesense10_page(&drv, MODE_SENSE_CONTROL_PG, (char*)subpage,
		size, CONTROL_PG_DP_SUBPG, True, True);
	if(rc) {
		printk("lin_tape: set_blk_prot cannot get mode pg 0x0A "\
			"sub 0xF0\n");
		goto EXIT_LABEL;
	} /* if */

	subpage->page.ps = 0;
	subpage->page.spf = 1;
	subpage->page.lb_prot_method = lbp->lbp_method;

	if(lbp->lbp_method) {
		subpage->page.lb_prot_info_length = lbp->lbp_info_length;
		subpage->page.rbdp = lbp->rbdp;
		subpage->page.lbp_r = lbp->lbp_r;
		subpage->page.lbp_w = lbp->lbp_w;
	} else {
		subpage->page.lb_prot_info_length =
		subpage->page.rbdp =
		subpage->page.lbp_r =
		subpage->page.lbp_w = 0;
	} /* else */

	rc = tape_modeselect10(&drv, (unchar*)subpage, False, size);
	if(rc) {
		printk("lin_tape: set_blk_prot cannot set mode pg 0x0A "\
			"sub 0xF0\n");
		goto EXIT_LABEL;
	} /* if */

	store_lbp(drv, lbp);

EXIT_LABEL:

	*drvp = drv;
	DbgPrint(("lin_tape: set_blk_prot exit: %d\n", rc));
	return rc;
} /* tape_set_blk_prot() */

/*******************************************************************************
*                                                                              *
*  NAME: store_lbp                                                             *
*                                                                              *
*  PURPOSE: Store logical block protection setting for reset on failover       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive for which the LBP setting is to be stored                  *
*    lbp: the structure to set, or NULL if setting is false                    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void store_lbp(struct tape_descriptor* drv,
	struct logical_block_protection* lbp)
{
	struct tape_descriptor* orig = drv, * trav = drv;

	WRITE_LOCK(&tape_des_lock);
	do {
		if(lbp) {
			trav->lbp_set = True;
			memcpy(&trav->lbp, lbp,
				sizeof(struct logical_block_protection));
		} else {
			trav->lbp_set = False;
			memset(&trav->lbp, '\0',
				sizeof(struct logical_block_protection));
		} /* else */
		trav = trav->next_path;
	} while(trav != orig);
	WRITE_UNLOCK(&tape_des_lock);

	return;
} /* store_lbp() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_query_eot_warn                                                  *
*                                                                              *
*  PURPOSE: Return current value of eot_warn field                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which eot_warn is being queried                       *
*    command: the IOCTL code itself                                            *
*    arg: the eot_warn structure into which to copy the data                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_query_eot_warn(struct tape_descriptor** drvp, uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct eot_warn* wrn = NULL;

	if(_IOC_SIZE(command) != sizeof(struct eot_warn)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	wrn = kmalloc(sizeof(struct eot_warn), GFP_KERNEL);
	if(!wrn) {
		printk("stioc_query_eot_warn: MEMALLOC FAILURE!\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(wrn, '\0', sizeof(struct eot_warn));

	wrn->warn = (*drvp)->eot_warn;

	if(copy_to_user((void*)arg, (void*)wrn, sizeof(struct eot_warn))) {
		printk("stioc_query_eot_warn: cannot copy to user!\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(wrn) kfree(wrn);
	return rc;
} /* stioc_query_eot_warn() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_set_eot_warn                                                    *
*                                                                              *
*  PURPOSE: Set current value of eot_warn field                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which eot_warn is being set                           *
*    command: the IOCTL code itself                                            *
*    arg: the eot_warn structure from which to copy the data                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_set_eot_warn(struct tape_descriptor** drvp, uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct eot_warn* wrn = NULL;

	if(_IOC_SIZE(command) != sizeof(struct eot_warn)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	wrn = kmalloc(sizeof(struct eot_warn), GFP_KERNEL);
	if(!wrn) {
		printk("stioc_set_eot_warn: MEMALLOC FAILURE!\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(wrn, '\0', sizeof(struct eot_warn));

	if(copy_from_user((void*)wrn, (void*)arg, sizeof(struct eot_warn))) {
		printk("stioc_set_eot_warn: cannot copy from user!\n");
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	(*drvp)->eot_warn = wrn->warn;

EXIT_LABEL:

	if(wrn) kfree(wrn);
	return rc;
} /* stioc_set_eot_warn() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_verify_tape_data                                                *
*                                                                              *
*  PURPOSE: Issue VERIFY command to drive                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: the struct verify_data                                               *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stioc_verify_tape_data(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct tape_descriptor* drv = *drvp;
	struct verify_data* vd = (struct verify_data*)buf;
	struct pos_info* read_pos_buf = NULL;
	int rc = STATUS_SUCCESS, timeout = 0;

	if(!drv->cdb || !buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;
	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->verify.op_code = VERIFY_TAPE_DATA;
	drv->cdb->verify.fixed = vd->fixed;
	drv->cdb->verify.bytcmp = vd->bytcmp;
	drv->cdb->verify.immed = vd->immed;
	drv->cdb->verify.vbf = vd->vbf;
	drv->cdb->verify.vlbpm = vd->vlbpm;
	drv->cdb->verify.vte = vd->vte;
	if(!vd->vte)
		memcpy(drv->cdb->verify.verify_length, vd->verify_length, 3);
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = VERIFY_TAPE_DATA;

	/* following timeout is established for 3592.  3580 is TBD */
	timeout = (vd->vte || vd->vbf) ? 21600 * HZ : 1080 * HZ;
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_NONE, 0, False);

	/* Drive position invalidated by VERIFY.  Will check and update below */
	drv->pos_valid = False;

	if(rc != STATUS_SUCCESS) {
		printk("verify_tape_data: rc %d\n", rc);
		goto EXIT_LABEL;
	} /* if */

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct pos_info));

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);
	if(rc == STATUS_SUCCESS && drv->eom_blk_id) {
		if(read_pos_buf->first_blk_pos < drv->eom_blk_id) {
			drv->deferred_eom = False;
			drv->eom_blk_id = 0;
		} /* if */
	} /* if */
	drv->eom = False;

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);

	*drvp = drv;
	return rc;
} /* stioc_verify_tape_data() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_verify_on_failover                                             *
*                                                                              *
*  PURPOSE: If verify with immediate bit is in progress on DPF, handle it      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int handle_verify_on_failover(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS, i = 0, orig_pos = 0;

	if(!drv || !drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(drv->cdb->verify.immed) {
		drv->sense_valid = False;
		for(i = 0; i < 10; i++) {
			rc = tape_send_erp_cmd(drv, REQUEST_SENSE, 0, 0);

			if(rc == STATUS_SUCCESS && drv->sense_data->valid
				&& drv->sense_data->asc == 0x00
				&& drv->sense_data->ascq == 0x1C) {
				/* VERIFY operation still in progress */
				rc = STATUS_NO_RETRY;
				goto EXIT_LABEL;
			} /* if */
		} /* for */
	} /* if */

	orig_pos = drv->curr_pos;
	rc = tape_send_erp_cmd(drv, READ_POSITION, 0, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(orig_pos != drv->curr_pos)
		rc = tape_send_erp_cmd(drv, LOCATE, 0,
			drv->cdb->verify.vte ? drv->curr_pos - 1 : orig_pos);

EXIT_LABEL:

	return rc;
} /* handle_verify_on_failover() */

/*******************************************************************************
*                                                                              *
*  NAME: copy_parms_to_new_device                                              *
*                                                                              *
*  PURPOSE: Copy information from old tape descriptor to new tape descriptor   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    old_dev_index: index of previous tape descriptor                          *
*    new_dev_index: index of current tape descriptor                           *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void copy_parms_to_new_device(int old_dev_index, int new_dev_index)
{
	struct tape_descriptor* prv = tape_devs[old_dev_index]->drv,
		* cur = tape_devs[new_dev_index]->drv;

	WRITE_LOCK(&tape_devs[cur->dev_index]->drv_entry_lock);
	cur->cdb = prv->cdb;
	cur->cdb_len = prv->cdb_len;
	cur->buffer = prv->buffer;
	cur->buf_len = prv->buf_len;
	cur->direct_io = prv->direct_io;
	cur->sg_list = prv->sg_list;
	cur->rw_buf = prv->rw_buf;
	cur->rw_buf_order = prv->rw_buf_order;
	cur->rw_buf_size = prv->rw_buf_size;
	cur->busy = prv->busy & 0xEC;
	cur->eom = prv->eom;
	cur->eom_blk_id = prv->eom_blk_id;
	cur->eod = prv->eod;
	cur->last_op = prv->last_op;
	cur->last_tape_motion = prv->last_tape_motion;
	cur->result = prv->result;
	cur->prevent_removal = prv->prevent_removal;
	cur->drv_open = prv->drv_open;
	cur->opened = prv->opened;
	cur->cmd_in_open = prv->cmd_in_open;
	cur->blk_size = prv->blk_size;
	cur->blk_num = prv->blk_num;
	cur->buffered_mode = prv->buffered_mode;
	cur->capacity_scaling = prv->capacity_scaling;
	cur->capacity_scaling_hex_val = prv->capacity_scaling_hex_val;
	cur->compression = prv->compression;
	cur->density = prv->density;
	cur->ending_fm = prv->ending_fm;
	cur->logic_write_protect = prv->logic_write_protect;
	cur->medium_type = prv->medium_type;
	cur->write_protect = prv->write_protect;
	cur->set_density_valid = prv->set_density_valid;
	cur->set_default_density = prv->set_default_density;
	cur->set_pending_density = prv->set_pending_density;
	cur->data_safe_mode = prv->data_safe_mode;
	cur->pews[0] = prv->pews[0];
	cur->pews[1] = prv->pews[1];
	cur->busy_retry = prv->busy_retry;
	cur->curr_pos = prv->curr_pos;
	cur->pos_valid = prv->pos_valid;
	cur->access_mode = prv->access_mode;
	cur->acf_mode = prv->acf_mode;
	cur->autoload = prv->autoload;
	cur->autoload_mode = prv->autoload_mode;
	cur->max_blk_size = prv->max_blk_size;
	cur->max_xfer_len = prv->max_xfer_len;
	cur->min_blk_size = prv->min_blk_size;
	cur->read_mode = prv->read_mode;
	cur->sili = prv->sili;
	cur->record_space_mode = prv->record_space_mode;
	cur->retain_reservation = prv->retain_reservation;
	cur->rew_immed = prv->rew_immed;
	cur->special_file_type = prv->special_file_type;
	cur->trailer_labels = prv->trailer_labels;
	cur->wfm_immed = prv->wfm_immed;
	cur->limit_read_recov = prv->limit_read_recov;
	cur->limit_write_recov = prv->limit_write_recov;
	cur->lbp_set = prv->lbp_set;
	memcpy(&cur->lbp,&prv->lbp,sizeof(struct logical_block_protection));
	cur->curr_reserve_type = prv->curr_reserve_type;
	cur->explicit_reserved = prv->explicit_reserved;
    cur->eot_warn = prv->eot_warn;
	cur->deferred_eom = prv->deferred_eom;
	cur->fm_detected = prv->fm_detected;
	cur->ili = prv->ili;
	cur->resid_valid = prv->resid_valid;
	cur->resid = prv->resid;
	cur->sense_data = prv->sense_data;
	cur->sense_valid = prv->sense_valid;
	cur->in_erp = prv->in_erp;
	cur->wfm_on_close_safe = prv->wfm_on_close_safe;
	cur->cmd_wo_ua_erp = prv->cmd_wo_ua_erp;
	cur->last_errno = prv->last_errno;
	cur->trace = prv->trace;
	cur->disable_simmim_log = prv->disable_simmim_log;
	cur->auto_logsense = prv->auto_logsense;
    cur->disable_auto_drive_dump = prv->disable_auto_drive_dump;
	cur->trc_hook = prv->trc_hook;
	cur->simmim_avail = prv->simmim_avail;
	cur->simmim = prv->simmim;
	cur->simmim_len = prv->simmim_len;
	cur->full_logsense_avail = prv->full_logsense_avail;
	cur->logsense = prv->logsense;
	cur->logsense_len = prv->logsense_len;
	cur->dmp_avail = prv->dmp_avail;
	cur->dmp = prv->dmp;
	cur->dmp_len = prv->dmp_len;
	cur->dmp_size_order = prv->dmp_size_order;
	cur->tape_log = prv->tape_log;
	cur->tape_log_len = prv->tape_log_len;
	cur->full_trc_head = prv->full_trc_head;
	cur->full_trc_tail = prv->full_trc_tail;
	cur->encryption_capable = prv->encryption_capable;
	cur->encryption_state = prv->encryption_state;
	cur->application_set_encryption = prv->application_set_encryption;
	cur->encryption_proxy_entitled = prv->encryption_proxy_entitled;
	cur->sys_encryption_proxy = prv->sys_encryption_proxy;
	cur->sys_encryption_write = prv->sys_encryption_write;
	cur->encryption_keypath_configured = prv->encryption_keypath_configured;
	cur->server_test = prv->server_test;
	cur->do_encryption = prv->do_encryption;
	cur->do_failover = prv->do_failover;
	cur->wakeup_result = prv->wakeup_result;
	cur->ready_wakeup = prv->ready_wakeup;
	cur->wakeup_info = prv->wakeup_info;
	cur->amu_enable = prv->amu_enable;
	cur->op_sg = prv->op_sg;
	cur->op_sglist = prv->op_sglist;
	WRITE_UNLOCK(&tape_devs[cur->dev_index]->drv_entry_lock);

	return;
} /* copy_parms_to_new_device() */

/*******************************************************************************
*                                                                              *
*  NAME: delete_parms_on_device                                                *
*                                                                              *
*  PURPOSE: Reset tape descriptor to default values (mainly 0)                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev_index: index of tape descriptor                                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void delete_parms_on_device(int dev_index)
{
	struct tape_descriptor* drv = tape_devs[dev_index]->drv;

	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->busy = 0;
	drv->direct_io = False;
	drv->cdb = NULL;
	drv->cdb_len = 0;
	drv->buffer = NULL;
	drv->buf_len = 0;
	drv->rw_buf = NULL;
	drv->rw_buf_order = -1;
	drv->rw_buf_size = 0;
	drv->eom = 0;
	drv->eom_blk_id = 0;
	drv->eod = 0;
	drv->last_op = NO_COMMAND;
	drv->last_tape_motion = NO_COMMAND;
	drv->result = 0;
	drv->opened = 0;
	drv->cmd_in_open = 0;
	drv->blk_size = 0;
	drv->blk_num = 0;
	drv->buffered_mode = True;
	drv->capacity_scaling = 0;
	drv->capacity_scaling_hex_val = 0;
	drv->compression = True;
	drv->density = 0;
	drv->ending_fm = 0;
	drv->logic_write_protect = False;
	drv->medium_type = 0;
	drv->write_protect = 0;
	drv->set_density_valid = 0;
	drv->set_default_density = 0;
	drv->set_pending_density = 0;
	drv->curr_pos = 0;
	drv->pos_valid = 0;
	drv->access_mode = ReadAndWrite;
	drv->acf_mode = 0;
	drv->autoload = 0;
	drv->autoload_mode = 0;
	drv->read_mode = AIXMode;
	drv->sili = 0;
	drv->record_space_mode = 0;
	drv->retain_reservation = 0;
	drv->rew_immed = 0;
	drv->special_file_type = RewindDevice;
	drv->trailer_labels = True;
	drv->wfm_immed = 0;
	drv->limit_read_recov = 0;
	drv->limit_write_recov = 0;
	drv->curr_reserve_type = LinTapeNoReserve;
	drv->explicit_reserved = 0;
	drv->deferred_eom = 0;
	drv->fm_detected = 0;
	drv->ili = 0;
	drv->resid_valid = 0;
	drv->resid = 0;
	drv->sense_data = NULL;
	drv->sense_valid = 0;
	drv->in_erp = 0;
	drv->wfm_on_close_safe = 0;
	drv->cmd_wo_ua_erp = 0;
	drv->last_errno = 0;
	drv->trace = True;
	drv->disable_simmim_log = 0;
	drv->auto_logsense = 0;
	drv->trc_hook = 0;
	drv->simmim_avail = 0;
	drv->simmim = 0;
	drv->simmim_len = 0;
	drv->full_logsense_avail = 0;
	drv->logsense = 0;
	drv->logsense_len = 0;
	drv->dmp_avail = 0;
	drv->dmp = 0;
	drv->dmp_len = 0;
	drv->dmp_size_order = 0;
	drv->tape_log = 0;
	drv->tape_log_len = 0;
	drv->full_trc_head = NULL;
	drv->full_trc_tail = NULL;
	drv->encryption_capable = 0;
	drv->application_set_encryption = 0;
	drv->encryption_state = STATE_UNKNOWN;
	drv->encryption_keypath_configured = False;
	drv->server_test = 0;
	drv->do_encryption = False;
	drv->do_failover = 0;
	drv->ready_wakeup = False;
	drv->wakeup_result = 0;
	drv->amu_enable = 0xFF;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	return;
} /* delete_parms_on_device() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_update_descriptor_on_erp                                         *
*                                                                              *
*  PURPOSE: Copy information from old tape descriptor to new tape descriptor   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    old_dev_index: index of previous tape descriptor                          *
*    new_dev_index: index of current tape descriptor                           *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void tape_update_descriptor_on_erp(int old_dev_index, int new_dev_index)
{
	struct tape_descriptor* prv = tape_devs[old_dev_index]->drv,
		* cur = tape_devs[new_dev_index]->drv;

	WRITE_LOCK(&tape_devs[cur->dev_index]->drv_entry_lock);
	cur->blk_size = prv->blk_size;
	cur->blk_num = prv->blk_num;
	cur->buffered_mode = prv->buffered_mode;
	cur->capacity_scaling = prv->capacity_scaling;
	cur->capacity_scaling_hex_val = prv->capacity_scaling_hex_val;
	cur->compression = prv->compression;
	cur->density = prv->density;
	cur->ending_fm = prv->ending_fm;
	cur->logic_write_protect = prv->logic_write_protect;
	cur->medium_type = prv->medium_type;
	cur->write_protect = prv->write_protect;
	cur->set_density_valid = prv->set_density_valid;
	cur->set_default_density = prv->set_default_density;
	cur->set_pending_density = prv->set_pending_density;
	cur->max_blk_size = prv->max_blk_size;
	cur->max_xfer_len = prv->max_xfer_len;
	cur->min_blk_size = prv->min_blk_size;
	cur->read_mode = prv->read_mode;
	cur->sili = prv->sili;
	cur->encryption_capable = prv->encryption_capable;
	cur->encryption_state = prv->encryption_state;
	cur->application_set_encryption = prv->application_set_encryption;
	cur->encryption_proxy_entitled = prv->encryption_proxy_entitled;
	cur->sys_encryption_proxy = prv->sys_encryption_proxy;
	cur->sys_encryption_write = prv->sys_encryption_write;
	cur->encryption_keypath_configured = prv->encryption_keypath_configured;
	cur->server_test = prv->server_test;
	cur->do_encryption = prv->do_encryption;
    cur->prevent_removal = prv->prevent_removal;
	cur->lbp_set = prv->lbp_set;
	memcpy(&cur->lbp,&prv->lbp,sizeof(struct logical_block_protection));
	cur->amu_enable = prv->amu_enable;
	WRITE_UNLOCK(&tape_devs[cur->dev_index]->drv_entry_lock);

	return;
} /* tape_update_descriptor_on_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: encryption_diag                                                       *
*                                                                              *
*  PURPOSE: Run encryption diagnostics                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    diag: in/out structure containing type and result of diagnostics          *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int encryption_diag(struct tape_descriptor **drvp,
	struct encryption_diagnostics* diag)
{
	struct tape_descriptor *drv = *drvp;
	unchar* diag_data = NULL;
	int rc = -EIO, action_code = 0, return_code = 0, i = 0;

	DbgPrint(("lin_tape: encryption_diag test %d\n", diag->diag_type));

	if(diag->diag_type == SERVER_PING_DIAG) {
		drv->server_test = diag->diag_type;
		rc = tape_perform_encryption_generic(&drv, 60000,
			&action_code, &return_code);
		DbgPrint(("lin_tape: action_code=%d, return_code=%d\n",
			action_code, return_code));

		if(rc == STATUS_SUCCESS) {
			if(action_code == DD_RETRY_COMMAND)
				diag->diag_cc = return_code;
			else diag->diag_errno = return_code;
		} /* if */
		goto EXIT_LABEL;
	} else if(diag->diag_type != BASIC_ENCRYPTION_DIAG &&
		diag->diag_type != FULL_ENCRYPTION_DIAG) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* else if */

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	diag_data = kmalloc(10, GFP_KERNEL);
	if(!diag_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < 2; i++) {
		memset(diag_data, '\0', 10);
		diag_data[0] = 0x80;   /* page code */
		diag_data[3] = 0x06;   /* length */
		diag_data[4] = 0xE0;   /* diag id msb */
		diag_data[5] = 0x00;   /* diag id lsb */
		diag_data[6] = 0x01;   /* cartridge required */
		diag_data[8] = 0x20;
		diag_data[9] = !i ? 0x0F :
			(diag->diag_type == BASIC_ENCRYPTION_DIAG) ? 1 : 0;

		memset(drv->cdb, '\0', MAX_CDB_SIZE);
		drv->cdb->cdb_send_diagnostic.op_code = SEND_DIAGNOSTIC;
		drv->cdb->cdb_send_diagnostic.pf = True;
		drv->cdb->cdb_send_diagnostic.parm_list_length[0] = 0x00;
		drv->cdb->cdb_send_diagnostic.parm_list_length[1] = 10;
		drv->cdb_len = CDB_GENERIC_6;
		drv->last_op = SEND_DIAGNOSTIC;
		drv->buffer = diag_data;
		drv->buf_len = 10;
		rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
			SendDiagnosticTimeOut), DMA_TO_DEVICE, 0, False);

		if(!i) { /* first time through the loop */
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
			drv->server_test = diag->diag_type;
			continue;
		} /* if */

		/* second and final time through the loop */
		if(rc == STATUS_SUCCESS) diag->diag_cc = drv->wakeup_result;
		else diag->diag_errno = drv->encryption_return_code;
		drv->server_test = 0;
	} /* for */

EXIT_LABEL:

	if(diag_data) kfree(diag_data);
	drv->server_test = 0;
	*drvp = drv;

	DbgPrint(("lin_tape: encryption_diag exit: %d\n", rc));
	return rc;
} /* encryption_diag() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_perform_encryption_generic                                       *
*                                                                              *
*  PURPOSE: Post encryption request to daemon and wait for result              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    timeout: seconds between command issue and completion                     *
*    action_code: based on result from daemon                                  *
*    return_code: based on result from daemon                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_perform_encryption_generic(struct tape_descriptor** drvp,
	int timeout, int* action_code, int* return_code)
{
	struct tape_descriptor* drv = *drvp;
	char* trc = NULL;
	int rc = -EIO;
	unchar ttype = lin_tape_trace_type, srv_test = 0;

	DbgPrint(("lin_tape: tape_perform_encryption_generic entry\n"));

	drv->do_encryption = False;

	/* if daemon is not running, return failure */
	READ_LOCK(&lin_tape_trace_lock);
	if(ttype == LIN_TAPE_INVALID_TRACE){
		printk("lin_tape: lin_taped daemon is not running, "\
			"IBMtape%d cannot perform encryption tasks.\n",
			drv->dev_index);
		READ_UNLOCK(&lin_tape_trace_lock);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&lin_tape_trace_lock);
	srv_test = (unchar)drv->server_test;
	rc = tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_ENCRYPT,
		&srv_test, sizeof(unchar));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	DbgPrint(("lin_tape: IBMtape%d is to wait.\n", drv->dev_index));
	rc = wait_event_interruptible_timeout(drv->wait_queue,
		!drv || !drvp || drv->ready_wakeup, timeout);
	if(!drv || !drvp) {
		printk("Drive removed while waiting\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if(rc < 0) {
		printk("lin_tape: IBMtape%d encryption task failed due to "\
			"unexpected interrupt. rc = %d\n", drv->dev_index, rc);
		rc = -EIO;
		goto EXIT_LABEL;
	} else if(rc == 0) {
		printk("lin_tape: IBMtape%d encryption task failed due to "\
			"timeout.\n", drv->dev_index);
		rc = -EIO;
		goto EXIT_LABEL;
	} else {
		DbgPrint(("lin_tape: IBMtape%d encryption task completed\n",
			drv->dev_index));

		WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		drv->ready_wakeup = False;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		*action_code = drv->wakeup_result;
		*return_code = drv->encryption_return_code;
	} /* else */

	if(drv->wakeup_result < STATUS_SUCCESS) {
		printk("lin_tape: IBMtape%d failover failed during "\
			"encryption.\n", drv->dev_index);
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(drv->dev_index != drv->cur_path) {
		DbgPrint(("lin_tape: Encryption failover: new path: "\
			"IBMtape%d\n", drv->cur_path));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg3(drv->tape_log, &drv->tape_log_len,
				TRC_TP_DPF_SUCCESS, drv->dev_index,
				drv->cur_path);
		else if(ttype == LIN_TAPE_FULL_TRACE) {
			trc = kmalloc(4, GFP_KERNEL);
			if(trc) {
				trc[0] = drv->dev_index >> 8;
				trc[1] = drv->dev_index & 0xFF;
				trc[2] = drv->cur_path >> 8;
				trc[3] = drv->cur_path & 0xFF;
				tape_log_full_trace(drv, TRC_TP_DPF_SUCCESS,
					4, trc);
				kfree(trc);
			} /* if */
		} /* else if */

		tape_update_to_new_current_path(&drv);

	} /* if */

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	*drvp = drv;
	DbgPrint(("lin_tape: tape_perform_encryption_generic exit: %d\n", rc));
	return rc;
} /* tape_perform_encryption_generic() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_encryption_write_setting                                         *
*                                                                              *
*  PURPOSE: Set encryption write setting to on or off                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_encryption_write_setting(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	union cdb* tmp_cdb = NULL;
	unchar* tcdb = NULL, * mode_buffer = NULL, * sense_data = NULL;
	int rc = -EIO, mode_select_needed = False, mode_buffer_size = 0,
		result = 0, resid;

	DbgPrint(("lin_tape: tape_encryption_write_setting entry\n"));

	if(!drv) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->encryption_capable ||
		(drv->sys_encryption_write != STATE_ON &&
		drv->sys_encryption_write != STATE_OFF)) {
		rc = STATUS_SUCCESS;
		goto EXIT_LABEL;
	} /* if */

	/* mode sense to get the current settings */
	tmp_cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tmp_cdb) {
		drv->encryption_state = STATE_UNKNOWN;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	mode_buffer_size = sizeof(struct mode_parm_header_10) +
		sizeof(struct rw_control_page);

	mode_buffer = kmalloc(mode_buffer_size, GFP_KERNEL);
	if(!mode_buffer) {
		drv->encryption_state = STATE_UNKNOWN;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data) {
		drv->encryption_state = STATE_UNKNOWN;
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_data, '\0', SCSI_SENSE_BUFFERSIZE);
	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	memset(mode_buffer, '\0', mode_buffer_size);

	tmp_cdb->cdb_modesense_10.op_code = MODE_SENSE_10;
	tmp_cdb->cdb_modesense_10.dbd = True;
	tmp_cdb->cdb_modesense_10.page_code = RWControlPageNumber;
	tmp_cdb->cdb_modesense_10.alloc_length[0] = 0;
	tmp_cdb->cdb_modesense_10.alloc_length[1] = mode_buffer_size;

	tcdb = (unchar*)tmp_cdb;

	rc = -1;
	while(rc != STATUS_SUCCESS && rc != UNIT_ATTENTION) {
		rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)mode_buffer,
			mode_buffer_size, compute_timeout(drv->drv_id,
			ModeSenseTimeOut), DMA_FROM_DEVICE, &result,
			sense_data, &resid);
	} /* while */

	if(STATUS_SUCCESS != rc) {
		drv->encryption_state = STATE_UNKNOWN;
		goto EXIT_LABEL;
	} /* if */

	/* issue mode select command, if necessary */
	mode_select_needed = False;
	mode_buffer[26] |= 0x01;

	if(drv->sys_encryption_write == STATE_ON) {
		if((mode_buffer[28] & 0x03) != 0x01) {
			mode_buffer[28] = (unchar)((mode_buffer[28] & 0xFC)
				| 0x01);
			mode_select_needed = True;
		} /* if */

		if(drv->drv_id == IBM_3592) {
			if((mode_buffer[30] & 0x09) != 0x08) {
				mode_buffer[30] = mode_buffer[30] | 0x08;
				mode_buffer[30] = mode_buffer[30] & 0xFE;
				mode_select_needed = True;
			} /* if */

			if(!(mode_buffer[30] & 0x02)) {
				mode_buffer[30] = mode_buffer[30] | 0x02;
				mode_select_needed = True;
			} /* if */
		} else if((mode_buffer[30] & 0xFE) != 0xFE) {
			mode_buffer[30] = mode_buffer[30] & 0xFE;
			mode_select_needed = True;
		} /* else */
	} else if((mode_buffer[28] & 0x03) != 0) {
		mode_buffer[28] = (unchar)(mode_buffer[28] & 0xFC);
		mode_select_needed = True;
	} /* else if */

	mode_buffer[0] = 0;
	mode_buffer[1] = 0;

	rc = STATUS_SUCCESS;

	if(!mode_select_needed) {
		drv->encryption_state = drv->sys_encryption_write;
		goto EXIT_LABEL;
	} /* if */

	memset(tmp_cdb, '\0', MAX_CDB_SIZE);
	tmp_cdb->cdb_modeselect_10.op_code = MODE_SELECT_10;
	tmp_cdb->cdb_modeselect_10.pf = True;
	tmp_cdb->cdb_modeselect_10.parm_list_length[0] = 0;
	tmp_cdb->cdb_modeselect_10.parm_list_length[1] = mode_buffer_size;
	tcdb = (unchar*)tmp_cdb;

	rc = tape_exe_scsi_cmd(drv, tcdb, 10, (void*)mode_buffer,
		mode_buffer_size, compute_timeout(drv->drv_id,
		ModeSelectTimeOut), DMA_TO_DEVICE, &result,
		sense_data, &resid);

	if(rc == STATUS_SUCCESS)
		drv->encryption_state = drv->sys_encryption_write;
	else {
		printk("lin_tape: encryption write setting failed, %d\n", rc);
		drv->encryption_state = STATE_UNKNOWN;
	}  /* else */

EXIT_LABEL:

	if(sense_data) kfree(sense_data);
	if(mode_buffer) kfree(mode_buffer);
	if(tmp_cdb) kfree(tmp_cdb);
	DbgPrint(("lin_tape: tape_encryption_write_setting exit: %d\n", rc));
	return rc;
} /* tape_encryption_write_setting() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_query_amu                                                    *
*                                                                              *
*  PURPOSE: retrieve archive mode unthread                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive for which we want to query amu         *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_query_amu(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct feature_switches_subpage* page = NULL;
	struct archive_mode_unthread amu;

	DbgPrint(("lin_tape_query_amu entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct archive_mode_unthread)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	memset(&amu, '\0', sizeof(struct archive_mode_unthread));

	page = kmalloc(sizeof(struct feature_switches_subpage), GFP_KERNEL);
	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', sizeof(struct feature_switches_subpage));

	rc = tape_modesense10_page(&drv, DEVICE_ATTRIBUTES_PAGE,
		(unchar*)page, sizeof(struct feature_switches_subpage), FEATURE_SWITCHES, True, False);

	if(rc != STATUS_SUCCESS) {
		DbgPrint(("tape_modesense_page failed: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	amu.amu_enable = page->earchive;
	if(copy_to_user((void*)arg, (void*)&amu,
			sizeof(struct archive_mode_unthread))) {
			rc = -EFAULT;
	} /* if */

EXIT_LABEL:

	if(page) kfree(page);

	DbgPrint(("lin_tape_query_amu exit: %d\n", rc));
	return rc;
} /* lin_tape_query_amu() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_amu                                                      *
*                                                                              *
*  PURPOSE: set archive mode unthread                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for the drive for which we want to query amu         *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_set_amu(struct tape_descriptor* drv, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, alloc_len;
	unchar* page = NULL;
	struct archive_mode_unthread amu;

	DbgPrint(("lin_tape_set_amu entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct archive_mode_unthread)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	alloc_len =  sizeof(struct feature_switches_subpage) +
                 sizeof(struct mode_parm_header_10);
	page = kmalloc(alloc_len, GFP_KERNEL);

	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', alloc_len);

	rc = tape_modesense10_page(&drv, DEVICE_ATTRIBUTES_PAGE,
		(unchar*)page, alloc_len, FEATURE_SWITCHES, True, True);
	if(rc != STATUS_SUCCESS) {
		DbgPrint(("tape_modesense_page failed: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

    if (copy_from_user((void*)&amu,(void*)arg,
         sizeof(struct archive_mode_unthread))) {
        printk("lin_tape_set_amu: cannot copy from user\n");
        rc = -EFAULT;
        goto EXIT_LABEL;
    }

	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->earchive = amu.amu_enable;
	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->spf = 1;
	((struct feature_switches_subpage *)(page + sizeof(struct mode_parm_header_10)))->ps = 1;

	rc = tape_modeselect10(&drv, (unchar*)page, 1, alloc_len);
	if (rc != STATUS_SUCCESS){
        DbgPrint(("tape_modeselect_page failed: %d\n", rc));
        goto EXIT_LABEL;
	}

EXIT_LABEL:

	if(page) kfree(page);

	DbgPrint(("lin_tape_set_amu exit: %d\n", rc));
	return rc;
} /* lin_tape_set_amu() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_load_unload                                                  *
*                                                                              *
*  PURPOSE: Load or unload cartridge from drive with hold, reten and immed     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*    command: the IOCTL code                                                   *
*    arg: the data for the request                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_load_unload(struct tape_descriptor **drvp, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor *drv = *drvp;
	struct tape_load_unload loadunload;

	DbgPrint(("lin_tape_load_unload entry\n"));

	if(_IOC_SIZE(command) != sizeof(struct tape_load_unload)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

    if (copy_from_user((void*)&loadunload,(void*)arg,
         sizeof(struct tape_load_unload))) {
        printk("lin_tape_load_unload: cannot copy from user\n");
        rc = -EFAULT;
        goto EXIT_LABEL;
    }

	if(!drv->cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->cdb, '\0', MAX_CDB_SIZE);
	drv->cdb->cdb_load_unload.op_code = LOAD_UNLOAD;
	drv->cdb->cdb_load_unload.immediate = loadunload.immediate;
	drv->cdb->cdb_load_unload.load = loadunload.load;
	drv->cdb->cdb_load_unload.hold = loadunload.hold;
	drv->cdb->cdb_load_unload.eot = loadunload.eot;
	drv->cdb->cdb_load_unload.reten = loadunload.retention;
	drv->cdb_len = CDB_GENERIC_6;
	drv->last_op = LOAD_UNLOAD;
	drv->buffer = NULL;
	drv->buf_len = 0;
	drv->pos_valid = False;

	rc = tape_send_scsi_cmd(&drv, compute_timeout(drv->drv_id,
		loadunload.load ?  LoadTimeOut : UnloadTimeOut), DMA_NONE, 0, False);

	if(drv && drv->sense_valid && drv->sense_data->sense_key == NOT_READY && !loadunload.load)
		rc = STATUS_SUCCESS;

	if(rc == STATUS_SUCCESS) {
		drv->eom = False;
		drv->eom_blk_id = 0;
		drv->deferred_eom = False;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* lin_tape_load_unload() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_get_timeouts                                                     *
*                                                                              *
*  PURPOSE: Retrieve auto timeout values                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t tape_get_timeouts(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS, result=0, resid;
#define DATA_LEN 80*20+4
    unchar data[DATA_LEN];
    uint offset=4, i, data_len, rec_timeout;
    unchar *descp, *sense_data, *cdb;
	union cdb* tcdb = NULL;

	DbgPrint(("lin_tape: tape_get_timeout\n"));

	tcdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
	if(!tcdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tcdb, '\0', MAX_CDB_SIZE);

	sense_data = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_data){
		printk("lin_tape: failed to allocate sense_data\n");
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset((void*)sense_data, '\0', SCSI_SENSE_BUFFERSIZE);
	memset(drv->timeout, 0, sizeof(drv->timeout));
	cdb = (unchar*)tcdb;
	tcdb->cdb_report_soc.op_code = 0xA3;
	tcdb->cdb_report_soc.service_action = 0x0C;
	tcdb->cdb_report_soc.rctd=1;
    tcdb->cdb_report_soc.reporting_opt=0;
    tcdb->cdb_report_soc.allocation_length[2]=(unchar)((DATA_LEN) >> 8)&0xFF;
    tcdb->cdb_report_soc.allocation_length[3]=(unchar)(DATA_LEN)&0xFF;
	for (i=0;i<3;i++){
        	rc = tape_exe_scsi_cmd(drv, cdb, CDB_GENERIC_12, (void*)data, DATA_LEN,
		60, DMA_FROM_DEVICE,&result, sense_data, &resid);
        if(rc == STATUS_SUCCESS) break;
	}
	if (rc!=STATUS_SUCCESS){
		printk("lin_tape: %s unable to get auto timeouts\n",
			drv->devid);
        drv->autotimeout=False;
		goto EXIT_LABEL;
	} /* if */
	drv->autotimeout=True;
	descp = data+offset;
    data_len=data[0]<<24|data[1]<<16|data[2]<<8|data[3];
    while (offset<data_len){
        rec_timeout=descp[19]|descp[18]<<8|descp[17]<<16|descp[16]<<24;
        if(drv->timeout[descp[0]==0])
            drv->timeout[descp[0]]=rec_timeout+5;
        else if(drv->timeout[descp[0]]<rec_timeout)
            drv->timeout[descp[0]]=rec_timeout+5;
        offset+=20;
        descp+=20;
    }

EXIT_LABEL:

    if (tcdb) kfree(tcdb);
	drv->buf_len = 0;
	drv->buffer = NULL;

	return rc;
} /* tape_get_timeouts() */
