/*******************************************************************************
 *   lin_tape_scsi_config.c
 *
 *  (C) Copyright International Business Machines  Corp., 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ******************************************************************************/

#include "lin_tape_tape.h"
#include "lin_tape_scsi_config.h"
#include "lin_tape_support_list.h"
#include "lin_tape_scsi_ioctl.h"
#include "convert.h"

/*******************************************************************************
*                                                                              *
*                               DRIVER ATTRIBUTES                              *
*                                                                              *
*******************************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: tape_version_show                                                     *
*                                                                              *
*  PURPOSE: Display lin_tape version                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t version_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, strlen(LIN_TAPE_VERSION) + 1, "%s\n",
		LIN_TAPE_VERSION);

	return rc;
} /* tape_version_show() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RO(version);
#else
DRIVER_ATTR(version, 0444, version_show, NULL);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_debug_show                                                   *
*                                                                              *
*  PURPOSE: Display debug driver parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_debug_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", lin_tape_debug);

	return rc;

} /* lin_tape_debug_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_debug_store                                                  *
*                                                                              *
*  PURPOSE: Write to debug driver parameter                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_debug_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		lin_tape_debug = 0;
	else
		lin_tape_debug = 1;
	if(buf[0] == '3')
		lin_tape_debug = 3;

	return count;

} /* lin_tape_debug_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(lin_tape_debug);
#else
DRIVER_ATTR(lin_tape_debug, 0660, lin_tape_debug_show, lin_tape_debug_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_clean_ghosts_show                                            *
*                                                                              *
*  PURPOSE: Display clean ghosts driver parameter                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_clean_ghosts_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", lin_tape_clean_ghosts);

	return rc;

} /* lin_tape_clean_ghosts_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_clean_ghosts_store                                           *
*                                                                              *
*  PURPOSE: Write to clean ghosts driver parameter                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_clean_ghosts_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '1'){
		lin_tape_clean_ghosts = 1;
	}
    else{
        lin_tape_clean_ghosts = 0;
    }
	return count;

} /* lin_tape_clean_ghosts_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(lin_tape_clean_ghosts);
#else
DRIVER_ATTR(lin_tape_clean_ghosts, 0660, lin_tape_clean_ghosts_show, lin_tape_clean_ghosts_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: auto_logging_show                                                     *
*                                                                              *
*  PURPOSE: Display auto logging parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t auto_logging_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", auto_logging);

	return rc;

} /* auto_logging_show() */

/*******************************************************************************
*                                                                              *
*  NAME: auto_logging_store                                                    *
*                                                                              *
*  PURPOSE: Write to auto logging parameter                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t auto_logging_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		auto_logging = 0;
	else
		auto_logging = 1;

	return count;

} /* auto_logging_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(auto_logging);
#else
DRIVER_ATTR(auto_logging, 0660, auto_logging_show, auto_logging_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: petro_show                                                            *
*                                                                              *
*  PURPOSE: Display petro driver parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t petro_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", petro);

	return rc;
} /* petro_show() */

/*******************************************************************************
*                                                                              *
*  NAME: petro_store                                                           *
*                                                                              *
*  PURPOSE: Write to petro driver parameter                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t petro_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		petro = 0;
	else
		petro = 1;

	return count;

} /* petro_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(petro);
#else
DRIVER_ATTR(petro, 0660, petro_show, petro_store);
#endif
/*******************************************************************************
*                                                                              *
*  NAME: support_dio_show                                                      *
*                                                                              *
*  PURPOSE: Display support_dio driver parameter                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t support_dio_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", support_dio);

	return rc;

} /* support_dio_show() */

/*******************************************************************************
*                                                                              *
*  NAME: support_dio_store                                                     *
*                                                                              *
*  PURPOSE: Write to support_dio driver parameter                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t support_dio_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		support_dio = 0;
	else
		support_dio = 1;

	return count;

} /* support_dio_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(support_dio);
#else
DRIVER_ATTR(support_dio, 0660, support_dio_show, support_dio_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: wait_on_config_show                                                   *
*                                                                              *
*  PURPOSE: Display wait_on_config driver parameter                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t wait_on_config_show(struct device_driver* driver,
					char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", wait_on_config);

	return rc;
} /* wait_on_config_show() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RO(wait_on_config);
#else
DRIVER_ATTR(wait_on_config, 0444, wait_on_config_show, NULL);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: changer_open_reserve_show                                             *
*                                                                              *
*  PURPOSE: Display changer_open_reserve driver parameter                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t changer_open_reserve_show(struct device_driver* driver,
					char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", changer_open_reserve);

	return rc;
} /* changer_open_reserve_show() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_open_reserve_store                                            *
*                                                                              *
*  PURPOSE: Write to changer_open_reserve driver parameter                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t changer_open_reserve_store(struct device_driver* driver,
					const char* buf, size_t count)
{
	if(buf[0] == '0')
		changer_open_reserve = 0;
	else
		changer_open_reserve = 1;

	return count;
} /* changer_open_reserve_show() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(changer_open_reserve);
#else
DRIVER_ATTR(changer_open_reserve, 0660, changer_open_reserve_show,
	changer_open_reserve_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: default_sys_encryption_proxy_show                                     *
*                                                                              *
*  PURPOSE: Display default_sys_encryption_proxy driver parameter              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t default_sys_encryption_proxy_show(struct device_driver* driver,
								char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", default_sys_encryption_proxy);

	return rc;
} /* default_sys_encryption_proxy_show() */

/*******************************************************************************
*                                                                              *
*  NAME: default_sys_encryption_proxy_store                                    *
*                                                                              *
*  PURPOSE: Write to default_sys_encryption_proxy driver parameter             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t default_sys_encryption_proxy_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		default_sys_encryption_proxy = 0;
	else
		default_sys_encryption_proxy = 1;

	return count;
} /* default_sys_encryption_proxy_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(default_sys_encryption_proxy);
#else
DRIVER_ATTR(default_sys_encryption_proxy, 0644,
		default_sys_encryption_proxy_show,
		default_sys_encryption_proxy_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: default_sys_encryption_write_show                                     *
*                                                                              *
*  PURPOSE: Display default_sys_encryption_write driver parameter              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t default_sys_encryption_write_show(struct device_driver* driver,
								char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", default_sys_encryption_write);

	return rc;
} /* default_sys_encryption_write_show() */

/*******************************************************************************
*                                                                              *
*  NAME: default_sys_encryption_write_store                                    *
*                                                                              *
*  PURPOSE: Write to default_sys_encryption_write driver parameter             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t default_sys_encryption_write_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		default_sys_encryption_write = 0;
	else if(buf[0] == '1')
		default_sys_encryption_write = 1;
	else
		default_sys_encryption_write = 2;

	return count;
} /* default_sys_encryption_write_show() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(default_sys_encryption_write);
#else
DRIVER_ATTR(default_sys_encryption_write, 0644,
		default_sys_encryption_write_show,
		default_sys_encryption_write_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: disable_density_on_open_show                                          *
*                                                                              *
*  PURPOSE: Display disable density parameter                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t disable_density_on_open_show(struct device_driver* driver,
	char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", disable_density_on_open);

	return rc;

} /* disable_density_on_open_show() */

/*******************************************************************************
*                                                                              *
*  NAME: disable_density_on_open_store                                         *
*                                                                              *
*  PURPOSE: Write to disable density parameter                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t disable_density_on_open_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		disable_density_on_open = 0;
	else
		disable_density_on_open = 1;

	return count;

} /* disable_density_on_open_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(disable_density_on_open);
#else
DRIVER_ATTR(disable_density_on_open, 0660, disable_density_on_open_show,
	disable_density_on_open_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: dynamic_attributes_show                                               *
*                                                                              *
*  PURPOSE: Display debug driver parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t dynamic_attributes_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", dynamic_attributes);

	return rc;

} /* dynamic_attributes_show() */

/*******************************************************************************
*                                                                              *
*  NAME: dynamic_attributes_store                                              *
*                                                                              *
*  PURPOSE: Write to debug driver parameter                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t dynamic_attributes_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		dynamic_attributes = 0;
	else
		dynamic_attributes = 1;

	return count;

} /* dynamic_attributes_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(dynamic_attributes);
#else
DRIVER_ATTR(dynamic_attributes, 0660, dynamic_attributes_show,
	dynamic_attributes_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME:  disable_auto_drive_dump_show                                         *
*                                                                              *
*  PURPOSE: Disable Autodump                                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t disable_auto_drive_dump_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", disable_auto_drive_dump);

	return rc;

} /* disable_auto_drive_dump_show() */

/*******************************************************************************
*                                                                              *
*  NAME: disable_auto_drive_dump_store                                         *
*                                                                              *
*  PURPOSE: Disable Autodump                                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t disable_auto_drive_dump_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		disable_auto_drive_dump = 0;
	else
		disable_auto_drive_dump = 1;

	return count;

} /* disable_auto_drive_dump_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(disable_auto_drive_dump);
#else
DRIVER_ATTR(disable_auto_drive_dump, 0660, disable_auto_drive_dump_show, disable_auto_drive_dump_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME:  busy_retry_show                                                      *
*                                                                              *
*  PURPOSE: Retry on buys driver parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data copied into buffer                                  *
*                                                                              *
*******************************************************************************/

static ssize_t busy_retry_show(struct device_driver* driver, char* buf)
{
	int rc;

	rc = snprintf(buf, 10, "%d\n", busy_retry);

	return rc;

} /* busy_retry_show() */

/*******************************************************************************
*                                                                              *
*  NAME: busy_retry_store                                                      *
*                                                                              *
*  PURPOSE: Retry on buys driver parameter                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    driver: unused                                                            *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t busy_retry_store(struct device_driver* driver,
				const char* buf, size_t count)
{
	if(buf[0] == '0')
		busy_retry = 0;
	else
		busy_retry = 1;

	return count;

} /* busy_retry_store() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DRIVER_ATTR_RW(busy_retry);
#else
DRIVER_ATTR(busy_retry, 0660, busy_retry_show, busy_retry_store);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static struct attribute *lin_tape_drv_attrs[] = {
	&driver_attr_version.attr,
	&driver_attr_lin_tape_debug.attr,
	&driver_attr_lin_tape_clean_ghosts.attr,
	&driver_attr_auto_logging.attr,
	&driver_attr_petro.attr,
	&driver_attr_support_dio.attr,
	&driver_attr_wait_on_config.attr,
	&driver_attr_changer_open_reserve.attr,
	&driver_attr_default_sys_encryption_proxy.attr,
	&driver_attr_default_sys_encryption_write.attr,
	&driver_attr_disable_density_on_open.attr,
	&driver_attr_dynamic_attributes.attr,
	&driver_attr_disable_auto_drive_dump.attr,
	&driver_attr_busy_retry.attr,
	NULL,
};

static struct attribute_group lin_tape_groupdrv = {
	.attrs = lin_tape_drv_attrs,
};

static const struct attribute_group *lin_tape_drv_groups[] = {
	&lin_tape_groupdrv,
	NULL,
};
#endif
/*******************************************************************************
*                                                                              *
*                           CLASS DEVICE ATTRIBUTES                            *
*                                                                              *
*******************************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_device_encryption_proxy_show                           *
*                                                                              *
*  PURPOSE: Display encryption_proxy class device parameter                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    device: the device for which the parameter is being shown                 *
*    attr: the device attribute we are printing                                *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t sys_encryption_proxy_show(
		struct device * device, struct device_attribute* attr,
		char* buf)
{
    struct scsi_device *sdev = to_scsi_device(device);
    struct tape_descriptor* tape = dev_get_drvdata(device);

    if (!sdev || sdev->type != TYPE_TAPE || !tape ) return 0;
	if(tape) return sprintf(buf, "%d\n", tape->sys_encryption_proxy);
	return 0;
} /* sys_encryption_proxy_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_device_encryption_proxy_set                            *
*                                                                              *
*  PURPOSE: Set encryption_proxy class device parameter                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    device: the device for which the parameter is being set                   *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t sys_encryption_proxy_store(
	struct device * device, struct device_attribute* attr,
	const char* buf, size_t count)
{
	int i, j;
	int new_setting;
	struct scsi_device *sdev = to_scsi_device(device);
	/* perform validity checking on the input string */

	if(!sdev || sdev->type != TYPE_TAPE || (buf[0] < '0') || (buf[0] > '1'))
		return strlen(buf);

	new_setting = buf[0] - '0';

	WRITE_LOCK(&tape_des_lock);
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(tape_devs[i] == NULL || tape_devs[i]->drv == NULL)
			continue;
		if((tape_devs[i]->drv->rew_class_dev != device) &&
			(tape_devs[i]->drv->nonrew_class_dev != device) &&
			(tape_devs[i]->drv->nonresv_class_dev != device))
			continue;

		/* if got to this point, tape_devs[i] is the correct device. */
		for(j = 0; j < MaxLinTapeDevices; j++) {
			if(tape_devs[j] == NULL || tape_devs[i]->drv == NULL)
				continue;

			if((tape_devs[i]->drv->serial_num_len > 0) &&
				(tape_devs[i]->drv->serial_num_len
					== tape_devs[j]->drv->serial_num_len)
				&& (memcmp(tape_devs[i]->drv->serial_num,
					tape_devs[j]->drv->serial_num,
					tape_devs[i]->drv->serial_num_len) == 0)) {
				/* this setting change refers to this device */
				tape_devs[j]->drv->sys_encryption_proxy
					= new_setting;
			} /* if */
		} /* for */
		break;
	} /* for */

	WRITE_UNLOCK(&tape_des_lock);

	return strlen(buf);
} /* sys_encryption_proxy_set() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DEVICE_ATTR_RW(sys_encryption_proxy);
#else
static DEVICE_ATTR(sys_encryption_proxy,
		0660,
		sys_encryption_proxy_show,
		sys_encryption_proxy_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_device_encryption_write_show                           *
*                                                                              *
*  PURPOSE: Display encryption_write class device parameter                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    device: the device for which the parameter is being shown                 *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t sys_encryption_write_show(
	struct device* device, struct device_attribute* attr, char* buf)
{
    struct scsi_device *sdev = to_scsi_device(device);
	struct tape_descriptor* tape = dev_get_drvdata(device);

    if (!sdev || sdev->type != TYPE_TAPE || !tape) return 0;
	if(tape) return sprintf(buf, "%d\n", tape->sys_encryption_write);
	return 0;
} /* sys_encryption_write_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_device_encryption_write_set                            *
*                                                                              *
*  PURPOSE: Set encryption_write class device parameter                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    device: the device for which the parameter is being set                   *
*    buf: kernel-filled buffer from which to copy data                         *
*    count: length of data in buffer                                           *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t sys_encryption_write_store(
	struct device* device, struct device_attribute* attr,
	const char* buf, size_t count)
{
	int i, j;
	int new_setting;
	/* perform validity checking on the input string */

	if((buf[0] < '0') || (buf[0] > '2'))
		return strlen(buf);

	new_setting = buf[0] - 0x30;

	WRITE_LOCK(&tape_des_lock);
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(tape_devs[i] == NULL || tape_devs[i]->drv == NULL)
			continue;
		if((tape_devs[i]->drv->rew_class_dev != device) &&
			(tape_devs[i]->drv->nonrew_class_dev != device) &&
			(tape_devs[i]->drv->nonresv_class_dev != device))
			continue;

		/* if got to this point, tape_devs[i] is the correct device. */
		for(j = 0; j < MaxLinTapeDevices; j++) {
			if(!tape_devs[j] || tape_devs[i]->drv == NULL)
				continue;

			if((tape_devs[i]->drv->serial_num_len > 0) &&
				(tape_devs[i]->drv->serial_num_len
				== tape_devs[j]->drv->serial_num_len)
				&&(memcmp(tape_devs[i]->drv->serial_num,
                tape_devs[j]->drv->serial_num,
				tape_devs[i]->drv->serial_num_len) == 0)) {

				/* this setting change refers to this device */
				if(tape_devs[j]->drv->sys_encryption_write
					!= new_setting) {
					/* change necessary */
					tape_devs[j]->drv->sys_encryption_write
						= new_setting;
				} /* if */
			} /* if */
		} /* for */
		break;
	} /* for */

	WRITE_UNLOCK(&tape_des_lock);

	return strlen(buf);
} /* lin_tape_class_device_encryption_write_set() */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DEVICE_ATTR_RW(sys_encryption_write);
#else
static DEVICE_ATTR(sys_encryption_write,
		0660,
		sys_encryption_write_show,
		sys_encryption_write_store);
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_sernum_show                                                  *
*                                                                              *
*  PURPOSE: Display serial_num device parameter                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device struct representing the device for which to show sernum   *
*    attr: the serial_num attribute                                            *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t serial_num_show(struct device* dev, struct device_attribute* attr,
	char* buf)
{
	struct tape_descriptor* tape = dev_get_drvdata(dev);
	struct changer_descriptor* chg = dev_get_drvdata(dev);
	if(!tape) return 0;

	if(!memcmp(tape->lname, "IBMtape", strlen("IBMtape"))
		&& tape->serial_num_len) {
		memcpy(buf, tape->serial_num, tape->serial_num_len);
		buf[tape->serial_num_len] = '\n';
		return strlen(buf);
	} else if(!memcmp(tape->lname, "IBMchanger", strlen("IBMchanger"))
		&& chg->serial_num_len) {
		memcpy(buf, chg->serial_num, chg->serial_num_len);
		buf[chg->serial_num_len] = '\n';
		return strlen(buf);
	} /* else if */

	return 0;
} /* serial_num_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ww_node_name_show                                            *
*                                                                              *
*  PURPOSE: Display ww_node_name device parameter                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device struct representing the device for which to show wwnn     *
*    attr: the wwnn attribute                                                  *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t ww_node_name_show(struct device* dev,
	struct device_attribute* attr, char* buf)
{
	int i = 0, valid = False;
	struct tape_descriptor* tape = dev_get_drvdata(dev);
	if(!tape) return 0;

	memcpy(buf, "0x", strlen("0x"));
	for(i = 0; i < WWN_LEN; i++) {
		sprintf(&buf[strlen(buf)], "%02X", tape->wwnn[i] & 0xFF);
		if(tape->wwnn[i]) valid = True;
	} /* for */

	strcat(buf, "\n");
	if(valid) return strlen(buf);

	memset(buf, '\0', 32);
	return 0;
} /* ww_node_name_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ww_port_name_show                                            *
*                                                                              *
*  PURPOSE: Display ww_port_name device parameter                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device struct representing the device for which to show wwpn     *
*    attr: the wwpn attribute                                                  *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t ww_port_name_show(struct device* dev,
	struct device_attribute* attr, char* buf)
{
	int i = 0, valid = False;
	struct tape_descriptor* tape = dev_get_drvdata(dev);
	if(!tape) return 0;

	memcpy(buf, "0x", strlen("0x"));
	for(i = 0; i < WWN_LEN; i++) {
		sprintf(&buf[strlen(buf)], "%02X", tape->wwpn[i] & 0xFF);
		if(tape->wwpn[i]) valid = True;
	} /* for */

	strcat(buf, "\n");
	if(valid) return strlen(buf);

	memset(buf, '\0', 32);
	return 0;
} /* ww_port_name_show() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_primary_path_show                                            *
*                                                                              *
*  PURPOSE: Display primary_path device parameter                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device struct representing the device for which to show primary  *
*    attr: the primary_path attribute                                          *
*    buf: kernel-allocated buffer to receive data                              *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*******************************************************************************/

static ssize_t primary_path_show(struct device* dev,
	struct device_attribute* attr, char* buf)
{
	struct tape_descriptor* tape = dev_get_drvdata(dev);
	struct changer_descriptor* chg = dev_get_drvdata(dev);
	if(!tape) return 0;

	if(!memcmp(tape->lname, "IBMtape", strlen("IBMtape"))) {
		if(tape->alt_pathing) {
			if(tape->checked_multipath) {
				if(tape->isPrimary) strcpy(buf, "Primary\n");
				else strcpy(buf, "Alternate\n");
			} else {
				strcpy(buf, "Unchecked\n");
			} /* else */
		} else {
			strcpy(buf, "NA\n");
		} /* else */

		return strlen(buf);
	} else if(!memcmp(tape->lname, "IBMchanger", strlen("IBMchanger"))) {

		if(chg->alt_pathing) {
			if(chg->checked_multipath) {
				if(chg->isPrimary) strcpy(buf, "Primary\n");
				else strcpy(buf, "Alternate\n");
			} else {
				strcpy(buf, "Unchecked\n");
			} /* else */
		} else {
			strcpy(buf, "NA\n");
		} /* else */

		return strlen(buf);
	} /* else if */

	return 0;
} /* primary_path_show() */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static DEVICE_ATTR(serial_num, S_IRUGO, serial_num_show, NULL);
static DEVICE_ATTR(ww_node_name, S_IRUGO, ww_node_name_show, NULL);
static DEVICE_ATTR(ww_port_name, S_IRUGO, ww_port_name_show, NULL);
static DEVICE_ATTR(primary_path, S_IRUGO, primary_path_show, NULL);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
static DEVICE_ATTR_RO(serial_num);
static DEVICE_ATTR_RO(ww_node_name);
static DEVICE_ATTR_RO(ww_port_name);
static DEVICE_ATTR_RO(primary_path);

static struct attribute *lin_tape_dev_attrs[] = {
	&dev_attr_sys_encryption_proxy.attr,
	&dev_attr_sys_encryption_write.attr,
	&dev_attr_serial_num.attr,
	&dev_attr_ww_node_name.attr,
	&dev_attr_ww_port_name.attr,
	&dev_attr_primary_path.attr,
	NULL,
};

static struct attribute_group lin_tape_group = {
	.attrs = lin_tape_dev_attrs,
};

static const struct attribute_group *lin_tape_dev_groups[] = {
	&lin_tape_group,
	NULL,
};
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_dev_release                                            *
*                                                                              *
*  PURPOSE: Free a class_device structure                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the class_device to free                                             *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_class_dev_release(struct device *dev)
{
	kfree(dev);
	dev = NULL;
} /* lin_tape_class_dev_release() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_class_release                                                *
*                                                                              *
*  PURPOSE: Free a class structure                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    class: the class to free                                                  *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_class_release(struct class *class)
{
	kfree(class);
	class = NULL;
} /* lin_tape_class_release() */

int lin_tape_sdev_get(struct scsi_device* sdev, int* ct)
{
	int rc = -ENXIO;

	if(sdev && ct) {
		rc = scsi_device_get(sdev);
		if(rc == STATUS_SUCCESS) (*ct)++;
        if (!scsi_block_when_processing_errors(sdev)) {
                printk("get dev scsi block error\n");
            lin_tape_sdev_put(sdev, ct);
            rc = -EIO;
        }
	} /* if */
	if (*ct > 1) printk ("lin_tape_sdev_get ct %d rc %d\n", *ct, rc);
	return rc;
} /* lin_tape_sdev_get() */

void lin_tape_sdev_put(struct scsi_device* sdev, int* ct)
{
	if(!sdev || !ct || (*ct) <= 0) return;
	scsi_device_put(sdev);
	(*ct)--;
	return;
} /* lin_tape_sdev_put() */

struct dyn_attrs dyn_attrs[] = {
	{DRA_DSFN, DRA_DSFN_LEN, "DRA_DSFN"},
	{DRA_DPF_ACTIVE_PATH, DRA_ACTIVE_PATH_LEN, "DRA_DPF_ACTIVE_PATH"},
	{DRA_HOST_NAME, DRA_HOST_NAME_LEN, "DRA_HOST_NAME"},
	{DRA_OS, DRA_OS_LEN, "DRA_OS"},
	{DRA_OS_VER, DRA_OS_VER_LEN, "DRA_OS_VER"},
	{DRA_DD_NAME, DRA_DD_NAME_LEN, "DRA_DD_NAME"},
	{DRA_DD_VER, DRA_DD_VER_LEN, "DRA_DD_VER"},
	{DRA_PID, DRA_PID_LEN, "DRA_PID"},
	{0, 0, NULL}
};

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_init                                                         *
*                                                                              *
*  PURPOSE: Initializes data associated with a tape device                     *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: length of data in buffer                                           *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    1. create and initialize a struct class (from the Linux Device Model)     *
*    2. register kernel module                                                 *
*    3. register the driver to the SCSI subsystem and add sysfs driver entries *
*    4. create class interface to SCSI subsystem                               *
*    5. initialize global variables                                            *
*                                                                              *
*******************************************************************************/

static int lin_tape_init()
{
	int rc = STATUS_SUCCESS;
	dev_t *first_dev = NULL;
	int lin_tape_class_allocd = False, chrdev_regd = False,
		driver_regd = False, interface_regd = False;

	DbgPrint(("lin_tape: lin_tape_init entry.\n"));

	rc = initialize_lin_tape_class();
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	lin_tape_class_allocd = True;

	rc = alloc_and_register(&first_dev, &chrdev_regd, &driver_regd,
		&interface_regd);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	create_driver_attributes();
#endif
	lin_tape_create_proc();

	printk("lin_tape %s is loaded.\n", LIN_TAPE_VERSION);

	if(support_dio == 1)
		printk("lin_tape: support_dio is enabled.\n");

	if(petro == 1)
		printk("lin_tape: petro is enabled.\n");

EXIT_LABEL:

	if(rc != STATUS_SUCCESS) {
		lin_tape_undo_init(interface_regd, chrdev_regd, driver_regd,
			lin_tape_class_allocd, first_dev);
	} /* if */

	if(first_dev) kfree(first_dev);

	DbgPrint(("lin_tape: lin_tape_init exit: %d\n", rc));
	return rc;
}  /* lin_tape_init() */

/*******************************************************************************
*                                                                              *
*  NAME: initialize_lin_tape_class                                             *
*                                                                              *
*  PURPOSE: Set up the lin_tape_class Linux Device Model class for sysfs       *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int initialize_lin_tape_class()
{
	int rc = 0;

	lin_tape_class = class_create(THIS_MODULE, "lin_tape");
	if(!lin_tape_class) {
		printk("lin_tape: cannot create scsi lin_tape class.\n");
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;

} /* initialize_lin_tape_class() */

/*******************************************************************************
*                                                                              *
*  NAME: alloc_and_register                                                    *
*                                                                              *
*  PURPOSE: Allocate and register basic driver variables and interfaces        *
*                                                                              *
*  PARAMETERS: (all output)                                                    *
*                                                                              *
*    first_dev: First device in newly-allocated device region                  *
*    chrdev_regd: region successfully registered                               *
*    driver_regd: driver successfully registered                               *
*    interface_regd: SCSI interface successfully registered                    *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int alloc_and_register(dev_t** first_dev, int* chrdev_regd, int* driver_regd,
		int* interface_regd)
{
	int rc = STATUS_SUCCESS;

	*first_dev = kmalloc(sizeof(dev_t), GFP_KERNEL);
	if(!*first_dev) {
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	rc = alloc_chrdev_region(*first_dev, LinTapeBaseMinor,
		LinTapeMaxCount, "lin_tape");
	if(rc != STATUS_SUCCESS) {
		rc = 1;
		goto EXIT_LABEL;
	} /* if */
	*chrdev_regd = True;

	WRITE_LOCK(&lin_tape_lock);
	lin_tape_major = MAJOR(**first_dev);
	WRITE_UNLOCK(&lin_tape_lock);
	DbgPrint(("lin_tape: lin_tape is registered under major number %d\n",
		lin_tape_major));

	rc = scsi_register_driver(&lin_tape_driver.gendrv);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: scsi_register_driver failed.\n");
		goto EXIT_LABEL;
	} /* if */
	*driver_regd = True;

	WRITE_LOCK(&lin_tape_lock);
	lin_tape_cleanup_module = False;
	if(lin_tape_log == NULL) {
		lin_tape_log_len = 0;
		lin_tape_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		if(!lin_tape_log){
			printk("lin_tape: log space not available.\n");
		} /* if */
	} /* if */
	WRITE_UNLOCK(&lin_tape_lock);

	WRITE_LOCK(&lin_tape_trace_lock);
	lin_tape_trace_type = LIN_TAPE_INVALID_TRACE;
	WRITE_UNLOCK(&lin_tape_trace_lock);

	if (lin_tape_as_sfmp)
		rc = join_register_interface(&lin_tape_class_interface);
	else
	/* the following call will register the driver to receive
		notification of new SCSI devices via add and release functions.
		lin_tape_add_class_device will in turn call device_create to add
		individual devices to the /dev device tree */
	rc = scsi_register_interface(&lin_tape_class_interface);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: cannot register class interface.\n");
		goto EXIT_LABEL;
	} /* if */
	*interface_regd = True;

	/* create daemon pseudo file cdevs */
	lin_tape_create_cdev(&lin_tape_daemon_dev, PSEUDO_MINOR, 0,
		"/dev/IBMtape");

EXIT_LABEL:

	return rc;
} /* alloc_and_register() */

/*******************************************************************************
*                                                                              *
*  NAME: create_driver_attributes                                              *
*                                                                              *
*  PURPOSE: Create standard driver attributes in sysfs                         *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

void create_driver_attributes(void)
{
	int rc = 0; /* to avoid warn_unused_result */
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_version);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_lin_tape_debug);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_auto_logging);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_petro);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_changer_open_reserve);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_default_sys_encryption_proxy);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_default_sys_encryption_write);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_support_dio);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_wait_on_config);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_disable_density_on_open);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_dynamic_attributes);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_busy_retry);
	rc = driver_create_file(&lin_tape_driver.gendrv,
		&driver_attr_disable_auto_drive_dump);
} /* create_driver_attributes() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_undo_init                                                    *
*                                                                              *
*  PURPOSE: If initialization fails, undo the steps already done               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    interface_regd: SCSI interface was registered                             *
*    chrdev_regd: character-device region was allocated and/or registered      *
*    driver_regd: the driver was registered with the SCSI subsystem            *
*    lin_tape_class_allocd: the class was set up in sysfs                      *
*    first_dev: first device in region; used to unregister                     *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

void lin_tape_undo_init(int interface_regd, int chrdev_regd, int driver_regd,
			int lin_tape_class_allocd, dev_t* first_dev)
{
	if(interface_regd){
		if (lin_tape_as_sfmp) join_unregister_interface(&lin_tape_class_interface);
		else scsi_unregister_interface(&lin_tape_class_interface);
	}

	WRITE_LOCK(&lin_tape_lock);
	if(lin_tape_log) {
		kfree(lin_tape_log);
		lin_tape_log = NULL;
		lin_tape_log_len = 0;
	} /* if */
	WRITE_UNLOCK(&lin_tape_lock);

	if(chrdev_regd)
		unregister_chrdev_region(*first_dev, LinTapeMaxCount);

	if(driver_regd)
		scsi_unregister_driver(&lin_tape_driver.gendrv);

	if(lin_tape_class_allocd) {
		class_destroy(lin_tape_class);
		lin_tape_class = NULL;
	} /* if */
} /* lin_tape_undo_init() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_exit                                                         *
*                                                                              *
*  PURPOSE: Unregister class and driver                                        *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

static void lin_tape_exit()
{
	dev_t *first_dev = NULL;
	int i = 0;

	DbgPrint(("lin_tape: lin_tape_exit entry.\n"));

	WRITE_LOCK(&lin_tape_lock);
	lin_tape_cleanup_module = True;
	WRITE_UNLOCK(&lin_tape_lock);

	if (lin_tape_as_sfmp)
		join_unregister_interface(&lin_tape_class_interface);
	else scsi_unregister_interface(&lin_tape_class_interface);

	DbgPrint(("lin_tape: scsi_unregister_interface done.\n"));

	first_dev = kmalloc(sizeof(dev_t), GFP_KERNEL);
	if(!first_dev) goto EXIT_LABEL;

	*first_dev = MKDEV(lin_tape_major, LinTapeBaseMinor);
	unregister_chrdev_region(*first_dev, LinTapeMaxCount);
	DbgPrint(("lin_tape: unregister_chrdev_region done.\n"));
	kfree(first_dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	DbgPrint(("lin_tape: removing driver attribute files...\n"));
	driver_remove_file(&lin_tape_driver.gendrv, &driver_attr_version);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_lin_tape_debug);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_auto_logging);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_changer_open_reserve);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_petro);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_default_sys_encryption_proxy);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_default_sys_encryption_write);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_support_dio);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_wait_on_config);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_disable_density_on_open);
	driver_remove_file(&lin_tape_driver.gendrv,
		&driver_attr_dynamic_attributes);
#endif
	WRITE_LOCK(&tape_des_lock);
	if(tape_devs) {
		for(i = 0; i < MaxLinTapeDevices; i++) {
			if(tape_devs[i]){
                if (tape_devs[i]->drv){
                    lin_tape_drive_cleanup(tape_devs[i]->drv);
                    kfree(tape_devs[i]->drv);
                    tape_devs[i]->drv = NULL;
                }
			kfree(tape_devs[i]);
			tape_devs[i] = NULL;
            }/* if */
		} /* for */
		kfree(tape_devs);
		tape_devs = NULL;
	} /* if */

	if(tsm_dev) lin_tape_delete_tsm_pseudo();

	if(special_dev) lin_tape_delete_special_dev();

	WRITE_UNLOCK(&tape_des_lock);

	WRITE_LOCK(&changer_des_lock);
	if(changer_devs) {
		for(i = 0; i < MaxLinTapeChangerDevices; i++) {
			if(changer_devs[i]) {
                if (changer_devs[i]->chg){
                    lin_tape_changer_cleanup(i);
                    kfree(changer_devs[i]->chg);
                    changer_devs[i]->chg = NULL;
                }
                kfree(changer_devs[i]);
                changer_devs[i] = NULL;
			}
		} /* for */
		kfree(changer_devs);
		changer_devs = NULL;
	} /* if */
	WRITE_UNLOCK(&changer_des_lock);

	DbgPrint(("lin_tape: unregister_driver...\n"));
	scsi_unregister_driver(&lin_tape_driver.gendrv);
	DbgPrint(("lin_tape: unregister_driver done.\n"));

	lin_tape_remove_proc();

	if(lin_tape_daemon_dev) {
		cdev_del(lin_tape_daemon_dev);
		lin_tape_daemon_dev = NULL;
	} /* if */

	WRITE_LOCK(&lin_tape_lock);
	lin_tape_major = 0;
	if(lin_tape_log) kfree(lin_tape_log);

	lin_tape_log = NULL;
	lin_tape_log_len = 0;
	WRITE_UNLOCK(&lin_tape_lock);

	class_destroy(lin_tape_class);

	lin_tape_class = NULL;

	WRITE_LOCK(&lin_tape_trace_lock);
	lin_tape_trace_type = LIN_TAPE_INVALID_TRACE;
	WRITE_UNLOCK(&lin_tape_trace_lock);

	printk("lin_tape %s is unloaded.\n", LIN_TAPE_VERSION);

EXIT_LABEL:

	WRITE_LOCK(&lin_tape_lock);
	lin_tape_cleanup_module = False;
	WRITE_UNLOCK(&lin_tape_lock);

	DbgPrint(("lin_tape: lin_tape_exit exit\n"));
	return;
} /* lin_tape_exit() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_supported_device                                             *
*                                                                              *
*  PURPOSE: Determine whether device is supported by this driver               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    device: the device we will claim or reject                                *
*                                                                              *
*  RETURNS: tape_id if valid or 0 if invalid                                   *
*                                                                              *
*******************************************************************************/

int lin_tape_supported_device(struct scsi_device* device)
{
	int i = 0;
	int j = 0;
	int cmplength = 0;
	int tape_id = 0;

	if((device->type != TYPE_TAPE) && (device->type != TYPE_MEDIUM_CHANGER))
		return 0;

	for(i = 0; supported_vendor[i]; i++) {

		if(!strncmp(device->vendor, supported_vendor[i],
			strlen(supported_vendor[i]))) {

			for(j = 0; supported_device[j]; j++) {

				if(!strncmp(device->model, supported_device[j],
					strlen(supported_device[j]))
				&& strlen(supported_device[j]) > cmplength) {

					tape_id = supported_devid[j];
					cmplength = strlen(supported_device[j]);

				} /* if */

			} /* for */

		} /* if */

	} /* for */

	return tape_id;

} /* lin_tape_supported_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_probe                                                        *
*                                                                              *
*  PURPOSE: Avoid claiming the unsupported devices; part of struct scsi_driver *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device we will claim or reject                                   *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

static int lin_tape_probe(struct device *dev)
{
	if(!lin_tape_supported_device(to_scsi_device(dev))) return -ENODEV;

	return 0;

} /* lin_tape_probe() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_remove                                                       *
*                                                                              *
*  PURPOSE: Corresponds to lin_tape_probe; part of struct scsi_driver          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device we will claim or reject                                   *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

static int lin_tape_remove(struct device *dev)
{
	if(!lin_tape_supported_device(to_scsi_device(dev))) return -ENODEV;

	return 0;

} /* lin_tape_remove() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_add_class_device                                             *
*                                                                              *
*  PURPOSE: Called by the kernel when SCSI interface detects a new device      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    class_dev: the class_device associated with this request                  *
*    class_intf: the interface that triggered the calling of this function     *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

static int lin_tape_add_class_device(struct device *class_dev,
					struct class_interface* class_intf)
{
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: lin_tape_add_class_device entry.\n"));

	tape_log_msg(lin_tape_log, &lin_tape_log_len, 208);

	rc = lin_tape_config_device(class_dev ? class_dev->parent : NULL);

	DbgPrint(("lin_tape: lin_tape_add_class_device exit: %d\n", rc));
	return rc;
} /* lin_tape_add_class_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_config_device                                                *
*                                                                              *
*  PURPOSE: Add a new logical device and bind the driver to the device         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device we are adding                                             *
*                                                                              *
*  RETURNS: 0 on success or error code                                         *
*                                                                              *
*******************************************************************************/

int lin_tape_config_device(struct device* dev)
{
	int rc = STATUS_SUCCESS;
	int tape_id = 0;
	struct scsi_device* device = NULL;

	DbgPrint(("lin_tape: lin_tape_config_device entry.\n"));

	if(dev) device = to_scsi_device(dev);

	if(!device || !device->host) {
		if (device) printk("lin_tape: something came wrong at scsi device host is NULL.\n");
		else printk("lin_tape: to_scsi_device returns NULL.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	tape_id = lin_tape_supported_device(device);
	if(!tape_id) {
		DbgPrint(("lin_tape: do not support %s %s %s:\n",
			device->vendor, device->model, device->rev));
		DbgPrint(("lin_tape: scsi bus:%d, channel:%d, id:%d, lun:%d\n",
			device->host->host_no, device->channel, device->id,
			(int)device->lun));

		rc = -ENODEV;
		goto EXIT_LABEL;
	} /* if */

	printk("lin_tape: detected device at host %d, channel %d, id %d, "\
		"lun %d\n", device->host->host_no, device->channel, device->id,
		(int)device->lun);
	printk("%.3s %.11s %.4s\n", device->vendor, device->model, device->rev);
	/* bind driver to this device */
	if(dev) lin_tape_bind_device(dev);

	if(device->type == TYPE_TAPE) {
		rc = lin_tape_config_drive(tape_id, device);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} else if(device->type == TYPE_MEDIUM_CHANGER) {
		rc = lin_tape_config_changer(tape_id, device);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* else if */

	if(wait_on_config)
		msleep(wait_on_config);

EXIT_LABEL:

	DbgPrint(("lin_tape: lin_tape_config_device exit: %d\n", rc))
	return rc;

} /* lin_tape_config_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_bind_device                                                  *
*                                                                              *
*  PURPOSE: Bind the newly discovered SCSI device to this driver               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: the device we are adding                                             *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_bind_device(struct device* dev)
{
	int rc = 0;

	if(dev->driver) {
		if(dev->driver != &(lin_tape_driver.gendrv)) {

			device_release_driver(dev);
		} /* if */

	} else DbgPrint(("lin_tape: the device currently is not bound "\
			"to any driver.\n"));

	if(dev->driver != &(lin_tape_driver.gendrv)) {
		dev->driver = &(lin_tape_driver.gendrv);
		rc = device_bind_driver(dev);
	} /* if */

} /* lin_tape_bind_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_config_drive                                                 *
*                                                                              *
*  PURPOSE: Perform tape drive configuration                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    tape_id: the drive model we are configuring                               *
*    sdev: the scsi_device object passed by the interface                      *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_config_drive(int tape_id, struct scsi_device* sdev)
{
	int rc = STATUS_SUCCESS, i = 0;
	char rtype1[RESERVE_TYPE_LENGTH];
	char rtype2[RESERVE_TYPE_LENGTH];
	memset(rtype1, '\0', RESERVE_TYPE_LENGTH);
	memcpy(rtype1, "reserve_6", strlen("reserve_6"));

	memset(rtype2, '\0', RESERVE_TYPE_LENGTH);
	memcpy(rtype2, "persistent", strlen("persistent"));

	WRITE_LOCK(&tape_des_lock);

	if(!tape_devs) {
		rc = allocate_tdesc_array();
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(memcmp(tape_reserve_type, rtype1, RESERVE_TYPE_LENGTH)
		&& memcmp(tape_reserve_type, rtype2, RESERVE_TYPE_LENGTH)) {
			memset(tape_reserve_type, '\0', RESERVE_TYPE_LENGTH);
			memcpy(tape_reserve_type, "reserve_6",
				strlen("reserve_6"));
		} /* if */

	} /* if */
	for(i = 0; i < MaxLinTapeDevices; i++) {

		if(tape_devs[i] && tape_devs[i]->drv) continue;
		rc = alloc_and_init_tdesc(sdev, tape_id, &i, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		if(!tsm_dev && !memcmp(tape_reserve_type, "persistent",
			strlen("persistent"))) {
			rc = lin_tape_create_tsm_pseudo(sdev);
		}

		if (!special_dev) {
		    rc=lin_tape_create_special_dev(sdev);
		}
		break;
	} /* for */
	if(i >= MaxLinTapeDevices)
		printk("lin_tape: maximum drives exceeded, cannot attach.\n");

EXIT_LABEL:

	WRITE_UNLOCK(&tape_des_lock);
	return rc;
} /* lin_tape_config_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_config_changer                                               *
*                                                                              *
*  PURPOSE: Perform medium changer configuration                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    tape_id: the changer model we are configuring                             *
*    device: the scsi_device object passed by the interface                    *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_config_changer(int tape_id, struct scsi_device* device)
{
	int rc = STATUS_SUCCESS, i = 0;

	WRITE_LOCK(&changer_des_lock);

	if(!changer_devs) {
		rc = allocate_cdesc_array();
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(changer_devs[i] && changer_devs[i]->chg)
			continue;

		rc = alloc_and_init_cdesc(device, tape_id, &i, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		break;
	} /* for */

	if(i >= MaxLinTapeChangerDevices)
		printk("lin_tape: maximum changers exceeded, cannot attach.\n");

EXIT_LABEL:

	WRITE_UNLOCK(&changer_des_lock);
	return rc;
} /* lin_tape_config_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: allocate_tdesc_array                                                  *
*                                                                              *
*  PURPOSE: allocate the tape descriptors array                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*                                                                              *
*******************************************************************************/

int allocate_tdesc_array()
{
	int rc = STATUS_SUCCESS;

	tape_devs = kmalloc(MaxLinTapeDevices
		* sizeof(struct descriptor_tape*), GFP_KERNEL);

	if(!tape_devs) {
		printk("lin_tape: cannot allocate tape descriptors\n");
		tape_log_msg(lin_tape_log, &lin_tape_log_len,
			TRC_TPS_NOALLOC);

		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(tape_devs, '\0',
		MaxLinTapeDevices * sizeof(struct descriptor_tape *));

EXIT_LABEL:

	return rc;
} /* allocate_tdesc_array() */

/*******************************************************************************
*                                                                              *
*  NAME: allocate_cdesc_array                                                  *
*                                                                              *
*  PURPOSE: allocate the changer descriptors array                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding changer_des_lock for write          *
*                                                                              *
*******************************************************************************/

int allocate_cdesc_array()
{
	int rc = STATUS_SUCCESS;

	changer_devs = kmalloc(MaxLinTapeChangerDevices
		* sizeof(struct descriptor_changer*), GFP_KERNEL);

	if(!changer_devs) {
		printk("lin_tape: cannot allocate changer descriptors\n");
		tape_log_msg(lin_tape_log, &lin_tape_log_len,
			TRC_CGS_NOALLOC);

		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(changer_devs, '\0',
		MaxLinTapeChangerDevices * sizeof(struct descriptor_changer*));

	EXIT_LABEL:

	return rc;
} /* allocate_cdesc_array() */

/*******************************************************************************
*                                                                              *
*  NAME: alloc_and_init_tdesc                                                  *
*                                                                              *
*  PURPOSE: allocate and initialize a tape descriptor                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: the scsi_device associated with this request                        *
*    tape_id: the type of tape device                                          *
*    i: the index of the device in the tape_devs array                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*                                                                              *
*******************************************************************************/

int alloc_and_init_tdesc(struct scsi_device* sdev, int tape_id,
                         int* i, boolean rescan)
{
	int rc = STATUS_SUCCESS;
	unchar* trc = NULL;
	struct tape_descriptor *tdesc;

	if(!tape_devs[*i]) {
		tape_devs[*i] = kmalloc(sizeof(struct descriptor_tape),
			GFP_KERNEL);

        if(!tape_devs[*i]) {
            printk("lin_tape: cannot allocate tape descriptor\n");
            tape_log_msg(lin_tape_log, &lin_tape_log_len,
			TRC_TP_NOALLOC);
            rc = -ENOMEM;
            goto EXIT_LABEL;
        } /* if */
        tape_devs[*i]->drv = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
        tape_devs[*i]->drv_entry_lock = RW_LOCK_UNLOCKED;
#else
        tape_devs[*i]->drv_entry_lock = __RW_LOCK_UNLOCKED(tape_devs[*i]->drv_entry_lock);
#endif
	}/* if */
    WRITE_LOCK(&tape_devs[*i]->drv_entry_lock);
	if (!tape_devs[*i]->drv){
        tape_devs[*i]->drv = kmalloc(sizeof(struct tape_descriptor),
			GFP_KERNEL);

        if (!tape_devs[*i]->drv){
            WRITE_UNLOCK(&tape_devs[*i]->drv_entry_lock);
            printk("lin_tape: cannot allocate tape descriptor\n");
            tape_log_msg(lin_tape_log, &lin_tape_log_len,
			TRC_TP_NOALLOC);
            rc = -ENOMEM;
            goto EXIT_LABEL;
        }
	}/* if */
	memset(tape_devs[*i]->drv, '\0', sizeof(struct tape_descriptor));
	tape_devs[*i]->drv->drv_id = tape_id;
    tape_devs[*i]->drv->dev_index = *i;
    WRITE_UNLOCK(&tape_devs[*i]->drv_entry_lock);
	rc = lin_tape_init_drive_des(tape_devs[*i]->drv, sdev, i, rescan);
	tdesc = tape_devs[*i]->drv;
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape_init_drive_des returned %d.\n", rc);
        WRITE_LOCK(&tape_devs[*i]->drv_entry_lock);
		if (tape_devs[*i]->drv) kfree(tape_devs[*i]->drv);
		tape_devs[*i]->drv = NULL;
        WRITE_UNLOCK(&tape_devs[*i]->drv_entry_lock);
	} else {
		/* update the device table */
		trc = kmalloc(1 + tdesc->serial_num_len,
			GFP_KERNEL);

		if(!trc) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		trc[0] = tdesc->serial_num_len;
		if(tdesc->serial_num_len)
			memcpy(&trc[1], tdesc->serial_num,
				tdesc->serial_num_len);
		tape_setup_and_log_full(tdesc,
			tdesc->dev_index, TRC_TP_SERIAL_NUM,
			trc, 1 + tdesc->serial_num_len);

		kfree(trc);
		trc = NULL;

	} /* else */

EXIT_LABEL:

	return rc;
} /* alloc_and_init_tdesc() */

/*******************************************************************************
*                                                                              *
*  NAME: alloc_and_init_cdesc                                                  *
*                                                                              *
*  PURPOSE: allocate and initialize a changer descriptor                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: the scsi_device associated with this request                        *
*    tape_id: the type of changer device                                       *
*    i: the index of the device in the changer_devs array                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding changer_des_lock for write          *
*                                                                              *
*******************************************************************************/

int alloc_and_init_cdesc(struct scsi_device* sdev, int tape_id,
                         int* i, boolean rescan)
{
	int rc = STATUS_SUCCESS;
	unchar* trc = NULL;
	struct changer_descriptor *cdesc;

	if(!changer_devs[*i]) {
		changer_devs[*i] = kmalloc(sizeof(struct descriptor_changer),
			GFP_KERNEL);

        if(!changer_devs[*i]) {
            printk("lin_tape: unable to allocate changer "\
                "descriptor\n");
            tape_log_msg(lin_tape_log, &lin_tape_log_len, TRC_CG_NOALLOC);
            rc = -ENOMEM;
            goto EXIT_LABEL;
        } /* if */
        changer_devs[*i]->chg = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
        changer_devs[*i]->entry_lock = RW_LOCK_UNLOCKED;
#else
        changer_devs[*i]->entry_lock = __RW_LOCK_UNLOCKED(changer_devs[*i]->entry_lock);
#endif
	} /* if */
    WRITE_LOCK(&changer_devs[*i]->entry_lock);
    if (!changer_devs[*i]->chg){
		changer_devs[*i]->chg = kmalloc(sizeof(struct changer_descriptor),
			GFP_KERNEL);
        if(!changer_devs[*i]->chg) {
            WRITE_UNLOCK(&changer_devs[*i]->entry_lock);
            printk("lin_tape: unable to allocate changer "\
                "descriptor\n");
            tape_log_msg(lin_tape_log, &lin_tape_log_len, TRC_CG_NOALLOC);
            rc = -ENOMEM;
            goto EXIT_LABEL;
        } /* if */
        memset(changer_devs[*i]->chg, '\0', sizeof(struct descriptor_changer));
    }
	changer_devs[*i]->chg->chg_id = tape_id;
	changer_devs[*i]->chg->dev_index = *i;
    WRITE_UNLOCK(&changer_devs[*i]->entry_lock);
	rc = lin_tape_init_changer_des(changer_devs[*i]->chg, sdev, i, rescan);
    cdesc = changer_devs[*i]->chg;
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape_init_changer_des returned %d.\n", rc);
        WRITE_LOCK(&changer_devs[*i]->entry_lock);
		if (changer_devs[*i]->chg) kfree(changer_devs[*i]->chg);
		changer_devs[*i]->chg = NULL;
		WRITE_UNLOCK(&changer_devs[*i]->entry_lock);
	} else {
		/* update the device and path table */
		printk("lin_tape: attached IBMchanger%d\n", *i);
		trc = kmalloc(1 + cdesc->serial_num_len,
			GFP_KERNEL);
		if(!trc) {
            rc = -ENOMEM;
            goto EXIT_LABEL;
		}

		trc[0] = cdesc->serial_num_len;
		if(cdesc->serial_num_len)
			memcpy(&trc[1], cdesc->serial_num,
				cdesc->serial_num_len);
		changer_setup_and_log_full(cdesc,
			cdesc->dev_index, TRC_CG_SERIAL_NUM,
			trc, 1 + cdesc->serial_num_len);

		kfree(trc);
		trc = NULL;
	} /* else */

EXIT_LABEL:

	return rc;
} /* alloc_and_init_cdesc() */

/*******************************************************************************
*                                                                              *
*  NAME: cleanup_and_free_tdesc                                                *
*                                                                              *
*  PURPOSE: free tape descriptor                                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    idx: the index within the tape_descriptor array                           *
*    lock_acquired: already own des_lock, do not attempt to lock               *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void cleanup_and_free_tdesc(int idx)
{
	int host, channel;
	int k = 0;
	struct tape_descriptor* trav = NULL, *drv;

	wait_queue_head_t wq;
	init_waitqueue_head(&wq);

    if (!tape_devs || !tape_devs[idx] || !tape_devs[idx]->drv){
        return;
    }
    READ_LOCK(&tape_devs[idx]->drv_entry_lock);
	host = tape_devs[idx]->drv->host;
	channel = tape_devs[idx]->drv->channel;
	while( drive_busy(tape_devs[idx]->drv, F_ADD_REMOVE)
            && k < TOT_RETRIES * 32) {
		printk("lin_tape: IBMtape%d busy %d. Waiting ten seconds\n",
         idx, tape_devs[idx]->drv->busy);
        READ_UNLOCK(&tape_devs[idx]->drv_entry_lock);
		wait_event_interruptible_timeout(wq,
			!tape_devs[idx] || !tape_devs[idx]->drv ||
			!(tape_devs[idx]->drv->busy - F_ADD_REMOVE) , HZ * 100);
        READ_LOCK(&tape_devs[idx]->drv_entry_lock);
		k++;
	} /* while */

	if (!tape_devs[idx] ||
        !tape_devs[idx]->drv ||
        !tape_devs[idx]->drv->ghost_device) {
        READ_UNLOCK(&tape_devs[idx]->drv_entry_lock);
        return;
    }
	if (drive_busy(tape_devs[idx]->drv, F_ADD_REMOVE)){
		tape_devs[idx]->drv->remove_pending = True;
        READ_UNLOCK(&tape_devs[idx]->drv_entry_lock);
		return;
	}
	drv = tape_devs[idx]->drv;
    READ_UNLOCK(&tape_devs[idx]->drv_entry_lock);

	if (tape_devs[idx]->drv->rew_dev){
		cdev_del(tape_devs[idx]->drv->rew_dev);
		tape_devs[idx]->drv->rew_dev = NULL;
	}

	if (tape_devs[idx]->drv->nonrew_dev){
		cdev_del(tape_devs[idx]->drv->nonrew_dev);
		tape_devs[idx]->drv->nonrew_dev = NULL;
	}

	if(tape_devs[idx]->drv->nonresv_dev) {
		cdev_del(tape_devs[idx]->drv->nonresv_dev);
		tape_devs[idx]->drv->nonresv_dev = NULL;
	}

	lin_tape_drive_cleanup(drv);

	WRITE_LOCK(&tape_des_lock);
	for(trav = tape_devs[idx]->drv;
		trav->next_path != tape_devs[idx]->drv;
		trav = trav->next_path) continue;
	trav->next_path = trav->next_path->next_path;
	tape_devs[idx]->drv->next_path = tape_devs[idx]->drv;
    WRITE_UNLOCK(&tape_des_lock);

    WRITE_LOCK(&tape_devs[idx]->drv_entry_lock);
	kfree(tape_devs[idx]->drv);
	tape_devs[idx]->drv = NULL;
    WRITE_UNLOCK(&tape_devs[idx]->drv_entry_lock);

	if(!lin_tape_cleanup_module && lin_tape_trace_type != LIN_TAPE_INVALID_TRACE) {
		tape_log_msg5(lin_tape_log, &lin_tape_log_len, TRC_MSG_TP_RM,
			idx >> 8, idx & 0xFF, host, channel);
	} /* if */

	printk("lin_tape: IBMtape%d removed.\n", idx);

} /* cleanup_and_free_tdesc() */

/*******************************************************************************
*                                                                              *
*  NAME: cleanup_and_free_cdesc                                                *
*                                                                              *
*  PURPOSE: free changer descriptor                                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    idx: the index within the changer_descriptor array                        *
*    lock_acquired: already own des_lock, do not attempt to lock               *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void cleanup_and_free_cdesc(int idx)
{
	int host, k = 0, channel;
	struct changer_descriptor* trav = NULL;
	wait_queue_head_t wq;

	init_waitqueue_head(&wq);

	if (!changer_devs || !changer_devs[idx] || !changer_devs[idx]->chg){
        return;
	}
    WRITE_LOCK(&changer_devs[idx]->entry_lock);
	host = changer_devs[idx]->chg->host;
	channel = changer_devs[idx]->chg->channel;
	while(changer_busy(changer_devs[idx]->chg,F_ADD_REMOVE)
		&& k < TOT_RETRIES * 32) {
		printk("lin_tape: IBMchanger%d busy %d. Waiting ten seconds\n",
         idx,changer_devs[idx]->chg->busy);
        WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
		wait_event_interruptible_timeout(wq,
			!changer_devs[idx] || !changer_devs[idx]->chg ||
			!(changer_devs[idx]->chg->busy - F_ADD_REMOVE), HZ * 100);
        WRITE_LOCK(&changer_devs[idx]->entry_lock);
        k++;
	} /* while */

	if (!changer_devs[idx] || !changer_devs[idx]->chg ||
        !changer_devs[idx]->chg->ghost_device){
        WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
        return;
    }
	if (changer_busy(changer_devs[idx]->chg,F_ADD_REMOVE)||
        (changer_devs[idx]->chg->cur_path == idx &&
         changer_devs[idx]->chg->chg_open )){
		changer_devs[idx]->chg->remove_pending = True;
        WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
		return;
	}

    WRITE_UNLOCK(&changer_devs[idx]->entry_lock);

	if (changer_devs[idx]->chg->chg_dev){
		cdev_del(changer_devs[idx]->chg->chg_dev);
		changer_devs[idx]->chg->chg_dev = NULL;
	}

	lin_tape_changer_cleanup(idx);

	WRITE_LOCK(&changer_des_lock);
	for(trav = changer_devs[idx]->chg;
		trav->next_path != changer_devs[idx]->chg;
		trav = trav->next_path) continue;

	trav->next_path = trav->next_path->next_path;
	changer_devs[idx]->chg->next_path = changer_devs[idx]->chg;
	WRITE_UNLOCK(&changer_des_lock);
    WRITE_LOCK(&changer_devs[idx]->entry_lock);
	kfree(changer_devs[idx]->chg);
	changer_devs[idx]->chg = NULL;
    WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
	if(!lin_tape_cleanup_module && lin_tape_trace_type != LIN_TAPE_INVALID_TRACE){
		tape_log_msg5(lin_tape_log, &lin_tape_log_len, TRC_MSG_CG_RM,
			idx >> 8, idx & 0xFF, host, channel);

	} /* if */

	printk("lin_tape: IBMchanger%d removed.\n", idx);

} /* cleanup_and_free_cdesc() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_remove_class_device                                          *
*                                                                              *
*  PURPOSE: remove a device                                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    class_dev: the struct class_device or struct device to remove             *
*    class_intf: the interface that contains pointers to add/remove for dvc    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

static void lin_tape_remove_class_device(struct device *class_dev,
	struct class_interface* class_intf)
{
	struct scsi_device *device = NULL;

	DbgPrint(("lin_tape: lin_tape_remove_class_device entry.\n"));

	/* lin_taped daemon is not running if cleaup_module is executing */
	if(lin_tape_log && !lin_tape_cleanup_module)
		tape_log_msg(lin_tape_log, &lin_tape_log_len, TRC_RM_CLASS_DVC);

	device = to_scsi_device(class_dev->parent);

	if(device == NULL){
		goto EXIT_LABEL;
	}

	if((device->type != TYPE_TAPE)
		&& (device->type != TYPE_MEDIUM_CHANGER)){
		goto EXIT_LABEL;
	}

	lin_tape_unconfig_device(class_dev->parent);

EXIT_LABEL:

	DbgPrint(("lin_tape: lin_tape_remove_class_device exit: void\n"));
	return;
} /* lin_tape_remove_class_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_unconfig_device                                              *
*                                                                              *
*  PURPOSE: remove a device from the driver device table                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: device to unconfigure                                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_unconfig_device(struct device *dev)
{
	struct scsi_device* device = NULL;

	DbgPrint(("lin_tape: lin_tape_unconfig_device entry.\n"));

	if(dev) device = to_scsi_device(dev);
	if(!device) goto EXIT_LABEL;

	if(device->type == TYPE_TAPE) {
		unconfig_drive_paths(dev);

	} else if(device->type == TYPE_MEDIUM_CHANGER) {
		unconfig_changer_paths(dev);
	} /* else if */

	if (wait_on_config) msleep(wait_on_config);

EXIT_LABEL:

	DbgPrint(("lin_tape_unconfig_device exit: void\n"));
	return;
} /* lin_tape_unconfig_device() */

/*******************************************************************************
*                                                                              *
*  NAME: unconfig_drive_paths                                                  *
*                                                                              *
*  PURPOSE: if all paths to a drive can be removed, remove them                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: device to unconfigure                                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*                                                                              *
*******************************************************************************/

void unconfig_drive_paths(struct device* dev)
{
	struct scsi_device *device = NULL;
	int i = 0, j = 0, pidx = 0, cur = 0, altp = False;
	int paths_array[MaxScsiPath];
	struct tape_descriptor* dri = NULL, * drj = NULL;

	if(dev) device = to_scsi_device(dev);
	if(!device) goto EXIT_LABEL;

	WRITE_LOCK(&tape_des_lock);
	if (!tape_devs){
        WRITE_UNLOCK(&tape_des_lock);
        return;
	}
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || ! tape_devs[i]->drv)
            continue;
		dri = tape_devs[i]->drv;
		if (dri->dev_obj == device)
            break;
	}/* for */
    WRITE_UNLOCK(&tape_des_lock);
	if ( i < MaxLinTapeDevices ){
	    WRITE_LOCK(&tape_devs[i]->drv_entry_lock);
        dri->ghost_device = True;
        dri->busy |= F_ADD_REMOVE;
        altp = dri->alt_pathing;
	    WRITE_UNLOCK(&tape_devs[i]->drv_entry_lock);
		device_release_driver(dev);
		lin_tape_unconfig_drive(&dri->dev_obj->sdev_gendev,dri);
		if(altp){
			tape_setup_and_log_full(dri, dri->dev_index,
					TRC_TP_GHOST, NULL, 0);
		}
		else {
			cleanup_and_free_tdesc(i);
			if (tape_devs[i]->drv)
                dri->busy &= 0xFE;
			goto EXIT_LABEL;
		} /* if */
		drj = dri;
		WRITE_LOCK(&tape_devs[i]->drv_entry_lock);
        dri->busy &= 0xFE;
        WRITE_UNLOCK(&tape_devs[i]->drv_entry_lock);
        WRITE_LOCK(&tape_des_lock);
		do {
			/* unconfig all paths for a device at once or
				none at all */
			if(!drj->ghost_device){
                WRITE_UNLOCK(&tape_des_lock);
				goto EXIT_LABEL;
			}
			paths_array[pidx++] = drj->dev_index;
			drj = drj->next_path;
		} while(drj != dri);
        WRITE_UNLOCK(&tape_des_lock);
		for(j = 0; j < pidx; j++) {
			cur = paths_array[j];
			cleanup_and_free_tdesc(cur);
		} /* for */
		goto EXIT_LABEL;
	} /* if found */

EXIT_LABEL:

	return;
} /* unconfig_drive_paths() */

/*******************************************************************************
*                                                                              *
*  NAME: unconfig_changer_paths                                                *
*                                                                              *
*  PURPOSE: if all paths to a changer can be removed, remove them              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dev: device to unconfigure                                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*    This function must be called while holding changer_des_lock for write     *
*                                                                              *
*******************************************************************************/

void unconfig_changer_paths(struct device* dev)
{
	struct scsi_device *device = NULL;
	int i = 0, j = 0, pidx = 0, cur = 0, altp = False;
	int paths_array[MaxScsiPath];
	struct changer_descriptor* chi = NULL, * chj = NULL;

	if(dev) device = to_scsi_device(dev);
	if(!device || !changer_devs) goto EXIT_LABEL;

	WRITE_LOCK(&changer_des_lock);
	if (!changer_devs){
        WRITE_UNLOCK(&changer_des_lock);
        return;
	}
	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg
            || changer_devs[i]->chg->dev_obj != device) continue;
		chi = changer_devs[i]->chg;
        break;
	}
	WRITE_UNLOCK(&changer_des_lock);
	if (chi){
        WRITE_LOCK(&changer_devs[i]->entry_lock);
        chi->ghost_device = True;
        if (chi->busy)
            printk("lin_tape: IBMChanger%d busy - %d\n", chi->dev_index, chi->busy);
        chi->busy |= F_ADD_REMOVE;
        altp = chi->alt_pathing;
        WRITE_UNLOCK(&changer_devs[i]->entry_lock);
		device_release_driver(dev);
        lin_tape_unconfig_changer(
            &changer_devs[i]->chg->dev_obj->sdev_gendev,
            changer_devs[i]->chg->dev_obj, i);

		if (altp){
			changer_setup_and_log_full(chi,
				chi->dev_index, TRC_CG_GHOST, NULL, 0);
		}
		else {
			cleanup_and_free_cdesc(i);
			if (changer_devs[i]->chg){
                WRITE_LOCK(&changer_devs[i]->entry_lock);
                chi->busy &= 0xFE;
                WRITE_UNLOCK(&changer_devs[i]->entry_lock);
			}
			goto EXIT_LABEL;
		} /* if */

		chj = chi;
        WRITE_LOCK(&changer_devs[i]->entry_lock);
        chi->busy &= 0xFE;
        WRITE_UNLOCK(&changer_devs[i]->entry_lock);
        WRITE_LOCK(&changer_des_lock);
		do {
			/* unconfig all paths at once or none at all */
            if(!chj->ghost_device){
                WRITE_UNLOCK(&changer_des_lock);
                goto EXIT_LABEL;
            }
			paths_array[pidx++] = chj->dev_index;
			if(pidx == MaxScsiPath) break;
			chj = chj->next_path;
		} while(chj != chi);
        WRITE_UNLOCK(&changer_des_lock);

		for(j = 0; j < pidx; j++) {
			cur = paths_array[j];
            cleanup_and_free_cdesc(cur);
		} /* for */
		goto EXIT_LABEL;
	} /* for */

EXIT_LABEL:

	return;
} /* unconfig_changer_paths() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_unconfig_drive                                               *
*                                                                              *
*  PURPOSE: remove a drive from the driver device table                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dvc: the sysfs device object we are removing                              *
*    sdev: the sysfs scsi_device object we are removing                        *
*    idx: the index within the driver device table that points to this device  *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_unconfig_drive(struct device* dvc, struct tape_descriptor *drv)
{
	int rc = 0, k=0, idx=drv->dev_index;
	wait_queue_head_t wq;
	init_waitqueue_head(&wq);

	if (!dvc) goto EXIT_LABEL;

	DbgPrint(("lin_tape: IBMtape%d alt_pathing %d\n", idx,
			tape_devs[idx]->drv->alt_pathing));

    WRITE_LOCK(&tape_devs[idx]->drv_entry_lock);
	while( (drive_busy(tape_devs[idx]->drv, F_ADD_REMOVE) || tape_devs[idx]->drv->cmd_in_open)
            && k < 30) {
		printk("lin_tape: Unconfig IBMtape%d busy %d. Waiting ten seconds\n",
         idx, tape_devs[idx]->drv->busy);
        WRITE_UNLOCK(&tape_devs[idx]->drv_entry_lock);
		wait_event_interruptible_timeout(wq,
			!tape_devs[idx] || !tape_devs[idx]->drv, HZ * 100);
        WRITE_LOCK(&tape_devs[idx]->drv_entry_lock);
		k++;
	} /* while */

    while(drv->sdev_ct > 0){
        lin_tape_sdev_put(drv->dev_obj,
                        &drv->sdev_ct);
    }
	drv->dev_obj = NULL;
	if (drv->cur_path != drv->dev_index){
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        delete_parms_on_device(drv->dev_index);
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	}
    WRITE_UNLOCK(&tape_devs[idx]->drv_entry_lock);

	if(drv->rew_class_dev){
		lin_tape_uncfg_file_tp_rm(&drv->rew_class_dev, idx,
			drv->rewflags);
		drv->rewflags = 0;
	}
	if(drv->nonrew_class_dev){
		lin_tape_uncfg_file_tp_rm(&drv->nonrew_class_dev,
			idx + LinTapeNonRewindOffset, drv->nonrewflags);
		drv->nonrewflags = 0;
	}
	if(drv->nonresv_class_dev){
		lin_tape_uncfg_file_tp_rm(&drv->nonresv_class_dev,
			idx + LinTapeNonReserveOffset, drv->nonrsvflags);
		drv->nonrsvflags = 0;
	}

EXIT_LABEL:

	return rc;

} /* lin_tape_unconfig_drive() */


/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_uncfg_file_tp_rm                                             *
*                                                                              *
*  PURPOSE: remove file-attributes from sysfs and perform other cleanup        *
*                                                                              *
*    sdev: the sysfs scsi_device object we are removing                        *
*    dvc: the sysfs device object we are removing                              *
*    idx: the index within the driver device table that points to this device  *
*    flags: a bitmap of attributes that need to be deleted                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_uncfg_file_tp_rm(struct device** dvc, int idx, ulong flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	if(flags & CFG_SYSFS_SN)
		device_remove_file(*dvc, &dev_attr_serial_num);
	if(flags & CFG_SYSFS_WWNN)
		device_remove_file(*dvc, &dev_attr_ww_node_name);
	if(flags & CFG_SYSFS_WWPN)
		device_remove_file(*dvc, &dev_attr_ww_port_name);
	if(flags & CFG_SYSFS_PATH)
		device_remove_file(*dvc, &dev_attr_primary_path);
	if(flags & CFG_SYSFS_EPXY)
		device_remove_file(*dvc, &dev_attr_sys_encryption_proxy);
	if(flags & CFG_SYSFS_EWRT)
		device_remove_file(*dvc, &dev_attr_sys_encryption_write);
#endif
	device_destroy(lin_tape_class, MKDEV(lin_tape_major, idx));
	*dvc = NULL;
} /* lin_tape_uncfg_file_tp_rm() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_unconfig_changer                                             *
*                                                                              *
*  PURPOSE: remove a changer from the driver device table                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dvc: the sysfs device object we are removing                              *
*    sdev: the sysfs scsi_device object we are removing                        *
*    idx: the index within the driver device table that points to this device  *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_unconfig_changer(struct device* dvc, struct scsi_device* sdev,
	int idx)
{
	int rc = STATUS_SUCCESS, k=0;
	wait_queue_head_t wq;
	init_waitqueue_head(&wq);

	if (!dvc) goto EXIT_LABEL;

	DbgPrint(("lin_tape: IBMchanger%d alt_pathing %d\n", idx,
			changer_devs[idx]->chg->alt_pathing));

    WRITE_LOCK(&changer_devs[idx]->entry_lock);
	while( changer_busy(changer_devs[idx]->chg, F_ADD_REMOVE)
            && k < 30) {
		printk("lin_tape: Unconf IBMchanger%d busy %d. Waiting ten seconds\n",
         idx, changer_devs[idx]->chg->busy);
        WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
		wait_event_interruptible_timeout(wq,
			!changer_devs[idx] || !changer_devs[idx]->chg, HZ * 100);
        WRITE_LOCK(&changer_devs[idx]->entry_lock);
		k++;
	} /* while */

    while(changer_devs[idx]->chg->sdev_ct > 0){
        lin_tape_sdev_put(changer_devs[idx]->chg->dev_obj,
                         &changer_devs[idx]->chg->sdev_ct);
    }
	changer_devs[idx]->chg->dev_obj = NULL;
    WRITE_UNLOCK(&changer_devs[idx]->entry_lock);

    if(changer_devs[idx]->chg->chg_class_dev) {
		lin_tape_uncfg_file_cg_rm(idx,
				changer_devs[idx]->chg->chgflags);
		changer_devs[idx]->chg->chgflags = 0;
	} /* if */

EXIT_LABEL:

	return rc;

} /* lin_tape_unconfig_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_uncfg_file_cg_rm                                             *
*                                                                              *
*  PURPOSE: remove file-attributes from sysfs and perform other cleanup        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: the sysfs scsi_device object we are removing                        *
*    idx: the index within the driver device table that points to this device  *
*    flags: determines which device attributes to delete                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_uncfg_file_cg_rm(int idx, ulong flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	if(flags & CFG_SYSFS_SN)
		device_remove_file(changer_devs[idx]->chg->chg_class_dev,
			&dev_attr_serial_num);
	if(flags & CFG_SYSFS_PATH)
		device_remove_file(changer_devs[idx]->chg->chg_class_dev,
			&dev_attr_primary_path);
#endif
	device_destroy(lin_tape_class, MKDEV(lin_tape_major,
		idx + LinTapeChangerOffset));
	changer_devs[idx]->chg->chg_class_dev = NULL;
} /* lin_tape_uncfg_file_cg_rm() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_cleanup                                                *
*                                                                              *
*  PURPOSE: clean up drive table entry                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*   dev_index: the device index of the tape drive we are removing              *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_drive_cleanup(struct tape_descriptor *drv)
{
	DbgPrint(("lin_tape_drive_cleanup entry: IBMtape%d alt_pathing %d\n", drv->dev_index,
		drv->alt_pathing));

	WRITE_LOCK(&drv->full_trc_lock);
	if(drv->full_trc_head) {
		struct tape_full_trace_unit *tmp;

		while(drv->full_trc_head) {
			tmp = drv->full_trc_head;
			drv->full_trc_head =
				drv->full_trc_head->next;
			kfree(tmp);
		} /* while */
	} /* if */

	WRITE_UNLOCK(&drv->full_trc_lock);

	if(drv->cdb != NULL) {
		kfree((char *)drv->cdb);
		drv->cdb = NULL;
	} /* if */

	if(drv->sense_data != NULL) {
		kfree(drv->sense_data);
		drv->sense_data = NULL;
	} /* if */

	if(drv->tape_log) kfree(drv->tape_log);

	drv->tape_log = NULL;
	drv->tape_log_len = 0;

	if((drv->dmp) && (drv->dmp_size_order >= 0))
		free_pages((ulong)drv->dmp,
			drv->dmp_size_order);

	drv->dmp = NULL;
	drv->dmp_len = 0;
	drv->dmp_size_order = 0;
	drv->op_sg = 0;

	if(drv->logsense) kfree(drv->logsense);

	drv->logsense = NULL;
	drv->logsense_len = 0;
	drv->doing = 0;

	if(drv->simmim) kfree(drv->simmim);

	drv->simmim = NULL;
	drv->simmim_len = 0;

	DbgPrint(("lin_tape_drive_cleanup exit: void\n"));
	return;
} /* lin_tape_drive_cleanup() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_changer_cleanup                                              *
*                                                                              *
*  PURPOSE: clean up changer table entry                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*   dev_index: the device index of the changer we are removing                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_changer_cleanup(int dev_index)
{
	int i = dev_index;

	DbgPrint(("lin_tape_changer_cleanup entry: IBMchanger%d "\
		"alt_pathing %d\n", i, changer_devs[i]->chg->alt_pathing));

	WRITE_LOCK(&changer_devs[i]->chg->full_trc_lock);
	if(changer_devs[i]->chg->full_trc_head) {
		struct tape_full_trace_unit *tmp;

		while(changer_devs[i]->chg->full_trc_head) {
			tmp = changer_devs[i]->chg->full_trc_head;
			changer_devs[i]->chg->full_trc_head =
				changer_devs[i]->chg->full_trc_head->next;
			kfree(tmp);
			tmp = NULL;
		} /* while */
	} /* if */

	WRITE_UNLOCK(&changer_devs[i]->chg->full_trc_lock);

	WRITE_LOCK(&changer_devs[i]->entry_lock);
	if(changer_devs[i]->chg->cdb != NULL) {
		kfree((char*)changer_devs[i]->chg->cdb);
		changer_devs[i]->chg->cdb = NULL;
	} /* if */

	if(changer_devs[i]->chg->sense_data != NULL) {
		kfree(changer_devs[i]->chg->sense_data);
		changer_devs[i]->chg->sense_data = NULL;
	} /* if */

	if(changer_devs[i]->chg->chg_log){
		kfree(changer_devs[i]->chg->chg_log);
		changer_devs[i]->chg->chg_log = NULL;
	} /* if */

	changer_devs[i]->chg->chg_log = NULL;
	changer_devs[i]->chg->chg_log_len = 0;
	changer_devs[i]->chg->doing = 0;

	WRITE_UNLOCK(&changer_devs[i]->entry_lock);

	DbgPrint(("lin_tape_changer_cleanup exit: void\n"));
	return;
} /* lin_tape_changer_cleanup() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_create_tsm_pseudo                                            *
*                                                                              *
*  PURPOSE: create file for use with TSM IOCTLs                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*   sdev: the SCSI device to be used for sysfs device creation (arbitrary)     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_create_tsm_pseudo(struct scsi_device* sdev)
{
	int rc = STATUS_SUCCESS;

	rc = lin_tape_create_cdev(&lin_tape_tsm_dev, TSM_MINOR, 0,
		"/dev/TSMtape");
	if(rc != STATUS_SUCCESS) {
		DbgPrint((": could not create TSM pseudo device: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	tsm_dev = kmalloc(sizeof(struct tape_descriptor), GFP_KERNEL);
	if(!tsm_dev) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(tsm_dev, '\0', sizeof(struct tape_descriptor));
	tsm_dev->rew_class_dev
		= device_create(lin_tape_class, &sdev->sdev_gendev,
			MKDEV(lin_tape_major, TSM_MINOR), NULL,
			"TSMtape");
	if(!tsm_dev->rew_class_dev) {
		printk("lin_tape: cannot create device TSMTape\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;

} /* lin_tape_create_tsm_pseudo() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_delete_tsm_pseudo                                            *
*                                                                              *
*  PURPOSE: delete file for use with TSM IOCTLs                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_delete_tsm_pseudo()
{

	if(tsm_dev) {
		device_destroy(lin_tape_class, MKDEV(lin_tape_major,
			TSM_MINOR));
		kfree(tsm_dev);
		tsm_dev = NULL;
	} /* if */

	if(lin_tape_tsm_dev) {
		cdev_del(lin_tape_tsm_dev);
		lin_tape_tsm_dev = NULL;
	} /* if */

	return;
} /* lin_tape_delete_tsm_pseudo() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_create_special_dev                                           *
*                                                                              *
*  PURPOSE: create file for use with preemption IOCTLs                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*   sdev: the SCSI device to be used for sysfs device creation (arbitrary)     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_create_special_dev (struct scsi_device* sdev)
{
	int rc = STATUS_SUCCESS;

	rc = lin_tape_create_cdev(&lin_tape_special_dev, SPECIAL_MINOR, 0,
		"/dev/IBMSpecial");
	if(rc != STATUS_SUCCESS) {
		DbgPrint((": could not create special device: %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	special_dev = kmalloc(sizeof(struct tape_descriptor), GFP_KERNEL);
	if(!special_dev) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(special_dev, '\0', sizeof(struct tape_descriptor));
	special_dev->rew_class_dev
		= device_create(lin_tape_class, &sdev->sdev_gendev,
			MKDEV(lin_tape_major, SPECIAL_MINOR), NULL,
			"IBMSpecial");

	if(!special_dev->rew_class_dev) {
		printk("lin_tape: cannot create device IBMSpecial\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;

} /* lin_tape_create_special_dev() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_delete_special_dev                                           *
*                                                                              *
*  PURPOSE: delete file for preemption use                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_delete_special_dev()
{

	if(special_dev) {
		device_destroy(lin_tape_class, MKDEV(lin_tape_major,
			SPECIAL_MINOR));
		kfree(special_dev);
		special_dev = NULL;
	} /* if */

	if(lin_tape_special_dev) {
		cdev_del(lin_tape_special_dev);
		lin_tape_special_dev = NULL;
	} /* if */

	return;
} /* lin_tape_delete_special_dev() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_create_proc                                                  *
*                                                                              *
*  PURPOSE: create /proc/scsi/IBMtape and IBMchanger entries                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_create_proc()
{
	DbgPrint(("lin_tape: lin_tape_create_proc entry\n"));

	WRITE_LOCK(&lin_tape_lock);
	if(!tape_proc_entry) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		tape_proc_entry =
			create_proc_entry("scsi/IBMtape", 0, NULL);
#else
	    tape_proc_entry =
			proc_create("scsi/IBMtape", 0, NULL, &tape_proc_ops);
#endif

		if(!tape_proc_entry) {
			DbgPrint(("lin_tape: cannot initialize "\
				"/proc/scsi/IBMtape\n"));
			goto EXIT_LABEL;
		} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		tape_proc_entry->read_proc = tape_read_proc;
#endif
	} /* if */

	if(!changer_proc_entry) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		 changer_proc_entry =
		         create_proc_entry("scsi/IBMchanger", 0, NULL);
#else
	     changer_proc_entry =
		         proc_create("scsi/IBMchanger", 0, NULL, &changer_proc_ops);
#endif
		if(!changer_proc_entry) {
			DbgPrint(("lin_tape: cannot initialize "\
				"/proc/scsi/IBMchanger\n"));
			goto EXIT_LABEL;
		} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		changer_proc_entry->read_proc = changer_read_proc;
#endif
	} /* if */

	if(!status_proc_entry && lin_tape_status) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		status_proc_entry =
			create_proc_entry("scsi/lin_tape", 0, NULL);
#else
	    status_proc_entry =
			proc_create("scsi/lin_tape", 0, NULL, &status_proc_ops);
#endif

		if(!status_proc_entry) {
			DbgPrint(("lin_tape: cannot initialize "\
				"/proc/scsi/lin_tape\n"));
			goto EXIT_LABEL;
		} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		status_proc_entry->read_proc = status_read_proc;
#endif
	} /* if */


EXIT_LABEL:

	WRITE_UNLOCK(&lin_tape_lock);

	DbgPrint(("lin_tape: lin_tape_create_proc exit: void\n"));
	return;
} /* lin_tape_create_proc() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_remove_proc                                                  *
*                                                                              *
*  PURPOSE: remove /proc/scsi/IBMtape and IBMchanger entries                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_remove_proc()
{
	WRITE_LOCK(&lin_tape_lock);

	if(tape_proc_entry) {
		remove_proc_entry("scsi/IBMtape", NULL);
		tape_proc_entry = NULL;
	} /* if */

	if(changer_proc_entry) {
		remove_proc_entry("scsi/IBMchanger", NULL);
		changer_proc_entry = NULL;
	} /* if */

    if(status_proc_entry) {
		remove_proc_entry("scsi/lin_tape", NULL);
		status_proc_entry = NULL;
	} /* if */

	WRITE_UNLOCK(&lin_tape_lock);

	return;
} /* lin_tape_remove_proc() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_read_proc                                                        *
*                                                                              *
*  PURPOSE: Display /proc/scsi/IBMtape                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    buffer: buffer into which to copy the data                                *
*    start: for this function, just points to same location as buffer          *
*    offset: file offset to read                                               *
*    length: amount of data to read                                            *
*    eof: set by the driver to indicate no further data to be read             *
*    data: driver-specific parameter (unused by this driver)                   *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/
static int tape_read_proc
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
                         (char* buffer, char** start, off_t offset, int length,
                        		                          int* eof, void* data)
#else
                         (struct seq_file *bufferfile, void *data)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	int buf_offset = 0, linelen = 0;
#endif
	int i = 0, start_dev = 0, found = 0;
	unchar* model = NULL, * pre_buf = NULL;
	unchar temp_buf[96];
	char* sn = NULL;
	boolean tape_des_locked = False;

	DbgPrint(("lin_tape: tape_read_proc entry\n"));

	model = kmalloc(12, GFP_KERNEL);
	sn = kmalloc(MaxSerialNumberSize + 1, GFP_KERNEL);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	pre_buf = kmalloc(ONE_K / 4, GFP_KERNEL);
	if(!model || !pre_buf || !sn || !buffer || !start || !eof) {
		    if(start) *start = buffer;
		    if(eof) *eof = True;
		goto EXIT_LABEL;
	} /* if */
	memset(pre_buf, '\0', ONE_K / 4);
#else
    if (!model || !sn || !bufferfile)
		goto EXIT_LABEL;
#endif

	memset(temp_buf, '\0', 96);
	memset(model, '\0', 12);
	memset(sn, '\0', MaxSerialNumberSize + 1);

	i = 0;
	WRITE_LOCK(&tape_des_lock);
	tape_des_locked = True;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(!offset && rd_pc_initial(buffer, &buf_offset, length, start, False))
		goto EXIT_LABEL;
	if(rd_pc_start_dev(&start_dev, eof, offset, buffer, &buf_offset, False))
		goto EXIT_LABEL;
#else
	if (rd_pc_initial(bufferfile, False))
		goto EXIT_LABEL;
#endif

	if(!tape_devs) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		*eof = True;
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*start = buffer;
#else
       seq_printf(bufferfile, "\n");
#endif
		goto EXIT_LABEL;
	} /* if */

	for(i = start_dev; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;
		found = False;

		if(tape_devs[i]->drv->dev_index == i) found = True;

		if(found && tape_devs[i]->drv->proc_read) continue;

		if (!tape_devs[i]->drv->ghost_device){
			memcpy(model, tape_devs[i]->drv->dev_obj->model, 12);
            model[11] = '\0';
		}
        else {model[0] = '\0';}

		memcpy(sn, tape_devs[i]->drv->serial_num, MaxSerialNumberSize);
		while(strlen(sn) &&
			(sn[strlen(sn) - 1] == ' ' ||
			sn[strlen(sn) - 1] == '\n' ||
			sn[strlen(sn) - 1] == '\t'))
			sn[strlen(sn) - 1] = '\0';
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		linelen = snprintf(pre_buf, ONE_K / 4,
#else
        seq_printf(bufferfile,
#endif
			"\n%-8d%-12s%-25s%-26s%-16s%-10s",
			i,
			model,

			tape_devs[i]->drv->ghost_device ? "REMOVED" :
			tape_devs[i]->drv->serial_num_len ? sn : "NA",

			tape_devs[i]->drv->ghost_device ? "" :
			found ? tape_devs[i]->drv->pname : "NA",

			scsi_addr_str(tape_devs[i]->drv, temp_buf,
				False),

			tape_devs[i]->drv->ghost_device ? ""       :
			!found ? "NA"                         :
			!tape_devs[i]->drv->serial_num_len ? "NA"  :
			!tape_devs[i]->drv->alt_pathing    ? "NA"  :
			tape_devs[i]->drv->fenced ? "Disabled"     :
			tape_devs[i]->drv->isPrimary ? "Primary"   :
			"Alternate");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset > length) {
			buf_offset = length;
			break;
		} else if(buf_offset + linelen >= length) {
			memset(&buffer[buf_offset], ' ', length - buf_offset);
			buf_offset = length;
			break;
		} /* if */
		buf_offset += sprintf(&buffer[buf_offset], "%s", pre_buf);
#endif
		tape_devs[i]->drv->proc_read = True;
	} /* for */

	tape_des_locked = False;
	WRITE_UNLOCK(&tape_des_lock);

	if(i == start_dev || i >= MaxLinTapeDevices) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*eof = True;
		if(i >= MaxLinTapeDevices) *start = buffer;
#else
		seq_printf(bufferfile, "\n");
#endif
		goto EXIT_LABEL;
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*start = buffer;
#endif

EXIT_LABEL:

	if(pre_buf) kfree(pre_buf);
	if(model) kfree(model);
	if(sn) kfree(sn);

	if(tape_des_locked)
	{
		WRITE_UNLOCK(&tape_des_lock);
		tape_des_locked = False;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(buf_offset == 1 && *eof) buf_offset = 0;
	DbgPrint(("tape_read_proc exit: %d\n", buf_offset));
	return buf_offset;
#else
	DbgPrint(("tape_read_proc exit\n"));
	return 0;
#endif
} /* tape_read_proc() */


/*******************************************************************************
*                                                                              *
*  NAME: changer_read_proc                                                     *
*                                                                              *
*  PURPOSE: Display /proc/scsi/IBMchanger                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    buffer: buffer into which to copy the data                                *
*    start: for this function, just points to same location as buffer          *
*    offset: file offset to read                                               *
*    length: amount of data to read                                            *
*    eof: set by the driver to indicate no further data to be read             *
*    data: driver-specific parameter (unused by this driver)                   *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/
static int changer_read_proc
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
                            (char* buffer, char** start, off_t offset,
				             int length, int* eof, void* data)
#else
                            (struct seq_file *bufferfile, void *data)
#endif
{
	int i = 0, start_dev = 0, found = 0;
	unchar* model = NULL, temp_buf[96];
	char* sn = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	int buf_offset = 0, linelen = 0;
	unchar * pre_buf = NULL;
#endif

	boolean changer_des_locked = False;

	DbgPrint(("lin_tape: changer_read_proc entry\n"));

	i = 0;

	model = kmalloc(12, GFP_KERNEL);
	sn = kmalloc(MaxSerialNumberSize + 1, GFP_KERNEL);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	pre_buf = kmalloc(ONE_K / 4, GFP_KERNEL);
	if(!model || !pre_buf || !sn || !buffer || !start || !eof) {
		if(start) *start = buffer;
		if(eof) *eof = True;
		goto EXIT_LABEL;
	} /* if */
	memset(pre_buf, '\0', ONE_K / 4);
#else
	if (!model || !sn || !bufferfile)
		goto EXIT_LABEL;
#endif
	memset(temp_buf, '\0', 96);
	memset(model, '\0', 12);
	memset(sn, '\0', MaxSerialNumberSize + 1);

	WRITE_LOCK(&changer_des_lock);
	changer_des_locked = True;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(!offset && rd_pc_initial(buffer, &buf_offset, length, start, True))
		goto EXIT_LABEL;
	if(rd_pc_start_dev(&start_dev, eof, offset, buffer, &buf_offset, True))
		goto EXIT_LABEL;
#else
	if (rd_pc_initial(bufferfile, True))
		goto EXIT_LABEL;
#endif

	if(!changer_devs) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		*eof = True;
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*start = buffer;
#endif
		goto EXIT_LABEL;
	} /* if */

	for(i = start_dev; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg) continue;
		found = False;

		if(changer_devs[i]->chg->dev_index == i) found = True;

		if(found && changer_devs[i]->chg->proc_read) continue;

		if (!changer_devs[i]->chg->ghost_device){
            memcpy(model, changer_devs[i]->chg->dev_obj->model, 12);
            model[11] = '\0';
		}
		else model[0] = '\0';

		memcpy(sn, changer_devs[i]->chg->serial_num, MaxSerialNumberSize);
		while(strlen(sn) &&
			(sn[strlen(sn) - 1] == ' ' ||
			sn[strlen(sn) - 1] == '\n' ||
			sn[strlen(sn) - 1] == '\t'))
			sn[strlen(sn) - 1] = '\0';
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		linelen = snprintf(pre_buf, ONE_K / 4,
#else
        seq_printf(bufferfile,
#endif
                               "\n%-8d%-12s%-25s%-26s%-16s%-10s",
			i,
			model,

			changer_devs[i]->chg->ghost_device ? "REMOVED" :
			changer_devs[i]->chg->serial_num_len ? sn : "NA",

			changer_devs[i]->chg->ghost_device ? "" :
			found ? changer_devs[i]->chg->pname : "NA",

			scsi_addr_str((struct tape_descriptor*)
				changer_devs[i]->chg,
				temp_buf,
				True),

			changer_devs[i]->chg->ghost_device                 ? ""   :
			!found                                        ? "NA" :
			changer_devs[i]->chg->chg_id == IBM_3590
				|| changer_devs[i]->chg->chg_id == IBM_3592 ? "NA" :
			!changer_devs[i]->chg->alt_pathing                  ? "NA" :
			changer_devs[i]->chg->fenced               ? "Disabled"    :
			changer_devs[i]->chg->isPrimary            ? "Primary"     :
			"Alternate");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset > length) {
			buf_offset = length;
			break;
		} else if(buf_offset + linelen >= length) {
			memset(&buffer[buf_offset], ' ', length - buf_offset);
			buf_offset = length;
			break;
		} /* if */

		buf_offset += sprintf(&buffer[buf_offset], "%s", pre_buf);
#endif
		changer_devs[i]->chg->proc_read = True;
	} /* for */

	changer_des_locked = False;
	WRITE_UNLOCK(&changer_des_lock);

	if(i == start_dev || i >= MaxLinTapeChangerDevices) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*eof = True;
		if(i >= MaxLinTapeChangerDevices) *start = buffer;
#else
        seq_printf(bufferfile, "\n");
#endif
		goto EXIT_LABEL;
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*start = buffer;
#endif

EXIT_LABEL:

	if(model) kfree(model);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(pre_buf) kfree(pre_buf);
#endif
	if(sn) kfree(sn);

	if(changer_des_locked) WRITE_UNLOCK(&changer_des_lock);
	changer_des_locked = False;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(buf_offset == 1 && *eof) buf_offset = 0;
	DbgPrint(("changer_read_proc exit: %d\n", buf_offset));

	return buf_offset;
#else
	DbgPrint(("changer_read_proc exit\n"));
	return 0;
#endif
} /* changer_read_proc() */

/*******************************************************************************
*                                                                              *
*  NAME: status_read_proc                                                      *
*                                                                              *
*  PURPOSE: Display /proc/scsi/lin_tape                                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    buffer: buffer into which to copy the data                                *
*    start: for this function, just points to same location as buffer          *
*    offset: file offset to read                                               *
*    length: amount of data to read                                            *
*    eof: set by the driver to indicate no further data to be read             *
*    data: driver-specific parameter (unused by this driver)                   *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/
static int status_read_proc
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
                         (char* buffer, char** start, off_t offset, int length,
                        		                          int* eof, void* data)
#else
                         (struct seq_file *bufferfile, void *data)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	int buf_offset = 0, linelen = 0;
#endif
	int i = 0, start_dev = 0, found = 0;
	unchar  * pre_buf = NULL;
	unchar temp_buf[96];
    boolean tape_des_locked = False;

	DbgPrint(("lin_tape: status_read_proc entry\n"));

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	pre_buf = kmalloc(ONE_K / 4, GFP_KERNEL);
	if(!pre_buf || !buffer || !start || !eof) {
		    if(start) *start = buffer;
		    if(eof) *eof = True;
		goto EXIT_LABEL;
	} /* if */
	memset(pre_buf, '\0', ONE_K / 4);
#else
    if (!bufferfile)
		goto EXIT_LABEL;
#endif

	memset(temp_buf, '\0', 96);

	i = 0;

//	WRITE_LOCK(&tape_des_lock);
//	tape_des_locked = True;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(!offset && rdstatus_pc_initial(buffer, &buf_offset, length, start, False))
		goto EXIT_LABEL;
	if(rdstatus_pc_start_dev(&start_dev, eof, offset, buffer, &buf_offset, False))
		goto EXIT_LABEL;
#else
	if (rdstatus_pc_initial(bufferfile, False))
		goto EXIT_LABEL;
#endif

	if(!tape_devs) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		*eof = True;
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*start = buffer;
#else
       seq_printf(bufferfile, "\n");
#endif
		goto EXIT_LABEL;
	} /* if */

	for(i = start_dev; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;

		found = False;

		if(tape_devs[i]->drv->dev_index == i) found = True;
        else printk("READ PROC i %d diff from %d devindex\n",i,tape_devs[i]->drv->dev_index);
		if(found && tape_devs[i]->drv->proc_read) continue;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		linelen = snprintf(pre_buf, ONE_K / 4,
#else
        seq_printf(bufferfile,
#endif
			"\n%-8d%-8d%-8d%-8s%-8s%-8d%-8d%-8d%-8d",
			tape_devs[i]->drv->dev_index,
            tape_devs[i]->drv->busy,
			tape_devs[i]->drv->ready_wakeup,
			tape_devs[i]->drv->ghost_device ? "Yes" :"No",
			tape_devs[i]->drv->remove_pending? "Yes":"No",
			tape_devs[i]->drv->cur_path,
			tape_devs[i]->drv->cmd_in_open,
			tape_devs[i]->drv->drv_open,
			tape_devs[i]->drv->curr_reserve_type);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset > length) {
			buf_offset = length;
			break;
		} else if(buf_offset + linelen >= length) {
			memset(&buffer[buf_offset], ' ', length - buf_offset);
			buf_offset = length;
			break;
		} /* if */
		buf_offset += sprintf(&buffer[buf_offset], "%s", pre_buf);
#endif
		tape_devs[i]->drv->proc_read = True;
	} /* for */

//	tape_des_locked = False;
//	WRITE_UNLOCK(&tape_des_lock);

	if(i == start_dev || i >= MaxLinTapeDevices) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
		if(buf_offset == length) buf_offset--;
		buf_offset += sprintf(&buffer[buf_offset], "\n");
		*eof = True;
		if(i >= MaxLinTapeDevices) *start = buffer;
#else
		seq_printf(bufferfile, "\n");
#endif
		goto EXIT_LABEL;
	} /* if */


#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*start = buffer;
#endif

EXIT_LABEL:

	if(pre_buf) kfree(pre_buf);

	if(tape_des_locked)
	{
		WRITE_UNLOCK(&tape_des_lock);
		tape_des_locked = False;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(buf_offset == 1 && *eof) buf_offset = 0;
	DbgPrint(("status_read_proc exit: %d\n", buf_offset));
	return buf_offset;
#else
	DbgPrint(("status_read_proc exit\n"));
	return 0;
#endif
} /* status_read_proc() */

/*******************************************************************************
*                                                                              *
*  NAME: rd_pc_initial                                                         *
*                                                                              *
*  PURPOSE: Retrieval of data for /proc/scsi/IBM* beginning at offset 0        *
*                                                                              *
*    buffer: into which to copy the data                                       *
*    buf_offset: current pointer to end of data                                *
*    length: amount to read, as specified by the call to tape_read_proc        *
*    start: points to buffer                                                   *
*    chg: this call is intended for the medium changers                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int rd_pc_initial
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
                 (char* buffer, int* buf_offset, int length, char** start,
	              boolean chg)
#else
                 (struct seq_file *bufferfile, boolean chg)
#endif
{
	int rc = 1, i = 0, max_devices = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	char tmp_buf[96];
	int count = 0;
#endif
	DbgPrint(("rd_pc_initial entry\n"));

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	memset(tmp_buf, '\0', 96);
#else
	if (!bufferfile) goto EXIT_LABEL;
#endif
	max_devices = chg ? MaxLinTapeChangerDevices : MaxLinTapeDevices;

	/* set proc_read flag for each logical device to false */
	if(chg ? changer_devs != NULL : tape_devs != NULL) {
		for(i = 0; i < max_devices; i++) {
			if(chg && changer_devs[i] && changer_devs[i]->chg)
				changer_devs[i]->chg->proc_read = False;
			else if(!chg && tape_devs[i] && tape_devs[i]->drv)
				tape_devs[i]->drv->proc_read = False;
		} /* for */
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	count = sprintf(tmp_buf, "lin_tape version: %s", LIN_TAPE_VERSION);
	if(count > length) goto EXIT_LABEL;
	count = sprintf(buffer,
#else
    seq_printf(bufferfile,
#endif
    		               "lin_tape version: %s", LIN_TAPE_VERSION);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset = count;
	count = sprintf(tmp_buf,"\nlin_tape major number: %d", lin_tape_major);
	if(*buf_offset + count > length) goto EXIT_LABEL;
	count = sprintf(buffer + *buf_offset,
#else
	seq_printf(bufferfile,
#endif
			              "\nlin_tape major number: %d", lin_tape_major);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;

	count = sprintf(tmp_buf, "\nAttached %s Devices:",
		    chg ? "Changer" : "Tape");
	if(*buf_offset + count > length) goto EXIT_LABEL;

	count = sprintf(buffer + *buf_offset,
#else
	seq_printf(bufferfile,
#endif
			                     "\nAttached %s Devices:",
                                 chg ? "Changer" : "Tape");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;

	count = sprintf(tmp_buf, "\n%-8s%-12s%-25s%-26s%-16s%-10s",
			"Number", "model", "SN", "HBA", "SCSI", "FO Path");
	if(*buf_offset + count > length) goto EXIT_LABEL;

	count = sprintf(buffer + *buf_offset,
#else
    seq_printf(bufferfile,
#endif
    		                  "\n%-8s%-12s%-25s%-26s%-16s%-10s",
                              "Number", "model", "SN", "HBA", "SCSI", "FO Path");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;
#endif
	rc = 0;

EXIT_LABEL:

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(rc) {
		memset(buffer + *buf_offset, ' ', length - *buf_offset);
		*buf_offset = length;
		*start = buffer;
	} /* if */
#endif
	DbgPrint(("rd_pc_initial  exit: %d\n", rc));
	return rc;
} /* rd_pc_initial() */

/*******************************************************************************
*                                                                              *
*  NAME: rd_pc_start_dev                                                       *
*                                                                              *
*  PURPOSE: Return the next tape device to be printed as a part of read_proc   *
*                                                                              *
*    start_dev: (output) the device to print                                   *
*    eof: set if there are no more devices to read                             *
*    offset: Offset at which to begin the read                                 *
*    buffer: the buffer to fill and be printed by read_proc                    *
*    buf_offset: char-index of the end of data in buffer                       *
*    chg: this call is intended for the medium changers                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: 0 if there is more to print, 1 if we have reached end-of-data      *
*                                                                              *
*******************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
int rd_pc_start_dev(int* start_dev, int* eof, off_t offset, char* buffer,
                    int* buf_offset, boolean chg)
{
	int found = False, rc = 0, i = 0, max_devices = 0;

	DbgPrint(("rd_pc_start_dev entry\n"));

	max_devices = chg ? MaxLinTapeChangerDevices : MaxLinTapeDevices;

	if(!offset) {
		*start_dev = 0;
		goto EXIT_LABEL;
	} /* if */

	*start_dev = max_devices;

	if(chg ? changer_devs != NULL : tape_devs != NULL) {
		i = -1;
		while(!found && ++i < max_devices) {
			if(chg) {
				if(!changer_devs[i] || !changer_devs[i]->chg) continue;
				if(!changer_devs[i]->chg->proc_read) found = True;
			} else {
				if(!tape_devs[i] || !tape_devs[i]->drv) continue;
				if(!tape_devs[i]->drv->proc_read) found = True;
			} /* else */
		} /* while */

		if(found) *start_dev = chg ?
			changer_devs[i]->chg->dev_index : tape_devs[i]->drv->dev_index;
	} /* if */

	if(*start_dev >= max_devices) {
		*eof = True;
		*buf_offset += sprintf(&buffer[*buf_offset], "\n");
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("rd_pc_start_dev exit: %d\n", rc));
	return rc;
} /* rd_pc_start_dev() */
#endif

/*******************************************************************************
*                                                                              *
*  NAME: scsi_addr_str                                                         *
*                                                                              *
*  PURPOSE: initialize a tape descriptor                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the tape descriptor for which to get the SCSI address                *
*    buffer: the character buffer to copy into                                 *
*    cgr: if True, this is a changer device, if False, it is a tape device     *
*                                                                              *
*  RETURNS: buffer                                                             *
*                                                                              *
*******************************************************************************/
char* scsi_addr_str(struct tape_descriptor* drv, char* buffer, boolean cgr)
{
	struct changer_descriptor* chg = (struct changer_descriptor*)drv;

	if(cgr)
		sprintf(buffer, "%d:%d:%d:%d",
			chg->host, chg->channel, chg->target, chg->lun);
	else
		sprintf(buffer, "%d:%d:%d:%d",
			drv->host, drv->channel, drv->target, drv->lun);
	return buffer;
} /* scsi_addr_str() */

/*******************************************************************************
*                                                                              *
*  NAME: rdstatus_pc_initial                                                   *
*                                                                              *
*  PURPOSE: Retrieval of data for /proc/scsi/lin_tape beginning at offset 0    *
*                                                                              *
*    buffer: into which to copy the data                                       *
*    buf_offset: current pointer to end of data                                *
*    length: amount to read, as specified by the call to status_read_proc      *
*    start: points to buffer                                                   *
*    chg: this call is intended for the medium changers                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int rdstatus_pc_initial
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
                 (char* buffer, int* buf_offset, int length, char** start,
	              boolean chg)
#else
                 (struct seq_file *bufferfile, boolean chg)
#endif
{
	int rc = 1, i = 0, max_devices = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	char tmp_buf[96];
	int count = 0;
#endif
	DbgPrint(("status rd_pc_initial entry\n"));

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	memset(tmp_buf, '\0', 96);
#else
	if (!bufferfile) goto EXIT_LABEL;
#endif
	max_devices = chg ? MaxLinTapeChangerDevices : MaxLinTapeDevices;

	/* set proc_read flag for each logical device to false */
	if(chg ? changer_devs != NULL : tape_devs != NULL) {
		for(i = 0; i < max_devices; i++) {
			if(chg && changer_devs[i] && changer_devs[i]->chg)
				changer_devs[i]->chg->proc_read = False;
			else if(!chg && tape_devs[i] && tape_devs[i]->drv)
				tape_devs[i]->drv->proc_read = False;
		} /* for */
	} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	count = sprintf(tmp_buf, "lin_tape version: %s", LIN_TAPE_VERSION);
	if(count > length) goto EXIT_LABEL;
	count = sprintf(buffer,
#else
    seq_printf(bufferfile,
#endif
    		               "lin_tape version: %s", LIN_TAPE_VERSION);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset = count;
	count = sprintf(tmp_buf,"\nlin_tape major number: %d", lin_tape_major);
	if(*buf_offset + count > length) goto EXIT_LABEL;
	count = sprintf(buffer + *buf_offset,
#else
	seq_printf(bufferfile,
#endif
			              "\nlin_tape major number: %d", lin_tape_major);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;

	count = sprintf(tmp_buf, "\nAttached %s Devices:",
		    chg ? "Changer" : "Tape");
	if(*buf_offset + count > length) goto EXIT_LABEL;

	count = sprintf(buffer + *buf_offset,
#else
	seq_printf(bufferfile,
#endif
			                     "\nAttached %s Devices:",
                                 chg ? "Changer" : "Tape");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;

	count = sprintf(tmp_buf, "\n%-8s%-8s%-8s%-8s%-8s%-8s%-8s%-8s%-8s",
			"Number", "Busy", "WakeUp", "Ghost", "Remove", "CurrPath","InOpen","Open","Reserve");
	if(*buf_offset + count > length) goto EXIT_LABEL;

	count = sprintf(buffer + *buf_offset,
#else
    seq_printf(bufferfile,
#endif
    		                  "\n%-8s%-8s%-8s%-8s%-8s%-8s%-8s%-8s%-8s",
                              "Number", "Busy", "WakeUp", "Ghost", "Remove", "CurrPath","InOpen","Open","Reserve");
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	*buf_offset += count;
#endif
	rc = 0;

EXIT_LABEL:

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
	if(rc) {
		memset(buffer + *buf_offset, ' ', length - *buf_offset);
		*buf_offset = length;
		*start = buffer;
	} /* if */
#endif
	DbgPrint(("status rd_pc_initial  exit: %d\n", rc));
	return rc;
} /* rdstatus_pc_initial() */

/*******************************************************************************
*                                                                              *
*  NAME: rdstatus_pc_start_dev                                                 *
*                                                                              *
*  PURPOSE: Return the next tape device to be printed as a part of status_proc *
*                                                                              *
*    start_dev: (output) the device to print                                   *
*    eof: set if there are no more devices to read                             *
*    offset: Offset at which to begin the read                                 *
*    buffer: the buffer to fill and be printed by status_proc                  *
*    buf_offset: char-index of the end of data in buffer                       *
*    chg: this call is intended for the medium changers                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: 0 if there is more to print, 1 if we have reached end-of-data      *
*                                                                              *
*******************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
int rdstatus_pc_start_dev(int* start_dev, int* eof, off_t offset, char* buffer,
                    int* buf_offset, boolean chg)
{
	int found = False, rc = 0, i = 0, max_devices = 0;

	DbgPrint(("status rd_pc_start_dev entry\n"));

	max_devices = chg ? MaxLinTapeChangerDevices : MaxLinTapeDevices;

	if(!offset) {
		*start_dev = 0;
		goto EXIT_LABEL;
	} /* if */

	*start_dev = max_devices;

	if(chg ? changer_devs != NULL : tape_devs != NULL) {
		i = -1;
		while(!found && ++i < max_devices) {
			if(chg) {
				if(!changer_devs[i]||!changer_devs[i]->chg) continue;
				if(!changer_devs[i]->chg->proc_read) found = True;
			} else {
				if(!tape_devs[i] || !tape_devs[i]->drv) continue;
				if(!tape_devs[i]->drv->proc_read) found = True;
			} /* else */
		} /* while */

		if(found) *start_dev = chg ?
			changer_devs[i]->chg->dev_index : tape_devs[i]->drv->dev_index;
	} /* if */

	if(*start_dev >= max_devices) {
		*eof = True;
		*buf_offset += sprintf(&buffer[*buf_offset], "\n");
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("status rd_pc_start_dev exit: %d\n", rc));
	return rc;
} /* rdstatus_pc_start_dev() */
#endif

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_init_drive_des                                               *
*                                                                              *
*  PURPOSE: initialize a tape descriptor                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the tape descriptor to initialize                                    *
*    sdev: the sysfs scsi_device object associated with the descriptor         *
*    i: the index of the drive in the tape descriptor array                    *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*                                                                              *
*******************************************************************************/

int lin_tape_init_drive_des(struct tape_descriptor* drv,
	struct scsi_device* sdev, int* i, boolean rescan)
{
	int rc = STATUS_SUCCESS;
	boolean rew_class_dvc_regd = False, rew_cdev_creatd = False,
		nonrew_class_dvc_regd = False, nonrew_cdev_creatd = False,
		nonrsv_class_dvc_regd = False, nonrsv_cdev_creatd = False;
	ulong rewflags = CFG_PRIM_DVC, nrewflags = CFG_ALT_DVC,
		nrsvflags = CFG_ALT_DVC;

	if(persistent_n_device)
		nrewflags = CFG_ALT_DVC_N;

	DbgPrint(("lin_tape: lin_tape_init_drive_des entry\n"));
	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	lin_tape_set_initial_drv_values(drv, sdev);
	/************************* rewindable device *************************/

	drv->serial_num_len = 0;
	memset(drv->serial_num, '\0', MaxSerialNumberSize);
	memcpy(drv->serial_num, "NA", 2);
    drv->busy |= F_ADD_REMOVE;
	WRITE_UNLOCK((&tape_devs[drv->dev_index]->drv_entry_lock));
	lin_tape_get_inq_p0x83_data(drv, False);
	rc = check_for_dup_drive(&drv, sdev, i, rescan);
	if (rc == 1) {
        rc = STATUS_SUCCESS;
        goto EXIT_LABEL;
	}
	tape_get_timeouts(drv);
	rc = lin_tape_create_cdev(&drv->rew_dev, drv->dev_index, 0,
		"/dev/IBMtape%d");
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	rew_cdev_creatd = True;
	rc = lin_tape_register_class_device(&drv->rew_class_dev, drv, sdev,
		"IBMtape%d", drv->dev_index, 0, &rewflags);
	rew_class_dvc_regd = True;
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	drv->rewflags = rewflags;

	/*********************** nonrewindable device ************************/

	rc = lin_tape_create_cdev(&drv->nonrew_dev, drv->dev_index,
		LinTapeNonRewindOffset, "/dev/IBMtape%dn");
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	nonrew_cdev_creatd = True;

	rc = lin_tape_register_class_device(&drv->nonrew_class_dev, drv,
		sdev, "IBMtape%dn", drv->dev_index, LinTapeNonRewindOffset,
		&nrewflags);
	nonrew_class_dvc_regd = True;
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	drv->nonrewflags = nrewflags;

	/************************ nonreserve device **************************/

	if(petro == 1) {
		rc = lin_tape_create_cdev(&drv->nonresv_dev,
			drv->dev_index, LinTapeNonReserveOffset,
			"/dev/IBMtape%dnrsv");
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		nonrsv_cdev_creatd = True;

		rc = lin_tape_register_class_device(&drv->nonresv_class_dev,
			drv, sdev, "IBMtape%dnrsv", drv->dev_index,
			LinTapeNonReserveOffset,
			&nrsvflags);
		nonrsv_class_dvc_regd = True;
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		drv->nonrsvflags = nrsvflags;
	} /* if */

EXIT_LABEL:

	if(rc != STATUS_SUCCESS && drv) {
		if(rew_cdev_creatd) cdev_del(drv->rew_dev);
		if(nonrew_cdev_creatd) cdev_del(drv->nonrew_dev);
		if(nonrsv_cdev_creatd) cdev_del(drv->nonresv_dev);

		if(drv && sdev && rew_class_dvc_regd)
			lin_tape_undo_class_dvc(&drv->rew_class_dev, sdev,
				drv->dev_index, 0, rewflags);

		if(drv && sdev && nonrew_class_dvc_regd)
			lin_tape_undo_class_dvc(&drv->nonrew_class_dev, sdev,
				drv->dev_index, LinTapeNonRewindOffset,
				nrewflags);

		if(drv && sdev && nonrsv_class_dvc_regd)
			lin_tape_undo_class_dvc(&drv->nonresv_class_dev,
				sdev, drv->dev_index,
				LinTapeNonReserveOffset, nrsvflags);

		drv->rew_dev = NULL;
		drv->nonrew_dev = NULL;
		drv->nonresv_dev = NULL;

		drv->rew_class_dev = NULL;
		drv->nonrew_class_dev = NULL;
		drv->nonresv_class_dev = NULL;
	} /* if */
	WRITE_LOCK((&tape_devs[drv->dev_index]->drv_entry_lock));
	if (drv){
    drv->busy &= 0xFE;
	}
    WRITE_UNLOCK((&tape_devs[drv->dev_index]->drv_entry_lock));
	DbgPrint(("lin_tape: lin_tape_init_drive_des exit: %d\n", rc));
	return rc;
} /* lin_tape_init_drive_des() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_create_cdev                                                  *
*                                                                              *
*  PURPOSE: create a Linux internal character device                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    cd: (out) the cdev created                                                *
*    idx: the device index number                                              *
*    offset: from idx, into the tape descriptor table                          *
*    name: the /dev name of the device                                         *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_create_cdev(struct cdev** cd, int idx, int offset, char* name)
{
	int rc = STATUS_SUCCESS;
	char abs_name[128] = {0};
	if(strlen(name) < 120) sprintf(abs_name, name, idx);

	DbgPrint(("lin_tape_create_cdev entry: %s\n", abs_name))
	if(!abs_name[0] ) {
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if (*cd != NULL) return STATUS_SUCCESS;

	*cd = cdev_alloc();
	if(!*cd) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	kobject_set_name(&(*cd)->kobj, abs_name);
	(*cd)->owner = THIS_MODULE;
	(*cd)->ops = &lin_tape_ops;

	rc = cdev_add(*cd, MKDEV(lin_tape_major, idx + offset), 1);
	if(rc != STATUS_SUCCESS) {
		kobject_put(&(*cd)->kobj);
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: lin_tape_create_cdev exit: %d\n", rc));
	return rc;
} /* lin_tape_create_cdev() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_register_class_device                                        *
*                                                                              *
*  PURPOSE: register a struct device or a struct class_device with sysfs       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    cd: (out) the class device that this function creates                     *
*    drv: the tape descriptor associated with this device                      *
*    sdev: the scsi device associated with this device                         *
*    name: the name to be associated with the device in /dev                   *
*    idx: the device index, 0, 1, etc.                                         *
*    offset: the offset into the descriptor table for the different dvc types  *
*    flags: dictates which kind of attributes will be created for this device  *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_register_class_device(struct device** cd,
	struct tape_descriptor* drv, struct scsi_device* sdev, char* name,
	int idx, int offset, ulong* flags)
{
	int rc = STATUS_SUCCESS;
	char abs_name[128] = {0};
	ulong nflags = 0;

	if(strlen(name) < 120) sprintf(abs_name, name, idx);
	DbgPrint(("lin_tape_register_class_device entry: %s\n", abs_name));
	nflags = *flags;
	*flags = 0;

	if(lin_tape_class) {
		if(nflags == CFG_PRIM_DVC || nflags == CFG_CHG_DVC)
			dev_set_drvdata(&sdev->sdev_gendev, drv);
		*cd = device_create(lin_tape_class, &sdev->sdev_gendev,
			MKDEV(lin_tape_major, idx + offset), drv, abs_name);
		if(!*cd) {
			printk("lin_tape: cannot create device %s\n", abs_name);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
		if(nflags & CFG_SYSFS_SN) {
			if(!device_create_file(*cd, &dev_attr_serial_num))
				*flags |= CFG_SYSFS_SN;
			else
				printk("lin_tape: cannot create serial number "\
					"attribute\n");
		} /* if */

		if(nflags & CFG_SYSFS_WWNN) {
			if(!device_create_file(*cd, &dev_attr_ww_node_name))
				*flags |= CFG_SYSFS_WWNN;
			else
				printk("lin_tape: cannot create wwnn "\
					"attribute\n");
		} /* if */

		if(nflags & CFG_SYSFS_WWPN) {
			if(!device_create_file(*cd, &dev_attr_ww_port_name))
				*flags |= CFG_SYSFS_WWPN;
			else
				printk("lin_tape: cannot create wwpn "\
					"attribute\n");
		} /* if */

		if(nflags & CFG_SYSFS_PATH) {
			if(!device_create_file(*cd, &dev_attr_primary_path))
				*flags |= CFG_SYSFS_PATH;
			else
				printk("lin_tape: cannot create primary_path "\
					"attribute\n");
		} /* if */
#else
        rc = device_add_groups(*cd,lin_tape_dev_groups);
#endif
	} /* if */

	if(*cd) {
		dev_set_drvdata(*cd, drv);
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
		if(nflags & CFG_SYSFS_EPXY) {
			int tmprc; /* to avoid unused rc warning */
			tmprc = device_create_file(*cd,
				&dev_attr_sys_encryption_proxy);
			*flags |= CFG_SYSFS_EPXY;
		} /* if */

		if(nflags & CFG_SYSFS_EWRT) {
			int tmprc; /* to avoid unused rc warning */
			tmprc = device_create_file(*cd,
				&dev_attr_sys_encryption_write);
			*flags |= CFG_SYSFS_EWRT;
		} /* if */
#endif
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape_register_class_device exit: %d\n", rc));
	return rc;

} /* lin_tape_register_class_device() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_undo_class_dvc                                               *
*                                                                              *
*  PURPOSE: undo a partially-completed class_device init in case of failure    *
*                                                                              *
*    cd: the class_device that we began to initialize                          *
*    sdev: the scsi_device associated with this class device                   *
*    idx: the index into the driver table                                      *
*    offs: the offset for the specific type of device, (rew, nonrew, etc.)     *
*    prim_dvc: undo operations that were performed on a primary (rew) device   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_undo_class_dvc(struct device** cd, struct scsi_device* sdev,
	int idx, int offs, ulong flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	if(flags & CFG_SYSFS_SN)
		device_remove_file(*cd, &dev_attr_serial_num);
	if(flags & CFG_SYSFS_WWNN)
		device_remove_file(*cd, &dev_attr_ww_node_name);
	if(flags & CFG_SYSFS_WWPN)
		device_remove_file(*cd, &dev_attr_ww_port_name);
	if(flags & CFG_SYSFS_PATH)
		device_remove_file(*cd, &dev_attr_primary_path);
	if(flags & CFG_SYSFS_EPXY)
		device_remove_file(*cd, &dev_attr_sys_encryption_proxy);
	if(flags & CFG_SYSFS_EWRT)
		device_remove_file(*cd, &dev_attr_sys_encryption_write);
#endif
	device_destroy(lin_tape_class, MKDEV(lin_tape_major, idx + offs));
} /* lin_tape_undo_class_dvc() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_initial_drv_values                                       *
*                                                                              *
*  PURPOSE: set up a (caller-allocated) drive descriptor with initial values   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the descriptor for this drive                                        *
*    sdev: the scsi_device associated with this drive                          *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*    This function is called while holding drv lock for write                  *
*                                                                              *
*******************************************************************************/

void lin_tape_set_initial_drv_values(struct tape_descriptor* drv,
	struct scsi_device* sdev)
{
	int name_length = 0;
	struct new_utsname* uname = utsname();

    /* write lock */
	drv->dev_obj = sdev;
	drv->sdev_ct = 0;
	drv->rew_dev = NULL;
	drv->nonrew_dev = NULL;
	drv->nonresv_dev = NULL;

	drv->rew_class_dev = NULL;
	drv->nonrew_class_dev = NULL;
	drv->nonresv_class_dev = NULL;

	drv->cdb = NULL;
	drv->cdb_len = 0;
	if(uname)
		memcpy(&drv->uname_info, uname, sizeof(struct new_utsname));
	drv->host = sdev->host->host_no;
	drv->channel = sdev->channel;
	drv->target = sdev->id;
	drv->lun = sdev->lun;
	drv->direct_io = False;
	drv->buffer = NULL;
	drv->buf_len = 0;
	drv->sg_list = NULL;
	drv->supported_sg_tablesize = min(max_hw_segs(sdev), max_ph_segs(sdev));
	drv->supported_sg_tablesize = min(drv->supported_sg_tablesize,
						LinTapeMaxScatterList);
	if(!drv->supported_sg_tablesize)
		drv->supported_sg_tablesize = LinTapeMinScatterList;
	DbgPrint(("lin_tape: supported_sg_tablesize: %d\n",
		drv->supported_sg_tablesize));
	drv->rw_buf = NULL;
	drv->rw_buf_order = -1;
	drv->rw_buf_size = 0;

	/* update max scsi tranfer size -- limited by HBA max transfer size*/
	drv->max_xfer_len = (ulong)min(max_hw_segs(sdev), max_ph_segs(sdev))
			* max_seg_size(sdev);
	DbgPrint(("lin_tape: max_hw_segments: %d, max_phys_segments: %d, "\
		"max_segment_size: %d\n", max_hw_segs(sdev), max_ph_segs(sdev),
		max_seg_size(sdev)));

	drv->busy = False;
	drv->eom = False;
	drv->eom_blk_id = 0;
	drv->eod = False;
	drv->last_op = NO_COMMAND;
	drv->last_tape_motion = NO_COMMAND;
	drv->result = 0;
	drv->drv_open = False;
	drv->opened = 0;
	drv->cmd_in_open = 0;
	drv->prevent_removal = 0;
	drv->blk_size = 0;	        /* default block size             */
	drv->blk_num = 0;
	drv->buffered_mode = True;	/* default buffer mode            */
	drv->capacity_scaling = False;	/* default capacity scaling value */
	drv->compression = True;	/* default compression on         */
	drv->density = 0;	        /* default density value          */
	drv->data_safe_mode = 0;
	drv->pews[0]=drv->pews[1]=0;
	drv->ending_fm = 0;
	drv->logic_write_protect = False;/* default logic write protect   */
	drv->medium_type = False;
	drv->write_protect = False;	 /* default medium protection     */

	drv->curr_pos = 0;
	drv->pos_valid = False;

	drv->access_mode = ReadAndWrite;
	drv->acf_mode = False;
	drv->autoload = False;

	drv->max_blk_size = False;
	drv->min_blk_size = False;
	drv->read_mode = AIXMode;	/* default read mode             */
	drv->sili = False;	        /* sili bit support              */
	drv->record_space_mode = False;
	drv->retain_reservation = False;
	drv->rew_immed = False;
	drv->special_file_type = RewindDevice; /* default rewind special file */
	drv->trailer_labels = True;	      /* default trailer label ON */
	drv->wfm_immed = False;
	drv->limit_read_recov = petro; /* False unless petro is set */
	drv->limit_write_recov = petro; /* False unless petro is set */
	drv->lbp_set = False;
	drv->amu_enable = 0xFF;
	drv->eot_warn = 0;
	drv->wfm_on_close_safe = 0;
	drv->cmd_wo_ua_erp = False;
	drv->op_sg = 0;
	drv->op_sglist = NULL;
	drv->encryption_capable = False;
	drv->encryption_state = STATE_UNKNOWN;
	drv->application_set_encryption = False;
	drv->encryption_proxy_entitled = False;
	drv->sys_encryption_proxy = default_sys_encryption_proxy;
	drv->sys_encryption_write = default_sys_encryption_write;
	drv->encryption_keypath_configured = False;
	drv->server_test = 0;
	drv->do_encryption = False;

	drv->disable_simmim_log = !auto_logging;
	drv->auto_logsense = auto_logging;
	drv->disable_auto_drive_dump = disable_auto_drive_dump;
	drv->busy_retry = busy_retry;
	drv->trace = True;
	drv->trc_hook = False;
	memset(drv->devid, 0, MAX_VOLUME_LABEL_SIZE + 1);
	drv->simmim = NULL;
	drv->simmim_len = 0;
	drv->dmp = NULL;
	drv->dmp_len = 0;
	drv->dmp_size_order = 0;
	drv->tape_log = NULL;
	drv->tape_log_len = 0;
	drv->logsense = NULL;
	drv->logsense_len = 0;
	drv->doing = 0;
	drv->simmim_avail = False;
	drv->dmp_avail = False;
	drv->full_logsense_avail = False;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
	drv->simmim_lock = RW_LOCK_UNLOCKED;
	drv->dmp_lock = RW_LOCK_UNLOCKED;
	drv->logsense_lock = RW_LOCK_UNLOCKED;
	drv->full_trc_lock = RW_LOCK_UNLOCKED;
#else
	drv->simmim_lock = __RW_LOCK_UNLOCKED(drv->simmim_lock);
	drv->dmp_lock = __RW_LOCK_UNLOCKED(drv->dmp_lock);
	drv->logsense_lock = __RW_LOCK_UNLOCKED(drv->logsense_lock);
	drv->full_trc_lock = __RW_LOCK_UNLOCKED(drv->full_trc_lock);
#endif
	drv->full_trc_head = NULL;
	drv->full_trc_tail = NULL;

	drv->deferred_eom = False;
	drv->fm_detected = False;
	drv->ili = False;
	drv->resid = False;
	drv->resid_valid = False;
	drv->sense_data = NULL;
	drv->sense_valid = False;
	drv->in_erp = False;
	drv->ghost_device = False;
	drv->remove_pending = False;

	drv->curr_reserve_type = LinTapeNoReserve;
	drv->explicit_reserved = False;
	memset(drv->reservation_key, 0, 8);
	drv->pr_generation = 0;

	/* initialize path info for fail over */
	drv->do_failover = False;
	drv->alt_pathing = False;
	drv->cur_path = drv->dev_index;
	drv->isPrimary = True;
	drv->checked_multipath = False;
	drv->next_path = drv;
	drv->fenced = False;
	drv->enabled = True;
	init_waitqueue_head(&drv->wait_queue);
	drv->ready_wakeup = False;
	drv->wakeup_result = 0;
	memset(drv->lname, '\0', MaxLogicName);
	sprintf(drv->lname, "IBMtape%d",
		drv->dev_index);
	memset(drv->pname, '\0', MaxLogicName);
	if(drv->dev_obj
		&& (drv->dev_obj->host)
		&& (drv->dev_obj->host->hostt)
		&& (drv->dev_obj->host->hostt->name)) {
		DbgPrint(("lin_tape: hba %s\n",
			drv->dev_obj->host->hostt->name));
		name_length = min(MaxLogicName,
				(int)strlen(drv->dev_obj->host->hostt->name));
		memcpy(drv->pname, drv->dev_obj->host->hostt->name,
			name_length);
		if(name_length == MaxLogicName)
			drv->pname[name_length-1] = '\0';
	} /* if */
	drv->autotimeout=0;
} /* lin_tape_set_initial_drv_values() */

/*******************************************************************************
*                                                                              *
*  NAME: check_for_dup_drive                                                   *
*                                                                              *
*  PURPOSE: see if a newly allocated drive is failing back to old drive        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the descriptor for this drive                                        *
*    sdev: the scsi_device associated with this drive                          *
*    i: the index into the tape descriptor array of the drive                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS                                                     *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*******************************************************************************/

int check_for_dup_drive(struct tape_descriptor** drvp,
	struct scsi_device* sdev, int* i, boolean rescan)
{
	int j = 0, rc = -1;
	struct tape_descriptor* trav = NULL, *drv = *drvp;
	wait_queue_head_t wq;
	boolean new_scsi = True;

	init_waitqueue_head(&wq);
	for(j = 0; j < MaxLinTapeDevices; j++) {
		if(tape_devs[j] && tape_devs[j]->drv
            && tape_devs[j]->drv != drv
            && !drv->remove_pending
			&& drv->serial_num_len
			&& drv->serial_num_len
				== tape_devs[j]->drv->serial_num_len
			&& !memcmp(drv->serial_num, tape_devs[j]->drv->serial_num,
				drv->serial_num_len)
            && sdev->host->host_no == tape_devs[j]->drv->host
            && sdev->channel == tape_devs[j]->drv->channel
            && sdev->id == tape_devs[j]->drv->target
            && sdev->lun == tape_devs[j]->drv->lun) {
                new_scsi = False;
                break;
        } /* if */
	} /* for */

	if(new_scsi) goto EXIT_LABEL;
    WRITE_LOCK(&tape_devs[j]->drv_entry_lock);
    drv = tape_devs[j]->drv;
    if (drv->busy)
        DbgPrint(("lin_tape: Duplicated IBMtape%d busy %d\n",
                  drv->dev_index,drv->busy));
    drv->busy |= F_ADD_REMOVE;
    WRITE_UNLOCK(&tape_devs[j]->drv_entry_lock);
    if (!drv->ghost_device){
        if (drv->dev_obj != sdev && drv->dev_obj){
            DbgPrint(("lin_tape: Detected duplicated with new scsi device\n"));
            WRITE_UNLOCK(&tape_des_lock);
            lin_tape_unconfig_drive(
                 &(*drvp)->dev_obj->sdev_gendev,
                 *drvp);
            WRITE_LOCK(&tape_des_lock);
        }
        else rc = 1;
        DbgPrint(("lin_tape: Detected duplicated path for drive index %d\n", j));
    } /* if */
    else {
        DbgPrint(("lin_tape: Resurrecting path for drive index %d\n", j));
        drv->ghost_device = False;
        drv->cur_path = j;
    }
    *drvp = drv;
    drv->dev_obj = sdev;
    WRITE_LOCK(&tape_devs[*i]->drv_entry_lock);
    kfree(tape_devs[*i]->drv);
    tape_devs[*i]->drv = NULL;
    WRITE_UNLOCK(&tape_devs[*i]->drv_entry_lock);
    *i = j;
    if (rc == 1) return rc;
    rc = STATUS_SUCCESS;

    for(trav = drv->next_path; trav != drv; trav = trav->next_path) {
            if(!trav->ghost_device) drv->cur_path = trav->cur_path;
            if(trav->isPrimary) drv->isPrimary = False;
    } /* for */

EXIT_LABEL:
	return rc;
} /* check_for_dup_drive() */


/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_init_changer_des                                             *
*                                                                              *
*  PURPOSE: initialize a medium changer descriptor                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the descriptor to initialize                                         *
*    sdev: the sysfs scsi_device object associated with the descriptor         *
*    i: the index of the device in the changer table                           *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding changer_des_lock for write          *
*                                                                              *
*******************************************************************************/

int lin_tape_init_changer_des(struct changer_descriptor* chg,
	struct scsi_device* sdev, int* i, boolean rescan)
{
	int rc = STATUS_SUCCESS;
	boolean chg_class_dvc_regd = False, chg_cdev_creatd = False;
	ulong flags = CFG_CHG_DVC;

	DbgPrint(("lin_tape_init_changer_des entry.\n"));
    WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
	lin_tape_set_initial_chg_values(chg, sdev);

	copy_wwn_from_tape_devs(chg);

	chg->cdb = NULL;
	chg->cdb_len = 0;
	chg->serial_num_len = 0;
	memset(chg->serial_num, 0, MaxSerialNumberSize);
	memcpy(chg->serial_num, "NA", 2);
	chg->busy |= F_ADD_REMOVE;
	WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
	lin_tape_get_inq_p0x83_data((struct tape_descriptor*)chg, True);

	if (dualaccessor_cpf == 1){
		DbgPrint(("lin_tape: CPF for dual accessor mode\n"));
		rc = check_for_dup_changer_by_wwnn(&chg, sdev, i);
	} else {
		rc = check_for_dup_changer(&chg, sdev, i, rescan);
	}
	chg = changer_devs[*i]->chg;
	if (rc == 1) {
        rc = STATUS_SUCCESS;
        goto EXIT_LABEL;
	}
	rc = lin_tape_create_cdev(&chg->chg_dev, chg->dev_index,
		LinTapeChangerOffset, "/dev/IBMchanger%d");
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	chg_cdev_creatd = True;

	rc = lin_tape_register_class_device(&chg->chg_class_dev,
		(struct tape_descriptor*)chg, sdev, "IBMchanger%d",
		chg->dev_index, LinTapeChangerOffset, &flags);
	chg_class_dvc_regd = True;
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	chg->chgflags = flags;

EXIT_LABEL:

	if(rc != STATUS_SUCCESS) {
		if(chg_cdev_creatd) cdev_del(chg->chg_dev);
		if(chg && sdev && chg_class_dvc_regd)
			lin_tape_undo_class_dvc(&chg->chg_class_dev, sdev,
				chg->dev_index, LinTapeChangerOffset,
				flags);
	} /* if */

    if (chg){
        WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
        chg->busy &= 0xFE;
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
    }
	DbgPrint(("lin_tape_init_changer_des exit: %d\n", rc));
	return rc;
} /* lin_tape_init_changer_des() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_initial_chg_values                                       *
*                                                                              *
*  PURPOSE: set up a (caller-allocated) chgr descriptor with initial values    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the descriptor for this changer                                      *
*    sdev: the scsi_device associated with this changer                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_initial_chg_values(struct changer_descriptor* chg,
	struct scsi_device* sdev)
{
	int name_length = 0;
	struct new_utsname* uname = utsname();

	chg->dev_obj = sdev;
	chg->sdev_ct = 0;
	chg->chg_class_dev = NULL;
    chg->chg_dev = NULL;

	if(uname)
		memcpy(&chg->uname_info, uname, sizeof(struct new_utsname));
	chg->host = sdev->host->host_no;
	chg->channel = sdev->channel;
	chg->target = sdev->id;
	chg->lun = sdev->lun;

	chg->buffer = NULL;
	chg->buf_len = 0;

	chg->busy = False;
	chg->chg_open = False;
	chg->open_count = 0;
	chg->last_op = NO_COMMAND;
	chg->result = 0;

	chg->curr_reserve_type = LinTapeNoReserve;
	chg->explicit_reserved = False;

	chg->sense_data = NULL;
	chg->sense_valid = False;
	chg->resid = 0;
	chg->resid_valid = False;
	chg->in_erp = False;
	chg->cmd_wo_ua_erp = False;
	chg->ghost_device = False;
	chg->remove_pending = False;

	chg->comp_code = 0;
	chg->comp_code_valid = False;
	chg->last_errno = 0;
	chg->barcode = 0;

	chg->trace = True;
	chg->chg_log = NULL;
	chg->chg_log_len = 0;
	memset(chg->devid, '\0', MAX_VOLUME_LABEL_SIZE + 1);
	chg->doing = 0;

	chg->full_trc_head = NULL;
	chg->full_trc_tail = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
	chg->full_trc_lock = RW_LOCK_UNLOCKED;
#else
	chg->full_trc_lock = __RW_LOCK_UNLOCKED(chg->full_trc_lock);
#endif

	chg->do_failover = False;
	chg->alt_pathing = False;
	chg->isPrimary = True;
	chg->checked_multipath = False;
	chg->next_path = chg;
	chg->fenced = False;
	chg->enabled = True;
	chg->cur_path = chg->dev_index;
	init_waitqueue_head(&chg->wait_queue);
	chg->ready_wakeup = False;
	chg->wakeup_result = 0;
    memset(chg->pname,'\0',MaxLogicName);
    memset(chg->lname,'\0',MaxLogicName);
	sprintf(chg->lname, "IBMchanger%d", chg->dev_index);
	if((chg->dev_obj)
		&& (chg->dev_obj->host)
		&& (chg->dev_obj->host->hostt)
		&& (chg->dev_obj->host->hostt->name)){
		DbgPrint(("hba: %s\n", chg->dev_obj->host->hostt->name));
		name_length = min(MaxLogicName,
				(int)strlen(chg->dev_obj->host->hostt->name));
		memcpy(chg->pname, chg->dev_obj->host->hostt->name,
			name_length);
		if(name_length == MaxLogicName)
			chg->pname[name_length - 1] = '\0';
	} /* if */

} /* lin_tape_set_initial_chg_values() */

/*******************************************************************************
*                                                                              *
*  NAME: check_for_dup_changer                                                 *
*                                                                              *
*  PURPOSE: see if a newly allocated changer is failing back to old changer    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the descriptor for this drive                                        *
*    sdev: the scsi_device associated with this changer                        *
*    i: the index into the changer descriptor array of the changer             *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    This function is called while holding tape_des_lock for write             *
*                                                                              *
*******************************************************************************/

int check_for_dup_changer(struct changer_descriptor** chgp,
	struct scsi_device* sdev, int* i, boolean rescan)
{
	int j = 0, rc = -1, sn_len = 0;
	struct changer_descriptor* trav = NULL, *chg = *chgp;
	wait_queue_head_t wq;
	boolean new_scsi = True;
	init_waitqueue_head(&wq);

	sn_len = min((int)chg->serial_num_len, 12);
	for(j = 0; j < MaxLinTapeChangerDevices; j++) {
		if(changer_devs[j] && changer_devs[j]->chg
			&& changer_devs[j]->chg != chg
			&& !chg->remove_pending
			&& chg->serial_num_len
			&& chg->serial_num_len
				== changer_devs[j]->chg->serial_num_len
			&& !memcmp(chg->serial_num,
				changer_devs[j]->chg->serial_num, sn_len)
			&& sdev->host->host_no == changer_devs[j]->chg->host
            && sdev->channel == changer_devs[j]->chg->channel
			&& sdev->id == changer_devs[j]->chg->target
			&& sdev->lun == changer_devs[j]->chg->lun) {
				new_scsi = False;
				break;
		} /* if */
	} /* for */

	if(new_scsi) goto EXIT_LABEL;
    WRITE_LOCK(&changer_devs[j]->entry_lock);
    chg = changer_devs[j]->chg;
    *chgp = chg;
    if (chg->busy)
        DbgPrint(("lin_tape: Duplicated IBMchanger%d busy %d\n",
                  chg->dev_index, chg->busy));
    chg->busy |= F_ADD_REMOVE;
    WRITE_UNLOCK(&changer_devs[j]->entry_lock);
    if (!chg->ghost_device){
        if (chg->dev_obj != sdev && chg->dev_obj){
            DbgPrint(("lin_tape: Detected duplicated with new scsi device\n"));
            WRITE_UNLOCK(&changer_des_lock);
            lin_tape_unconfig_changer(
                &changer_devs[j]->chg->dev_obj->sdev_gendev,
                changer_devs[j]->chg->dev_obj, j);
            WRITE_LOCK(&changer_des_lock);
        }
        else {
            rc = 1;
            DbgPrint(("lin_tape: Detected duplicated path for changer index %d\n", j));
        }
    } /* if */
    else{
        DbgPrint(("lin_tape: Resurrecting path for changer index %d\n", j));
        chg->ghost_device = False;
        chg->cur_path = j;
    }
    chg->dev_obj = sdev;

    WRITE_LOCK(&changer_devs[*i]->entry_lock);
    kfree(changer_devs[*i]->chg);
    changer_devs[*i]->chg = NULL;
    WRITE_UNLOCK(&changer_devs[*i]->entry_lock);
    *i = j;
    if (rc == 1) return rc;
    rc = STATUS_SUCCESS;

	for(trav = chg->next_path; trav != chg; trav = trav->next_path) {
		if(!trav->ghost_device) chg->cur_path = trav->cur_path;
		if(trav->isPrimary) chg->isPrimary = False;
	} /* for */

EXIT_LABEL:

	return rc;
} /* check_for_dup_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: check_for_dup_changer_by_wwnn                                         *
*                                                                              *
*  PURPOSE: see if a newly allocated changer is failing back to old changer    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the descriptor for this drive                                        *
*    sdev: the scsi_device associated with this changer                        *
*    i: the index into the changer descriptor array of the changer             *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/
int check_for_dup_changer_by_wwnn(struct changer_descriptor** chg,
				  struct scsi_device* sdev, int* i)
{
	int j = 0, k = -1, rc = STATUS_SUCCESS, trc_len = 0, tmp_rc;
	struct changer_descriptor* trav = NULL;
	wait_queue_head_t wq;
	boolean new_scsi = False;
	unchar* trc = NULL;

	init_waitqueue_head(&wq);

	DbgPrint(("lin_tape: check_for_dup_cahnger_by_wwnn: \n"));

	for(j = 0; j < MaxLinTapeChangerDevices; j++) {

	  if (changer_devs[j] && changer_devs[j]->chg){
	    int m = 0;
	    DbgPrint(("lin_tape: (*chg)->wwnn=0x"));
	    for(m = 0; m < WWN_LEN; m++) {
	      DbgPrint(("%02X", (*chg)->wwnn[m] & 0xFF));
	    } /* for */
	    DbgPrint((", changer_devs[%d]->wwnn=0x",j));
	    for(m = 0; m < WWN_LEN; m++) {
	      DbgPrint(("%02X", changer_devs[j]->chg->wwnn[m] & 0xFF));
	    } /* for */
	    DbgPrint(("\n"));
	  }

	  if(changer_devs[j] && changer_devs[j]->chg
			&& changer_devs[j]->chg != *chg
			&& changer_devs[j]->chg->ghost_device
			&& !memcmp((*chg)->wwnn,
				changer_devs[j]->chg->wwnn, WWN_LEN)) {

			if(k < 0) {
			    k = j;
			}
			if(sdev->host->host_no == changer_devs[j]->chg->host
				&& sdev->channel == changer_devs[j]->chg->channel
				&& sdev->id == changer_devs[j]->chg->target
				&& sdev->lun == changer_devs[j]->chg->lun) {
				k = j;
				break;
			} /* if */
		} /* if */
	} /* for */

	if(j >= MaxLinTapeChangerDevices && k < 0) goto EXIT_LABEL;
	if(j >= MaxLinTapeChangerDevices) new_scsi = True;

	j = k;
	DbgPrint(("lin_tape: %s path for changer %d\n", (new_scsi)? "New":"Resurecting",j));

	if(new_scsi) {
		trc_len = 4;
		trc = kmalloc(trc_len, GFP_KERNEL);
		if(!trc) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		trc[0] = changer_devs[j]->chg->host;
		trc[1] = changer_devs[j]->chg->channel;
		trc[2] = changer_devs[j]->chg->target;
		trc[3] = changer_devs[j]->chg->lun;
	} /* if */

	if(changer_devs[j]->chg->checked_multipath) {
		(*chg)->checked_multipath = True;
		(*chg)->alt_pathing = True;
		for(trav = changer_devs[j]->chg->next_path;
			trav->next_path != changer_devs[j]->chg;
			trav = trav->next_path) continue;
		trav->next_path = *chg;
		(*chg)->next_path = changer_devs[j]->chg;
		(*chg)->open_count = changer_devs[j]->chg->open_count;
	} /* if */

	cleanup_and_free_cdesc(j);

	changer_devs[j]->chg = *chg;
	*chg = NULL;
	chg = &changer_devs[j]->chg;
	*i = (*chg)->dev_index = (*chg)->cur_path = j;
	(*chg)->host = sdev->host->host_no;
	(*chg)->channel = sdev->channel;
	(*chg)->target = sdev->id;
	(*chg)->lun = sdev->lun;

	memset((*chg)->lname, '\0', MaxLogicName);
	sprintf((*chg)->lname, "IBMchanger%d", (*chg)->dev_index);

	for(trav = (*chg)->next_path; trav != *chg; trav = trav->next_path) {
		if(!trav->ghost_device) (*chg)->cur_path = trav->cur_path;
		if(trav->isPrimary) (*chg)->isPrimary = False;
	} /* for */

	if(new_scsi) {
		tmp_rc = rc;
		rc = changer_setup_and_log_full(*chg, (*chg)->dev_index,
			TRC_CG_NEW_SCSI_ADDR, trc, trc_len);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		rc = tmp_rc;
	} /* if */

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;
} /* check_for_dup_changer_by_wwnn() */

/*    This function is called while holding des_lock for write     */
void lin_tape_get_inq_p0x83_data(struct tape_descriptor* drv, int changer)
{
	union cdb* cdb = NULL;
	char* page = NULL;
	struct changer_descriptor* chg = (struct changer_descriptor*)drv;
	int rc = 0, inquiry_data_len = 4, id_type_offset = 1,
		id_len_offset = 3, i = 0, pkg_offs = 35;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,4,38)
    int retry = -1;
#endif // LINUX_VERSION_CODE
	struct scsi_device *sdev;
	struct tape_sense_data *sense_data = NULL;
    /* write lock */
	DbgPrint(("lin_tape_get_inq_p0x83_data entry\n"));
	if(!drv) goto EXIT_LABEL;

	cdb = kmalloc(sizeof(union cdb), GFP_KERNEL);
	if(!cdb) goto EXIT_LABEL;
	memset(cdb, '\0', sizeof(union cdb));
	drv->cdb = cdb;

	sense_data = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
	if(!sense_data) goto EXIT_LABEL;
	memset(sense_data, '\0', sizeof(struct tape_sense_data));

	page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!page) goto EXIT_LABEL;
	memset(page, '\0', MAX_INQ_LEN);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,4,38)
    do {
    if (rc != 0) msleep(5000);
	if(changer) {
		sdev = chg->dev_obj;
		WRITE_UNLOCK(&changer_des_lock);
		chg->sense_data = sense_data;
		rc = changer_inquiry_page(&chg, DeviceIdentificationPage, page);
		WRITE_LOCK(&changer_des_lock);
	} else {
		sdev = drv->dev_obj;
		WRITE_UNLOCK(&tape_des_lock);
        drv->sense_data = sense_data;
		rc = tape_inquiry_page(&drv, DeviceIdentificationPage, page);
		WRITE_LOCK(&tape_des_lock);
	} /* else */
	retry ++;
    } while (rc && retry < 36 &&
            (
                (
                 sense_data->sense_key == 0x02 &&
                        ((sense_data->asc == 0x04 &&
                          sense_data->ascq < 0x03)
                        ||
                        (sense_data->asc == 0x3E &&
                         sense_data->ascq == 0x00))
                )
                ||
                (
                 sense_data->sense_key == 0xB &&
                 sense_data->asc == 0x29 &&
                 sense_data->ascq == 0x07
                )
            )
        );
#else
		if (changer)	{
            sdev = chg->dev_obj;
            WRITE_UNLOCK(&changer_des_lock);
		}
		else	{
		    sdev = drv->dev_obj;
            WRITE_UNLOCK(&tape_des_lock);
		}
		rc = scsi_get_vpd_page(sdev, 0x83, page, MAX_INQ_LEN);
		if ( rc < 0) rc = scsi_get_vpd_page(sdev, 0x80, page, MAX_INQ_LEN);
		if (changer) { WRITE_LOCK(&changer_des_lock);}
        else { WRITE_LOCK(&tape_des_lock);}
#endif
	if(rc < 0) 	goto EXIT_LABEL;

	i = inquiry_data_len;
	inquiry_data_len += page[inquiry_data_len - 1] & 0xFF;
	while(i < inquiry_data_len) {
		switch(page[i + id_type_offset] & 0x0F) {
		case 1: /* vendor ID, product ID, serial number */
			rc=lin_tape_inq0x83_ser(page, drv, i, changer);
			if (rc < 0){
				DbgPrint(("lin_tape_get_inq_p0x83_data: %s_inquiry_page "\
					"failed with %d\n", changer ? "changer" : "tape", rc));
				goto EXIT_LABEL;
			}
			break;
		case 3: /* WWNN or WWPN */
			lin_tape_inq0x83_wwn(page, drv, i, changer);
			break;
		default:
			break;
		} /* switch */

		i += id_len_offset;
		i += (page[i] & 0xFF) + 1;
	} /* while */

	if(!changer) {
		memset(page, '\0', MAX_INQ_LEN);
		WRITE_UNLOCK(&tape_des_lock);
		rc = scsi_get_vpd_page(drv->dev_obj, DriveRevisionLevelsPage, page, MAX_INQ_LEN);
        WRITE_LOCK(&tape_des_lock);
		if(!rc && !memcmp(&page[pkg_offs], "hl", strlen("hl")))
			drv->v2 = True;
		DbgPrint(("lin_tape: this is a %s drive\n", drv->v2 ?
				"v2" : "v1"));
	} /* if */

EXIT_LABEL:

	if(cdb) kfree(cdb);
	if(page) kfree(page);
	if(sense_data) kfree(sense_data);
	drv->cdb = NULL;
	if (!changer) drv->sense_data = NULL;
	else chg->sense_data = NULL;
	DbgPrint(("lin_tape_get_inq_p0x83_data exit\n"));
	return;

} /* lin_tape_get_inq_p0x83_data() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_inq0x83_ser                                                  *
*                                                                              *
*  PURPOSE: process serial number returned from inquiry p. 0x83                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    page: the data returned by the call to inquiry p. 0x83                    *
*    drv: the tape/changer descriptor                                          *
*    i: index of the current p. 0x83 ID descriptor                             *
*    changer: boolean indicating that this is a changer rather than a drive    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_inq0x83_ser(char* page, struct tape_descriptor* drv, int i,
	int changer)
{
	int serial_num_len = 0, id_len_offset = 3, vendor_id_len = 8,
		product_id_len = 16, data_offset = 4;
	struct changer_descriptor* chg = (struct changer_descriptor*)drv;
	unchar *inq_page = NULL;
	int rc = STATUS_SUCCESS;

	if (!changer){
	serial_num_len = (page[i + id_len_offset] & 0xFF) - vendor_id_len
		- product_id_len;

	memcpy(drv->serial_num,
		   &page[i + data_offset + vendor_id_len + product_id_len],
		   serial_num_len);

    drv->serial_num_len = serial_num_len;
	DbgPrint(("lin_tape_get_inq_p0x83_data: serial_num %s\n",
		drv->serial_num));

	memset(drv->devid, '\0', MAX_VOLUME_LABEL_SIZE + 1);
	sprintf(&drv->devid[0], "IBMtape%d------", drv->dev_index);

	if (drv->serial_num_len >= 5)
	memcpy(&drv->devid[13], &(drv->serial_num[drv->serial_num_len-5]), 5);
	else memcpy(&drv->devid[13], "NoSer", strlen("NoSer"));

	}
	else{//changer

		/* get device serial number from p.0x80 due to new validation requirement*/
		inq_page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
		if(!inq_page) {
			return -ENOMEM;
		} /* if */
		memset(inq_page, '\0', MAX_INQ_LEN);
        /* NOTE: this function is called only on init, so we hold the des
		lock.  If it is ever called in any other way, the lock
		should be checked */
		WRITE_UNLOCK(&changer_des_lock);
    	rc = changer_inquiry_page(&chg, UnitSerialPage, inq_page);
		WRITE_LOCK(&changer_des_lock);
		if (chg){
		if(rc == STATUS_SUCCESS )
			chg->serial_num_len = min((int)inq_page[3], MaxSerialNumberSize);
		if (rc != STATUS_SUCCESS || chg->serial_num_len < 7)
		{//get it from p0x83
			serial_num_len = (page[i + id_len_offset] & 0xFF) - vendor_id_len
				- product_id_len;

			memcpy(chg->serial_num,
				   &page[i + data_offset + vendor_id_len + product_id_len],
				   serial_num_len);

			chg->serial_num_len = serial_num_len;
			DbgPrint(("lin_tape_get_inq_p0x83_data: serial_num %s\n",
					chg->serial_num));

            memset(chg->devid, '\0', MAX_VOLUME_LABEL_SIZE + 1);
            sprintf(&chg->devid[0], "IBMChgr%d------", chg->dev_index);

            if (chg->serial_num_len >= 5)
            memcpy(&drv->devid[13], &(drv->serial_num[drv->serial_num_len-5]), 5);
            else memcpy(&drv->devid[13], "NoSer", strlen("NoSer"));

			rc = STATUS_SUCCESS;
		} /* if */
		else{
		    memcpy(chg->serial_num, &inq_page[4], chg->serial_num_len);
		    DbgPrint(("lin_tape_get_inq_p0x80_data: serial_num %s\n",
		    	chg->serial_num));
		} /* else */
		}
	}

	return rc;

} /* lin_tape_inq0x83_ser() */


/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_inq0x83_wwn                                                  *
*                                                                              *
*  PURPOSE: process ww names returned from inquiry p. 0x83                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    page: the data returned by the call to inquiry p. 0x83                    *
*    drv: the tape/changer descriptor                                          *
*    i: index of the current p. 0x83 ID descriptor                             *
*    changer: boolean indicating that this is a changer rather than a drive    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_inq0x83_wwn(char* page, struct tape_descriptor* drv, int i,
	int changer)
{
	int id_type_offset = 1, data_offset = 4, j = 0;
	char buffer[32] = {0};

	if(changer) return;
	switch((page[i + id_type_offset] & 0x30) >> 4) {
	case 0: /* WWNN */
		memcpy(drv->wwnn, &page[i + data_offset], WWN_LEN);

		if(lin_tape_debug) {
			memset(buffer, '\0', 32);
			memcpy(buffer, "0x", strlen("0x"));
			for(j = 0; j < WWN_LEN; j++) {
				sprintf(&buffer[strlen(buffer)], "%02X",
					drv->wwnn[j] & 0xFF);
			} /* for */

			DbgPrint(("lin_tape_get_inq_p0x83_data: "\
				"world-wide node name: %s\n", buffer));
		} /* if */
		break;

	case 1: /* WWPN */
		memcpy(drv->wwpn, &page[i + data_offset], WWN_LEN);

		if(lin_tape_debug) {
			memset(buffer, '\0', 32);
			memcpy(buffer, "0x", strlen("0x"));
			for(j = 0; j < WWN_LEN; j++) {
				sprintf(&buffer[strlen(buffer)], "%02X",
					drv->wwpn[j] & 0xFF);
			} /* for */

			DbgPrint(("lin_tape_get_inq_p0x83_data: "\
				"world-wide port name: %s\n", buffer));
		} /* if */
		break;
	default:
		break;
	} /* switch */

} /* lin_tape_inq0x83_wwn() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_rescan                                                       *
*                                                                              *
*  PURPOSE: rescan device and update device table.  Invoked by rescan in sysfs *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    dvc: the device object that is being rescanned                            *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

static void lin_tape_rescan(struct device* dvc)
{
	struct scsi_device *sdev = NULL;
	unchar tape_id = 0;

	DbgPrint(("lin_tape_rescan entry\n"));

	sdev = to_scsi_device(dvc);
	if(!sdev) {
		printk("lin_tape: to_scsi_device returned NULL.\n");
		goto EXIT_LABEL;
	} /* if */

	if((sdev->type != TYPE_TAPE) && (sdev->type != TYPE_MEDIUM_CHANGER))
		goto EXIT_LABEL;

	tape_id = lin_tape_supported_device(sdev);

	if(!tape_id) {
		printk("lin_tape: do not support %s %s %s:\n",
			sdev->vendor, sdev->model, sdev->rev);
		printk("scsi bus:%d, channel:%d, id:%d, lun:%d\n",
			sdev->host->host_no, sdev->channel, sdev->id,
			(int)sdev->lun);

		goto EXIT_LABEL;
	} /* if */

	/* find out the device table entry */
	if(sdev->type == TYPE_TAPE) {
		lin_tape_rescan_tape(sdev, tape_id);
	} else if(sdev->type == TYPE_MEDIUM_CHANGER) {
		lin_tape_rescan_changer(sdev, tape_id);
	} /* else if */

EXIT_LABEL:

	DbgPrint(("lin_tape_rescan exit\n"));
	return;
} /* lin_tape_rescan() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_rescan_tape                                                  *
*                                                                              *
*  PURPOSE: handle sysfs rescan for tape drives                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: scsi_device for this request                                        *
*    tape_id: index for the tape descriptor                                    *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_rescan_tape(struct scsi_device* sdev, int tape_id)
{
	int i = 0, rc = 0;

	DbgPrint(("lin_tape_rescan tape entry\n"));

	WRITE_LOCK(&tape_des_lock);
	if(!tape_devs) {
		rc = allocate_tdesc_array();
		if(rc != STATUS_SUCCESS)    goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeDevices; i++) {
        if(!tape_devs[i] || !tape_devs[i]->drv) {
				rc = alloc_and_init_tdesc(sdev, tape_id,
                                    &i, True);
				break;
		} /* if */
	} /* for */

	if(i >= MaxLinTapeDevices){
		DbgPrint(("lin_tape: exceeded maximum drive limit, "\
			"cannot attach. Rescan recommended\n"));
	} /* if */

    if (lin_tape_clean_ghosts || i == MaxLinTapeDevices){
	for(i = 0; i < MaxLinTapeDevices; i++) {
        if(tape_devs[i] && tape_devs[i]->drv) {
            READ_LOCK(&tape_devs[i]->drv_entry_lock);
            if (tape_devs[i]->drv->ghost_device
                && !tape_devs[i]->drv->busy
                && tape_devs[i]->drv->cur_path != i){
                WRITE_UNLOCK(&tape_des_lock);
                READ_UNLOCK(&tape_devs[i]->drv_entry_lock);
                cleanup_and_free_tdesc(i);
                WRITE_LOCK(&tape_des_lock);
            } /* if */
            else READ_UNLOCK(&tape_devs[i]->drv_entry_lock);
        } /* if */
	} /* for */
    }

EXIT_LABEL:

    WRITE_UNLOCK(&tape_des_lock);
	DbgPrint(("lin_tape_rescan tape exit\n"));
	return;
} /* lin_tape_rescan_tape() */



/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_rescan_changer                                               *
*                                                                              *
*  PURPOSE: handle sysfs rescan for changers                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    sdev: scsi_device for this request                                        *
*    tape_id: index for the changer descriptor                                 *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_rescan_changer(struct scsi_device* sdev, int tape_id)
{
	int i = 0, rc = 0;

	DbgPrint(("lin_tape_rescan changer entry\n"));

	WRITE_LOCK(&changer_des_lock);
	if(!changer_devs) {
		rc = allocate_cdesc_array();
		if(rc != STATUS_SUCCESS)    goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MaxLinTapeDevices; i++) {
        if(!changer_devs[i] || !changer_devs[i]->chg) {
				rc = alloc_and_init_cdesc(sdev, tape_id,
                                    &i, True);
				break;
		} /* if */
	} /* for */

	if(i >= MaxLinTapeDevices){
		DbgPrint(("lin_tape: exceeded maximum drive limit, "\
			"cannot attach. Rescan again\n"));
	} /* if */

    if (lin_tape_clean_ghosts || i == MaxLinTapeDevices)
	for(i = 0; i < MaxLinTapeDevices; i++) {
        if(changer_devs[i] && changer_devs[i]->chg) {
            READ_LOCK(&changer_devs[i]->entry_lock);
            if (changer_devs[i]->chg->ghost_device
                && !changer_devs[i]->chg->busy
                && changer_devs[i]->chg->cur_path != i){
                READ_UNLOCK(&changer_devs[i]->entry_lock);
                cleanup_and_free_tdesc(i);
            } /* if */
            else READ_UNLOCK(&changer_devs[i]->entry_lock);
        } /* if */
	} /* for */

EXIT_LABEL:

    WRITE_UNLOCK(&changer_des_lock);
	DbgPrint(("lin_tape_rescan changer exit\n"));
	return;
} /* lin_tape_rescan_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open                                                         *
*                                                                              *
*  PURPOSE: Open entry point for drives and changers                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    inode: device special file metadata structure                             *
*    filp: device special file pointer                                         *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

static int lin_tape_open(struct inode *inode, struct file *filp)
{
	unsigned int minor;
	int rc = -EIO;
	unchar ttype = lin_tape_trace_type;

	minor = iminor(inode);
	DbgPrint2(("lin_tape: lin_tape_open entry: minor %d\n", minor));
	if(PSEUDO_TAPE(minor)) /* /dev/IBMtape for daemon */
		rc = lin_tape_open_pseudo();

	else if(TSM_TAPE(minor)) /* /dev/TSMtape for TSM IOCTLs */
		rc = lin_tape_open_tsm();

	else if(SPECIAL_DEV(minor)) /*/dev/IBMSpecial for preemption */
		rc = lin_tape_open_special();

	else if(minor < LinTapeChangerOffset)
		rc=lin_tape_open_drive(minor, ttype, filp);

	else if(minor < LinTapeMaxCount)
		rc = lin_tape_open_changer(minor, ttype, filp);

	else {
		printk("lin_tape: unsupported device.\n");
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len, 234);
		rc = -ENXIO;
	} /* else */

	goto EXIT_LABEL;

EXIT_LABEL:

	DbgPrint2(("lin_tape_open exit: %d\n", rc));
	return rc;
} /* lin_tape_open() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open_pseudo                                                  *
*                                                                              *
*  PURPOSE: Open entry point for pseudo device (used with daemon)              *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_open_pseudo()
{
	unchar* trc = NULL;
	int rc = STATUS_SUCCESS, i;

	if(lin_tape_trace_type != LIN_TAPE_INVALID_TRACE) {
		printk("lin_tape: lin_taped is already opened.\n");
		rc = -EAGAIN;
		goto EXIT_LABEL;
	} /* if */

	trc = kmalloc(1 + MaxSerialNumberSize, GFP_KERNEL);
	if(!trc) goto EXIT_LABEL;

	if(tape_devs) {
		READ_LOCK(&tape_des_lock);
		for(i = 0; i < MaxLinTapeDevices; i++) {
			if(!tape_devs[i] || !tape_devs[i]->drv) continue;

			trc[0] = tape_devs[i]->drv->serial_num_len;

			if(tape_devs[i]->drv->serial_num_len)
				memcpy(&trc[1], tape_devs[i]->drv->serial_num,
					tape_devs[i]->drv->serial_num_len);

			tape_setup_and_log_full(tape_devs[i]->drv,
				tape_devs[i]->drv->dev_index, TRC_TP_SERIAL_NUM,
				trc, 1 + tape_devs[i]->drv->serial_num_len);

			if(tape_devs[i]->drv->drv_open)
				tape_setup_and_log_full(tape_devs[i]->drv,
					tape_devs[i]->drv->dev_index, TRC_TP_OPEN,
					NULL, -2);

			if(tape_devs[i]->drv->curr_reserve_type)
				tape_setup_and_log_full(tape_devs[i]->drv,
					tape_devs[i]->drv->dev_index,
					TRC_TP_RESERVED, NULL, 0);
		} /* for */
		READ_UNLOCK(&tape_des_lock);
	} /* if */

	memset(trc, '\0', 1 + MaxSerialNumberSize);

	if(changer_devs) {
		READ_LOCK(&changer_des_lock);
		for(i = 0; i<MaxLinTapeChangerDevices; i++) {
			if(!changer_devs[i] || !changer_devs[i]->chg) continue;

			trc[0] = changer_devs[i]->chg->serial_num_len;
			if(changer_devs[i]->chg->serial_num_len)
				memcpy(&trc[1], changer_devs[i]->chg->serial_num,
					changer_devs[i]->chg->serial_num_len);

			changer_setup_and_log_full(changer_devs[i]->chg,
				changer_devs[i]->chg->dev_index, TRC_CG_SERIAL_NUM,
				trc, 1 + changer_devs[i]->chg->serial_num_len);

			if(changer_devs[i]->chg->chg_open)
				changer_setup_and_log_full(changer_devs[i]->chg,
					changer_devs[i]->chg->dev_index,
					TRC_CG_OPEN, NULL, -2);

			if(changer_devs[i]->chg->curr_reserve_type)
				changer_setup_and_log_full(changer_devs[i]->chg,
					changer_devs[i]->chg->dev_index,
					TRC_CG_RESERVED, NULL, 0);
		} /* for */
		READ_UNLOCK(&changer_des_lock);
	} /* if */

EXIT_LABEL:

	if(trc) kfree(trc);
	return rc;
} /* lin_tape_open_pseudo() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open_special                                                 *
*                                                                              *
*  PURPOSE: Open entry point for special dev                                   *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_open_special()
{
	int rc = STATUS_SUCCESS;

	if(atomic_read(&special_tp_open)) {
		printk("lin_tape: Special device is already opened.\n");
		rc = -EAGAIN;
		goto EXIT_LABEL;
	} /* if */

	atomic_inc(&special_tp_open);

EXIT_LABEL:
	return rc;

} /* lin_tape_open_special() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open_tsm                                                     *
*                                                                              *
*  PURPOSE: Open entry point for TSM pseudo device (used for TSM IOCTLs)       *
*                                                                              *
*  PARAMETERS: None                                                            *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_open_tsm()
{
	int rc = STATUS_SUCCESS;

	if(atomic_read(&tsm_tp_open)) {
		printk("lin_tape: TSM pseudo device is already opened.\n");
		rc = -EAGAIN;
		goto EXIT_LABEL;
	} /* if */

	atomic_inc(&tsm_tp_open);

EXIT_LABEL:

	return rc;

} /* lin_tape_open_tsm() */

#define LOCK_DES {READ_LOCK(&tape_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&tape_des_lock);}
#define LOCK_DRV {READ_LOCK(&drv->drv_entry_lock); drv_locked = True;}
#define UNLOCK_DRV {drv_locked = False; READ_UNLOCK(&drv->drv_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open_drive                                                   *
*                                                                              *
*  PURPOSE: Perform open operations for drives                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    minor: the minor device number, from which the dev index is derived       *
*    ttype: trace type                                                         *
*    filp: Linux special file pointer                                          *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_open_drive(int minor, unchar ttype, struct file* filp)
{

	struct tape_descriptor *drv = NULL;
	int idx = -1;
	int rc = STATUS_SUCCESS, device_gotten = False, in = False;
	unchar reserve_on_open = True;

	READ_LOCK(&tape_des_lock);
	rc = lin_tape_get_drive_des(&drv, minor, ttype, &reserve_on_open);
	if(rc != STATUS_SUCCESS) {
            READ_UNLOCK(&tape_des_lock);
            goto EXIT_LABEL;
	}
	idx = drv->dev_index;
	if (!tape_devs[drv->cur_path] ||
        !(drv = tape_devs[drv->cur_path]->drv)||
          drv->remove_pending) {
        printk("lin_tape_open IBMtape%d removed or being removed\n",idx);
        READ_UNLOCK(&tape_des_lock);
        rc = -ENXIO;
        goto EXIT_LABEL;
	}
    READ_UNLOCK(&tape_des_lock);
    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if (drv->busy || drv->drv_open || drv->cmd_in_open){
        DbgPrint(("lin_tape_open: IBMtape%d busy %d\n",drv->dev_index, drv->busy));
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	}
	drv->cmd_in_open = in = True;
    WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	rc = lin_tape_find_drv_path(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if (drv->busy || drv->drv_open){
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EBUSY);

		printk("lin_tape: the device %s or another path to the same "\
			"drive is busy %d.\n", drv->devid, drv->busy);

		rc = -EBUSY;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		goto EXIT_LABEL;
	} /* if */

	if(drv->remove_pending || drv->ghost_device) {
		printk("lin_tape: the device %s being removed busy %d.\n",
            drv->devid,drv->busy);

		rc = -ENXIO;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		goto EXIT_LABEL;
	} /* if */

    DbgPrint(("lin_tape_open: IBMtape%d using current path %d\n",
              idx,drv->dev_index));
    rc = lin_tape_sdev_get(drv->dev_obj, &drv->sdev_ct);
	if( rc != STATUS_SUCCESS) {
        printk("lin_tape: IBMtape%d scsi device error %d\n",drv->dev_index,rc);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        goto EXIT_LABEL;
	} /* if */
    drv->busy |= F_OPEN;
    device_gotten = True;
    WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(ttype == LIN_TAPE_REGULAR_TRACE && drv->tape_log == NULL) {
		drv->tape_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		drv->tape_log_len = 0;
	} /* if */

	rc = lin_tape_drv_open_check(drv, ttype);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	if(minor < LinTapeNonRewindOffset)
		drv->special_file_type = RewindDevice;
	else
		drv->special_file_type = NonRewindDevice;

	if(drv->trace) {
		switch(ttype) {
		case LIN_TAPE_REGULAR_TRACE:
			tape_log_msg3(lin_tape_log, &lin_tape_log_len,
				TRC_TP_ATMPT_OPEN, drv->dev_index >> 8,
				drv->dev_index & 0xFF);
			break;

		case LIN_TAPE_FULL_TRACE:
			tape_setup_and_log_full(drv, drv->cur_path,
				TRC_SCSI_PATH, NULL, 0);

			tape_setup_and_log_full(drv, drv->dev_index,
				TRC_TP_ATMPT_OPEN, NULL, -4);
			break;

		default: break;
		} /* switch */
	} /* if */

	/* attempt to open IBMtape# */
	rc = lin_tape_drive_open(&drv, filp, reserve_on_open);
EXIT_LABEL:
    if (in) drv->cmd_in_open = False;
    if (drv && device_gotten){
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        if (rc != STATUS_SUCCESS){
            DbgPrint(("lin_tape_open_drive %d.  Removing refcount\n", rc));
            lin_tape_sdev_put(drv->dev_obj, &drv->sdev_ct);
        }
        drv->busy &= (0xFF-F_OPEN);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    } /* if */
	return rc;

} /* lin_tape_open_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_drive_des                                                *
*                                                                              *
*  PURPOSE: Get drv tape_descriptor from minor device number                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: receives the struct tape_descriptor*                                 *
*    minor: the minor device number, from which the drv is obtained            *
*    ttype: trace type                                                         *
*    reserve_on_open: (output) reserve this device when opening                *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_get_drive_des(struct tape_descriptor** drv, int minor,
	unchar ttype, unchar* reserve_on_open)
{
	int rc = STATUS_SUCCESS, idx = -1;

	if(!tape_devs) {
		DbgPrint(("lin_tape: corrupted tape entry table\n"));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_BAD_DEVS);

		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(minor < LinTapeNonRewindOffset)
        idx = minor;
	else {
		if(petro == 1 && minor >= LinTapeNonReserveOffset) {
			idx = minor - LinTapeNonReserveOffset;
			*reserve_on_open = False;
		} else idx = minor - LinTapeNonRewindOffset;
	} /* else */

	if (idx >= 0 && tape_devs[idx])
        *drv = tape_devs[idx]->drv;

	if(!*drv) {
		printk("lin_tape: device removed or not configured.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;
} /* lin_tape_get_drive_des() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_find_drv_path                                                *
*                                                                              *
*  PURPOSE: Find the best (or default) path to use for the operation           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: (input/output) tape descriptor                                       *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_find_drv_path(struct tape_descriptor** drv)
{
	int rc = STATUS_SUCCESS;

	/* Check if another path is better */

	if (!(*drv)->alt_pathing){
        return rc;
	}

	tape_get_best_path(drv);

	if(!*drv) {
		printk("lin_tape: drive has been removed or memory allocation error.\n");
		rc = -ENXIO;
	} /* if */

	return rc;

} /* lin_tape_find_drv_path() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drv_open_check                                               *
*                                                                              *
*  PURPOSE: Check this descriptor for common errors prior to attempting open   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor                                                      *
*    ttype: trace type                                                         *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_drv_open_check(struct tape_descriptor* drv, unchar ttype)
{
	int rc = STATUS_SUCCESS;

	if(drv->ghost_device) {
		printk("lin_tape: IBMtape%d remove in progress.\n",
			drv->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

/*	if(drv->drv_open) {
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EAGAIN);

		printk("lin_tape: the device %s or another path to the same "\
			"drive is already open.\n", drv->devid);

		rc = -EAGAIN;
		goto EXIT_LABEL;
	}*/ /* if */

	/* check the integrity of the buffer before open */
	if(drv->rw_buf) {
		printk("lin_tape: rw_buf is corrupted!\n");

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_BAD_RW_BUF);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_BAD_RW_BUF, 0, NULL);
			tape_log_rc(drv, EFAULT);
		}
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(drv->rw_buf_size != 0 || drv->rw_buf_order != -1) {
		printk("lin_tape: lin_tape_open rw_buf controls "\
			"are corrupted! rw_buf_size: %lu, rw_buf_order: %ld\n",
			drv->rw_buf_size, drv->rw_buf_order);

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_BAD_RW_CTLS);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_BAD_RW_CTLS, 0, NULL);
			tape_log_rc(drv, EFAULT);
		} /* else if */
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(drv->sg_list) {
		printk("lin_tape: scatter gather is corrupted!\n");
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_BAD_SG);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_BAD_SG, 0, NULL);
			tape_log_rc(drv, EFAULT);
		} /* else if */

		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;

} /* lin_tape_drv_open_check() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_DRV
#undef UNLOCK_DRV

#define LOCK_DES {READ_LOCK(&changer_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&changer_des_lock);}
#define LOCK_CHG {READ_LOCK(&chg->chg_entry_lock); chg_locked = True;}
#define UNLOCK_CHG {chg_locked = False; READ_UNLOCK(&chg->chg_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_open_changer                                                 *
*                                                                              *
*  PURPOSE: Perform open operations for changers                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    minor: the minor device number, from which the dev index is derived       *
*    ttype: trace type                                                         *
*    filp: Linux special file pointer                                          *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_open_changer(int minor, unchar ttype, struct file* filp)
{
	struct changer_descriptor *chg = NULL;
	int rc = STATUS_SUCCESS, device_gotten = False, idx = -1;

	READ_LOCK(&changer_des_lock);

	rc = lin_tape_get_changer_des(&chg, minor, ttype);
	if(rc != STATUS_SUCCESS || !chg) {
        READ_UNLOCK(&changer_des_lock);
            goto EXIT_LABEL;
	}
    idx = chg->dev_index;

    if (!changer_devs[chg->cur_path] ||
        !(chg = changer_devs[chg->cur_path]->chg)){
		DbgPrint(("lin_tape_open: current path %d for IBMchanger%d removed\n",
				chg->cur_path,idx));
        READ_UNLOCK(&changer_des_lock);
        rc = -ENXIO;
        goto EXIT_LABEL;
	} /* if */
    READ_UNLOCK(&changer_des_lock);
	WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);

	if(chg->remove_pending ){
		printk("lin_tape: the device %s is being removed busy %d.\n",
            chg->devid, chg->busy);
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

    if (!chg->chg_open){
        if(chg->busy ){
            printk("lin_tape: the device %s is busy %d.\n",
                chg->devid, chg->busy);
            WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
            rc = -EBUSY;
            goto EXIT_LABEL;
        } /* if */

        if(!chg->dev_obj || chg->ghost_device ||
            (rc = lin_tape_sdev_get(chg->dev_obj, &chg->sdev_ct)!=STATUS_SUCCESS)) {
            if(chg->alt_pathing) {
                WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
                rc = changer_failover_ops(&chg, HZ * 1800, 1);
                if (rc == STATUS_SUCCESS)
                    rc = lin_tape_sdev_get(chg->dev_obj, &chg->sdev_ct);
                WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
            } /* if */
            if(rc != STATUS_SUCCESS || !chg->alt_pathing) {
                printk("lin_tape: lin_tape_open: changer removed.\n");
                WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
                rc = -ENXIO;
                goto EXIT_LABEL;
            } /* if */
        } /* if */
        DbgPrint(("lin_tape_open: IBMChanger%d using current %d\n",
                  idx,chg->dev_index));
        device_gotten = True;
        if (chg->busy)
            DbgPrint(("lin_tape: IBMChanger%d busy on open %d\n", chg->dev_index, chg->busy));
        chg->busy |= F_OPEN;
    } /* if */
    else
        changer_devs[chg->cur_path]->chg->open_count++;
    WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);

	if(ttype == LIN_TAPE_REGULAR_TRACE && chg->chg_log == NULL) {
		chg->chg_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		chg->chg_log_len = 0;
	} /* if */

	switch(ttype) {
	case LIN_TAPE_REGULAR_TRACE:
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_CG_ATMPT_OPEN, chg->dev_index >> 8,
			chg->dev_index & 0xFF);
		break;

	case LIN_TAPE_FULL_TRACE:
		changer_setup_and_log_full(chg, chg->cur_path,
			TRC_SCSI_PATH, NULL, 0);

		changer_setup_and_log_full(chg, chg->cur_path,
			TRC_CG_ATMPT_OPEN, NULL, -4);
		break;

	default: break;
	} /* switch */

	/* attempt to open IBMChanger# */
    if (device_gotten)
        rc = lin_tape_changer_open(&chg, filp);

EXIT_LABEL:

	if (chg && device_gotten){
        WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
        if(rc != STATUS_SUCCESS) {
            DbgPrint(("lin_tape_open_changer %d.  Removing refcount\n",
                        rc));
            lin_tape_sdev_put(chg->dev_obj, &chg->sdev_ct);
        }
        chg->busy &= (0xFF-F_OPEN);
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
	} /* if */

	return rc;
} /* lin_tape_open_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_changer_des                                              *
*                                                                              *
*  PURPOSE: Get chg descriptor from minor device number                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: receives the struct changer_descriptor*                              *
*    minor: the minor device number, from which the chg is obtained            *
*    ttype: trace type                                                         *
*                                                                              *
*  RETURNS: 0 on success or error                                              *
*                                                                              *
*******************************************************************************/

int lin_tape_get_changer_des(struct changer_descriptor** chg, int minor,
	unchar ttype)
{
	int rc = 0, idx = minor-LinTapeChangerOffset;
	if(!changer_devs) {
		printk("lin_tape: lin_tape_open corrupted tape entry table\n");

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_BAD_DEVS);

		rc = -EIO;
		goto EXIT_LABEL;
	} else if(minor >= LinTapeMaxCount) {
		rc = -EIO;
		goto EXIT_LABEL;
	}

	if(!changer_devs[idx] || !changer_devs[idx]->chg) {
		printk("lin_tape: lin_tape_open device is not configured "\
			"or has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	}
	*chg = changer_devs[idx]->chg;

EXIT_LABEL:

	return rc;
} /* lin_tape_get_changer_des() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_CHG
#undef UNLOCK_CHG

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_close                                                        *
*                                                                              *
*  PURPOSE: close entry point for drives and changers                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    inode: i-node for the special file for this device                        *
*    filp: file pointer for this device                                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

static int lin_tape_close(struct inode *inode, struct file *filp)
{
	unsigned int minor;
	int rc = -EIO;

	minor = iminor(inode);
	DbgPrint2(("lin_tape_close entry: minor %d\n", minor));
	if(PSEUDO_TAPE(minor)) {
		WRITE_LOCK(&lin_tape_trace_lock);
		lin_tape_trace_type = LIN_TAPE_INVALID_TRACE;
		WRITE_UNLOCK(&lin_tape_trace_lock);
		rc = STATUS_SUCCESS;
	} else if(TSM_TAPE(minor)) {
		atomic_dec(&tsm_tp_open);
		rc = STATUS_SUCCESS;
	}else if(SPECIAL_DEV(minor)){
		atomic_dec(&special_tp_open);
		special_drv = NULL;
		rc = STATUS_SUCCESS;
	}else if(minor < LinTapeChangerOffset){
		rc = lin_tape_close_drive(filp, minor);
	} else if(minor < LinTapeMaxCount){
		rc = lin_tape_close_changer(filp, minor);
	} else {
		DbgPrint(("lin_tape: unsupported device.\n"));
		rc = -ENXIO;
	} /* else */

	DbgPrint2(("lin_tape_close exit: %d\n", rc));
	return rc;
} /* lin_tape_close() */

#define LOCK_DES {READ_LOCK(&tape_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&tape_des_lock);}
#define LOCK_DRV {READ_LOCK(&drv->drv_entry_lock); drv_locked = True;}
#define UNLOCK_DRV {drv_locked = False; READ_UNLOCK(&drv->drv_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_close_drive                                                  *
*                                                                              *
*  PURPOSE: close a tape drive                                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    filp: the file pointer of the special file we are attempting to close     *
*    minor: the minor device number of the drive we are attempting to remove   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_close_drive(struct file* filp, int minor)
{
	struct tape_descriptor* drv = NULL;
	int rc = -EIO, dev_index = -1, attempt_close = False;
	unchar ttype = lin_tape_trace_type, reserve_on_open = TRUE;

	READ_LOCK(&tape_des_lock);
	if(!tape_devs) {
		DbgPrint(("lin_tape_close: corrupted tape entry table\n"));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_TBL_CORRUPT);
		READ_UNLOCK(&tape_des_lock);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(minor < LinTapeNonRewindOffset)
		dev_index = minor;
	else {
		if((petro == 1) && (minor >= LinTapeNonReserveOffset)) {
			dev_index = minor - LinTapeNonReserveOffset;
			reserve_on_open = FALSE;
		} /* if */
		else dev_index = minor - LinTapeNonRewindOffset;
	} /* else */

	if(dev_index < 0 || !tape_devs[dev_index] ||
        !tape_devs[dev_index]->drv ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path] ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path]->drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_close: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	drv = tape_devs[tape_devs[dev_index]->drv->cur_path]->drv;
    dev_index = drv->dev_index;
    READ_UNLOCK(&tape_des_lock);
    WRITE_LOCK(&tape_devs[dev_index]->drv_entry_lock);
	if (!drv->drv_open){
        WRITE_UNLOCK(&tape_devs[dev_index]->drv_entry_lock);
		printk("lin_tape: lin_tape_close drive is not open.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	}
	if(drv->busy) {
        DbgPrint(("lin_tape_close: Index %d, Busy %d\n",
                  drv->dev_index, drv->busy));
		/* lin_tape_close failed, device busy */
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_CLOSE_BSY);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_CLOSE_BSY, 0, NULL);
			tape_log_rc(drv, EBUSY);
		}
        WRITE_UNLOCK(&tape_devs[dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	drv->busy |= F_CLOSE;
	attempt_close = True;
	WRITE_UNLOCK(&tape_devs[dev_index]->drv_entry_lock);

	/* attempt to close IBMtape# */
	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
				TRC_TP_ATMPT_CLOSE, dev_index >> 8,
				dev_index & 0xFF);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {

		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);

		tape_setup_and_log_full(drv, dev_index, TRC_TP_ATMPT_CLOSE,
			NULL, -4);
	} /* else if */
	rc = lin_tape_drive_close(&drv, filp, reserve_on_open);
EXIT_LABEL:

    if (drv && attempt_close) {
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        drv->busy &= (0xFF-F_CLOSE);
        if(drv->remove_pending) {
            printk("lin_tape_close: IBMtape%d being removed.\n", drv->dev_index);
            WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
            cleanup_and_free_tdesc(drv->dev_index);
        } /* if */
        else WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    }
	return rc;
} /* lin_tape_close_drive() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_DRV
#undef UNLOCK_DRV

#define LOCK_DES {READ_LOCK(&changer_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&changer_des_lock);}
#define LOCK_CHG {READ_LOCK(&chg->chg_entry_lock); chg_locked = True;}
#define UNLOCK_CHG {chg_locked = False; READ_UNLOCK(&chg->chg_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_close_changer                                                *
*                                                                              *
*  PURPOSE: close a medium changer                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    filp: the file pointer of the special file we are attempting to close     *
*    minor: the minor device number of the changer we are attempting to close  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_close_changer(struct file* filp, int minor)
{
	struct changer_descriptor *chg = NULL;
	int rc = -EIO;
	int dev_index = minor-LinTapeChangerOffset;
	unchar ttype = lin_tape_trace_type;

	READ_LOCK(&changer_des_lock);
	if(!changer_devs || minor >= LinTapeMaxCount ||
        !changer_devs[dev_index] || !changer_devs[dev_index]->chg ||
        !changer_devs[changer_devs[dev_index]->chg->cur_path] ||
        !(chg=changer_devs[changer_devs[dev_index]->chg->cur_path]->chg) ||
        chg->open_count <= 0) {
		printk("lin_tape_close: corrupted tape entry table\n");

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_TBL_CORRUPT);
        READ_UNLOCK(&changer_des_lock);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
	READ_UNLOCK(&changer_des_lock);

	if( chg->busy && chg->open_count == 1) {
        DbgPrint(("lin_tape: IBMChanger%d busy %d\n", chg->dev_index, chg->busy))
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_CLOSE_BSY);
		else if(ttype == LIN_TAPE_FULL_TRACE) {
			changer_log_full_trace(chg, TRC_CLOSE_BSY, 0, NULL);
			changer_log_rc(chg, EBUSY);
		}
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */

	if (chg->open_count > 1){
        chg->open_count --;
        rc = STATUS_SUCCESS;
	}
    WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);

	/* attempt to close IBMChanger# */
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_CG_ATMPT_CLOSE, dev_index >> 8, dev_index & 0xFF);
	else if(ttype == LIN_TAPE_FULL_TRACE) {
		changer_setup_and_log_full(chg, chg->cur_path, TRC_SCSI_PATH,
			NULL, 0);

		changer_setup_and_log_full(chg, dev_index, TRC_CG_ATMPT_CLOSE,
			NULL, -4);
	} /* else if */

	if (rc != STATUS_SUCCESS)
        rc = lin_tape_changer_close(&chg, filp);

EXIT_LABEL:
    if (chg && chg->remove_pending &&chg->open_count == 0){
        cleanup_and_free_cdesc(chg->dev_index);
    }

	return rc;
} /* lin_tape_close_changer() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_CHG
#undef UNLOCK_CHG

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ioctl                                                        *
*                                                                              *
*  PURPOSE: IOCTL entry point for drives and changers                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    inode: i-node for the special file to which this request is sent          *
*    filp: file pointer of the special file to which this request is sent      *
*    command: the IOCTL code                                                   *
*    arg: the data required to carry out the IOCTL                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
static int lin_tape_ioctl(struct inode* inode, struct file* filp, uint command,
	ulong arg)
#else
static long lin_tape_ioctl(struct file* filp, uint command, ulong arg)
#endif
{
	int rc = -EIO;
	int minor = -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	struct inode *inode = file_inode(filp);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
	struct inode* inode = filp->f_dentry->d_inode;
#endif
#endif

	minor = iminor(inode);

	if(command != LIN_TAPE_POLL_TRACE){
		DbgPrint2(("lin_tape_ioctl entry: minor %d command 0x%x\n",
			minor, (int)command));
    }
	if(PSEUDO_TAPE(minor)) {
		rc = lin_tape_daemon_ioctl(command, arg);
	} else if(TSM_TAPE(minor)) {
		rc = lin_tape_tsm_ioctl(command, arg);
	} else if(SPECIAL_DEV(minor)) {
		rc = lin_tape_ioctl_special(&special_drv, command, arg);
	} else if(minor < LinTapeChangerOffset) {
		rc = lin_tape_ioctl_drive(minor, filp, command, arg);
	} else {
		rc = lin_tape_ioctl_changer(minor, filp, command, arg);
	} /* else */

	if(command != LIN_TAPE_POLL_TRACE){
    DbgPrint2(("lin_tape_ioctl exit: %d\n", rc));
	}
	return rc;
} /* lin_tape_ioctl() */

#define LOCK_DES {READ_LOCK(&tape_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&tape_des_lock);}
#define LOCK_DRV {READ_LOCK(&drv->drv_entry_lock); drv_locked = True;}
#define UNLOCK_DRV {drv_locked = False; READ_UNLOCK(&drv->drv_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ioctl_drive                                                  *
*                                                                              *
*  PURPOSE: IOCTL handler for drives                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    minor: device minor number for this request                               *
*    filp: file pointer of the special file to which this request is sent      *
*    command: the IOCTL code                                                   *
*    arg: the data required to carry out the IOCTL                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int lin_tape_ioctl_drive(int minor, struct file* filp, uint command, ulong arg)
{
	int rc = -EIO, dev_index = -1;
	struct tape_descriptor* drv = NULL;
	boolean device_gotten = False;

	READ_LOCK(&tape_des_lock);
	if(!tape_devs) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: corrupted tape entry table\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(minor < LinTapeNonRewindOffset) {
		dev_index = minor;
	} else {
		if(petro == 1 && minor >= LinTapeNonReserveOffset) {
            dev_index = minor - LinTapeNonReserveOffset;
		} else dev_index = minor - LinTapeNonRewindOffset;
	} /* else */

	if(dev_index < 0 || !tape_devs[dev_index] ||
        !tape_devs[dev_index]->drv ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path]
        || !tape_devs[tape_devs[dev_index]->drv->cur_path]->drv ||
        tape_devs[tape_devs[dev_index]->drv->cur_path]->drv->remove_pending) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_ioctl: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	/* use the cur_path */
	drv = tape_devs[tape_devs[dev_index]->drv->cur_path]->drv;
    READ_UNLOCK(&tape_des_lock);
    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	if(	drive_busy(tape_devs[drv->dev_index]->drv, F_ADD_REMOVE)) {
        printk("lin_tape_ioctl: IBMtape%d busy %d\n", drv->dev_index, drv->busy);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	if(drv->remove_pending) {
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		printk("lin_tape_ioctl: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
    drv->busy |= F_IOCTL;
    device_gotten = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	rc = lin_tape_drive_ioctl(&drv, filp, command, arg);

EXIT_LABEL:

    if (device_gotten && drv){
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        drv->busy &= (0xFF-F_IOCTL);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    }
	return rc;
} /* lin_tape_ioctl_drive() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_DRV
#undef UNLOCK_DRV

#define LOCK_DES {READ_LOCK(&changer_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&changer_des_lock);}
#define LOCK_CHG {READ_LOCK(&chg->chg_entry_lock); chg_locked = True;}
#define UNLOCK_CHG {chg_locked = False; READ_UNLOCK(&chg->chg_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ioctl_changer                                                *
*                                                                              *
*  PURPOSE: IOCTL handler for medium changers                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    minor: device minor number for this request                               *
*    filp: file pointer of the special file to which this request is sent      *
*    command: the IOCTL code                                                   *
*    arg: the data required to carry out the IOCTL                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_ioctl_changer(int minor, struct file* filp, uint command,
	ulong arg)
{
	int rc = -EIO;
	int dev_index = minor - LinTapeChangerOffset;
	struct changer_descriptor* chg = NULL;

	READ_LOCK(&changer_des_lock);
	if(!changer_devs || !changer_devs[dev_index] ||
        !changer_devs[dev_index]->chg ||
        !changer_devs[changer_devs[dev_index]->chg->cur_path] ||
        !(chg = changer_devs[changer_devs[dev_index]->chg->cur_path]->chg)||
        chg->remove_pending) {
        READ_UNLOCK(&changer_des_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&changer_des_lock);

	rc = lin_tape_changer_ioctl(&chg, filp, command, arg);

EXIT_LABEL:

	return rc;

} /* lin_tape_ioctl_changer() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_CHG
#undef UNLOCK_CHG

#define LOCK_DES {READ_LOCK(&tape_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&tape_des_lock);}
#define LOCK_DRV {READ_LOCK(&drv->drv_entry_lock); drv_locked = True;}
#define UNLOCK_DRV {drv_locked = False; READ_UNLOCK(&drv->drv_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_read                                                         *
*                                                                              *
*  PURPOSE: tape drive read handler                                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    file: file pointer of the special file to which this request is sent      *
*    user_location: data area for read data to be placed                       *
*    xfer_len: number of bytes to read                                         *
*    offset: unused by this driver                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_read(struct file* file, char* user_location,
	size_t xfer_len, loff_t* offset)
{
	ssize_t rc = -EIO;
	struct tape_descriptor* drv = NULL;
	int minor = -1, dev_index = -1, device_gotten = False;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	minor = iminor(file_inode(file));
#else
	minor = iminor(file->f_dentry->d_inode);
#endif
	DbgPrint2(("lin_tape_read entry: minor %d\n", minor));

	if(minor >= LinTapeChangerOffset) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	READ_LOCK(&tape_des_lock);
	if(!tape_devs) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_read: no tape devices table\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(minor < LinTapeNonRewindOffset){
		dev_index = minor;
	}
	else {
        if(petro == 1 && minor >= LinTapeNonReserveOffset){
            dev_index = minor - LinTapeNonReserveOffset;
        }
        else {
            dev_index = minor - LinTapeNonRewindOffset;
        }
	}

	if(dev_index < 0 || !tape_devs[dev_index] ||
        !tape_devs[dev_index]->drv ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path] ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path]->drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_read: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	/* use the cur_path */
	drv = tape_devs[tape_devs[dev_index]->drv->cur_path]->drv;
    dev_index = drv->dev_index;
    READ_UNLOCK(&tape_des_lock);
	WRITE_LOCK(&tape_devs[dev_index]->drv_entry_lock);
	if(drv->remove_pending ||
        drive_busy(tape_devs[drv->dev_index]->drv, F_ADD_REMOVE)) {
        DbgPrint(("lin_tape_read: IBMtape%d busy %d or being removed\n",
                  dev_index, tape_devs[dev_index]->drv->busy));
        WRITE_UNLOCK(&tape_devs[dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	drv->busy |= F_READ;
	device_gotten = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	rc = lin_tape_drive_read(&drv, file, user_location,
				xfer_len, offset);
EXIT_LABEL:

	DbgPrint2(("lin_tape_read exit: %d\n", (int)rc));

    if (device_gotten && drv){
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        drv->busy &= (0xFF-F_READ);
        /* check if class_device_remove has been called */
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    }
	return rc;
} /* lin_tape_read() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_write                                                        *
*                                                                              *
*  PURPOSE: tape drive write handler                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    file: file pointer of the special file to which this request is sent      *
*    user_location: data area for read data to be placed                       *
*    xfer_len: number of bytes to write                                        *
*    offset: unused by this driver                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

static ssize_t lin_tape_write(struct file* file, const char* user_location,
	size_t xfer_len, loff_t* offset)
{
	ssize_t rc = -EIO;
	struct tape_descriptor *drv = NULL;
	int minor = -1, dev_index = -1, device_gotten = True;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	minor = iminor(file_inode(file));
#else
	minor = iminor(file->f_dentry->d_inode);
#endif

	DbgPrint2(("lin_tape_write entry: minor %d\n", minor));

	if(minor >= LinTapeChangerOffset) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	READ_LOCK(&tape_des_lock);
	if(!tape_devs) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_write: no tape devices table\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(minor < LinTapeNonRewindOffset)
		dev_index = minor;
	else if(petro == 1 && minor >= LinTapeNonReserveOffset)
		dev_index = minor - LinTapeNonReserveOffset;
	else dev_index = minor - LinTapeNonRewindOffset;

	if(dev_index < 0 || !tape_devs[dev_index] ||
        !tape_devs[dev_index]->drv ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path] ||
        !tape_devs[tape_devs[dev_index]->drv->cur_path]->drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_write: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	/* use the cur_path */

	drv = tape_devs[tape_devs[dev_index]->drv->cur_path]->drv;
    dev_index = drv->dev_index;
    WRITE_LOCK(&tape_devs[dev_index]->drv_entry_lock);
    READ_UNLOCK(&tape_des_lock);
	if(drv->remove_pending || drive_busy(tape_devs[dev_index]->drv, F_ADD_REMOVE)) {
        DbgPrint(("lin_tape_write: IBMtape%d busy %d or being removed\n",
                  drv->dev_index, drv->busy));
        WRITE_UNLOCK(&tape_devs[dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	drv->busy |= F_WRITE;
	device_gotten = True;
    WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	rc = lin_tape_drive_write(&drv, file, user_location,
				xfer_len, offset);

EXIT_LABEL:

	DbgPrint2(("lin_tape_write exit: %d\n", (int)rc));

    if (drv && device_gotten){
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        drv->busy &= (0xFF-F_WRITE);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    }
	return rc;
} /* lin_tape_write() */

void copy_wwn_from_tape_devs(struct changer_descriptor* chg)
{
  int j = 0, i = -1;
  char buffer[32], buff[32];

  if (!tape_devs) return;

  if (!chg) return;

  for(j = 0; j < MaxLinTapeDevices; j++) {
	  if(tape_devs[j] && tape_devs[j]->drv){
		  if ( chg->host == tape_devs[j]->drv->host
			   && chg->channel == tape_devs[j]->drv->channel
			   && chg->target == tape_devs[j]->drv->target)
		  {
		  	  DbgPrint(("lin_tape: lname=%s, pname=%s, host=%d, channel=%d, target=%d, lun=%d\n",
		  			  chg->lname,
					  chg->pname,
					  chg->host,
					  chg->channel,
					  chg->target,
					  chg->lun));

	  	 	   memset(buffer, '\0', 32);
			   memcpy(buffer, "0x", strlen("0x"));
			   for(i = 0; i < WWN_LEN; i++) {
				sprintf(&buffer[strlen(buffer)], "%02X",
					tape_devs[j]->drv->wwnn[i] & 0xFF);
			   }
   	  	 	   memset(buff, '\0', 32);
			   memcpy(buff, "0x", strlen("0x"));
			   for(i = 0; i < WWN_LEN; i++) {
				sprintf(&buff[strlen(buff)], "%02X",
					tape_devs[j]->drv->wwpn[i] & 0xFF);
			   }

		  	  DbgPrint(("lin_tape: lname=%s, pname=%s, host=%d, channel=%d, target=%d, lun=%d, wwnn=%s, wwpn=%s\n",
		  			  tape_devs[j]->drv->lname,
					  tape_devs[j]->drv->pname,
					  tape_devs[j]->drv->host,
					  tape_devs[j]->drv->channel,
					  tape_devs[j]->drv->target,
					  tape_devs[j]->drv->lun,
					  buffer,
					  buff));

		  	  memcpy(chg->wwnn, tape_devs[j]->drv->wwnn, WWN_LEN);
		  	  memcpy(chg->wwpn, tape_devs[j]->drv->wwpn, WWN_LEN);
		  }
	  }
  }
}

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_DRV
#undef UNLOCK_DRV

void lin_tape_pfoID_hctl_update (struct pfo_ID *pfo_id,
								struct tape_descriptor *drv)
{
	if (pfo_id && drv){
		pfo_id->hctl.host = drv->host;
		pfo_id->hctl.channel = drv->channel;
		pfo_id->hctl.target = drv->target;
		pfo_id->hctl.lun = drv->lun;
	}
}

void lin_tape_chg_pfoID_hctl_update (struct pfo_ID *pfo_id,
								struct changer_descriptor *drv)
{
	if (pfo_id && drv){
		pfo_id->hctl.host = drv->host;
		pfo_id->hctl.channel = drv->channel;
		pfo_id->hctl.target = drv->target;
		pfo_id->hctl.lun = drv->lun;
	}
}

int pfo_minor_match(struct pfo_ID *pfo_id, int typed){

	int dev;

	if (typed){
		dev = pfo_id->pfo_minor - LinTapeNonRewindOffset;
		if( tape_devs[dev] && tape_devs[dev]->drv
			&& tape_devs[dev]->drv->serial_num_len == pfo_id->serial_num_len &&
			memcmp(tape_devs[dev]->drv->serial_num, pfo_id->serial_num,
				   tape_devs[dev]->drv->serial_num_len) == 0 &&
			memcmp(tape_devs[dev]->drv->wwnn, pfo_id->wwnn, WWN_LEN) == 0){
				return 1;
			}
	}
	else{
		dev = pfo_id->pfo_minor - LinTapeChangerOffset;
		if( changer_devs[dev] && changer_devs[dev]->chg &&
			changer_devs[dev]->chg->serial_num_len == pfo_id->serial_num_len &&
			memcmp(changer_devs[dev]->chg->serial_num, pfo_id->serial_num,
					changer_devs[dev]->chg->serial_num_len) == 0 /*&&
			memcmp(changer_devs[dev]->wwnn, pfo_id->wwnn, WWN_LEN) == 0*/){
				return 1;
			}
	}
	return 0;
}

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_drive_pfodes                                             *
*                                                                              *
*  PURPOSE: Search the drive descriptor on non-rewind descriptors              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: (out) the drive descriptor                                           *
*    ttype: trace type                                                         *
*    reserve_on_open: if driver is set to reserve implicitly                   *
*    pfo: the pfo data with device identification                              *
*    minor: (out) the minor for this drv                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int lin_tape_get_drive_pfodes(struct tape_descriptor** drv, unchar ttype,
		                      unchar* reserve_on_open, struct pfo_ID *pfo_id,
		                      int *minor)
{
	int rc = STATUS_SUCCESS, i;
	*minor=-1;

	if (!pfo_id){
		rc = -EINVAL;
		goto EXIT_LABEL;
	}
	else pfo_id->pfo_minor = *minor;

	if(!tape_devs) {
		DbgPrint(("lin_tape: corrupted tape entry table\n"));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_BAD_DEVS);

		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i<LinTapeNonRewindOffset; i++) {
		if(tape_devs[i] == NULL || tape_devs[i]->drv == NULL)
			continue;
		if((tape_devs[i]->drv->serial_num_len > 0) &&
			(tape_devs[i]->drv->serial_num_len
				== pfo_id->serial_num_len)
			&& (memcmp(tape_devs[i]->drv->serial_num,
				pfo_id->serial_num,
				tape_devs[i]->drv->serial_num_len) == 0)) {
				if ( memcmp(tape_devs[i]->drv->wwnn, pfo_id->wwnn, WWN_LEN) == 0){
				*minor = i + LinTapeNonRewindOffset;
				break;
			}
		} /* if */
	} /* for */

	if (*minor < LinTapeNonRewindOffset || *minor >= LinTapeChangerOffset ){
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	//For pfo interface NonRewind devices will be used
	*drv = tape_devs[*minor - LinTapeNonRewindOffset]->drv;

	if(!*drv) {
		DbgPrint(("lin_tape: device removed or not configured.\n"));
		rc = -ENXIO;
	} /* if */

	if(!(*drv)->alt_pathing) {
		do{
          if ((*drv)->host == pfo_id->hctl.host &&
            (*drv)->channel == pfo_id->hctl.channel &&
            (*drv)->target == pfo_id->hctl.target &&
            (*drv)->lun == pfo_id->hctl.lun){
              goto EXIT_LABEL;
            }
            i++;
            for(; i < LinTapeNonRewindOffset; i++) {
        		if(!tape_devs[i] || !tape_devs[i]->drv )
        			continue;
        		if((tape_devs[i]->drv->serial_num_len > 0) &&
        			(tape_devs[i]->drv->serial_num_len
        				== pfo_id->serial_num_len)    &&
        			(memcmp(tape_devs[i]->drv->serial_num,
        			        pfo_id->serial_num,
        				    tape_devs[i]->drv->serial_num_len) == 0) &&
        			memcmp(tape_devs[i]->drv->wwnn, pfo_id->wwnn, WWN_LEN) == 0){
    				  *minor = i + LinTapeNonRewindOffset;
    				  *drv = tape_devs[i]->drv;
                    break;
                } /* if */
            } /* for */
		}while(i < LinTapeNonRewindOffset);/* do while */
		*drv = NULL;
		*minor = -1;
		rc = -ENXIO;
	}

EXIT_LABEL:

	return rc;
} /* lin_tape_get_drive_pfodes() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_changer_pfodes                                           *
*                                                                              *
*  PURPOSE: Search the changer descriptor                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: (out) the changer descriptor                                         *
*    minor: (out) the minor for this chg                                       *
*    ttype: trace type                                                         *
*    pfo_id: the pfo data with device identification                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int lin_tape_get_changer_pfodes(struct changer_descriptor** chg, int *minor,
	unchar ttype, struct pfo_ID *pfo_id)
{
	int rc = STATUS_SUCCESS, i, dev_index;
	*minor = -1;

	if (!pfo_id){
		rc = -EINVAL;
		goto EXIT_LABEL;
	}
	else pfo_id->pfo_minor = *minor;

	if(!changer_devs) {
		DbgPrint(("lin_tape: corrupted tape entry table\n"));
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_BAD_DEVS);
		rc = -EIO;
		goto EXIT_LABEL;
	}

	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(changer_devs[i] == NULL || changer_devs[i]->chg == NULL)
			continue;
		if((changer_devs[i]->chg->serial_num_len > 0) &&
		   (changer_devs[i]->chg->serial_num_len
					== pfo_id->serial_num_len)
			&& (memcmp(changer_devs[i]->chg->serial_num,
						pfo_id->serial_num,
						changer_devs[i]->chg->serial_num_len)
						== 0)) {
				*minor = i+LinTapeChangerOffset;
				break;
		} /* if */
	} /* for */
    dev_index = *minor - LinTapeChangerOffset;
	if(*minor < 0 || *minor >= LinTapeMaxCount ||
        !changer_devs[dev_index] || !changer_devs[dev_index]->chg) {
		DbgPrint(("lin_tape: Changer device is not configured "\
			"or has been removed.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	}
    (*chg)=changer_devs[dev_index]->chg;
	if(!(*chg)->alt_pathing) {
		do{
          if ((*chg)->host == pfo_id->hctl.host &&
            (*chg)->channel == pfo_id->hctl.channel &&
            (*chg)->target == pfo_id->hctl.target &&
            (*chg)->lun == pfo_id->hctl.lun){
              goto EXIT_LABEL;
            }
            i++;
            for(; i < MaxLinTapeChangerDevices; i++) {
              if(!changer_devs[i] || !changer_devs[i]->chg )
                continue;
              if((changer_devs[i]->chg->serial_num_len > 0) &&
                 (changer_devs[i]->chg->serial_num_len
						== pfo_id->serial_num_len)     &&
                 (memcmp(changer_devs[i]->chg->serial_num,
                         pfo_id->serial_num,
                         changer_devs[i]->chg->serial_num_len)== 0)) {
                    *minor = i+LinTapeChangerOffset;
                	*chg = changer_devs[i]->chg;
                    break;
              } /* if */
            } /* for */
		}while(i < MaxLinTapeChangerDevices);/* do while */
		*chg = NULL;
		*minor = -1;
		rc = -ENXIO;
	}

EXIT_LABEL:

	return rc;
} /* lin_tape_get_changer_pfodes() */


/**************************************JOIN INTERFACE*************************/
#define LOCK_DES {READ_LOCK(&tape_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&tape_des_lock);}
#define LOCK_DRV {READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock); drv_locked = True;}
#define UNLOCK_DRV {drv_locked = False; READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);}

/*******************************************************************************
*                                                                              *
*  NAME: pfo_open_drive                                                        *
*                                                                              *
*  PURPOSE: pfo interface to open a drive                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*                                           READ_UNLOCK(&tape_des_lock);                                       *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_open_drive(struct pfo_ID *pfo_id)
{
	struct tape_descriptor *drv = NULL;
	struct file filp;
	int rc = STATUS_SUCCESS, device_gotten = False, minor=-1, in=False;
	unchar ttype = lin_tape_trace_type;
	unchar reserve_on_open = True;

	READ_LOCK(&tape_des_lock);
	rc = lin_tape_get_drive_pfodes(&drv, ttype, &reserve_on_open, pfo_id, &minor);
	if(!drv || rc != STATUS_SUCCESS) {
            READ_UNLOCK(&tape_des_lock);
            goto EXIT_LABEL;
	}
	if(!drv || !tape_devs[drv->cur_path] || !tape_devs[drv->cur_path]->drv
        || tape_devs[drv->cur_path]->drv->remove_pending) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: lin_tape_open drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	drv = tape_devs[drv->cur_path]->drv;
    READ_UNLOCK(&tape_des_lock);
	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if (drv->busy || drv->drv_open || drv->cmd_in_open){
		rc = -EBUSY;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		goto EXIT_LABEL;
	}
	drv->cmd_in_open = in = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

 	rc = lin_tape_find_drv_path(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if (drv->busy || drv->drv_open){
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EBUSY);

		printk("lin_tape: the device %s or another path to the same "\
			"drive is busy %d.\n", drv->devid,drv->busy);

		rc = -EBUSY;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		goto EXIT_LABEL;
	} /* if */

	if(drv->remove_pending || drv->ghost_device) {
		printk("lin_tape: the device %s being removed busy %d.\n",
            drv->devid,drv->busy);

		rc = -ENXIO;
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		goto EXIT_LABEL;
	} /* if */

	rc = lin_tape_sdev_get(drv->dev_obj, &drv->sdev_ct);
	if( rc != STATUS_SUCCESS) {
        printk("lin_tape: IBMtape%d scsi device error %d\n",drv->dev_index,rc);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        goto EXIT_LABEL;
	} /* if */
    drv->busy |= F_OPEN;
    device_gotten = True;

    WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(ttype == LIN_TAPE_REGULAR_TRACE && drv->tape_log == NULL) {
		drv->tape_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		drv->tape_log_len = 0;
	} /* if */

    rc = lin_tape_drv_open_check(drv, ttype);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
    filp.f_flags = pfo_id->f_flags;
	drv->special_file_type = NonRewindDevice;

	if(drv->trace) {
		switch(ttype) {
		case LIN_TAPE_REGULAR_TRACE:
			tape_log_msg3(lin_tape_log, &lin_tape_log_len,
				TRC_TP_ATMPT_OPEN, drv->dev_index >> 8,
				drv->dev_index & 0xFF);
			break;

		case LIN_TAPE_FULL_TRACE:
			tape_setup_and_log_full(drv, drv->cur_path,
				TRC_SCSI_PATH, NULL, 0);

			tape_setup_and_log_full(drv, drv->dev_index,
				TRC_TP_ATMPT_OPEN, NULL, -4);
			break;

		default: break;
		} /* switch */
	} /* if */

	/* attempt to open IBMtape# */
	rc = lin_tape_drive_open(&drv, &filp, reserve_on_open);

	EXIT_LABEL:
    if (in) drv->cmd_in_open = False;
	if (drv && device_gotten){
       WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        if (rc == STATUS_SUCCESS) {
            lin_tape_pfoID_hctl_update(pfo_id, drv);
            pfo_id->pfo_minor = minor;
        }
        if (rc != STATUS_SUCCESS){
            DbgPrint(("lin_tape_open_drive %d.  Removing refcount\n", rc));
            lin_tape_sdev_put(drv->dev_obj, &drv->sdev_ct);
        }
        drv->busy &= (0xFF-F_OPEN);
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	}

	return rc;

} /* pfo_open_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: pfo_tape_pass_thru                                                    *
*                                                                              *
*  PURPOSE: pfo interface to pass cdb to drive                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*    cdb:                                                                      *
*    cmd_length:                                                               *
*    buffer:                                                                   *
*    buffer_length:                                                            *
*    data_direction:                                                           *
*    timeout:                                                                  *
*    sense:                                                                    *
*    sense_length:                                                             *
*    sense_avail:                                                              *
*    resid:                                                                    *
*    errors:                                                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_tape_pass_thru(struct pfo_ID *pfo_id, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense,
	unchar *sense_length, unchar *sense_avail, int *resid,
	int *errors)
{
	int rc = -EIO, minor = -1;
	struct tape_descriptor* drv = NULL;
	unchar ttype = lin_tape_trace_type;

	if (!pfo_id) {
		rc = -EINVAL;
	}
	minor = pfo_id->pfo_minor;
	READ_LOCK(&tape_des_lock);
	//For pfo interface NonRewind devices will be used
	if (minor >= LinTapeNonRewindOffset
		&& tape_devs
		&& pfo_minor_match(pfo_id, 1))
			drv = tape_devs[minor - LinTapeNonRewindOffset]->drv;
	else{
        READ_UNLOCK(&tape_des_lock);
		DbgPrint(("lin_tape_pfo_passthru: corrupted minor or tape table\n"));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_TBL_CORRUPT);
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		DbgPrint(("lin_tape: device removed or not configured.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	DbgPrint(("CDB 0x%x sense SK/ASC/ASCQ %x %x %x resid 0x%x errors %x\n",
           cdb[0], sense[2] & 0x0F, sense[12], sense[13],*resid,*errors));
	drv = tape_devs[drv->cur_path]->drv;
	if(!drv || drv->remove_pending) {
        READ_UNLOCK(&tape_des_lock);
		DbgPrint(("lin_tape_ioctl: drive has been removed.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(drv->busy) {
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */

    WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	rc = lin_tape_pass_thru_from_pfoID(&drv, cdb, cmd_length, buffer,
			      buffer_length, data_direction, timeout, sense, sense_length,
				  sense_avail, resid, errors);

EXIT_LABEL:
    if (drv){
		lin_tape_pfoID_hctl_update(pfo_id, drv);
		if (rc == -ENXIO)
			pfo_id->pfo_minor = -1;
		else pfo_id->pfo_minor = drv->dev_index;
		DbgPrint(("CDB 0x%x sense SK/ASC/ASCQ %x %x %x resid %x errors %x rc %x\n",
             cdb[0], sense[2] & 0x0F, sense[12], sense[13],*resid,*errors, rc));
    }
	return rc;
} /* pfo_tape_pass_thru() */

/*******************************************************************************
*                                                                              *
*  NAME: pfo_close_drive                                                       *
*                                                                              *
*  PURPOSE: pfo interface to close a drive                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_close_drive(struct pfo_ID *pfo_id)
{
	struct tape_descriptor* drv = NULL;
	struct file filp;
	int rc = -EIO;
	int dev_index = 0, minor=-1, attempt_close = False;
	unchar ttype = lin_tape_trace_type;
	unchar reserve_on_open = TRUE;

	if (!pfo_id) {
		rc = -EINVAL;
	}
	minor = pfo_id->pfo_minor;

	READ_LOCK(&tape_des_lock);
	if (minor >= LinTapeNonRewindOffset
		&& tape_devs
		&& pfo_minor_match(pfo_id, 1))
			drv = tape_devs[minor - LinTapeNonRewindOffset]->drv;
	else{
        READ_UNLOCK(&tape_des_lock);
		DbgPrint(("lin_tape_close: corrupted minor or tape table\n"));

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_TBL_CORRUPT);
		rc = -EIO;
		goto EXIT_LABEL;
	}

	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_close: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	drv = tape_devs[drv->cur_path]->drv;
	if(!drv) {
	    READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: lin_tape_close drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(drv->busy) {
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		/* lin_tape_close failed, device busy */
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_CLOSE_BSY);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_CLOSE_BSY, 0, NULL);
			tape_log_rc(drv, EBUSY);
		}
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	drv->busy |= F_CLOSE;
	attempt_close = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	/* attempt to close IBMtape# */
	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
				TRC_TP_ATMPT_CLOSE, dev_index >> 8,
				dev_index & 0xFF);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {

		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);

		tape_setup_and_log_full(drv, dev_index, TRC_TP_ATMPT_CLOSE,
			NULL, -4);
	} /* else if */
	rc = lin_tape_drive_close(&drv, &filp, reserve_on_open);

EXIT_LABEL:

	if(rc == STATUS_SUCCESS){
		if (drv) lin_tape_pfoID_hctl_update(pfo_id, drv);
		if (pfo_id) pfo_id->pfo_minor = -1;
	}
    if (drv && attempt_close) {
        WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
        drv->busy &= (0xFF-F_CLOSE);
        if(drv->remove_pending) {
            printk("lin_tape_close: IBMtape%d being removed.\n", drv->dev_index);
            WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
            cleanup_and_free_tdesc(drv->dev_index);
        } /* if */
        else WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    }
	return rc;
} /* pfo_close_drive() */

#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_DRV
#undef UNLOCK_DRV

#define LOCK_DES {READ_LOCK(&changer_des_lock); des_locked = True;}
#define UNLOCK_DES {des_locked = False; READ_UNLOCK(&changer_des_lock);}
#define LOCK_CHG {READ_LOCK(&changer_devs[chg->dev_index]->entry_lock); chg_locked = True;}
#define UNLOCK_CHG {chg_locked = False; READ_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);}
/*******************************************************************************
*                                                                              *
*  NAME: pfo_open_changer                                                      *
*                                                                              *
*  PURPOSE: pfo interface to open a changer                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_open_changer(struct pfo_ID *pfo_id)
{
	struct changer_descriptor *chg = NULL;
	struct file filp;
	int rc = STATUS_SUCCESS, device_gotten = False, minor;
	unchar ttype = lin_tape_trace_type;

	READ_LOCK(&changer_des_lock);
	rc = lin_tape_get_changer_pfodes(&chg, &minor, ttype, pfo_id);
	if(rc != STATUS_SUCCESS ){
            READ_UNLOCK(&changer_des_lock);
            goto EXIT_LABEL;
	}

	if(chg == NULL || minor < 0 ||
        !changer_devs[chg->cur_path] ||
        !changer_devs[chg->cur_path]->chg) {
        READ_UNLOCK(&changer_des_lock);
		DbgPrint(("lin_tape: Changer device is not configured "\
			"or has been removed.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	filp.f_flags = pfo_id->f_flags;
	chg = changer_devs[chg->cur_path]->chg;

	if(!chg || chg->remove_pending) {
        READ_UNLOCK(&changer_des_lock);
		DbgPrint(("lin_tape: lin_tape_open changer has been removed.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
    READ_UNLOCK(&changer_des_lock);
	WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
    if (!chg->chg_open){
        if(!chg->dev_obj || chg->ghost_device ||
            (rc = lin_tape_sdev_get(chg->dev_obj, &chg->sdev_ct)!=STATUS_SUCCESS)) {
            if(chg->alt_pathing) {
                WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
                rc = changer_failover_ops(&chg, HZ * 1800, 1);
                if (rc == STATUS_SUCCESS)
                    rc = lin_tape_sdev_get(chg->dev_obj, &chg->sdev_ct);
                WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
            } /* if */
            if(rc != STATUS_SUCCESS || !chg->alt_pathing) {
                printk("lin_tape: lin_tape_open: changer removed.\n");
                WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
                rc = -ENXIO;
                goto EXIT_LABEL;
            } /* if */
        } /* if */
        device_gotten = True;
        if (chg->busy)
            DbgPrint(("lin_tape: IBMChanger%d busy on open %d\n", chg->dev_index, chg->busy));
        chg->busy |= F_OPEN;
    } /* if */
    else
        changer_devs[chg->cur_path]->chg->open_count++;
    WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
    DbgPrint(("lin_tape_open: IBMchanger%d attempt to open\n", chg->dev_index));
	if(ttype == LIN_TAPE_REGULAR_TRACE && chg->chg_log == NULL) {
		chg->chg_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		chg->chg_log_len = 0;
	} /* if */

	switch(ttype) {
	case LIN_TAPE_REGULAR_TRACE:
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_CG_ATMPT_OPEN, chg->dev_index >> 8,
			chg->dev_index & 0xFF);
		break;

	case LIN_TAPE_FULL_TRACE:
		changer_setup_and_log_full(chg, chg->cur_path,
			TRC_SCSI_PATH, NULL, 0);

		changer_setup_and_log_full(chg, chg->cur_path,
			TRC_CG_ATMPT_OPEN, NULL, -4);
		break;

	default: break;
	} /* switch */

	/* attempt to open IBMChanger# */

	if (device_gotten)
	rc = lin_tape_changer_open(&chg, &filp);

EXIT_LABEL:

	if (chg && device_gotten){
        WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
        if(rc != STATUS_SUCCESS) {
            DbgPrint(("lin_tape_open_changer %d.  Removing refcount\n",
                        rc));
            lin_tape_sdev_put(chg->dev_obj, &chg->sdev_ct);
        }
        chg->busy &= (0xFF-F_OPEN);
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
	} /* if */

	if (rc == STATUS_SUCCESS) {
		lin_tape_chg_pfoID_hctl_update(pfo_id, chg);
		pfo_id->pfo_minor = minor;
	}

	return rc;
} /* pfo_open_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: pfo_close_changer                                                      *
*                                                                              *
*  PURPOSE: pfo interface to close a changer                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_close_changer(struct pfo_ID *pfo_id)
{
	struct changer_descriptor *chg = NULL;
	struct file filp;
	int rc = -EIO, minor = -1;
	int dev_index;
	unchar ttype = lin_tape_trace_type;

	if (!pfo_id) {
		rc = -EINVAL;
	}

	minor = pfo_id->pfo_minor;
	READ_LOCK(&changer_des_lock);
	if(minor >= LinTapeChangerOffset
	   && changer_devs
	   && pfo_minor_match(pfo_id, 0)
        && changer_devs[minor-LinTapeChangerOffset]
        && (chg=changer_devs[minor-LinTapeChangerOffset]->chg)
        && changer_devs[chg->cur_path])
			chg = changer_devs[chg->cur_path]->chg;
	READ_UNLOCK(&changer_des_lock);

	if(chg == NULL) {
		DbgPrint(("lin_tape: unexcpected minor or device is not configured "\
			"or has been removed.\n"));
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	filp.f_flags = pfo_id->f_flags;
	/* check if class_device_remove has been called */
	dev_index = chg->dev_index;
	WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
	if( chg->busy && chg->open_count == 1) {
        DbgPrint(("lin_tape: IBMChanger%d busy %d\n", chg->dev_index, chg->busy))
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_CLOSE_BSY);
		else if(ttype == LIN_TAPE_FULL_TRACE) {
			changer_log_full_trace(chg, TRC_CLOSE_BSY, 0, NULL);
			changer_log_rc(chg, EBUSY);
		}
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */

	if (chg->open_count > 1){
        chg->open_count --;
        rc = STATUS_SUCCESS;
	}
    WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);

	/* attempt to close IBMChanger# */
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_CG_ATMPT_CLOSE, dev_index >> 8, dev_index & 0xFF);
	else if(ttype == LIN_TAPE_FULL_TRACE) {
		changer_setup_and_log_full(chg, chg->cur_path, TRC_SCSI_PATH,
			NULL, 0);

		changer_setup_and_log_full(chg, dev_index, TRC_CG_ATMPT_CLOSE,
			NULL, -4);
	} /* else if */
    if (rc != STATUS_SUCCESS)
	rc = lin_tape_changer_close(&chg, &filp);

EXIT_LABEL:

    if (chg && chg->remove_pending && chg->open_count == 0){
        cleanup_and_free_cdesc(chg->dev_index);
    } /* if */

	if (rc == STATUS_SUCCESS) {
		lin_tape_chg_pfoID_hctl_update(pfo_id, chg);
		pfo_id->pfo_minor = -1;
	} /* if */
	else if (rc == -ENXIO) pfo_id->pfo_minor = -1;

	return rc;
} /* pfo_close_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: pfo_changer_pass_thru                                                 *
*                                                                              *
*  PURPOSE: pfo interface to pass cdb to changer                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    pfo_id: the pfo data with device identification                           *
*    cdb:                                                                      *
*    cmd_length:                                                               *
*    buffer:                                                                   *
*    buffer_length:                                                            *
*    data_direction:                                                           *
*    timeout:                                                                  *
*    sense:                                                                    *
*    sense_length:                                                             *
*    sense_avail:                                                              *
*    resid:                                                                    *
*    errors:                                                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int pfo_changer_pass_thru(struct pfo_ID *pfo_id, unchar* cdb,
	unchar cmd_length, unchar* buffer, uint *buffer_length,
	uint data_direction, uint timeout, unchar* sense,
	unchar *sense_length, unchar *sense_avail, int *resid,
	int *errors)
{
	int rc = -EIO;
	int minor = -1;
	struct changer_descriptor* chg = NULL;
	unchar ttype = lin_tape_trace_type;

	if (!pfo_id) {
		rc = -EINVAL;
	}

	minor = pfo_id->pfo_minor;

	READ_LOCK(&tape_des_lock);
	if(minor >= LinTapeChangerOffset
	   && changer_devs
	   && pfo_minor_match(pfo_id, 0)
        && changer_devs[minor-LinTapeChangerOffset]
        && (chg=changer_devs[minor-LinTapeChangerOffset]->chg)
        && changer_devs[chg->cur_path])
			chg = changer_devs[chg->cur_path]->chg;

	if(chg == NULL || chg->remove_pending) {
		DbgPrint(("lin_tape_pfo_passthru: Device removed\n"));
        READ_UNLOCK(&tape_des_lock);
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log,
				&lin_tape_log_len, TRC_TBL_CORRUPT);
		rc = -ENXIO;
		goto EXIT_LABEL;
	}

	DbgPrint(("CDB 0x%x sense SK/ASC/ASCQ %x %x %x resid 0x%x errors %x\n",
           cdb[0], sense[2] & 0x0F, sense[12], sense[13],*resid,*errors));

	if(chg->busy) {
        READ_UNLOCK(&tape_des_lock);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);
	rc = lin_tape_changer_pass_thru_from_pfoID(&chg,
								cdb,
								cmd_length,
								buffer,
								buffer_length,
								data_direction,
								timeout,
								sense,
								sense_length,
								sense_avail,
								resid,
								errors);

EXIT_LABEL:

	lin_tape_chg_pfoID_hctl_update(pfo_id, chg);
	if (rc == -ENXIO){
		pfo_id->pfo_minor = -1;
	}
	DbgPrint(("CDB 0x%x sense SK/ASC/ASCQ %x %x %x resid %x errors %x rc %x\n",
           cdb[0], sense[2] & 0x0F, sense[12], sense[13],*resid,*errors, rc));
	return rc;
} /* pfo_changer_pass_thru() */


#undef LOCK_DES
#undef UNLOCK_DES
#undef LOCK_CHG
#undef UNLOCK_CHG
/*****************************************************************************/


module_init(lin_tape_init);
module_exit(lin_tape_exit);
