/******************************************************************************
 *   lin_tape_scsi_tape.c
 *
 *   (C) Copyright International Business Machines  Corp. 2007-2020
 *   This file is part of the IBM Linux Enhanced SCSI Tape and
 *   Medium Changer Device Driver.
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *****************************************************************************/

#include "lin_tape_scsi_tape.h"
#include "lin_tape_scsi_ioctl.h"
#include "lin_tape_extra_ioctl.h"
#include "lin_tape_timeout_tape.h"
#include "convert.h"
/**************************************************************
 *                                                            *
 * Internal Private Functions                                 *
 *                                                            *
 **************************************************************/

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_write_dynamic_attrs                                          *
*                                                                              *
*  PURPOSE: set the dynamic runtime attributes for this ITL nexus              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: drive for which we want to set the attributes                       *
*    clear: if True, clear the values rather than set them                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_write_dynamic_attrs(struct tape_descriptor** drvp, boolean clear)
{
	int rc = -EIO, tot_len = 0, i = 0, timeout = 0;
	boolean cdb_allocd = False;
	struct dyn_attr_parm_list* parm_list = NULL;
	struct tape_descriptor* drv = *drvp;

	if(!dynamic_attributes || !drv ) goto EXIT_LABEL;

	if(!drv->cdb) {
		drv->cdb = kmalloc(MAX_CDB_SIZE, GFP_KERNEL);
		if(!drv->cdb) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		cdb_allocd = True;
	} /* if */
	memset(drv->cdb, '\0', MAX_CDB_SIZE);

	/* length calculated as the size of the containing structure, minus the
		array-of-one placeholder byte, plus the size of the data */
	tot_len = sizeof(struct dyn_attr_parm_list) - sizeof(char);
	for(i = 0; dyn_attrs[i].val_str; i++)
		tot_len += sizeof(struct dyn_attr) - sizeof(char) +
			(clear ? 0 : dyn_attrs[i].len);

	parm_list = kmalloc(tot_len, GFP_KERNEL);
	if(!parm_list) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(parm_list, '\0', tot_len);

	rc = lin_tape_dyn_attrs_parm_list(drv, (char*)parm_list, tot_len,
		clear);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->buffer = (void*)parm_list;
	drv->buf_len = tot_len;
	drv->cdb->cdb_write_dynamic_attribute.op_code =
		SCSIOP_WRITE_DYNAMIC_RUNTIME;
	for(i = 0; i < 4; i++)
		drv->cdb->cdb_write_dynamic_attribute.parameter_list_length[i] =
			(tot_len >> (32 - (i + 1) * BITS_PER_BYTE)) & 0xFF;
	drv->cdb_len = CDB_GENERIC_12;
	drv->last_op = SCSIOP_WRITE_DYNAMIC_RUNTIME;

	timeout = compute_timeout(drv->drv_id, WriteDynamicRTTimeout);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, 0, False);

	DbgPrint(("lin_tape_write_dynamic_attrs: %d\n", rc));

EXIT_LABEL:

	if(parm_list) kfree(parm_list);
	if(cdb_allocd && drv ) {
		kfree(drv->cdb);
		drv->cdb = NULL;
	} /* if */

	drv->buffer = NULL;
	drv->buf_len = 0;
	*drvp = drv;
	return rc;
} /* lin_tape_write_dynamic_attrs() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_dyn_attrs_parm_list                                          *
*                                                                              *
*  PURPOSE: Fill out the parameter list for setting dynamic attributes         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: drive for which we want to set the attributes                        *
*    parm_list: the list to fill out                                           *
*    len: the length of the parm_list                                          *
*    clear: if True, clear the values rather than set them                     *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_dyn_attrs_parm_list(struct tape_descriptor* drv, char* parm_list,
	int len, boolean clear)
{
	int rc = STATUS_SUCCESS, idx = DRA_LIST_HDR_LEN, i = 0, path_ct = 0;
	char tmp[ONE_K/16];
	struct dyn_attr* attr = NULL;
	struct new_utsname* uname = NULL;

	uname = kmalloc(sizeof(struct new_utsname), GFP_KERNEL);
	if(!uname) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memcpy(uname, &drv->uname_info, sizeof(struct new_utsname));


	for(i = 0; dyn_attrs[i].val_str; i++) {
		memset(tmp, '\0', ONE_K/16);
		attr = (struct dyn_attr*)&parm_list[idx];
		attr->attr_id[0] = (dyn_attrs[i].id >> 8) & 0xFF;
		attr->attr_id[1] = dyn_attrs[i].id & 0xFF;
		attr->format = 1;
		if(clear) {
			attr->attr_length[0] = attr->attr_length[1] =
			attr->attr_length[2] = attr->attr_length[3] = 0;
			idx += sizeof(struct dyn_attr) - sizeof(char);
			continue;
		} /* if */

		attr->attr_length[0] = (dyn_attrs[i].len >> 24) & 0xFF;
		attr->attr_length[1] = (dyn_attrs[i].len >> 16) & 0xFF;
		attr->attr_length[2] = (dyn_attrs[i].len >> 8) & 0xFF;
		attr->attr_length[3] = dyn_attrs[i].len & 0xFF;
		switch(dyn_attrs[i].id) {
			case DRA_DSFN:
				sprintf(attr->attr_value, "IBMtape%d",
					drv->dev_index);
				break;

			case DRA_DPF_ACTIVE_PATH:
				if(alternate_pathing)
					path_ct = get_dpf_path(drv);
				sprintf(tmp, "%d", path_ct);
				if(strlen(tmp) <= DRA_ACTIVE_PATH_LEN)
					memcpy(attr->attr_value, tmp,
						strlen(tmp));
				else
					memcpy(attr->attr_value, "???",
						strlen("???"));
				break;

			case DRA_HOST_NAME:
				memcpy(attr->attr_value, uname->nodename,
					min((int)strlen(uname->nodename),
						(int)dyn_attrs[i].len));
				break;

			case DRA_OS:
				memcpy(attr->attr_value, uname->sysname,
					min((int)strlen(uname->sysname),
						(int)dyn_attrs[i].len));
				break;

			case DRA_OS_VER:
				memcpy(attr->attr_value, uname->release,
					min((int)strlen(uname->release),
						(int)dyn_attrs[i].len));

				if(attr->attr_value[dyn_attrs[i].len - 1]
					!= '\0' || strlen(attr->attr_value) +
					strlen(" ") + strlen(uname->machine) >
					dyn_attrs[i].len) break;

				strcat(attr->attr_value, " ");
				strcat(attr->attr_value, uname->machine);
				break;

			case DRA_DD_NAME:
				memcpy(attr->attr_value, "lin_tape",
					strlen("lin_tape"));
				break;

			case DRA_DD_VER:
				memcpy(attr->attr_value, LIN_TAPE_VERSION,
					strlen(LIN_TAPE_VERSION));
				break;

			case DRA_PID:
				sprintf(attr->attr_value, "%d",
					current->tgid);
				break;
		} /* switch */

		idx += sizeof(struct dyn_attr) - sizeof(char) +
			dyn_attrs[i].len;
	} /* for */

EXIT_LABEL:

	if(uname) kfree(uname);
	return rc;
} /* lin_tape_dyn_attrs_parm_list() */

/*******************************************************************************
*                                                                              *
*  NAME: get_dpf_path                                                          *
*                                                                              *
*  PURPOSE: Get the number of the path that this drive represents              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: drive for which we want to get the path                              *
*                                                                              *
*  RETURNS: path                                                               *
*                                                                              *
*******************************************************************************/

int get_dpf_path(struct tape_descriptor* drv)
{
	int low_idx = 0, done = False, path_ct = 0;
	struct tape_descriptor* trav = drv, * orig = drv, * lowest = drv;

	do {
		if(trav->dev_index < lowest->dev_index) lowest = trav;
		trav = trav->next_path;
	} while(trav != orig);
	path_ct = 1;
	if(lowest == drv) done = True;

	while(!done) {
		low_idx = lowest->dev_index;
		trav = orig;
		do {
			if(trav->dev_index > lowest->dev_index) {
				lowest = trav;
				trav = orig;
			} else {
				trav = trav->next_path;
			} /* else */
		} while(trav != orig);

		do {
			if(trav->dev_index < lowest->dev_index
				&& trav->dev_index > low_idx) lowest = trav;
			trav = trav->next_path;
		} while(trav != orig);
		if(lowest == drv) done = True;
		path_ct++;
	} /* while */

	return path_ct;
} /* get_dpf_path() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_open                                                   *
*                                                                              *
*  PURPOSE: open a tape drive                                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*    filp: the file pointer representing the drive in the kernel               *
*    reserve_on_open: automatically reserve                                    *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_drive_open(struct tape_descriptor** drvp, struct file* filp,
	unchar reserve_on_open)
{
	unchar ready = False;
	int rc = -EIO, i = 0, dyn_attrs_set=FALSE, rsvd=FALSE;
	struct tape_descriptor* drv = *drvp;
	boolean ua_recvd = False;

	/* get access mode */
	drv->access_mode = (filp->f_flags & O_ACCMODE);

	rc = set_descriptor_data_on_open(&drv);
	if(rc != STATUS_SUCCESS) {
            goto EXIT_LABEL;
	}

 	rc = drv_validate_inquiry(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(!reserve_on_open) {
		drv->curr_reserve_type = LinTapeNoReserve;
		report_open_drive(drv, rc);
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < 20; i++) {
		rc = tape_tur(&drv);
		if(rc == STATUS_SUCCESS) {
			ready = True;
			break;
		} /* if */

		if(drv->sense_data->sense_key != UNIT_ATTENTION) break;
		else ua_recvd = True;
	} /* for */

	if(ua_recvd && drv->prevent_removal) {
		if(tape_send_erp_cmd(drv, PREVENT_MEDIUM_REMOVAL, 0, 0))
			DbgPrint(("lin_tape: cannot restore prevent "\
				"medium removal\n"));
	} /* if */

	if((!(filp->f_flags & O_NONBLOCK)) && (!(filp->f_flags & O_NDELAY)) &&
		rc != STATUS_SUCCESS) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if((drv->drv_id == IBM_3580 && drv_gen(drv) >= 5)
		|| (drv->drv_id == IBM_3592 && drv_gen(drv) >= 4)) {
		if (dynamic_attributes){
		    rc = lin_tape_write_dynamic_attrs(&drv, False);
			if(rc == STATUS_SUCCESS) {
                    dyn_attrs_set = True;
			}
			else if(!drv || !drv->dev_obj || drv->remove_pending) {
				     printk("lin_tape: lin_tape_drive_open: drive removed.\n");
				     goto EXIT_LABEL;
			} /* else if */
		}
	} /* if */
    rsvd = True;
	if(TP_RSV_PERSISTENT(drv)) {
		if(drv->explicit_reserved &&
			drv->curr_reserve_type == LinTapePersistentReserve) {
			rc = check_persistent_reservation(&drv);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} else {
			if(drv->curr_reserve_type != LinTapeReserve6)
				drv->curr_reserve_type = LinTapeNoReserve;
			rc = establish_persistent_reservation(&drv);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* else */
	} else {
		/* attempt to reserve the device, allowing two retries */
		rc = tape_reserve(&drv);
		if(rc != STATUS_SUCCESS) rc = tape_reserve(&drv);

		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: unable to reserve the device\n");
			drv->curr_reserve_type = LinTapeNoReserve;
			drv->explicit_reserved = False;
			rc = -EBUSY;
			goto EXIT_LABEL;
		} else drv->curr_reserve_type = LinTapeReserve6;
	} /* else */

	if(drv->opened == False) {
		drv->opened = True;
		rc = tape_set_all_parms(&drv);
	} /* if */

	rc = tape_get_media_parm(&drv);

	if(rc == STATUS_SUCCESS && ready && WriteAccess(drv->access_mode)
		&& (drv->write_protect || drv->logic_write_protect)) {
		rc = -EACCES;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_update_block_limits(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->max_xfer_len = (ulong)min(drv->max_xfer_len, drv->max_blk_size);
	if(drv->max_xfer_len < LinTapeDefaultTransferLength)
		drv->max_xfer_len = LinTapeDefaultTransferLength;

	rc = density_and_encryption_on_open(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = tape_initial_position(&drv, filp, ready);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if ((drv->drv_id == IBM_3592 && drv_gen(drv) >4) ||
        (drv->drv_id == IBM_3580 && drv_gen(drv) >6)){
        erp_amu(&drv,NULL);
	}

	report_open_drive(drv, rc);

EXIT_LABEL:

	if(rc != STATUS_SUCCESS && dyn_attrs_set)
		lin_tape_write_dynamic_attrs(&drv, True);

	if(rc != STATUS_SUCCESS) {
		handle_open_failure(&drv, rsvd);
	} /* if */

	*drvp = drv;
	DbgPrint(("lin_tape: lin_tape_drive_open exit: %d\n", rc));
	return rc;
} /* lin_tape_drive_open() */

/*******************************************************************************
*                                                                              *
*  NAME: set_descriptor_data_on_open                                           *
*                                                                              *
*  PURPOSE: allocate memory and initialize other variables in drive descriptor *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_descriptor_data_on_open(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	rc = allocate_tdescriptor_buffers(drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	drv->in_erp = False;

	drv->buffer = NULL;
	drv->buf_len = 0;

	drv->last_errno = STATUS_SUCCESS;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* set_descriptor_data_on_open() */

/*******************************************************************************
*                                                                              *
*  NAME: allocate_tdescriptor_buffers                                          *
*                                                                              *
*  PURPOSE: Allocate cdb, sense, and tape_log buffers in tape descriptor       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which to allocate buffers                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int allocate_tdescriptor_buffers(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	if(!drv->cdb) {
		drv->cdb = kmalloc(sizeof(union cdb), GFP_KERNEL);
		if(!drv->cdb) {
			if(ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_OPEN_NO_CDB);
			else if(ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENOMEM);

			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	memset(drv->cdb, '\0', sizeof(union cdb));

	if(!drv->sense_data) {
		drv->sense_data = kmalloc(sizeof(struct tape_sense_data),
			GFP_KERNEL);
		if(!drv->sense_data) {
			if(ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_OPEN_NO_SENSE);
			else if(ttype == LIN_TAPE_FULL_TRACE) {
				tape_log_full_trace(drv, TRC_OPEN_NO_SENSE, 0,
					NULL);
				tape_log_rc(drv, ENOMEM);
			} /* else if */

			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	memset(drv->sense_data, '\0', sizeof(struct tape_sense_data));
	drv->resid = False;
	drv->sense_valid = False;
	drv->resid_valid = False;

	if(!drv->tape_log) {
		drv->tape_log = kmalloc(MaxMessageLogSize, GFP_KERNEL);
		if(drv->tape_log) drv->tape_log_len = 0;
	} /* if */

EXIT_LABEL:

	return rc;
} /* allocate_tdescriptor_buffers() */

/*******************************************************************************
*                                                                              *
*  NAME: drv_validate_inquiry                                                  *
*                                                                              *
*  PURPOSE: Validate vendor ID, product ID, and serial number on open          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int drv_validate_inquiry(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	unchar* inq_data = NULL, * data = NULL;
	int rc = STATUS_SUCCESS, inq_size = 0, id_len = 0, offset = 0;

	/* get the product ID and vendor. */
	inq_data = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(inq_data, '\0', MAX_INQ_LEN);

	rc = tape_inquiry_page(&drv, DeviceIdentificationPage, inq_data);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape open failed: tape_inquiry_page: %d.\n", rc);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	inq_size = inq_data[3] + INQ_HEADER_LEN;
	for(offset = INQ_HEADER_LEN; offset < inq_size; offset += id_len) {
		if((inq_data[offset + 1] & 0x0F) == 1) break;
		id_len = 4 + (inq_data[offset + 3] & 0xFF);
	} /* for */

	if(offset >= inq_size) {
		printk("lin_tape: Could not find dev ID descriptor\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* check vendor id */
	data = &inq_data[offset + 4];
	/* check the vendor id */
	if(memcmp(data, "IBM", 3) &&
	   memcmp(data, "QUANTUM", 7) &&
	   memcmp(data, "ADIC", 4) &&
	   memcmp(data, "HP", 2)) {
		printk("lin_tape: open failed: vendor mismatch\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* check product id */
	data = &inq_data[offset + 12];
	if((drv->drv_id == IBM_3590 && memcmp(data, "03590", strlen("03590")))
		||
	(drv->drv_id == IBM_3592 && memcmp(data, "03592", strlen("03592")))
		||
	(drv->drv_id == IBM_3580 && memcmp(data, "ULT", strlen("ULT")) &&
		memcmp(data, "HH LTO", strlen("HH LTO")) &&
		memcmp(data, "Ult", strlen("Ult")))) {
		printk("lin_tape open failed: product mismatch\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* check serial number */
	data = &inq_data[offset + 28];
	if(memcmp(data, drv->serial_num, drv->serial_num_len)){
        rc = -EINVAL;
		printk("lin_tape open failed: SN mismatch \n");
	}

EXIT_LABEL:

	if(inq_data) kfree(inq_data);
	*drvp = drv;
	return rc;
} /* drv_validate_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: check_persistent_reservation                                          *
*                                                                              *
*  PURPOSE: If drive is already reserved, see if we own the reservation        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int check_persistent_reservation(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	unchar* key = NULL;
	char pr_type = 0;
	unsigned int pr_generation = 0;
	int rc = STATUS_SUCCESS;

	key = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!key) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_read_reservation(&drv, key, &pr_generation, &pr_type);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: tape_read_reservation failed, rc %d\n", rc);
		goto EXIT_LABEL;
	} /* if */

	if(memcmp(drv->reservation_key, key, RESERVE_KEY_LENGTH)
		|| drv->pr_generation != pr_generation
		|| pr_type != EXCLUSIVE_ACCESS) {
		printk("lin_tape: mismatched reservation\n");
		drv->explicit_reserved = False;
		drv->curr_reserve_type = LinTapeNoReserve;
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(key) kfree(key);
	*drvp = drv;
	return rc;
} /* check_persistent_reservation() */

/*******************************************************************************
*                                                                              *
*  NAME: establish_persistent_reservation                                      *
*                                                                              *
*  PURPOSE: establish persistent reserve, converting from regular if necessary *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int establish_persistent_reservation(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	unsigned int pr_generation = 0;
	int rc = STATUS_SUCCESS, temp_rc = STATUS_SUCCESS;
	char* rkey = NULL;

	if(drv->curr_reserve_type == LinTapeReserve6) {
		rc = tape_release(&drv);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	drv->explicit_reserved = False;
	drv->curr_reserve_type = LinTapeNoReserve;

	rkey = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!rkey) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(rkey, '\0', RESERVE_KEY_LENGTH);

	/* register a reservation key */
	get_random_bytes(rkey, RESERVE_KEY_LENGTH);
	rc = tape_register_reservation_key(&drv, rkey);
	if(rc != STATUS_SUCCESS) {
		DbgPrint(("lin_tape: tape_register_reservation_key %d\n", rc));
		goto EXIT_LABEL;
	} /* if */

	/* create a reservation */
	rc = tape_create_persist_reserve(&drv);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: tape_create_persist_reserve: %d\n", rc);

		/* remove the key just registered */
		temp_rc = tape_register_reservation_key(&drv, NULL);

		temp_rc = tape_read_reservation(&drv, NULL, &pr_generation,
			NULL);
		if(temp_rc == STATUS_SUCCESS)
			drv->pr_generation = pr_generation;

		goto EXIT_LABEL;
	} /* if */

	rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
	if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
	drv->curr_reserve_type = LinTapePersistentReserve;

EXIT_LABEL:

	if(rkey) kfree(rkey);
	*drvp = drv;
	return rc;
} /* establish_persistent_reservation() */

/*******************************************************************************
*                                                                              *
*  NAME: density_and_encryption_on_open                                        *
*                                                                              *
*  PURPOSE: Set initial density and encryption parameters for the drive        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int density_and_encryption_on_open(struct tape_descriptor** drvp)
{
	struct density_data_t* density_info = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	/* for 3952-2 & later, clear set_density_valid and set density to max */
	if(!disable_density_on_open) {
		drv->set_density_valid = 0;
		if(!strncmp(drv->dev_obj->model, "03592", strlen("03592")) &&
			strncmp(&drv->dev_obj->model[strlen("03592")], "J1A",
				strlen("J1A"))) {
			density_info = kmalloc(sizeof(struct density_data_t),
				GFP_KERNEL);
			if(!density_info) {
				rc = -ENOMEM;
				goto EXIT_LABEL;
			} /* if */

			memset(density_info, '\0',
				sizeof(struct density_data_t));
			density_info->default_density = 0;
			density_info->pending_density = 0;

			rc = tape_set_density(&drv, (void*)density_info, 0);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */
	} /* if */

	/* check drive encryption capability */
	if(drv->drv_id == IBM_3592 || drv->drv_id == IBM_3580)
		tape_get_encryption_capability(&drv);
	DbgPrint(("%s: encryption capable is %d\n", drv->devid,
		drv->encryption_capable));

	if(drv->encryption_capable) {
		tape_get_encryption_enablement(&drv);
		DbgPrint(("%s: encryption enablement is %d\n", drv->devid,
			drv->encryption_state));
	} /* if */

	/* If system-managed encryption, set up write parameter */
	rc = tape_encryption_write_setting(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

EXIT_LABEL:

	if(density_info) kfree(density_info);
	*drvp = drv;
	return rc;
} /* density_and_encryption_on_open() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_initial_position                                                 *
*                                                                              *
*  PURPOSE: Read and set the initial position of the tape drive on open        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we are attempting to open                                 *
*    filp: the file pointer representing the drive in the kernel               *
*    ready: drive has responded with SUCCESS to TUR & is ready for operations  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_initial_position(struct tape_descriptor** drvp, struct file* filp,
	unchar ready)
{
	struct pos_info *read_pos_buf = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	if(ready) {
		/* update the medium position setting */
		read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
		if(read_pos_buf) {
			memset(read_pos_buf, '\0', sizeof(struct pos_info));
			rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM,
				False);
		} /* if */
	} /* if */

	if(filp->f_flags & O_APPEND) {
		rc = tape_rewind(&drv, 0);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(drv->drv_id == IBM_3590 || drv->drv_id == IBM_3592) {
			rc = tape_set_pos(&drv, SpaceSequentialFM, 2);
			goto EXIT_LABEL;
		} /* if */

		while(True) {
			rc = tape_set_pos(&drv, SpaceFileMark, 1);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
			rc = tape_set_pos(&drv, SpaceRecord, 1);
			if(rc == STATUS_SUCCESS) continue;

			if(drv->fm_detected) {
				drv->fm_detected = False;
				rc = STATUS_SUCCESS;
			} /* if */
			break;
		} /* while */
	} /* if */

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);

	*drvp = drv;
	return rc;
} /* tape_initial_position() */

/*******************************************************************************
*                                                                              *
*  NAME: report_open_drive                                                     *
*                                                                              *
*  PURPOSE: If open was successful, set flags and post to log                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive we were attempting to open                                 *
*    rc: the return code to log                                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void report_open_drive(struct tape_descriptor* drv, int rc)
{
	char ttype = lin_tape_trace_type;

	drv->eom = False;
	drv->eom_blk_id = 0;
	drv->eod = False;
	drv->deferred_eom = False;

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_TAPEX_OPENED, drv->dev_index >> 8,
				drv->dev_index & 0xFF);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_setup_and_log_full(drv, drv->dev_index, TRC_TAPEX_OPENED,
			NULL, -4);

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_TP_DVC_OPEN);

	if(ttype == LIN_TAPE_FULL_TRACE) tape_log_rc(drv, -rc);

	tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_OPEN, NULL, -2);

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) tape_log_rc(drv, -rc);

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->drv_open = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

EXIT_LABEL:

	return;
} /* report_open_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_open_failure                                                   *
*                                                                              *
*  PURPOSE: Handle error processing when open fails                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive we were attempting to open                                *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void handle_open_failure(struct tape_descriptor** drvp, int rsv)
{
	struct tape_descriptor* drv = *drvp;
	unsigned int pr_generation = 0;
	int rc = 0;

	if (!drv) return;

	if (rsv){
	if(drv->curr_reserve_type != LinTapeNoReserve
		&& !drv->explicit_reserved) {

		switch(drv->curr_reserve_type) {
		case LinTapeReserve6:
			rc = tape_release(&drv);
			if(rc != STATUS_SUCCESS)
				rc = tape_release(&drv);
			if(rc == STATUS_SUCCESS)
				drv->curr_reserve_type = LinTapeNoReserve;
			break;

		case LinTapePersistentReserve:
			/* register key 0 to clear the reservation */
			rc = tape_register_reservation_key(&drv,NULL);
			if(rc == STATUS_SUCCESS) {
				drv->curr_reserve_type = LinTapeNoReserve;

				rc = tape_read_reservation(&drv,NULL,
					&pr_generation, NULL);
				if(rc == STATUS_SUCCESS)
					drv->pr_generation = pr_generation;
			} /* if */
			break;

		default:
			break;
		} /* switch */
	} /* if */
	}

	drv->sense_valid = False;
	if(drv->sense_data) {
		kfree(drv->sense_data);
		drv->sense_data = NULL;
	} /* if */

	if(drv->cdb) {
		kfree(drv->cdb);
		drv->cdb = NULL;
	} /* if */

	if(drv->buffer) {
		kfree(drv->buffer);
		drv->buffer = NULL;
	} /* if */

	drv->buf_len = False;

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->drv_open = False;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	*drvp = drv;
	return;
} /* handle_open_failure() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_changer_open                                                 *
*                                                                              *
*  PURPOSE: Open a medium changer                                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which we want to perform the request                *
*    filp: the file representing the device in the kernel                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_changer_open(struct changer_descriptor **chgp, struct file *filp)
{
	int rc = -EIO, i = 0;
	unchar ttype = lin_tape_trace_type;

	DbgPrint(("lin_tape: lin_tape_changer_open entry\n"));

	READ_LOCK(&changer_des_lock);
	if(!(*chgp)) {
        READ_UNLOCK(&changer_des_lock);
		printk("lin_tape: lin_tape_changer_open: changer removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&changer_des_lock);

	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

	rc = chg_alloc_cdb_and_sense((*chgp));
	if(rc != STATUS_SUCCESS) {
        WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
        goto EXIT_LABEL;
    }
	(*chgp)->in_erp = False;
	(*chgp)->buffer = NULL;
	(*chgp)->buf_len = 0;
	(*chgp)->comp_code = STATUS_SUCCESS;
	(*chgp)->comp_code_valid = False;
	(*chgp)->last_errno = STATUS_SUCCESS;
    WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

	rc = chg_validate_vendor_and_serial(&(*chgp));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	(*chgp)->cmd_wo_ua_erp = False;
	for(i = 0; i < 20; i++) {
		rc = changer_tur(&(*chgp));
		if(rc != STATUS_SUCCESS) {
			if((*chgp)->sense_data->sense_key != UNIT_ATTENTION)
				break;
		} else break;
	} /* for */
	(*chgp)->cmd_wo_ua_erp = False;

	DbgPrint(("lin_tape: changer_open_reserve %d\n", changer_open_reserve));

	/* attempt to reserve the device, allowing two retries */
	if(changer_open_reserve) {
		rc = changer_reserve(&(*chgp));

		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: lin_tape_changer_open unable to "\
				"reserve the changer\n");

			if(ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_CG_NO_RESERVE);
			else if(ttype == LIN_TAPE_FULL_TRACE)
				changer_log_full_trace((*chgp), TRC_CG_NO_RESERVE,
					0, NULL);

			if(ttype == LIN_TAPE_FULL_TRACE)
				changer_log_rc((*chgp), EBUSY);
			rc = -EBUSY;
			goto EXIT_LABEL;
		} else (*chgp)->curr_reserve_type = LinTapeReserve6;
	} /* if */

	/* IBMchanger# opened */
	if(ttype == LIN_TAPE_REGULAR_TRACE) {
		tape_log_msg3(lin_tape_log, &lin_tape_log_len,
			TRC_CHG_OPENED, (*chgp)->dev_index >> 8,
			(*chgp)->dev_index & 0xFF);
		tape_log_msg((*chgp)->chg_log, &((*chgp)->chg_log_len),
			TRC_CHG_DVC_OPEN);
	} else if(ttype == LIN_TAPE_FULL_TRACE) {
		changer_setup_and_log_full((*chgp), (*chgp)->dev_index, TRC_CHG_OPENED,
			NULL, -4);
	} /* else if */

	if(ttype == LIN_TAPE_FULL_TRACE) changer_log_rc((*chgp), -rc);
	changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_CG_OPEN, NULL, -2);

EXIT_LABEL:

	if((*chgp) && !(*chgp)->open_count && rc != STATUS_SUCCESS) {
		(*chgp)->sense_valid = False;
		if((*chgp)->sense_data) {
			kfree((*chgp)->sense_data);
			(*chgp)->sense_data = NULL;
		} /* if */

		if((*chgp)->cdb) {
			kfree((*chgp)->cdb);
			(*chgp)->cdb = NULL;
		} /* if */
	} /* if */

	if((*chgp) && rc == STATUS_SUCCESS) {
		WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		(*chgp)->open_count++;
		(*chgp)->chg_open = True;
		WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	} /* if */

	DbgPrint(("lin_tape: lin_tape_changer_open exit: %d\n", rc));
	return rc;
} /* lin_tape_changer_open() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_alloc_cdb_and_sense                                               *
*                                                                              *
*  PURPOSE: Allocate CDB and sense buffer if necessary on changer open         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer to open                                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chg_alloc_cdb_and_sense(struct changer_descriptor* chg)
{
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	if(!chg->cdb)
		chg->cdb = kmalloc(sizeof(union cdb), GFP_KERNEL);
	if(!chg->cdb) {
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_OPEN_NO_CDB);
		else if(ttype == LIN_TAPE_FULL_TRACE)
			changer_log_rc(chg, ENOMEM);
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(chg->cdb, '\0', sizeof(union cdb));

	if(!chg->sense_data)
		chg->sense_data = kmalloc(sizeof(struct tape_sense_data),
			GFP_KERNEL);
	if(!chg->sense_data) {
		printk("lin_tape: lin_tape_changer_open: unable to allocate "\
			"space for sensedata");

		if(chg->cdb) {
			kfree(chg->cdb);
			chg->cdb = NULL;
		} /* if */

		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_OPEN_NO_SENSE);
		else if(ttype == LIN_TAPE_FULL_TRACE) {
			changer_log_full_trace(chg, TRC_OPEN_NO_SENSE, 0, NULL);
			changer_log_rc(chg, ENOMEM);
		} /* else if */

		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(chg->sense_data, '\0', sizeof(struct tape_sense_data));
	chg->sense_valid = False;

EXIT_LABEL:

	return rc;
} /* chg_alloc_cdb_and_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_validate_vendor_and_serial                                        *
*                                                                              *
*  PURPOSE: Make sure that the changer we are opening meets vendor & SN reqs   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer to open                                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chg_validate_vendor_and_serial(struct changer_descriptor** chgp)
{
	int rc = STATUS_SUCCESS;
	unchar* inq_data = NULL;

	inq_data = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(inq_data, '\0', MAX_INQ_LEN);
	rc = changer_std_inquiry(&(*chgp), inq_data);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: changer_open unable to get the device "\
			"inquiry data.\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* check the vendor id */
	if(memcmp(&inq_data[8], "IBM", 3) &&
	   memcmp(&inq_data[8], "QUANTUM", 7) &&
	   memcmp(&inq_data[8], "ADIC", 4) &&
	   memcmp(&inq_data[8], "HP", 2)) {
		printk("lin_tape: IBMchanger open failed, vendor does not "\
			"match previous configuration.\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	kfree(inq_data);
	inq_data = NULL;

	inq_data = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq_data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	/* check the serial number */
	rc = changer_inquiry_page(&(*chgp), UnitSerialPage, inq_data);
	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: changer_open unable to get the device "\
			"inquiry data.\n");
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(memcmp(&inq_data[4], (*chgp)->serial_num, (*chgp)->serial_num_len)) {
		printk("lin_tape: IBMchanger open failed, serial "\
			"number does not match previous "\
			"configuration.\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(inq_data) kfree(inq_data);

	return rc;

} /* chg_validate_vendor_and_serial() */

int drive_busy(struct tape_descriptor *drv, unchar flag){
    unchar busy = drv->busy;
    if (drv->dev_index != drv->cur_path) busy = 0;
    else {
        if (drv->busy & flag)
            busy = drv->busy-flag;
    }
    DbgPrint(("lin_tape: Drive busy: %d flag %d\n",busy, flag));
    return (int) busy;
}
/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_close                                                  *
*                                                                              *
*  PURPOSE: Cleanup and close a tape drive                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which we want to perform the request                  *
*    filp: the file representing the device in the kernel                      *
*    reserve_on_open: if False, do not attempt release                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_drive_close(struct tape_descriptor **drvp, struct file *filp,
	unchar reserve_on_open)
{
	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	DbgPrint(("lin_tape: lin_tape_drive_close entry\n"));
	READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if((!drv->dev_obj || drv->ghost_device) && !drv->remove_pending) {
		if(drv->alt_pathing) {
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = tape_perform_failover_generic(&drv, 1800 * HZ);
			READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		} /* if */

		if(rc != STATUS_SUCCESS || !drv->alt_pathing) {
			printk("lin_tape: lin_tape_drive_close: drive "\
				"removed.\n");
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_TP_OUT_RANGE);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENXIO);
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
    READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    if (!drv->remove_pending){
	if( dynamic_attributes &&
		((drv->drv_id == IBM_3580 && drv_gen(drv) >= 5) ||
		(drv->drv_id == IBM_3592 && drv_gen(drv) >=4)))
		lin_tape_write_dynamic_attrs(&drv, True);

	if(!petro || (petro == 1 && reserve_on_open))
		rc = lin_tape_flush(&drv, filp);

	drv->cmd_wo_ua_erp = True;
	if(drv->curr_reserve_type != LinTapeNoReserve &&
		drv->explicit_reserved == False) {
		rc = drive_release_on_close(&drv, rc);
		if(rc == STATUS_SUCCESS)
			drv->curr_reserve_type = LinTapeNoReserve;
		else printk("lin_tape: IBMtape%d failed release on close.\n",
				drv->cur_path);
	} /* if */
	drv->cmd_wo_ua_erp = False;

	if((drv->encryption_state && drv->application_set_encryption) ||
		(drv->encryption_proxy_entitled &&
		drv->encryption_keypath_configured &&
		drv->sys_encryption_proxy)) tape_discard_data_key(&drv);
    }
	drv->application_set_encryption = False;
	drv->encryption_state = STATE_UNKNOWN;

	cleanup_buffers_on_close(drv);

    WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	/* release hba module */
    lin_tape_sdev_put(drv->dev_obj, &drv->sdev_ct);
	drv->drv_open = False;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	log_tape(drv, TRC_TP_CLOSED);

	tape_setup_and_log_full(drv, drv->cur_path, TRC_TP_RM_HBA_CT,
		NULL, -2);
	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, 0);

EXIT_LABEL:

	*drvp = drv;

	DbgPrint(("lin_tape: lin_tape_drive_close exit: %d\n", rc));
	return rc;
} /* lin_tape_drive_close() */

/*******************************************************************************
*                                                                              *
*  NAME: drive_release_on_close                                                *
*                                                                              *
*  PURPOSE: Issue automatic release when a drive is closed                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which we want to perform the request                  *
*    rc: default return code                                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int drive_release_on_close(struct tape_descriptor** drvp, int rc)
{
	unsigned int pr_generation = 0;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	switch(drv->curr_reserve_type) {
	case LinTapeReserve6:
		rc = tape_release(&drv);
		if(rc != STATUS_SUCCESS)
			rc = tape_release(&drv);
		if(rc == STATUS_SUCCESS) {
			drv->curr_reserve_type = LinTapeNoReserve;
			break;
		} /* if */

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_RELEASE_FAILED);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_RELEASE_FAILED, 0, NULL);

		break;

	case LinTapePersistentReserve:
		/* register key 0 to clear the reservation */
		rc = tape_register_reservation_key(&drv, NULL);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d "\
				"tape_register_reservation_key failed, "\
				"rc %d\n", drv->cur_path, rc);
			break;
		} /* if */

		drv->curr_reserve_type = LinTapeNoReserve;

		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;

	default:
		break;
	} /* switch */

	*drvp = drv;
	return rc;
} /* drive_release_on_close() */

/*******************************************************************************
*                                                                              *
*  NAME: cleanup_buffers_on_close                                              *
*                                                                              *
*  PURPOSE: Free memory and set counters back to zero                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive for which we want to perform the request                   *
*    rc: default return code                                                   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void cleanup_buffers_on_close(struct tape_descriptor* drv)
{
	char ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;

	if((drv->rw_buf && (!drv->rw_buf_size || drv->rw_buf_order == -1))
		||
	(!drv->rw_buf && (drv->rw_buf_size || drv->rw_buf_order != -1))) {
		printk("lin_tape: IBMtape%d lin_tape_drive_close rw buffer "\
			"corrupted! rw_buf_size %lu, rw_buf_order %ld\n",
			drv->cur_path, drv->rw_buf_size, drv->rw_buf_order);

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_RW_BUF_CORRUPT);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_RW_BUF_CORRUPT, 0, NULL);

	} else {
		if(drv->rw_buf)
			free_pages((ulong)drv->rw_buf, drv->rw_buf_order);
		drv->rw_buf_size = 0;
		drv->rw_buf_order = -1;
		drv->rw_buf = NULL;
	} /* else */

	if(drv->sg_list) {
		rc = tape_del_sg(drv);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d lin_tape_drive_close "\
				"unable to destory scatter gather\n",
				drv->cur_path);
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log,
					&lin_tape_log_len, TRC_SG_STUCK);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_full_trace(drv, TRC_SG_STUCK, 0, NULL);
		} /* if */
	} /* if */

	if(drv->sense_data) {
		kfree(drv->sense_data);
		drv->sense_data = NULL;
	} /* if */
	drv->sense_valid = False;
	if(drv->cdb) {
		kfree(drv->cdb);
		drv->cdb = NULL;
	} /* if */
	if(drv->buffer) {
		kfree(drv->buffer);
		drv->buffer = NULL;
	} /* if */
	drv->buf_len = 0;

	return;
} /* cleanup_buffers_on_close() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_changer_close                                                *
*                                                                              *
*  PURPOSE: Cleanup and close a changer device                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which we want to perform the request                *
*    filp: the file representing the device in the kernel                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_changer_close(struct changer_descriptor **chgp, struct file *filp)
{

	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;
	char call_description[] = "close";

	DbgPrint(("lin_tape_changer_close entry\n"));
	if(changer_wait_to_set_busy(&(*chgp), call_description,F_ADD_REMOVE) != STATUS_SUCCESS) {
		printk("lin_tape_close failed.  Device busy\n");
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */
	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
    (*chgp)->busy |= F_CLOSE;
    if((!(*chgp)->dev_obj || (*chgp)->ghost_device) && !(*chgp)->remove_pending) {
        if((*chgp)->alt_pathing) {
            WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
            rc = changer_failover_ops(&(*chgp), HZ * 1800, 0);
        	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
        } /* if */
        if(rc != STATUS_SUCCESS || !(*chgp)->alt_pathing) {
            WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
            printk("lin_tape: lin_tape_open: changer removed.\n");
            rc = -ENXIO;
            goto EXIT_LABEL;
        } /* if */
    } /* if */
    WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	if(!(*chgp)->remove_pending &&
        (*chgp)->curr_reserve_type != LinTapeNoReserve
		&& !(*chgp)->explicit_reserved
		&& (*chgp)->curr_reserve_type == LinTapeReserve6) {

		rc = changer_release(&(*chgp));
		if(rc != STATUS_SUCCESS)
			rc = changer_release(&(*chgp));
		if(rc == STATUS_SUCCESS)
			(*chgp)->curr_reserve_type = LinTapeNoReserve;
		else {
			if(ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_RELEASE_FAILED);
			else if(ttype == LIN_TAPE_FULL_TRACE)
				changer_log_full_trace((*chgp), TRC_RELEASE_FAILED,
					0, NULL);
			printk("lin_tape: IBMchanger%d failed release "\
				"on close\n", (*chgp)->cur_path);
		} /* else */
	} /* if */

	if((*chgp)->sense_data) {
		kfree((*chgp)->sense_data);
		(*chgp)->sense_data = NULL;
	} /* if */
	(*chgp)->sense_valid = False;
	if((*chgp)->cdb) {
		kfree((*chgp)->cdb);
		(*chgp)->cdb = NULL;
	} /* if */
	if((*chgp)->buffer) {
		kfree((*chgp)->buffer);
		(*chgp)->buffer = NULL;
	} /* if */
	(*chgp)->buf_len = 0;

	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg((*chgp)->chg_log, &((*chgp)->chg_log_len), TRC_CHG_CLOSED);
	else if(ttype == LIN_TAPE_FULL_TRACE)
		changer_log_rc((*chgp), 0);

	changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_CHG_RM_HBA_CT,
		NULL, -2);

    WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	if (*chgp){
        (*chgp)->open_count = 0;
        (*chgp)->chg_open = False;
        (*chgp)->busy &= (0xFF-F_CLOSE);
        lin_tape_sdev_put((*chgp)->dev_obj, &(*chgp)->sdev_ct);
	}
    WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

EXIT_LABEL:

	DbgPrint(("lin_tape_changer_close exit: %d, open_count %d\n",
		rc, (*chgp) ? (*chgp)->open_count : 424242));
	return rc;
} /* lin_tape_changer_close() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_flush                                                        *
*                                                                              *
*  PURPOSE: Flush the tape drive and handle write filemark and repositioning   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which we want to perform the request                  *
*    filp: the file representing the device in the kernel                      *
*    command: the IOCTL request                                                *
*    arg: user-specified argument, typically a pointer to a data buffer        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_flush(struct tape_descriptor **drvp, struct file *filp)
{
	int rc = STATUS_SUCCESS;
	char ttype;
	struct tape_descriptor* drv = *drvp;

	READ_LOCK(&lin_tape_trace_lock);
	ttype = lin_tape_trace_type;
	READ_UNLOCK(&lin_tape_trace_lock);

	if(drv->last_tape_motion == WRITE_6 && drv->wfm_on_close_safe) {

		rc = tape_wfm(&drv, False, 2);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(drv->special_file_type == NonRewindDevice) {
			rc = tape_set_pos(&drv, SpaceFileMark, -1);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */
	} else if(drv->last_tape_motion == WRITE_FILEMARKS &&
		drv->ending_fm == 1 && drv->wfm_on_close_safe) {
		rc = tape_wfm(&drv, False, 1);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(drv->special_file_type == NonRewindDevice) {
			rc = tape_set_pos(&drv, SpaceFileMark, -1);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_FLUSH_FM);
	} else if(drv->last_tape_motion == READ_6 &&
		drv->special_file_type == NonRewindDevice &&
		!drv->fm_detected) {
		rc = tape_set_pos(&drv, SpaceFileMark, 1);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(lin_tape_log, &lin_tape_log_len,
				TRC_FLUSH_FWD);
	} /* else if */

	if(drv->special_file_type == RewindDevice) {
		rc = tape_rewind(&drv, drv->rew_immed);
		if(rc == STATUS_SUCCESS)
			drv->blk_num = 0;
		else if(drv->sense_valid &&
			drv->sense_data &&
			drv->sense_data->sense_key == 2) {
			drv->blk_num = 0;
			rc = STATUS_SUCCESS;
		} else goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	drv->ending_fm = drv->wfm_on_close_safe = 0;
	*drvp = drv;
	return rc;
} /* lin_tape_flush() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_ioctl                                                  *
*                                                                              *
*  PURPOSE: Handle IOCTL requests for tape drives                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which we want to perform the request                  *
*    filp: the file representing the device in the kernel                      *
*    command: the IOCTL request                                                *
*    arg: user-specified argument, typically a pointer to a data buffer        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_drive_ioctl(struct tape_descriptor **drvp, struct file *filp,
	uint command, ulong arg)
{
	int rc = -EIO, type = 0, cmd = 0;
	char ttype = lin_tape_trace_type;
	struct mtop* mt_cmd = NULL;
	struct tape_descriptor* drv = *drvp;
	boolean immed = False;

	READ_LOCK(&tape_des_lock);
	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: lin_tape_drive_ioctl: drive removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

	if(petro == 1 && drv->curr_reserve_type == LinTapeNoReserve &&
		command != SIOC_INQUIRY && command != SIOC_INQUIRY_PAGE) {
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(!drv->dev_obj || drv->ghost_device || !scsi_block_when_processing_errors(drv->dev_obj)) {
		if(drv->alt_pathing) {
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = tape_perform_failover_generic(&drv, 1800 * HZ);
			READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		} /* if */

		if(rc != STATUS_SUCCESS || !drv->alt_pathing) {
			printk("lin_tape: lin_tape_drive_read: drive "\
				"removed.\n");
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_TP_OUT_RANGE);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENXIO);
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
    READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);

	if(!drv->cdb || !drv->sense_data || !drv->tape_log) {
		rc = allocate_tdescriptor_buffers(drv);
		if(rc != STATUS_SUCCESS) {
			DbgPrint(("lin_tape_drive_ioctl: cannot allocate "\
				"buffer\n"));
			goto EXIT_LABEL;
		} /* if */
	} /* if */


	type = _IOC_TYPE(command);
	cmd = _IOC_NR(command);

	if(type == _IOC_TYPE(MTIOCGET) && cmd == _IOC_NR(MTIOCGET)) {
		rc = mtiocget(&drv, command, arg);
	} else if(type == _IOC_TYPE(MTIOCPOS) && cmd == _IOC_NR(MTIOCPOS)) {
		rc = mtiocpos(&drv, arg);
	} else if(type == _IOC_TYPE(MTIOCTOP) && cmd == _IOC_NR(MTIOCTOP)) {

		if(_IOC_SIZE(command) != sizeof(struct mtop)) {
			drv->last_errno = EINVAL;

			if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, EINVAL);
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		mt_cmd = kmalloc(sizeof(struct mtop), GFP_KERNEL);
		if(!mt_cmd) {
			drv->last_errno = ENOMEM;

			if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENOMEM);
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		if(copy_from_user((char*)mt_cmd, (char*)arg,
			sizeof(struct mtop))) {
			drv->last_errno = EFAULT;

			if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, EFAULT);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		switch(mt_cmd->mt_op) {
		case MTRESET:
			log_tape(drv, TRC_MT_RESET);
			rc = tape_reset_drive(&drv);
			/* reset drive will clear the reservation */
			if(rc == STATUS_SUCCESS) {
				drv->curr_reserve_type = LinTapeNoReserve;
				memset(drv->reservation_key, '\0',
					RESERVE_KEY_LENGTH);
			} /* if */
			break;

		case MTFSF:
			log_tape(drv, TRC_MT_FSF);
			rc = tape_set_pos(&drv, SpaceFileMark,
				(long) mt_cmd->mt_count);
			break;

		case MTBSF:
			log_tape(drv, TRC_MT_BSF);
			rc = tape_set_pos(&drv, SpaceFileMark,
				-((long)mt_cmd->mt_count));
			break;

		case MTFSR:
			log_tape(drv, TRC_MT_FSR);
			rc = tape_set_pos(&drv, SpaceRecord,
				(long)mt_cmd->mt_count);
			break;

		case MTBSR:
			log_tape(drv, TRC_MT_BSR);
			rc = tape_set_pos(&drv, SpaceRecord,
				-((long)mt_cmd->mt_count));
			break;

		case MTWEOF:
		case MTWEOFI:
			immed = (petro == 1 && mt_cmd->mt_count == 1
					&& drv->wfm_immed) ||
				(mt_cmd->mt_op == MTWEOFI);

			log_tape(drv, TRC_MT_WEOF);
			if(!WriteAccess(drv->access_mode)) {
				rc = -EPERM;
				break;
			} /* if */

			rc = tape_wfm(&drv, immed, (ulong)mt_cmd->mt_count);
			break;

		case MTREW:
			log_tape(drv, TRC_MT_REW);
			rc = tape_rewind(&drv, drv->rew_immed);
			break;

		case MTOFFL:
			log_tape(drv, TRC_MT_OFFL);
			rc = tape_load_unload(&drv, False, False);
			break;

		case MTNOP:
			log_tape(drv, TRC_MT_NOP);
			rc = tape_wfm(&drv, False, False);
			break;

		case MTRETEN:
			log_tape(drv, TRC_MT_RETEN);
			rc = tape_rewind(&drv, False);
			break;

		case MTBSFM:
			log_tape(drv, TRC_MT_BSFM);
			/* ReverseSpaceFileMark positions before the filemark */
			rc = tape_set_pos(&drv, SpaceFileMark,
				-((long)mt_cmd->mt_count));
			if(rc == STATUS_SUCCESS)
				rc = tape_set_pos(&drv, SpaceFileMark, 1);
			break;

		case MTFSFM:
			log_tape(drv, TRC_MT_FSFM);
			/* ForwardSpaceFileMark positions after the filemark */
			rc = tape_set_pos(&drv, SpaceFileMark,
				(long)mt_cmd->mt_count);
			if(rc == STATUS_SUCCESS)
				rc = tape_set_pos(&drv, SpaceFileMark, -1);
			break;

		case MTEOM:
			log_tape(drv, TRC_MT_EOM);
			rc = tape_set_pos(&drv, SpaceEOM,
				(long)mt_cmd->mt_count);
			break;

		case MTERASE:
			log_tape(drv, TRC_MT_ERASE);
			if(!WriteAccess(drv->access_mode)) {
				rc = -EPERM;
				break;
			} /* if */
			rc = tape_erase(&drv, True, False);
			break;

		case MTRAS1:
			log_tape(drv, TRC_MT_RAS1);
			rc = tape_self_test(&drv);
			break;

		case MTRAS2:
			rc = -ENOSYS;
			break;

		case MTRAS3:
			rc = -ENOSYS;
			break;

		case MTSETBLK:
			log_tape(drv, TRC_MT_SETBLK);
			rc = tape_set_media_parm(&drv, BlockSizeOnly,
				True, False, (ulong)mt_cmd->mt_count);
			break;

		case MTSETDENSITY:
			rc = -ENOSYS;
			break;

		case MTSEEK:
			log_tape(drv, TRC_MT_SEEK);
			rc = tape_set_pos(&drv, LocateBlock,
				(long)mt_cmd->mt_count);
			break;

		case MTTELL:
			log_tape(drv, TRC_MT_TELL);
			rc = mttell(&drv, mt_cmd, arg);
			break;

		case MTSETDRVBUFFER:
			log_tape(drv, TRC_MT_SETDRVBUF);
			rc = tape_set_media_parm(&drv, BufferModeOnly,
				mt_cmd->mt_count, False, False);
			break;

		case MTFSS:
			rc = -ENOSYS;
			break;

		case MTBSS:
			rc = -ENOSYS;
			break;

		case MTWSM:
			rc = -ENOSYS;
			break;

		case MTLOCK:
			log_tape(drv, TRC_MT_LOCK);
			rc = tape_prevent_medium_remove(&drv, True);
			break;

		case MTUNLOCK:
			log_tape(drv, TRC_MT_UNLOCK);
			rc = tape_prevent_medium_remove(&drv, False);
			break;

		case MTLOAD:
			log_tape(drv, TRC_MT_LOAD);
			rc = tape_load_unload(&drv, False, True);
			break;

		case MTUNLOAD:
			log_tape(drv, TRC_MT_UNLOAD);
			rc = tape_load_unload(&drv, False, False);
			break;

		case MTCOMPRESSION:
			log_tape(drv, TRC_MT_COMPRESS);
			rc = tape_set_compression(&drv, mt_cmd->mt_count);
			break;

		case MTSETPART:
			rc = -ENOSYS;
			break;

		case MTMKPART:
			rc = -ENOSYS;
			break;

		default:
			DbgPrint(("lin_tape: lin_tape_drive_ioctl: Invalid "\
				"MT IOCTL Call\n"));
			log_tape(drv, TRC_INV_MT);
			rc = -EINVAL;
			break;
		} /* switch */
	} else {
		switch(command) {
		case SCSI_IOCTL_TEST_UNIT_READY:
			log_tape(drv, TRC_SCSI_TUR);
			rc = tape_tur(&drv);
			break;

		case SCSI_IOCTL_SEND_COMMAND:
		case SCSI_IOCTL_BENCHMARK_COMMAND:
		case SCSI_IOCTL_SYNC:
		case SCSI_IOCTL_START_UNIT:
		case SCSI_IOCTL_STOP_UNIT:
		case SCSI_IOCTL_DOORLOCK:
		case SCSI_IOCTL_DOORUNLOCK:
			if(drv->trace)
				tape_log_msg(drv->tape_log,
					&drv->tape_log_len, TRC_SCSI_IOC);
			rc = scsi_ioctl(drv->dev_obj, command, (void*)arg);
			break;

		case MTDEVICE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(int), TRC_MTDVC, False, True, False,
				False, True, mtdevice);
			break;

		case STIOCDM:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stdm_s), TRC_DISPLAY_MSG,
				False, True, True, True, False, stiocdm);
			break;

		case STIOCTOP:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stop), TRC_STOP,
				False, True, True, True, False, stioctop);
			break;

		case STIOCHGP:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stchgp), TRC_CHGP,
				False, True, True, True, False, stiochgp);
			break;

		case SIOC_INQUIRY:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct inquiry_data), TRC_INQ,
				False, True, True, False, True, sioc_inquiry);
			break;

		case SIOC_REQSENSE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct request_sense), TRC_REQSENSE,
				False, True, True, False, True, sioc_reqsense);
			break;

		case SIOC_RESERVE:
			rc = sioc_reserve(&drv);
			break;

		case SIOC_RELEASE:
			rc = sioc_release(&drv);
			break;

		case SIOC_TEST_UNIT_READY:
			log_tape(drv, TRC_SIOC_TUR);
			rc = tape_tur(&drv);
			break;

		case SIOC_LOG_SENSE_PAGE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct log_sense_page),
				TRC_LOG_SNS_PG, False, True, True,
				True, True, sioc_log_sense_page);
			break;

		case SIOC_LOG_SENSE10_PAGE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct log_sense10_page),
				TRC_LOG_SNS10_PG, False, True, True,
				True, True, sioc_log_sense10_page);
			break;

		case SIOC_ENH_LOG_SENSE:
			rc = sioc_enh_log_sense(&drv, arg);
			break;

		case SIOC_MODE_SENSE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct mode_sense),
				TRC_MODE_SNS_PG, False, True, True,
				True, True, sioc_mode_sense);
			break;

		case SIOC_MODE_SENSE_PAGE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct mode_sense_page),
				TRC_MODE_SNS_PG, False, True, True,
				True, True, sioc_mode_sense_page);
			break;

		case SIOC_INQUIRY_PAGE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct inquiry_page), TRC_INQ_PG, False,
				True, False, True, True, sioc_inquiry_page);
			break;

		case STIOCSETP:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stchgp_s), TRC_SETP, False,
				True, True, True, False, stiocsetp);
			break;

		case STIOCQRYP:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stchgp_s), TRC_QRYP, False,
				True, True, False, True, stiocqryp);
			break;

		case STIOCSYNC:
			log_tape(drv, TRC_SYNC);
			rc = tape_wfm(&drv, False, 0);
			break;

		case STIOCQRYPOS:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stpos_s), TRC_QRYPOS, False,
				True, True, False, True, stiocqrypos);
			break;

		case STIOCSETPOS:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stpos_s), TRC_SETPOS, False,
				True, True, True, False, stiocsetpos);
			break;

		case STIOCQRYSENSE:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct stsense_s), TRC_QRYSENSE, False,
				True, True, True, True, stiocqrysense);
			break;

		case STIOCQRYINQUIRY:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct st_inquiry),
				TRC_QRYINQ, False, True, True, False, True,
				tape_std_inquiry);
			break;

		case STIOC_LOCATE:
			rc = drvioc_exe(&drv, command, arg, sizeof(uint),
				TRC_LOCATE, False, True, True, True, False,
				stioc_locate);
			break;

		case STIOC_LOCATE_16:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct set_tape_position),
				-1, True, False, True, True, False,
				stioc_locate);
			break;

		case STIOC_READ_POSITION:
			rc = stioc_read_position(&drv, arg);
			break;

		case STIOC_RESET_DRIVE:
			log_tape(drv, TRC_RESET_DRV);
			rc = tape_reset_drive(&drv);
			break;

		case STIOC_PREVENT_MEDIUM_REMOVAL:
			log_tape(drv, TRC_PRV_MEDIUM);
			rc = tape_prevent_medium_remove(&drv, True);
			break;

		case STIOC_ALLOW_MEDIUM_REMOVAL:
			log_tape(drv, TRC_ALLOW_MEDIUM);
			rc = tape_prevent_medium_remove(&drv, False);
			break;

		case STIOC_REPORT_DENSITY_SUPPORT:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct report_density_support),
				TRC_RPT_DENSITY, False, True, True, True, True,
				stioc_report_density_support);
			break;

		case STIOC_GET_DENSITY:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct density_data_t), -1, False,
				False, True, False, True, tape_get_density);
			break;

		case STIOC_SET_DENSITY:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct density_data_t), -1, False,
				False, True, True, False, tape_set_density);
			break;

		case GET_ENCRYPTION_STATE:
			rc = get_encryption_state(&drv, arg);
			break;

		case SET_ENCRYPTION_STATE:
			rc = set_encryption_state(&drv, arg);
			break;

		case SET_DATA_KEY:
			rc = set_data_key(&drv, arg);
			break;

		case DATA_ENCRYPTION_DIAGNOSTICS:
			rc = data_encryption_diagnostics(&drv, arg);
			break;

		case QUERY_DRIVER_VERSION:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct QueryDriverVersion),
				TRC_QVER_CMD, False, True, True, False, True,
				lin_tape_query_driver_version);
			break;

		case SIOC_QUERY_PATH:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct scsi_path), TRC_QRY_PATH,
				False, True, True, False, True,
				tape_query_path);
			break;

		case SIOC_DEVICE_PATHS:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct device_paths), TRC_DVC_PATHS,
				False, True, True, False, True,
				tape_show_paths);
			break;

		case SIOC_ENABLE_PATH:
			rc = lin_tape_tsioc_enable_path(&drv, arg);
			break;

		case SIOC_DISABLE_PATH:
			rc = lin_tape_tsioc_disable_path(&drv, arg);
			break;

		case SCSI_PASS_THROUGH:
			rc = lin_tape_tscsi_pass_through(&drv, command, arg);
			break;

		case SIOC_PASS_THROUGH:
			rc = lin_tape_tsioc_pass_through(&drv, command, arg);
			break;

		case STIOC_READ_RESERVEKEYS:
			rc = stioc_read_reserve_keys(&drv, command, arg);
			break;

		case STIOC_READ_RESERVATIONS:
			rc = stioc_read_reservations(&drv, command, arg);
			break;

		case STIOC_REGISTER_KEY:
			rc = stioc_register_key(&drv);
			break;

		case STIOC_REMOVE_REGISTRATION:
			rc = stioc_remove_registration(&drv);
			break;

		case STIOC_CLEAR_ALL_REGISTRATIONS:
			rc = stioc_clear_all_registrations(&drv);
			break;

		case DDIOC_SET_RESERVE_KEY:
			rc = lin_tape_tsm_set_reserve_key(drv, command, arg);
			break;

		case DDIOC_QUERY_PR_KEY:
			rc = lin_tape_tsm_query_pr_key(drv, command, arg, 0);
			break;

		case STIOC_QUERY_PARTITION:
			rc = lin_tape_query_partition(drv, command, arg);
			break;

		case STIOC_CREATE_PARTITION:
			rc = lin_tape_create_partition(drv, command, arg);
			break;

		case STIOC_SET_ACTIVE_PARTITION:
			rc = lin_tape_set_active_partition(&drv, command, arg);
			break;

		case STIOC_ALLOW_DATA_OVERWRITE:
			rc = lin_tape_allow_data_overwrite(drv, command, arg);
			break;

		case STIOC_READ_POSITION_EX:
			rc = tape_get_pos_ex(&drv, command, arg);
			break;

		case STIOC_QUERY_BLK_PROTECTION:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct logical_block_protection), -1,
				False, False, True, False, True,
				tape_qry_blk_prot);
			break;

		case STIOC_SET_BLK_PROTECTION:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct logical_block_protection), -1,
				False, False, True, True, False,
				tape_set_blk_prot);
			break;

		case STIOC_QUERY_EOT_WARN:
			rc = stioc_query_eot_warn(&drv, command, arg);
			break;

		case STIOC_SET_EOT_WARN:
			rc = stioc_set_eot_warn(&drv, command, arg);
			break;

		case STIOC_VERIFY_TAPE_DATA:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct verify_data), -1, False, False,
				True, True, False, stioc_verify_tape_data);
			break;

		case STIOC_QUERY_RAO:
			rc = drvioc_exe(&drv, command, arg,
				sizeof(struct query_rao_info),TRC_RAO, False, True,
				True, True, True, stioc_query_rao);
			break;

		case STIOC_GENERATE_RAO:
			rc = stioc_generate_rao(&drv, arg);
			break;

		case STIOC_RECEIVE_RAO:
			rc = stioc_receive_rao(&drv, arg);
			break;

		case STIOC_QUERY_AMU:
			rc = lin_tape_query_amu(drv, command, arg);
			break;

		case STIOC_SET_AMU:
			rc = lin_tape_set_amu(drv, command, arg);
			break;

        case STIOC_LOAD_UNLOAD:
            rc = lin_tape_load_unload(&drv, command, arg);
            break;

		default:
			printk("lin_tape: lin_tape_drive_ioctl invalid "\
				"ioctl.\n");
			log_tape(drv, TRC_INV_IOC);
			rc = -EINVAL;

			break;
		} /* switch */
	} /* else */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, -rc);

	/* if it is a non-sense command, update last errno */
	if(command != SIOC_REQSENSE && command != STIOCQRYSENSE &&
		command != DDIOC_RETURN_ERR) drv->last_errno = -rc;

EXIT_LABEL:

	if(mt_cmd) kfree(mt_cmd);

	*drvp = drv;
	return rc;

} /* lin_tape_drive_ioctl() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_ioctl_special                                                *
*                                                                              *
*  PURPOSE: Handle Special IOCTLs (preemption)                                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the IOCTL code                                                   *
*    arg: as passed in to the ioctl() call                                     *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

int lin_tape_ioctl_special(struct tape_descriptor **drv, uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct sp_dev sp_drv_sn;

	DbgPrint(("lin_tape_special_ioctl entry\n"));

	if (!*drv && command!=STIOC_SET_SPDEV) rc=-EINVAL;
	else{
	switch(command) {
	case STIOC_SET_SPDEV:
		if(_IOC_SIZE(command) != sizeof(struct sp_dev)) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */
		if(copy_from_user(&sp_drv_sn, (char*)arg, sizeof(struct sp_dev))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	    rc = lin_tape_get_drv_by_sn(drv, sp_drv_sn.device_serial);
	    if(rc != STATUS_SUCCESS) {
	    	*drv = NULL;
	    	rc = -ENXIO;
	    }
		break;
	case STIOC_READ_RESERVEKEYS:
		rc = stioc_read_reserve_keys(drv, command, arg);
		break;

	case STIOC_READ_RESERVATIONS:
		rc = stioc_read_reservations(drv, command, arg);
		break;

	case STIOC_REGISTER_KEY:
		rc = stioc_register_key(drv);
		break;

	case STIOC_REMOVE_REGISTRATION:
		rc = stioc_remove_registration(drv);
		break;

	case STIOC_CLEAR_ALL_REGISTRATIONS:
		rc = stioc_clear_all_registrations(drv);
		break;

	default:
		DbgPrint(("lin_tape: lin_tape_ioctl_special unknown cmd 0x%x\n",
			command));
		rc = -EINVAL;
		break;
	} /* switch */
	}

EXIT_LABEL:
	DbgPrint(("lin_tape_ioctl_special exit: %d\n", rc));
	return rc;
} /* lin_tape_ioctl_special() */

/*******************************************************************************
*                                                                              *
*  NAME: log_tape                                                              *
*                                                                              *
*  PURPOSE: Post a log message of an IOCTL call                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive for which we want to perform the request                   *
*    trc_msg: message to post to daemon                                        *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_tape(struct tape_descriptor* drv, char trc_msg)
{
	char ttype = lin_tape_trace_type;
	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len, trc_msg);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, trc_msg, 0, NULL);
} /* log_tape() */

/*******************************************************************************
*                                                                              *
*  NAME: drvioc_exe                                                            *
*                                                                              *
*  PURPOSE: Prepare and send a drive IOCTL request                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: the drive for which we want to perform the request                  *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*    buflen: expected size of command argument                                 *
*    trc_msg: message to post to daemon                                        *
*    cpy_method: to user or to kernel                                          *
*    trace: if true, attempt to log                                            *
*    chk_sz: enforce the length of the input buffer                            *
*    fm: copy data from user space                                             *
*    to: copy data to user space                                               *
*    func: the function that will perform the main part of the command         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int drvioc_exe(struct tape_descriptor** drvp, uint command, ulong arg,
	ssize_t buflen, char trc_msg, char cpy_method, boolean trace,
	boolean chk_sz, boolean fm, boolean to, drv_method func)
{
	int rc = -EIO;
	void* buf = NULL;
	struct tape_descriptor* drv = *drvp;

	if(trace) log_tape(drv, trc_msg);

	if(chk_sz && _IOC_SIZE(command) != buflen) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(buflen, GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', buflen);

	if(fm && copy_from_user((char*)buf, (char*)arg, buflen)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */


	rc = func(&drv, (void*)buf, cpy_method);

	if(to && rc == STATUS_SUCCESS &&
		copy_to_user((char*)arg, (char*)buf, buflen)) rc = -EFAULT;


EXIT_LABEL:

	*drvp = drv;
	if(buf) kfree(buf);
	return rc;
} /* drvioc_exe() */

/*******************************************************************************
*                                                                              *
*  NAME: mttell                                                                *
*                                                                              *
*  PURPOSE: Handle MTTELL IOCTL                                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    mt_cmd: caller-allocated mtop structure to copy to user space             *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t mttell(struct tape_descriptor** drvp, struct mtop* mt_cmd, ulong arg)
{
	struct pos_info* read_pos_buf = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct pos_info));

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	mt_cmd->mt_count = read_pos_buf->first_blk_pos;

	if(copy_to_user((char*)arg, (char*)mt_cmd, sizeof(struct mtop)))
		rc = -EFAULT;

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);

	*drvp = drv;
	return rc;
} /* mttell() */

/*******************************************************************************
*                                                                              *
*  NAME: mtiocget                                                              *
*                                                                              *
*  PURPOSE: Return mtstatus information to user in mtget struct                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int mtiocget(struct tape_descriptor** drvp, uint command, ulong arg)
{
	struct mtget *mt_status = NULL;
	struct pos_info *read_pos_buf = NULL;
	struct tape_sense_data *buf = NULL;
	struct tape_descriptor* drv = *drvp;
	int len = 0, resid = 0, i = 0, rc = STATUS_SUCCESS;

	log_tape(drv, TRC_MTIOCGET);

	if(_IOC_SIZE(command) != sizeof(struct mtget)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	mt_status = kmalloc(sizeof(struct mtget), GFP_KERNEL);
	if(!mt_status) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(mt_status, '\0', sizeof(struct mtget));

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(read_pos_buf == NULL) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct pos_info));

	buf = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
	if(buf == NULL) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset(buf, '\0', sizeof(struct tape_sense_data));

	mt_status->mt_type = MT_ISSCSI2;
	mt_status->mt_resid = 0;
	mt_status->mt_gstat = 0;
	mt_status->mt_erreg = 0;
	mt_status->mt_fileno = -1;

	/* update blk_size, wp, density_code, buffered_mode */
	rc = tape_get_media_parm(&drv);
	if(rc == STATUS_SUCCESS) {
		if(drv->write_protect)
			mt_status->mt_gstat |= GMT_WR_PROT(0xFFFFFFFF);
		mt_status->mt_dsreg = ((drv->blk_size << MT_ST_BLKSIZE_SHIFT)
				& MT_ST_BLKSIZE_MASK)
			| ((drv->density << MT_ST_DENSITY_SHIFT)
				& MT_ST_DENSITY_MASK);
	} else mt_status->mt_dsreg = False;

	if(drv->eom) mt_status->mt_gstat |= GMT_EOT(0xFFFFFFFF);

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);
	if(rc == STATUS_SUCCESS) {
		if(read_pos_buf->bop)
			mt_status->mt_gstat |= GMT_BOT(0xFFFFFFFF);
		if(read_pos_buf->eop)
			mt_status->mt_gstat |= GMT_EOT(0xFFFFFFFF);
		mt_status->mt_blkno = read_pos_buf->first_blk_pos;
	} else mt_status->mt_blkno = -1;

	if(drv->fm_detected) mt_status->mt_gstat |= GMT_EOF(0xFFFFFFFF);
	if(drv->eod) mt_status->mt_gstat |= GMT_EOD(0xFFFFFFFF);
	rc = tape_request_sense(&drv, &len, &resid, (void*)buf);
	if(rc == STATUS_SUCCESS) {
		if(buf->fm)
			mt_status->mt_gstat |= GMT_EOF(0xFFFFFFFF);
		if(buf->eom)
			mt_status->mt_gstat |= GMT_EOT(0xFFFFFFFF);
	} /* if */

	if(drv->write_protect || drv->logic_write_protect)
		mt_status->mt_gstat |= GMT_WR_PROT(0xFFFFFFFF);

	for(i = 0; i < 20; i++) {
		rc = tape_tur(&drv);
		if(rc == STATUS_SUCCESS) {
			mt_status->mt_gstat |=  GMT_ONLINE(0xFFFFFFFF);
			break;
		} else if(drv->sense_valid &&
			drv->sense_data->sense_key == NOT_READY
			&& drv->sense_data->asc == 0x3A
			&& drv->sense_data->ascq == 0x00) {
			mt_status->mt_gstat |= GMT_DR_OPEN(0xFFFFFFFF);
			break;
		}  /* else if */
	} /* for */

	rc = copy_to_user((char*)arg, (char*)mt_status, sizeof(struct mtget))
		?  -EFAULT : STATUS_SUCCESS;

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);
	if(buf) kfree(buf);
	if(mt_status) kfree(mt_status);

	*drvp = drv;
	return rc;
} /* mtiocget() */

/*******************************************************************************
*                                                                              *
*  NAME: mtiocpos                                                              *
*                                                                              *
*  PURPOSE: Return position to user in mtpos struct                            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int mtiocpos(struct tape_descriptor** drvp, ulong arg)
{
	struct mtpos *pos = NULL;
	struct pos_info *read_pos_buf = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	log_tape(drv, TRC_MTIOCPOS);

	if(!arg) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	pos = kmalloc(sizeof(struct mtpos), GFP_KERNEL);
	if(!pos) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(pos, '\0', sizeof(struct mtpos));

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct mtpos));

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);
	if(rc == STATUS_SUCCESS) {
		pos->mt_blkno = (long)read_pos_buf->first_blk_pos;
		if(copy_to_user((char*)arg, (char*)pos, sizeof(struct mtpos)))
			rc = -EFAULT;
	} /* if */

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);
	if(pos) kfree(pos);

	*drvp = drv;
	return rc;
} /* mtiocpos() */

/*******************************************************************************
*                                                                              *
*  NAME: mtdevice                                                              *
*                                                                              *
*  PURPOSE: Return MTDEVICE number                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user dev_num integer into which to copy mtdevice number              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t mtdevice(struct tape_descriptor** drvp, void* buf, char unused)
{
	unchar *page = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS, plength = 0, offset = 0, i = 0;
	int dev_num = *(int*)buf;

	page = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!page) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(page, '\0', MAX_INQ_LEN);

	rc = tape_inquiry_page(&drv, UnitSerialPage, page);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	plength = page[INQ_HEADER_LEN - 1];

	if(plength < 5 || page[1] != UnitSerialPage) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* use the last 5 digit of the serial number */
	offset = INQ_HEADER_LEN + plength - 5;
	for(i = offset; i < INQ_HEADER_LEN + plength; i++) {
		if(!is_hex(page[i])) {
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */
	} /* for */

	dev_num = char_to_hex(page[offset]) << 20;
	dev_num |= char_to_hex(page[offset + 1]) << 16;
	dev_num |= char_to_hex(page[offset + 2]) << 12;
	dev_num |= char_to_hex(page[offset + 3]) << 8;
	dev_num |= char_to_hex(page[offset + 4]) << 4;
	dev_num &= 0xFFFFF0;

	*(int*)buf = dev_num;

EXIT_LABEL:

	if(page) kfree(page);

	*drvp = drv;
	return rc;
} /* mtdevice() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocdm                                                               *
*                                                                              *
*  PURPOSE: Handle STIOCDM (display message) IOCTL                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stdm structure from which to get the message to display         *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocdm(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct stdm_s* dmbuf = (struct stdm_s*)buf;
	struct tape_descriptor* drv = *drvp;
	int i = 0, rc = STATUS_SUCCESS;

	if(!dmbuf->dm_msg0 || !dmbuf->dm_msg1) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < MAXMSGLEN; i++) {
		if(dmbuf->dm_msg0[i] >= 'a' && dmbuf->dm_msg0[i] <= 'z')
			dmbuf->dm_msg0[i] = dmbuf->dm_msg0[i] - 'a' + 'A';
		if(dmbuf->dm_msg0[i] == '\0')
			dmbuf->dm_msg0[i] = ' ';

		if(dmbuf->dm_msg1[i] >= 'a' && dmbuf->dm_msg1[i] <= 'z')
			dmbuf->dm_msg1[i] = dmbuf->dm_msg1[i] - 'a' + 'A';
		if(dmbuf->dm_msg1[i] == '\0')
			dmbuf->dm_msg1[i] = ' ';
	} /* for */

	rc = tape_display_msg(&drv, dmbuf->dm_func, (char*)dmbuf->dm_msg0,
		(char*)dmbuf->dm_msg1);
EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* stiocdm() */

/*******************************************************************************
*                                                                              *
*  NAME: stioctop                                                              *
*                                                                              *
*  PURPOSE: Handle STIOCTOP IOCTL                                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stop structure from which to get the parameter data             *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stioctop(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct stop* st_cmd = (struct stop*)buf;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	switch(st_cmd->st_op) {
	case STOFFL:
		log_tape(drv, TRC_STOFFL);
		rc = tape_load_unload(&drv, False, False);
		break;

	case STREW:
		log_tape(drv, TRC_STREW);
		rc = tape_rewind(&drv, drv->rew_immed);
		break;

	case STERASE:
		log_tape(drv, TRC_STERASE);
		if(!WriteAccess(drv->access_mode)) {
			rc = -EPERM;
			break;
		} /* if */
		rc = tape_erase(&drv, !st_cmd->st_count, False);
		break;

	case STRETEN:
		log_tape(drv, TRC_STRETEN);
		rc = tape_rewind(&drv, drv->rew_immed);
		break;

	case STWEOF:
		log_tape(drv, TRC_STWEOF);
		if(!WriteAccess(drv->access_mode)) {
			rc = -EPERM;
			break;
		} /* if */

		if(petro==1 && st_cmd->st_count==1 && drv->wfm_immed)
			rc = tape_wfm(&drv, True, (ulong)st_cmd->st_count);
		else
			rc = tape_wfm(&drv, False, (ulong) st_cmd->st_count);
		break;

	case STFSF:
		log_tape(drv, TRC_STFSF);
		rc = tape_set_pos(&drv, SpaceFileMark, (long)st_cmd->st_count);
		break;

	case STRSF:
		log_tape(drv, TRC_STRSF);
		rc = tape_set_pos(&drv, SpaceFileMark,
			-((long)st_cmd->st_count));
		break;

	case STFSR:
		log_tape(drv, TRC_STFSR);
		rc = tape_set_pos(&drv, SpaceRecord, (long)st_cmd->st_count);
		break;

	case STRSR:
		log_tape(drv, TRC_STRSR);
		rc = tape_set_pos(&drv, SpaceRecord, -((long)st_cmd->st_count));
		break;

	case STINSRT:
		log_tape(drv, TRC_STINSRT);
		rc = tape_load_unload(&drv, False, True);
		break;

	case STEJECT:
		log_tape(drv, TRC_STEJECT);
		rc = tape_load_unload(&drv, False, False);
		break;

	case STTUR:
		log_tape(drv, TRC_STTUR);
		rc = tape_tur(&drv);
		break;

	case STLOAD:
		log_tape(drv, TRC_STLOAD);
		rc = tape_load_unload(&drv, False, True);
		break;

	case STSEOD:
		log_tape(drv, TRC_STSEOD);
		rc = tape_set_pos(&drv, SpaceEOM, False);
		break;

	case STFSSF:
		log_tape(drv, TRC_STFSSF);
		rc = tape_set_pos(&drv, SpaceSequentialFM,
			((long)st_cmd->st_count));
		break;

	case STRSSF:
		log_tape(drv, TRC_STRSSF);
		rc = tape_set_pos(&drv, SpaceSequentialFM,
			((long)st_cmd->st_count));
		break;

	default:
		rc = -EINVAL;
		break;
	} /* switch */

	*drvp = drv;
	return rc;

} /* stioctop() */

/*******************************************************************************
*                                                                              *
*  NAME: stiochgp                                                              *
*                                                                              *
*  PURPOSE: Handle STIOCHGP IOCTL                                              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stchgp structure from which to get the parameter data           *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiochgp(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct stchgp* ch_parm = (struct stchgp*)buf;
	return tape_set_media_parm(drvp, BlockSizeOnly, True, False,
		(ulong)ch_parm->st_blksize);

} /* stiochgp() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_inquiry                                                          *
*                                                                              *
*  PURPOSE: Handle SIOC_INQUIRY IOCTL                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user inquiry_data structure into which to copy data                  *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_inquiry(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct inquiry_data* inq_buf = (struct inquiry_data*)buf;
	struct inquiry* ibuf = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	ibuf = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!ibuf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(ibuf, '\0', MAX_INQ_LEN);

	rc = tape_std_inquiry(&drv, (void*)ibuf, False);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	inq_buf->qual = ibuf->dev_type_qualifier;
	inq_buf->type = ibuf->dev_type;
	inq_buf->rm = ibuf->rm;
	inq_buf->mod = ibuf->dev_type_modifier;
	inq_buf->iso = ibuf->iso;
	inq_buf->ecma = ibuf->ecma;
	inq_buf->ansi = ibuf->ansi;
	inq_buf->aenc = ibuf->aenc;
	inq_buf->trmiop = ibuf->trmiop;
	inq_buf->rdf = ibuf->response_data_format;
	inq_buf->len = ibuf->additional_length;
	inq_buf->mchngr = ibuf->mchngr;
	inq_buf->reladr = ibuf->relative_addressing;
	inq_buf->wbus32 = ibuf->wide_32bit;
	inq_buf->wbus16 = ibuf->wide_16bit;
	inq_buf->sync = ibuf->sync;
	inq_buf->linked = ibuf->linked_cmds;
	inq_buf->cmdque = ibuf->cmd_queue;
	inq_buf->sftre = ibuf->soft_reset;
	memcpy(inq_buf->vid, ibuf->vendor_id, 8);
	memcpy(inq_buf->pid, ibuf->product_id, 16);
	memcpy(inq_buf->revision, ibuf->revision, 4);
	memcpy(inq_buf->vendor1, ibuf->vendor_specific, 20);

EXIT_LABEL:

	if(ibuf) kfree(ibuf);
	*drvp = drv;
	return rc;
} /* sioc_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_reqsense                                                         *
*                                                                              *
*  PURPOSE: Handle SIOC_REQSENSE IOCTL                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user request_sense structure into which to copy data                 *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_reqsense(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct request_sense* sense_buf = (struct request_sense*)buf;
	struct tape_descriptor* drv = *drvp;
	struct tape_sense_data* sbuf = NULL;
	int len = 0, resid = 0, rc = STATUS_SUCCESS;
	boolean allocd = False;

	if(drv->sense_valid
		/* See if a TUR causes drv->sense_valid to be True */
		|| (tape_tur(&drv) != STATUS_SUCCESS && drv->sense_valid)) {
		drv->sense_valid = False;
		sbuf = drv->sense_data;
	} else {
		sbuf = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
		if(!sbuf) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		allocd = True;
		memset(sbuf, '\0', sizeof(struct tape_sense_data));
		rc = tape_request_sense(&drv, &len, &resid, (void*)sbuf);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* else */

	sense_buf->valid = sbuf->valid;
	sense_buf->err_code = sbuf->err_code;
	sense_buf->segnum = sbuf->seg_num;
	sense_buf->fm = sbuf->fm;
	sense_buf->eom = sbuf->eom;
	sense_buf->ili = sbuf->ili;
	sense_buf->key = sbuf->sense_key;
	sense_buf->info = (sbuf->information[0] << 24)
		+ (sbuf->information[1] << 16)
		+ (sbuf->information[2] << 8)
		+ sbuf->information[3];
	sense_buf->addlen = sbuf->additional_sense_len;
	sense_buf->cmdinfo = (sbuf->cmd_spec[0] << 24)
		+ (sbuf->cmd_spec[1] << 16)
		+ (sbuf->cmd_spec[2] << 8)
		+ sbuf->cmd_spec[3];
	sense_buf->asc = sbuf->asc;
	sense_buf->ascq = sbuf->ascq;
	sense_buf->fru = sbuf->field_replaceable;
	sense_buf->sksv = sbuf->sksv;
	sense_buf->cd = sbuf->cd;
	sense_buf->bpv = sbuf->bpv;
	sense_buf->sim = sbuf->bit_ptr;
	sense_buf->field[0] = sbuf->sksv_field_ptr;
	sense_buf->field[1] = sbuf->sense_key_spec;
	sense_buf->vendor[0] = sbuf->vendor_specific1;
	sense_buf->vendor[1] = sbuf->vendor_specific2;
	sense_buf->vendor[2] = sbuf->vendor_specific3;
	sense_buf->vendor[3] = sbuf->vendor_specific4;
	sense_buf->vendor[4] = sbuf->vendor_specific5;
	sense_buf->vendor[5] = sbuf->vendor_specific6;
	memcpy(&sense_buf->vendor[6],
		sbuf->vendor_specific_sense,
		MAX_ADDITIONAL_SENSE - 6);
	rc = STATUS_SUCCESS;

EXIT_LABEL:

	if(allocd) kfree(sbuf);
	*drvp = drv;
	return rc;
} /* sioc_reqsense() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_reserve                                                          *
*                                                                              *
*  PURPOSE: Handle SIOC_RESERVE IOCTL                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int sioc_reserve(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	unchar *key = NULL;
	unsigned int pr_generation = 0;
	char pr_type = 0;
	int rc = STATUS_SUCCESS;
	char* rkey = NULL;

	log_tape(drv, TRC_RESERVE);

	if(!TP_RSV_PERSISTENT(drv)) {
		rc = tape_reserve(&drv);
		if(rc == STATUS_SUCCESS) {
			drv->curr_reserve_type = LinTapeReserve6;
			drv->explicit_reserved = True;
		} /* if */
		goto EXIT_LABEL;
	} /* if */

	key = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!key) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(key, '\0', RESERVE_KEY_LENGTH);

	rkey = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!rkey) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(rkey, '\0', RESERVE_KEY_LENGTH);

	if(drv->curr_reserve_type == LinTapePersistentReserve) {
		/* read the reservation */
		rc = tape_read_reservation(&drv, key, &pr_generation, &pr_type);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		if(!memcmp(drv->reservation_key, key, RESERVE_KEY_LENGTH)
			&& drv->pr_generation == pr_generation
			&& pr_type == EXCLUSIVE_ACCESS) {
			rc = STATUS_SUCCESS;
			drv->explicit_reserved = True;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	/* register a reservation key */
	get_random_bytes(rkey, RESERVE_KEY_LENGTH);
	rc = tape_register_reservation_key(&drv, rkey);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* create a reservation */
	rc = tape_create_persist_reserve(&drv);
	if(rc != STATUS_SUCCESS) {
		/* undo previous work */
		rc = tape_register_reservation_key(&drv, NULL);
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
		goto EXIT_LABEL;
	} /* if */

	/* update pr_generation field */
	rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
	if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;

	drv->curr_reserve_type = LinTapePersistentReserve;
	drv->explicit_reserved = True;

EXIT_LABEL:

	if(key) kfree(key);
	if(rkey) kfree(rkey);
	*drvp = drv;
	return rc;
} /* sioc_reserve() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_release                                                          *
*                                                                              *
*  PURPOSE: Handle SIOC_RELEASE IOCTL                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int sioc_release(struct tape_descriptor** drvp)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;
	unsigned int pr_generation = 0;

	log_tape(drv, TRC_RELEASE);
	if(drv->curr_reserve_type == LinTapePersistentReserve) {
		/* failover device */
		/* register reservation with NULL key clears the reserve */
		rc = tape_register_reservation_key(&drv, NULL);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;

		drv->curr_reserve_type = LinTapeNoReserve;
		drv->explicit_reserved = False;
	} else {
		/* non-failover device */
		rc = tape_release(&drv);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		drv->curr_reserve_type = LinTapeNoReserve;
		drv->explicit_reserved = False;
	} /* else */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* sioc_release() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_log_sense_page                                                   *
*                                                                              *
*  PURPOSE: Handle SIOC_LOG_SENSE_PAGE IOCTL                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user log_sense_page structure into which to copy data                *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_log_sense_page(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct log_sense_page* page = (struct log_sense_page*)buf;
	uint length = 0;
	int rc = STATUS_SUCCESS;

	memset(page->data, '\0', LOGSENSEPAGE);
	rc = tape_logsense_page(drvp, page->page_code, 0, page->data, &length);
	if(rc == STATUS_SUCCESS) page->len = (ushort)length;

	return rc;
} /* sioc_log_sense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_log_sense10_page                                                 *
*                                                                              *
*  PURPOSE: Handle SIOC_LOG_SENSE10_PAGE IOCTL (supporting subpages)           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user log_sense10_page structure into which to copy data              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_log_sense10_page(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct log_sense10_page* page = (struct log_sense10_page*)buf;
	uint length = (uint) page->len;
	int rc = STATUS_SUCCESS;

	if (length > LOGSENSEPAGE) length = LOGSENSEPAGE;
	memset(page->data, '\0', LOGSENSEPAGE);

    rc=tape_logsense_page(drvp, page->page_code, page->subpage_code,
		                page->data, &length);
	if(rc == STATUS_SUCCESS) page->len = (ushort)length;

	return rc;
} /* sioc_log_sense10_page() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_enh_log_sense                                                    *
*                                                                              *
*  PURPOSE: Handle SIOC_ENH_LOG_SENSE IOCTL (supporting subpages and extended  *
* 			return size.)												       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user enh_log_sense structure copy for kernel space                   *
*    unused:                                                                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_enh_log_sense(struct tape_descriptor** drvp, ulong user)
{
	struct tape_descriptor* drv = *drvp;

	uint length;
	int rc = STATUS_SUCCESS;
	char *info = NULL;
	int structsize;
	struct enh_log_sense* page = NULL;

	log_tape(drv, TRC_ENH_LOG_SNS);

	structsize = sizeof(struct enh_log_sense);

	page = kmalloc(structsize, GFP_KERNEL);
    if(!page){
		printk("lin_tape: ENH_LOG_SENSE unable to alloc %x bytes "\
			"in kernel\n", structsize);
		rc = -ENOMEM;
		goto EXIT_LABEL;
    }
	if(copy_from_user((char*)page, (char*)user, structsize)){
		rc = -EFAULT;
		goto EXIT_LABEL;
	}
	length = page->len;
	if(length < 4){
		printk("lin_tape: SIOC_ENH_LOG_SENSE "\
			"must have a length >= header size\n");
		rc = -EINVAL;
		goto EXIT_LABEL;
	}
	info = kmalloc(length, GFP_KERNEL);
	if(!info){
		printk("lin_tape: ENH_LOG_SENSE unable to alloc %x bytes "\
			"in kernel\n", length);
		rc = -ENOMEM;
		goto EXIT_LABEL;
	}
	memset(info, '\0', length);

	rc = tape_logsense_page(drvp, page->page_code, page->subpage_code,
		info, &length);
	if(rc == STATUS_SUCCESS) {
		page->len = (ushort)length;

		rc = copy_to_user((char*)user, (char*)page,
			(structsize - sizeof(char *)));
		if(!rc){
			rc=copy_to_user((char*)page->logdatap, info, length);
			if(rc) printk("lin_tape: Error copying to user space: %d", rc);
		}
	}

EXIT_LABEL:
	if(info) kfree(info);
	if(page) kfree(page);

	return rc;
} /* sioc_enh_log_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_mode_sense                                                       *
*                                                                              *
*  PURPOSE: Handle SIOC_MODE_SENSE IOCTL                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user mode_sense structure into which to copy data                    *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_mode_sense(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct mode_sense* page = (struct mode_sense*)buf;

	memset(page->data, '\0', MAX_MDSNS_LEN);
	page->cmd_code = MODE_SENSE_10;
	return tape_modesense10_page(drvp, page->page_code, page->data,
		MAX_MDSNS_LEN, page->subpage_code, False, True);
} /* sioc_mode_sense() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_mode_sense_page                                                  *
*                                                                              *
*  PURPOSE: Handle SIOC_MODE_SENSE_PAGE IOCTL                                  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user mode_sense_page structure into which to copy data               *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_mode_sense_page(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct mode_sense_page* page = (struct mode_sense_page*)buf;

	memset(page->data, '\0', MAX_MDSNS_LEN);
	return tape_modesense_page(drvp, page->page_code, page->data,
		MAX_MDSNS_LEN, True, False);
} /* sioc_mode_sense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: sioc_inquiry_page                                                     *
*                                                                              *
*  PURPOSE: Handle SIOC_INQUIRY_PAGE IOCTL                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user inquiry_page structure into which to copy data                  *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t sioc_inquiry_page(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct inquiry_page* page = (struct inquiry_page*)buf;

	memset(page->data, '\0', MAX_INQ_LEN);
	return tape_inquiry_page(drvp, page->page_code, page->data);
} /* sioc_inquiry_page() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocsetp                                                             *
*                                                                              *
*  PURPOSE: Handle STIOCSETP IOCTL                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stchgp_s structure from which to set data                       *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocsetp(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct stchgp_s* parm_buf = (struct stchgp_s*)buf;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	/* get the media paramters and compress setting */
	rc = tape_update_all_parms(&drv);
	if(rc < 0) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	rc = stiocsetp_check_inval(drv, parm_buf);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = stiocsetp_cap_scal(&drv, parm_buf);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = tape_data_safe_mode_and_pew(&drv, parm_buf);
	if(rc == -EPERM) rc = STATUS_SUCCESS;
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* update the medium settings */
	if(parm_buf->buffered_mode != drv->buffered_mode
		|| parm_buf->density_code != drv->density
		|| parm_buf->blksize != drv->blk_size
		|| (!drv->blk_size && drv->sili && parm_buf->read_sili_bit)) {
		rc = tape_set_media_parm(&drv, AllModifications,
			parm_buf->buffered_mode, parm_buf->density_code,
			parm_buf->blksize);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: unable to set medium settings\n");
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	drv->blk_size = parm_buf->blksize;
	drv->trace = parm_buf->trace;
	drv->buffered_mode = parm_buf->buffered_mode;
	drv->disable_simmim_log = parm_buf->disable_sim_logging;
	drv->disable_auto_drive_dump = parm_buf->disable_auto_drive_dump;
	drv->auto_logsense = parm_buf->logging;

	/* disable or enable compression */
	if(parm_buf->compression != drv->compression) {
		rc = tape_set_compression(&drv,
			parm_buf->compression);

		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: tape_set_all_parms unable to "\
				"change compression\n");
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	drv->compression = parm_buf->compression;
	drv->trailer_labels = parm_buf->trailer_labels;
	drv->rew_immed = parm_buf->rewind_immediate;
	drv->busy_retry = parm_buf->busy_retry;

	drv->trace = parm_buf->trace;
	if(petro == 1) drv->wfm_immed = parm_buf->wfm_immediate;

	if(parm_buf->logical_write_protect != drv->logic_write_protect
		&& (drv->drv_id == IBM_3590 || drv->drv_id == IBM_3592)) {

		rc = tape_tur(&drv);
		if(rc == STATUS_SUCCESS) {
			rc = tape_set_medium_sense(&drv,
				parm_buf->logical_write_protect);
			if(rc != STATUS_SUCCESS) {
				printk("lin_tape: unable to set medium sense\n");
				goto EXIT_LABEL;
			} /* if */
			drv->logic_write_protect =
				parm_buf->logical_write_protect;
		} /* if */
	} /* if */

	drv->read_mode = parm_buf->read_past_filemark;

	if(drv->sili != parm_buf->read_sili_bit) {
		drv->sili = parm_buf->read_sili_bit;

		DbgPrint(("lin_tape: SILI bit is set to %d\n", drv->sili));

		if(!drv->blk_size && drv->sili) {
			rc = tape_set_media_parm(&drv, BlockSizeOnly, True,
				False, 4);
			drv->blk_size = 0;
		} else if(!drv->blk_size && !drv->sili) {
			rc = tape_set_media_parm(&drv, BlockSizeOnly, True,
				False, 0);
		} /* else if */
	} /* if */

	/* set the read/write recovery timeout if applicable */
	if(drv->limit_read_recov != parm_buf->limit_read_recov
		|| drv->limit_write_recov != parm_buf->limit_write_recov) {
		rc = tape_set_rw_error_recovery_page(&drv,
			parm_buf->limit_read_recov,
			parm_buf->limit_write_recov);
		if(rc == STATUS_SUCCESS) {
			drv->limit_write_recov = parm_buf->limit_write_recov;
			drv->limit_read_recov = parm_buf->limit_read_recov;
		} else printk("lin_tape: %s tape_set_all_parms unable to "\
			"change recovery timeouts\n", drv->devid);
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* stiocsetp() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocsetp_check_inval                                                 *
*                                                                              *
*  PURPOSE: Check for invalid parameters passed by user to STIOCSETP           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    parm_buf: buffer with the parameters to set                               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stiocsetp_check_inval(struct tape_descriptor* drv,
	struct stchgp_s* parm_buf)
{
	int rc = -EINVAL;

	if(parm_buf->max_scsi_xfer < drv->blk_size) goto EXIT_LABEL;

	if(parm_buf->record_space_mode != SCSI_SPACE_MODE) {
		printk("lin_tape: stiocsetp invalid space mode");
		goto EXIT_LABEL;
	} /* if */

	if(parm_buf->logical_write_protect != NO_PROTECT
		&& drv->drv_id == IBM_3580) goto EXIT_LABEL;

	if(drv->drv_id == IBM_3580) {
		if(parm_buf->capacity_scaling == SCALE_VALUE
			&& parm_buf->capacity_scaling_value != Capacity100)
			goto EXIT_LABEL;

		if(parm_buf->capacity_scaling != SCALE_100)
			goto EXIT_LABEL;
	} /* if */

	if(parm_buf->blksize != 0 && (parm_buf->blksize > drv->max_blk_size
		|| parm_buf->blksize < drv->min_blk_size
		|| parm_buf->blksize > drv->max_xfer_len)) goto EXIT_LABEL;

	if(parm_buf->read_past_filemark != 0 &&
		parm_buf->read_past_filemark != 1) goto EXIT_LABEL;

	if(parm_buf->capacity_scaling > SCALE_VALUE) goto EXIT_LABEL;

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	return rc;

} /* stiocsetp_check_inval() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocsetp_cap_scal                                                    *
*                                                                              *
*  PURPOSE: Scale capacity.  This is a supporter function to STIOCSETP         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    parm_buf: buffer with the parameters to set                               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stiocsetp_cap_scal(struct tape_descriptor** drvp,
	struct stchgp_s* parm_buf)
{
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	if(parm_buf->capacity_scaling != SCALE_VALUE) {
		/* SCALE_100, SCALE_75, SCALE_50, SCALE_25 */
		if(parm_buf->capacity_scaling != drv->capacity_scaling) {
			if(parm_buf->capacity_scaling == SCALE_100)
				rc = tape_capacity_scaling(&drv,
					parm_buf->capacity_scaling,
					Capacity100);
			else if(parm_buf->capacity_scaling == SCALE_75)
				rc = tape_capacity_scaling(&drv,
					parm_buf->capacity_scaling,
					Capacity75);
			else if(parm_buf->capacity_scaling == SCALE_50)
				rc = tape_capacity_scaling(&drv,
					parm_buf->capacity_scaling,
					Capacity50);
			else if(parm_buf->capacity_scaling == SCALE_25)
				rc = tape_capacity_scaling(&drv,
					parm_buf->capacity_scaling,
					Capacity25);
			else
				rc = -EINVAL;

			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */
	} else if(drv->capacity_scaling != SCALE_VALUE
		|| (drv->capacity_scaling_hex_val
			!= parm_buf->capacity_scaling_value)) {
		rc = tape_capacity_scaling(&drv, parm_buf->capacity_scaling,
			parm_buf->capacity_scaling_value);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;

} /* stiocsetp_cap_scal() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocqryp                                                             *
*                                                                              *
*  PURPOSE: Handle STIOCQRYP IOCTL                                             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stchgp_s structure into which to copy data                      *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocqryp(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct tape_descriptor* drv = *drvp;
	struct stchgp_s* parm_buf = (struct stchgp_s*)buf;
	unchar *temp_mode = NULL;
	int rc = STATUS_SUCCESS;

	rc = tape_update_all_parms(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	parm_buf->blksize = drv->blk_size;
	parm_buf->trace = drv->trace;
	parm_buf->hkwrd = drv->trc_hook;
	parm_buf->autoload = drv->autoload_mode;
	parm_buf->buffered_mode = drv->buffered_mode;
	parm_buf->compression = drv->compression;
	parm_buf->trailer_labels = drv->trailer_labels;
	parm_buf->rewind_immediate = drv->rew_immed;
	parm_buf->logging = drv->auto_logsense;
	parm_buf->write_protect = drv->write_protect;
	parm_buf->min_blksize = drv->min_blk_size;
	parm_buf->max_blksize = drv->max_blk_size;
	parm_buf->max_scsi_xfer = drv->max_xfer_len;
	parm_buf->acf_mode = drv->acf_mode;
	parm_buf->record_space_mode = SCSI_SPACE_MODE;
	parm_buf->logical_write_protect = drv->logic_write_protect;

	parm_buf->capacity_scaling = drv->capacity_scaling;
	parm_buf->capacity_scaling_value = drv->capacity_scaling_hex_val;

	parm_buf->limit_read_recov = drv->limit_read_recov;
	parm_buf->limit_write_recov = drv->limit_write_recov;
	parm_buf->retain_reservation = drv->retain_reservation;
	parm_buf->alt_pathing = False;
	parm_buf->emulate_autoloader = False;
	parm_buf->medium_type = drv->medium_type;
	parm_buf->density_code = drv->density;
	parm_buf->disable_sim_logging = drv->disable_simmim_log;
	parm_buf->disable_auto_drive_dump = drv->disable_auto_drive_dump;
	parm_buf->read_sili_bit = drv->sili;
	parm_buf->read_past_filemark = drv->read_mode;
	parm_buf->wfm_immediate = drv->wfm_immed;
	parm_buf->data_safe_mode = drv->data_safe_mode;
	memcpy(parm_buf->pews, drv->pews, 2);
	parm_buf->busy_retry = drv->busy_retry;
	if(TP_RSV_PERSISTENT(drv)) parm_buf->reserve_type = 1;
	else parm_buf->reserve_type = 0;

	if(drv->drv_id != IBM_3590 && drv->drv_id != IBM_3592) goto EXIT_LABEL;

	temp_mode = kmalloc(MAX_MDSNS_LEN, GFP_KERNEL);
	if(!temp_mode) goto EXIT_LABEL;

	memset(temp_mode, '\0', MAX_MDSNS_LEN);
	rc = tape_modesense_page(&drv, 0x20, temp_mode, MAX_MDSNS_LEN, True,
		False);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	switch(temp_mode[2]) {
	case 0:
		drv->acf_mode=ACF_NONE;
		break;
	case 1:
		drv->acf_mode=ACF_SYSTEM;
		break;
	case 2:
		drv->acf_mode=ACF_RANDOM;
		break;
	case 3:
		drv->acf_mode=ACF_MANUAL;
		break;
	case 4:
		drv->acf_mode=ACF_ACCUMULATE;
		break;
	case 5:
		drv->acf_mode=ACF_AUTOMATIC;
		break;
	default:
		drv->acf_mode=ACF_NONE;
		break;
	} /* switch */
	parm_buf->acf_mode = drv->acf_mode;

EXIT_LABEL:

	if(temp_mode) kfree(temp_mode);
	*drvp = drv;
	return rc;
} /* stiocqryp() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocqrypos                                                           *
*                                                                              *
*  PURPOSE: Handle STIOCQRYPOS IOCTL                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: user stpos_s structure into which to copy data                       *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocqrypos(struct tape_descriptor** drvp, void* buf, char unused)
{
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;
	struct stpos_s* pos = (struct stpos_s*)buf;
	struct pos_info *read_pos_buf = NULL;

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct pos_info));

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);
	pos->partition_number = False;

	if(rc == STATUS_SUCCESS) {
		pos->eot = read_pos_buf->eop;
		pos->curpos = read_pos_buf->first_blk_pos;
		pos->tapepos = read_pos_buf->first_blk_pos;
		pos->lbot = read_pos_buf->last_blk_pos;
		pos->num_blocks = read_pos_buf->blks_in_buf;
		pos->num_bytes = read_pos_buf->bytes_in_buf;
		pos->bot = read_pos_buf->bop;
	} /* if */

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);

	*drvp = drv;
	return rc;
} /* stiocqrypos() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocsetpos                                                           *
*                                                                              *
*  PURPOSE: Handle STIOCSETPOS IOCTL                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: stpos_s structure copied from userspace                              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocsetpos(struct tape_descriptor** drvp, void* buf, char unused)
{
	struct stpos_s* pos = (struct stpos_s*)buf;
	return tape_set_pos(drvp, LocateBlock, (long)pos->curpos);
} /* stiocsetpos() */

/*******************************************************************************
*                                                                              *
*  NAME: stiocqrysense                                                         *
*                                                                              *
*  PURPOSE: Handle STIOCQRYSENSE IOCTL                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: the buffer into which to copy the sense data                         *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stiocqrysense(struct tape_descriptor** drvp, void* buf, char unused)
{
	int rc = STATUS_SUCCESS;
	struct stsense_s* sense_buf = (struct stsense_s*)buf;
	struct tape_descriptor* drv = *drvp;

	memset(sense_buf->sense, '\0', MAXSENSE);

	if(sense_buf->sense_type == FRESH) {
		rc = tape_request_sense(&drv, &sense_buf->len,
			&sense_buf->residual_count, sense_buf->sense);
		goto EXIT_LABEL;
	} /* if */

	if(sense_buf->sense_type != LASTERROR) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(drv->sense_valid) {
		drv->sense_valid = False;
		memcpy(sense_buf->sense, drv->sense_data,
			sizeof(struct tape_sense_data));
		sense_buf->len = (int)drv->sense_data->additional_sense_len + 8;
		if(drv->sense_data->valid) {
			sense_buf->residual_count =
				(int)((drv->sense_data->information[0]<<24)
				+ (drv->sense_data->information[1]<<16)
				+ (drv->sense_data->information[2]<<8)
				+ (drv->sense_data->information[3]));
		} else sense_buf->residual_count = False;
	} else {
		sense_buf->len = False;
		sense_buf->residual_count = False;
	} /* else */

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* stiocqrysense() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_locate                                                          *
*                                                                              *
*  PURPOSE: Handle STIOC_LOCATE IOCTL                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: the position requested by the user                                   *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stioc_locate(struct tape_descriptor** drvp, void* buf, char loc_16)
{
	struct set_tape_position* pos = (struct set_tape_position*)buf;
	return loc_16 ? tape_set_pos(drvp,
			pos->logical_id_type == LOGICAL_ID_BLOCK_TYPE ?
			Locate16Block : Locate16File,
			(long long)pos->logical_id) :
		tape_set_pos(drvp, LocateBlock, *(long long*)buf);
} /* stioc_locate() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_read_position                                                   *
*                                                                              *
*  PURPOSE: Handle STIOC_READ_POSITION IOCTL                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_read_position(struct tape_descriptor** drvp, ulong arg)
{
	struct pos_info* read_pos_buf = NULL;
	struct tape_descriptor* drv = *drvp;
	uint blk_id = 0;
	int rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len, TRC_READ_POS);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_READ_POS, 0, NULL);

	read_pos_buf = kmalloc(sizeof(struct pos_info), GFP_KERNEL);
	if(!read_pos_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(read_pos_buf, '\0', sizeof(struct pos_info));

	rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);

	if(rc == STATUS_SUCCESS) {
		blk_id = (uint)read_pos_buf->first_blk_pos;
		if(copy_to_user((char*)arg, (char*)&blk_id, sizeof(blk_id)))
			rc = -EFAULT;
	} /* if */

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);
	*drvp = drv;
	return rc;
} /* stioc_read_position() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_report_density_support                                          *
*                                                                              *
*  PURPOSE: Handle STIOC_REPORT_DENSITY_SUPPORT IOCTL                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    buf: report_density_support structure to receive data formatted for user  *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t stioc_report_density_support(struct tape_descriptor** drvp, void* buf,
	char unused)
{
	struct report_density_support* density_buf =
		(struct report_density_support*)buf;
	struct tape_descriptor* drv = *drvp;
	struct report_density* report = NULL;
	int i = 0, rc = STATUS_SUCCESS;

	memset(density_buf->reports, '\0',
		(sizeof(struct density_report) * MAX_DENSITY_REPORTS));

	report = kmalloc(sizeof(struct report_density), GFP_KERNEL);
	if(!report) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(report, '\0', sizeof(struct report_density));

	rc = tape_report_density(&drv, density_buf->media,
		&density_buf->number_reports, (void*)report);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	for(i = 0; i < density_buf->number_reports; i++) {
		density_buf->reports[i].primary_density_code =
			report->descriptor[i].primary_density;
		density_buf->reports[i].secondary_density_code =
			report->descriptor[i].second_density;
		density_buf->reports[i].wrtok =
			report->descriptor[i].wrtok;
		density_buf->reports[i].dup =
			report->descriptor[i].dup;
		density_buf->reports[i].deflt =
			report->descriptor[i].deflt;
		density_buf->reports[i].bits_per_mm =
			(report->descriptor[i].bits_per_mm[0] << 16)
			+ (report->descriptor[i].bits_per_mm[1] << 8)
			+ report->descriptor[i].bits_per_mm[2];
		density_buf->reports[i].media_width =
			(report->descriptor[i].media_width[0] << 8)
			+ report->descriptor[i].media_width[1];
		density_buf->reports[i].tracks =
			(report->descriptor[i].tracks[0] << 8)
			+ (report->descriptor[i].tracks[1]);

		density_buf->reports[i].capacity =
			(report->descriptor[i].capacity[0] << 24)
			+ (report->descriptor[i].capacity[1] << 16)
			+ (report->descriptor[i].capacity[2] << 8)
			+ (report->descriptor[i].capacity[3]);

		memcpy(density_buf->reports[i].assigning_org,
			report->descriptor[i].assign_org, 8);
		memcpy(density_buf->reports[i].density_name,
			report->descriptor[i].density_name, 8);
		memcpy(density_buf->reports[i].description,
			report->descriptor[i].description, 20);
	} /* for */

EXIT_LABEL:

	if(report) kfree(report);
	*drvp = drv;
	return rc;
} /* stioc_report_density_support() */

/*******************************************************************************
*                                                                              *
*  NAME: get_encryption_state                                                  *
*                                                                              *
*  PURPOSE: Get encryption state for system-managed encryption                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int get_encryption_state(struct tape_descriptor** drvp, ulong arg)
{
	struct encryption_status* status = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	status = kmalloc(sizeof(struct encryption_status), GFP_KERNEL);
	if(!status) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(status, '\0', sizeof(struct encryption_status));

	status->encryption_capable = drv->encryption_capable;
	if(!drv->encryption_capable) {
		status->encryption_method = METHOD_NONE;
		status->encryption_state = STATE_OFF;
	} else rc = tape_get_encryption_state(&drv, status);

	if(copy_to_user((char*)arg, (char*)status,
		sizeof(struct encryption_status))) rc = -EFAULT;

EXIT_LABEL:

	if(status) kfree(status);

	*drvp = drv;
	return rc;
} /* get_encryption_state() */

/*******************************************************************************
*                                                                              *
*  NAME: set_encryption_state                                                  *
*                                                                              *
*  PURPOSE: Set encryption state for system-managed encryption                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_encryption_state(struct tape_descriptor** drvp, ulong arg)
{
	struct encryption_status* status = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	if(!drv->encryption_capable) {
		rc = -EINVAL;
		DbgPrint(("%s: not encryption capable.\n", drv->devid));
		goto EXIT_LABEL;
	} /* if */

	status = kmalloc(sizeof(struct encryption_status), GFP_KERNEL);
	if(!status) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(status, '\0', sizeof(struct encryption_status));

	rc = tape_get_encryption_state(&drv, status);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(status->encryption_method != METHOD_APPLICATION) {
		printk("%s: drive is not currently set to "\
			"application-managed encryption.\n", drv->devid);
		rc = -EACCES;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user((char*)status, (char*)arg,
		sizeof(struct encryption_status))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(status->encryption_method != METHOD_APPLICATION) {
		printk("%s: cannot set encryption state for non-application "\
			"method.\n", drv->devid);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(status->encryption_state != STATE_OFF &&
		status->encryption_state != STATE_ON) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_set_encryption_state(&drv, status);
	if(rc == STATUS_SUCCESS) {
		drv->application_set_encryption = True;
		drv->encryption_state = status->encryption_state;
	} /* if */

EXIT_LABEL:

	if(status) kfree(status);

	*drvp = drv;
	return rc;
} /* set_encryption_state() */

/*******************************************************************************
*                                                                              *
*  NAME: set_data_key                                                          *
*                                                                              *
*  PURPOSE: Set data key for system-managed encryption                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int set_data_key(struct tape_descriptor** drvp, ulong arg)
{
	struct data_key* data_key = NULL;
	struct encryption_status* status = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	if(!drv->encryption_capable) {
		rc = -EINVAL;
		DbgPrint(("%s: not encryption capable.\n", drv->devid));
		goto EXIT_LABEL;
	} /* if */

	/* query the drive's setting first */
	status = kmalloc(sizeof(struct encryption_status), GFP_KERNEL);
	if(!status) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(status, '\0', sizeof(struct encryption_status));

	rc = tape_get_encryption_state(&drv, status);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(status->encryption_method != METHOD_APPLICATION) {
		printk("%s: drive is not currently set to "\
			"application-managed encryption.\n", drv->devid);
		rc = -EACCES;
		goto EXIT_LABEL;
	} /* if */

	/* copy data key from application */
	data_key = kmalloc(sizeof(struct data_key), GFP_KERNEL);
	if(!data_key) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(data_key, '\0', sizeof(struct data_key));

	if(copy_from_user((unchar*)data_key, (unchar*)arg,
		sizeof(struct data_key))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_set_data_key(&drv, data_key);
	DbgPrint(("%s: set data key, rc = %d\n", drv->devid, rc));

EXIT_LABEL:

	if(data_key) kfree(data_key);
	if(status)   kfree(status);

	*drvp = drv;
	return rc;
} /* set_data_key() */

/*******************************************************************************
*                                                                              *
*  NAME: data_encryption_diagnostics                                           *
*                                                                              *
*  PURPOSE: Run system-managed encryption diagnostic test                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int data_encryption_diagnostics(struct tape_descriptor** drvp, ulong arg)
{
	struct encryption_status* status = NULL;
	struct encryption_diagnostics* diag = NULL;
	struct tape_descriptor* drv = *drvp;
	int rc = STATUS_SUCCESS;

	if(!drv->encryption_capable) {
		rc = -EINVAL;
		DbgPrint(("%s: not encryption capable.\n", drv->devid));
		goto EXIT_LABEL;
	} /* if */

	status = kmalloc(sizeof(struct encryption_status), GFP_KERNEL);
	if(!status) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(status, '\0', sizeof(struct encryption_status));

	diag = kmalloc(sizeof(struct encryption_diagnostics), GFP_KERNEL);
	if(!diag) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_get_encryption_state(&drv, status);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(METHOD_SYSTEM != status->encryption_method) {
		printk("%s: drive is not currently set to system-managed "\
			"encryption.\n", drv->devid);
		rc = -EACCES;
		goto EXIT_LABEL;
	} /* if */

	memset(diag, '\0', sizeof(struct encryption_diagnostics));

	if(copy_from_user((char*)diag, (char*)arg,
		sizeof(struct encryption_diagnostics))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = encryption_diag(&drv, diag);

	if(copy_to_user((char*)arg, (char*)diag,
		sizeof(struct encryption_diagnostics))) rc = -EFAULT;

EXIT_LABEL:

	if(status) kfree(status);
	if(diag) kfree(diag);

	*drvp = drv;
	return rc;
} /* data_encryption_diagnostics() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsioc_enable_path                                            *
*                                                                              *
*  PURPOSE: Enable a drive path                                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_tsioc_enable_path(struct tape_descriptor** drvp, ulong arg)
{
	unchar ttype = lin_tape_trace_type;
	int count = 0, i = 0, rc = -EINVAL;
	struct tape_descriptor* drv = *drvp;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_ENABLE_IOCTL);

	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_ENABLE_IOCTL, 0, NULL);

	DbgPrint(("lin_tape: Enabling path %d for IBMtape%d\n", (int)arg,
		drv->dev_index));

	if(arg < 1) goto EXIT_LABEL;

	count = 0;
	WRITE_LOCK(&tape_des_lock);

	/* find the path */
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv || !tape_devs[i]->drv->serial_num_len ||
			tape_devs[i]->drv->serial_num_len != drv->serial_num_len ||
			memcmp(tape_devs[i]->drv->serial_num, drv->serial_num,
				drv->serial_num_len)) continue;
		count++;
		if(count == (int)arg) break;
	} /* for */

	WRITE_UNLOCK(&tape_des_lock);

	if(i == MaxLinTapeDevices) goto EXIT_LABEL;

	DbgPrint(("lin_tape: Enabling path IBMtape%d\n", i));

	/* have daemon mark device as enabled */
	rc = tape_setup_and_log_full(tape_devs[i]->drv, tape_devs[i]->drv->dev_index,
		TRC_TP_ENABLED, NULL, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	tape_devs[i]->drv->fenced = False;
	tape_devs[i]->drv->enabled = True;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* lin_tape_tsioc_enable_path() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsioc_disable_path                                           *
*                                                                              *
*  PURPOSE: Disable a drive path                                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_tsioc_disable_path(struct tape_descriptor** drvp, ulong arg)
{
	unchar ttype = lin_tape_trace_type;
	int i = 0, count = 0, already_disabled = 0, rc = -EINVAL;
	struct tape_descriptor* drv = *drvp;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_DISABLE_IOCTL);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_DISABLE_IOCTL, 0, NULL);

	DbgPrint(("lin_tape: Disabling path %d for IBMtape%d\n", (int)arg,
		drv->dev_index));

	if(arg < 1) goto EXIT_LABEL;
	count = 0;

	WRITE_LOCK(&tape_des_lock);

	/* find the path */
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv || !tape_devs[i]->drv->serial_num_len ||
			tape_devs[i]->drv->serial_num_len != drv->serial_num_len ||
			memcmp(tape_devs[i]->drv->serial_num, drv->serial_num,
				drv->serial_num_len)) continue;

		count++;
		if(count == (int)arg) break;
	} /* for */

	WRITE_UNLOCK(&tape_des_lock);

	if(i == MaxLinTapeDevices) goto EXIT_LABEL;

	DbgPrint(("lin_tape: Disabling path IBMtape%d\n", i));

	/* have daemon mark device as disabled */
	rc = tape_setup_and_log_full(tape_devs[i]->drv, tape_devs[i]->drv->dev_index,
		TRC_TP_DISABLED, NULL, 0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	already_disabled = !tape_devs[i]->drv->enabled;
	tape_devs[i]->drv->enabled = False;

	rc = STATUS_SUCCESS;

	if(drv != tape_devs[i]->drv) goto EXIT_LABEL;

	/* the current path is being disabled */
	/*  check with the daemon for a better path */
	tape_get_best_path(&drv);
	if(drv == tape_devs[i]->drv) {
		/* didn't get a better path */
		if(!already_disabled) {
			tape_devs[i]->drv->enabled = True;
			tape_setup_and_log_full(tape_devs[i]->drv,
				tape_devs[i]->drv->dev_index, TRC_TP_ENABLED,
					NULL, 0);
		} /* if */
		rc = -EINVAL;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* lin_tape_tsioc_disable_path() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tscsi_pass_through                                           *
*                                                                              *
*  PURPOSE: Handle SCSI_PASS_THROUGH request for drives                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_tscsi_pass_through(struct tape_descriptor** drvp, uint command,
	ulong arg)
{
	PSCSIPassThrough buf = NULL;
	int buffer_length = 0, resid = 0, result = 0, rc = STATUS_SUCCESS;
	unchar sense_length = 0; unchar sense_avail = 0,
		ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &(drv->tape_log_len), TRC_PTCMD);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_PTCMD, 0, NULL);

	if(_IOC_SIZE(command) != sizeof(SCSIPassThrough)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(sizeof(SCSIPassThrough), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', sizeof(SCSIPassThrough));

	if(copy_from_user((char*)buf, (char*)arg, sizeof(SCSIPassThrough))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	buffer_length = buf->BufferLength;
	sense_length = 0;

	rc = tape_pass_thru(&drv, buf->CDB, buf->CommandLength, buf->Buffer,
		&buffer_length, buf->DataDirection, buf->TimeOut, NULL,
		&sense_length, &sense_avail, &resid, &result, False);

	buf->TargetStatus = status_byte(result);
	buf->MessageStatus = msg_byte(result);
	buf->HostStatus = host_byte(result);
	buf->DriverStatus = driver_byte(result);

	if(buf->CDB[0] == READ_6) buf->BufferLength = buffer_length;

	if(sense_avail) {
		buf->SenseDataValid = True;
		buf->SenseKey = drv->sense_data->sense_key;
		buf->ASC = drv->sense_data->asc;
		buf->ASCQ = drv->sense_data->ascq;
	} else {
		buf->SenseDataValid = FALSE;
		buf->SenseKey = 0;
		buf->ASC = 0;
		buf->ASCQ = 0;
	} /* else */

	if(rc == STATUS_SUCCESS) {
		if(copy_to_user((char*)arg, (char*)buf,
			sizeof(SCSIPassThrough))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

EXIT_LABEL:

	if(buf) kfree(buf);
	*drvp = drv;

	return rc;
} /* lin_tape_tscsi_pass_through() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_tsioc_pass_through                                           *
*                                                                              *
*  PURPOSE: Handle SIOC_PASS_THROUGH request for drives                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_tsioc_pass_through(struct tape_descriptor** drvp, uint command,
	ulong arg)
{
	struct sioc_pass_through* pass_through = NULL;
	unchar sense_avail = 0, * cdb = NULL, ttype = lin_tape_trace_type;
	int rc = STATUS_SUCCESS;
	struct tape_descriptor* drv = *drvp;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log, &drv->tape_log_len, TRC_PTCMD);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_PTCMD, 0, NULL);

	if(_IOC_SIZE(command) != sizeof(struct sioc_pass_through)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	pass_through = kmalloc(sizeof(struct sioc_pass_through), GFP_KERNEL);
	if(!pass_through) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(pass_through, '\0', sizeof(struct sioc_pass_through));

	if(copy_from_user((char*)pass_through, (char*)arg,
		sizeof(struct sioc_pass_through))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	cdb = kmalloc(pass_through->cmd_length, GFP_KERNEL);
	if(!cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(copy_from_user((char*)cdb, (char*)pass_through->cdb,
		pass_through->cmd_length)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_pass_thru(&drv, cdb, pass_through->cmd_length,
		pass_through->buffer, &pass_through->buffer_length,
		pass_through->data_direction, pass_through->timeout,
		pass_through->sense, &pass_through->sense_length, &sense_avail,
		&pass_through->resid, &pass_through->result, False);

	pass_through->target_status = status_byte(pass_through->result);
	pass_through->msg_status = msg_byte(pass_through->result);
	pass_through->host_status = host_byte(pass_through->result);
	pass_through->driver_status = driver_byte(pass_through->result);

	if(rc == STATUS_SUCCESS) {
		if(copy_to_user((char*)arg, (char*)pass_through,
			sizeof(struct sioc_pass_through))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

EXIT_LABEL:

	if(pass_through) kfree(pass_through);
	if(cdb) kfree(cdb);
	*drvp = drv;

	return rc;
} /* lin_tape_tsioc_pass_through() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_read_reserve_keys                                               *
*                                                                              *
*  PURPOSE: Read persistent reservation information from drive                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_read_reserve_keys(struct tape_descriptor** drvp, uint command,
	ulong arg)
{
	struct read_keys *buf = NULL;
	struct persist_read_keys *key_buf = NULL;
	struct persist_read_keys *key_tmp = NULL;
	struct tape_descriptor* drv = *drvp;
	ulong order = 0, key_len = 0, buf_size = 0;
	int rc = 0;

	if(drv->trace)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_STIOC_READ_KEYS);

	if(drv->curr_reserve_type == LinTapeReserve6 ||
		_IOC_SIZE(command) != sizeof(struct read_keys)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(sizeof(struct read_keys), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	key_buf = kmalloc(sizeof(struct persist_read_keys) - sizeof(char),
		GFP_KERNEL);
	if(!key_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(buf, '\0', sizeof(struct read_keys));
	memset(key_buf, '\0', sizeof(struct persist_read_keys) - sizeof(char));

	if(copy_from_user((char*)buf, (char*)arg, sizeof(struct read_keys))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* get the length of key list first */
	rc = tape_read_keys(&drv, key_buf, sizeof(struct persist_read_keys)
		- sizeof(char));
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	key_len = (key_buf->additional_length[0] << 24)
		+ (key_buf->additional_length[1] << 16)
		+ (key_buf->additional_length[2] << 8)
		+ key_buf->additional_length[3];

	buf->generation = (key_buf->generation[0] << 24)
		+ (key_buf->generation[1] << 16)
		+ (key_buf->generation[2] << 8)
		+ key_buf->generation[3];
	buf->length = key_len;

	if(!buf->reserve_key_list) {
		if(copy_to_user((char*)arg, (char*)buf,
			sizeof(struct read_keys))) rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* get the key values */
	key_len += (sizeof(struct persist_read_keys) - sizeof(char));
	if(key_len > KmallocThreshhold) {
		order = get_mem_ord(key_len, &buf_size);
		key_tmp = (struct persist_read_keys*)
			__get_free_pages(GFP_KERNEL, order);
	} else key_tmp = kmalloc(key_len, GFP_KERNEL);

	if(!key_tmp) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(key_tmp, '\0', key_len);
	memcpy(key_tmp, key_buf, sizeof(struct persist_read_keys)
		- sizeof(char));

	kfree(key_buf);
	key_buf = key_tmp;
	key_tmp = NULL;

	rc = tape_read_keys(&drv, key_buf, key_len);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* copy the generation and length first */
	if(copy_to_user((char*)arg, (char*)buf, 8)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* copy key values */
	if(copy_to_user(buf->reserve_key_list, key_buf->key_list,
		buf->length)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(key_buf) {
		if(key_len <= KmallocThreshhold) kfree(key_buf);
		else free_pages((ulong)key_buf, order);
	} /* if */
	if(buf) kfree(buf);

	*drvp = drv;
	return rc;
} /* stioc_read_reserve_keys() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_read_reservations                                               *
*                                                                              *
*  PURPOSE: Read persistent reservation information from drive                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_drive_ioctl                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_read_reservations(struct tape_descriptor** drvp, uint command,
	ulong arg)
{
	struct read_reserves *buf = NULL;
	struct reserve_descriptor *descriptors = NULL;
	struct reserve_descriptor *temp1 = NULL;
	struct persist_read_reservs *res_buf = NULL, *res_buf_tmp = NULL;
	struct tape_descriptor* drv = *drvp;
	ulong order = 0, res_len = 0, buf_size = 0;
	int count = 0, i = 0, rc = 0;

	if(drv->trace)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_STIOC_READ_RSVS);

	if(drv->curr_reserve_type == LinTapeReserve6 ||
		_IOC_SIZE(command) != sizeof(struct read_reserves)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(sizeof(struct read_reserves), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', sizeof(struct read_reserves));

	res_buf = kmalloc(sizeof(struct persist_read_reservs) - sizeof(char),
		GFP_KERNEL);
	if(!res_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(res_buf, '\0',
		sizeof(struct persist_read_reservs) - sizeof(char));

	if(copy_from_user((char*)buf, (char*)arg,
		sizeof(struct read_reserves))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* get the length of key list first */
	rc = tape_read_reserves(&drv, res_buf, PERSIST_PAGE_HEAD_SIZE);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	res_len = (res_buf->additional_length[0] << 24)
		+ (res_buf->additional_length[1] << 16)
		+ (res_buf->additional_length[2] << 8)
		+ res_buf->additional_length[3];

	buf->generation = (res_buf->generation[0] << 24)
		+ (res_buf->generation[1] << 16)
		+ (res_buf->generation[2] << 8)
		+ res_buf->generation[3];
	buf->length = res_len;

	res_len += PERSIST_PAGE_HEAD_SIZE;

	if(!buf->reserve_list) {
		if(copy_to_user((char*)arg, (char*)buf,
			sizeof(struct read_reserves))) rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* get the reservation descriptors */
	if(res_len > KmallocThreshhold) {
		order = get_mem_ord(res_len, &buf_size);
		res_buf_tmp = (struct persist_read_reservs*)
			__get_free_pages(GFP_KERNEL, order);
		if(!res_buf_tmp) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(res_buf_tmp, '\0', buf_size);
		memcpy(res_buf_tmp, res_buf, PERSIST_PAGE_HEAD_SIZE);
		kfree(res_buf);
		res_buf = res_buf_tmp;
		res_buf_tmp = NULL;

		descriptors = (struct reserve_descriptor*)
			__get_free_pages(GFP_KERNEL, order);
	} else {
		res_buf_tmp = kmalloc(res_len, GFP_KERNEL);
		if(!res_buf_tmp) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(res_buf_tmp, '\0', res_len);

		memcpy(res_buf_tmp, res_buf, PERSIST_PAGE_HEAD_SIZE);
		kfree(res_buf);
		res_buf = res_buf_tmp;
		res_buf_tmp = NULL;

		descriptors = kmalloc(res_len, GFP_KERNEL);
	} /* else */

	if(!descriptors) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	rc = tape_read_reserves(&drv, res_buf, res_len);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	/* copy the generation and length first */
	if(copy_to_user((char*)arg, (char*)buf, 8)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* copy key descriptors */
	count = res_len / sizeof(struct persist_reserve);
	temp1 = descriptors;

	for(i = 0; i < count; i++) {
		memcpy(descriptors->key,
			res_buf->key_descriptor_list[i].reserve_key,
			RESERVE_KEY_LENGTH);
		descriptors->scope_spec_addr
			= (res_buf->key_descriptor_list[i].scope_addr[0] << 24)
			+ (res_buf->key_descriptor_list[i].scope_addr[1] << 16)
			+ (res_buf->key_descriptor_list[i].scope_addr[2] << 8)
			+ (res_buf->key_descriptor_list[i].scope_addr[3]
				& 0xFF);
		descriptors->scope = res_buf->key_descriptor_list[i].scope;
		descriptors->type = res_buf->key_descriptor_list[i].type;
		descriptors->ext_length
			= (res_buf->key_descriptor_list[i].extra_length[0] << 8)
			+ (res_buf->key_descriptor_list[i].extra_length[1]
				& 0xFF);

		if(copy_to_user(buf->reserve_list, descriptors,
			sizeof(struct reserve_descriptor))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		descriptors++;
		buf->reserve_list++;
	} /* for */
	descriptors = temp1;

EXIT_LABEL:

	if(buf) kfree(buf);
	if(res_len <= KmallocThreshhold) {
		if(res_buf) kfree(res_buf);
		if(descriptors) kfree(descriptors);
	} else {
		if(res_buf) free_pages((ulong)res_buf, order);
		if(descriptors) free_pages((ulong)descriptors, order);
	} /* else */

	*drvp = drv;
	return rc;
} /* stioc_read_reservations() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_register_key                                                    *
*                                                                              *
*  PURPOSE: Register key with drive to be used in persistent reserve ops       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_register_key(struct tape_descriptor** drvp)
{
	int rc = 0;
	unsigned int pr_generation = 0;
	char* rkey = NULL;
	struct tape_descriptor* drv = *drvp;

	rkey = kmalloc(RESERVE_KEY_LENGTH, GFP_KERNEL);
	if(!rkey) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	if(!TP_RSV_PERSISTENT(drv)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */
	get_random_bytes(rkey, RESERVE_KEY_LENGTH);
	rc = tape_register_reservation_key(&drv, rkey);
	if(rc == STATUS_SUCCESS) {
		/* update pr_generation field */
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
	} /* if */

EXIT_LABEL:

	if(rkey) kfree(rkey);
	*drvp = drv;
	return rc;
} /* stioc_register_key() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_remove_registration                                             *
*                                                                              *
*  PURPOSE: Remove persistent reservation from a drive                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_remove_registration(struct tape_descriptor** drvp)
{
	int rc = 0;
	unsigned int pr_generation = 0;
	struct tape_descriptor* drv = *drvp;

	if(drv->curr_reserve_type != LinTapePersistentReserve) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	/* register key 0 for the current initiator */
	rc = tape_register_reservation_key(&drv, NULL);
	if(rc == STATUS_SUCCESS) {
		drv->curr_reserve_type = LinTapeNoReserve;

		/* update pr_generation field */
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* stioc_remove_registration() */

/*******************************************************************************
*                                                                              *
*  NAME: stioc_clear_all_registrations                                         *
*                                                                              *
*  PURPOSE: Handle IOCTL request to clear persistent registrations from drive  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int stioc_clear_all_registrations(struct tape_descriptor** drvp)
{
	int rc = -EIO;
	unsigned int pr_generation = 0;
	struct tape_descriptor* drv = *drvp;

	if(drv->trace)
		tape_log_msg(drv->tape_log, &drv->tape_log_len,
			TRC_STIOC_CLEAR_REGS);

	if(drv->curr_reserve_type == LinTapeReserve6) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */
	rc = tape_clear_all_registrations(&drv);
	if(rc == STATUS_SUCCESS) {
		drv->curr_reserve_type = LinTapeNoReserve;

		/* update pr_generation field */
		rc = tape_read_reservation(&drv, NULL, &pr_generation, NULL);
		if(rc == STATUS_SUCCESS) drv->pr_generation = pr_generation;
	} /* if */

EXIT_LABEL:
	*drvp = drv;
	return rc;
} /* stioc_clear_all_registrations() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_changer_ioctl                                                *
*                                                                              *
*  PURPOSE: perform changer IOCTL operations                                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor associated with the request                      *
*    filp: the file pointer for the device                                     *
*    command: the type of request                                              *
*    arg: the generic data area used to perform IO                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_changer_ioctl(struct changer_descriptor **chgp,
	struct file *filp, uint command, ulong arg)
{
	int rc = -EIO;
	unchar ttype = lin_tape_trace_type;
	char call_description[40];
    boolean gotten = False;
	memset(call_description, '\0', 40);

	DbgPrint(("lin_tape: lin_tape_changer_ioctl entry: cmd 0x%x\n",
		command));

	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	if(!(*chgp)->dev_obj || (*chgp)->ghost_device || !scsi_block_when_processing_errors((*chgp)->dev_obj)) {
        if((*chgp)->alt_pathing) {
			WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
			rc = changer_failover_ops(&(*chgp), HZ * 1800, 0);
			WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
		} /* if */
		if(!(*chgp)->alt_pathing || rc != STATUS_SUCCESS) {
			printk("lin_tape: lin_tape_changer_ioctl: changer "\
				"removed.\n");
			WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

    switch(command) {
	case SIOC_INQUIRY:
		strcpy(call_description, "SIOC_INQUIRY");
		break;

	case SIOC_INQUIRY_PAGE:
		strcpy(call_description, "SIOC_INQUIRY_PAGE");
		break;

	case SIOC_REQSENSE:
		strcpy(call_description, "SIOC_REQSENSE");
		break;

	case SIOC_RESERVE: case SIOC_RELEASE:
		strcpy(call_description, "SIOC_RESERVE");
		break;

	case SIOC_TEST_UNIT_READY: case SCSI_IOCTL_TEST_UNIT_READY:
		strcpy(call_description, "SIOC_TEST_UNIT_READY");
		break;

	case SIOC_LOG_SENSE_PAGE:
		strcpy(call_description, "SIOC_LOG_SENSE_PAGE");
		break;
	case SIOC_LOG_SENSE10_PAGE:
		strcpy(call_description, "SIOC_LOG_SENSE10_PAGE");
		break;

	case SIOC_MODE_SENSE_PAGE:
		strcpy(call_description, "SIOC_MODE_SENSE_PAGE");
		break;

	case SIOC_QUERY_PATH:
		strcpy(call_description, "SIOC_QUERY_PATH");
		break;

	case SIOC_DEVICE_PATHS:
		strcpy(call_description, "SIOC_DEVICE_PATHS");
		break;

	case SIOC_ENABLE_PATH:
		strcpy(call_description, "SIOC_ENABLE_PATH");
		break;

	case SIOC_DISABLE_PATH:
		strcpy(call_description, "SIOC_DISABLE_PATH");
		break;

	case READ_ELEMENT_INFO: case SMCIOC_ELEMENT_INFO:
		strcpy(call_description, "READ_ELEMENT_INFO");
		break;

	case SMCIOC_POS_TO_ELEM:
		strcpy(call_description, "SMCIOC_POS_TO_ELEM");
		break;

	case SMCIOC_INIT_ELEM_STAT_RANGE:
		strcpy(call_description, "SMCIOC_INIT_ELEM_STAT_RANGE");
		break;

	case READ_ELEMENT_STATUS: case SMCIOC_INVENTORY:
		strcpy(call_description, "READ_ELEMENT_STATUS");
		break;

	case SMCIOC_READ_CARTRIDGE_LOCATION:
		strcpy(call_description, "SMCIOC_READ_CARTRIDGE_LOCATION");
		break;

	case SMCIOC_MOVE_MEDIUM:
		strcpy(call_description, "SMCIOC_MOVE_MEDIUM");
		break;

	case SMCIOC_INIT_ELEM_STAT:
		strcpy(call_description, "SMCIOC_INIT_ELEM_STAT");
		break;

	case SMCIOC_PREVENT_MEDIUM_REMOVAL: case SMCIOC_ALLOW_MEDIUM_REMOVAL:
		strcpy(call_description, "SMCIOC_PREVENT_MEDIUM_REMOVAL");
		break;

	case SMCIOC_EXCHANGE_MEDIUM:
		strcpy(call_description, "SMCIOC_EXCHANGE_MEDIUM");
		break;

	case SMCIOC_READ_ELEMENT_DEVIDS:
		strcpy(call_description, "SMCIOC_READ_ELEMENT_DEVIDS");
		break;

	case SMCIOC_LOAD_MEDIUM: case SMCIOC_UNLOAD_MEDIUM:
		strcpy(call_description, "SMCIOC_LOAD_MEDIUM");
		break;

	case SCSI_IOCTL_SEND_COMMAND:
		strcpy(call_description, "SCSI_IOCTL_SEND_COMMAND");
		break;

	case QUERY_DRIVER_VERSION:
		strcpy(call_description, "QUERY_DRIVER_VERSION");
		break;

	case SCSI_PASS_THROUGH:
		strcpy(call_description, "SCSI_PASS_THROUGH");
		break;

	case SIOC_PASS_THROUGH:
		strcpy(call_description, "SIOC_PASS_THROUGH");
		break;

	case DDIOC_INQUIRY:
		strcpy(call_description, "DDIOC_INQUIRY");
		break;

	case DDIOC_RETURN_ERR:
		strcpy(call_description, "DDIOC_RETURN_ERR");
		break;

	case LBIOC_RETURN_ELEMENT_COUNT:
		strcpy(call_description, "LBIOC_RETURN_ELEMENT_COUNT");
		break;

	case LBIOC_LIBRARY_INVENTORY:
		strcpy(call_description, "LBIOC_LIBRARY_INVENTORY");
		break;

	case LBIOC_MOVE_MEDIUM:
		strcpy(call_description, "LBIOC_MOVE_MEDIUM");
		break;

	case LBIOC_LIBRARY_AUDIT:
		strcpy(call_description, "LBIOC_LIBRARY_AUDIT");
		break;

	case LBIOC_EXTEND_RETRACT_IO_PT:
		strcpy(call_description, "LBIOC_EXTEND_RETRACT_IO_PT");
		break;

	case LBIOC_POSITION:
		strcpy(call_description, "LBIOC_POSITION");
		break;

	case LBIOC_LIBRARY_INFO:
		strcpy(call_description, "LBIOC_LIBRARY_INFO");
		break;

	case DDIOC_XINQUIRY:
		strcpy(call_description, "DDIOC_XINQUIRY");
		break;

	case LBIOC_XDRIVE_INVENTORY:
		strcpy(call_description, "LBIOC_XDRIVE_INVENTORY");
		break;

	case DDIOC_LOG_SENSE:
		strcpy(call_description, "DDIOC_LOG_SENSE");
		break;

	case LBIOC_ENH_LIBRARY_INVENTORY:
		strcpy(call_description, "LBIOC_ENH_LIBRARY_INVENTORY");
		break;

	default:
		strcpy(call_description, "default");
		break;
	} /* switch */
    if ((*chgp)->busy)
        DbgPrint(("lin_tape_changer_ioctl: IBMchanger%d busy %d\n",
                  (*chgp)->dev_index,(*chgp)->busy));
	WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	rc = changer_wait_to_set_busy(&(*chgp),call_description,0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
    gotten = True;
    (*chgp)->busy |= F_IOCTL;
    WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

	if(!(*chgp)->cdb || !(*chgp)->sense_data) {
		rc = chg_alloc_cdb_and_sense((*chgp));
		if(rc != STATUS_SUCCESS) {
			DbgPrint(("lin_tape_changer_ioctl: cannot allocate "\
				"buffer\n"));
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_SCSI_PATH,
			NULL, 0);

	/* every non-sense command should reset the last errno and compcode */
	if(command != SIOC_REQSENSE && command != DDIOC_RETURN_ERR) {
		(*chgp)->last_errno = STATUS_SUCCESS;
		(*chgp)->comp_code_valid = False;
		(*chgp)->comp_code = STATUS_SUCCESS;
	} /* if */

	switch(command) {
	case SIOC_INQUIRY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct inquiry_data), TRC_SIOC_INQUIRY,
			False, True, False, True, chg_sioc_inquiry);
		break;

	case SIOC_INQUIRY_PAGE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct inquiry_page), TRC_SIOC_INQUIRY_PG,
			False, True, True, True, chg_sioc_inquiry_page);
		break;

	case SIOC_REQSENSE:
		rc = chg_sioc_reqsense(&(*chgp), command, arg);
		break;

	case SIOC_RESERVE: case SIOC_RELEASE:
		rc = chg_sioc_reserve_release(&(*chgp), command == SIOC_RESERVE);
		break;

	case SIOC_TEST_UNIT_READY: case SCSI_IOCTL_TEST_UNIT_READY:
		log_changer((*chgp), TRC_SIOC_CHG_TUR, True);
		rc = changer_tur(&(*chgp));
		break;

	case SIOC_LOG_SENSE_PAGE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct log_sense_page), TRC_SIOC_LG_SNS_PG,
			False, True, True, True, chg_sioc_log_sense_page);
		break;

	case SIOC_LOG_SENSE10_PAGE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct log_sense_page), TRC_SIOC_LG_SNS_PG,
			False, True, True, True, chg_sioc_log_sense10_page);
		break;

	case SIOC_MODE_SENSE_PAGE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct mode_sense_page), TRC_SIOC_MD_SNS_PG,
			False, True, True, True, chg_sioc_mode_sense_page);
		break;

	case SIOC_QUERY_PATH:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct scsi_path), TRC_QRY_PATH,
			False, True, False, True, changer_query_path);
		break;

	case SIOC_DEVICE_PATHS:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct device_paths), TRC_DVC_PATHS,
			False, True, False, True, changer_show_paths);
		break;

	case SIOC_ENABLE_PATH:
		rc = lin_tape_csioc_enable_path((*chgp), arg);
		break;

	case SIOC_DISABLE_PATH:
		rc = lin_tape_csioc_disable_path((*chgp), arg);
		break;

	case READ_ELEMENT_INFO: case SMCIOC_ELEMENT_INFO:
		rc = smcioc_element_info(&(*chgp), command, arg);
		break;

	case SMCIOC_POS_TO_ELEM:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct pos_to_elem), TRC_SMCIOC_POS,
			False, True, True, False, smcioc_pos_to_elem);
		break;

	case SMCIOC_INIT_ELEM_STAT_RANGE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct element_range), TRC_SMCIOC_IES_RANGE,
			False, True, True, False, smcioc_init_elem_stat_range);
		break;

	case READ_ELEMENT_STATUS: case SMCIOC_INVENTORY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct inventory), TRC_SMCIOC_RES,
			False, True, True, False, smcioc_inventory);
		break;

	case SMCIOC_READ_CARTRIDGE_LOCATION:
		rc = changer_read_cartridge_location((*chgp), (char*)arg, FALSE);
		break;

	case SMCIOC_MOVE_MEDIUM:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct move_medium), TRC_SMCIOC_MVMED,
			False, True, True, False, smcioc_move_medium);
		break;

	case SMCIOC_INIT_ELEM_STAT:
		rc = changer_init_element(&(*chgp));
		break;

	case SMCIOC_PREVENT_MEDIUM_REMOVAL: case SMCIOC_ALLOW_MEDIUM_REMOVAL:
		rc = changer_prevent_allow_remove(&(*chgp),
			command == SMCIOC_PREVENT_MEDIUM_REMOVAL ? 1 : 0);
		break;

	case SMCIOC_EXCHANGE_MEDIUM:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct exchange_medium), TRC_SMCIOC_EXMED,
			False, True, True, False, smcioc_exchange_medium);
		break;

	case SMCIOC_READ_ELEMENT_DEVIDS:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct read_element_devids), TRC_SMCIOC_DEVIDS,
			False, True, True, False, smcioc_read_element_devids);
		break;

	case SMCIOC_LOAD_MEDIUM: case SMCIOC_UNLOAD_MEDIUM:
		rc = smcioc_load_unload_medium(&(*chgp), command, arg,
			command == SMCIOC_LOAD_MEDIUM);
		break;

	case SCSI_IOCTL_SEND_COMMAND:
		log_changer((*chgp), TRC_SMCIOC_CMD, True);
		rc = scsi_ioctl((*chgp)->dev_obj, command, (void*)arg);
		break;

	case QUERY_DRIVER_VERSION:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(struct QueryDriverVersion), TRC_QVER_CMD, False,
			True, False, True,
			(chg_method)lin_tape_query_driver_version);
		break;

	case SCSI_PASS_THROUGH:
		rc = lin_tape_cscsi_pass_through(&(*chgp), command, arg);
		break;

	case SIOC_PASS_THROUGH:
		rc = lin_tape_csioc_pass_through(&(*chgp), command, arg);
		break;

	case DDIOC_INQUIRY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(dd_old_inquiry_t), TRC_DINQ_ENT, False,
			False, False, True, changer_ddioc_inquiry);
		break;

	case DDIOC_RETURN_ERR:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(dd_return_error_state_t), TRC_RERR_ENT, False,
			False, False, True, changer_ddioc_ret_err);
		break;

	case LBIOC_RETURN_ELEMENT_COUNT:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_element_count_t), TRC_RELC_ENT, False,
			False, False, True, changer_lbioc_ret_elem_cnt);
		break;

	case LBIOC_LIBRARY_INVENTORY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_inventory_t), TRC_LINV_ENT, False,
			False, True, True, changer_lbioc_lib_inventory);
		break;

	case LBIOC_MOVE_MEDIUM:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_move_medium_t), TRC_LMVM_ENT, False,
			False, True, True, changer_lbioc_move_medium);
		break;

	case LBIOC_LIBRARY_AUDIT:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_audit_t), TRC_LAUD_ENT, False,
			False, True, True, changer_lbioc_audit);
		break;

	case LBIOC_EXTEND_RETRACT_IO_PT:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_extend_retract_t), TRC_LERI_ENT, False,
			False, False, True, changer_lbioc_ext_ret_ioport);
		break;

	case LBIOC_POSITION:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_position_t), TRC_LPOS_ENT, False,
			False, True, True, changer_lbioc_position);
		break;

	case LBIOC_LIBRARY_INFO:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_library_info_t), TRC_LINF_ENT, False,
			False, False, True, changer_lbioc_lib_info);
		break;

	case DDIOC_XINQUIRY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(dd_inquiry_t), TRC_XINQ_ENT, False,
			False, False, True, changer_ddioc_xinquiry);
		break;

	case LBIOC_XDRIVE_INVENTORY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_Xdrive_inventory_t), TRC_XINV_ENT, False,
			False, True, True, changer_lbioc_xdrive_inventory);
		break;

	case DDIOC_LOG_SENSE:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(dd_log_sense_page_t), TRC_LSNS_ENT, False,
			False, True, True, changer_lbioc_logsensepage);
		break;

	case LBIOC_ENH_LIBRARY_INVENTORY:
		rc = chgioc_exe(&(*chgp), command, arg,
			sizeof(lb_enh_inventory_t), TRC_EIVT_ENT, False,
			False, True, True, changer_lbioc_lib_enh_inventory);
		break;

	default:
		log_changer((*chgp), TRC_SMCIOC_INV, True);
		rc = -ENOSYS;
		break;
	} /* switch */

	if(rc != STATUS_SUCCESS) {
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg2((*chgp)->chg_log, &((*chgp)->chg_log_len),
				TRC_SMCIOC_ERR, -rc);
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE) changer_log_rc((*chgp), -rc);

	/* if it is a non-sense command, update the LastErrno */
	if(command != SIOC_REQSENSE && command != DDIOC_RETURN_ERR)
		(*chgp)->last_errno = -rc;

    WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	(*chgp)->busy &= (0xFF-F_IOCTL);
	WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

EXIT_LABEL:

	DbgPrint(("lin_tape: lin_tape_changer_ioctl exit: %d\n", rc));
	return rc;
} /* lin_tape_changer_ioctl() */

/*******************************************************************************
*                                                                              *
*  NAME: changer_wait_to_set_busy                                             *
*                                                                              *
*  PURPOSE: enter wait loop until changer becomes free to perform operation    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgr: changer descriptor associated with the request                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int changer_wait_to_set_busy(struct changer_descriptor** chgp,
                              char *instruction, unchar flag)
{
	int rc = STATUS_SUCCESS, bsy_ct = -1, wait_time = 1800;
	int idx = (*chgp)->dev_index;
	wait_queue_head_t wq;
    init_waitqueue_head(&wq);

    WRITE_LOCK(&changer_devs[idx]->entry_lock);
	while(changer_busy((*chgp),flag) && bsy_ct < wait_time){
        WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
		bsy_ct++;
		printk("lin_tape: IBMchanger%d busy %d: ct %d, (if < %d, wait)\n",
			(*chgp)->dev_index, (*chgp)->busy,bsy_ct, wait_time);
		rc = wait_event_interruptible_timeout(wq, !*chgp || !(*chgp)->busy, HZ);
        WRITE_LOCK(&changer_devs[idx]->entry_lock);
		if(!chgp || !(*chgp) || !changer_devs[idx] ||
            !changer_devs[idx]->chg || (*chgp)->remove_pending) {
			printk("lin_tape_changer_ioctl: changer "\
				"removed while waiting\n");
			rc = -ENXIO;
            WRITE_UNLOCK(&changer_devs[idx]->entry_lock);
			goto EXIT_LABEL;
		} /* if */
	}
    WRITE_UNLOCK(&changer_devs[idx]->entry_lock);

	if(bsy_ct >= wait_time) {
		printk("lin_tape_changer_ioctl: changer busy for > %d secs.  "\
			"Aborting command\n", wait_time);
		rc = -EBUSY;
		goto EXIT_LABEL;
	} /* if */

	DbgPrint(("lin_tape_changer set busy holder is %s\n", instruction));
	rc = STATUS_SUCCESS;

EXIT_LABEL:

	return rc;
} /* changer_wait_and_set_busy() */

int changer_busy(struct changer_descriptor *chg, unchar flag){
    unchar busy = chg->busy;
    if (chg->dev_index != chg->cur_path) busy = 0;
    else{
        if (chg->busy & flag)
            busy = chg->busy-flag;
    }
    DbgPrint(("lin_tape: Changer %d busy: %d flag %d\n",chg->dev_index, busy, flag));
    return (int) busy;
}

/*******************************************************************************
*                                                                              *
*  NAME: log_changer                                                           *
*                                                                              *
*  PURPOSE: Post a log message of an IOCTL call                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which we want to perform the request                 *
*    trc_msg: message to post to daemon                                        *
*    rtrace: if True and regular trace is specified, attempt to log            *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void log_changer(struct changer_descriptor* chg, char trc_msg, boolean rtrace)
{
	char ttype = lin_tape_trace_type;
	if(ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(chg->chg_log, &chg->chg_log_len, trc_msg);
	else if(rtrace && ttype == LIN_TAPE_FULL_TRACE)
		changer_log_full_trace(chg, trc_msg, 0, NULL);
} /* log_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: chgioc_exe                                                            *
*                                                                              *
*  PURPOSE: Prepare and send a changer IOCTL request                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which we want to perform the request                *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*    buflen: expected size of command argument                                 *
*    trc_msg: message to post to daemon                                        *
*    cpy_method: to user or to kernel                                          *
*    rtrace: if true and regular trace is specified, attempt to log            *
*    fm: copy data from user space                                             *
*    to: copy data to user space                                               *
*    func: the function that will perform the main part of the command         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chgioc_exe(struct changer_descriptor** chgp, uint command, ulong arg,
	ssize_t buflen, char trc_msg, char cpy_method, boolean rtrace,
	boolean fm, boolean to, chg_method func)
{
	int rc = -EIO;
	void* buf = NULL;

	log_changer((*chgp), trc_msg, rtrace);

	if(_IOC_SIZE(command) != buflen || (char *)arg == NULL) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(buflen, GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', buflen);

	if(fm && copy_from_user((char*)buf, (char*)arg, buflen)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = func(&(*chgp), (void*)buf, cpy_method);

	if(to && copy_to_user((char*)arg, (char*)buf, buflen)
		&& rc == STATUS_SUCCESS) rc = -EFAULT;

EXIT_LABEL:

	if(buf) kfree(buf);
	return rc;
} /* chgioc_exe() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_inquiry                                                      *
*                                                                              *
*  PURPOSE: Obtain standard inquiry from changer                               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    buf: the inquiry buffer for input and output                              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t chg_sioc_inquiry(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	int rc = STATUS_SUCCESS;
	struct inquiry_data* inq_buf = (struct inquiry_data*)buf;
	struct inquiry* inq = NULL;

	inq = kmalloc(MAX_INQ_LEN, GFP_KERNEL);
	if(!inq) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(inq, '\0', MAX_INQ_LEN);

	rc = changer_std_inquiry(&(*chgp), (void*)inq);

	if(rc == STATUS_SUCCESS) {
		inq_buf->qual = inq->dev_type_qualifier;
		inq_buf->type = inq->dev_type;
		inq_buf->rm = inq->rm;
		inq_buf->mod = inq->dev_type_modifier;
		inq_buf->iso = inq->iso;
		inq_buf->ecma = inq->ecma;
		inq_buf->ansi = inq->ansi;
		inq_buf->aenc = inq->aenc;
		inq_buf->trmiop = inq->trmiop;
		inq_buf->rdf = inq->response_data_format;
		inq_buf->len = inq->additional_length;
		inq_buf->mchngr = inq->mchngr;
		inq_buf->reladr = inq->relative_addressing;
		inq_buf->wbus32 = inq->wide_32bit;
		inq_buf->wbus16 = inq->wide_16bit;
		inq_buf->sync = inq->sync;
		inq_buf->linked = inq->linked_cmds;
		inq_buf->cmdque = inq->cmd_queue;
		inq_buf->sftre = inq->soft_reset;
		memcpy(inq_buf->vid, inq->vendor_id, 8);
		memcpy(inq_buf->pid, inq->product_id, 16);
		memcpy(inq_buf->revision, inq->revision, 4);
		memcpy(inq_buf->vendor1, inq->vendor_specific, 20);
	} /* if */

EXIT_LABEL:

	if(inq) kfree(inq);
	return rc;
} /* chg_sioc_inquiry() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_inquiry_page                                                 *
*                                                                              *
*  PURPOSE: Issue an inquiry on the page specified by buf                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    buf: the inquiry buffer for input and output                              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t chg_sioc_inquiry_page(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct inquiry_page* page = (struct inquiry_page*)buf;
	memset(page->data, '\0', MAX_INQ_LEN);
	return changer_inquiry_page(chgp, page->page_code, page->data);
} /* chg_sioc_inquiry_page() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_reqsense                                                     *
*                                                                              *
*  PURPOSE: Obtain sense data from changer if necessary and copy to user space *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chg_sioc_reqsense(struct changer_descriptor** chgp, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct request_sense* sense_buf = NULL;
	struct tape_sense_data* buf = NULL;
	int len = 0, resid = 0;
	boolean buf_allocd = False;

	log_changer((*chgp), TRC_SIOC_REQSENSE, True);

	if(_IOC_SIZE(command) != sizeof(struct request_sense)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	sense_buf = kmalloc(sizeof(struct request_sense), GFP_KERNEL);
	if(!sense_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buf, '\0', sizeof(struct request_sense));

	if((*chgp)->sense_valid) {
		buf = (*chgp)->sense_data;
		(*chgp)->sense_valid = False;
		rc = STATUS_SUCCESS;
	} else {
		buf = kmalloc(sizeof(struct tape_sense_data), GFP_KERNEL);
		if(!buf) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(buf, '\0', sizeof(struct tape_sense_data));
		buf_allocd = True;

		rc = changer_request_sense(&(*chgp), &len, &resid, (void*)buf);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* else */

	sense_buf->valid = buf->valid;
	sense_buf->err_code = buf->err_code;
	sense_buf->segnum = buf->seg_num;
	sense_buf->fm = buf->fm;
	sense_buf->eom = buf->eom;
	sense_buf->ili = buf->ili;
	sense_buf->key = buf->sense_key;
	sense_buf->info = (buf->information[0] << 24)
		+ (buf->information[1] << 16)
		+ (buf->information[2] << 8)
		+ buf->information[3];
	sense_buf->addlen = buf->additional_sense_len;
	sense_buf->cmdinfo = (buf->cmd_spec[0] << 24)
		+ (buf->cmd_spec[1] << 16)
		+ (buf->cmd_spec[2] << 8)
		+ buf->cmd_spec[3];
	sense_buf->asc = buf->asc;
	sense_buf->ascq = buf->ascq;
	sense_buf->fru = buf->field_replaceable;
	sense_buf->sksv = buf->sksv;
	sense_buf->cd = buf->cd;
	sense_buf->bpv = buf->bpv;
	sense_buf->sim = buf->bit_ptr;
	sense_buf->field[0] = buf->sksv_field_ptr;
	sense_buf->field[1] = buf->sense_key_spec;
	sense_buf->vendor[0] = buf->vendor_specific1;
	sense_buf->vendor[1] = buf->vendor_specific2;
	sense_buf->vendor[2] = buf->vendor_specific3;
	sense_buf->vendor[3] = buf->vendor_specific4;
	sense_buf->vendor[4] = buf->vendor_specific5;
	sense_buf->vendor[5] = buf->vendor_specific6;
	memcpy(&sense_buf->vendor[6], buf->vendor_specific_sense,
		MAX_ADDITIONAL_SENSE - 6);

	if(rc == STATUS_SUCCESS && copy_to_user((char*)arg, (char*)sense_buf,
		sizeof(struct request_sense))) rc = -EFAULT;

EXIT_LABEL:

	if(sense_buf) kfree(sense_buf);
	if(buf_allocd) kfree(buf);
	return rc;
} /* chg_sioc_reqsense() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_reserve_release                                              *
*                                                                              *
*  PURPOSE: Handle reserve and release operations at a high level              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    buf: the log_page buffer for input and output                             *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int chg_sioc_reserve_release(struct changer_descriptor** chgp,
	boolean reserve)
{
	int rc = STATUS_SUCCESS;

	log_changer(*chgp, reserve ? TRC_SIOC_RESERVE : TRC_SIOC_RELEASE, True);

	rc = reserve ? changer_reserve(chgp) : changer_release(chgp);
	if(rc == STATUS_SUCCESS) {
		(*chgp)->curr_reserve_type = reserve ? LinTapeReserve6
			: LinTapeNoReserve;
		(*chgp)->explicit_reserved = reserve ? True : False;
	} /* if */

	return rc;
} /* chg_sioc_reserve_release() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_log_sense_page                                               *
*                                                                              *
*  PURPOSE: Issue a log sense on the page specified by buf                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    buf: the log_page buffer for input and output                             *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t chg_sioc_log_sense_page(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	int rc = STATUS_SUCCESS;
	uint length = 0;
	struct log_sense_page* page = (struct log_sense_page*)buf;

	memset(page->data, '\0', LOGSENSEPAGE);
	rc = changer_logsense_page(chgp, True, page->page_code,
		page->data, &length);
	if(rc == STATUS_SUCCESS) page->len = (ushort)length;

	return rc;
} /* chg_sioc_log_sense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_log_sense10_page                                             *
*                                                                              *
*  PURPOSE: Handle SIOC_LOG_SENSE10_PAGE IOCTL (supporting subpages)           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: changer descriptor associated with the request                      *
*    buf: user log_sense10_page structure into which to copy data              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t chg_sioc_log_sense10_page(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct log_sense10_page* page = (struct log_sense10_page*)buf;
	uint length = (uint) page->len;
	int rc = STATUS_SUCCESS;

	if (!page->data) return -EINVAL;
	memset(page->data, '\0', length);

    rc=changer_logsense10_page(chgp, page->page_code, page->subpage_code,
			            (char *)page->data, page->parm_pointer, &length);
	if(rc == STATUS_SUCCESS) page->len = (ushort)length;

	return rc;
} /* chg_sioc_log_sense10_page() */

/*******************************************************************************
*                                                                              *
*  NAME: chg_sioc_mode_sense_page                                              *
*                                                                              *
*  PURPOSE: Issue a mode sense on the page specified by buf                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: device object used to call the IOCTL                                *
*    buf: the mode_page buffer for input and output                            *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t chg_sioc_mode_sense_page(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct mode_sense_page* page = (struct mode_sense_page*)buf;

	memset(page->data, '\0', MAX_MDSNS_LEN);
	return changer_modesense_page(chgp, page->page_code,
		page->data, False);
} /* chg_sioc_mode_sense_page() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_csioc_enable_path                                            *
*                                                                              *
*  PURPOSE: Enable changer path 'arg'                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: device object used to call the IOCTL                                 *
*    arg: an integer representing the number of the path to enable             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_csioc_enable_path(struct changer_descriptor* chg, ulong arg)
{
	int i = 0, count = 0, rc = -EINVAL;
	struct changer_descriptor* chgi = NULL;

	log_changer(chg, TRC_ENABLE_IOCTL, True);
	if(arg < 1) goto EXIT_LABEL;

	WRITE_LOCK(&changer_des_lock);

	/* find the path */
	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg
        || !changer_devs[i]->chg->serial_num_len
		|| changer_devs[i]->chg->serial_num_len != chg->serial_num_len
		|| memcmp(changer_devs[i]->chg->serial_num, chg->serial_num,
			chg->serial_num_len)) continue;

		count++;
		if(count == (int)arg) break;
	} /* for */

	WRITE_UNLOCK(&changer_des_lock);

	if(i >= MaxLinTapeChangerDevices) goto EXIT_LABEL;

	/* This is the path.  Have daemon mark device as enabled */
	chgi = changer_devs[i]->chg;
	rc = changer_setup_and_log_full(chgi, chgi->dev_index, TRC_CHG_ENABLED,
		NULL, 0);

	chgi->fenced = False;
	chgi->enabled = True;

	rc = STATUS_SUCCESS;

EXIT_LABEL:

	return rc;
} /* lin_tape_csioc_enable_path() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_csioc_disable_path                                           *
*                                                                              *
*  PURPOSE: Disable changer path 'arg'                                         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: device object used to call the IOCTL                                 *
*    arg: an integer representing the number of the path to disable            *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_csioc_disable_path(struct changer_descriptor* chg, ulong arg)
{
	int rc = -EINVAL, i = 0, count = 0, already_disabled = 0;
	struct changer_descriptor* chgi = NULL;

	log_changer(chg, TRC_DISABLE_IOCTL, True);
	if((signed)arg < 1) goto EXIT_LABEL;

	WRITE_LOCK(&changer_des_lock);

	/* find the path */
	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg
        || !changer_devs[i]->chg->serial_num_len
		|| changer_devs[i]->chg->serial_num_len != chg->serial_num_len
		|| memcmp(changer_devs[i]->chg->serial_num, chg->serial_num,
			chg->serial_num_len)) continue;

		count++;
		if(count == (int)arg) break;
	} /* for */

	WRITE_UNLOCK(&changer_des_lock);

	if(i >= MaxLinTapeChangerDevices) goto EXIT_LABEL;

	/* have daemon mark device as disabled */
	chgi = changer_devs[i]->chg;
	rc = changer_setup_and_log_full(chgi, chgi->dev_index,
		TRC_CHG_DISABLED, NULL, 0);

	already_disabled = !chgi->enabled;
	chgi->enabled = False;

	rc = STATUS_SUCCESS;

	if(chg != chgi) goto EXIT_LABEL;

	/* the current path is being disabled */
	/*  check with the daemon for a better path */
	changer_get_best_path(&chg);
	if(chg == chgi) {
		/* didn't get a better path */
		rc = -EINVAL;
		if(already_disabled) goto EXIT_LABEL;

		chgi->enabled = True;

		rc = changer_setup_and_log_full(chgi, chgi->dev_index,
			TRC_CHG_ENABLED, NULL, 0);
	} /* if */

EXIT_LABEL:

	return rc;
} /* lin_tape_csioc_disable_path() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_element_info                                                   *
*                                                                              *
*  PURPOSE: Get the position and number of each type of element                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to get the element information                *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int smcioc_element_info(struct changer_descriptor** chgp, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS, elinf_len = sizeof(struct element_info);
	struct element_info *info_buf = NULL;
	struct elem_assignment *data = NULL;

	log_changer((*chgp), TRC_SMCIOC_READ_EL_INFO, True);

	if(_IOC_SIZE(command) != elinf_len) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	info_buf = kmalloc(elinf_len, GFP_KERNEL);
	if(!info_buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(info_buf, '\0', elinf_len);

	data = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(data, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(&(*chgp), ElementAddressPage, (void*)data,
		sizeof(struct elem_assignment));

	if(rc != STATUS_SUCCESS) {
		(*chgp)->last_errno = -rc;
		goto EXIT_LABEL;
	} /* if */

	info_buf->robot_addr = (data->medium_transport_elem_addr[0] << 8)
		+ data->medium_transport_elem_addr[1];
	info_buf->robots = (data->transport_elem_num[0] << 8)
		+ data->transport_elem_num[1];
	info_buf->slot_addr = (data->first_storage_addr[0] << 8)
		+ data->first_storage_addr[1];
	info_buf->slots = (data->storage_elem_num[0] << 8)
		+  data->storage_elem_num[1];
	info_buf->ie_addr = (data->first_ie_port_addr[0] << 8)
		+ data->first_ie_port_addr[1];
	info_buf->ie_stations = (data->ie_port_elem_num[0] << 8)
		+ data->ie_port_elem_num[1];
	info_buf->drive_addr = (data->first_data_transfer_addr[0] << 8)
		+ data->first_data_transfer_addr[1];
	info_buf->drives = (data->data_transfer_num[0] << 8)
		+ data->data_transfer_num[1];

	if(copy_to_user((char*)arg, (char*)info_buf, elinf_len)) rc = -EFAULT;

	(*chgp)->last_errno = -rc;

EXIT_LABEL:

	if(data) kfree(data);
	if(info_buf) kfree(info_buf);
	return rc;
} /* smcioc_element_info() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_pos_to_elem                                                    *
*                                                                              *
*  PURPOSE: Issue a position to element command                                *
*                                                                             *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to issue position                             *
*    buf: a pos_to_elem structure                                              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_pos_to_elem(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct pos_to_elem *elem_buf = buf;

	return changer_pos_element(chgp, elem_buf->robot,
		elem_buf->destination, elem_buf->invert);
} /* smcioc_pos_to_elem() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_init_elem_stat_range                                           *
*                                                                              *
*  PURPOSE: Issue an initialize element status with range command              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to issue IES                                  *
*    buf: an element_range structure                                           *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_init_elem_stat_range(struct changer_descriptor** chgp,
	void* buf, char unused)
{
	struct element_range *elem_buf = buf;

	return changer_init_elem_range(chgp, elem_buf->element_address,
		elem_buf->number_elements);
} /* smcioc_init_elem_stat_range() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_inventory                                                      *
*                                                                              *
*  PURPOSE: Check parameters and issue a read devids command                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to read the element IDs                       *
*    buf: an inventory structure                                               *
*    cpy_method: to user space or kernel space                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_inventory(struct changer_descriptor** chgp, void* buf,
	char cpy_method)
{
	struct inventory *elem_buf = buf;

	return changer_inventory(chgp, elem_buf->robot_status,
		elem_buf->slot_status, elem_buf->ie_status,
		elem_buf->drive_status, cpy_method);
} /* smcioc_inventory() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_move_medium                                                    *
*                                                                              *
*  PURPOSE: Issue move medium command                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to move the medium                            *
*    buf: a move_medium structure                                              *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_move_medium(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct move_medium *elem_buf = buf;

	return changer_move_medium(chgp, elem_buf->robot, elem_buf->source,
		elem_buf->destination, elem_buf->invert);
} /* smcioc_move_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_exchange_medium                                                *
*                                                                              *
*  PURPOSE: Issue exchange medium command                                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to exchange the medium                        *
*    buf: an exchange_medium structure                                         *
*    unused: so we can pass this as a function pointer                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_exchange_medium(struct changer_descriptor** chgp, void* buf,
	char unused)
{
	struct exchange_medium *elem_buf = buf;

	return changer_exchange_medium(chgp, elem_buf->robot, elem_buf->source,
		elem_buf->destination1, elem_buf->destination2,
		elem_buf->invert1, elem_buf->invert2);
} /* smcioc_exchange_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_read_element_devids                                            *
*                                                                              *
*  PURPOSE: Check parameters and issue a read devids command                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which to read the element IDs                       *
*    buf: a read_element_devids structure                                      *
*    cpy_method: to user space or kernel space                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t smcioc_read_element_devids(struct changer_descriptor** chgp, void* buf,
	char cpy_method)
{
	int rc = STATUS_SUCCESS;
	struct read_element_devids* elem_buf = buf;

	if(!elem_buf->number_elements || !elem_buf->drive_devid) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = changer_read_devid(chgp, elem_buf->element_address,
		elem_buf->number_elements, elem_buf->drive_devid, cpy_method);

EXIT_LABEL:
	return rc;
} /* smcioc_read_element_devids() */

/*******************************************************************************
*                                                                              *
*  NAME: smcioc_load_unload_medium                                             *
*                                                                              *
*  PURPOSE: Execute (un)load medium for smaller libraries                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer descriptor for which to move the medium                 *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*    load: True to load, False to unload                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int smcioc_load_unload_medium(struct changer_descriptor** chgp, uint command,
	ulong arg, boolean load)
{
	int rc = STATUS_SUCCESS, i = 0;
	struct elem_assignment* data = NULL;
	struct element_status* slot_elem = NULL;
	ushort robot_addr = 0, drive_addr = 0, src_tgt = (ushort)arg;
	uint slots = 0, drives = 0;
	ulong alloc_len = 0, order = 0;

	log_changer((*chgp), load ? TRC_SMCIOC_LOAD : TRC_SMCIOC_UNLOAD, True);
	if((*chgp)->chg_id != IBM_3590 && (*chgp)->chg_id != IBM_3581 &&
		(*chgp)->chg_id != BDT_3581 && (*chgp)->chg_id != IBM_3573 &&
		(*chgp)->chg_id != HP_MSLG3 && (*chgp)->chg_id != HP_1x8G2) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	data = kmalloc(sizeof(struct elem_assignment), GFP_KERNEL);
	if(!data) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(data, '\0', sizeof(struct elem_assignment));

	rc = changer_modesense_page(&(*chgp), ElementAddressPage, (void*)data,
		sizeof(struct elem_assignment));

	if(rc == STATUS_SUCCESS) {
		robot_addr = (data->medium_transport_elem_addr[0] << 8)
			+ data->medium_transport_elem_addr[1];
		drive_addr = (data->first_data_transfer_addr[0] << 8)
			+ data->first_data_transfer_addr[1];
		slots = (data->storage_elem_num[0] << 8)
			+ data->storage_elem_num[1];
		drives = (data->data_transfer_num[0] << 8)
			+ data->data_transfer_num[1];
	} /* if */

	if(rc) goto EXIT_LABEL;
	if(drives != 1) {
		rc = -ENOSYS;
		goto EXIT_LABEL;
	} /* if */

	if(!src_tgt) {
		/* pick any slot */
		if(slots * sizeof(struct element_status) < KmallocThreshhold) {
			alloc_len = slots * sizeof(struct element_status);
			slot_elem = kmalloc(alloc_len, GFP_KERNEL);
		} else {
			order = get_mem_ord(slots*sizeof(struct element_status),
				&alloc_len);
			slot_elem = (void*)__get_free_pages(GFP_KERNEL,order);
		} /* else */

		if(!slot_elem) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(slot_elem, '\0', alloc_len);

		rc = changer_inventory(&(*chgp), NULL, slot_elem, NULL, NULL, True);

		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

		for(i = 0; i < slots; i++) {
			if(load ? slot_elem[i].full : !slot_elem[i].full) {
				src_tgt = slot_elem[i].address;
				break;
			} /* if */
		} /* for */

		if(!src_tgt) {
			rc = -EIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	rc = changer_move_medium(&(*chgp), robot_addr, load ? src_tgt : drive_addr,
		load ? drive_addr : src_tgt, False);

EXIT_LABEL:

	if(data) kfree(data);

	if(slot_elem) {
		if(alloc_len < KmallocThreshhold) kfree(slot_elem);
		else free_pages((ulong)slot_elem, order);
	} /* if */

	return rc;
} /* smcioc_load_unload_medium() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_query_driver_version                                         *
*                                                                              *
*  PURPOSE: Copy version information to version structure                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    unused1: so we can pass this as a function pointer                        *
*    version: the version structure to which to copy                           *
*    unused2: so we can pass this as a function pointer                        *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

ssize_t lin_tape_query_driver_version(struct tape_descriptor** unused1,
	void* buf, char unused2)
{
	struct QueryDriverVersion* version = buf;
	sscanf(LIN_TAPE_VERSION, "%u.%u.%u", &version->MajorRelease,
		&version->MinorRelease, &version->IncrementRelease);
	return STATUS_SUCCESS;

} /* lin_tape_query_driver_version() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_cscsi_pass_through                                           *
*                                                                              *
*  PURPOSE: Prepare and send a changer passthrough IOCTL request               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which we want to perform the request                *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_cscsi_pass_through(struct changer_descriptor** chgp,
	uint command, ulong arg)
{

	PSCSIPassThrough buf = NULL;
	int buffer_length = 0, rc = 0;
	unchar sense_length = 0, sense_avail = 0;
	int resid = 0, result = 0;

	log_changer((*chgp), TRC_PTCMD, True);

	if(_IOC_SIZE(command) != sizeof(SCSIPassThrough)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	buf = kmalloc(sizeof(SCSIPassThrough), GFP_KERNEL);
	if(!buf) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(buf, '\0', sizeof(SCSIPassThrough));

	if(copy_from_user((char*)buf, (char*)arg, sizeof(SCSIPassThrough))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* for old style passthrough, buffer length is not updated */
	buffer_length = buf->BufferLength;
	sense_length = 0;

	rc = changer_pass_thru(&(*chgp), buf->CDB, buf->CommandLength,
		buf->Buffer, &buffer_length, buf->DataDirection, buf->TimeOut,
		NULL, &sense_length, &sense_avail, &resid, &result, 0, False);

	buf->TargetStatus = status_byte(result);
	buf->MessageStatus = msg_byte(result);
	buf->HostStatus = host_byte(result);
	buf->DriverStatus = driver_byte(result);

	if(sense_avail) {
		buf->SenseDataValid = True;
		buf->SenseKey = (*chgp)->sense_data->sense_key;
		buf->ASC = (*chgp)->sense_data->asc;
		buf->ASCQ = (*chgp)->sense_data->ascq;
	} else {
		buf->SenseDataValid = FALSE;
		buf->SenseKey = 0;
		buf->ASC = 0;
		buf->ASCQ = 0;
	} /* else */

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	if(copy_to_user((char *) arg, (char *) buf, sizeof(SCSIPassThrough)))
		rc = -EFAULT;

EXIT_LABEL:
	if(buf) kfree(buf);
	return rc;
} /* lin_tape_cscsi_pass_through() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_csioc_pass_through                                           *
*                                                                              *
*  PURPOSE: Prepare and send a changer passthrough IOCTL request               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chgp: the changer for which we want to perform the request                *
*    command: the type of request                                              *
*    arg: the argument passed into lin_tape_changer_ioctl                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_csioc_pass_through(struct changer_descriptor** chgp, uint command,
	ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct sioc_pass_through *pass_through = NULL;
	unchar* cdb = NULL;
	unchar sense_avail = 0;
	int ioc_len = sizeof(struct sioc_pass_through);

	log_changer((*chgp), TRC_PTCMD, True);

	if(_IOC_SIZE(command) != ioc_len) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	pass_through = kmalloc(ioc_len, GFP_KERNEL);
	if(!pass_through) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(pass_through, '\0', ioc_len);

	if(copy_from_user((char*)pass_through, (char*)arg, ioc_len)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	cdb = kmalloc(pass_through->cmd_length, GFP_KERNEL);
	if(!cdb) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	if(copy_from_user((char*)cdb, (char*)pass_through->cdb,
		pass_through->cmd_length)) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = changer_pass_thru(&(*chgp), cdb, pass_through->cmd_length,
		pass_through->buffer, &pass_through->buffer_length,
		pass_through->data_direction, pass_through->timeout,
		pass_through->sense, &pass_through->sense_length,
		&sense_avail, &pass_through->resid, &pass_through->result, 0, False);

	pass_through->target_status = status_byte(pass_through->result);
	pass_through->msg_status = msg_byte(pass_through->result);
	pass_through->host_status = host_byte(pass_through->result);
	pass_through->driver_status = driver_byte(pass_through->result);

	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	if(copy_to_user((char*)arg, (char*)pass_through, ioc_len)) rc = -EFAULT;

EXIT_LABEL:
	if(pass_through) kfree(pass_through);
	if(cdb) kfree(cdb);
	return rc;
} /* lin_tape_csioc_pass_through() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_daemon_ioctl                                                 *
*                                                                              *
*  PURPOSE: Perform a request on behalf of the lin_taped daemon                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the type of request                                              *
*    arg: the command-specific usermode argument                               *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_daemon_ioctl(uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;

    DbgPrint3(("lin_tape_ioctl command 0x%x\n",(int)command));
	switch (command) {

	case LIN_TAPE_SET_TRACE_TYPE:
		rc = lin_tape_set_trace_type(command, arg);
		break;

	case LIN_TAPE_GET_DRIVE_NUM:
		rc = lin_tape_get_drive_num(arg);
		break;

	case LIN_TAPE_GET_CHANGER_NUM:
		rc = lin_tape_get_changer_num(arg);
		break;

	case LIN_TAPE_GET_FO_PARMS:
		rc = lin_tape_get_fo_parms(arg);
		break;

	case LIN_TAPE_SET_FO:
		lin_tape_set_fo(arg);
		break;

	case LIN_TAPE_SET_SFMP:
		lin_tape_set_sfmp(arg);
		break;

	case LIN_TAPE_GET_PETRO:
		rc = copy_to_user((char*)arg, (char*)&(petro), sizeof(int)) ?
			-EFAULT : STATUS_SUCCESS;
		break;

	case LIN_TAPE_SET_EA:
		if((int)arg >= MaxLinTapeDevices ||
            !tape_devs[(int)arg] ||
            !tape_devs[(int)arg]->drv) break;
		tape_devs[(int)arg]->drv->encryption_proxy_entitled = True;
		break;

	case LIN_TAPE_POLL_TRACE:
		rc = lin_tape_poll_trace(command, arg);
		break;

	case RETRIEVE_DRIVE_DUMP:
		rc = lin_tape_retrieve_drive_dump(command, arg);
		break;

	case RETRIEVE_LOG_DATA:
		rc = lin_tape_retrieve_log_data(command, arg);
		break;

	case RETRIEVE_SIMMIM_DATA:
		rc = lin_tape_retrieve_simmim_data(command, arg);
		break;

	case LIN_TAPE_WAKE_UP:
		rc = lin_tape_wake_up(arg);
		break;

	case LIN_TAPE_SEND_COMMAND:
		rc = lin_tape_send_command(arg);
		break;

	case LIN_TAPE_HANDLE_ERP:
		rc = lin_tape_handle_erp(arg);
		break;

	default:
		DbgPrint(("lin_tape: Invalid daemon ioctl.\n"));
		rc = -EINVAL;
		break;
	} /* switch */

	return rc;
} /* lin_tape_daemon_ioctl() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_trace_type                                               *
*                                                                              *
*  PURPOSE: Set the policy for writing trace messages to the log               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_set_trace_type(uint command, ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct tape_set_trace_type *ttype = NULL;

	if(_IOC_SIZE(command) != sizeof(struct tape_set_trace_type)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	ttype = kmalloc(sizeof(struct tape_set_trace_type), GFP_KERNEL);
	if(!ttype) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(ttype, '\0', sizeof(struct tape_set_trace_type));

	if(copy_from_user((char*)ttype, (char*)arg,
		sizeof(struct tape_set_trace_type))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	WRITE_LOCK(&lin_tape_trace_lock);
	lin_tape_trace_type = ttype->trace_type;
	WRITE_UNLOCK(&lin_tape_trace_lock);

EXIT_LABEL:

	if(ttype) kfree(ttype);
	return rc;
} /* lin_tape_set_trace_type() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_drive_num                                                *
*                                                                              *
*  PURPOSE: Retrieve number of drives currently attached                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_get_drive_num(ulong arg)
{
	uint num = 0;
	int i = 0, rc = 0;

	READ_LOCK(&tape_des_lock);
	if(tape_devs) {
		for(i = 0; i < MaxLinTapeDevices; i++) {
			if(tape_devs[i] && tape_devs[i]->drv) num++;
		} /* for */
	} /* if */
	READ_UNLOCK(&tape_des_lock);

	rc = copy_to_user((char*)arg, (char*)&num, sizeof(uint))
		? -EFAULT : STATUS_SUCCESS;
	return rc;
} /* lin_tape_get_drive_num() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_changer_num                                              *
*                                                                              *
*  PURPOSE: Retrieve number of changers currently attached                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_get_changer_num(ulong arg)
{
	uint num = 0;
	int i = 0, rc = 0;

	READ_LOCK(&changer_des_lock);
	if(changer_devs) {
		for(i = 0; i < MaxLinTapeChangerDevices; i++) {
			if(changer_devs[i] && changer_devs[i]->chg) num++;
		} /* for */
	} /* if */
	READ_UNLOCK(&changer_des_lock);

	rc = copy_to_user((char*)arg, (char*)&num, sizeof(uint))
		?  -EFAULT : STATUS_SUCCESS;

	return rc;
} /* lin_tape_get_changer_num() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_fo_parms                                                 *
*                                                                              *
*  PURPOSE: Retrieve current multipathing state                                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_get_fo_parms(ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct tape_get_fo_parms* parms = NULL;

	parms = kmalloc(sizeof(struct tape_get_fo_parms), GFP_KERNEL);
	if(!parms) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(parms, '\0', sizeof(struct tape_get_fo_parms));

	parms->alternate_pathing = alternate_pathing;
	if(dpf_keys) {
		parms->dpf_keys_avail = True;

		memcpy(parms->dpf_keys, dpf_keys, min((int)strlen(dpf_keys),
			MAX_DPF_KEYS * 17));
	} /* if */

	rc = copy_to_user((char*)arg, (char*)parms,
		sizeof(struct tape_get_fo_parms)) ?  -EFAULT : STATUS_SUCCESS;

EXIT_LABEL:

	if(parms)kfree(parms);
	return rc;
} /* lin_tape_get_fo_parms() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_fo                                                       *
*                                                                              *
*  PURPOSE: Set up multipathing for the device at index arg                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_fo(ulong arg)
{
	printk("LIN_TAPE_SET_FO entry, device minor number is %d.\n", (int)arg);

	if((int)arg < LinTapeChangerOffset) lin_tape_set_fo_drive(arg);
	else {
		  if (dualaccessor_cpf == 1){
		    DbgPrint(("lin_tape: CPF for dual accessor mode\n"));
		    lin_tape_set_fo_changer_by_wwnn(arg - LinTapeChangerOffset);
		  }else    lin_tape_set_fo_changer(arg - LinTapeChangerOffset);
	}

	return;
} /* lin_tape_set_fo() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_fo_drive                                                 *
*                                                                              *
*  PURPOSE: Set up multipathing for the drive at index arg                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_fo_drive(ulong arg)
{
	int i = (int)arg, j = 0;
	struct tape_descriptor* trav = NULL, * orig = NULL;

	WRITE_LOCK(&tape_des_lock);

	if(tape_devs[i] && tape_devs[i]->drv) tape_devs[i]->drv->alt_pathing = True;

	for(j = 0; j < MaxLinTapeDevices; j++) {
		if(!tape_devs[j] || !tape_devs[j]->drv
			|| i == j
			|| tape_devs[j]->drv->remove_pending
			|| !tape_devs[j]->drv->checked_multipath
			|| tape_devs[i]->drv->serial_num_len <= 0
			|| tape_devs[i]->drv->serial_num_len
				!= tape_devs[j]->drv->serial_num_len
			|| memcmp(tape_devs[i]->drv->serial_num,
				tape_devs[j]->drv->serial_num,
				tape_devs[i]->drv->serial_num_len)) continue;

		if(tape_devs[j]->drv->isPrimary) {
			DbgPrint(("lin_tape: IBMtape%d found an alternate "\
				"path IBMtape%d.\n", j, i));
			tape_devs[i]->drv->cur_path = tape_devs[j]->drv->cur_path;
			tape_devs[i]->drv->isPrimary = False;
		} /* if */

		trav = orig = tape_devs[j]->drv;
		while(True) {
			if(trav == tape_devs[i]->drv) break;
			if(trav->next_path == orig) {
				tape_devs[i]->drv->next_path = trav->next_path;
				trav->next_path = tape_devs[i]->drv;
				break;
			} /* if */

			trav = trav->next_path;
		} /* while */

		break;
	} /* for */

	if(tape_devs[i] && tape_devs[i]->drv) tape_devs[i]->drv->checked_multipath = True;

	WRITE_UNLOCK(&tape_des_lock);

	return;
} /* lin_tape_set_fo_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_fo_changer                                               *
*                                                                              *
*  PURPOSE: Set up multipathing for the changer at index arg                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_fo_changer(ulong arg)
{
	int i = (int)arg, j = 0;
	struct changer_descriptor* trav = NULL, * orig = NULL;

	WRITE_LOCK(&changer_des_lock);

	if(changer_devs[i] && changer_devs[i]->chg) changer_devs[i]->chg->alt_pathing = True;

	/* search changer_devs table to see if there is an S/N match */
	for(j = 0; j < MaxLinTapeChangerDevices; j++) {
		if(!changer_devs[j] || !changer_devs[j]->chg
			|| i == j
			|| changer_devs[j]->chg->remove_pending
			|| !changer_devs[j]->chg->checked_multipath
			|| changer_devs[i]->chg->serial_num_len <= 0
			|| changer_devs[i]->chg->serial_num_len
				!= changer_devs[j]->chg->serial_num_len
			|| memcmp(changer_devs[i]->chg->serial_num,
				changer_devs[j]->chg->serial_num,
				changer_devs[i]->chg->serial_num_len)) continue;

		if(changer_devs[j]->chg->isPrimary) {
			DbgPrint(("lin_tape: IBMchanger%d found an alternate "\
				"path IBMchanger%d.\n", j, i));
			changer_devs[i]->chg->cur_path = changer_devs[j]->chg->cur_path;
			changer_devs[i]->chg->isPrimary = False;
		} /* if */

		trav = orig = changer_devs[j]->chg;
		while(True) {
			if(trav == changer_devs[i]->chg) break;
			if(trav->next_path == orig) {
				changer_devs[i]->chg->next_path = trav->next_path;
				trav->next_path = changer_devs[i]->chg;
				changer_devs[i]->chg->open_count = orig->open_count;
				break;
			} /* if */

			trav = trav->next_path;
		} /* while */

		break;
	} /* for */

	if(changer_devs[i] && changer_devs[i]->chg) changer_devs[i]->chg->checked_multipath = True;

	WRITE_UNLOCK(&changer_des_lock);

	return;
} /* lin_tape_set_fo_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_fo_changer_by_wwnn                                       *
*                                                                              *
*  PURPOSE: Set up multipathing for the changer at index arg                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_fo_changer_by_wwnn(ulong arg)
{
	int i = (int)arg, j = 0;
	struct changer_descriptor* trav = NULL, * orig = NULL;

	WRITE_LOCK(&changer_des_lock);

	if(changer_devs[i]) changer_devs[i]->chg->alt_pathing = True;

	DbgPrint(("lin_tape: lin_tape_set_fo_changer_by_wwnn: \n"));
	DbgPrint(("lin_tape: src: changer_devs[%d]->wwnn=0x",i));
	if (changer_devs[i]){
	    int m = 0;
	    for(m = 0; m < WWN_LEN; m++) {
	      DbgPrint(("%02X", changer_devs[i]->chg->wwnn[m] & 0xFF));
	    } /* for */
	    DbgPrint(("\n"));
	}

	/* search changer_devs table to see if there is an S/N match */
	for(j = 0; j < MaxLinTapeChangerDevices; j++) {

	  if (changer_devs[j]){
	    int m = 0;
	    DbgPrint(("lin_tape: tgt: changer_devs[%d]->wwnn=0x",j));
	    for(m = 0; m < WWN_LEN; m++) {
	      DbgPrint(("%02X", changer_devs[j]->chg->wwnn[m] & 0xFF));
	    } /* for */
	    DbgPrint(("\n"));
	  }

		if(!changer_devs[j]
			|| i == j
			|| !changer_devs[j]->chg->checked_multipath
			|| memcmp(changer_devs[i]->chg->wwnn,
				changer_devs[j]->chg->wwnn,
				WWN_LEN)) continue;

		if(changer_devs[j]->chg->isPrimary) {
			DbgPrint(("lin_tape: IBMchanger%d found an alternate "\
				"path IBMchanger%d.\n", j, i));
			changer_devs[i]->chg->cur_path = changer_devs[j]->chg->cur_path;
			changer_devs[i]->chg->isPrimary = False;
		} /* if */

		trav = orig = changer_devs[j]->chg;
		while(True) {
			if(trav == changer_devs[i]->chg) break;
			if(trav->next_path == orig) {
				changer_devs[i]->chg->next_path = trav->next_path;
				trav->next_path = changer_devs[i]->chg;
				changer_devs[i]->chg->open_count = orig->open_count;
				break;
			} /* if */

			trav = trav->next_path;
		} /* while */

		break;
	} /* for */

	if(changer_devs[i]->chg) changer_devs[i]->chg->checked_multipath = True;

	WRITE_UNLOCK(&changer_des_lock);

	return;
} /* lin_tape_set_fo_changer_by_wwnn() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_set_sfmp                                                     *
*                                                                              *
*  PURPOSE: Set up sfmp for pfo to enable sg/st device file/paths              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void lin_tape_set_sfmp(ulong arg)
{
# define MAX_32 32
	int i, fo;
	unchar data[MAX_32];

	if (!lin_tape_as_sfmp ) return;

	i = (int)arg;

	memset(data, 0, MAX_32);

	if(i < LinTapeChangerOffset && i >= 0){
		if (!tape_devs || !tape_devs[i] || !tape_devs[i]->drv) {
			return;
		}
		data[0]='D';
		data[1]=WWN_LEN;
		READ_LOCK(&tape_des_lock);
		memcpy(data+2, tape_devs[i]->drv->wwnn, WWN_LEN);
		fo=tape_devs[i]->drv->alt_pathing;
		READ_UNLOCK(&tape_des_lock);
		pfo_failover_device(data, fo);
	}
	else{
		if (i<0 || i>(LinTapeChangerOffset+MaxLinTapeChangerDevices) || !changer_devs ||
            !changer_devs[i-LinTapeChangerOffset] || !changer_devs[i-LinTapeChangerOffset]->chg)
			return;

		data[0]='C';
		data[1]=changer_devs[i-LinTapeChangerOffset]->chg->serial_num_len;
		READ_LOCK(&changer_des_lock);
		memcpy(data+2, changer_devs[i-LinTapeChangerOffset]->chg->serial_num,
            changer_devs[i-LinTapeChangerOffset]->chg->serial_num_len);
		fo=changer_devs[i-LinTapeChangerOffset]->chg->alt_pathing;
		READ_UNLOCK(&changer_des_lock);
		pfo_failover_device(data, fo);
	}

	return;
} /* lin_tape_set_sfmp() */


/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_poll_trace                                                   *
*                                                                              *
*  PURPOSE: Copy any drive trace messages from kernel space to user space      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the command passed into lin_tape_daemon_ioctl                    *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_poll_trace(uint command, ulong arg)
{
	struct tape_poll* poll = NULL;
	struct tape_trace* trc = NULL;
	void* tptr = NULL;
	int j = 0, rc = STATUS_SUCCESS;
	char ttype = lin_tape_trace_type;

	if(_IOC_SIZE(command) != sizeof(struct tape_poll)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	poll = kmalloc(sizeof(struct tape_poll), GFP_KERNEL);
	if(poll) memset(poll, '\0', sizeof(struct tape_poll));
	else {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* else */

	trc = kmalloc(sizeof(struct tape_trace), GFP_KERNEL);
	if(trc) memset(trc, '\0', sizeof(struct tape_trace));
	else {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* else */

	if(copy_from_user((char *) poll, (char *) arg,
		sizeof(struct tape_poll))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!poll->log_trace) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	tptr = poll->log_trace;
	ttype = lin_tape_trace_type;

	/* if trace type is not invalid,
	* it traces message for driver itself
	*/
	if(ttype != LIN_TAPE_INVALID_TRACE) {
		if(lin_tape_log_len > 0 && lin_tape_log != NULL
			&& j < poll->dev_cnt) {
			sprintf(&trc->devid[0], "IBMtapeDevDriver");
			trc->dmp_len = 0;
			trc->simmim_len = 0;
			trc->trc_len = 0;
			trc->msg_len = lin_tape_log_len;
			memcpy(&trc->msg[0], lin_tape_log, trc->msg_len);
			lin_tape_log_len = 0;

			if(copy_to_user(tptr, trc, sizeof(struct tape_trace))) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			tptr += sizeof(struct tape_trace);
			j++;
		} /* if */
	} /* if */

	if(tape_devs) rc = lin_tape_poll_trace_drive(poll, trc, &tptr, &j);
	if(changer_devs) rc = lin_tape_poll_trace_changer(poll, trc, &tptr, &j);

	poll->dev_cnt = j;
	if(copy_to_user((char*)arg, (char*)poll, sizeof(struct tape_poll))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	rc = STATUS_SUCCESS;

EXIT_LABEL:
	if(poll) kfree(poll);
	if(trc) kfree(trc);
	return rc;
} /* lin_tape_poll_trace() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_poll_trace_drive                                             *
*                                                                              *
*  PURPOSE: Copy any drive trace messages from kernel space to user space      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    poll: contains maximum number of devices for which to obtain data         *
*    trc: the trace data itself                                                *
*    tptr: pointer into user space, trc will be copied here                    *
*    j: (output) the number of messages copied                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_poll_trace_drive(struct tape_poll* poll, struct tape_trace* trc,
	void** tptr, int* j)
{
	int rc = STATUS_SUCCESS, i = 0;
	struct tape_descriptor* drv = NULL;
	struct tape_full_trace_unit* unit = NULL;
	char ttype = lin_tape_trace_type;

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;

		drv = tape_devs[i]->drv;
		READ_LOCK(&drv->full_trc_lock);
		unit = drv->full_trc_head;
		READ_UNLOCK(&drv->full_trc_lock);

		if((!unit && (drv->tape_log_len <= 0 || !drv->tape_log))
			|| *j >= poll->dev_cnt) continue;

		memset(trc, '\0', sizeof(struct tape_trace));

		memcpy(&trc->devid[0], &drv->devid[0], MAX_VOLUME_LABEL_SIZE);

		trc->scsi_path[0] = drv->host;
		trc->scsi_path[1] = drv->channel;
		trc->scsi_path[2] = drv->target;
		trc->scsi_path[3] = drv->lun;

		/* check for dump data */
		READ_LOCK(&drv->dmp_lock);
		if(drv->dmp_len > 0 && (drv->dmp || drv->op_sg == DUMP_READY)){
			trc->dmp_len = drv->dmp_len;
			trc->op_sg = drv->op_sg;
		}
		else trc->dmp_len = 0;
		READ_UNLOCK(&drv->dmp_lock);

		/* check simmim data */
		READ_LOCK(&drv->simmim_lock);
		if(drv->simmim_len > 0 && drv->simmim)
			trc->simmim_len = drv->simmim_len;
		else trc->simmim_len = 0;
		READ_UNLOCK(&drv->simmim_lock);

		/* check full logsense pages */
		READ_LOCK(&drv->logsense_lock);
		if(drv->logsense_len > 0 && drv->logsense)
			trc->logsense_len = drv->logsense_len;
		else trc->logsense_len = 0;
		READ_UNLOCK(&drv->logsense_lock);

		/* check message for regular trace */
		if(ttype == LIN_TAPE_REGULAR_TRACE) {
			trc->msg_len = drv->tape_log_len;
			memcpy(&trc->msg[0], drv->tape_log, trc->msg_len);
			drv->tape_log_len = 0;
		} /* if */

		/* check full trace and error logging */
		if(!unit) {
			trc->trc_len = 0;
			memset(&trc->full_trc[0], '\0', FullTraceUnitSize);
		} else {
			WRITE_LOCK(&drv->full_trc_lock);

			adjust_queued_dpf(drv, unit);

			if(drv->full_trc_tail == drv->full_trc_head)
				drv->full_trc_tail = NULL;
			drv->full_trc_head = drv->full_trc_head->next;
			WRITE_UNLOCK(&drv->full_trc_lock);

			trc->trc_len = unit->cur_len;

			if(unit->cur_len > 0)
				memcpy(&trc->full_trc[0], unit->unit_data,
					min((int)(unit->cur_len),
					FullTraceUnitSize));
			kfree(unit);
			unit = NULL;
		} /* else */

		if(copy_to_user(*tptr, trc, sizeof(struct tape_trace))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
		(*tptr) += sizeof(struct tape_trace);
		(*j)++;

	} /* for */

EXIT_LABEL:

	return rc;

} /* lin_tape_poll_trace_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_poll_trace_changer                                           *
*                                                                              *
*  PURPOSE: Copy any changer trace messages from kernel space to user space    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    poll: contains maximum number of devices for which to obtain data         *
*    trc: the trace data itself                                                *
*    tptr: pointer into user space, trc will be copied here                    *
*    j: (output) the number of messages copied                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_poll_trace_changer(struct tape_poll* poll, struct tape_trace* trc,
	void** tptr, int* j)
{
	int rc = STATUS_SUCCESS, i = 0;
	struct changer_descriptor* chg = NULL;
	struct tape_full_trace_unit* unit = NULL;
	char ttype = lin_tape_trace_type;

	for(i = 0; i < MaxLinTapeChangerDevices; i++) {
		if(!changer_devs[i] || !changer_devs[i]->chg) continue;

		chg = changer_devs[i]->chg;
		READ_LOCK(&chg->full_trc_lock);
		unit = chg->full_trc_head;
		READ_UNLOCK(&chg->full_trc_lock);

		if((!unit && (chg->chg_log_len <= 0
			|| !chg->chg_log))
			|| *j >= poll->dev_cnt) continue;

		memcpy(&trc->devid[0], &chg->devid[0], MAX_VOLUME_LABEL_SIZE);

		trc->scsi_path[0] = chg->host;
		trc->scsi_path[1] = chg->channel;
		trc->scsi_path[2] = chg->target;
		trc->scsi_path[3] = chg->lun;

		trc->dmp_len = 0;
		trc->simmim_len = 0;
		trc->logsense_len = 0;

		/* check for regular trace data */
		if(ttype == LIN_TAPE_REGULAR_TRACE) {
			trc->msg_len = chg->chg_log_len;
			memcpy(&trc->msg[0], chg->chg_log, trc->msg_len);
			chg->chg_log_len = 0;
		} /* if */

		/* check for full trace and error logging data */
		if(!unit) {
			trc->trc_len = 0;
			memset(&trc->full_trc[0], '\0', FullTraceUnitSize);
		} else {
			WRITE_LOCK(&chg->full_trc_lock);
			adjust_queued_cpf(chg, unit);

			if(chg->full_trc_tail == chg->full_trc_head)
				chg->full_trc_tail=NULL;
			chg->full_trc_head = chg->full_trc_head->next;
			WRITE_UNLOCK(&chg->full_trc_lock);

			trc->trc_len = unit->cur_len;
			if(unit->cur_len > 0)
				memcpy(&trc->full_trc[0], unit->unit_data,
					min((int)(unit->cur_len),
						FullTraceUnitSize));
			kfree(unit);
			unit = NULL;
		} /* else */

		if(copy_to_user(*tptr, trc, sizeof(struct tape_trace))) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
		(*tptr) += sizeof(struct tape_trace);
		(*j)++;
	} /* for */

EXIT_LABEL:

	return rc;

} /* lin_tape_poll_trace_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: adjust_queued_dpf                                                     *
*                                                                              *
*  PURPOSE: Decrement count on dpf messages being unqueued.                    *
*           Strip out any non-failover messages if appropriate                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive descriptor for which data is being polled                  *
*    unit: the full trace unit that is being dequeued                          *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void adjust_queued_dpf(struct tape_descriptor* drv,
	struct tape_full_trace_unit* unit)
{
	int i = 0, j = 0, k = 0;
	ushort trc_len = 0;

	if(drv->failover_msgs_queued) {
		i = 0;
		while(i < unit->cur_len) {
			if(unit->unit_data[i] == TRC_TP_ATMPT_FAILOVER)
				drv->failover_msgs_queued--;
			i++;
			trc_len = ((unit->unit_data[i] << 8)
				| (unit->unit_data[i + 1] & 0xFF));
			i += 2;
			i += trc_len;
		} /* while */
	} /* if */

	if(drv->remove_pending) {
		/* SCSI device object has been deleted, tape descriptor is
			pending deletion.  We need to let failover messages
			through so any failover caused by the device deletion
			will trigger wakeup.  Remove all other messages */
		i = 0;
		while(i < unit->cur_len) {
			trc_len = ((unit->unit_data[i + 1] << 8)
				| (unit->unit_data[i + 2] & 0xFF)) + 3;
			if(unit->unit_data[i] == TRC_TP_ATMPT_FAILOVER) {
				i += trc_len;
			} else {
				for(k = 0, j = i + trc_len;
					j < unit->cur_len;
					j++, k++) {
					unit->unit_data[i + k]
						= unit->unit_data[j];
				} /* for */
				unit->cur_len -= trc_len;
			} /* else */
		} /* while */

		memset(&unit->unit_data[unit->cur_len], '\0',
			FullTraceUnitSize - unit->cur_len);
	} /* if */
} /* adjust_queued_dpf() */

/*******************************************************************************
*                                                                              *
*  NAME: adjust_queued_cpf                                                     *
*                                                                              *
*  PURPOSE: Decrement count on cpf messages being unqueued.                    *
*           Strip out any non-failover messages if appropriate                 *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer descriptor for which data is being polled                *
*    unit: the full trace unit that is being dequeued                          *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void adjust_queued_cpf(struct changer_descriptor* chg,
	struct tape_full_trace_unit* unit)
{
	int i = 0, j = 0, k = 0;
	ushort trc_len = 0;

	if(chg->failover_msgs_queued) {
		i = 0;
		while(i < unit->cur_len) {
			if(unit->unit_data[i] == TRC_CG_ATMPT_FAILOVER)
				chg->failover_msgs_queued--;
			i++;
			trc_len = ((unit->unit_data[i] << 8)
				| (unit->unit_data[i + 1] & 0xFF));
			i += 2;
			i += trc_len;
		} /* while */
	} /* if */

	if(chg->remove_pending) {
		/* SCSI device object has been deleted, chg descriptor is
			pending deletion.  We need to let failover messages
			through so any failover caused by the device deletion
			will trigger wakeup.  Remove all other messages */
		i = 0;
		while(i < unit->cur_len) {
			trc_len = ((unit->unit_data[i + 1] << 8)
				| (unit->unit_data[i + 2] & 0xFF)) + 3;
			if(unit->unit_data[i] == TRC_CG_ATMPT_FAILOVER) {
				i += trc_len;
			} else {
				for(k = 0, j = i + trc_len;
					j < unit->cur_len;
					j++, k++) {
					unit->unit_data[i + k]
						= unit->unit_data[j];
				} /* for */
				unit->cur_len -= trc_len;
			} /* else */
		} /* while */

		memset(&unit->unit_data[unit->cur_len], '\0',
			FullTraceUnitSize - unit->cur_len);
	} /* if */
} /* adjust_queued_cpf() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_retrieve_drive_dump                                          *
*                                                                              *
*  PURPOSE: obtain any outstanding dump and send it to the daemon              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the command passed into lin_tape_daemon_ioctl                    *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_retrieve_drive_dump(uint command, ulong arg)
{
	struct tape_retrv_dump* retrv_dmp = NULL;
	int i = 0, rc = STATUS_SUCCESS, dmp_lockd = False;
	struct tape_descriptor* drv = NULL;

	if(_IOC_SIZE(command) != sizeof(struct tape_retrv_dump)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	retrv_dmp = kmalloc(sizeof(struct tape_retrv_dump), GFP_KERNEL);
	if(!retrv_dmp) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(retrv_dmp, '\0', sizeof(struct tape_retrv_dump));

	if(copy_from_user((char*)retrv_dmp, (char*)arg,
		sizeof(struct tape_retrv_dump))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!retrv_dmp->dmp_len || !retrv_dmp->dmp) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = STATUS_SUCCESS;
	if(!tape_devs) goto EXIT_LABEL;

	/* check all the tape devices for drive dumps */
	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;
		drv = tape_devs[i]->drv;
		READ_LOCK(&drv->dmp_lock);
		dmp_lockd = True;

		/* we need check if this data has already been
		* retrieved by another path.
		*/
		if(!drv->dmp_len && !memcmp(&retrv_dmp->devid[11],
				&drv->devid[11], 5)) {
			READ_UNLOCK(&drv->dmp_lock);
			dmp_lockd = False;
			rc = -ENODATA;
			continue;
		} /* if */

		if(drv->dev_obj && drv->dmp_len > 0
			&& (drv->dmp || drv->op_sg == DUMP_READY)
			&& !memcmp(&retrv_dmp->devid[0], &drv->devid[0],
				MAX_VOLUME_LABEL_SIZE)) {

			if(drv->dmp_len > retrv_dmp->dmp_len) {
				rc = -ENOMEM;
				printk("lin_tape: dump size larger than "\
					"user allocated size for IBMtape%d\n",
					drv->dev_index);
				READ_UNLOCK(&drv->dmp_lock);
				dmp_lockd = False;
				goto EXIT_LABEL;
			} /* if */

			if(drv->op_sglist && drv->dmp == NULL){
				rc = tape_copy_ops_sg_buffer(drv, retrv_dmp->dmp,
						drv->dmp_len);
			}else if(copy_to_user(retrv_dmp->dmp, drv->dmp,	drv->dmp_len)) {
				rc = -EFAULT;
				READ_UNLOCK(&drv->dmp_lock);
				dmp_lockd = False;
				goto EXIT_LABEL;
			} /* else if */
			if (rc != STATUS_SUCCESS){
				READ_UNLOCK(&drv->dmp_lock);
				dmp_lockd = False;
				goto EXIT_LABEL;
			}

			retrv_dmp->dmp_len = drv->dmp_len;

			if(copy_to_user((char*)arg,(char*)retrv_dmp,
				sizeof(struct tape_retrv_dump))) {
				rc = -EFAULT;
				READ_UNLOCK(&drv->dmp_lock);
				dmp_lockd = False;
				goto EXIT_LABEL;
			} /* if */

			READ_UNLOCK(&drv->dmp_lock);
			dmp_lockd = False;

			WRITE_LOCK(&drv->dmp_lock);
			if(drv->dmp && drv->dmp_size_order >= 0)
				free_pages((ulong)drv->dmp,
					drv->dmp_size_order);
			else if(drv->op_sg == DUMP_READY)
				tape_del_ops_sglist(drv);
			drv->dmp = NULL;
			drv->dmp_len = 0;
			drv->dmp_size_order = 0;
			WRITE_UNLOCK(&drv->dmp_lock);
			drv->op_sg =0;
			rc = STATUS_SUCCESS;
			break;
		} else {
			READ_UNLOCK(&drv->dmp_lock);
			dmp_lockd = False;
		} /* else */
	} /* for */

EXIT_LABEL:

	if(dmp_lockd) READ_UNLOCK(&drv->dmp_lock);
	if(retrv_dmp) kfree(retrv_dmp);
	return rc;
} /* lin_tape_retrieve_drive_dump() */

/*******************************************************************************
*                                                                              *
*  NAME: setup_sg_for_ops                                                      *
*                                                                              *
*  PURPOSE: set up scatter gather list for operations needing large amount     *
*           of memory                                                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    dump_size: Size of the dump to be sent                                    *
*    xfer_cnt: number of SCSI requests to send                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int setup_sg_for_ops(struct tape_descriptor* drv, size_t xfer_cnt)
{
	int rc = STATUS_SUCCESS, i = 0;
	size_t remain = 0;
	ulong copy_size = 0;


	/* If the transfer length is less than existing scatter size, we
	will need to resize the scatter list */
	if(!drv->op_sglist || xfer_cnt > drv->op_sglist->total_sg_size) {
		rc = tape_build_op_sg(drv, (ulong)xfer_cnt);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d tape_build_sg ops xfer_cnt "\
				"%lu failed for %d: %d\n", drv->dev_index,
				(ulong)xfer_cnt, drv->op_sg, rc);
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	}else if (xfer_cnt <= drv->op_sglist->total_sg_size) {
		printk("lin_tape: op_sglist is currently in use by %d\n",
				drv->op_sg);
		return -EINVAL;
	}/* if */

	/* make sure everything looks okay to avoid any kernel panic */
	if(!drv->op_sglist || !drv->op_sglist->total_sgl ||
		!drv->op_sglist->total_sg_size) {
		printk("lin_tape: IBMtape%d lin_tape_drive_ops corrupted "\
			"sglist for %d.\n", drv->dev_index, drv->op_sg);
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* determine how many scatter gather we need */
	i = 0;
	remain = xfer_cnt;
	drv->op_sglist->use_sg = 0;
	while(remain > 0 && i < drv->supported_sg_tablesize) {
		if(!drv->op_sglist ||
			!drv->op_sglist->gather[i].length ||
			!sg_page(&drv->op_sglist->gather[i])) {
			printk("lin_tape: IBMtape%d lin_tape_drive_ops "\
				"corrupted sglist.\n", drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		copy_size = (drv->op_sglist->gather[i].length < remain) ?
			drv->op_sglist->gather[i].length : remain;

		memcpy(&drv->op_sglist->sgl_for_io[i],
			&drv->op_sglist->gather[i],
			sizeof(struct scatterlist));

		drv->op_sglist->sgl_for_io[i].length = copy_size;

		remain -= copy_size;	/* decrease the remaining length */
		drv->op_sglist->use_sg++;
		i++;
	} /* while */

	/* if we still have some left over, that means we went over the
	scatter list which could only mean corrupted scatter list */
	if(remain > 0) {
		printk("lin_tape: IBMtape%d lin_tape_drive_ops data overflow "\
			"detected on %d.\n", drv->dev_index, drv->op_sg);
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;
} /* setup_sg_for_ops() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_build_op_sg                                                      *
*                                                                              *
*  PURPOSE: Build sg list operations large data transfers                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: the number of bytes for the new allocation                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/

int tape_build_op_sg(struct tape_descriptor *drv, size_t xfer_length)
{
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: tape_build_sg entry\n"));
	if(!xfer_length) {
		printk("lin_tape: IBMtape%d tape_build_sg 0 transferlength\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(drv->op_sglist && (!drv->op_sglist->total_sgl ||
		!drv->op_sglist->total_sg_size)) {
		printk("lin_tape: IBMtape%d tape_build_sg scatterlist is "\
			"corrupted: sg %p sgl %d size %d\n", drv->dev_index,
			drv->op_sglist, drv->op_sglist->total_sgl,
			drv->op_sglist->total_sg_size);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(drv->op_sglist && xfer_length <= drv->op_sglist->total_sgl) {
		printk("lin_tape: IBMtape%d tape_build_sg driver logic error\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->op_sglist)
		rc = tape_init_op_sg(drv, xfer_length);
	else {
		printk("lin_tape: IBMtape%d tape_build_op_sg "\
			"already exists\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* else */

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: IBMtape%d tape_build_op_sg failed requesting "\
			"buffer size %lu.\n", drv->dev_index,
			(ulong)xfer_length);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:
	DbgPrint(("lin_tape: tape_build_sg exit: %d\n", rc));
	return rc;
} /* tape_build_op_sg() */


/*******************************************************************************
*                                                                              *
*  NAME: tape_init_op_sg                                                          *
*                                                                              *
*  PURPOSE: Allocate sg list for read/write                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    alloc_length: the number of bytes for the new allocation                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/

int tape_init_op_sg(struct tape_descriptor *drv, size_t alloc_length)
{
	int i = 0, rc = STATUS_SUCCESS;
	ulong size = 0;	/* allocation size */
	ulong order = 0; /*  2^order length */
	long remain = 0; /* remain to be allocated length */
	long cur_alloc_length = 0;

	DbgPrint(("lin_tape: tape_init_sg entry\n"));

	drv->op_sglist = kmalloc(sizeof(struct tape_sgl),
		GFP_KERNEL | __GFP_HIGH);

	if(!drv->op_sglist) {
		printk("lin_tape: IBMtape%d tape_init_op_sg alloc scatterlist\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->op_sglist, '\0', sizeof(struct tape_sgl));

	i = 0;
	remain = alloc_length;
	order = get_mem_ord(alloc_length, &size);
	while(remain > 0 && i < drv->supported_sg_tablesize) {
		unchar *address = NULL;

		if(sg_page(&drv->op_sglist->gather[i])) {
			printk("lin_tape: IBMtape%d tape_init_op_sg gather %d "\
				"memory leak.\n", drv->dev_index, i);
			break;
		} /* if */

		/* try to allocate length */
		if(remain > DUMP_CHUNK_SIZE) {
			cur_alloc_length = min((int)size,
					DUMP_CHUNK_SIZE);
			order = get_mem_ord(cur_alloc_length, &size);
		} /* if */

		address = (unchar *) __get_free_pages(GFP_KERNEL, order);

		if(address) {
			sg_set_page(&drv->op_sglist->gather[i],
				virt_to_page(address), size,
				(unsigned long)address & ~PAGE_MASK);
			drv->op_sglist->total_sg_size += size;
			drv->op_sglist->total_sgl++;
			remain -= size;
			i++;

		} else {

			order--;
			size /= 2;

			if((alloc_length - drv->op_sglist->total_sg_size) >
				((drv->supported_sg_tablesize -
				drv->op_sglist->total_sgl) * size)) {
				printk("lin_tape: IBMtape%d tape_init_op_sg "\
					"requested size is greater than "\
					"allocation limit, total_sgl %d "\
					"totalsg_size  %u, gather size %lu\n",
					drv->dev_index, drv->op_sglist->total_sgl,
					drv->op_sglist->total_sg_size, size);
				break;
			}
		} /* else */
	} /* while */

	/* if allocation failed, then cleanup */
	if(remain > 0) {
		i = 0;
		while(i < drv->supported_sg_tablesize &&
			drv->op_sglist->gather[i].length) {

			if(!sg_page(&drv->op_sglist->gather[i])) {
				drv->op_sglist->gather[i].length = 0;
			} else {
				/* determine order for every scatter gather */
				order = get_mem_ord(drv->op_sglist->gather[i]
					.length, &size);

				free_pages((ulong)
					(sg_virt(&drv->op_sglist->gather[i])),
					order);
				sg_set_page(&drv->op_sglist->gather[i],
					NULL, 0, 0);
				drv->op_sglist->total_sgl--;
				drv->op_sglist->total_sg_size -= size;
				i++;
			} /* else */
		} /* while */

		if(drv->op_sglist->total_sgl || drv->op_sglist->total_sg_size)
			printk("lin_tape: IBMtape%d tape_init_op_sg warning: "\
					"possible memory leak, total_sgl %d, "\
					"total_sg_size %u\n", drv->dev_index,
				drv->op_sglist->total_sgl,
				drv->op_sglist->total_sg_size);

		kfree(drv->op_sglist);
		drv->op_sglist = NULL;

		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_init_op_sg exit: %d\n", rc));
	return rc;
} /* tape_init_op_sg() */


/*******************************************************************************
*                                                                              *
*  NAME: tape_del_ops_sglist                                                   *
*                                                                              *
*  PURPOSE: Deallocate scatter gather list for opertations                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the user pages to deallocate         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/
int tape_del_ops_sglist(struct tape_descriptor *drv)
{
	int i = 0, rc = STATUS_SUCCESS;
	ulong order = 0;	/* 2^order length */
	ulong size = 0;
	struct tape_sgl *sg_list;

	DbgPrint(("lin_tape: tape_del_ops_sglist entry\n"));

	sg_list = drv->op_sglist;

	if(!sg_list) {
		printk("lin_tape: IBMtape%d tape_del_ops_sglist is NULL\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < sg_list->total_sgl; i++) {
		if(sg_list->gather[i].length) {
			if(!sg_page(&sg_list->gather[i]))
				sg_list->gather[i].length = 0;

			/* determine the order for every scatter gather */
			order = get_mem_ord(sg_list->gather[i].length,
				&size);

			free_pages((ulong)(sg_virt(&sg_list->gather[i])),
				order);
			sg_set_page(&sg_list->gather[i], NULL, 0, 0);
		} else {
			if(sg_page(&sg_list->gather[i]))
				printk("lin_tape: IBMtape%d tape_del_ops_sglist "\
					"detected memory leak\n",
				drv->dev_index);
		} /* else */
	} /* for */

	kfree(drv->op_sglist);
	drv->op_sglist = NULL;
	drv->op_sg = 0;

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_del_ops_sglist exit: %d\n", rc));
	return rc;

} /* tape_del_ops_sglist() */


/*******************************************************************************
*                                                                              *
*  NAME: tape_copy_ops_sg_buffer                                               *
*                                                                              *
*  PURPOSE: copy the data from driver's sg buffer to user's buffer             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    user_location: data area in user space to which data should be copied     *
*    buf_length: number of bytes to copy                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_copy_ops_sg_buffer(struct tape_descriptor* drv, char* user_location,
	size_t buf_length)
{
	int rc = STATUS_SUCCESS;
	int i = 0;
	ulong remain = 0;
	size_t copy_size = 0;
	char* temp_location = user_location;

	remain = buf_length;

	while(remain > 0 && i < drv->supported_sg_tablesize) {
		if(!drv->op_sglist || !drv->op_sglist->gather[i].length
			|| !sg_page(&drv->op_sglist->gather[i])) {
			printk("lin_tape: IBMtape%d "\
				"lin_tape_copy_operation_buffer corrupted "\
				"scatter gather, returned EFAULT.\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		/* copying either scatter list size or the remaining */
		copy_size = (drv->op_sglist->gather[i].length < remain) ?
			drv->op_sglist->gather[i].length : remain;

		if(copy_to_user(temp_location,
				sg_virt(&drv->op_sglist->gather[i]),
				copy_size)) {
			printk("lin_tape: IBMtape%d sg copy failed\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		temp_location += copy_size;
		remain -= copy_size;
		drv->op_sglist->use_sg--;
		i++;
	} /* while */

	/* if we still have some left over, that means we went over the
	* scatter gathers which could only mean corrupted scatter list
	* We could still have some unused scatter list when reading
	* less than expected
	*/
	if(remain > 0 || drv->op_sglist->use_sg < 0) {
		printk("lin_tape: IBMtape%d tape_copy_ops_buffer "\
			"data overflowed scatterlist\n",
			drv->dev_index);

		if(remain > 0)
			printk("lin_tape: IBMtape%d "\
				"tape_copy_ops_buffer remain is "\
				"%ld, returned EFAULT\n",
				drv->dev_index, remain);
		if(drv->op_sglist && drv->op_sglist->use_sg < 0)
			printk("lin_tape: IBMtape%d "\
				"tape_copy_ops_buffer use_sg is %d",
				drv->dev_index,
				drv->op_sglist->use_sg);

		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(rc != STATUS_SUCCESS)
		printk("lin_tape: tape_copy_op_sg_buffer rc: %d\n", rc);
	return rc;
} /* tape_copy_ops_sg_buffer() */


/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_retrieve_log_data                                            *
*                                                                              *
*  PURPOSE: obtain any outstanding log data and send it to the daemon          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the command passed into lin_tape_daemon_ioctl                    *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_retrieve_log_data(uint command, ulong arg)
{
	struct tape_retrv_logsense *retrv_logsense = NULL;
	struct tape_descriptor* drv = NULL;
	int i = 0, rc = 0, lgsns_lockd = False;

	if(_IOC_SIZE(command) != sizeof(struct tape_retrv_logsense)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	retrv_logsense = kmalloc(sizeof(struct tape_retrv_logsense),
		GFP_KERNEL);
	if(!retrv_logsense) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(retrv_logsense, '\0', sizeof(struct tape_retrv_logsense));
	if(copy_from_user((char*)retrv_logsense, (char*)arg,
		sizeof(struct tape_retrv_logsense))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!retrv_logsense->logsense_len || !retrv_logsense->logsense) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = STATUS_SUCCESS;
	if(!tape_devs) goto EXIT_LABEL;

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;
		drv = tape_devs[i]->drv;
		READ_LOCK(&drv->logsense_lock);
		lgsns_lockd = True;

		if(!drv->logsense_len
			&& !memcmp(&retrv_logsense->devid[11],
				&drv->devid[11], 5)) {
			rc = -ENODATA;
			goto EXIT_LABEL;
		} /* if */

		if(drv->dev_obj && drv->logsense_len > 0 && drv->logsense
			&& !memcmp(retrv_logsense->devid, drv->devid,
			MAX_VOLUME_LABEL_SIZE)) {

			if(drv->logsense_len > retrv_logsense->logsense_len) {
				rc = -ENOMEM;
				goto EXIT_LABEL;
			} /* if */

			if(copy_to_user(retrv_logsense->logsense, drv->logsense,
					drv->logsense_len)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */

			retrv_logsense->logsense_len = drv->logsense_len;

			if(copy_to_user((char*)arg, (char*)retrv_logsense,
					sizeof(struct tape_retrv_logsense))) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			READ_UNLOCK(&drv->logsense_lock);
			lgsns_lockd = False;

			WRITE_LOCK(&drv->logsense_lock);
			if(drv->logsense) kfree(drv->logsense);
			drv->logsense = NULL;
			drv->logsense_len = 0;
			WRITE_UNLOCK(&drv->logsense_lock);

			rc = STATUS_SUCCESS;
			break;
		} else {
			READ_UNLOCK(&drv->logsense_lock);
			lgsns_lockd = False;
		} /* else */
	} /* for */

EXIT_LABEL:

	if(lgsns_lockd) READ_UNLOCK(&drv->logsense_lock);
	if(retrv_logsense) kfree(retrv_logsense);
	return rc;
} /* lin_tape_retrieve_log_data() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_retrieve_simmim_data                                         *
*                                                                              *
*  PURPOSE: obtain any outstanding simmim data and send it to the daemon       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    command: the command passed into lin_tape_daemon_ioctl                    *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_retrieve_simmim_data(uint command, ulong arg)
{
	struct tape_retrv_simmim *retrv_simmim = NULL;
	struct tape_descriptor* drv = NULL;
	int i = 0, rc = STATUS_SUCCESS, simm_lockd = False;

	if(_IOC_SIZE(command) != sizeof(struct tape_retrv_simmim)) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	retrv_simmim = kmalloc(sizeof(struct tape_retrv_simmim), GFP_KERNEL);
	if(!retrv_simmim) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(retrv_simmim, '\0', sizeof(struct tape_retrv_simmim));
	if(copy_from_user((char*)retrv_simmim, (char*)arg,
		sizeof(struct tape_retrv_simmim))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(!retrv_simmim->simmim_len || !retrv_simmim->simmim) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = STATUS_SUCCESS;
	if(!tape_devs) goto EXIT_LABEL;

	for(i = 0; i < MaxLinTapeDevices; i++) {
		if(!tape_devs[i] || !tape_devs[i]->drv) continue;
		drv = tape_devs[i]->drv;
		READ_LOCK(&drv->simmim_lock);
		simm_lockd = True;

		if(!drv->simmim_len && !memcmp(&retrv_simmim->devid[11],
			&drv->devid[11], 5)) {
			rc = -ENODATA;
			goto EXIT_LABEL;
		} /* if */

		if(drv->dev_obj && drv->simmim_len > 0 && drv->simmim
			&& !memcmp(&retrv_simmim->devid[0], &drv->devid[0],
				MAX_VOLUME_LABEL_SIZE)) {

			if(drv->simmim_len > retrv_simmim->simmim_len) {
				rc = -ENOMEM;
				goto EXIT_LABEL;
			} /* if */

			if(copy_to_user(retrv_simmim->simmim, drv->simmim,
				drv->simmim_len)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */

			retrv_simmim->simmim_len = drv->simmim_len;

			if(copy_to_user((char*)arg, (char*)retrv_simmim,
				sizeof(struct tape_retrv_simmim))) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */

			READ_UNLOCK(&drv->simmim_lock);
			simm_lockd = False;

			WRITE_LOCK(&drv->simmim_lock);
			if(drv->simmim) kfree(drv->simmim);
			drv->simmim = NULL;
			drv->simmim_len = 0;
			WRITE_UNLOCK(&drv->simmim_lock);

			rc = STATUS_SUCCESS;
			break;
		} else {
			READ_UNLOCK(&drv->simmim_lock);
			simm_lockd = False;
		} /* else */
	} /* for */

EXIT_LABEL:

	if(simm_lockd) READ_UNLOCK(&drv->simmim_lock);
	if(retrv_simmim) kfree(retrv_simmim);
	return rc;
} /* lin_tape_retrieve_simmim_data() */
/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_wake_up                                                      *
*                                                                              *
*  PURPOSE: handle daemon notification of task completion; wake up kmode proc. *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/
int lin_tape_wake_up(ulong arg)
{
	struct device_wakeup_t dev_wakeup;
	int prv = 0, new = 0, rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: lin_tape_wake_up entry.\n"));

	memset(&dev_wakeup, '\0', sizeof(struct device_wakeup_t));

	if(copy_from_user((char*)&dev_wakeup, (char*)arg,
		sizeof(struct device_wakeup_t))) {
		printk("lin_tape: received a wakeup call, but "\
			"copy_from_user failed.\n");
		rc = -EFAULT;
	} /* if */
	prv = (int)dev_wakeup.prv_path;
	new = (int)dev_wakeup.new_path;
	if(dev_wakeup.dev_type == TYPE_TAPE) {
		if(!tape_devs || !tape_devs[prv] || !tape_devs[prv]->drv) {
			printk("lin_tape: received a wakeup call with "\
				"invalid drive index %d or %d - %p\n", prv, new, tape_devs);
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		rc = -EINVAL;
		if(dev_wakeup.result >= STATUS_SUCCESS && tape_devs[prv]->drv->drv_open){
            if (tape_devs[new]&&tape_devs[new]->drv&&
                tape_devs[new]->drv->dev_obj &&
                KOBJ_CT(&tape_devs[new]->drv->dev_obj->sdev_gendev.kobj))
                    rc = lin_tape_sdev_get(tape_devs[new]->drv->dev_obj, &tape_devs[new]->drv->sdev_ct);
            if (rc == STATUS_SUCCESS)
                if( tape_devs[prv]->drv->dev_obj && KOBJ_CT(&tape_devs[prv]->drv->dev_obj->sdev_gendev.kobj))
                    lin_tape_sdev_put(tape_devs[prv]->drv->dev_obj, &tape_devs[prv]->drv->sdev_ct);
		}

		WRITE_LOCK(&tape_devs[prv]->drv_entry_lock);
		tape_devs[prv]->drv->wakeup_result = dev_wakeup.result;
		tape_devs[prv]->drv->encryption_return_code = dev_wakeup.result2;
		if(dev_wakeup.result >= STATUS_SUCCESS) {
			tape_devs[prv]->drv->cur_path = new;
			tape_devs[prv]->drv->wakeup_info = dev_wakeup.information;
		} /* if */
		DbgPrint(("lin_tape: current path is %d\n",
			tape_devs[prv]->drv->cur_path));
		tape_devs[prv]->drv->ready_wakeup = True;
		WRITE_UNLOCK(&tape_devs[prv]->drv_entry_lock);

		wake_up_interruptible(&tape_devs[prv]->drv->wait_queue);
		rc = STATUS_SUCCESS;

	} else if(dev_wakeup.dev_type == TYPE_MEDIUM_CHANGER) {
		if(!changer_devs || !changer_devs[prv] || !changer_devs[prv]->chg) {
			printk("lin_tape: received a wakeup call with "\
				"invalid changer index %d or %d\n", prv, new);
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		if(changer_devs[new] && changer_devs[new]->chg && prv != new &&
            dev_wakeup.result == STATUS_SUCCESS && changer_devs[new]->chg->chg_open){
            if (changer_devs[new]->chg->dev_obj &&
                KOBJ_CT(&changer_devs[new]->chg->dev_obj->sdev_gendev.kobj))
                    rc = lin_tape_sdev_get(changer_devs[new]->chg->dev_obj, &changer_devs[new]->chg->sdev_ct);
            if (rc == STATUS_SUCCESS)
                if(changer_devs[prv]->chg->dev_obj && KOBJ_CT(&changer_devs[prv]->chg->dev_obj->sdev_gendev.kobj))
                    lin_tape_sdev_put(changer_devs[prv]->chg->dev_obj, &changer_devs[prv]->chg->sdev_ct);
        }
		WRITE_LOCK(&changer_devs[prv]->entry_lock);
		changer_devs[prv]->chg->wakeup_result = dev_wakeup.result;
		if(dev_wakeup.result == STATUS_SUCCESS && changer_devs[new] &&
            changer_devs[new]->chg)
			changer_devs[prv]->chg->cur_path = new;

		changer_devs[prv]->chg->ready_wakeup = True;
		DbgPrint(("lin_tape: current path is %d\n",
			changer_devs[prv]->chg->cur_path));
		WRITE_UNLOCK(&changer_devs[prv]->entry_lock);
		rc = STATUS_SUCCESS;
		wake_up_interruptible(&changer_devs[prv]->chg->wait_queue);
	} else {
		printk("lin_tape: received a wakeup call with invalid device "\
			"type: %d\n", dev_wakeup.dev_type);
		rc = -EINVAL;
	} /* else */

EXIT_LABEL:

	return rc;
} /* lin_tape_wake_up() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_send_command                                                 *
*                                                                              *
*  PURPOSE: send a SCSI command to the device on behalf of the daemon          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_send_command(ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct device_command_t* dev_command = NULL;
	ulong order = 0;
	ulong buf_size = 0;
	unchar* buffer = NULL;

	dev_command = kmalloc(sizeof(struct device_command_t), GFP_KERNEL);
	if(!dev_command) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(dev_command, '\0', sizeof(struct device_command_t));

	if(copy_from_user((char*)dev_command, (char*)arg,
			sizeof(struct device_command_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* memory buffer management if there is buffer involved */
	if(dev_command->data_direction != SCSI_DATA_NONE) {
		if(!dev_command->buffer_length || !dev_command->buffer) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		if(dev_command->buffer_length > KmallocThreshhold) {
			order = get_mem_ord(dev_command->buffer_length,
				&buf_size);
			buffer = (char*)__get_free_pages(GFP_KERNEL, order);
		} else {
			buffer = kmalloc(dev_command->buffer_length,GFP_KERNEL);
			buf_size = dev_command->buffer_length;
		} /* else */
		if(!buffer) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
		memset(buffer, '\0', buf_size);

		if(dev_command->data_direction == SCSI_DATA_OUT) {
			if(copy_from_user(buffer, dev_command->buffer,
				dev_command->buffer_length)) {
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
		} /* if */
	} /* if */

	if(dev_command->dev_type == TYPE_TAPE) {
		rc = lin_tape_send_cmd_drive(arg, buffer, buf_size,
			dev_command);
	} else if(dev_command->dev_type == TYPE_MEDIUM_CHANGER) {
		rc = lin_tape_send_cmd_changer(arg, buffer, buf_size,
			dev_command);
	} else {
		DbgPrint(("lin_tape: LIN_TAPE_SEND_COMMAND unsupported "\
			"device type %d\n", dev_command->dev_type));
		rc = -EINVAL;
	} /* else */

EXIT_LABEL:

	if(buffer) {
		if(dev_command->buffer_length > KmallocThreshhold)
			free_pages((ulong)buffer, order);
		else
			kfree(buffer);
		buffer = NULL;
	} /* if */

	if(dev_command) {
		kfree(dev_command);
		dev_command = NULL;
	} /* if */

	return rc;
} /* lin_tape_send_command() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_send_cmd_drive                                               *
*                                                                              *
*  PURPOSE: send a SCSI command to the drive on behalf of the daemon           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*    buffer: the data buffer                                                   *
*    buf_size: size of data buffer                                             *
*    dev_command: SCSI command and other information from user                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_send_cmd_drive(ulong arg, unchar* buffer, int buf_size,
	struct device_command_t* dev_command)
{
	int rc = STATUS_SUCCESS, pr_size = 0, pr_count = 0, result = 0,
		resid = 0;
	unsigned int pr_gen = 0;
	boolean obj_gotten = False;
	struct tape_descriptor* drv = NULL;
	unchar* sense_buffer = NULL;
	READ_LOCK(&tape_des_lock);
	if (!tape_devs || !tape_devs[dev_command->logical_dev_num] ||
        ! (drv = tape_devs[dev_command->logical_dev_num]->drv) ||
        !drv->dev_obj || drv->ghost_device) {
        READ_UNLOCK(&tape_des_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	/* make sure device matches */
	if(memcmp(dev_command->serial_num, drv->serial_num, drv->serial_num_len)) {
		printk("lin_tape: lin_tape_send_cmd_drive mismatched "\
			"device serial number %s, %s\n",
			dev_command->serial_num, drv->serial_num);
		rc = -EINVAL;
		READ_UNLOCK(&tape_des_lock);
		goto EXIT_LABEL;
	} /* if */
    READ_UNLOCK(&tape_des_lock);
	WRITE_LOCK(&tape_devs[dev_command->logical_dev_num]->drv_entry_lock);
	rc = lin_tape_sdev_get(drv->dev_obj, &drv->sdev_ct);
	if(rc != STATUS_SUCCESS) {
        WRITE_UNLOCK(&tape_devs[dev_command->logical_dev_num]->drv_entry_lock);
		printk("lin_tape: lin_tape_sdev_get failed on IBMtape%d\n",
			drv->dev_index);
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */
	if (tape_devs[dev_command->logical_dev_num]->drv->busy)
        DbgPrint(("lin_tape: IBMtape%d is busy %d\n",dev_command->logical_dev_num,
                  tape_devs[dev_command->logical_dev_num]->drv->busy));
    tape_devs[dev_command->logical_dev_num]->drv->busy |= F_SEND;
	obj_gotten = True;
    WRITE_UNLOCK(&tape_devs[dev_command->logical_dev_num]->drv_entry_lock);

	/* send scsi command down to device */
	sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(!sense_buffer) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	result = tape_execute_scsi_command(drv->dev_obj, dev_command->scsi_cdb,
		dev_command->data_direction, buffer, buf_size, 0,
		dev_command->timeout_value * HZ, sense_buffer, &resid);
	dev_command->result = result;
	if (result == -ENOMEM){
		rc = result;
		goto EXIT_LABEL;
	}

	if(result != STATUS_SUCCESS) {
		rc = -EIO;
		if((driver_byte(result) & DRIVER_SENSE) && sense_buffer) {
			memcpy(dev_command->sense, sense_buffer,
				min(SCSI_SENSE_BUFFERSIZE, 255));
			dev_command->sense_length =
				min(SCSI_SENSE_BUFFERSIZE, 255);
		} /* if */
	} else {
		if(dev_command->scsi_cdb[0] == PERSISTENT_RESERVE_IN
			&& dev_command->scsi_cdb[1] == QUERY_RESERVATIONS) {
			pr_gen = (buffer[0] << 24) + (buffer[1] << 16) +
				(buffer[2] << 8) + buffer[3];
			pr_size = (buffer[4] << 24) + (buffer[5] << 16) +
				(buffer[6] << 8) + buffer[7];
			pr_count = pr_size / 16;
			if(pr_count == 1) {
				drv->pr_generation = pr_gen;
				memcpy(drv->reservation_key, &buffer[8],
					RESERVE_KEY_LENGTH);

			} /* if */
		} /* if */
	} /* else */

	/* copy data to user space */
	if(copy_to_user((char*)arg, (char*)dev_command,
		sizeof(struct device_command_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(dev_command->data_direction == SCSI_DATA_IN) {
		if(copy_to_user(dev_command->buffer, buffer,
			dev_command->buffer_length)) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

EXIT_LABEL:
    WRITE_LOCK(&tape_devs[dev_command->logical_dev_num]->drv_entry_lock);
	if(obj_gotten && tape_devs[dev_command->logical_dev_num]->drv) {
        lin_tape_sdev_put(drv->dev_obj, &drv->sdev_ct);
        tape_devs[dev_command->logical_dev_num]->drv->busy &= (0xFF-F_SEND);
	}
	WRITE_UNLOCK(&tape_devs[dev_command->logical_dev_num]->drv_entry_lock);

	if(sense_buffer) {
		kfree(sense_buffer);
		sense_buffer = NULL;
	} /* if */
	return rc;
} /* lin_tape_send_cmd_drive() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_send_cmd_changer                                             *
*                                                                              *
*  PURPOSE: send a SCSI command to the changer on behalf of the daemon         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*    buffer: the data buffer                                                   *
*    buf_size: size of data buffer                                             *
*    dev_command: SCSI command and other information from user                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_send_cmd_changer(ulong arg, unchar* buffer, int buf_size,
	struct device_command_t* dev_command)
{
	int rc = STATUS_SUCCESS, result = 0, resid = 0;
	boolean obj_gotten = False;
	struct changer_descriptor* chg = NULL;
	unchar* sense_buffer = NULL;
	int busy = 0;
	char stringcdb[50];
	memset(stringcdb, '\0', 50);

	if(!changer_devs[dev_command->logical_dev_num] ||
        !changer_devs[dev_command->logical_dev_num]->chg ||
        !changer_devs[dev_command->logical_dev_num]->chg->dev_obj ||
        changer_devs[dev_command->logical_dev_num]->chg->ghost_device) {
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	chg = changer_devs[dev_command->logical_dev_num]->chg;
	sprintf(stringcdb, "lin_tape_send_cmd_changer 0x%02x", dev_command->scsi_cdb[0]);
	rc = changer_wait_to_set_busy(&chg, stringcdb,0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	busy=1;

	/* make sure device matches */
	if(memcmp(dev_command->serial_num, chg->serial_num, 16)) {
		printk("lin_tape: LIN_TAPE_SEND_COMMAND mismatched "\
			"device serial number, %s, %s\n",
			dev_command->serial_num,
			chg->serial_num);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
	rc = lin_tape_sdev_get(chg->dev_obj, &chg->sdev_ct);
	if(rc != STATUS_SUCCESS) {
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
		printk("lin_tape: lin_tape_sdev_get failed on "\
			"IBMchanger%d\n", chg->dev_index);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	if (chg->busy)
        DbgPrint(("lin_tape: IBMChanger%d busy %d\n",chg->dev_index, chg->busy));
    chg->busy |= F_SEND;
	obj_gotten = True;
	WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);

	/* send scsi command down to device */
	sense_buffer = kmalloc(SCSI_SENSE_BUFFERSIZE, GFP_KERNEL);
	if(sense_buffer == NULL) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	memset(sense_buffer, '\0', SCSI_SENSE_BUFFERSIZE);

	result = tape_execute_scsi_command(chg->dev_obj, dev_command->scsi_cdb,
		dev_command->data_direction, buffer, buf_size, 0,
		dev_command->timeout_value * HZ, sense_buffer, &resid);

	dev_command->result = result;

	if (result == -ENOMEM){
		rc = result;
		goto EXIT_LABEL;
	}

	if(result != STATUS_SUCCESS) {
		rc = -EIO;
		if((driver_byte(result) & DRIVER_SENSE) && sense_buffer) {
			memcpy(dev_command->sense, sense_buffer,
				min(SCSI_SENSE_BUFFERSIZE, 255));
			dev_command->sense_length =
				min(SCSI_SENSE_BUFFERSIZE, 255);
		} /* if */
	} /* if */

	/* copy data to user space */
	if(copy_to_user((char*)arg, (char*)dev_command,
		sizeof(struct device_command_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(dev_command->data_direction == SCSI_DATA_IN) {
		if(copy_to_user((unchar*)dev_command->buffer, buffer,
			dev_command->buffer_length)) {
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

EXIT_LABEL:

	if(obj_gotten) {
        WRITE_LOCK(&changer_devs[chg->dev_index]->entry_lock);
        lin_tape_sdev_put(chg->dev_obj, &chg->sdev_ct);
        chg->busy &= (0xFF-F_SEND);
        WRITE_UNLOCK(&changer_devs[chg->dev_index]->entry_lock);
	}

	if(sense_buffer) {
		kfree(sense_buffer);
		sense_buffer = NULL;
	} /* if */

	return rc;
} /* lin_tape_send_cmd_changer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_handle_erp                                                   *
*                                                                              *
*  PURPOSE: Handle error recovery procedure and associated processes           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    arg: the argument passed into lin_tape_daemon_ioctl                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_handle_erp(ulong arg)
{
	int rc = STATUS_SUCCESS;
	struct device_erp_t* dev_erp = NULL;
	struct tape_descriptor *drv = NULL;

	DbgPrint(("lin_tape: lin_tape_handle_erp entry.\n"));

	dev_erp = kmalloc(sizeof(struct device_erp_t), GFP_KERNEL);
	if(!dev_erp) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	memset(dev_erp, '\0', sizeof(struct device_erp_t));

	if(copy_from_user((char*)dev_erp, (char*)arg,
			sizeof(struct device_erp_t))) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	if(dev_erp->dev_type == TYPE_TAPE) {
        if (!tape_devs || !tape_devs[dev_erp->logical_dev_num]
            || !tape_devs[dev_erp->logical_dev_num]->drv){
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */

		drv = tape_devs[dev_erp->logical_dev_num]->drv;
		/* make sure device matches */
		if(memcmp(dev_erp->serial_num, drv->serial_num, 16)) {
			printk("lin_tape: lin_tape_handle_erp mismatched "\
				"device serial number, %s, %s\n",
			dev_erp->serial_num, drv->serial_num);
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		DbgPrint(("lin_tape: HANDLE ERP %x\n", dev_erp->command));
		tape_update_descriptor_on_erp(drv->cur_path, drv->dev_index);
		rc = tape_send_erp_cmd(drv, dev_erp->command, 0, 0);
	} /* if */

EXIT_LABEL:

	if(dev_erp) kfree(dev_erp);
	return rc;
} /* lin_tape_handle_erp() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_read                                                   *
*                                                                              *
*  PURPOSE: perform tape read operations                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    filp: the file pointer for the device                                     *
*    user_location: user buffer into which to copy the read data               *
*    xfer_length: bytes of data to read                                        *
*    offset: unused                                                            *
*                                                                              *
*  RETURNS: bytes read or error code                                           *
*                                                                              *
*******************************************************************************/

ssize_t lin_tape_drive_read(struct tape_descriptor** drvp, struct file* filp,
	char* user_location, size_t xfer_length, loff_t* offset)
{
	ssize_t rc = -EIO;
	size_t per_xfer_len = 0;
	size_t xfer_cnt = 0;
	ulong xfer_offset = 0;
	int done = 0;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;
	boolean unmap = False, err = True, lg_rc = True;

	DbgPrint(("lin_tape: lin_tape_drive_read entry.\n"));

	rc = lin_tape_drive_read_check(&drv);
	if(rc != STATUS_SUCCESS) {
		err = lg_rc = False;
		goto EXIT_LABEL;
	} /* if */
	if(drv->blk_size){
		rc = prepare_fixed_length_read(drv, xfer_length, &per_xfer_len,
			&xfer_cnt);
	}
	else
		rc = prepare_var_length_read(drv, xfer_length, &per_xfer_len,
			&xfer_cnt);
	if(rc != STATUS_SUCCESS) {
		err = lg_rc = False;
		goto EXIT_LABEL;
	} /* if */

	rc = lin_tape_direct_or_buffered(drv, (const char*)user_location,
		xfer_cnt, True, False, &unmap);

	if(rc != STATUS_SUCCESS) {
		/* Unable use DMA method */
		rc = setup_sg_for_read(drv, xfer_cnt);
		if(rc != STATUS_SUCCESS) {
			err = lg_rc = False;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	if(!drv->cdb) {
		printk("lin_tape: IBMtape%d lin_tape_drive_read: failed to "\
			"locate cdb space.\n", drv->dev_index);
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len,
				TRC_RD_NO_CDB);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_RD_NO_CDB, 0, NULL);
		} /* else if */

		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */

	xfer_offset = 0;
	while(True) {
		rc = lin_tape_perform_read(&drv, &per_xfer_len, user_location,
			xfer_length, &xfer_offset, &xfer_cnt, &err, &lg_rc,
			&done, &unmap);
		if(rc < STATUS_SUCCESS) goto EXIT_LABEL;
		else if(done == 1) goto EXIT_LABEL;
		else if(done == 2) break;
	} /* while */

	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	drv->last_errno = STATUS_SUCCESS;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, STATUS_SUCCESS);
	err = lg_rc = False;

	rc = xfer_length;

EXIT_LABEL:

	if(unmap) lin_tape_release_user_pages(drv, True);

	if(err) drv->last_errno = -rc;

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) tape_log_rc(drv, -rc);

	*drvp = drv;

	DbgPrint(("lin_tape: lin_tape_drive_read exit: %d\n", (int)rc));
	return rc;
} /* lin_tape_drive_read() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_read_check                                             *
*                                                                              *
*  PURPOSE: initial check to see if read may proceed                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_drive_read_check(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	READ_LOCK(&tape_des_lock);
	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: lin_tape_drive_read: drive removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

	if(petro == 1 && drv->curr_reserve_type == LinTapeNoReserve) {
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);

	READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
    if(drv->remove_pending) {
		READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
    }
	if(!drv->dev_obj || drv->ghost_device) {
		if(drv->alt_pathing) {
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = tape_perform_failover_generic(&drv, 1800 * HZ);
			READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		} /* if */
		if(rc != STATUS_SUCCESS || !drv->alt_pathing) {
			printk("lin_tape: lin_tape_drive_read: drive "\
				"removed.\n");
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_TP_OUT_RANGE);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENXIO);
			READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
    READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(!ReadAccess(drv->access_mode)) {
		drv->last_errno = EPERM;
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EPERM);
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* lin_tape_drive_read_check() */

/*******************************************************************************
*                                                                              *
*  NAME: prepare_fixed_length_read                                             *
*                                                                              *
*  PURPOSE: check params and set xfer variables for fixed length read          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: size to transfer                                             *
*    per_xfer_len: size to transfer per each SCSI request                      *
*    xfer_cnt: number of SCSI requests to send                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int prepare_fixed_length_read(struct tape_descriptor* drv, size_t xfer_length,
	size_t* per_xfer_len, size_t* xfer_cnt)
{
	int rc = STATUS_SUCCESS;
	unchar msg = 0, ttype = lin_tape_trace_type;
	int def_len = LinTapeDefaultTransferLength;

	if(drv->blk_size > drv->max_xfer_len ||
		drv->blk_size > drv->max_blk_size) {
		printk("lin_tape: IBMtape%d prepare_fixed_length_read: "\
			"request size exceeds max xfer size\n", drv->dev_index);
		rc = -EINVAL;
		msg = TRC_RD_BLK_LEN;
		goto EXIT_LABEL;
	} /* if */
	if(!xfer_length || xfer_length % drv->blk_size) {
		printk("lin_tape: IBMtape%d prepare_fixed_length_read: "\
			"fixed request not multiple of size\n", drv->dev_index);
		rc = -EINVAL;
		msg = TRC_RD_INV_FXD;
		goto EXIT_LABEL;
	} /* if */
	if((drv->blk_size & (ulong)(0x01)) && xfer_length != drv->blk_size) {
		printk("lin_tape: IBMtape%d prepare_fixed_length_read: "\
			"multiple odd block transfer\n", drv->dev_index);
		rc = -EINVAL;
		msg = TRC_RD_ODD_FXD;
		goto EXIT_LABEL;
	} /* if */
	if(drv->sili) {
		printk("lin_tape: IBMtape%d prepare_fixed_length_read: "\
			"SILI bit ON.\n", drv->dev_index);
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */
	/* Setup the length per transfer and transfered length */
	if(xfer_length > def_len) {
		DbgPrint(("IBMtape%d prepare_fixed_length_read: "\
			"xfer len greater than default.\n", drv->dev_index));
		msg = TRC_RD_EXC_DEF;

		if(drv->blk_size <= def_len) {
			*per_xfer_len = (size_t)(def_len / drv->blk_size);
			*xfer_cnt = drv->blk_size * (*per_xfer_len);
		} else {
			*per_xfer_len = 1;
			*xfer_cnt = drv->blk_size;
		} /* else */

	} else {
		*per_xfer_len = (size_t)(xfer_length / drv->blk_size);
		*xfer_cnt = xfer_length;
	} /* else */
EXIT_LABEL:

	if(rc < 0) {
		drv->last_errno = -rc;
	} /* if */

	if(msg) {
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len, msg);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, msg, 0, NULL);
			if(rc < 0) tape_log_rc(drv, -rc);
		} /* else if */

	} /* if */

	return rc;
} /* prepare_fixed_length_read() */

/*******************************************************************************
*                                                                              *
*  NAME: prepare_var_length_read                                               *
*                                                                              *
*  PURPOSE: check params and set xfer variables for var length read            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: size to transfer                                             *
*    per_xfer_len: size to transfer per each SCSI request                      *
*    xfer_cnt: number of SCSI requests to send                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int prepare_var_length_read(struct tape_descriptor* drv, size_t xfer_length,
	size_t* per_xfer_len, size_t* xfer_cnt)
{
	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;

	if(xfer_length > drv->max_xfer_len) {
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len,
				TRC_RD_INV_VAR);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			tape_log_full_trace(drv, TRC_RD_INV_VAR, 0, NULL);
			tape_log_rc(drv, EINVAL);
		} /* else if */

		drv->last_errno = EINVAL;
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	*per_xfer_len = *xfer_cnt = xfer_length;

EXIT_LABEL:

	return rc;

} /* prepare_var_length_read() */

/*******************************************************************************
*                                                                              *
*  NAME: setup_sg_for_read                                                     *
*                                                                              *
*  PURPOSE: set up scatter gather list for multiple read transfers             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_cnt: number of SCSI requests to send                                 *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int setup_sg_for_read(struct tape_descriptor* drv, size_t xfer_cnt)
{
	int rc = STATUS_SUCCESS, i = 0;
	unchar msg = 0;
	long remain = 0;
	ulong copy_size = 0;
	unchar ttype = lin_tape_trace_type;

	/* If the transfer length is less than existing scatter size, we
	will need to resize the scatter list */
	if(!drv->sg_list || xfer_cnt > drv->sg_list->total_sg_size) {
		rc = tape_build_sg(drv, (ulong)xfer_cnt);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d tape_build_sg xfer_cnt "\
				"%lu failed: %d\n", drv->dev_index,
				(ulong)xfer_cnt, rc);
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	/* make sure everything looks okay to avoid any kernel panic */
	if(!drv->sg_list || !drv->sg_list->total_sgl ||
		!drv->sg_list->total_sg_size) {
		printk("lin_tape: IBMtape%d lin_tape_drive_read corrupted "\
			"sglist.\n", drv->dev_index);
		msg = TRC_RD_CRPT_SG;
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* determine how many scatter gather we need */
	i = 0;
	remain = xfer_cnt;
	drv->sg_list->use_sg = 0;
	while(remain > 0 && i < drv->supported_sg_tablesize) {
		if(!drv->sg_list ||
			!drv->sg_list->gather[i].length ||
			!sg_page(&drv->sg_list->gather[i])) {
			printk("lin_tape: IBMtape%d lin_tape_drive_read "\
				"corrupted sglist.\n", drv->dev_index);
			msg = TRC_RD_CRPT_SG;
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		copy_size = (drv->sg_list->gather[i].length < remain) ?
			drv->sg_list->gather[i].length : remain;

		memcpy(&drv->sg_list->sgl_for_io[i],
			&drv->sg_list->gather[i],
			sizeof(struct scatterlist));

		drv->sg_list->sgl_for_io[i].length = copy_size;

		remain -= copy_size;	/* decrease the remaining length */
		drv->sg_list->use_sg++;
		i++;
	} /* while */

	/* if we still have some left over, that means we went over the
	scatter list which could only mean corrupted scatter list */
	if(remain > 0) {
		printk("lin_tape: IBMtape%d lin_tape_drive_read data "\
			"overflow detected.\n", drv->dev_index);
		msg = TRC_RD_OVERFLOW;
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	if(rc < 0) {
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
			if(msg) {
				tape_log_full_trace(drv, msg, 0, NULL);
			} /* if */
			tape_log_rc(drv, -rc);
		} else if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE
			&& msg) {
			tape_log_msg(drv->tape_log, &drv->tape_log_len, msg);
		} /* else if */
		drv->last_errno = -rc;
	} /* if */

	return rc;
} /* setup_sg_for_read() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_perform_read                                                 *
*                                                                              *
*  PURPOSE: set up the CDB, execute the read command and handle errors         *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape_descriptor on which to perform the read                        *
*    per_xfer_len: length of each read request sent to the HBA                 *
*    user_location: buffer to read into                                        *
*    xfer_length: total bytes requested                                        *
*    xfer_offset: into user_location                                           *
*    xfer_cnt: bytes actually read                                             *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*    unmap (output): if True, user pages were mapped by this function          *
*                                                                              *
*  RETURNS: bytes read or error code                                           *
*                                                                              *
*******************************************************************************/

int lin_tape_perform_read(struct tape_descriptor** drvp, size_t* per_xfer_len,
	char* user_location, size_t xfer_length, ulong* xfer_offset,
	size_t* xfer_cnt, boolean* err, boolean* lg_rc, int* done,
	boolean* unmap)
{
	int rc = 0, timeout = 0, read_length = 0;
	struct tape_descriptor* drv = *drvp;

	memset((void*)drv->cdb, '\0', MAX_CDB_SIZE);

	(drv->cdb)->cdb_read_write.op_code = READ_6;
	if(drv->blk_size) {
		(drv->cdb)->cdb_read_write.fixed = True;
		(drv->cdb)->cdb_read_write.sili = False;
	} else {
		(drv->cdb)->cdb_read_write.fixed = False;
		(drv->cdb)->cdb_read_write.sili = drv->sili;
	} /* else */

	(drv->cdb)->cdb_read_write.xfer_length[0] =
		(unchar)(*per_xfer_len >> 16 & 0xFF);
	(drv->cdb)->cdb_read_write.xfer_length[1] =
		(unchar)(*per_xfer_len >> 8 & 0xFF);
	(drv->cdb)->cdb_read_write.xfer_length[2] =
		(unchar)(*per_xfer_len & 0xFF);

	drv->last_op = READ_6;
	drv->cdb_len = CDB_GENERIC_6;

	timeout = compute_timeout(drv->drv_id, ReadTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_FROM_DEVICE, *xfer_cnt, 0);

	if(rc != STATUS_SUCCESS && drv->blk_size && drv->resid_valid &&
		drv->resid > 0) {
		read_length = *xfer_cnt - drv->resid;

		if(read_length > 0) {
			/* encryption partial read          */
			/* consider this good, and move on. */
			rc = STATUS_SUCCESS;
			*xfer_cnt = read_length;
		} /* if */
	} /* if */

	if(rc == STATUS_SUCCESS)
		rc = handle_successful_read(drv, user_location, xfer_length,
			xfer_offset, xfer_cnt, per_xfer_len, err, lg_rc, done,
			unmap);
	else
		rc = handle_read_failure(&drv, user_location, xfer_offset,
			*xfer_cnt, err, lg_rc, done, rc);

	*drvp = drv;
	return rc;
} /* lin_tape_perform_read() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_successful_read                                                *
*                                                                              *
*  PURPOSE: handle post-processing for a successful read request               *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape_descriptor on which the read was performed                      *
*    user_location: buffer to read into                                        *
*    xfer_length: total bytes requested                                        *
*    xfer_offset: into user_location                                           *
*    xfer_cnt: bytes actually read                                             *
*    per_xfer_len: length of each read request sent to the HBA                 *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*    unmap (output): if True, user pages were mapped by this function          *
*                                                                              *
*  RETURNS: bytes read or error code                                           *
*                                                                              *
*******************************************************************************/

int handle_successful_read(struct tape_descriptor* drv, char* user_location,
	size_t xfer_length, ulong* xfer_offset, size_t* xfer_cnt,
	size_t* per_xfer_len, boolean* err, boolean* lg_rc, int* done,
	boolean* unmap)
{
	int rc = STATUS_SUCCESS;

	/* update the medium position settings */
	if(drv->eod || drv->fm_detected || drv->eom) {
		drv->pos_valid = False;
	} else if(drv->pos_valid) {
		if(drv->blk_size) drv->curr_pos += *per_xfer_len;
		else drv->curr_pos++;
	} /* else if */

	if(!drv->blk_size && drv->sili) {
		rc = suc_var_read_sili(drv, user_location + *xfer_offset,
			*xfer_cnt, err, lg_rc, done);
		if(rc < STATUS_SUCCESS || *done) goto EXIT_LABEL;
	} /* if */

	/* copy read data to user space */
	if(!drv->direct_io) {
		rc = tape_copy_read_buffer(drv, user_location + *xfer_offset,
			*xfer_cnt, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	*xfer_offset += *xfer_cnt;
	if(*xfer_offset >= xfer_length) {
		*done = 2;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->blk_size) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	rc = successful_partial_fixed_read(drv, user_location, xfer_length,
		xfer_offset, xfer_cnt, per_xfer_len, unmap);

EXIT_LABEL:

	return rc;

} /* handle_successful_read() */

/*******************************************************************************
*                                                                              *
*  NAME: suc_var_read_sili                                                     *
*                                                                              *
*  PURPOSE: handle post-processing for a var-length read with sili on          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor for drive from which we read                         *
*    buffer_to_copy: back to user space                                        *
*    xfer_cnt: original number of bytes requested                              *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    sili bit support:                                                         *
*      If this is a variable block read and the SILI bit is on, then a length  *
*      mismatch will not return check conditions; therefore we must manually   *
*      check the transfer length returned by the SCSI request.  If the resid   *
*      count is negative, then we have an underlength read and return EIO.     *
*      Some adapters do not report back the right residual value, so we must   *
*      manually test each adapter before we claim this support.                *
*                                                                              *
*******************************************************************************/

int suc_var_read_sili(struct tape_descriptor* drv, char* buffer_to_copy,
	size_t xfer_cnt, boolean* err, boolean* lg_rc, int* done)
{
	int rc = STATUS_SUCCESS;
	long read_length = 0;

	if(!drv->resid_valid) {
		drv->pos_valid = False;
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	if(drv->resid > 0) {
		/* underlength -- record smaller than requested */
		read_length = xfer_cnt - drv->resid;

		if(read_length <= 0) {
			/* < 0  indicates problem with HBA driver
			= 0 only if the record on tape has 0 bytes */
			drv->pos_valid = False;
			rc = -EIO;
			goto EXIT_LABEL;
		} else if(!drv->direct_io) {
			rc = tape_copy_read_buffer(drv, buffer_to_copy,
				read_length, False);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* else if */

		drv->last_errno = STATUS_SUCCESS;
		*err = *lg_rc = False;
		*done = 1;
		rc = read_length;
		goto EXIT_LABEL;
	} else if(drv->resid < 0) {
		/* overlength -- record greater than requested,
		and the BLOCK LENGTH field in the mode sense
		descriptor block is 0. Copy the data to user space.
		Since it is an overlength read, xfer_cnt bytes of data
		must have been read */
		if(!drv->direct_io) {
			rc = tape_copy_read_buffer(drv, buffer_to_copy,
				xfer_cnt, False);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		rc = -EOVERFLOW;
		goto EXIT_LABEL;
	} /* else if */

EXIT_LABEL:

	return rc;
} /* suc_var_read_sili() */

/*******************************************************************************
*                                                                              *
*  NAME: successful_partial_fixed_read                                         *
*                                                                              *
*  PURPOSE: handle post-processing for a partial successful fixed-length read  *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape_descriptor on which the read was performed                      *
*    user_location: buffer to read into                                        *
*    xfer_length: total bytes requested                                        *
*    xfer_offset: into user_location                                           *
*    xfer_cnt: bytes actually read                                             *
*    per_xfer_len: length of each read request sent to the HBA                 *
*    unmap (output): if True, user pages were mapped by this function          *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int successful_partial_fixed_read(struct tape_descriptor* drv,
	char* user_location, size_t xfer_length, ulong* xfer_offset,
	size_t* xfer_cnt, size_t* per_xfer_len, boolean* unmap)
{
	int rc = STATUS_SUCCESS, def_xfer = LinTapeDefaultTransferLength, i = 0;
	long remain = 0, copy_size = 0;
	char ttype = lin_tape_trace_type;

	/* if the bytes remaining still exceed the default transfer length */
	if(xfer_length - *xfer_offset > def_xfer) {
		if(drv->blk_size <= def_xfer) {
			*per_xfer_len = (size_t)(def_xfer / drv->blk_size);
			*xfer_cnt = drv->blk_size * (*per_xfer_len);
		} else {
			*per_xfer_len = 1;
			*xfer_cnt = drv->blk_size;
		} /* else */
	} else {
		/* the leftover can be transfered in one read */
		*per_xfer_len = (size_t)((xfer_length - *xfer_offset) /
			drv->blk_size);
		*xfer_cnt = xfer_length - *xfer_offset;
	} /* else */

	if(*unmap) {
		rc = lin_tape_release_user_pages(drv, True);
		*unmap = False;
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

	rc = lin_tape_direct_or_buffered(drv, user_location + *xfer_offset,
		*xfer_cnt, True, True, unmap);

	/* direct_io means pages are already mapped. !sg_list is error */
	if(drv->direct_io || !drv->sg_list) goto EXIT_LABEL;

	/* calculate the new number of sg pages needed */
	drv->sg_list->use_sg = 0;
	remain = *xfer_cnt;
	i = 0;

	while(remain > 0 && i < drv->supported_sg_tablesize) {
		if(!drv->sg_list || !drv->sg_list->gather[i].length ||
			!sg_page(&drv->sg_list->gather[i])) {
			printk("lin_tape: IBMtape%d lin_tape_drive_read "\
				"corrupted scatter gather, return EFAULT.\n",
				drv->dev_index);

			if(drv->sg_list)
				printk("lin_tape: sg[%d] length %d page %p\n",
					i, drv->sg_list->gather[i].length,
					sg_page(&drv->sg_list->gather[i]));
			else printk("lin_tape: sg_list is NULL.\n");

			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(drv->tape_log,
					&drv->tape_log_len, TRC_RD_CRPT_SG);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
				tape_log_full_trace(drv, TRC_RD_CRPT_SG, 0,
				NULL);
			} /* else if */

			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */

		copy_size = (drv->sg_list->gather[i].length < remain) ?
			drv->sg_list->gather[i].length : remain;

		remain -= copy_size;
		drv->sg_list->use_sg++;
		i++;
	} /* while */

	/* if we still have bytes left, we exceeded the (corrupted) sg list */
	if(remain > 0) {
		printk("lin_tape: IBMtape%d lin_tape_drive_read data "\
			"overflow detected, return EFAULT\n", drv->dev_index);
		printk("lin_tape: remainder is %ld\n", remain);

		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len,
				TRC_RD_SG_OVRRN);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_RD_SG_OVRRN, 0, NULL);

		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;

} /* successful_partial_fixed_read() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_read_failure                                                   *
*                                                                              *
*  PURPOSE: handle post-processing for a failed read request                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive from which we attempted to read       *
*    user_location: the user buffer into which to read                         *
*    xfer_offset: offset from user_location into which to copy                 *
*    xfer_cnt: number of bytes to read                                         *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*    rc: default rc to use if not overwritten                                  *
*                                                                              *
*  RETURNS: bytes read or error code                                           *
*                                                                              *
*******************************************************************************/

int handle_read_failure(struct tape_descriptor** drvp, char* user_location,
	ulong* xfer_offset, size_t xfer_cnt, boolean* err, boolean* lg_rc,
	int* done, int rc)
{
	struct tape_descriptor* drv = *drvp;
	DbgPrint(("lin_tape: handling exception on %s-length read\n",
		drv->blk_size ? "fixed" : "variable"));

	if(drv->blk_size)
		rc = handle_fixed_length_read_failure(&drv, user_location,
			xfer_offset, xfer_cnt, err, lg_rc, done, rc);
	else
		rc = handle_var_length_read_failure(&drv, user_location,
			xfer_offset, xfer_cnt, err, lg_rc, done, rc);

	*drvp = drv;
	return rc;

} /* handle_read_failure() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_fixed_length_read_failure                                      *
*                                                                              *
*  PURPOSE: handle post-processing for a failed fixed-length read request      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive from which we attempted to read       *
*    user_location: the user buffer into which to read                         *
*    xfer_offset: offset from user_location into which to copy                 *
*    xfer_cnt: number of bytes to read                                         *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*    rc: default rc to use if not overwritten                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    If we are doing fixed block read and get a check condition then           *
*      If the check condition is filemark detected:                            *
*        If AIX mode:                                                          *
*          Return the data read and back space file mark                       *
*        If SUN mode:                                                          *
*          Return the data read and continue                                   *
*      If the check condition is not filemark detected:                        *
*        Return error, and let application decided what to do                  *
*                                                                              *
*******************************************************************************/

int handle_fixed_length_read_failure(struct tape_descriptor** drvp,
	char* user_location, ulong* xfer_offset, size_t xfer_cnt, boolean* err,
	boolean* lg_rc, int* done, int rc)
{
	long read_length = 0;
	char* temp_location = user_location + *xfer_offset;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	if(!drv->resid_valid || !drv->fm_detected || !drv->sense_valid ||
		drv->sense_data->sense_key != 0x00 ||
		(drv->sense_data->asc != 0x00 &&
		drv->sense_data->ascq != 0x01)) {
		drv->pos_valid = False;
		goto EXIT_LABEL;
	} /* if */

	if(drv->read_mode == AIXMode && drv->fm_detected) {
		rc = tape_set_pos(&drv, SpaceFileMark, -1);

		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d read: unable to reverse "\
				"space filemark\n", drv->dev_index);
			drv->pos_valid = False;
			goto EXIT_LABEL;
		} /* if */
	} else {
		/* update the medium position setting */
		tape_send_erp_cmd(drv, READ_POSITION, 0, 0);
	} /* else */

	/* if there is a residual then calculate bytes read */
	read_length = xfer_cnt - drv->resid;

	if(read_length < 0) {
		drv->pos_valid = False;
		goto EXIT_LABEL;
	} else if(read_length > 0) {
		/* copy data to user space */
		if(!drv->direct_io) {
			rc = tape_copy_read_buffer(drv, temp_location,
				read_length, False);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		*xfer_offset += read_length;

		drv->last_errno = STATUS_SUCCESS;
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, STATUS_SUCCESS);
		*err = *lg_rc = False;
		*done = 1;
		rc = *xfer_offset;
		goto EXIT_LABEL;
	} /* else if */

	/* read_length == 0
	This should not happen because drv->resid_valid is True.
	Return data that has been successfully read so far */

	drv->pos_valid = False;

	drv->last_errno = STATUS_SUCCESS;

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, STATUS_SUCCESS);
	*err = *lg_rc = False;
	*done = 1;

	rc = *xfer_offset;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* handle_fixed_length_read_failure() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_var_length_read_failure                                        *
*                                                                              *
*  PURPOSE: handle post-processing for a failed fixed-length read request      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor for the drive from which we attempted to read       *
*    user_location: the user buffer into which to read                         *
*    xfer_offset: offset from user_location into which to copy                 *
*    xfer_cnt: number of bytes to read                                         *
*    err (output): if True, set error at end of read; if False, do not set     *
*    lg_rc (output): if True, log error at end; if False, do not log           *
*    done (output): if True, we are done with read                             *
*    rc: default rc to use if not overwritten                                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*  NOTES:                                                                      *
*                                                                              *
*    If we are doing variable block read and get a check condition then        *
*      If the check condition is incorrect length indicator:                   *
*        If the residual value is greather than 0 (under length)               *
*          Return only the data read                                           *
*        If the residual value is less than 0 (over length)                    *
*          Return ENOMEM                                                       *
*        If the check condition is not incorrect length indicator:             *
*          Return Error                                                        *
*                                                                              *
*******************************************************************************/

int handle_var_length_read_failure(struct tape_descriptor** drvp,
	char* user_location, ulong* xfer_offset, size_t xfer_cnt, boolean* err,
	boolean* lg_rc, int* done, int rc)
{
	long read_length = 0;
	char *tmp_loc = user_location + *xfer_offset;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	if(drv->resid_valid && drv->resid < 0) {
		if(drv->pos_valid) drv->curr_pos++;

		/* copy the actual bytes read to user space */
		if(!drv->direct_io) {
			rc = tape_copy_read_buffer(drv, tmp_loc, xfer_cnt, False);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		rc = -EOVERFLOW;
		goto EXIT_LABEL;
	} /* if */

	if(drv->fm_detected && drv->sense_data->asc == 0x00 &&
		drv->sense_data->ascq == 0x01) {
		/* update the medium position setting */
		DbgPrint(("lin_tape: FileMark detected.\n"));

		rc = tape_send_erp_cmd(drv, READ_POSITION, 0, 0);

		drv->last_errno = EIO;

		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EIO);
		*err = *lg_rc = False;
		*done = 1;
		rc = 0;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->sense_data->ili || !drv->resid_valid ||
		drv->sense_data->sense_key || drv->sense_data->asc ||
		drv->sense_data->ascq) {
		drv->pos_valid = False;
		goto EXIT_LABEL;
	} /* if */

	/* if there is a residual then calculate bytes read */
	read_length = xfer_cnt - drv->resid;

	if(read_length < 0) {
		drv->pos_valid = False;
		goto EXIT_LABEL;
	} else if(read_length > 0) {
		if(drv->pos_valid) drv->curr_pos++;

		if(!drv->direct_io) {
			rc = tape_copy_read_buffer(drv, tmp_loc, read_length, False);
			if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
		} /* if */

		*xfer_offset = *xfer_offset + read_length;

		drv->last_errno = STATUS_SUCCESS;

		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, STATUS_SUCCESS);
		*err = *lg_rc = False;
		*done = 1;
		rc = *xfer_offset;
		goto EXIT_LABEL;
	} /* else if */

	/* read_length == 0 */
	drv->pos_valid = False;
	drv->last_errno = -rc;

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) tape_log_rc(drv, -rc);
	*err = *lg_rc = False;
	rc = -EIO;
	goto EXIT_LABEL;

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* handle_var_length_read_failure() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_write                                                  *
*                                                                              *
*  PURPOSE: write function                                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*    filp: the file pointer for the drive to which to write                    *
*    user_location: the user buffer from which to copy the data to write       *
*    xfer_length: size to transfer                                             *
*    offset: unused                                                            *
*                                                                              *
*  RETURNS: length transferred or error code                                   *
*                                                                              *
*******************************************************************************/

ssize_t lin_tape_drive_write(struct tape_descriptor** drvp, struct file* filp,
	const char* user_location, size_t xfer_length, loff_t* offset)
{
	ssize_t rc = -EIO;
	size_t per_xfer_len = xfer_length;
	size_t xfer_cnt = 0;
	ulong xfer_offset = 0;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;
	unchar done = False;
	boolean unmap = False;

	DbgPrint(("lin_tape: lin_tape_drive_write entry.\n"));

	rc = lin_tape_drive_write_check(&drv);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;

	rc = lin_tape_write_validate_parms(drv, xfer_length,
		&per_xfer_len, &xfer_cnt);
	if(rc != STATUS_SUCCESS) goto ERROR_LABEL;

	rc = lin_tape_write_check_eom(drv);
	if(rc != STATUS_SUCCESS) goto ERROR_LABEL;

	rc = lin_tape_direct_or_buffered(drv, user_location, xfer_cnt, False,
		False, &unmap);

	/* if we don't have scatter gather yet */
	if(rc != STATUS_SUCCESS
		&& (!drv->sg_list || xfer_cnt > drv->sg_list->total_sg_size)) {
		/* get at least xfer_length scatterlist */
		rc = tape_build_sg(drv, (ulong)xfer_cnt);
		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d tape_build_sg "\
				"with xfer_cnt %lu failed: %d.\n",
				drv->dev_index, (ulong)xfer_cnt, (int)rc);
			rc = -ENOMEM;
			goto ERROR_LABEL;
		} /* if */
	} /* if */

	if(!drv->direct_io) {
		rc = tape_copy_write_buffer(drv, user_location, xfer_cnt, False);
		if(rc != STATUS_SUCCESS) goto ERROR_LABEL;
	} /* if */

	/* write all the data */
	xfer_offset = 0;
	while(True) {
		rc = lin_tape_perform_write(&drv, &per_xfer_len, &xfer_cnt,
			&xfer_offset, user_location, xfer_length, &done);
		if(rc < STATUS_SUCCESS) goto ERROR_LABEL;
		else if(done == 1) goto SUCCESS_LABEL;
		else if(done == 2) break;
	} /* while */

	*drvp = drv;
	rc = xfer_length;

SUCCESS_LABEL:

	drv->last_errno = STATUS_SUCCESS;

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, STATUS_SUCCESS);

	goto EXIT_LABEL;

ERROR_LABEL:
	if (drv){
	drv->last_errno = -rc;

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, -rc);
	}
	goto EXIT_LABEL;

EXIT_LABEL:

	DbgPrint(("lin_tape_drive_write exit: %d\n", (int)rc));
	return rc;
} /* lin_tape_drive_write() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_drive_write_check                                            *
*                                                                              *
*  PURPOSE: initial check to see if write may proceed                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: tape descriptor associated with the request                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_drive_write_check(struct tape_descriptor** drvp)
{
	int rc = STATUS_SUCCESS;
	unchar ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	READ_LOCK(&tape_des_lock);
	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape_drive_write: drive has been removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

	if(petro == 1 && drv->curr_reserve_type == LinTapeNoReserve) {
		/* writes not allowed if drive is not reserved, which should
		only happen with a petro "nrsv" device */
		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE) {
		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);
		tape_setup_and_log_full(drv, 0, TRC_WT_ENT, NULL, -6);
	} /* if */

	READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if (drv->remove_pending){
        READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	    rc = -ENXIO;
        goto EXIT_LABEL;
	}
	if(!drv->dev_obj || drv->ghost_device) {
		if(drv->alt_pathing) {
            READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = tape_perform_failover_generic(&drv, 1800 * HZ);
			READ_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		} /* if */
		if(!drv->alt_pathing || rc != STATUS_SUCCESS) {
			printk("lin_tape_drive_write: drive has been "\
				"removed.\n");
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_TP_OUT_RANGE);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENXIO);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	if(!WriteAccess(drv->access_mode)) {
		drv->last_errno = EPERM;
		if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_rc(drv, EPERM);

		printk("lin_tape: IBMtape%d does not have write access.\n",
			drv->dev_index);

		rc = -EPERM;
		goto EXIT_LABEL;
	} /* if */

	if(drv->encryption_capable &&

		((drv->sys_encryption_write == STATE_ON &&
		drv->encryption_state != STATE_ON) ||

		(drv->sys_encryption_write == STATE_OFF &&
		drv->encryption_state != STATE_OFF))) {

		/* Incorrect encryption write settings */
		rc = tape_encryption_write_setting(&drv);
		if(rc != STATUS_SUCCESS)
		{
			if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, -rc);
			printk("lin_tape: IBMtape%d encryption write "\
				"settings are incorrectly set - encryption "\
				"state = %d\n", drv->dev_index,
				drv->encryption_state);
			drv->last_errno = -rc;
            READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			goto EXIT_LABEL;
		} /* if */
	} /* if */
    READ_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

EXIT_LABEL:

	*drvp = drv;
	return rc;
} /* lin_tape_drive_write_check() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_write_validate_parms                                         *
*                                                                              *
*  PURPOSE: initial check to see if write may proceed                          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: number of bytes to write                                     *
*    per_xfer_len (in/out): number of bytes for each transfer                  *
*    xfer_cnt (in/out) bytes to write based on block size                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_write_validate_parms(struct tape_descriptor* drv,
	size_t xfer_length, size_t* per_xfer_len, size_t* xfer_cnt)
{
	int rc = STATUS_SUCCESS;
	int def_xfer = LinTapeDefaultTransferLength;

	if(drv->blk_size) {
		/* fixed block mode */
		/* validate parameters:
		* blk_size must be less than max transfer length
		* xfer_length must be at least one block size
		* xfer_length must be multiple of block size
		* no multiple odd block transfer
		*/
		if(drv->blk_size > drv->max_xfer_len ||
			drv->blk_size > drv->max_blk_size) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		if(xfer_length < drv->blk_size) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		if(xfer_length % drv->blk_size) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		if((drv->blk_size & (ulong)(0x01)) &&
			xfer_length != drv->blk_size) {
			/* odd size blocks limited to one transfer */
			rc = -EINVAL;
			goto EXIT_LABEL;
		} /* if */

		/* setup the length per transfer and transferred length */
		if(xfer_length <= def_xfer) {
			*per_xfer_len = (size_t)(xfer_length / drv->blk_size);
			*xfer_cnt = xfer_length;
		} else if(drv->blk_size <= def_xfer) {
			*per_xfer_len = (size_t)(def_xfer / drv->blk_size);
			*xfer_cnt = drv->blk_size * (*per_xfer_len);
		} else {
			*per_xfer_len = 1;
			*xfer_cnt = drv->blk_size;
		} /* else */

	} else {
		/* variable block mode
		*
		* not allowing variable block transfer greater than the
		* max transfer length because artificial auto-blocking
		* gives applications the wrong impression of the physical
		* structure of the data on the tape
		*/
		if(xfer_length > drv->max_xfer_len ||
			xfer_length > drv->max_blk_size) {
			rc = -EINVAL;
			goto EXIT_LABEL;
		} else {
			*per_xfer_len = xfer_length;
			*xfer_cnt = xfer_length;
		} /* else */
	} /* else */

EXIT_LABEL:

	return rc;

} /* lin_tape_write_validate_parms() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_write_check_eom                                              *
*                                                                              *
*  PURPOSE: fail write if tape has reached end of media warning                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_write_check_eom(struct tape_descriptor* drv)
{
	int rc = STATUS_SUCCESS;

	/* check if we hit early warning zone on the last command */
	if(drv->deferred_eom == 1 && !drv->eot_warn) {
		drv->deferred_eom++;
		rc = -ENOSPC;
		goto EXIT_LABEL;
	} /* if */

	/* if trailer label is not turned on then we fail the command */
	if(!drv->trailer_labels && drv->deferred_eom) {
		rc = -ENOSPC;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;
} /* lin_tape_write_check_eom() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_direct_or_buffered                                           *
*                                                                              *
*  PURPOSE: determine whether to do direct IO or buffered IO                   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    user_location: user data area to map, if direct IO is used                *
*    xfer_cnt: the size of the data area to map or allocate                    *
*    rw: zero if memory will only be read, non-zero if it may be written to    *
*    pg_mapd (output): True if buffer was mapped in this function              *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_direct_or_buffered(struct tape_descriptor* drv,
	const char* user_location, size_t xfer_cnt, char rw,
	boolean verify_read, boolean* pg_mapd)
{
	int rc = -EIO;

	/* if support_dio is enabled, try direct I/O */
	if(support_dio) {
		rc = ((unsigned long)user_location &
			queue_dma_alignment(drv->dev_obj->request_queue));
		if(rc == STATUS_SUCCESS) {
			rc = lin_tape_get_user_pages(drv, user_location,
				xfer_cnt, rw);
		} /* if */
		if(rc == STATUS_SUCCESS) drv->direct_io = *pg_mapd = True;
		else {
			if(drv->direct_io && drv->sg_list) {
				kfree(drv->sg_list);
				drv->sg_list = NULL;
			} /* if */
			drv->direct_io = False;
		}  /* else */
	} /* if */
	/* if we cannot do direct io, allocate kernel buffer */
	if(rc != STATUS_SUCCESS && !drv->sg_list) {
		if(drv->rw_buf_size < xfer_cnt || !drv->rw_buf) {
			if(verify_read) {
				printk("lin_tape: IBMtape%d rw buffer "\
					"corrupted on read, return EFAULT.\n",
					drv->dev_index);
				printk("lin_tape: rw_buf %p, buf size %lu, "\
					"xfer_cnt %lu\n", drv->rw_buf,
					drv->rw_buf_size, (ulong)xfer_cnt);
				rc = -EFAULT;
			} else if(xfer_cnt <= LinTapeSingleRWBufferSize) {
				rc = tape_rw_buffer(drv, (ulong)xfer_cnt);
			} else if(drv->rw_buf && drv->rw_buf_order >= 0
				&& drv->rw_buf_size > 0) {
				/* free the rw buffer; use scatter gather */
				free_pages((ulong)drv->rw_buf,
					drv->rw_buf_order);

				drv->rw_buf_size = 0;
				drv->rw_buf_order = -1;
				drv->rw_buf = NULL;
			} /* else if */
		} else rc = STATUS_SUCCESS;
	} /* if */
	return rc;
} /* lin_tape_direct_or_buffered() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_rw_buffer                                                        *
*                                                                              *
*  PURPOSE: allocate buffer for read/write                                     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: size to allocate                                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_rw_buffer(struct tape_descriptor *drv, size_t xfer_length)
{
	ulong order;
	ulong buf_size;
	int rc = STATUS_SUCCESS;

	DbgPrint(("tape_rw_buffer entry\n"));

	if(drv->rw_buf && (drv->rw_buf_size == 0 || drv->rw_buf_order == -1)) {
		printk("lin_tape: IBMtape%d corrupt buffer\n", drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(drv->rw_buf) free_pages((ulong)drv->rw_buf, drv->rw_buf_order);

	drv->rw_buf = NULL;

	/* determine the order needed for transferlength
		eg: PAGE_SIZE * 2^order > = xfer_length */
	order = get_mem_ord(xfer_length, &buf_size);

	drv->rw_buf = (unchar*)__get_free_pages(GFP_KERNEL, order);
	drv->rw_buf_size = buf_size;
	drv->rw_buf_order = order;

	if(drv->rw_buf == NULL) {
		drv->rw_buf_size = 0;
		drv->rw_buf_order = -1;
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("tape_rw_buffer exit: %d\n", rc));
	return rc;
} /* tape_rw_buffer() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_copy_write_buffer                                                *
*                                                                              *
*  PURPOSE: copy the data from user's buffer to driver's buffer                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    user_location: data area in user space from which data should be copied   *
*    buf_length: number of bytes to copy                                       *
*    fpfo: flag set True if request comes from pfo                             *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_copy_write_buffer(struct tape_descriptor* drv,
	const char* user_location, size_t buf_length, boolean fpfo)
{
	int rc = STATUS_SUCCESS;
	const char* temp_location = user_location;
	int i = 0;
	ulong remain = 0;
	size_t copy_size = 0;

	if(drv->sg_list) {
		i = 0;
		remain = buf_length;
		drv->sg_list->use_sg = 0;
		while(remain > 0 && i < drv->supported_sg_tablesize) {
			copy_size = (drv->sg_list->gather[i].length < remain) ?
				drv->sg_list->gather[i].length : remain;

			memcpy(&drv->sg_list->sgl_for_io[i],
				&drv->sg_list->gather[i],
				sizeof(struct scatterlist));

			drv->sg_list->sgl_for_io[i].length = copy_size;

			if(!fpfo && copy_from_user(sg_virt(&drv->sg_list->sgl_for_io[i]),
				temp_location, copy_size)) {
				printk("lin_tape: IBMtape%d scatter gather "\
					"copy_from_user failed on gather %d, "\
					"returned EFAULT.\n", drv->dev_index,
					i);
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			else if(fpfo) memcpy(sg_virt(&drv->sg_list->sgl_for_io[i]),
									temp_location, copy_size);
			temp_location += copy_size;
			remain -= copy_size;
			drv->sg_list->use_sg++;
			i++;
		} /* while */

		if(remain > 0) {
			printk("IBMtape%d write: remain %ld\n", drv->dev_index,
				remain);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} else if(drv->rw_buf) {
		if(!fpfo && copy_from_user(drv->rw_buf, user_location, buf_length)) {
			printk("lin_tape%d copy_from_user failed.\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
		else if(fpfo) memcpy(drv->rw_buf, user_location, buf_length);
	} else {
			printk("IBMtape%d corrupted rw buffer or sg list\n",
				drv->dev_index);
			rc = -EFAULT;
			goto EXIT_LABEL;
	} /* else */

EXIT_LABEL:

	if(rc != STATUS_SUCCESS)
		printk("lin_tape: tape_copy_write_buffer returning %d\n", rc);
	return rc;
} /* tape_copy_write_buffer() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_perform_write                                                *
*                                                                              *
*  PURPOSE: send a single write request to the SCSI bus                        *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drvp: drive on which to perform the write                                 *
*    per_xfer_len (in/out): length of write for each transfer                  *
*    xfer_cnt (in/out): count of the bytes written                             *
*    xfer_offset (in/out): offset within user_buffer from which to take data   *
*    user_location: the data buffer to write                                   *
*    xfer_length: length of data to write                                      *
*    done: (out) terminate execution                                           *
*                                                                              *
*  RETURNS: number of bytes written or error code                              *
*                                                                              *
*******************************************************************************/

int lin_tape_perform_write(struct tape_descriptor** drvp, size_t* per_xfer_len,
	size_t* xfer_cnt, ulong* xfer_offset, const char* user_location,
	size_t xfer_length, unchar* done)
{
	int rc = STATUS_SUCCESS, timeout = 0;
	unchar ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;

	if(!drv->cdb) {
		printk("lin_tape_perform_write: IBMtape%d failed to "\
			"allocate cdb space\n", drv->dev_index);
		if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len,
				TRC_WT_NO_CDB);
		else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_WT_NO_CDB, 0, NULL);
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */
	drv->wakeup_info = 0;
	memset((void*)drv->cdb, '\0', MAX_CDB_SIZE);

	(drv->cdb)->cdb_read_write.op_code = WRITE_6;
	(drv->cdb)->cdb_read_write.fixed = drv->blk_size ? True : False;
	(drv->cdb)->cdb_read_write.xfer_length[0]
		= (unchar)(*per_xfer_len >> 16 & 0xFF);
	(drv->cdb)->cdb_read_write.xfer_length[1]
		= (unchar)(*per_xfer_len >> 8 & 0xFF);
	(drv->cdb)->cdb_read_write.xfer_length[2]
		= (unchar)(*per_xfer_len & 0xFF);

	drv->last_op = WRITE_6;
	drv->cdb_len = CDB_GENERIC_6;
	drv->wfm_on_close_safe = True;

	timeout = compute_timeout(drv->drv_id, WriteTimeOut);
	rc = tape_send_scsi_cmd(&drv, timeout, DMA_TO_DEVICE, *xfer_cnt, 0);

	/* if it is direct I/O, unmap the user pages */
	if(drv->direct_io) lin_tape_release_user_pages(drv, False);

	if(rc == STATUS_SUCCESS) {
		rc = handle_successful_write(drv, per_xfer_len, xfer_cnt,
			xfer_offset, xfer_length, user_location, done);
	} else {
		DbgPrint(("lin_tape: lin_tape_perform_write unable to "\
			"complete the request: %d\n", rc));
		rc = handle_write_failure(drv, xfer_cnt, xfer_offset, done, rc);
	} /* else */

EXIT_LABEL:

	*drvp = drv;
	return rc;

} /* lin_tape_perform_write() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_successful_write                                               *
*                                                                              *
*  PURPOSE: update position and bookkeeping variables after successful write   *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: drive on which write was performed                                   *
*    per_xfer_len (in/out): length of write for each transfer                  *
*    xfer_cnt (in/out): count of the bytes written                             *
*    xfer_offset (in/out): offset within user_buffer from which to take data   *
*    xfer_length: length of data to write                                      *
*    user_location: the data buffer to write                                   *
*    done: (out) terminate execution                                           *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int handle_successful_write(struct tape_descriptor* drv, size_t* per_xfer_len,
	size_t* xfer_cnt, ulong* xfer_offset, size_t xfer_length,
	const char* user_location, unchar* done)
{
	int rc = STATUS_SUCCESS;
	int def_xfer = LinTapeDefaultTransferLength;

	if(drv->pos_valid) {
		if(drv->blk_size) drv->curr_pos += *per_xfer_len;
		else drv->curr_pos++;
	} /* if */

	*xfer_offset += *xfer_cnt;

	if(*xfer_offset >= xfer_length) {
		*done = 2;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->blk_size) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* if remaining greater than default xfer len */
	if(xfer_length - *xfer_offset > def_xfer) {
		if(drv->blk_size <= def_xfer) {
			*per_xfer_len = (size_t)(def_xfer/drv->blk_size);
			*xfer_cnt = drv->blk_size * (*per_xfer_len);
		} else {
			*per_xfer_len = 1;
			*xfer_cnt = drv->blk_size;
		} /* else */
	} else {
		/* remaining transfered in one write */
		*per_xfer_len = (size_t)((xfer_length - *xfer_offset)
			/drv->blk_size);
		*xfer_cnt = xfer_length - *xfer_offset;
	} /* else */

	/* if support_dio is enabled, try direct I/o */
	if(support_dio) {
		rc = ((unsigned long)user_location & queue_dma_alignment(
			drv->dev_obj->request_queue));
		if(rc == STATUS_SUCCESS) {
			rc = lin_tape_get_user_pages(drv, user_location
				+ *xfer_offset, *xfer_cnt, True);
		} /* if */

		if(rc != STATUS_SUCCESS) {
			if(drv->direct_io && drv->sg_list) {
				kfree(drv->sg_list);
				drv->sg_list = NULL;
			} /* if */
			drv->direct_io = False;
		} else {
			drv->direct_io = True;
		} /* else */
	} /* if */

	/* if we cannot do direct io, allocate kernel buffer */
	/* copy new set of data from user space */
	if(!drv->direct_io) {
		rc = tape_copy_write_buffer(drv,
			user_location + *xfer_offset, *xfer_cnt, False);
		if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	return rc;
} /* handle_successful_write() */

/*******************************************************************************
*                                                                              *
*  NAME: handle_write_failure                                                  *
*                                                                              *
*  PURPOSE: take care of bookkeeping when there is a failure on write          *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: drive on which write was performed                                   *
*    xfer_cnt: count of the bytes written                                      *
*    xfer_offset: offset within user_buffer from which to take data            *
*    done: (out) terminate execution                                           *
*    rc: default return code if this function does not overwrite it            *
*                                                                              *
*  RETURNS: number of bytes written or error code                              *
*                                                                              *
*******************************************************************************/

int handle_write_failure(struct tape_descriptor* drv, size_t* xfer_cnt,
	ulong* xfer_offset, unchar* done, int rc)
{
	unchar ttype = lin_tape_trace_type;
	struct pos_info* read_pos_buf = NULL;

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
		tape_log_msg(drv->tape_log,
			&drv->tape_log_len, TRC_WT_FAIL);

	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_full_trace(drv, TRC_WT_FAIL, 0, NULL);

	/* invalidate position valid flag */
	drv->pos_valid = False;

	if(!drv->eom || !drv->resid_valid) goto EXIT_LABEL;

	if(drv->sense_data->sense_key == VOLUME_OVERFLOW || drv->resid < 0) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	/* If we hit the eom check condition and all of the data requested
	was written then return success and report eom next time */
	if(drv->eot_warn) rc = -ENOSPC;

	if(drv->sense_data->sense_key == 0x00
		&& drv->sense_data->asc == 0x00
		&& drv->sense_data->ascq == 0x07)
		/* programmable early warning zone detected */
		rc = -ENOSPC;

	else if(!drv->deferred_eom && !drv->eom_blk_id) {
		read_pos_buf = kmalloc(sizeof(struct pos_info),
				GFP_KERNEL);
		if(!read_pos_buf) {
			printk("lin_tape: IBMtape%d allocate read_pos_buf "\
				"failed, return ENOMEM\n", drv->dev_index);
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */

		memset(read_pos_buf, '\0', sizeof(struct pos_info));

		tape_wfm(&drv, False, 0);
		rc = tape_get_pos(&drv, read_pos_buf, RP_SHORT_FORM, False);

		if(rc == STATUS_SUCCESS) {
			drv->deferred_eom = True;
			if(drv->eot_warn) rc = -ENOSPC;
			drv->eom_blk_id = read_pos_buf->first_blk_pos - 1;
		} else goto EXIT_LABEL;

	} else if(!drv->deferred_eom) drv->deferred_eom = True;

	if(drv->resid && *xfer_offset + *xfer_cnt - drv->resid < 0) rc = -EIO;
	else {
		if(!drv->eot_warn || rc != -ENOSPC)
			rc = *xfer_offset + *xfer_cnt - drv->resid;
		*done = 1;
	} /* else */

EXIT_LABEL:

	if(read_pos_buf) kfree(read_pos_buf);
	return rc;

} /* handle_write_failure() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_copy_read_buffer                                                 *
*                                                                              *
*  PURPOSE: copy the data from driver's buffer to user's buffer                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    user_location: data area in user space to which data should be copied     *
*    buf_length: number of bytes to copy                                       *
*    fpfo: True if request comes from pfo                                      *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int tape_copy_read_buffer(struct tape_descriptor* drv, char* user_location,
	size_t buf_length, boolean fpfo)
{
	int rc = STATUS_SUCCESS;
	int i = 0;
	ulong remain = 0;
	size_t copy_size = 0;
	char* temp_location = user_location;

	if(BUF_RW(drv)) {
		if(!fpfo && copy_to_user(user_location, drv->rw_buf, buf_length)) {
			printk("lin_tape: IBMtape%d rw buffer copy_to_user "\
				"failed, return EFAULT\n", drv->dev_index);

			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
		else if (fpfo) memcpy(user_location, drv->rw_buf, buf_length);
	} else if(SGL_RW(drv)) {
		i = 0;
		remain = buf_length;

		while(remain > 0 && i < drv->supported_sg_tablesize) {
			if(!drv->sg_list || !drv->sg_list->gather[i].length
				|| !sg_page(&drv->sg_list->gather[i])) {
				printk("lin_tape: IBMtape%d "\
					"lin_tape_copy_read_buffer corrupted "\
					"scatter gather, returned EFAULT.\n",
					drv->dev_index);
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */

			/* copying either scatter list size or the remaining */
			copy_size = (drv->sg_list->gather[i].length < remain) ?
				drv->sg_list->gather[i].length : remain;

			if(!fpfo && copy_to_user(temp_location,
					sg_virt(&drv->sg_list->gather[i]),
					copy_size)) {
				printk("lin_tape: IBMtape%d sg copy failed\n",
					drv->dev_index);
				rc = -EFAULT;
				goto EXIT_LABEL;
			} /* if */
			else if (fpfo) memcpy (temp_location,
								sg_virt(&drv->sg_list->gather[i]),
								copy_size);
			temp_location += copy_size;
			remain -= copy_size;
			drv->sg_list->use_sg--;
			i++;
		} /* while */

		/* if we still have some left over, that means we went over the
		* scatter gathers which could only mean corrupted scatter list
		* We could still have some unused scatter list when reading
		* less than expected
		*/
		if(remain > 0 || drv->sg_list->use_sg < 0) {
			printk("lin_tape: IBMtape%d tape_copy_read_buffer "\
				"data overflowed scatterlist\n",
				drv->dev_index);

			if(remain > 0)
				printk("lin_tape: IBMtape%d "\
					"tape_copy_read_buffer remain is "\
					"%ld, returned EFAULT\n",
					drv->dev_index, remain);
			if(drv->sg_list && drv->sg_list->use_sg < 0)
				printk("lin_tape: IBMtape%d "\
					"tape_copy_read_buffer use_sg is %d",
					drv->dev_index,
					drv->sg_list->use_sg);

			rc = -EFAULT;
			goto EXIT_LABEL;
		} /* if */
	} else {
		printk("lin_tape: IBMtape%d tape_copy_read_buffer fatal "\
			"error -- corrupted buffer on copy\n",
			drv->dev_index);

		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* else */

EXIT_LABEL:

	if(rc != STATUS_SUCCESS)
		printk("lin_tape: tape_copy_read_buffer returning %d\n", rc);
	return rc;
} /* tape_copy_read_buffer() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_build_sg                                                         *
*                                                                              *
*  PURPOSE: Build sg list for read/write, allocating or enlarging if necessary *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    xfer_length: the number of bytes for the new allocation                   *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/

int tape_build_sg(struct tape_descriptor *drv, size_t xfer_length)
{
	int rc = STATUS_SUCCESS;

	DbgPrint(("lin_tape: tape_build_sg entry\n"));
	if(!xfer_length) {
		printk("lin_tape: IBMtape%d tape_build_sg 0 transferlength\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(drv->sg_list && (!drv->sg_list->total_sgl ||
		!drv->sg_list->total_sg_size)) {
		printk("lin_tape: IBMtape%d tape_build_sg scatterlist is "\
			"corrupted: sg %p sgl %d size %d\n", drv->dev_index,
			drv->sg_list, drv->sg_list->total_sgl,
			drv->sg_list->total_sg_size);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(drv->sg_list && xfer_length <= drv->sg_list->total_sgl) {
		printk("lin_tape: IBMtape%d tape_build_sg driver logic error\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->sg_list)
		rc = tape_init_sg(drv, xfer_length);
	else {
		rc = tape_ext_sg(drv, xfer_length);

		if(rc != STATUS_SUCCESS) {
			printk("lin_tape: IBMtape%d tape_build_sg unable to "\
				"extend scatter list\n", drv->dev_index);

			rc = tape_del_sg(drv);
			if(rc != STATUS_SUCCESS) {
				printk("lin_tape: IBMtape%d tape_build_sg "\
					"unable to reset scatter list\n",
					drv->dev_index);
				rc = 1;
				goto EXIT_LABEL;
			} /* if */
			rc = tape_init_sg(drv, xfer_length);
		} /* if */
	} /* else */

	if(rc != STATUS_SUCCESS) {
		printk("lin_tape: IBMtape%d tape_build_sg failed requesting "\
			"buffer size %lu.\n", drv->dev_index,
			(ulong)xfer_length);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_build_sg exit: %d\n", rc));
	return rc;
} /* tape_build_sg() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_init_sg                                                          *
*                                                                              *
*  PURPOSE: Allocate sg list for read/write                                    *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    alloc_length: the number of bytes for the new allocation                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/

int tape_init_sg(struct tape_descriptor *drv, size_t alloc_length)
{
	int i = 0, rc = STATUS_SUCCESS;
	ulong size = 0;	/* allocation size */
	ulong order = 0; /*  2^order length */
	long remain = 0; /* remain to be allocated length */
	long cur_alloc_length = 0;

	DbgPrint(("lin_tape: tape_init_sg entry\n"));
	drv->sg_list = kmalloc(sizeof(struct tape_sgl),
		GFP_KERNEL | __GFP_HIGH);
	if(!drv->sg_list) {
		printk("lin_tape: IBMtape%d tape_init_sg alloc scatterlist\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	memset(drv->sg_list, '\0', sizeof(struct tape_sgl));

	i = 0;
	remain = alloc_length;
	order = get_mem_ord(alloc_length, &size);
	while(remain > 0 && i < drv->supported_sg_tablesize) {
		unchar *address = NULL;

		if(sg_page(&drv->sg_list->gather[i])) {
			printk("lin_tape: IBMtape%d tape_init_sg gather %d "\
				"memory leak.\n", drv->dev_index, i);
			break;
		} /* if */

		/* try to allocate length */
		if(remain > LinTapeSingleRWBufferSize) {
			cur_alloc_length = min((int)size,
				LinTapeSingleRWBufferSize);
			order = get_mem_ord(cur_alloc_length, &size);
		} /* if */

		address = (unchar *) __get_free_pages(GFP_KERNEL, order);

		if(address) {
			sg_set_page(&drv->sg_list->gather[i],
				virt_to_page(address), size,
				(unsigned long)address & ~PAGE_MASK);
			drv->sg_list->total_sg_size += size;
			drv->sg_list->total_sgl++;
			remain -= size;
			i++;

		} else {

			order--;
			size /= 2;

			if((alloc_length - drv->sg_list->total_sg_size) >
				((drv->supported_sg_tablesize -
				drv->sg_list->total_sgl) * size)) {
				printk("lin_tape: IBMtape%d tape_init_sg "\
					"requested size is greater than "\
					"allocation limit, total_sgl %d "\
					"totalsg_size  %u, gather size %lu\n",
					drv->dev_index, drv->sg_list->total_sgl,
					drv->sg_list->total_sg_size, size);
				break;
			}
		} /* else */
	} /* while */

	/* if allocation failed, then cleanup */
	if(remain > 0) {
		i = 0;
		while(i < drv->supported_sg_tablesize &&
			drv->sg_list->gather[i].length) {

			if(!sg_page(&drv->sg_list->gather[i])) {
				drv->sg_list->gather[i].length = 0;
			} else {
				/* determine order for every scatter gather */
				order = get_mem_ord(drv->sg_list->gather[i]
					.length, &size);

				free_pages((ulong)
					(sg_virt(&drv->sg_list->gather[i])),
					order);
				sg_set_page(&drv->sg_list->gather[i],
					NULL, 0, 0);
				drv->sg_list->total_sgl--;
				drv->sg_list->total_sg_size -= size;
				i++;
			} /* else */
		} /* while */

		if(drv->sg_list->total_sgl || drv->sg_list->total_sg_size)
			printk("lin_tape: IBMtape%d tape_init_sg warning: "\
					"possible memory leak, total_sgl %d, "\
					"total_sg_size %u\n", drv->dev_index,
				drv->sg_list->total_sgl,
				drv->sg_list->total_sg_size);

		kfree(drv->sg_list);
		drv->sg_list = NULL;
		drv->direct_io = False;

		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_init_sg exit: %d\n", rc));
	return rc;
} /* tape_init_sg() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_ext_sg                                                           *
*                                                                              *
*  PURPOSE: Extend sg list for read/write, assuming there is a scatterlist     *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the request                          *
*    alloc_length: the number of bytes for the new allocation                  *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/

int tape_ext_sg(struct tape_descriptor *drv, size_t alloc_length)
{
	int i = 0, rc = STATUS_SUCCESS;
	ulong order = 0; /* 2^order length */
	ulong size = 0;
	long remain = 0; /* remain to be allocated length */
	struct tape_sgl* sg = NULL;
	long cur_alloc_length = 0;
	unchar* address = NULL;

	DbgPrint(("lin_tape: tape_ext_sg entry\n"));

	if(!drv->sg_list) {
		printk("lin_tape: IBMtape%d tape_ext_sg missing scatterlist\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	if(!drv->sg_list->total_sg_size || !drv->sg_list->total_sgl) {
		printk("lin_tape: IBMtape%d tape_ext_sg logic error, "\
			"total_sg_size %u, total_sgl %d\n",
			drv->dev_index, drv->sg_list->total_sg_size,
			drv->sg_list->total_sgl);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	sg = drv->sg_list;

	if(alloc_length <= sg->total_sg_size) {
		printk("IBMtape%d tape_ext_sg logic error, alloc_length "\
			"%lu, TotalScattersize %d\n", drv->dev_index,
			(ulong)alloc_length, sg->total_sg_size);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	remain = alloc_length - sg->total_sg_size;

	if(sg->gather[sg->total_sgl].length ||
		sg_page(&sg->gather[sg->total_sgl])) {
		printk("lin_tape: IBMtape%d tape_ext_sg logic error beyond "\
			"the last recognized scatter gatther.\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	i = 0;
	while(i < drv->supported_sg_tablesize
		&& sg->gather[i].length
		&& sg_page(&sg->gather[i])) {
		size += sg->gather[i].length;
		i++;
	} /* while */

	if(size != sg->total_sg_size || i != sg->total_sgl) {
		printk("lin_tape: IBMtape%d tape_ext_sg corrupted "\
			"scatter list detected\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	order = get_mem_ord(remain, &size);

	/* extend the scatter gather list */
	while((remain > 0) && (i < drv->supported_sg_tablesize)) {
		address = NULL;

		if(sg_page(&sg->gather[i])) break;

		/* try to allocate length */
		if(remain > LinTapeSingleRWBufferSize) {
			cur_alloc_length = min((int)size,
				LinTapeSingleRWBufferSize);
			order = get_mem_ord(cur_alloc_length, &size);
		} /* if */

		address = (unchar*)__get_free_pages(GFP_KERNEL, order);

		if(address) {
			sg_set_page(&sg->gather[i], virt_to_page(address), size,
				(unsigned long)address & ~PAGE_MASK);
			sg->total_sg_size += size;
			sg->total_sgl++;
			remain -= size;
			i++;

		} else {
			order--;
			size /= 2;

			/* if the remaining size is greater than what
			we can allocate out of the remaining sg, then
			we won't even try */
			if(alloc_length - sg->total_sg_size > size *
				(drv->supported_sg_tablesize - sg->total_sgl)) {
				printk("lin_tape: IBMtape%d tape_ext_sg "\
					"requested size exceeds allocation "\
					"limit\n", drv->dev_index);
				break;
			} /* if */
		} /* else */
	}  /* while */

	/* if we failed to extend to the required length */
	if(remain > 0) {
		printk("lin_tape: IBMtape%d tape_ext_sg failed, %ld bytes "\
			"remain in request.\n", drv->dev_index, remain);

		rc = 1;
		goto EXIT_LABEL;
	} /* if */

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_ext_sg exit: %d\n", rc));
	return rc;
} /* tape_ext_sg() */

/*******************************************************************************
*                                                                              *
*  NAME: tape_del_sg                                                           *
*                                                                              *
*  PURPOSE: Deallocate scatter gather list for read/write                      *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the user pages to deallocate         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or 1 on error                                       *
*                                                                              *
*******************************************************************************/
int tape_del_sg(struct tape_descriptor *drv)
{
	int i = 0, rc = STATUS_SUCCESS;
	ulong order = 0;	/* 2^order length */
	ulong size = 0;

	DbgPrint(("lin_tape: tape_del_sg entry\n"));

	if(!drv->sg_list) {
		printk("lin_tape: IBMtape%d tape_del_sg sg_list is NULL\n",
			drv->dev_index);
		rc = 1;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < drv->sg_list->total_sgl; i++) {
		if(drv->sg_list->gather[i].length) {
			if(!sg_page(&drv->sg_list->gather[i]))
				drv->sg_list->gather[i].length = 0;

			/* determine the order for every scatter gather */
			order = get_mem_ord(drv->sg_list->gather[i].length,
				&size);

			free_pages((ulong)(sg_virt(&drv->sg_list->gather[i])),
				order);
			sg_set_page(&drv->sg_list->gather[i], NULL, 0, 0);
		} else {
			if(sg_page(&drv->sg_list->gather[i]))
				printk("lin_tape: IBMtape%d tape_del_sg "\
					"detected memory leak\n",
				drv->dev_index);
		} /* else */
	} /* for */

	kfree(drv->sg_list);
	drv->sg_list = NULL;
	drv->direct_io = False;

EXIT_LABEL:

	DbgPrint(("lin_tape: tape_del_sg exit: %d\n", rc));
	return rc;

} /* tape_del_sg() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_get_user_pages                                               *
*                                                                              *
*  PURPOSE: Allocate pages to an sg list                                       *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the user pages to allocate           *
*    user_location: user data to be mapped                                     *
*    xfer_count: number of bytes to transfer                                   *
*    rw: zero if memory will only be read, non-zero if it may be written to    *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_get_user_pages(struct tape_descriptor *drv,
	const char* user_location, size_t xfer_count, char rw)
{
	int rc = -EFAULT;
	int number_pages = 0; /* number of pages in user buffer */
	struct page **pages = NULL; /* list of page pointers retrieved */
	int mapped_pages = 0;
	int i = 0;
	ulong start_offset = 0;
	size_t remaining = 0;
	boolean pages_allocd = False;

	/* calculate how many user pages need to be mapped */
	start_offset = (unsigned long)user_location & ~PAGE_MASK;
	if(start_offset == 0) {
		number_pages = xfer_count >> PAGE_SHIFT;
		if(xfer_count % PAGE_SIZE) number_pages++;
	} else if(xfer_count >= PAGE_SIZE - start_offset) {
		remaining = xfer_count - (PAGE_SIZE - start_offset);
		number_pages = 1 + (remaining >> PAGE_SHIFT);
		if(remaining % PAGE_SIZE) number_pages++;
	} else number_pages = 1;

	if(number_pages > drv->supported_sg_tablesize) {
		rc = -EIO;
		goto EXIT_LABEL;
	} /* if */

	pages = kmalloc(drv->supported_sg_tablesize * sizeof(struct page*),
		GFP_KERNEL);
	if(!pages) {
		rc = -ENOMEM;
		goto EXIT_LABEL;
	} /* if */
	pages_allocd = True;

	memset(pages, '\0', drv->supported_sg_tablesize * sizeof(struct page*));

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,8,0)
	down_read(&current->mm->mmap_sem);
#else
	mmap_read_lock(current->mm);
#endif
#if LINUX_VERSION_CODE == KERNEL_VERSION(4,4,73) \
	|| (LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,92) && LINUX_VERSION_CODE < KERNEL_VERSION(4,4,103)) \
	|| defined(SLESKERNEL_4_4_103_6) \
	|| (LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,114) && !defined(SLESSP2)) \
	|| (LINUX_VERSION_CODE >= KERNEL_VERSION(5,3,0) && defined(SLESSP2)) \
	|| defined (SLESKERNEL_4_4_103_94)
	mapped_pages = get_user_pages((unsigned long)user_location, number_pages,
			                      rw, pages, NULL);

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
	mapped_pages = get_user_pages((ulong)user_location, number_pages,
			                      rw, 0, pages, NULL);
#else
	mapped_pages = get_user_pages(current, current->mm,
		(ulong)user_location, number_pages, rw, 0, pages, NULL);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,8,0)
	up_read(&current->mm->mmap_sem);
#else
	mmap_read_unlock(current->mm);
#endif

	if(mapped_pages < number_pages) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	/* update the scatter gather list */
	if(drv->sg_list && drv->sg_list->total_sg_size) {
		tape_del_sg(drv);
		drv->sg_list = NULL;
	} /* if */

	if(!drv->sg_list) {
		drv->sg_list = kmalloc(sizeof(struct tape_sgl), GFP_KERNEL);
		if(!drv->sg_list) {
			rc = -ENOMEM;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	memset(drv->sg_list, '\0', sizeof(struct tape_sgl));

	drv->sg_list->use_sg = mapped_pages;
	sg_set_page(&drv->sg_list->gather[0], pages[0],
		mapped_pages > 1
			? PAGE_SIZE - start_offset : xfer_count, start_offset);
	memcpy(&drv->sg_list->sgl_for_io[0], &drv->sg_list->gather[0],
			sizeof(struct scatterlist));

	remaining = xfer_count - drv->sg_list->gather[0].length;
	for(i = 1; i < mapped_pages; i++) {
		sg_set_page(&drv->sg_list->gather[i], pages[i],
			remaining < PAGE_SIZE ? remaining : PAGE_SIZE, 0);

		remaining -= drv->sg_list->gather[i].length;

		/* update sgl_for_io list */
		memcpy(&drv->sg_list->sgl_for_io[i],
			&drv->sg_list->gather[i],
			sizeof(struct scatterlist));

	} /* for */

	if(remaining > 0) rc = -EFAULT;
	else rc = STATUS_SUCCESS;

EXIT_LABEL:

	if(rc < STATUS_SUCCESS)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
		for(i = 0; i < mapped_pages; i++)
			put_page(pages[i]);
#else
		for(i = 0; i < mapped_pages; i++)
			page_cache_release(pages[i]);
#endif
	if(pages_allocd) kfree(pages);
	return rc;
} /* lin_tape_get_user_pages() */

/*******************************************************************************
*                                                                              *
*  NAME: lin_tape_release_user_pages                                           *
*                                                                              *
*  PURPOSE: Free pages in an sg list                                           *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: tape descriptor associated with the user pages to free               *
*    dirty: the page has been written to                                       *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

int lin_tape_release_user_pages(struct tape_descriptor *drv, char dirty)
{
	int i = 0, rc = 0;

	if(!drv->sg_list || !drv->sg_list->use_sg) {
		rc = -EFAULT;
		goto EXIT_LABEL;
	} /* if */

	for(i = 0; i < drv->sg_list->use_sg; i++) {
		if(sg_page(&drv->sg_list->gather[i])) {
			if(dirty && !PageReserved(sg_page(
				&drv->sg_list->gather[i]))) {
				SetPageDirty(sg_page(&drv->sg_list->gather[i]));
			} /* if */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
			put_page(sg_page(&drv->sg_list->gather[i]));
#else
			page_cache_release(sg_page(&drv->sg_list->gather[i]));
#endif
		} /* if */
	} /* for */

	tape_del_sg(drv);

EXIT_LABEL:

	return rc;
} /* lin_tape_release_user_pages() */

/*******************************************************************************
*                                                                              *
*  NAME: char_to_hex                                                           *
*                                                                              *
*  PURPOSE: convert character representation to hex representation             *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    x: the character to convert                                               *
*                                                                              *
*  RETURNS: the hex value                                                      *
*                                                                              *
*******************************************************************************/

char char_to_hex(char x)
{
	return (x >= '0' && x <= '9')
		? (x - '0')
		: ((x >= 'a' && x <= 'f') || (x >= 'A' && x <= 'F'))
		? ((x + 9) & 0x0F)
		: x;
} /* char_to_hex() */

/*******************************************************************************
*                                                                              *
*  NAME: is_hex                                                                *
*                                                                              *
*  PURPOSE: return whether the ASCII character is a valid hex digit            *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    x: the character to test                                                  *
*                                                                              *
*  RETURNS: True if hex, False if not                                          *
*                                                                              *
*******************************************************************************/

boolean is_hex(char x)
{
	return     (x >= '0' && x <= '9')
		|| (x >= 'A' && x <= 'F')
		|| (x >= 'a' && x <= 'f');
} /* is_hex() */

/*******************************************************************************
*                                                                              *
*  NAME: drv_gen                                                               *
*                                                                              *
*  PURPOSE: return the drive generation based on the product ID                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the tape descriptor for which to get the generation                  *
*                                                                              *
*  RETURNS: generation                                                         *
*                                                                              *
*******************************************************************************/

int drv_gen(struct tape_descriptor* drv)
{
	char* gen = NULL;
	if(!drv || !drv->dev_obj || !drv->dev_obj->model) return -1;

	gen = (char*)&drv->dev_obj->model[min(PROD_ID_LEN,
		(int)strlen(drv->dev_obj->model)) - 1];
	while((gen - 1) > drv->dev_obj->model && *(gen - 1) == ' ') gen--;
	if(gen - drv->dev_obj->model < strlen("J1A")) return -1;
	gen -= strlen("J1A");

	return drv->drv_id == IBM_3592 ?
			(!strncmp(gen, "J1A", strlen("J1A")) ? 1 :
			!strncmp(gen, "E05", strlen("E05")) ? 2 :
			!strncmp(gen, "E06", strlen("E06")) ? 3 :
			!strncmp(gen, "E07", strlen("E07")) ? 4 :
			!strncmp(gen, "E08", strlen("E08")) ? 5 :
			!strncmp(gen, "55F", strlen("55F")) ? 5 :
			!strncmp(gen, "55E", strlen("55E")) ? 5 :
			!strncmp(gen, "55G", strlen("55G")) ? 5 :
			!strncmp(gen, "60G", strlen("60G")) ? 6 :
			!strncmp(gen, "60F", strlen("60F")) ? 6 :
			!strncmp(gen, "60E", strlen("60E")) ? 6 :
			-1) :
		drv->drv_id == IBM_3580 ?  gen[2] - '0' :
		-1;

} /* drv_gen() */

/*******************************************************************************
*                                                                              *
*  NAME: set_drive_open                                                        *
*                                                                              *
*  PURPOSE: Set open flag in current drive and all of its paths                *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    drv: the drive for which to set the flag                                  *
*    v: the flag to set, True or False                                         *
*    lock_held: True if lock is already held, False if we need to acquire it   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void set_drive_open(struct tape_descriptor* drv, boolean v, boolean lock_held)
{
	struct tape_descriptor* orig = drv, * trav = drv;

	if(!lock_held) WRITE_LOCK(&tape_des_lock);

	do {
        WRITE_LOCK(&tape_devs[trav->dev_index]->drv_entry_lock);
		trav->drv_open = v;
        WRITE_UNLOCK(&tape_devs[trav->dev_index]->drv_entry_lock);
		trav = trav->next_path;
	} while(trav != orig);

	if(!lock_held) WRITE_UNLOCK(&tape_des_lock);

} /* set_drive_open() */

/*******************************************************************************
*                                                                              *
*  NAME: set_changer_open                                                      *
*                                                                              *
*  PURPOSE: Set open flag in current changer and all of its paths              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    chg: the changer for which to set the flag                                *
*    v: the flag to set, True or False                                         *
*    lock_held: True if lock is already held, False if we need to acquire it   *
*                                                                              *
*  RETURNS: void                                                               *
*                                                                              *
*******************************************************************************/

void set_changer_open(struct changer_descriptor* chg, boolean v,
	boolean lock_held)
{
	struct changer_descriptor* orig = chg, * trav = chg;

	if(!lock_held) WRITE_LOCK(&changer_des_lock);

	do {
		if(v) {
			trav->open_count++;
			trav->chg_open = v;
		} /* if */

		else {
			if(trav->open_count) trav->open_count--;
			if(!trav->open_count) trav->chg_open = v;
		} /* else */
		trav = trav->next_path;
	} while(trav != orig);

	if(!lock_held) WRITE_UNLOCK(&changer_des_lock);

} /* set_changer_open() */

/*******************************************************************************
*                                                                              *
*  NAME: printk_hex                                                            *
*                                                                              *
*  PURPOSE: debugging function: print the buffer pointed to by sz              *
*                                                                              *
*  PARAMETERS:                                                                 *
*                                                                              *
*    str: the buffer to print                                                  *
*    len: the length of the buffer in bytes                                    *
*    line: line number                                                         *
*                                                                              *
*  RETURNS: STATUS_SUCCESS or error code                                       *
*                                                                              *
*******************************************************************************/

void printk_hex(char* hex_buf, int len, int line)
{
	char* sz = NULL;
	int i = 0;

	printk("LINE %d, buffer: ", line);

	sz = kmalloc(len * 2 + 1, GFP_KERNEL);
	if(!sz) {
		printk("MEMALLOC FAILED\n");
		goto EXIT_LABEL;
	} /* if */
	memset(sz, '\0', len * 2 + 1);

	for(i = 0; i < len; i++)
		sprintf(&sz[strlen(sz)], "%02X", hex_buf[i] & 0xFF);

	printk("0x%s\n", sz);

EXIT_LABEL:

	if(sz) kfree(sz);
	return;
} /* printk_hex() */

int lin_tape_pass_thru_from_pfoID(struct tape_descriptor **drvp,
				unchar* cdb,
				unchar cmd_length,
				unchar* buffer,
				uint *buffer_length,
				uint data_direction,
				uint timeout,
				unchar* sense,
				unchar *sense_length,
				unchar *sense_avail,
				int *resid,
				int *errors)
{
	int rc = -EIO;
	char ttype = lin_tape_trace_type;
	struct tape_descriptor* drv = *drvp;
	boolean bsy = False;

	READ_LOCK(&tape_des_lock);
	if(!drv) {
        READ_UNLOCK(&tape_des_lock);
		printk("lin_tape: lin_tape_pfo: drive removed.\n");
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */
	READ_UNLOCK(&tape_des_lock);

	if(!cdb) {
		rc = -EINVAL;
		goto EXIT_LABEL;
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_setup_and_log_full(drv, drv->cur_path, TRC_SCSI_PATH,
			NULL, 0);

	WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	if(!drv->dev_obj || drv->ghost_device) {
		if(drv->alt_pathing) {
			WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			rc = tape_perform_failover_generic(&drv, 1800 * HZ);
			WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			if(rc == STATUS_COMPLETE) {
				rc = STATUS_SUCCESS;
			}else if(rc != STATUS_SUCCESS) {
				    rc = drv->result;
			}
		} /* if */
		if(!drv->alt_pathing || rc != STATUS_SUCCESS) {
			printk("lin_tape: lin_tape_pfo: drive "\
				"removed.\n");
			if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
				tape_log_msg(lin_tape_log, &lin_tape_log_len,
					TRC_TP_OUT_RANGE);
			else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
				tape_log_rc(drv, ENXIO);
			if (!drv->alt_pathing) rc = -ENXIO;
			WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	if (!drv) {
        WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
	}
	drv->busy |= F_IOCTL;
	bsy = True;
	WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);

	if(!drv->cdb || !drv->sense_data || !drv->tape_log) {
		rc = allocate_tdescriptor_buffers(drv);
		if(rc != STATUS_SUCCESS) {
			DbgPrint(("lin_tape_drive_ioctl: cannot allocate "\
				"buffer\n"));
			goto EXIT_LABEL;
		} /* if */
	} /* if */

	if(drv->trace && ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg(drv->tape_log, &drv->tape_log_len, TRC_PTCMD);
	else if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
			tape_log_full_trace(drv, TRC_PTCMD, 0, NULL);

	rc = tape_pass_thru(&drv, cdb, cmd_length,
					buffer, buffer_length,
					data_direction, timeout,
					sense, sense_length, sense_avail,
					resid, errors, True);

	EXIT_LABEL:

	if(drv->trace && ttype == LIN_TAPE_FULL_TRACE)
		tape_log_rc(drv, -rc);

	drv->last_errno = -rc;

	if(bsy && drv) {
		WRITE_LOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
		drv->busy &= (0xFF-F_IOCTL);
		WRITE_UNLOCK(&tape_devs[drv->dev_index]->drv_entry_lock);
	} /* if */

	*drvp = drv;
	return rc;

} /* lin_tape_pass_thru_from_pfoID() */

int lin_tape_changer_pass_thru_from_pfoID(struct changer_descriptor **chgp,
			unchar* cdb,
			unchar cmd_length,
			unchar* buffer,
			uint *buffer_length,
			uint data_direction,
			uint timeout,
			unchar* sense,
			unchar *sense_length,
			unchar *sense_avail,
			int *resid,
			int *errors	)
{
	int rc = -EIO;
	unchar ttype = lin_tape_trace_type;
	char call_description[40];

	memset(call_description, '\0', 40);

	DbgPrint(("lin_tape: lin_tape_changer_ioctl entry: cdb 0x%x\n",
		*cdb));

	READ_LOCK(&changer_des_lock);
	if(!(*chgp)) {
		printk("lin_tape: lin_tape_changer_ioctl: changer removed.\n");
		READ_UNLOCK(&changer_des_lock);
		rc = -ENXIO;
		goto EXIT_LABEL;
	} /* if */

	if(!(*chgp)->cdb || !(*chgp)->sense_data) {
		READ_UNLOCK(&changer_des_lock);
		rc = chg_alloc_cdb_and_sense((*chgp));
		if(rc != STATUS_SUCCESS) {
			DbgPrint(("lin_tape_changer_ioctl: cannot allocate "\
				"buffer\n"));
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	else READ_UNLOCK(&changer_des_lock);

	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	if((*chgp)->ghost_device) {
		if((*chgp)->alt_pathing) {
			WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
			rc = changer_failover_ops(&(*chgp), HZ * 1800, 0);
			if (!chgp || !(*chgp)){
				rc = -ENXIO;
				goto EXIT_LABEL;
			}
			if (chgp && *chgp) WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);;
		} /* if */

		if(!(*chgp)->alt_pathing || rc != STATUS_SUCCESS) {
			printk("lin_tape: lin_tape_changer_ioctl: changer "\
				"removed.\n");
			WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
			rc = -ENXIO;
			goto EXIT_LABEL;
		} /* if */
	} /* if */
	WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

	strcpy(call_description, "SIOC_PASS_THROUGH");

	rc = changer_wait_to_set_busy(&(*chgp),call_description,0);
	if(rc != STATUS_SUCCESS) goto EXIT_LABEL;
    (*chgp)->busy |= F_IOCTL;
	if(ttype == LIN_TAPE_FULL_TRACE)
		changer_setup_and_log_full((*chgp), (*chgp)->cur_path, TRC_SCSI_PATH,
			NULL, 0);

	(*chgp)->last_errno = STATUS_SUCCESS;
	(*chgp)->comp_code_valid = False;
	(*chgp)->comp_code = STATUS_SUCCESS;

	log_changer((*chgp), TRC_PTCMD, True);

	rc = changer_pass_thru(&(*chgp),
							cdb,
							cmd_length,
							buffer,
							buffer_length,
							data_direction,
							timeout,
							sense,
							sense_length,
							sense_avail,
							resid,
							errors,
							0,
							True);

EXIT_LABEL:

	if(rc != STATUS_SUCCESS) {
		if(ttype == LIN_TAPE_REGULAR_TRACE)
			tape_log_msg2((*chgp)->chg_log, &((*chgp)->chg_log_len),
				TRC_SMCIOC_ERR, -rc);
	} /* if */

	if(ttype == LIN_TAPE_FULL_TRACE) changer_log_rc((*chgp), -rc);

	(*chgp)->last_errno = -rc;

	WRITE_LOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);
	(*chgp)->busy &= (0xFF-F_IOCTL);
	WRITE_UNLOCK(&changer_devs[(*chgp)->dev_index]->entry_lock);

	DbgPrint(("lin_tape: lin_tape_changer_ioctl exit: %d\n", rc));
	return rc;
} /* lin_tape_changer_pass_thru_from_pfoID */


